Package: xfun
Type: Package
Title: Supporting Functions for Packages Maintained by 'Yihui Xie'
Version: 0.52
Authors@R: c(
  person("Yihui", "Xie", role = c("aut", "cre", "cph"), email = "xie@yihui.name", comment = c(ORCID = "0000-0003-0645-5666", URL = "https://yihui.org")),
  person("Wush", "Wu", role = "ctb"),
  person("Daijiang", "Li", role = "ctb"),
  person("Xianying", "Tan", role = "ctb"),
  person("Salim", "Brüggemann", role = "ctb", email = "salim-b@pm.me", comment = c(ORCID = "0000-0002-5329-5987")),
  person("Christophe", "Dervieux", role = "ctb"),
  person()
  )
Description: Miscellaneous functions commonly used in other packages maintained by 'Yihui Xie'.
Depends: R (>= 3.2.0)
Imports: grDevices, stats, tools
Suggests: testit, parallel, codetools, methods, rstudioapi, tinytex (>=
        0.30), mime, litedown (>= 0.4), commonmark, knitr (>= 1.50),
        remotes, pak, curl, xml2, jsonlite, magick, yaml, qs
License: MIT + file LICENSE
URL: https://github.com/yihui/xfun
BugReports: https://github.com/yihui/xfun/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
VignetteBuilder: litedown
NeedsCompilation: yes
Packaged: 2025-04-02 19:53:53 UTC; runner
Author: Yihui Xie [aut, cre, cph] (<https://orcid.org/0000-0003-0645-5666>,
    https://yihui.org),
  Wush Wu [ctb],
  Daijiang Li [ctb],
  Xianying Tan [ctb],
  Salim Brüggemann [ctb] (<https://orcid.org/0000-0002-5329-5987>),
  Christophe Dervieux [ctb]
Maintainer: Yihui Xie <xie@yihui.name>
Repository: CRAN
Date/Publication: 2025-04-02 20:40:02 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4; 2026-07-17 18:00:41 UTC; unix

Package: xfun
Type: Package
Title: Supporting Functions for Packages Maintained by 'Yihui Xie'
Version: 0.58
Authors@R: c(
  person("Yihui", "Xie", role = c("aut", "cre", "cph"), email = "xie@yihui.name", comment = c(ORCID = "0000-0003-0645-5666", URL = "https://yihui.org")),
  person("Wush", "Wu", role = "ctb"),
  person("Daijiang", "Li", role = "ctb"),
  person("Xianying", "Tan", role = "ctb"),
  person("Salim", "Brüggemann", role = "ctb", email = "salim-b@pm.me", comment = c(ORCID = "0000-0002-5329-5987")),
  person("Christophe", "Dervieux", role = "ctb"),
  person()
  )
Description: Miscellaneous functions commonly used in other packages maintained by 'Yihui Xie'.
Depends: R (>= 3.2.0)
Imports: grDevices, stats, tools
Suggests: testit, parallel, codetools, methods, rstudioapi, tinytex (>=
        0.30), mime, litedown (>= 0.6), commonmark, knitr (>= 1.50),
        remotes, pak, curl, xml2, jsonlite, magick, yaml, data.table,
        qs2
License: MIT + file LICENSE
URL: https://github.com/yihui/xfun
BugReports: https://github.com/yihui/xfun/issues
Encoding: UTF-8
VignetteBuilder: litedown
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-31 16:49:17 UTC; yihui
Author: Yihui Xie [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-0645-5666>, URL: https://yihui.org),
  Wush Wu [ctb],
  Daijiang Li [ctb],
  Xianying Tan [ctb],
  Salim Brüggemann [ctb] (ORCID: <https://orcid.org/0000-0002-5329-5987>),
  Christophe Dervieux [ctb]
Maintainer: Yihui Xie <xie@yihui.name>
Repository: CRAN
Date/Publication: 2026-06-01 05:10:03 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-06-01 05:37:19 UTC; unix
Archs: xfun.so.dSYM

Package: xfun
Type: Package
Title: Supporting Functions for Packages Maintained by 'Yihui Xie'
Version: 0.36
Authors@R: c(
  person("Yihui", "Xie", role = c("aut", "cre", "cph"), email = "xie@yihui.name", comment = c(ORCID = "0000-0003-0645-5666")),
  person("Wush", "Wu", role = "ctb"),
  person("Daijiang", "Li", role = "ctb"),
  person("Xianying", "Tan", role = "ctb"),
  person("Salim", "Brüggemann", role = "ctb", email = "salim-b@pm.me", comment = c(ORCID = "0000-0002-5329-5987")),
  person("Christophe", "Dervieux", role = "ctb"),
  person()
  )
Description: Miscellaneous functions commonly used in other packages maintained by 'Yihui Xie'.
Imports: stats, tools
Suggests: testit, parallel, codetools, rstudioapi, tinytex (>= 0.30),
        mime, markdown, knitr, htmltools, remotes, pak, rhub, renv,
        curl, jsonlite, magick, rmarkdown
License: MIT + file LICENSE
URL: https://github.com/yihui/xfun
BugReports: https://github.com/yihui/xfun/issues
Encoding: UTF-8
RoxygenNote: 7.2.3
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2022-12-21 15:05:22 UTC; yihui
Author: Yihui Xie [aut, cre, cph] (<https://orcid.org/0000-0003-0645-5666>),
  Wush Wu [ctb],
  Daijiang Li [ctb],
  Xianying Tan [ctb],
  Salim Brüggemann [ctb] (<https://orcid.org/0000-0002-5329-5987>),
  Christophe Dervieux [ctb]
Maintainer: Yihui Xie <xie@yihui.name>
Repository: CRAN
Date/Publication: 2022-12-21 15:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",xfun_strict_list)
S3method(format,xfun_record_results)
S3method(print,xfun_md_viewable)
S3method(print,xfun_raw_string)
S3method(print,xfun_record_results)
S3method(print,xfun_rename_seq)
S3method(print,xfun_strict_list)
S3method(record_print,default)
S3method(record_print,record_asis)
S3method(record_print,xfun_raw_string)
export(Rcmd)
export(Rscript)
export(Rscript_call)
export(alnum_id)
export(append_unique)
export(append_utf8)
export(as_strict_list)
export(attr2)
export(base64_decode)
export(base64_encode)
export(base64_uri)
export(base_pkgs)
export(bg_process)
export(broken_packages)
export(browser_print)
export(bump_version)
export(cache_exec)
export(cache_rds)
export(check_old_package)
export(check_package_name)
export(compare_Rcheck)
export(crandalf_check)
export(crandalf_results)
export(csv_options)
export(decimal_dot)
export(del_empty_dir)
export(dir_create)
export(dir_exists)
export(divide_chunk)
export(do_once)
export(download_cache)
export(download_file)
export(embed_dir)
export(embed_file)
export(embed_files)
export(env_option)
export(existing_files)
export(exit_call)
export(fenced_block)
export(fenced_div)
export(file_exists)
export(file_ext)
export(file_rename)
export(file_string)
export(find_globals)
export(find_locals)
export(format_bytes)
export(from_root)
export(github_api)
export(github_releases)
export(grep_sub)
export(gsub_dir)
export(gsub_ext)
export(gsub_file)
export(gsub_files)
export(html_escape)
export(html_tag)
export(html_value)
export(html_view)
export(in_dir)
export(install_dir)
export(install_github)
export(is_CRAN_incoming)
export(is_R_CMD_check)
export(is_abs_path)
export(is_arm64)
export(is_ascii)
export(is_blank)
export(is_linux)
export(is_macos)
export(is_rel_path)
export(is_sub_path)
export(is_unix)
export(is_web_path)
export(is_windows)
export(join_words)
export(js)
export(json_vector)
export(lazy_load)
export(lazy_save)
export(loadable)
export(magic_path)
export(make_fence)
export(mark_dirs)
export(md5)
export(md_table)
export(mime_type)
export(msg_cat)
export(n2w)
export(native_encode)
export(new_app)
export(new_record)
export(news2md)
export(normalize_path)
export(numbers_to_words)
export(optipng)
export(parse_only)
export(pkg_attach)
export(pkg_attach2)
export(pkg_available)
export(pkg_bib)
export(pkg_load)
export(pkg_load2)
export(proc_kill)
export(process_file)
export(proj_root)
export(prose_index)
export(protect_math)
export(rand_unit)
export(random_port)
export(raw_string)
export(read_all)
export(read_bin)
export(read_utf8)
export(record)
export(record_print)
export(relative_path)
export(rename_seq)
export(rest_api)
export(rest_api_raw)
export(retry)
export(rev_check)
export(root_rules)
export(rstudio_type)
export(same_path)
export(sans_ext)
export(serve_dir)
export(session_info)
export(set_envvar)
export(shrink_images)
export(sort_file)
export(split_lines)
export(split_source)
export(stop_app)
export(str_wrap)
export(strict_list)
export(strip_html)
export(submit_cran)
export(system3)
export(tab_content)
export(tabset)
export(taml_file)
export(taml_load)
export(taml_save)
export(tinify)
export(tinify_dir)
export(tojson)
export(tree)
export(try_error)
export(try_silent)
export(upload_ftp)
export(upload_imgur)
export(upload_win_builder)
export(url_accessible)
export(url_destination)
export(url_filename)
export(valid_syntax)
export(with_ext)
export(write_utf8)
export(yaml_body)
export(yaml_load)
import(grDevices)
import(stats)
import(utils)
useDynLib(xfun, .registration = TRUE)

# Generated by roxygen2: do not edit by hand

S3method("$",xfun_strict_list)
S3method(format,xfun_record_results)
S3method(print,xfun_md_viewable)
S3method(print,xfun_raw_string)
S3method(print,xfun_record_results)
S3method(print,xfun_rename_seq)
S3method(print,xfun_strict_list)
S3method(record_print,default)
S3method(record_print,record_asis)
S3method(record_print,xfun_raw_string)
export(Rcmd)
export(Rscript)
export(Rscript_call)
export(alnum_id)
export(append_unique)
export(append_utf8)
export(as_strict_list)
export(attr)
export(attr2)
export(base64_decode)
export(base64_encode)
export(base64_uri)
export(base_pkgs)
export(bg_process)
export(broken_packages)
export(bump_version)
export(cache_exec)
export(cache_rds)
export(check_old_package)
export(check_package_name)
export(compare_Rcheck)
export(crandalf_check)
export(crandalf_results)
export(csv_options)
export(decimal_dot)
export(del_empty_dir)
export(dir_create)
export(dir_exists)
export(divide_chunk)
export(do_once)
export(download_cache)
export(download_file)
export(embed_dir)
export(embed_file)
export(embed_files)
export(env_option)
export(existing_files)
export(exit_call)
export(fenced_block)
export(fenced_div)
export(file_exists)
export(file_ext)
export(file_rename)
export(file_string)
export(find_globals)
export(find_locals)
export(format_bytes)
export(from_root)
export(github_api)
export(github_releases)
export(grep_sub)
export(gsub_dir)
export(gsub_ext)
export(gsub_file)
export(gsub_files)
export(html_escape)
export(html_tag)
export(html_value)
export(html_view)
export(in_dir)
export(install_dir)
export(install_github)
export(is_CRAN_incoming)
export(is_R_CMD_check)
export(is_abs_path)
export(is_arm64)
export(is_ascii)
export(is_blank)
export(is_linux)
export(is_macos)
export(is_rel_path)
export(is_sub_path)
export(is_unix)
export(is_web_path)
export(is_windows)
export(join_words)
export(json_vector)
export(lazy_load)
export(lazy_save)
export(loadable)
export(magic_path)
export(make_fence)
export(mark_dirs)
export(md5)
export(md_table)
export(mime_type)
export(msg_cat)
export(n2w)
export(native_encode)
export(new_app)
export(new_record)
export(news2md)
export(normalize_path)
export(numbers_to_words)
export(optipng)
export(parse_only)
export(pkg_attach)
export(pkg_attach2)
export(pkg_available)
export(pkg_bib)
export(pkg_load)
export(pkg_load2)
export(proc_kill)
export(process_file)
export(proj_root)
export(prose_index)
export(protect_math)
export(rand_unit)
export(raw_string)
export(read_all)
export(read_bin)
export(read_utf8)
export(record)
export(record_print)
export(relative_path)
export(rename_seq)
export(rest_api)
export(rest_api_raw)
export(retry)
export(rev_check)
export(root_rules)
export(rstudio_type)
export(same_path)
export(sans_ext)
export(session_info)
export(set_envvar)
export(shrink_images)
export(sort_file)
export(split_lines)
export(split_source)
export(str_wrap)
export(strict_list)
export(strip_html)
export(submit_cran)
export(system3)
export(tabset)
export(taml_file)
export(taml_load)
export(taml_save)
export(tinify)
export(tinify_dir)
export(tojson)
export(tree)
export(try_error)
export(try_silent)
export(upload_ftp)
export(upload_imgur)
export(upload_win_builder)
export(url_accessible)
export(url_destination)
export(url_filename)
export(valid_syntax)
export(with_ext)
export(write_utf8)
export(yaml_body)
export(yaml_load)
import(grDevices)
import(stats)
import(utils)
useDynLib(xfun, .registration = TRUE)

# xfun

<!-- badges: start -->
[![R-CMD-check](https://github.com/yihui/xfun/workflows/R-CMD-check/badge.svg)](https://github.com/yihui/xfun/actions)
[![Coverage status](https://codecov.io/gh/yihui/xfun/branch/master/graph/badge.svg)](https://codecov.io/github/yihui/xfun?branch=master)
[![CRAN release](https://www.r-pkg.org/badges/version/xfun)](https://cran.r-project.org/package=xfun)
<!-- badges: end -->

This package contains several utility functions that I frequently use in other packages, and also miscellaneous functions that I use by myself from time to time. For more information, see https://yihui.org/xfun/.

library(testit)

assert('attr2() is strict', {
  z = structure(list(a = 1), foo = 2)
  (attr2(z, 'foo') %==% 2)
  (attr2(z, 'f') %==% NULL)
})

assert('in_dir() preserves the working directory', {
  owd = getwd()
  in_dir('.', setwd(tempdir()))
  (same_path(owd, getwd()))
})

library(testit)

assert('csv_options() parses chunk options to a list', {
  (csv_options('a-s-d,b=TRUE,c=def') %==% alist(label = 'a-s-d', b = TRUE, c = def))
  (csv_options('a,opt=c(1,3,5)') %==% alist(label = 'a', opt = c(1, 3, 5)))
  (csv_options('label="xx",opt=zz') %==% alist(label = 'xx', opt = zz))
  (csv_options('label=foo') %==% alist(label = 'foo'))
  (csv_options('a,b=2,c="qwer",asdf="efg"') %==%
      alist(label = 'a', b = 2, c = 'qwer', asdf = 'efg'))
  (csv_options('2a') %==% alist(label = '2a'))
  (csv_options('abc-function,fig.path="foo/bar-"') %==%
      alist(label = 'abc-function', fig.path = "foo/bar-"))
  (has_error(csv_options('a,b')))
  (has_error(csv_options('a,b,c=qwer')))
})

library(testit)

assert('native_encode() warns against characters that cannot be represented in native encoding', {
  latin_str = 'fa\u00e7ile'  # Latin-1
  cn_str = '\u4e2d\u6587\u5b57\u7b26'  # UTF-8 Chinese
  # when the locale is not UTF-8, the above strings cannot be converted to native encoding
  (isTRUE(l10n_info()[['UTF-8']]) ||
      has_warning({native_encode(cn_str); native_encode(latin_str)}))

  gb2312_raw = as.raw(c(0xd6, 0xd0, 0xce, 0xc4, 0xd7, 0xd6, 0xb7, 0xfb))
  is_gb2312_native = identical(gb2312_raw, charToRaw(enc2native(cn_str)))
  no_need_test = !(is_windows() && is_gb2312_native)
  cn_str_native = native_encode(cn_str)
  (no_need_test || Encoding(cn_str_native) %==% 'unknown')
  (no_need_test || charToRaw(cn_str_native) %==% gb2312_raw)
})

assert('is_ascii() can identify ascii strings', {
  ascii_str = c('aaa', 'bbb', 'ccc')
  latin_str = 'fa\u00e7ile'
  cn_str = '\u4e2d\u6587\u5b57\u7b26'
  mixed_str = c(ascii_str, latin_str)
  (is_ascii(ascii_str) %==% c(TRUE, TRUE, TRUE))
  (!is_ascii(latin_str))
  (!is_ascii(cn_str))
  (is_ascii(mixed_str) %==% c(TRUE, TRUE, TRUE, FALSE))
  (is_ascii(c(NA_character_, 'a')) %==% c(NA, TRUE))
})

library(testit)

assert("Rscript() correctly calls Rscript with system2()", {
  (Rscript(c("-e", "1+1"), stdout = TRUE) %==% "[1] 2")
})

assert("Rscript_call() correctly calls Rscript()", {
  (Rscript_call(function() 1+1) %==% 2)
  loaded = Rscript_call(
    function() loadedNamespaces(), options = "--default-packages=xfun"
  )
  ("xfun" %in% loaded)
})

library(xfun)

library(xfun)
(z = strict_list(aaa = "I am aaa", b = 1:5))
z$a  # NULL (strict matching)
z$aaa  # I am aaa
z$b
z$c = "you can create a new element"

z2 = unclass(z)  # a normal list
z2$a  # partial matching

z3 = as_strict_list(z2)  # a strict list again
z3$a  # NULL (strict matching) again!

library(xfun)
raw_string(head(LETTERS))
(x = c("a \"b\"", "hello\tworld!"))
raw_string(x)  # this is more likely to be what you want to see

f = system.file("LICENSE", package = "xfun")
xfun::file_string(f)
as.character(xfun::file_string(f))  # essentially a character string

f = system.file("LICENSE", package = "xfun")
xfun::base64_uri(f)

xfun::grep_sub('a([b]+)c', 'a\\U\\1c', c('abc', 'abbbc', 'addc', '123'), perl = TRUE)

library(xfun)
f = tempfile()
writeLines(c("hello", "world"), f)
gsub_file(f, "world", "woRld", fixed = TRUE)
file_string(f)

process_file(f, function(x) {
  rep(x, 3)  # repeat the content 3 times
})
file_string(f)

library(xfun)
p = c("abc.doc", "def123.tex", "path/to/foo.Rmd")
file_ext(p)
sans_ext(p)
with_ext(p, ".txt")
with_ext(p, c(".ppt", ".sty", ".Rnw"))
with_ext(p, "html")

xfun::is_macos()
xfun::is_unix()
xfun::is_linux()
xfun::is_windows()

library(testit)
library(parallel)
library(tinytex)
library(mime)

xfun::pkg_attach(c('testit', 'parallel', 'tinytex', 'mime'))

if (!requireNamespace('tinytex')) install.packages('tinytex')
library(tinytex)

xfun::pkg_attach2('tinytex')

n2w(0, cap = TRUE)
n2w(seq(0, 121, 11), and = TRUE)
n2w(1e+06)
n2w(1e+11 + 12345678)
n2w(-987654321)
n2w(1e+15 - 1)

xfun::session_info(c('xfun', 'litedown', 'tinytex'), dependencies = FALSE)

library(testit)

assert('find_globals() identifies global variables', {
  env = list2env(
    list(qwer = 1, g = 2, y = 3, d = 4, z = 5), parent = emptyenv()
  )
  # nothing from outside environment
  (find_globals('x=1', env) %==% character(0))
  # qwer must be created outside somewhere
  (find_globals('a=1; b=a; d=qwer', env) %==% 'qwer')
  (find_globals('a=function(){f=2;g}', env) %==% 'g')
  # y was assigned locally in z, but there is another y outside from nowhere
  (find_globals('z=function(){y=1};y', env) %==% 'y')
  # more complicated cases: operators, subscripts, ...
  (find_globals(c('a=1%*%1%o%2 %in% d', 'b=d%%10+3%/%2-z[1:3]'), env) %==% c('d', 'z'))
})

library(testit)

assert('invalid_utf8() works and respects NA', {
  (invalid_utf8(character()) %==% integer())
  x = 'fa\xE7ile'; Encoding(x) = 'latin1'
  (invalid_utf8(c('aaa', x, NA_character_, '', '\u4e2d\u6587')) %==% 2L)
})

assert('read/write_utf8() works', {
  ascii_txt = c('aa', 'bb', 'cc')
  utf8_txt = c('\u4e2d\u6587', '\u5927\u5bb6\u597d', '\u65e9\u996d')
  latin_txt = local({x = 'fa\xE7ile'; Encoding(x) = 'latin1'; x})
  mixed_txt = c(ascii_txt, latin_txt, utf8_txt)

  write_utf8(ascii_txt, con = (ascii_file = tempfile()))
  write_utf8(utf8_txt, con = (utf8_file = tempfile()))
  write_utf8(latin_txt, con = (latin_file = tempfile()))
  write_utf8(mixed_txt, con = (mixed_file = tempfile()))

  (read_utf8(ascii_file) %==% ascii_txt)
  (read_utf8(utf8_file) %==% utf8_txt)
  (read_utf8(latin_file) %==% latin_txt) # identical will not compare Encoding
  (Encoding(read_utf8(latin_file)) %==% 'UTF-8')
  (read_utf8(mixed_file) %==% mixed_txt)
  (Encoding(read_utf8(mixed_file)) %==% c(rep('unknown', 3), rep('UTF-8', 4)))

  mixed_file2 = tempfile()
  local({
    opts = options(encoding = 'native.enc'); on.exit(options(opts), add = TRUE)
    writeLines(mixed_txt, con = mixed_file2, useBytes = TRUE)
  })
  (suppressWarnings(read_utf8(mixed_file2)[4] != mixed_txt[4]))
  has_warning(read_utf8(mixed_file2))
  has_error(read_utf8(mixed_file2, error = TRUE))
})

assert('empty files produce character() via file_string()', {
  tmp = tempfile()
  writeLines(character(), tmp)
  (file_string(tmp) %==% raw_string(character()))
})

assert('grep_sub() matches elements and do substitution on them', {
  (grep_sub('a([b]+)c', 'a\\U\\1c', c('abc', 'abbbc', 'addc', '123'), perl = TRUE) %==%
     c('aBc', 'aBBBc'))
})

library(testit)

assert("loadable() works", {
  (loadable("base"))
  # (loadable("base", new_session = TRUE))
  (loadable("base", strict = FALSE))
  (has_error(loadable(c("base", "base2"))))
  (has_error(loadable(character())))
  (!loadable("#base"))
  # (!loadable("#base", new_session = TRUE))
  (!loadable("#base", strict = FALSE))
})

library(testit)

assert('file_ext() and sans_ext() work', {
  p = c('abc.doc', 'def123.tex#', 'path/to/foo.Rmd', 'backup.ppt~', 'pkg.tar.xz')
  (file_ext(p) %==% c('doc', 'tex#', 'Rmd', 'ppt~', 'tar.xz'))
  (sans_ext(p) %==% c('abc', 'def123', 'path/to/foo', 'backup', 'pkg'))
  (file_ext(c('foo.bar.gz', 'foo', 'file.nb.html')) %==% c('gz', '', 'nb.html'))
  # some special extensions
  p = c('abc.c++', 'def.c--', 'ghi.e##', 'jkl.FB2K-COMPONENT', 'mno.WITNESS_CAMPAIGN', 'pqr.H!')
  (file_ext(p, '-+!_#') %==% c('c++', 'c--', 'e##', 'FB2K-COMPONENT', 'WITNESS_CAMPAIGN', 'H!'))
  # by default, these extensions are not recognized
  (file_ext(p) %==% character(length(p)))
})

assert('with_ext() works for corner cases', {
  (with_ext(character(), 'abc') %==% character())
  (with_ext('abc', character()) %==% 'abc')
  (with_ext(NA_character_, 'abc') %==% NA_character_)
  (has_error(with_ext('abc', NA_character_)))
  (with_ext('abc', c('d', 'e')) %==% c('abc.d', 'abc.e'))
  (has_error(with_ext(c('a', 'b'), c('d', 'e', 'f'))))
  (with_ext(c('a', 'b'), c('d', 'e')) %==% c('a.d', 'b.e'))
  (with_ext(c('a', 'b'), c('d')) %==% c('a.d', 'b.d'))
  (with_ext(c('a', 'b', 'c'), c('', '.d', 'e.e')) %==% c('a', 'b.d', 'c.e.e'))
})

assert('same_path() works', {
  (is.na(same_path('~/foo', NA_character_)))
  (is.na(same_path(NA_character_, '~/foo')))
  (same_path('~/foo', file.path(Sys.getenv('HOME'), 'foo')))
  (!same_path(tempdir(), 'foo'))
})

assert('normalize_path() works', {
  f1 = tempfile()
  writeLines('test symlink', f1)
  f2 = paste0(f1, '~')
  res = file.symlink(f1, f2)  # this may fail (on Windows), i.e., res = FALSE
  # resolve symlink by default
  (!res || basename(normalize_path(f2)) %==% basename(f1))
  # do not resolve symlink
  (!res || basename(normalize_path(f2, resolve_symlink = FALSE)) %==% basename(f2))
  # resolve_symlink = FALSE should work with inputs like . and ..
  (normalize_path(c('.', '..'), resolve_symlink = FALSE) %==% normalize_path(c('.', '..')))
})

assert('url_filename() returns the file names in URLs', {
  (url_filename('https://yihui.org/images/logo.png') %==% 'logo.png')
  (url_filename(c(
    'https://yihui.org/index.html',
    'https://yihui.org/index.html?foo=bar',
    'https://yihui.org/index.html#about'
  )) %==% rep('index.html', 3))
})

assert('is_abs_path() recognizes absolute paths on Windows and *nix', {
  (!is_abs_path('abc/def'))
  (is_abs_path(if (.Platform$OS.type == 'windows') {
    c('D:\\abc', '\\\\netdrive\\somewhere')
  } else '/abc/def'))
})

assert('del_empty_dir() correctly deletes empty dirs', {
  # do nothing is NULL
  (del_empty_dir(NULL) %==% NULL)
  # remove if empty
  dir.create(temp_dir <- tempfile())
  del_empty_dir(temp_dir)
  (!dir_exists(temp_dir))
  # do not remove if not empty
  dir.create(temp_dir <- tempfile())
  writeLines('test', tempfile(tmpdir = temp_dir))
  (del_empty_dir(temp_dir) %==% NULL)
  (dir_exists(temp_dir))
  unlink(temp_dir, recursive = TRUE)
})

assert('mark_dirs add trailing / when necessary', {
  local({
    dir.create(tmp_dir <- tempfile())
    tmp_dir_slash = paste0(tmp_dir, '/')
    file.create(tmp_file <- tempfile(tmpdir = tmp_dir))
    (mark_dirs(c(tmp_dir, tmp_file)) %==% c(tmp_dir_slash, tmp_file))
    (mark_dirs(c(tmp_dir_slash, tmp_file)) %==% c(tmp_dir_slash, tmp_file))
    unlink(tmp_dir, recursive = TRUE)
  })
})

assert('relative_path() works', {
  (relative_path(c('foo/bar.txt', 'foo/baz.txt'), 'foo/') %==% c('bar.txt', 'baz.txt'))
  (relative_path('foo/bar.txt', 'foo') %==% 'bar.txt')
})

assert('proj_root() works', {
  # detect .Rproj root
  dir.create(tmp_dir <- tempfile())
  tmp_dir_slash <- paste0(tmp_dir, '/')
  file.create(f1 <- file.path(tmp_dir, 'test.Rproj'))
  writeLines(c('Version: 1.2.3', 'test: 321'), f1)

  (same_path(proj_root(tmp_dir), tmp_dir) %==% TRUE)
  unlink(f1)

  # detect package root
  file.create(f2 <- file.path(tmp_dir, 'DESCRIPTION'))
  writeLines(c('Package: abc', 'test: 321'), f2)
  dir.create(tmp_dir_child <- tempfile(tmpdir = tmp_dir))

  (same_path(proj_root(tmp_dir_child), tmp_dir) %==% TRUE)
  unlink(tmp_dir, recursive = TRUE)
})

assert('file_rename() works', {
  # work in temp dir
  dir.create(tmp_dir <- tempfile())
  owd = setwd(tmp_dir)

  # empty dir is not moved but deleted
  dir.create('dest')
  dir.create('empty')
  file_rename('empty', 'dest')
  (dir.exists(c('empty', 'dest')) %==% c(FALSE, TRUE))

  # files are moved correctly
  dir.create('filled')
  dummy_files = c('dummy1', 'dummy2')
  file.create(file.path('filled', dummy_files))
  file_rename('filled', 'dest1')
  (dir.exists('filled') %==% FALSE)
  (list.files('dest1') %==% dummy_files)

  # rename multiple dirs
  dir.create('dest2')
  file_rename(c('dest1', 'dest2'), c('dest3', 'dest4'))
  (dir.exists(sprintf('dest%d', 1:4)) %==% c(FALSE, FALSE, TRUE, TRUE))

  # rename files
  file_rename(file.path('dest3', dummy_files), dummy_files)
  (length(list.files('dest3'))  %==% 0L)
  (file.exists(dummy_files) %==% c(TRUE, TRUE))

  # remove temp dir
  setwd(owd)
  unlink(tmp_dir, recursive = TRUE)
})

library(testit)

assert('strict_list() is really strict', {
  s_list = strict_list(aaa = 1:3, bbb = c('hey', 'mom'))
  # the class name better not be changed in the future
  (inherits(s_list, 'xfun_strict_list'))
  # `$` returns the expected value if the name provided is complete
  (s_list$aaa %==% 1:3)
  # but partial match is prohibited
  (s_list$a %==% NULL)
  # `[` will return a list - will it be better if returns a strict list?
  (inherits(s_list['aaa'], 'list'))
  # and add an element won't change the class by accident
  s_list$ccc = 4
  s_list$ddd = 'abcd'
  (inherits(s_list, 'xfun_strict_list'))
})

assert('as_strict_list() converts a list to a strict list', {
  normal_list = list(aaa = 1:3, bbb = c('hey', 'mom'))
  s_list = as_strict_list(normal_list)
  # does the strict list have the same length as the normal list?
  (length(normal_list) %==% length(s_list))
  # is the converted strict list equal to the same object created by `strict_list()`?
  s_list = strict_list(aaa = 1:3, bbb = c('hey', 'mom'))
  (as_strict_list(normal_list) %==% s_list)
})

assert('print() returns the same output for strict and normal list', {
  normal_list = list(aaa = 1:2, bbb = c('hey', 'dad'))
  s_list = as_strict_list(normal_list)
  (capture.output(print(normal_list)) %==% capture.output(print(s_list)))
})

assert('raw_string() prints as expected', {
  rs = raw_string(c('a "b"', 'hello\tworld!'))
  (inherits(rs, 'xfun_raw_string'))
  output = capture.output(rs)
  (output %==% c('a "b"', 'hello\tworld!'))
})

assert('raw_string() returns 0-length xfun_raw_string when input is NULL', {
  rs1 = raw_string(NULL)
  (inherits(rs1, 'xfun_raw_string'))
  (length(rs1) == 0)
  (capture.output(rs1) %==% character(0))

  # string 'NULL' is treated appropriately
  rs2 = raw_string('NULL')
  (inherits(rs2, 'xfun_raw_string'))
  (length(rs2) == 1)
  (capture.output(rs2) %==% 'NULL')
})

assert('raw_string() inherits from character', {
  rs = raw_string(c('a "b"', 'hello\tworld!'))
  (inherits(rs, 'character'))
})

library(testit)

assert('is_blank() tests if a vector is blank', {
  (is_blank(''))
  (is_blank(NULL) %==% logical(0))
  (is_blank(c('', '   ', '\n', '\t')))
  (!is_blank('abc'))
  (is_blank(c('', 'abc')) %==% c(TRUE, FALSE))
})

assert('n2w converts numbers to words', {
  (n2w(0) %==% 'zero')
  # cap capitalizes the first letter
  (n2w(0, cap = TRUE) %==% 'Zero')
  # hyphen adds '-' for 21-99 (except 30, 40, ...)
  (n2w(21, cap = TRUE, hyphen = TRUE) %==% 'Twenty-one')
  (n2w(21, cap = TRUE, hyphen = FALSE) %==% 'Twenty one')
  # x can be negative integers
  (n2w(-21, cap = TRUE, hyphen = TRUE) %==% 'Minus twenty-one')
  (n2w(-21, cap = TRUE, hyphen = FALSE) %==% 'Minus twenty one')
  # and controls whether to have 'add' between hundreds and double digits
  (n2w(121, and = FALSE) %==% 'one hundred twenty-one')
  (n2w(121, and = TRUE) %==% 'one hundred and twenty-one')
  # x can be a vector with length > 1
  (n2w(c(10, 13, 99, 1e6)) %==% c('ten', 'thirteen', 'ninety-nine', 'one million'))
  # the number should be less than 1e15
  (has_error(n2w(1e15)))
})

assert('join_words() joins multiple words into a single string', {
  jw = function(...) unclass(join_words(...))
  (jw(NULL) %==% character())
  (jw(c('a')) %==% 'a')
  (jw(c('a', 'b')) %==% 'a and b')
  (jw(c('a', 'b'), and = "") %==% 'a, b')
  (jw(c('a', 'b', 'c')) %==% 'a, b, and c')
  (jw(c('a', 'b', 'c'), and = '') %==% 'a, b, c')
  (jw(c('a', 'b', 'c'), ' / ', '') %==% 'a / b / c')
  (jw(c('a', 'b', 'c'), before = '"') %==% '"a", "b", and "c"')
  (jw(c('a', 'b', 'c'), before = '``', after = "''") %==% "``a'', ``b'', and ``c''")
  (jw(c('a', 'b', 'c'), before = '``', after = "''", oxford_comma = FALSE) %==% "``a'', ``b'' and ``c''")
  rm(list = 'jw')
})

assert('split_lines() splits a character vector into lines', {
  (split_lines('a') %==% 'a')
  (split_lines('') %==% '')
  (split_lines(NULL) %==% NULL)
  (split_lines('a\n') %==% c('a', ''))
  (split_lines(c('a', 'b\nc')) %==% c('a', 'b', 'c'))
  (split_lines(c('', '\n')) %==% c('', '', ''))
  (split_lines('a\nb') %==% c('a', 'b'))
  (split_lines('a\nb\n\n') %==% c('a', 'b', '', ''))
  (split_lines(c('a\nb', '', ' ', 'c')) %==% c('a', 'b', '', ' ', 'c'))
})

assert('split_source() puts lines of the same expression into a list element', {
  (split_source('1+1') %==% list('1+1'))
  (split_source(c('1+1+', '1')) %==% list(c('1+1+', '1')))
  (split_source(c('1+1+', '1', 'TRUE')) %==% list(c('1+1+', '1'), 'TRUE'))
  x = c('# a', '# b', '1', '# c', 'if (T)', 'F')
  (split_source(x) %==% c(as.list(x[1:4]), list(x[5:6])))
  (split_source(x, merge_comments = TRUE) %==% list(x[1:3], x[4:6]))
  x = c(x, '1+1')
  (split_source(x, TRUE, line_number = TRUE)[2:3] %==% list(
    structure(x[4:6], lines = c(4L, 6L)),
    structure(x[7], lines = c(7L, 7L))
  ))
})

assert('split_source() should signal an error for incomplete code', {
  (has_error(split_source('1+1+')))
  (has_error(split_source(c('1+1', '1+1+'))))
})

assert('valid_syntax() tells if a code fragment is syntactically valid', {
  (valid_syntax('1+1'))
  (!valid_syntax('1+1+'))
  (valid_syntax('if(TRUE)1'))
  (!valid_syntax(c('if(T){', 'F')))
  (valid_syntax(c('if(T){', 'F}')))
})

assert('alnum_id() generates ID strings', {
  x = c('Hello world 123!', 'a  &b*^##c 456')
  (alnum_id(x) %==% c('hello-world-123', 'a-b-c-456'))
  (alnum_id(x, '[^[:alpha:]]+') %==% c('hello-world', 'a-b-c'))
})

# This script is executed via the command line `Rscript call-fun.R arg1 arg2
# arg3`, where arg1 is a path to an .rds file, which contains the function and
# its arguments saved as a list; arg2 is a path to an .rds file to which the
# returned value of the function call is saved; arg3 saves the error message.

local({
  if (length(a <- commandArgs(TRUE)) != 3)
    stop('The number of arguments passed to Rscript should be 3.')
  # save the error message on exit if an error occurred
  on.exit(if (!file.exists(a[2])) saveRDS(geterrmessage(), a[3]))
  x = readRDS(a[1])  # list(fun, args)
  f = x[[1]]
  if (is.character(f)) f = eval(parse(text = f), envir = globalenv())
  r = do.call(f, x[[2]], envir = globalenv())
  saveRDS(r, a[2])
})

library(testit)

pkgs = c('KernSmooth', 'xfun', 'knitr')
pkg_bib(pkgs, tempfile(), tweak = FALSE)

assert('& is escaped in title when pkg_bib(tweak = TRUE)', {
  (length(grep(' & ', grep(
    '^\\s*title =', capture.output(pkg_bib(pkgs, tweak = TRUE)), value = TRUE
  ))) %==% 0L)
})

library(testit)

yaml = '
a: 1
b: [1, 2, 3]
c: true
d:
  e: !expr 1+1
  f: null
'

yaml_long_indent = '
a: 1
b: [1, 2, 3]
c: true
d:
    e: !expr 1+1
    f: null
'


yaml_tabs = '
a: 1
b: [1, 2, 3]
c: true
d:
\te: !expr 1+1
\tf: null
'

# see https://github.com/yihui/xfun/issues/94
yaml_mre = '
output:
    latex:
        latex_engine: pdflatex
        options:
            toc: true
'


if (loadable('yaml')) assert('yaml_load() works with the yaml package', {
  (yaml_load(yaml) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (yaml_load(yaml, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
  f = function() {
    foo = 1:10
    yaml_load('a: !expr head(foo, 4)')
  }
  (f() %==% list(a = 1:4))
})

assert('taml_load() works', {
  (taml_load(yaml) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (taml_load(yaml, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
  f = function() {
    foo = 1:10
    taml_load('a: !expr head(foo, 4)')
  }
  (f() %==% list(a = 1:4))
  res = list(a = c('A', 'B', 'C D'))
  (taml_load('a: [A, B, C D]') %==% res)
  (taml_load('a: ["A", \'B\', "C D"]') %==% res)
  (taml_load("a: ['A', 'B', 'C D']") %==% res)
})

assert('taml_load() works with variable indent', {
  (taml_load(yaml_long_indent) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (taml_load(yaml_long_indent, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
})

assert('taml_load() works with tabs indent', {
  (taml_load(yaml_tabs) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (taml_load(yaml_tabs, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
})

assert('yaml_load() works with a complex output', {
  expected = list(
    output = list(
      latex = list(
        latex_engine = "pdflatex", options = list(toc = TRUE)
      )
    )
  )

  (yaml_load(yaml_mre) %==% expected)
  (yaml_load(yaml_mre, use_yaml = FALSE) %==% expected)
})

library(testit)

assert('.mime_cmd() works', {
  # test some common file types
  fs = list.files(R.home('doc'), '[.](html|pdf|png)$', recursive = TRUE, full.names = TRUE)
  fs = fs[!duplicated(file_ext(fs))]
  m = list(html = 'text/html', pdf = 'application/pdf', png = 'image/png')
  types = unname(m[file_ext(fs)])
  (lapply(fs, .mime_cmd) %==% types)
  (!is_windows() || lapply(fs, .mime_cmd, use_file = FALSE) %==% types)
})

library(testit)

assert('base64_encode() encodes the string correctly', {
  ref = c(
    "AQ==", "AQI=", "AQID", "AQIDBA==", "AQIDBAU=", "AQIDBAUG",
    "AQIDBAUGBw==", "AQIDBAUGBwg=", "AQIDBAUGBwgJ", "AQIDBAUGBwgJCg=="
  )
  (sapply(1:10, function(i) base64_encode(as.raw(1:i))) == ref)

  ref = c(
    "/w==", "//4=", "//79", "//79/A==", "//79/Ps=", "//79/Pv6",
    "//79/Pv6+Q==", "//79/Pv6+fg=", "//79/Pv6+fj3", "//79/Pv6+fj39g=="
  )
  (sapply(255:246, function(i) base64_encode(as.raw(255:i))) == ref)
})

assert('base64_decode() decodes the string correctly', {
  (sapply(c(1:10, 255:246), function(i) {
    input = as.raw(1:i)
    output = base64_encode(input)
    input2 = base64_decode(output)
    input %==% input2
  }))
})

assert('base64_decode() will not make R crash if the input is not valid', {
  (has_error(base64_decode("lz..??")))
})

assert('base64_encode_r() returns the same result as base64_encode()', {
  f = R_logo()
  (base64_encode_r(f) %==% base64_encode(f))
})

assert('base64_decode() does not accept a non-string input', {
  (has_error(base64_decode(x = 42)))
})

assert('base64_decode() does not accept both string and file input', {
  f = tempfile()
  (has_error(base64_decode(x = 'Kg==', from = f)))
})

assert('base64_decode() returns the same result when the same string is used as an input directly or from a file connection', {
  f = tempfile()
  writeLines(text = "Kg==", con = f, sep = "")
  (base64_decode(from = f) %==% base64_decode(x = 'Kg=='))
})

assert('base64_uri() returns proper data type', {
  f = R_logo()
  (!grepl('[.]svg$', f) || strsplit(base64_uri(f), split = ';')[[1]][1] %==% 'data:image/svg+xml')
})

library(testit)

assert("pkg_maintainers() works", {
  x = pkg_maintainers("xfun")
  (length(x) > 0)
  # %==% not working for some reason
  (x %==% c(xfun = packageDescription('xfun', fields = 'Maintainer')))
})

library(testit)

assert('file_ext() and sans_ext() work', {
  p = c('abc.doc', 'def123.tex#', 'path/to/foo.Rmd', 'backup.ppt~', 'pkg.tar.xz')
  (file_ext(p) %==% c('doc', 'tex#', 'Rmd', 'ppt~', 'tar.xz'))
  (sans_ext(p) %==% c('abc', 'def123', 'path/to/foo', 'backup', 'pkg'))
  (file_ext(c('foo.bar.gz', 'foo', 'file.nb.html')) %==% c('gz', '', 'nb.html'))
})

assert('with_ext() works for corner cases', {
  (with_ext(character(), 'abc') %==% character())
  (with_ext('abc', character()) %==% 'abc')
  (with_ext(NA_character_, 'abc') %==% NA_character_)
  (has_error(with_ext('abc', NA_character_)))
  (with_ext('abc', c('d', 'e')) %==% c('abc.d', 'abc.e'))
  (has_error(with_ext(c('a', 'b'), c('d', 'e', 'f'))))
  (with_ext(c('a', 'b'), c('d', 'e')) %==% c('a.d', 'b.e'))
  (with_ext(c('a', 'b'), c('d')) %==% c('a.d', 'b.d'))
  (with_ext(c('a', 'b', 'c'), c('', '.d', 'e.e')) %==% c('a', 'b.d', 'c.e.e'))
})

assert('same_path() works', {
  (is.na(same_path('~/foo', NA_character_)))
  (is.na(same_path(NA_character_, '~/foo')))
  (same_path('~/foo', file.path(Sys.getenv('HOME'), 'foo')))
  (!same_path(tempdir(), 'foo'))
})

assert('url_filename() returns the file names in URLs', {
  (url_filename('https://yihui.org/images/logo.png') %==% 'logo.png')
  (url_filename(c(
    'https://yihui.org/index.html',
    'https://yihui.org/index.html?foo=bar',
    'https://yihui.org/index.html#about'
  )) %==% rep('index.html', 3))
})

assert('is_abs_path() recognizes absolute paths on Windows and *nix', {
  (!is_abs_path('abc/def'))
  (is_abs_path(if (.Platform$OS.type == 'windows') {
    c('D:\\abc', '\\\\netdrive\\somewhere')
  } else '/abc/def'))
})

assert('del_empty_dir() correctly deletes empty dirs', {
  # do nothing is NULL
  (del_empty_dir(NULL) %==% NULL)
  # remove if empty
  dir.create(temp_dir <- tempfile())
  del_empty_dir(temp_dir)
  (!dir_exists(temp_dir))
  # do not remove if not empty
  dir.create(temp_dir <- tempfile())
  writeLines('test', tempfile(tmpdir = temp_dir))
  (del_empty_dir(temp_dir) %==% NULL)
  (dir_exists(temp_dir))
  unlink(temp_dir, recursive = TRUE)
})

assert('mark_dirs add trailing / when necessary', {
  local({
    dir.create(tmp_dir <- tempfile())
    tmp_dir_slash = paste0(tmp_dir, "/")
    file.create(tmp_file <- tempfile(tmpdir = tmp_dir))
    (mark_dirs(c(tmp_dir, tmp_file)) %==% c(tmp_dir_slash, tmp_file))
    (mark_dirs(c(tmp_dir_slash, tmp_file)) %==% c(tmp_dir_slash, tmp_file))
    unlink(tmp_dir, recursive = TRUE)
  })
})

#' Get the tags of Github releases of a repository
#'
#' Use the Github API (\code{\link{github_api}()}) to obtain the tags of the
#' releases.
#' @param repo The repository name of the form \code{user/repo}, e.g.,
#'   \code{"yihui/xfun"}.
#' @param tag A tag as a character string. If provided, it will be returned if
#'   the tag exists. If \code{tag = "latest"}, the tag of the latest release is
#'   returned.
#' @param pattern A regular expression to match the tags.
#' @param use_jsonlite Whether to use \pkg{jsonlite} to parse the releases info.
#' @export
#' @return A character vector of (GIT) tags.
#' @examplesIf interactive()
#' xfun::github_releases('yihui/xfun')
#' xfun::github_releases('gohugoio/hugo')
github_releases = function(
  repo, tag = '', pattern = 'v[0-9.]+', use_jsonlite = loadable('jsonlite')
) {
  if (tag != '') return(github_releases2(repo, tag, pattern))

  i = 1; v = character()
  repeat {
    res = github_api(
      sprintf('/repos/%s/tags', repo), NULL, list(per_page = 100, page = i),
      raw = !use_jsonlite
    )
    v2 = unlist(if (use_jsonlite) {
      lapply(res, `[[`, 'name')
    } else {
      reg_match('\\{"name":"([^"]+)",', res)
    })
    if (length(v2) == 0) break
    v = c(v, v2)
    if (length(v2) < 100) break  # not enough items for the next page
    i = i + 1
  }
  grep(sprintf('^%s$', pattern), unique(v), value = TRUE)
}

# extract the matched elements in the n-th pair of () in the regex
reg_match = function(p, x, n = 1, ...) {
  # TODO: gregexec was added in R 4.1.0; remove this workaround when we don't
  # need to support R < 4.1.0
  v = 'gregexec' %in% ls(baseenv())
  m = (if (v) base::gregexec else base::gregexpr)(p, x, ...)
  lapply(regmatches(x, m), function(x) {
    if (v) x[n + 1, ] else gsub(p, paste0('\\', n), x)
  })
}

# the fallback method to retrieve release tags (read HTML source)
github_releases2 = function(repo, tag = '', pattern = '[^"&]+') {
  read = function() suppressWarnings(
    read_utf8(sprintf('https://github.com/%s/releases/%s', repo, tag))
  )
  h = if (tag == '') read() else tryCatch(read(), error = function(e) '')
  r = sprintf('^.*?%s/releases/tag/(%s)".*', repo, pattern)
  unique(grep_sub(r, '\\1', h))
}

#' @details \code{github_api()} is a wrapper function based on
#'   \code{rest_api_raw()} to obtain data from the Github API:
#'   \url{https://docs.github.com/en/rest}. You can provide a personal access
#'   token (PAT) via the \code{token} argument, or via one of the environment
#'   variables \var{GITHUB_PAT}, \var{GITHUB_TOKEN}, \var{GH_TOKEN}. A PAT
#'   allows for a much higher rate limit in API calls. Without a token, you can
#'   only make 60 calls in an hour.
#' @param raw Whether to return the raw response or parse the response with
#'   \pkg{jsonlite}.
#' @rdname rest_api
#' @export
github_api = function(
  endpoint, token = '', params = list(), headers = NULL, raw = !loadable('jsonlite')
) {
  token = c(token, unname(Sys.getenv(envs <- c('GITHUB_PAT', 'GITHUB_TOKEN', 'GH_TOKEN'))))
  token = if (length(token <-  token[token != ''])) token[1] else ''
  names(token) = 'token'
  error = TRUE
  on.exit(if (error && token == '') message(
    'You may need to save a Github personal access token in one of the ',
    'environment variables: ', paste(envs, collapse = ', ')
  ))
  res = rest_api_raw('https://api.github.com', endpoint, token, params, headers)
  error = FALSE
  if (raw) res else jsonlite::fromJSON(res, FALSE)
}

git = function(...) {
  if (Sys.which('git') == '') stop('git is not available')
  # R's HOME var is different from the system's HOME on Windows:
  # https://github.com/yihui/crandalf/issues/24
  if (is_windows()) {
    env = set_envvar(c(HOME = Sys.getenv('USERPROFILE')))
    on.exit(set_envvar(env), add = TRUE)
  }
  system2('git', ...)
}

git_co = function(args = NULL, ...) {
  git(c('checkout', args), ...)
}

git_test_branch = function() {
  if (length(d <- git(c('diff', '--name-only'), stdout = TRUE))) stop(
    'The current branch has changes not stated for commit:\n',
    paste(d, collapse = '\n')
  )
}

gh = function(...) {
  if (Sys.which('gh') == '') stop('Github CLI not found: https://cli.github.com')
  system2('gh', ...)
}

gh_run = function(..., repo = NA) {
  gh(c(if (!is.na(repo)) c('-R', repo), 'run', ...), stdout = TRUE)
}

#' Try to use the system native encoding to represent a character vector
#'
#' Apply \code{enc2native()} to the character vector, and check if
#' \code{enc2utf8()} can convert it back without a loss. If it does, return
#' \code{enc2native(x)}, otherwise return the original vector with a warning.
#' @param x A character vector.
#' @note On platforms that supports UTF-8 as the native encoding
#'   (\code{\link{l10n_info}()[['UTF-8']]} returns \code{TRUE}), the conversion
#'   will be skipped.
#' @export
#' @examples
#' library(xfun)
#' s = intToUtf8(c(20320, 22909))
#' Encoding(s)
#'
#' s2 = native_encode(s)
#' Encoding(s2)
native_encode = function(x) {
  if (isTRUE(l10n_info()[['UTF-8']])) return(x)
  if (identical(enc2utf8(x2 <- enc2native(x)), x)) return(x2)
  warning('The character vector cannot be represented in the native encoding')
  x
}

#' Check if a character vector consists of entirely ASCII characters
#'
#' Converts the encoding of a character vector to \code{'ascii'}, and check if
#' the result is \code{NA}.
#' @param x A character vector.
#' @return A logical vector indicating whether each element of the character
#'   vector is ASCII.
#' @export
#' @examples library(xfun)
#' is_ascii(letters)  # yes
#' is_ascii(intToUtf8(8212))  # no
is_ascii = function(x) {
  out = !is.na(iconv(x, to = 'ascii'))
  out[is.na(x)] = NA
  out
}

#' Run \command{R CMD check} on the reverse dependencies of a package
#'
#' Install the source package, figure out the reverse dependencies on CRAN,
#' download all of their source packages, and run \command{R CMD check} on them
#' in parallel.
#'
#' Everything occurs under the current working directory, and you are
#' recommended to call this function under a designated directory, especially
#' when the number of reverse dependencies is large, because all source packages
#' will be downloaded to this directory, and all \file{*.Rcheck} directories
#' will be generated under this directory, too.
#'
#' If a source tarball of the expected version has been downloaded before (under
#' the \file{tarball} directory), it will not be downloaded again (to save time
#' and bandwidth).
#'
#' After a package has been checked, the associated \file{*.Rcheck} directory
#' will be deleted if the check was successful (no warnings or errors or notes),
#' which means if you see a \file{*.Rcheck} directory, it means the check
#' failed, and you need to take a look at the log files under that directory.
#'
#' The time to finish the check is recorded for each package. As the check goes
#' on, the total remaining time will be roughly estimated via \code{n *
#' mean(times)}, where \code{n} is the number of packages remaining to be
#' checked, and \code{times} is a vector of elapsed time of packages that have
#' been checked.
#'
#' If a check on a reverse dependency failed, its \file{*.Rcheck} directory will
#' be renamed to \file{*.Rcheck2}, and another check will be run against the
#' CRAN version of the package unless \code{options(xfun.rev_check.compare =
#' FALSE)} is set. If the logs of the two checks are the same, it means no new
#' problems were introduced in the package, and you can probably ignore this
#' particular reverse dependency. The function \code{compare_Rcheck()} can be
#' used to create a summary of all the differences in the check logs under
#' \file{*.Rcheck} and \file{*.Rcheck2}. This will be done automatically if
#' \code{options(xfun.rev_check.summary = TRUE)} has been set.
#'
#' A recommended workflow is to use a special directory to run
#' \code{rev_check()}, set the global \code{\link{options}}
#' \code{xfun.rev_check.src_dir} and \code{repos} in the R startup (see
#' \code{?\link{Startup}}) profile file \code{.Rprofile} under this directory,
#' and (optionally) set \code{R_LIBS_USER} in \file{.Renviron} to use a special
#' library path (so that your usual library will not be cluttered). Then run
#' \code{xfun::rev_check(pkg)} once, investigate and fix the problems or (if you
#' believe it was not your fault) ignore broken packages in the file
#' \file{00ignore}, and run \code{xfun::rev_check(pkg)} again to recheck the
#' failed packages. Repeat this process until all \file{*.Rcheck} directories
#' are gone.
#'
#' As an example, I set \code{options(repos = c(CRAN =
#' 'https://cran.rstudio.com'), xfun.rev_check.src_dir = '~/Dropbox/repo')} in
#' \file{.Rprofile}, and \code{R_LIBS_USER=~/R-tmp} in \file{.Renviron}. Then I
#' can run, for example, \code{xfun::rev_check('knitr')} repeatedly under a
#' special directory \file{~/Downloads/revcheck}. Reverse dependencies and their
#' dependencies will be installed to \file{~/R-tmp}, and \pkg{knitr} will be
#' installed from \file{~/Dropbox/repo/kintr}.
#' @param pkg The package name.
#' @param which Which types of reverse dependencies to check. See
#'   \code{tools::\link[tools]{package_dependencies}()} for possible values. The
#'   special value \code{'hard'} means the hard dependencies, i.e.,
#'   \code{c('Depends', 'Imports', 'LinkingTo')}.
#' @param recheck A vector of package names to be (re)checked. If not provided
#'   and there are any \file{*.Rcheck} directories left by certain packages
#'   (this often means these packages failed the last time), \code{recheck} will
#'   be these packages; if there are no \file{*.Rcheck} directories but a text
#'   file \file{recheck} exists, \code{recheck} will be the character vector
#'   read from this file. This provides a way for you to manually specify the
#'   packages to be checked. If there are no packages to be rechecked, all
#'   reverse dependencies will be checked.
#' @param ignore A vector of package names to be ignored in \command{R CMD
#'   check}. If this argument is missing and a file \file{00ignore} exists, the
#'   file will be read as a character vector and passed to this argument.
#' @param update Whether to update all packages before the check.
#' @param src The path of the source package directory.
#' @param src_dir The parent directory of the source package directory. This can
#'   be set in a global option if all your source packages are under a common
#'   parent directory.
#' @param timeout Timeout in seconds for \command{R CMD check} to check each
#'   package. The (approximate) total time can be limited by the global option
#'   \code{xfun.rev_check.timeout_total}.
#' @return A named numeric vector with the names being package names of reverse
#'   dependencies; \code{0} indicates check success, \code{1} indicates failure,
#'   and \code{2} indicates that a package was not checked due to global
#'   timeout.
#' @seealso \code{devtools::revdep_check()} is more sophisticated, but currently
#'   has a few major issues that affect me: (1) It always deletes the
#'   \file{*.Rcheck} directories
#'   (\url{https://github.com/r-lib/devtools/issues/1395}), which makes it
#'   difficult to know more information about the failures; (2) It does not
#'   fully install the source package before checking its reverse dependencies
#'   (\url{https://github.com/r-lib/devtools/pull/1397}); (3) I feel it is
#'   fairly difficult to iterate the check (ignore the successful packages and
#'   only check the failed packages); by comparison, \code{xfun::rev_check()}
#'   only requires you to run a short command repeatedly (failed packages are
#'   indicated by the existing \file{*.Rcheck} directories, and automatically
#'   checked again the next time).
#'
#'   \code{xfun::rev_check()} borrowed a very nice feature from
#'   \code{devtools::revdep_check()}: estimating and displaying the remaining
#'   time. This is particularly useful for packages with huge numbers of reverse
#'   dependencies.
#' @export
rev_check = function(
  pkg, which = 'all', recheck = NULL, ignore = NULL, update = TRUE,
  timeout = getOption('xfun.rev_check.timeout', 15 * 60),
  src = file.path(src_dir, pkg), src_dir = getOption('xfun.rev_check.src_dir')
) {
  if (length(src) != 1 || !dir_exists(src)) stop(
    'The package source dir (the "src" argument) must be an existing directory'
  )

  message('Installing the source package ', src)
  install_dir(path.expand(src))

  db = available.packages(type = 'source')

  # install packages that are not loadable (testing in parallel)
  p_install = function(pkgs) {
    pkgs_up = NULL
    if (update) {
      message('Updating all R packages...')
      pkgs_up = intersect(old.packages(checkBuilt = TRUE)[, 'Package'], pkgs)
      pkg_install(pkgs_up)
    }
    pkgs = setdiff(pkgs, pkgs_up)  # don't install pkgs that were just updated
    print(system.time(
      pkg_install(unlist(plapply(pkgs, function(p) if (!loadable(p, new_session = TRUE)) p)))
    ))
  }

  unlink('*.Rcheck2', recursive = TRUE)
  if (missing(recheck)) {
    dirs = list.files('.', '.+[.]Rcheck$')
    pkgs = gsub('.Rcheck$', '', dirs)
    recheck = if (length(pkgs) == 0 && file_exists('recheck')) {
      scan('recheck', 'character')
    } else pkgs
  }
  pkgs = if (length(recheck)) {
    p_install(pkg_dep(recheck, db, which = 'all'))
    recheck
  } else {
    res = check_deps(pkg, db, which)
    message('Installing dependencies of reverse dependencies')
    res$install = setdiff(res$install, ignore_deps())
    print(system.time(p_install(res$install)))
    res$check
  }
  pkgs = intersect(pkgs, rownames(db))  # make sure the pkgs are on CRAN

  lib_cran = './library-cran'
  on.exit(unlink(lib_cran, recursive = TRUE), add = TRUE)
  dir.create(lib_cran, showWarnings = FALSE)
  pkg_install(pkg, lib = lib_cran)  # the CRAN version of the package

  f = tempfile('check-done', fileext = '.rds')
  l = tempfile('check-lock'); on.exit(unlink(c(f, l)), add = TRUE)
  n = length(pkgs)
  if (n == 0) {
    message('No reverse dependencies to be checked for the package ', pkg); return()
  }

  if (missing(ignore) && file_exists('00ignore')) ignore = scan('00ignore', 'character')
  if (length(ignore)) {
    message('Ignoring packages: ', paste(ignore, collapse = ' '))
    unlink(sprintf('%s.Rcheck', ignore), recursive = TRUE)
    pkgs = setdiff(pkgs, ignore)
    if ((n <- length(pkgs)) == 0) {
      message('No packages left to be checked'); return()
    }
  }

  message('Downloading tarballs')
  tars = download_tarball(pkgs, db, dir = 'tarball')
  tars = setNames(tars, pkgs)

  t0 = Sys.time()
  tt = getOption('xfun.rev_check.timeout_total', Inf)
  message('Checking ', n, ' packages: ', paste(pkgs, collapse = ' '))

  res = plapply(pkgs, function(p) {
    d = sprintf('%s.Rcheck', p)
    if (!p %in% rownames(db)) {
      message('Checking ', p, ' (aborted since it is no longer on CRAN')
      unlink(d, recursive = TRUE)
      return()
    }

    timing = function() {
      # in case two packages finish at exactly the same time
      while (file_exists(l)) Sys.sleep(.1)
      file.create(l); on.exit(unlink(l), add = TRUE)
      done = c(if (file_exists(f)) readRDS(f), p)
      saveRDS(done, f)
      n2 = length(setdiff(pkgs, done))  # remaining packages
      t1 = Sys.time(); t2 = Sys.time() + n2 * (t1 - t0) / (n - n2)
      message(
        'Packages remaining: ', n2, '/', n, '; Expect to finish at ', t2,
        ' (', format(round(difftime(t2, t1))), ')'
      )
      # 0 (FALSE): success; 1: failure
      setNames(as.integer(dir_exists(d)), p)
    }

    if (!file_exists(z <- tars[p])) {
      dir.create(d, showWarnings = FALSE)
      return(timing())
    }

    # timeout; package not checked
    if (difftime(Sys.time(), t0, units = 'secs') > tt) {
      return(setNames(2L, p))
    }

    check_it = function(args = NULL, ...) {
      system2(
        file.path(R.home('bin'), 'R'),
        c(args, 'CMD', 'check', '--no-manual', shQuote(z)),
        stdout = FALSE, stderr = FALSE, timeout = timeout, ...
      )
    }
    check_it()

    if (!clean_Rcheck(d)) {
      if (!dir_exists(d)) {dir.create(d); return(timing())}
      # try to install missing LaTeX packages for vignettes if possible, then recheck
      vigs = list.files(
        file.path(d, 'vign_test', p, 'vignettes'), '[.](Rnw|Rmd)$',
        ignore.case = TRUE, full.names = TRUE
      )
      pkg_load2('tinytex')
      if (length(vigs) && any(file_exists(with_ext(vigs, 'log')))) {
        if (tinytex::is_tinytex()) for (vig in vigs) in_dir(dirname(vig), {
          Rscript(shQuote(c('-e', 'if (grepl("[.]Rnw$", f <- commandArgs(T), ignore.case = T)) knitr::knit2pdf(f) else rmarkdown::render(f)', basename(vig))))
        })
        check_it()
        if (clean_Rcheck(d)) return(timing())
      }
      # if there are still missing LaTeX packages, install them and recheck
      l0 = tinytex::tl_pkgs()
      lapply(
        list.files(d, '[.]log$', full.names = TRUE, recursive = TRUE),
        tinytex::parse_install, quiet = TRUE
      )
      if (!identical(l0, tinytex::tl_pkgs())) {
        check_it()
        if (clean_Rcheck(d)) return(timing())
      }
      # clean up the check log, and recheck with the current CRAN version of pkg
      cleanup = function() in_dir(d, {
        clean_log()
        # so that I can easily preview it in the Finder on macOS
        file_exists('00install.out') && file.rename('00install.out', '00install.log')
      })
      # ignore vignettes that failed to build for unknown reasons
      cleanup()
      if (clean_Rcheck(d)) return(timing())
      # whether to check the package against the CRAN version?
      if (!getOption('xfun.rev_check.compare', TRUE)) return(timing())
      file.rename(d, d2 <- paste0(d, '2'))
      check_it('--no-environ', env = tweak_r_libs(lib_cran))
      if (!dir_exists(d)) file.rename(d2, d) else {
        cleanup()
        if (identical_logs(c(d, d2))) unlink(c(d, d2), recursive = TRUE)
      }
    }
    timing()
  })
  if (getOption('xfun.rev_check.summary', FALSE)) {
    html = compare_Rcheck(); if (isTRUE(grepl('[.]html$', html))) browseURL(html)
  }
  unlist(res)
}

# remove the OK lines in the check log
clean_log = function() {
  if (!file_exists(l <- '00check.log')) return()
  x = grep('^[*].+OK$', read_utf8(l), invert = TRUE, value = TRUE)
  # don't want diffs in random tempdir/tempfile paths when comparing check logs
  x[grep(dirname(tempdir()), x, fixed = TRUE)] = 'RANDOM TEMPDIR/TEMPFILE PATH DELETED'
  # delete the download progress
  x = grep('^\\s*\\[\\d+%] Downloaded \\d+ bytes...\\s*$', x, invert = TRUE, value = TRUE)
  # delete lines of the form "address 0x1067143eb, cause 'illegal opcode'"
  x = grep("address 0x[[:xdigit:]]+, cause '[^']+'", x, invert = TRUE, value = TRUE)
  x = recheck_vig(x)
  x = tail(x, -2)
  writeLines(x, l)  # remove the first 2 lines (log dir name and R version)
  x
}

# sometimes R CMD check fails to build vignettes for unknown reasons; try to
# recheck the package in this case
recheck_vig = function(x) {
  if (!any(i1 <- (x == '* checking re-building of vignette outputs ... WARNING')))
    return(x)

  i1 = which(i1)[1]
  i2 = which(x == 'Execution halted')
  i2 = i2[i2 > i1]
  if (length(i2) == 0) return(x)

  i3 = grep('^[*] checking ', x)  # next checking item
  i3 = i3[i3 > i1]
  if (length(i3)) {
    i2 = i2[i2 < i3[1]]  # 'Execution halted' needs to appear before next '* checking'
    if (length(i2) == 0) return(x)
  }

  # if no explicit errors were found in processing vignettes (except pandoc
  # error), remove the relevant log
  i2 = tail(i2, 1)
  if (length(grep('pandoc document conversion failed with error', x[i1:i2])) > 0 ||
      length(grep('Error: processing vignette .+ failed with diagnostics:', x[i1:i2])) == 0)
    x = x[-(i1:i2)]
  x
}

# are the check logs identical under a series of *.Rcheck directories?
identical_logs = function(dirs) {
  if (length(dirs) < 2) return(FALSE)
  if (!all(file_exists(logs <- file.path(dirs, '00check.log')))) return(FALSE)
  x = read_utf8(logs[1])
  for (i in 2:length(dirs)) if (!identical(x, read_utf8(logs[i]))) return(FALSE)
  TRUE
}

# delete files/dirs that are usually not helpful
clean_Rcheck2 = function(dir = '.') {
  owd = setwd(dir); on.exit(setwd(owd), add = TRUE)
  ds = list.files('.', '.+[.]Rcheck$')
  for (d in c(ds, paste0(ds, '2'))) {
    f1 = list.files(d, full.names = TRUE)
    f2 = file.path(d, c('00_pkg_src', '00check.log', '00install.log'), fsep = '/')
    unlink(setdiff(f1, f2), recursive = TRUE)
  }
}

# add a new library path to R_LIBS_USER
tweak_r_libs = function(new) {
  x = read_all(existing_files(c('~/.Renviron', '.Renviron')))
  x = grep('^\\s*#', x, invert = TRUE, value = TRUE)
  x = gsub('^\\s+|\\s+$', '', x)
  x = x[x != '']
  i = grep('^R_LIBS_USER=.+', x)
  if (length(i)) {
    x[i[1]] = sub('(="?)', path_sep('\\1', new), x[i[1]])
    x
  } else {
    v = Sys.getenv('R_LIBS_USER')
    v = if (v == '') new else path_sep(new, v)
    c(paste0('R_LIBS_USER=', v), x)
  }
}

# separate paths by the path separator on a specific platform
path_sep = function(...) paste(..., sep = .Platform$path.sep)

# a shorthand of tools::package_dependencies()
pkg_dep = function(x, ...) {
  if (length(x)) unique(unlist(tools::package_dependencies(x, ...)))
}

# calculate the packages required to check a package
check_deps = function(x, db = available.packages(), which = 'all') {
  if (identical(which, 'hard')) which = c('Depends', 'Imports', 'LinkingTo')
  x0 = db[, 'Package']  # all available packages
  # packages that reverse depend on me
  x1 = pkg_dep(x, db, which, reverse = TRUE)
  x1 = intersect(x1, x0)
  # only check a sample of soft reverse dependencies (useful if there are too many)
  if (identical(which, 'all') && (n <- getOption('xfun.rev_check.sample', 100)) >= 0) {
    x2 = pkg_dep(x, db, c('Suggests', 'Enhances'), reverse = TRUE)
    x2 = intersect(x2, x0)
    if (n < length(x2)) x1 = c(setdiff(x1, x2), sample(x2, n))
  }
  # to R CMD check x1, I have to install all their dependencies
  x2 = pkg_dep(x1, db, 'all')
  # and for those dependencies, I have to install the default dependencies
  x3 = pkg_dep(x2, db, recursive = TRUE)
  list(check = x1, install = intersect(c(x1, x2, x3), x0))
}

#' Submit check jobs to crandalf
#'
#' Check the reverse dependencies of a package using the crandalf service:
#' \url{https://github.com/yihui/crandalf}. If the number of reverse
#' dependencies is large, they will be split into batches and pushed to crandalf
#' one by one.
#'
#' Due to the time limit of a single job on Github Actions (6 hours), you will
#' have to split the large number of reverse dependencies into batches and check
#' them sequentially on Github (at most 5 jobs in parallel). The function
#' \code{crandalf_check()} does this automatically when necessary. It requires
#' the \command{git} command to be available.
#'
#' The function \code{crandalf_results()} fetches check results from Github
#' after all checks are completed, merge the results, and show a full summary of
#' check results. It requires \code{gh} (Github CLI:
#' \url{https://cli.github.com/manual/}) to be installed and you also need to
#' authenticate with your Github account beforehand.
#' @param pkg The package name of which the reverse dependencies are to be
#'   checked.
#' @param size The number of reverse dependencies to be checked in each job.
#' @param jobs The number of jobs to run in Github Actions (by default, all jobs
#'   are submitted, but you can choose to submit the first few jobs).
#' @param which The type of dependencies (see \code{\link{rev_check}()}).
#' @export
crandalf_check = function(pkg, size = 400, jobs = Inf, which = 'all') {
  git_test_branch()
  git_co('main')
  on.exit(git_co('main'), add = TRUE)
  git_test_branch()

  # do everything inside the check-pkg branch
  b = paste0('check-', pkg)
  if (git_co(b, stderr = FALSE) != 0) {
    git_co(c('-b', b))
    file.create('recheck')
    git(c('add', 'recheck'))
    git(c('commit', '-m', shQuote(paste('Revcheck', pkg))))
    git('push')
    message(
      'Please create a pull request from the branch ', b,
      ' on Github and re-run xfun::crandalf_check("', pkg, '").'
    )
    return(invisible())
  }
  git(c('merge', 'main'))

  x = check_deps(pkg, which = which)$check
  n = length(x)
  if (n <= size) {
    message('No need to split ', n, ' reverse dependencies into batches of size ', size, '.')
    if (any(grepl('Your branch is ahead of ', git('status', stdout = TRUE)))) {
      git('push')
    } else if (Sys.which('gh') != '') {
      gh(c('workflow', 'run', 'rev-check.yaml', '--ref', b))
      message('Triggering rev-check.yaml job against ', b, ' branch in crandalf repo on Github.')
    } else {
      message('Remember to re-run the last job for the package ', pkg, ' on Github.')
    }
    return(invisible())
  }

  b = ceiling(n/size)
  i = rep(seq_len(b), each = size)[seq_len(n)]
  k = 1
  # use an id in the commit so that I know which jobs are for the same pkg
  id = format(Sys.time(), '%Y%m%d%H%M')
  for (p in head(split(x, i), jobs)) {
    message('Batch ', k)
    writeLines(p, 'recheck')
    git(c('add', 'recheck'))
    git(c('commit', '-m', shQuote(paste(
      c(id, 'checking:', head(p, 3), '...'), collapse = ' '
    ))))
    git('push')
    Sys.sleep(10)
    k = k + 1
  }
}

#' @param repo The crandalf repo on Github (of the form \code{user/repo} such as
#'   \code{"yihui/crandalf"}). Usually you do not need to specify it, unless you
#'   are not calling this function inside the crandalf project, because
#'   \command{gh} should be able to figure out the repo automatically.
#' @param limit The maximum of records for \command{gh run list} to retrieve.
#'   You only need a larger number if the check results are very early in the
#'   Github Action history.
#' @param wait Number of seconds to wait if not all jobs have been completed on
#'   Github. By default, this function checks the status every 5 minutes until
#'   all jobs are completed. Set \code{wait} to 0 to disable waiting (and throw
#'   an error immediately when any jobs are not completed).
#' @rdname crandalf_check
#' @export
crandalf_results = function(pkg, repo = NA, limit = 200, wait = 5 * 60) {
  res = crandalf_jobs(pkg, repo, limit)
  if (NROW(res) == 0) {
    stop('Did not find check results for ', pkg, ' from Github Actions.')
  }
  if (any(res[, 1] != 'completed')) {
    if (wait <= 0) stop('Please wait till all jobs have been completed on Github Actions.')
    status = NULL
    repeat {
      res = crandalf_jobs(pkg, repo, limit)
      if (all(res[, 1] == 'completed')) break
      if (is.null(status) || !identical(status, table(res[, 1]))) {
        status = table(res[, 1])
        timestamp()
        print(status)
      }
      Sys.sleep(wait)
    }
  }
  ids = grep_sub('^(\\d+) checking: .+', '\\1', res[, 3])
  i = if (length(ids) > 0) grep(sprintf('^%s checking: ', ids[1]), res[, 3]) else {
    head(which(res[, 2] == 'failure'), 1)
  }
  res = res[i, , drop = FALSE]
  res = res[res[, 2] == 'failure', , drop = FALSE]
  if (NROW(res) == 0) {
    stop('Did not find any failed results on Github Actions.')
  }
  for (i in seq_len(nrow(res))) {
    message('Downloading check results (', i, '/', nrow(res), ')')
    gh_run('download', res[i, 7], '-D', tempfile('crandalf-', '.'), repo = repo)
  }
  if (interactive()) browseURL(crandalf_merge(pkg))
}

# retrieve the first N jobs info
crandalf_jobs = function(pkg, repo = NA, limit = 200) {
  res = gh_run('list', '-L', limit, '-w', 'rev-check', repo = repo)
  res = res[grep(paste0('rev-check\tcheck-', pkg), res)]
  do.call(rbind, strsplit(res, '\t'))
}

crandalf_merge = function(pkg) {
  unlink(list.files('.', '[.]Rcheck2?$'), recursive = TRUE)
  x1 = x2 = x3 = NULL
  f1 = '00check_diffs.html'; f3 = 'latex.txt'
  for (d in list.files('.', '^crandalf-.+')) {
    if (!dir_exists(d)) next
    p = file.path(d, 'macOS-rev-check-results')
    if (file_exists(f <- file.path(p, f1))) {
      x = read_utf8(f)
      x1 = if (length(x1) == 0) x else {
        i1 = grep('<body>', x)[1]
        i2 = tail(grep('</body>', x), 1)
        i3 = tail(grep('</body>', x1), 1)
        append(x1, x[(i1 + 1):(i2 - 1)], i3 - 1)
      }
      file.remove(f)
    }
    if (file_exists(f <- file.path(p, 'recheck2'))) {
      x2 = c(x2, read_utf8(f))
      file.remove(f)
    }
    cs = list.files(p, '[.]Rcheck[2]?$', full.names = TRUE)
    file.rename(cs, basename(cs))
    if (file_exists(f <- file.path(p, f3))) {
      x3 = c(x3, read_utf8(f))
      file.remove(f)
    }
    unlink(d, recursive = TRUE)
  }
  write_utf8(x1, f1)  # the full summary

  # store newly detected missing latex packages in latex.txt and commit/push
  git_co('main')
  append_unique(x3, f3)
  find_missing_latex()
  git(c('commit', '-m', shQuote('add more latex packages'), f3))
  git('push')

  git_co(paste0('check-', pkg))
  r = '[.]Rcheck2$'
  write_utf8(sort(unique(c(x2, gsub(r, '', list.files('.', r))))), 'recheck')
  f1
}

# mclapply() with a different default for mc.cores and disable prescheduling
plapply = function(X, FUN, ...) {
  parallel::mclapply(
    X, FUN, ..., mc.cores = getOption('mc.cores', parallel::detectCores()),
    mc.preschedule = FALSE
  )
}

# download the source package from CRAN
download_tarball = function(p, db = available.packages(type = 'source'), dir = '.', retry = 3) {
  if (!dir_exists(dir)) dir.create(dir, recursive = TRUE)
  z = file.path(dir, sprintf('%s_%s.tar.gz', p, db[p, 'Version']))
  mapply(function(p, z) {
    # remove other versions of the package tarball
    unlink(setdiff(list.files(dir, sprintf('^%s_.+.tar.gz', p), full.names = TRUE), z))
    for (i in seq_len(retry)) {
      if (file_exists(z)) break
      try(download.file(paste(db[p, 'Repository'], basename(z), sep = '/'), z, mode = 'wb'))
    }
  }, p, z, SIMPLIFY = FALSE)
  z
}

# clean up *.Rcheck if there are no warnings, errors, or notes in the log
clean_Rcheck = function(dir, log = read_utf8(file.path(dir, '00check.log'))) {
  # do not check the status line
  if (length(grep('^Status: ', tail(log, 1)))) log = head(log, -1)
  if (length(grep('(WARNING|ERROR|NOTE)$', log)) == 0 ||
      length(grep('[*] checking whether package .+ can be installed ... ERROR', log)))
    unlink(dir, recursive = TRUE)
  !dir_exists(dir)
}

#' @rdname rev_check
#' @param status_only If \code{TRUE}, only compare the final statuses of the
#'   checks (the last line of \file{00check.log}), and delete \file{*.Rcheck}
#'   and \file{*.Rcheck2} if the statuses are identical, otherwise write out the
#'   full diffs of the logs. If \code{FALSE}, compare the full logs under
#'   \file{*.Rcheck} and \file{*.Rcheck2}.
#' @param output The output Markdown file to which the diffs in check logs will
#'   be written. If the \pkg{markdown} package is available, the Markdown file
#'   will be converted to HTML, so you can see the diffs more clearly.
#' @export
compare_Rcheck = function(status_only = TRUE, output = '00check_diffs.md') {
  if (length(dirs <- list.files('.', '.+[.]Rcheck$')) == 0) {
    # clean up the `recheck` file
    if (file_exists('recheck')) writeLines(character(), 'recheck')
    return()
  }
  d2 = function(d) c(d, paste0(d, '2'))
  logs = function(d) file.path(d2(d), '00check.log')
  res = NULL
  if (!status_only && Sys.which('diff') == '')
    warning("The command 'diff' is not available; will not calculate exact diffs in logs.")
  for (d in dirs) {
    f = existing_files(logs(d))
    if (status_only && length(f) == 2) {
      status_line = function(file) {
        x = tail(read_utf8(file), 1)
        if (grepl('^Status: ', x)) x else {
          warning('The last line of ', file, ' is not the status.')
          NULL
        }
      }
      # if the check with current CRAN version of package also failed, or the
      # two statues are the same, chances are we are good to go
      s1 = status_line(f[1])
      if (length(grep('Status: .*\\d+ ERROR', s1)) || identical(s1, status_line(f[2]))) {
        unlink(d2(d), recursive = TRUE); next
      }
    }
    res = c(
      res, paste('##', p <- sans_ext(d)), '',
      sprintf('[CRAN version](https://cran.rstudio.com/package=%s) (-) vs current version (+):\n', p),
      '```diff', file_diff(f), '```', ''
    )
    if (length(res2 <- cran_check_page(p, NULL))) res = c(
      res, 'CRAN check logs:\n\n```', head_tail(unique(unlist(strsplit(res2, '\n')))), '```\n'
    )
  }
  if (length(res) == 0) return()
  xfun::write_utf8(res, output)
  if (!loadable('markdown')) return(output)
  markdown::markdownToHTML(
    text = gsub('>', '+', gsub('^<', '-', res)),
    output = html_file <- with_ext(output, 'html'),
    header = c(
      "<link href='https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@9.17.1/build/styles/github.min.css' rel='stylesheet' type='text/css' />",
      "<script src='https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@9.17.1/build/highlight.min.js'></script>",
      "<script>hljs.initHighlightingOnLoad();</script>"
    ), encoding = 'UTF-8'
  )
  if (!getOption('xfun.rev_check.keep_md', FALSE)) unlink(output)
  html_file
}

# keep the first and last n elements in x, and omit the middle
head_tail = function(x, n = 10) {
  if (length(x) <= 2 * n) return(x)
  c(head(x, n), '....', tail(x, n))
}

# compute the diffs of two files; if diffs too large, dedup them
file_diff = function(files, len = 200, use_diff = Sys.which('diff') != '') {
  n = length(files)
  if (n == 0) return()
  if (n == 1) {
    f = tempfile(); on.exit(unlink(f), add = TRUE); file.create(f)
    files = c(f, files)
  }
  d = if (use_diff) {
    suppressWarnings(system2('diff', shQuote(files), stdout = TRUE))
  } else {
    c(paste('<', read_utf8(files[1])), '---', paste('>', read_utf8(files[2])))
  }
  if (length(d) >= len) unique(d) else d
}

# specify a list of package names to be ignored when installing all dependencies
ignore_deps = function() {
  if (file_exists('00ignore_deps')) scan('00ignore_deps', 'character')
}

# download a check summary of a package from CRAN
cran_check_page = function(pkg, con = '00check-cran.log') {
  u = sprintf('https://cran.rstudio.com/web/checks/check_results_%s.html', pkg)
  x = read_utf8(u)
  if (length(i <- grep('Check Details', x, ignore.case = TRUE)) == 0) return()
  x = x[i[1]:length(x)]
  x = gsub('<[^>]+>', '', x)
  x = gsub('&nbsp;', ' ', x)
  x = gsub('&gt;', '>', x)
  x = gsub('&lt;', '<', x)
  x = gsub('\\s+', ' ', x)
  x = paste(trimws(x), collapse = '\n')
  x = gsub('\n\n+', '\n\n', x)
  if (length(con) == 1) writeLines(x, con) else x
}

# download CRAN check summaries of all failed packages
cran_check_pages = function() {
  dirs = list.files('.', '[.]Rcheck$')
  for (d in dirs) {
    if (dir_exists(d)) in_dir(d, cran_check_page(gsub('[.]Rcheck$', '', d)))
  }
}

# parse the check log for missing LaTeX packages and install them
find_missing_latex = function() {
  dirs = list.files('.', '[.]Rcheck2?$')
  pkgs = NULL
  for (d in dirs) {
    if (dir_exists(d)) pkgs = c(pkgs, in_dir(
      d, tinytex::parse_packages('00check.log', quiet = c(TRUE, FALSE, FALSE))
    ))
  }
  pkgs = unique(pkgs)
  if (file_exists(f <- 'latex.txt')) append_unique(pkgs, f)
  pkgs
}

#' Test if a character vector consists of blank strings
#'
#' Return \code{TRUE} if all elements of a character vector are blank (white
#' spaces or empty strings).
#' @param x A character vector.
#' @return \code{TRUE} if the input is blank, or \code{FALSE} otherwise.
#' @export
#' @examples
#' xfun::is_blank('')
#' xfun::is_blank('abc')
#' xfun::is_blank(c('', '  ', '\n\t'))
#' xfun::is_blank(c('', ' ', 'abc'))
is_blank = function(x) {
  all(grepl('^\\s*$', x))
}

#' Convert numbers to English words
#'
#' This can be helpful when writing reports with \pkg{knitr}/\pkg{rmarkdown} if
#' we want to print numbers as English words in the output. The function
#' \code{n2w()} is an alias of \code{numbers_to_words()}.
#' @param x A numeric vector. Values should be integers. The absolute values
#'   should be less than \code{1e15}.
#' @param cap Whether to capitalize the first letter of the word. This can be
#'   useful when the word is at the beginning of a sentence. Default is
#'   \code{FALSE}.
#' @param hyphen Whether to insert hyphen (-) when the number is between 21 and
#'   99 (except 30, 40, etc.).
#' @param and Whether to insert \code{and} between hundreds and tens, e.g.,
#'   write 110 as \dQuote{one hundred and ten} if \code{TRUE} instead of
#'   \dQuote{one hundred ten}.
#' @return A character vector.
#' @author Daijiang Li
#' @export
#' @examples library(xfun)
#' n2w(0, cap = TRUE)
#' n2w(0:121, and = TRUE)
#' n2w(1e6)
#' n2w(1e11+12345678)
#' n2w(-987654321)
#' n2w(1e15-1)
numbers_to_words = function(x, cap = FALSE, hyphen = TRUE, and = FALSE) {

  if (!is.numeric(x)) stop('The input is not numeric.')
  if (any(abs(x) >= 1e15)) stop('The absolute value must be less than 1e15.')
  opts = options(scipen = 15); on.exit(options(opts), add = TRUE)  # avoid scientific notation
  if (any(x != floor(x))) stop('The numbers must be integer. ')

  zero_to_19 = c(
    'zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten',
    'eleven', 'twelve', paste0(c('thir', 'four', 'fif', 'six', 'seven', 'eigh', 'nine'), 'teen')
  )
  names(zero_to_19) = as.character(0:19)
  tens = c('twenty', 'thirty', 'forty', 'fifty', 'sixty', 'seventy', 'eighty', 'ninety')
  names(tens) = as.character(seq(20, 90, 10))
  marks = c('', 'thousand,', 'million,', 'billion,', 'trillion,')

  convert_1 = function(x_c) zero_to_19[x_c]  # 0 - 9

  # 10 - 99
  convert_2 = function(x_c) {
    x_cs = strsplit(x_c, split = '')[[1]]
    if (x_cs[1] == 1) return(zero_to_19[x_c])  # 10 - 19
    if (x_cs[2] == 0) return(tens[x_c])  # 20, 30, 40, ...
    # 21, 22, etc.
    paste(tens[as.integer(x_cs[1]) - 1], convert_1(x_cs[2]), sep = if (hyphen) '-' else ' ')
  }

  # 100 - 999
  convert_3 = function(x_c) {
    x_cs = strsplit(x_c, split = '')[[1]]
    n_hundreds = paste(convert_1(x_cs[1]), 'hundred', sep = ' ')
    out = if (x_cs[2] == '0') {
      if (x_cs[3] == '0') return(n_hundreds)  # x00
      convert_1(x_cs[3])  # x0x
    } else {
      convert_2(paste(x_cs[2:3], collapse = ''))  # xxx
    }
    paste(n_hundreds, out, sep = if (and) ' and ' else ' ')
  }

  convert_le3 = function(x_c) {
    x_c = gsub('^0+', '', x_c) # avoid something like 000, 001, 010; but also remove 0
    n = nchar(x_c)
    if (n == 0) return('')
    if (n == 1) return(convert_1(x_c))
    if (n == 2) return(convert_2(x_c))
    if (n == 3) return(convert_3(x_c))
  }

  convert_one = function(x) {
    minus = if (x >= 0) '' else {
      x = abs(x); 'minus '
    }
    if (x == 0) {
      out = 'zero'  # because convert_le3 removed all 0s
    } else {
      x_marks = strsplit(format(x, big.mark = ','), split = ',')[[1]]  # e.g. 123,456,789
      out = vapply(x_marks, convert_le3, character(1))  # group by 3 digits
      x_marks2 = marks[length(x_marks):1]  # units?
      x_marks2[which(out == '')] = ''  # e.g. 4,000,123, 000, remove millions
      out = paste(out, x_marks2, sep = ' ', collapse = ' ')  # zip together
    }
    out = paste0(minus, out)
    out = gsub('^ *|,? *$', '', out)  # trim heading/trailing space
    out = gsub(' {2,}', ' ', out)  # remove multiple spaces
    if (cap) out = sub('^([a-z])', '\\U\\1', out, perl = TRUE)
    out
  }

  if (length(x) > 1) vapply(x, convert_one, character(1)) else convert_one(x)
}

#' @export
#' @rdname numbers_to_words
n2w = numbers_to_words

# create a URL query string from named parameters
query_params = function(..., .list = list()) {
  x = if (missing(.list)) list(...) else .list
  x = paste(names(x), x, sep = '=', collapse = '&')
  if (x != '') paste0('?', x) else x
}

#' Split a character vector by line breaks
#'
#' Call \code{unlist(strsplit(x, '\n'))} on the character vector \code{x} and
#' make sure it works in a few edge cases: \code{split_lines('')} returns
#' \code{''} instead of \code{character(0)} (which is the returned value of
#' \code{strsplit('', '\n')}); \code{split_lines('a\n')} returns \code{c('a',
#' '')} instead of \code{c('a')} (which is the returned value of
#' \code{strsplit('a\n', '\n')}.
#' @param x A character vector.
#' @return All elements of the character vector are split by \code{'\n'} into
#'   lines.
#' @export
#' @examples xfun::split_lines(c('a', 'b\nc'))
split_lines = function(x) {
  if (length(grep('\n', x)) == 0L) return(x)
  x = gsub('\n$', '\n\n', x)
  x[x == ''] = '\n'
  unlist(strsplit(x, '\n'))
}

#' Split source lines into complete expressions
#'
#' Parse the lines of code one by one to find complete expressions in the code,
#' and put them in a list.
#' @param x A character vector of R source code.
#' @return A list of character vectors, and each vector contains a complete R
#'   expression.
#' @export
#' @examples xfun::split_source(c('if (TRUE) {', '1 + 1', '}', 'print(1:5)'))
split_source = function(x) {
  if ((n <- length(x)) < 1) return(list(x))
  i = i1 = i2 = 1
  res = list()
  while (i2 <= n) {
    piece = x[i1:i2]
    if (valid_syntax(piece)) {
      res[[i]] = piece; i = i + 1
      i1 = i2 + 1 # start from the next line
    }
    i2 = i2 + 1
  }
  if (i1 <= n) parse(text = piece)  # must be an error there
  res
}

#' Check if the syntax of the code is valid
#'
#' Try to \code{\link{parse}()} the code and see if an error occurs.
#' @param code A character vector of R source code.
#' @param silent Whether to suppress the error message when the code is not
#'   valid.
#' @return \code{TRUE} if the code could be parsed, otherwise \code{FALSE}.
#' @export
#' @examples xfun::valid_syntax('1+1')
#' xfun::valid_syntax('1+')
#' xfun::valid_syntax(c('if(T){1+1}', 'else {2+2}'), silent = FALSE)
valid_syntax = function(code, silent = TRUE) {
  !inherits(try(parse_only(code), silent = silent), 'try-error')
}

#' Bump version numbers
#'
#' Increase the last digit of version numbers, e.g., from \code{0.1} to
#' \code{0.2}, or \code{7.23.9} to \code{7.23.10}.
#' @param x A vector of version numbers (of the class \code{"numeric_version"}),
#'   or values that can be coerced to version numbers via
#'   \code{as.numeric_version()}.
#' @return A vector of new version numbers.
#' @export
#' @examples xfun::bump_version(c('0.1', '91.2.14'))
bump_version = function(x) {
  x = as.numeric_version(x)
  for (i in seq_along(x)) {
    v = x[i]
    n = length(unclass(v)[[1]])
    v[[1, n]] = v[[1, n]] + 1  # bump the last digit
    x[i] = v
  }
  x
}

#' Fix pairs of characters in a file
#'
#' For example, the curly braces may be wrong (the opening and closing braces
#' are swapped for some reason).
#' @param x A character vector (by default, read from \code{file}).
#' @param file Path to a text file.
#' @param chars A vector of characters of length 2. By default, it is a pair of
#'   curly double quotes.
#' @references \url{https://d.cosx.org/d/420794/5}
#' @noRd
#' @examples
#' files = list.files('.', '[.]R?md$', recursive = TRUE, full.names = TRUE)
#' for (f in files) {
#'   pair_chars(file = f)
#'   # curly single quotes
#'   pair_chars(file = f, chars = c('\U2018', '\U2019'))
#' }
pair_chars = function(x = read_utf8(file), file, chars = c('\U201c', '\U201d')) {
  if (length(chars) != 2) stop("'chars' must be of length 2 (i.e., a pair of characters)")
  is_file = !missing(file)
  r = paste(c('[', chars, ']'), collapse = '')
  k = gregexpr(r, x)
  m = regmatches(x, k)
  for (i in seq_along(m)) {
    n = length(m[[i]])
    if (n %% 2 != 0) {
      warning(
        'The characters do not appear in pairs in the text (',
        'line: ', i, if (is_file) c('; file: ', file), '):\n', x[i], '\n'
      )
      next
    }
    m[[i]] = rep(chars, length.out = n)
  }
  x2 = x
  regmatches(x, k) = m
  if (is_file) {
    if (!identical(x, x2)) xfun::write_utf8(x, file)
    invisible(x)
  } else x
}

#' Generate ID strings
#'
#' Substitute certain (by default, non-alphanumeric) characters with dashes and
#' remove extra dashes at both ends to generate ID strings.
#' @param x A character vector.
#' @param exclude A (Perl) regular expression to detect characters to be
#'   replaced by dashes. By default, non-alphanumeric characters are replaced.
#' @return A character vector of IDs.
#' @export
#' @examples
#' x = c('Hello world 123!', 'a  &b*^##c 456')
#' xfun::alnum_id(x)
#' xfun::alnum_id(x, '[^[:alpha:]]+')  # only keep alphabetical chars
alnum_id = function(x, exclude = '[^[:alnum:]]+') {
  tolower(gsub('^-+|-+$', '', gsub(exclude, '-', x, perl = TRUE)))
}

#' Test for types of operating systems
#'
#' Functions based on \code{.Platform$OS.type} and \code{Sys.info()} to test if
#' the current operating system is Windows, macOS, Unix, or Linux.
#' @rdname os
#' @export
#' @examples
#' library(xfun)
#' # only one of the following statements should be true
#' is_windows()
#' is_unix() && is_macos()
#' is_linux()
is_windows = function() .Platform$OS.type == 'windows'

#' @rdname os
#' @export
is_unix = function() .Platform$OS.type == 'unix'

#' @rdname os
#' @export
is_macos = function() unname(Sys.info()['sysname'] == 'Darwin')

#' @rdname os
#' @export
is_linux = function() unname(Sys.info()['sysname'] == 'Linux')

# This script is executed via the command line `Rscript call-fun.R arg1 arg2`,
# where arg1 is a path to an .rds file, which contains the function and its
# arguments saved as a list, and arg2 is a path to an .rds file to which the
# returned value of the function call is saved.

local({
  if (length(a <- commandArgs(TRUE)) != 2)
    stop('The number of arguments passed to Rscript should be 2.')
  x = readRDS(a[1])  # list(fun, args)
  f = x[[1]]
  if (is.character(f)) f = eval(parse(text = f), envir = globalenv())
  r = do.call(f, x[[2]], envir = globalenv())
  saveRDS(r, a[2])
})

#' Cache the value of an R expression to an RDS file
#'
#' Save the value of an expression to a cache file (of the RDS format). Next
#' time the value is loaded from the file if it exists.
#'
#' Note that the \code{file} argument does not provide the full cache filename.
#' The actual name of the cache file is of the form \file{BASENAME_HASH.rds},
#' where \file{BASENAME} is the base name provided via the \file{file} argument
#' (e.g., if \code{file = 'foo.rds'}, \code{BASENAME} would be \file{foo}), and
#' \file{HASH} is the MD5 hash (also called the \sQuote{checksum}) calculated
#' from the R code provided to the \code{expr} argument and the value of the
#' \code{hash} argument, which means when the code or the \code{hash} argument
#' changes, the \file{HASH} string may also change, and the old cache will be
#' invalidated (if it exists). If you want to find the cache file, look for
#' \file{.rds} files that contain 32 hexadecimal digits (consisting of 0-9 and
#' a-z) at the end of the filename.
#'
#' The possible ways to invalidate the cache are: 1) change the code in
#' \code{expr} argument; 2) delete the cache file manually or automatically
#' through the argument \code{rerun = TRUE}; and 3) change the value of the
#' \code{hash} argument. The first two ways should be obvious. For the third
#' way, it makes it possible to automatically invalidate the cache based on
#' changes in certain R objects. For example, when you run \code{cache_rds({ x +
#' y })}, you may want to invalidate the cache to rerun \code{{ x + y }} when
#' the value of \code{x} or \code{y} has been changed, and you can tell
#' \code{cache_rds()} to do so by \code{cache_rds({ x + y }, hash = list(x,
#' y))}. The value of the argument \code{hash} is expected to be a list, but it
#' can also take a special value, \code{"auto"}, which means
#' \code{cache_rds(expr)} will try to automatically figure out the global
#' variables in \code{expr}, return a list of their values, and use this list as
#' the actual value of \code{hash}. This behavior is most likely to be what you
#' really want: if the code in \code{expr} uses an external global variable, you
#' may want to invalidate the cache if the value of the global variable has
#' changed. Here a \dQuote{global variable} means a variable not created locally
#' in \code{expr}, e.g., for \code{cache_rds({ x <- 1; x + y })}, \code{x} is a
#' local variable, and \code{y} is (most likely to be) a global variable, so
#' changes in \code{y} should invalidate the cache. However, you know your own
#' code the best. If you want to be completely sure when to invalidate the
#' cache, you can always provide a list of objects explicitly rather than
#' relying on \code{hash = "auto"}.
#'
#' By default (the argument \code{clean = TRUE}), old cache files will be
#' automatically cleaned up. Sometimes you may want to use \code{clean = FALSE}
#' (set the R global option \code{options(xfun.cache_rds.clean = FALSE)} if you
#' want \code{FALSE} to be the default). For example, you may not have decided
#' which version of code to use, and you can keep the cache of both versions
#' with \code{clean = FALSE}, so when you switch between the two versions of
#' code, it will still be fast to run the code.
#' @param expr An R expression.
#' @param rerun Whether to delete the RDS file, rerun the expression, and save
#'   the result again (i.e., invalidate the cache if it exists).
#' @param file The \emph{base} (see Details) cache filename under the directory
#'   specified by the \code{dir} argument. If not specified and this function is
#'   called inside a code chunk of a \pkg{knitr} document (e.g., an R Markdown
#'   document), the default is the current chunk label plus the extension
#'   \file{.rds}.
#' @param dir The path of the RDS file is partially determined by
#'   \code{paste0(dir, file)}. If not specified and the \pkg{knitr} package is
#'   available, the default value of \code{dir} is the \pkg{knitr} chunk option
#'   \code{cache.path} (so if you are compiling a \pkg{knitr} document, you do
#'   not need to provide this \code{dir} argument explicitly), otherwise the
#'   default is \file{cache/}. If you do not want to provide a \code{dir} but
#'   simply a valid path to the \code{file} argument, you may use \code{dir =
#'   ""}.
#' @param hash A \code{list} object that contributes to the MD5 hash of the
#'   cache filename (see Details). It can also take a special character value
#'   \code{"auto"}. Other types of objects are ignored.
#' @param clean Whether to clean up the old cache files automatically when
#'   \code{expr} has changed.
#' @param ... Other arguments to be passed to \code{\link{saveRDS}()}.
#' @note Changes in the code in the \code{expr} argument do not necessarily
#'   always invalidate the cache, if the changed code is \code{\link{parse}d} to
#'   the same expression as the previous version of the code. For example, if
#'   you have run \code{cache_rds({Sys.sleep(5);1+1})} before, running
#'   \code{cache_rds({ Sys.sleep( 5 ) ; 1 + 1 })} will use the cache, because
#'   the two expressions are essentially the same (they only differ in white
#'   spaces). Usually you can add/delete white spaces or comments to your code
#'   in \code{expr} without invalidating the cache. See the package vignette
#'   \code{vignette('xfun', package = 'xfun')} for more examples.
#'
#'   When this function is called in a code chunk of a \pkg{knitr} document, you
#'   may not want to provide the filename or directory of the cache file,
#'   because they have reasonable defaults.
#'
#'   Side-effects (such as plots or printed output) will not be cached. The
#'   cache only stores the last value of the expression in \code{expr}.
#' @return If the cache file does not exist, run the expression and save the
#'   result to the file, otherwise read the cache file and return the value.
#' @export
#' @examples
#' f = tempfile()  # the cache file
#' compute = function(...) {
#'   res = xfun::cache_rds({
#'     Sys.sleep(1)
#'     1:10
#'   }, file = f, dir = '', ...)
#'   res
#' }
#' compute()  # takes one second
#' compute()  # returns 1:10 immediately
#' compute()  # fast again
#' compute(rerun = TRUE)  # one second to rerun
#' compute()
#' file.remove(f)
cache_rds = function(
  expr = {}, rerun = FALSE, file = 'cache.rds', dir = 'cache/',
  hash = NULL, clean = getOption('xfun.cache_rds.clean', TRUE), ...
) {
  if (loadable('knitr')) {
    if (missing(file) && !is.null(lab <- knitr::opts_current$get('label')))
      file = paste0(lab, '.rds')
    if (missing(dir) && !is.null(d <- knitr::opts_current$get('cache.path')))
      dir = d
  }
  path = paste0(dir, file)
  if (!grepl(r <- '([.]rds)$', path)) path = paste0(path, '.rds')
  code = deparse(substitute(expr))
  md5  = md5sum_obj(code)
  if (identical(hash, 'auto')) hash = global_vars(code, parent.frame(2))
  if (is.list(hash)) md5 = md5sum_obj(c(md5, md5sum_obj(hash)))
  path = sub(r, paste0('_', md5, '\\1'), path)
  if (rerun) unlink(path)
  if (clean) clean_cache(path)
  if (file_exists(path)) readRDS(path) else {
    obj = expr  # lazy evaluation
    dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
    saveRDS(obj, path, ...)
    obj
  }
}

# write an object to a file and return the md5 sum
md5sum_obj = function(x) {
  f = tempfile(); on.exit(unlink(f), add = TRUE)
  if (is.character(x)) writeLines(x, f) else saveRDS(x, f)
  tools::md5sum(f)
}

# clean up old cache files (those with the same base names as the new cache
# file, e.g., if the new file is FOO_0123abc...z.rds, then FOO_9876def...x.rds
# should be deleted)
clean_cache = function(path) {
  olds = list.files(dirname(path), '_[0-9a-f]{32}[.]rds$', full.names = TRUE)
  olds = c(olds, path)  # `path` may not exist; make sure it is in target paths
  base = basename(olds)
  keep = basename(path) == base  # keep this file (will cache to this file)
  base = substr(base, 1, nchar(base) - 37)  # 37 = 1 (_) + 32 (md5 sum) + 4 (.rds)
  unlink(olds[(base == base[keep][1]) & !keep])
}

# analyze code and find out global variables
find_globals = function(code) {
  fun = eval(parse_only(c('function(){', code, '}')))
  setdiff(codetools::findGlobals(fun), known_globals)
}

known_globals = c(
  '{', '[', '(', ':', '<-', '=', '+', '-', '*', '/', '%%', '%/%', '%*%', '%o%', '%in%'
)

# return a list of values of global variables in code
global_vars = function(code, env) {
  if (length(vars <- find_globals(code)) > 0) mget(vars, env)
}

# add a border to an image via ImageMagick
add_border = function(input, pixels = 1, color = 'black', output) {
  input = normalizePath(input)
  if (missing(output))
    output = paste0(sans_ext(input), '-output.', file_ext(input))
  system2('convert', shQuote(c(
    input, '-shave', paste(pixels, pixels, sep = 'x'), '-bordercolor', color,
    '-border', pixels, output)
  ))
  optipng(dirname(output))
}

#' Use the Tinify API to compress PNG and JPEG images
#'
#' Compress PNG/JPEG images with \samp{api.tinify.com}, and download the
#' compressed images. These functions require R packages \pkg{curl} and
#' \pkg{jsonlite}. \code{tinify_dir()} is a wrapper function of \code{tinify()}
#' to compress images under a directory.
#'
#' You are recommended to set the API key in \file{.Rprofile} or
#' \file{.Renviron}. After that, the only required argument of this function is
#' \code{input}. If the original images can be overwritten by the compressed
#' images, you may either use \code{output = identity}, or set the value of the
#' \code{history} argument in \file{.Rprofile} or \file{.Renviron}.
#' @param input A vector of input paths of images.
#' @param output A vector of output paths or a function that takes \code{input}
#'   and returns a vector of output paths (e.g., \code{output = \link{identity}}
#'   means \code{output = input}). By default, if the \code{history} argument is
#'   not a provided, \code{output} is \code{input} with a suffix \code{-min}
#'   (e.g., when \code{input = 'foo.png'}, \code{output = 'foo-min.png'}),
#'   otherwise \code{output} is the same as \code{input}, which means the
#'   original image files will be overwritten.
#' @param quiet Whether to suppress detailed information about the compression,
#'   which is of the form \samp{input.png (10 Kb) ==> output.png (5 Kb, 50\%);
#'   compression count: 42}. The percentage after \code{output.png} stands for
#'   the compression ratio, and the compression count shows the number of
#'   compressions used for the current month.
#' @param force Whether to compress an image again when it appears to have been
#'   compressed before. This argument only makes sense when the \code{history}
#'   argument is provided.
#' @param key The Tinify API key. It can be set via either the global option
#'   \code{xfun.tinify.key} (you may set it in \file{~/.Rprofile}) or the
#'   environment variable \code{R_XFUN_TINIFY_KEY} (you may set it in
#'   \file{~/.Renviron}).
#' @param history Path to a history file to record the MD5 checksum of
#'   compressed images. If the checksum of an expected output image exists in
#'   this file and \code{force = FALSE}, the compression will be skipped. This
#'   can help you avoid unnecessary API calls.
#' @return The output file paths.
#' @references Tinify API: \url{https://tinypng.com/developers}.
#' @seealso The \pkg{tinieR} package (\url{https://github.com/jmablog/tinieR/})
#'   is a more comprehensive implementation of the Tinify API, whereas
#'   \code{xfun::tinify()} has only implemented the feature of shrinking images.
#' @export
#' @examplesIf interactive()
#' f = xfun:::R_logo('jpg$')
#' xfun::tinify(f)  # remember to set the API key before trying this
tinify = function(
  input, output, quiet = FALSE, force = FALSE,
  key = getOption('xfun.tinify.key', Sys.getenv('R_XFUN_TINIFY_KEY')),
  history = getOption('xfun.tinify.history', Sys.getenv('R_XFUN_TINIFY_HISTORY'))
) {
  if (!(is.character(key) && length(key) == 1 && key != '')) stop(
    "The value of the 'key' argument must be a single non-empty character string."
  )
  if (length(input) == 0) return(invisible(input))
  if (any(i <- !file_exists(input))) stop(
    'Input file(s) not found: ', paste(input[i], collapse = ', ')
  )
  if (missing(output)) {
    output = if (is.character(history)) input else {
      paste0(sans_ext(input), '-min.', file_ext(input))
    }
  } else if (is.function(output)) output = output(input)

  # avoid optimizing the input image if its md5 checksum exists in history
  save_history = function(file) {
    if (!is.character(history) || history == '') return()
    dir_create(dirname(history))
    cat(paste0(tools::md5sum(file), '\n'), file = history, append = TRUE)
  }
  test_history = function(file) {
    is.character(history) && all(file_exists(c(history, file))) &&
      (tools::md5sum(file) %in% readLines(history))
  }

  auth = paste('Authorization: Basic', base64_encode(charToRaw(paste0('api:', key))))

  mapply(input, output, FUN = function(i, o) {
    if (!force && test_history(o)) {
      if (!quiet) message(
        'The image ', o, ' has been compressed before. ',
        'To compress it again, call tinify() with force = TRUE.'
      )
      return()
    }
    if (grepl('[.]png$', i, ignore.case = TRUE))
      optipng(files = i, stdout = if (quiet) FALSE else '')
    res = curl::curl_upload(i, 'https://api.tinify.com/shrink', httpheader = auth, verbose = FALSE)
    cnt = curl::parse_headers_list(res$headers)[['compression-count']]
    res = jsonlite::fromJSON(rawToChar(res$content))
    if (!is.character(u <- res$output$url)) stop2(
      "Failed to shrink '", i, "'", sprintf(': %s (%s)', res$error, res$message)
    )
    if (!quiet) message(sprintf(
      '%s (%s) ==> %s (%s, %.01f%%); compression count: %s',
      i, format_bytes(res$input$size), o, format_bytes(res$output$size),
      res$output$ratio * 100, if (length(cnt)) cnt else NA
    ))
    # back up the original image and restore it if download failed
    if (i == o) {
      b = paste0(i, '~')
      file.rename(i, b)
      on.exit(if (file_exists(o)) file.remove(b) else file.rename(b, i), add = TRUE)
    }
    curl::curl_download(u, o)
    save_history(o)
  })

  invisible(output)
}

#' @param dir A directory under which all \file{.png}, \file{.jpeg}, and
#'   \file{.webp} files are to be compressed.
#' @param ... Arguments passed to \code{\link{tinify}()}.
#' @rdname tinify
#' @export
tinify_dir = function(dir = '.', ...) {
  tinify(all_files('[.](png|jpe?g|webp)$', dir), ...)
}

#' Shrink images to a maximum width
#'
#' Use \code{\link[magick:image_resize]{magick::image_resize}()} to shrink an
#' image if its width is larger than the value specified by the argument
#' \code{width}, and optionally call \code{\link{tinify}()} to compress it.
#' @param width The desired maximum width of images.
#' @param dir The directory of images.
#' @param files A vector of image file paths. By default, this is all
#'   \file{.png}, \file{.jpeg}, and \file{.webp} images under \code{dir}.
#' @param tinify Whether to compress images using \code{\link{tinify}()}.
#' @export
#' @examples
#' f = xfun:::all_files('[.](png|jpe?g)$', R.home('doc'))
#' file.copy(f, tempdir())
#' f = file.path(tempdir(), basename(f))
#' magick::image_info(magick::image_read(f))  # some widths are larger than 300
#' xfun::shrink_images(300, files = f)
#' magick::image_info(magick::image_read(f))  # all widths <= 300 now
#' file.remove(f)
shrink_images = function(
  width = 800, dir = '.', files = all_files('[.](png|jpe?g|webp)$', dir),
  tinify = FALSE
) {
  for (f in files) {
    x = magick::image_read(f)
    if (magick::image_info(x)$width <= width) next
    x = magick::image_resize(x, sprintf('%dx', width))
    magick::image_write(x, f)
  }
  if (tinify) tinify(files, identity)
}

stop2 = function(...) stop(..., call. = FALSE)

warning2 = function(...) warning(..., call. = FALSE)

#' Obtain an attribute of an object without partial matching
#'
#' An abbreviation of \code{base::\link[base]{attr}(exact = TRUE)}.
#' @param ... Passed to \code{base::\link[base]{attr}()} (without the
#'   \code{exact} argument).
#' @export
#' @examples
#' z = structure(list(a = 1), foo = 2)
#' base::attr(z, 'f')  # 2
#' xfun::attr(z, 'f')  # NULL
#' xfun::attr(z, 'foo')  # 2
attr = function(...) base::attr(..., exact = TRUE)

#' Set environment variables
#'
#' Set environment variables from a named character vector, and return the old
#' values of the variables, so they could be restored later.
#'
#' The motivation of this function is that \code{\link{Sys.setenv}()} does not
#' return the old values of the environment variables, so it is not
#' straightforward to restore the variables later.
#' @param vars A named character vector of the form \code{c(VARIABLE = VALUE)}.
#'   If any value is \code{NA}, this function will try to unset the variable.
#' @return Old values of the variables (if not set, \code{NA}).
#' @export
#' @examples
#' vars = xfun::set_envvar(c(FOO = '1234'))
#' Sys.getenv('FOO')
#' xfun::set_envvar(vars)
#' Sys.getenv('FOO')
set_envvar = function(vars) {
  if (is.null(nms <- names(vars)) || any(nms == '')) stop(
    "The 'vars' argument must take a named character vector."
  )
  vals = Sys.getenv(nms, NA, names = TRUE)
  i = is.na(vars)
  suppressWarnings(Sys.unsetenv(nms[i]))
  if (length(vars <- vars[!i])) do.call(Sys.setenv, as.list(vars))
  invisible(vals)
}

#' Call \code{on.exit()} in a parent function
#'
#' The function \code{\link{on.exit}()} is often used to perform tasks when the
#' current function exits. This \code{exit_call()} function allows calling a
#' function when a parent function exits (thinking of it as inserting an
#' \code{on.exit()} call into the parent function).
#' @param fun A function to be called when the parent function exits.
#' @param n The parent frame number. For \code{n = 1}, \code{exit_call(fun)} is
#'   the same as \code{on.exit(fun())}; \code{n = 2} means adding
#'   \code{on.exit(fun())} in the parent function; \code{n = 3} means the
#'   grandparent, etc.
#' @param ... Other arguments to be passed to \code{on.exit()}.
#' @references This function was inspired by Kevin Ushey:
#'   \url{https://yihui.org/en/2017/12/on-exit-parent/}
#' @export
#' @examples
#' f = function(x) {
#'   print(x)
#'   xfun::exit_call(function() print('The parent function is exiting!'))
#' }
#' g = function(y) {
#'   f(y)
#'   print('f() has been called!')
#' }
#' g('An argument of g()!')
exit_call = function(fun, n = 2, ...) {
  do.call(
    on.exit, list(substitute(fun(), list(fun = fun)), add = TRUE, ...),
    envir = parent.frame(n)
  )
}

#' Evaluate an expression under a specified working directory
#'
#' Change the working directory, evaluate the expression, and restore the
#' working directory.
#' @param dir Path to a directory.
#' @param expr An R expression.
#' @export
#' @examples
#' library(xfun)
#' in_dir(tempdir(), {print(getwd()); list.files()})
in_dir = function(dir, expr) {
  owd = setwd(dir); on.exit(setwd(owd))
  expr
}

#' Test if an object is identical to \code{FALSE}
#'
#' A simple abbreviation of \code{identical(x, FALSE)}.
#' @param x An R object.
#' @export
#' @examples
#' library(xfun)
#' isFALSE(TRUE)  # false
#' isFALSE(FALSE)  # true
#' isFALSE(c(FALSE, FALSE))  # false
isFALSE = function(x) identical(x, FALSE)

#' Parse R code and do not keep the source
#'
#' An abbreviation of \code{parse(keep.source = FALSE)}.
#' @param code A character vector of the R source code.
#' @export
#' @return R \code{\link{expression}}s.
#' @examples library(xfun)
#' parse_only('1+1'); parse_only(c('y~x', '1:5 # a comment'))
#' parse_only(character(0))
parse_only = function(code) {
  if (length(code) == 0) return(expression())
  parse(text = code, keep.source = FALSE)
}

#' Try to evaluate an expression silently
#'
#' An abbreviation of \code{try(silent = TRUE)}.
#' @param expr An R expression.
#' @export
#' @examples library(xfun)
#' z = try_silent(stop('Wrong!'))
#' inherits(z, 'try-error')
try_silent = function(expr) try(expr, silent = TRUE)

#' Try an expression and see if it throws an error
#'
#' Use \code{\link{tryCatch}()} to check if an expression throws an error.
#' @inheritParams try_silent
#' @return \code{TRUE} (error) or \code{FALSE} (success).
#' @export
#' @examples
#' xfun::try_error(stop('foo'))  # TRUE
#' xfun::try_error(1:10)  # FALSE
try_error = function(expr) {
  err = FALSE
  tryCatch(expr, error = function(e) err <<- TRUE)
  err
}

#' Retry calling a function for a number of times
#'
#' If the function returns an error, retry it for the specified number of
#' times, with a pause between attempts.
#'
#' One application of this function is to download a web resource. Since the
#' download might fail sometimes, you may want to retry it for a few more times.
#' @param fun A function.
#' @param ... Arguments to be passed to the function.
#' @param .times The number of times.
#' @param .pause The number of seconds to wait before the next attempt.
#' @export
#' @examplesIf interactive()
#' # read the Github releases info of the repo yihui/xfun
#' xfun::retry(xfun::github_releases, 'yihui/xfun')
retry = function(fun, ..., .times = 3, .pause = 5) {
  for (i in seq_len(.times)) {
    if (!inherits(res <- tryCatch(fun(...), error = identity), 'error'))
      return(res)
    Sys.sleep(.pause)
  }
  stop(res$message, call. = FALSE)
}

gsubi = function(...) gsub(..., ignore.case = TRUE)

#' Turn the output of \code{\link{str}()} into a tree diagram
#'
#' The super useful function \code{str()} uses \verb{..} to indicate the level
#' of sub-elements of an object, which may be difficult to read. This function
#' uses vertical pipes to connect all sub-elements on the same level, so it is
#' clearer which elements belong to the same parent element in an object with a
#' nested structure (such as a nested list).
#' @param ... Arguments to be passed to \code{\link{str}()} (note that the
#'   \code{comp.str} is hardcoded inside this function, and it is the only
#'   argument that you cannot customize).
#' @return A character string as a \code{\link{raw_string}()}.
#' @export
#' @examples fit = lsfit(1:9, 1:9)
#' str(fit)
#' xfun::tree(fit)
#'
#' fit = lm(dist ~ speed, data = cars)
#' str(fit)
#' xfun::tree(fit)
#'
#' # some trivial examples
#' xfun::tree(1:10)
#' xfun::tree(iris)
tree = function(...) {
  x = capture.output(str(..., comp.str = '$ '))
  r = '^([^$-]+[$-] )(.*)$'
  x1 = gsub(r, '\\1', x)
  x2 = gsub(r, '\\2', x)
  x1 = gsub('[.][.]', '  ', x1)
  x1 = gsub('[$] $', '|-', x1)
  x1 = connect_pipes(x1)
  x3 = paste(x1, x2, sep = '')
  i = !grepl(r, x)
  x3[i] = x[i]
  raw_string(x3)
}

# for a tree diagram, connect the pipes on the same level, e.g., change

# |- ..
#   |- ..
#
#   |- ..

# to

# |- ..
#   |- ..
#   |
#   |- ..

# this task is not complicated, but just boring nested for-loops
connect_pipes = function(x) {
  ns = nchar(x); n = max(ns); m = length(x)
  if (n < 2 || m < 3) return(x)
  A = matrix('', nrow = m, ncol = n)
  x = strsplit(x, '')
  for (i in seq_len(m)) {
    A[i, seq_len(ns[i])] = x[[i]]
  }
  k = NULL
  for (j in seq_len(n - 1)) {
    for (i in seq_len(m - 2)) {
      if (!all(A[i, j + 0:1] == c('|', '-'))) next
      for (l in (i + 1):m) {
        cells = A[l, j + 0:1]
        if (all(cells == ' ')) {
          if (l == m) {
            k = NULL; break
          } else k = c(k, l)
        } else if (all(cells == c('|', '-'))) {
          break
        } else {
          k = NULL; break
        }
      }
      if (length(k) > 0) A[k, j] = '|'
      k = NULL
    }
  }
  apply(A, 1, paste, collapse = '')
}

pkg_file = function(...) system.file(..., package = 'xfun', mustWork = TRUE)

#' Format numbers of bytes using a specified unit
#'
#' Call the S3 method \code{format.object_size()} to format numbers of bytes.
#' @param x A numeric vector (each element represents a number of bytes).
#' @param units,... Passed to \code{\link[=format.object_size]{format}()}.
#' @return A character vector.
#' @export
#' @examples
#' xfun::format_bytes(c(1, 1024, 2000, 1e6, 2e8))
#' xfun::format_bytes(c(1, 1024, 2000, 1e6, 2e8), units = 'KB')
format_bytes = function(x, units = 'auto', ...) {
  vapply(x, function(b) {
    format(structure(b, class = 'object_size'), units = units, ...)
  }, character(1))
}

# get the function name of the parent call
func_name = function(which = 1) {
  x = sys.call(which)[[1]]
  deparse(x)[1]
}

# retrieve the release dates of packages
cran_pkg_dates = function(full = FALSE, maintainer = 'Yihui Xie') {
  info = tools::CRAN_package_db()
  pkgs = info[grep(maintainer, info$Maintainer), 'Package']
  info = setNames(vector('list', length(pkgs)), pkgs)
  for (p in pkgs) {
    message('Processing ', p)
    x = readLines(u <- sprintf('https://cran.rstudio.com/web/packages/%s/', p))
    i = which(x == '<td>Published:</td>')
    if (length(i) == 0) stop('Cannot find the publishing date from ', u)
    d = as.Date(gsub('</?td>', '', x[i[1] + 1]))
    x = try(readLines(u <- sprintf('https://cran.r-project.org/src/contrib/Archive/%s/', p)))
    if (inherits(x, 'try-error')) {
      info[[p]] = d; next
    }
    r = '.+</td><td align="right">(\\d{4,}-\\d{2}-\\d{2}) .+'
    d = c(d, as.Date(gsub(r, '\\1', grep(r, x, value = TRUE))))
    info[[p]] = sort(d, decreasing = TRUE)
  }
  if (full) info else sort(do.call(c, lapply(info, `[`, 1)), decreasing = TRUE)
}

# return packages that haven't been updated for X days, and can be updated on CRAN
cran_updatable = function(days = 90, maintainer = 'Yihui Xie') {
  info = cran_pkg_dates(TRUE, maintainer)
  flag = unlist(lapply(info, function(d) {
    sum(d > Sys.Date() - 180) < 6 && d[1] < Sys.Date() - days
  }))
  names(which(flag))
}


#' Some utility functions for checking packages
#'
#' Miscellaneous utility functions to obtain information about the package
#' checking environment.
#' @export
#' @keywords internal
is_R_CMD_check = function() {
  !is.na(check_package_name()) || tolower(Sys.getenv('_R_CHECK_LICENSE_')) == 'true'
}

#' @rdname is_R_CMD_check
#' @export
is_CRAN_incoming = function() {
  isTRUE(as.logical(Sys.getenv('_R_CHECK_CRAN_INCOMING_REMOTE_')))
}

#' @rdname is_R_CMD_check
#' @export
check_package_name = function() {
  Sys.getenv('_R_CHECK_PACKAGE_NAME_', NA)
}

# is R CMD check running on a package that has a version lower or equal to `version`?
#' @rdname is_R_CMD_check
#' @export
check_old_package = function(name, version) {
  if (is.na(pkg <- check_package_name()) || pkg != name) return(FALSE)
  tryCatch(packageVersion(name) <= version, error = function(e) FALSE)
}

# return package maintainers (with email addresses)
pkg_maintainers = function(pkgs) {
  info = tools::CRAN_package_db()
  info = info[match(pkgs, info$Package), c('Package', 'Maintainer')]
  setNames(info$Maintainer, info$Package)
}

#' Submit a source package to CRAN
#'
#' Build a source package and submit it to CRAN with the \pkg{curl} package.
#' @param file The path to the source package tarball. By default, the current
#'   working directory is treated as the package root directory, and
#'   automatically built into a tarball, which is deleted after submission. This
#'   means you should run \code{xfun::submit_cran()} in the root directory of a
#'   package project, unless you want to pass a path explicitly to the
#'   \code{file} argument.
#' @param comment Submission comments for CRAN. By default, if a file
#'   \file{cran-comments.md} exists, its content will be read and used as the
#'   comment.
#' @seealso \code{devtools::submit_cran()} does the same job, with a few more
#'   dependencies in addition to \pkg{curl} (such as \pkg{cli});
#'   \code{xfun::submit_cran()} only depends on \pkg{curl}.
#' @export
submit_cran = function(file = pkg_build(), comment = '') {
  # if the tarball is automatically created, delete it after submission
  if (missing(file)) on.exit(file.remove(file), add = TRUE)

  # read the maintainer's name/email
  dir_create(d <- tempfile())
  on.exit(unlink(d, recursive = TRUE), add = TRUE)
  desc = file.path(gsub('_.*', '', basename(file)), 'DESCRIPTION')
  untar(file, desc, exdir = d)
  info = read.dcf(file.path(d, desc), fields = 'Maintainer')[1, 1]
  info = unlist(strsplit(info, '( <|>)'))

  # read submission comments from cran-comments.md if exists
  if (missing(comment) && file_exists(f <- 'cran-comments.md')) {
    comment = file_string(f)
  }
  params = list(
    uploaded_file = curl::form_file(file), name = info[1], email = info[2],
    upload = 'Upload package'
  )
  params$comment = if (length(comment)) comment
  server = 'https://xmpalantir.wu.ac.at/cransubmit/index2.php'

  # submit the form
  h = curl::new_handle()
  curl::handle_setform(h, .list = params)
  res = curl::curl_fetch_memory(server, h)

  # find the pkg_id from the response page
  id = grep_sub(
    '(.*<input name="pkg_id" type="hidden" value=")([^"]+)(".*)', '\\2',
    rawToChar(res$content)
  )
  if (length(id) != 1) stop('Failed to submit ', file, ' to CRAN')

  # skip the review and submit directly
  h = curl::new_handle()
  curl::handle_setform(h, .list = list(pkg_id = id, submit = 'Submit package'))
  res = curl::curl_fetch_memory(server, h)
  if (grepl('>Step 3<', rawToChar(res$content))) message(
    'The package has been submitted. Please confirm the submission in email: ',
    params$email
  ) else message('The submission may be unsuccessful.')
}

#' Find the indices of lines in Markdown that are prose (not code blocks)
#'
#' Filter out the indices of lines between code block fences such as \verb{```}
#' (could be three or four or more backticks).
#' @param x A character vector of text in Markdown.
#' @param warn Whether to emit a warning when code fences are not balanced.
#' @note If the code fences are not balanced (e.g., a starting fence without an
#'   ending fence), this function will treat all lines as prose.
#' @return An integer vector of indices of lines that are prose in Markdown.
#' @export
#' @examples library(xfun)
#' prose_index(c('a', '```', 'b', '```', 'c'))
#' prose_index(c('a', '````', '```r', '1+1', '```', '````', 'c'))
prose_index = function(x, warn = TRUE) {
  idx = NULL; r = '^(\\s*```+).*'; s = ''
  for (i in setdiff(grep(r, x), grep('-->\\s*$', x))) {
    if (s == '') {
      s = gsub(r, '\\1', x[i]); idx = c(idx, i); next
    }
    # look for the next line with the same amount of backticks (end of block)
    if (grepl(paste0('^', s), x[i])) {
      idx = c(idx, i); s = ''
    }
  }
  xi = seq_along(x); n = length(idx)
  if (n == 0) return(xi)
  if (n %% 2 != 0) {
    if (warn) warning('Code fences are not balanced')
    # treat all lines as prose
    return(xi)
  }
  idx2 = matrix(idx, nrow = 2)
  idx2 = unlist(mapply(seq, idx2[1, ], idx2[2, ], SIMPLIFY = FALSE))
  xi[-idx2]
}

#' Protect math expressions in pairs of backticks in Markdown
#'
#' For Markdown renderers that do not support LaTeX math, we need to protect
#' math expressions as verbatim code (in a pair of backticks), because some
#' characters in the math expressions may be interpreted as Markdown syntax
#' (e.g., a pair of underscores may make text italic). This function detects
#' math expressions in Markdown (by heuristics), and wrap them in backticks.
#'
#' Expressions in pairs of dollar signs or double dollar signs are treated as
#' math, if there are no spaces after the starting dollar sign, or before the
#' ending dollar sign. There should be spaces before the starting dollar sign,
#' unless the math expression starts from the very beginning of a line. For a
#' pair of single dollar signs, the ending dollar sign should not be followed by
#' a number. With these assumptions, there should not be too many false
#' positives when detecing math expressions.
#'
#' Besides, LaTeX environments (\verb{\begin{*}} and \verb{\end{*}}) are also
#' protected in backticks.
#' @param x A character vector of text in Markdown.
#' @param token A character string to wrap math expressions at both ends. This
#'   can be a unique token so that math expressions can be reliably identified
#'   and restored after the Markdown text is converted.
#' @return A character vector with math expressions in backticks.
#' @note If you are using Pandoc or the \pkg{rmarkdown} package, there is no
#'   need to use this function, because Pandoc's Markdown can recognize math
#'   expressions.
#' @export
#' @examples library(xfun)
#' protect_math(c('hi $a+b$', 'hello $$\\alpha$$', 'no math here: $x is $10 dollars'))
#' protect_math(c('hi $$', '\\begin{equation}', 'x + y = z', '\\end{equation}'))
#' protect_math('$a+b$', '===')
protect_math = function(x, token = '') {
  i = prose_index(x)
  if (length(i)) x[i] = escape_math(x[i], token)
  x
}

escape_math = function(x, token = '') {
  # replace $x$ with `\(x\)` (protect inline math in <code></code>)
  m = gregexpr('(?<=^|[\\s])[$](?! )[^$]+?(?<! )[$](?![$0123456789])', x, perl = TRUE)
  regmatches(x, m) = lapply(regmatches(x, m), function(z) {
    if (length(z) == 0) return(z)
    z = sub('^[$]', paste0('`', token, '\\\\('), z)
    z = sub('[$]$', paste0('\\\\)', token, '`'), z)
    z
  })
  # replace $$x$$ with `$$x$$` (protect display math)
  m = gregexpr('(?<=^|[\\s])[$][$](?! )[^$]+?(?<! )[$][$]', x, perl = TRUE)
  regmatches(x, m) = lapply(regmatches(x, m), function(z) {
    if (length(z) == 0) return(z)
    paste0('`', token, z, token, '`')
  })
  # now, if there are still lines starting and ending with $$, they might be
  # math expressions of display style spanning multiple lines, e.g.,
  # $$\alpha +
  # \beta$$
  # we assume that $$ can only appear once on one line
  i = vapply(gregexpr('[$]', x), length, integer(1)) == 2
  if (any(i)) {
    x[i] = gsub('^([$][$])([^ ]+)', paste0('`', token, '\\1\\2'), x[i], perl = TRUE)
    x[i] = gsub('([^ ])([$][$])$', paste0('\\1\\2', token, '`'), x[i], perl = TRUE)
  }
  # equation environments (\begin and \end must match)
  i1 = grep('^\\\\begin\\{[^}]+\\}$', x)
  i2 = grep('^\\\\end\\{[^}]+\\}$', x)
  if (length(i1) == length(i2)) {
    # TODO: do not protect inner environments in case of nested environments (#57)
    x[i1] = paste0('`', token, x[i1])
    x[i2] = paste0(x[i2], token, '`')
  }
  x
}

#' Embed a file, multiple files, or directory on an HTML page
#'
#' For a file, first encode it into base64 data (a character string). Then
#' generate a hyperlink of the form \samp{<a href="base64 data"
#' download="filename">Download filename</a>}. The file can be downloaded when
#' the link is clicked in modern web browsers. For a directory, it will be
#' compressed as a zip archive first, and the zip file is passed to
#' \code{embed_file()}. For multiple files, they are also compressed to a zip
#' file first.
#'
#' These functions can be called in R code chunks in R Markdown documents with
#' HTML output formats. You may embed an arbitrary file or directory in the HTML
#' output file, so that readers of the HTML page can download it from the
#' browser. A common use case is to embed data files for readers to download.
#' @param path Path to the file(s) or directory.
#' @param name The default filename to use when downloading the file. Note that
#'   for \code{embed_dir()}, only the base name (of the zip filename) will be
#'   used.
#' @param text The text for the hyperlink.
#' @param ... For \code{embed_file()}, additional arguments to be passed to
#'   \code{htmltools::a()} (e.g., \code{class = 'foo'}). For \code{embed_dir()}
#'   and \code{embed_files()}, arguments passed to \code{embed_file()}.
#' @note Windows users may need to install Rtools to obtain the \command{zip}
#'   command to use \code{embed_dir()} and \code{embed_files()}.
#'
#'   These functions require R packages \pkg{mime} and \pkg{htmltools}. If you
#'   have installed the \pkg{rmarkdown} package, these packages should be
#'   available, otherwise you need to install them separately.
#'
#'   Currently Internet Explorer does not support downloading embedded files
#'   (\url{https://caniuse.com/#feat=download}). Chrome has a 2MB limit on the
#'   file size.
#' @return An HTML tag \samp{<a>} with the appropriate attributes.
#' @export
#' @examples
#' logo = xfun:::R_logo()
#' link = xfun::embed_file(logo, text = 'Download R logo')
#' link
#' if (interactive()) htmltools::browsable(link)
embed_file = function(path, name = basename(path), text = paste('Download', name), ...) {
  pkg_require(c('mime', 'htmltools'))
  h = base64_uri(path)
  htmltools::a(text, href = h, download = name, ...)
}

#' @rdname embed_file
#' @export
embed_dir = function(path, name = paste0(normalize_path(path), '.zip'), ...) {
  name  = gsub('/', '', basename(name))
  in_dir(path, {
    name = file.path(tempdir(), name); on.exit(file.remove(name), add = TRUE)
    zip(name, '.'); embed_file(name, ...)
  })
}

#' @rdname embed_file
#' @export
embed_files = function(path, name = with_ext(basename(path[1]), '.zip'), ...) {
  name = file.path(tempdir(), basename(name))
  on.exit(file.remove(name), add = TRUE)
  zip(name, path)
  embed_file(name, ...)
}

zip = function(name, ...) {
  if (utils::zip(name, ...) != 0) stop('Failed to create the zip archive ', name)
  invisible(0)
}

#' Get data from a REST API
#'
#' Read data from a REST API and optionally with an authorization token in the
#' request header. The function \code{rest_api_raw()} returns the raw text of
#' the response, and \code{rest_api()} will parse the response with
#' \code{jsonlite::fromJSON()} (assuming that the response is in the JSON
#' format).
#'
#' These functions are simple wrappers based on \code{\link{url}()} and
#' \code{\link{read_utf8}()}. Specifically, the \code{headers} argument is
#' passed to \code{url()}, and \code{read_utf8()} will send a \samp{GET} request
#' to the API server. This means these functions only support the \samp{GET}
#' method. If you need to use other HTTP methods (such as \samp{POST}), you have
#' to use other packages such as \pkg{curl} and \pkg{httr}.
#' @param ... Arguments to be passed to \code{rest_api_raw()}.
#' @return A character vector (the raw JSON response) or an R object parsed from
#'   the JSON text.
#' @export
#' @examplesIf interactive()
#' # a normal GET request
#' xfun::rest_api('https://httpbin.org', '/get')
#' xfun::rest_api_raw('https://httpbin.org', '/get')
#'
#' # send the request with an auth header
#' xfun::rest_api('https://httpbin.org', '/headers', 'OPEN SESAME!')
#'
#' # with query parameters
#' xfun::rest_api('https://httpbin.org', '/response-headers', params = list(foo = 'bar'))
#'
#' # get the rate limit info from Github
#' xfun::github_api('/rate_limit')
rest_api = function(...) {
  res = rest_api_raw(...)
  jsonlite::fromJSON(res, simplifyVector = FALSE)
}

#' @param root The API root URL.
#' @param endpoint The API endpoint.
#' @param token A named character string (e.g., \code{c(token = "xxxx")}), which
#'   will be used to create an authorization header of the form
#'   \samp{Authorization: NAME TOKEN} for the API call, where \samp{NAME} is the
#'   name of the string and \samp{TOKEN} is the string. If the string does not
#'   have a name, \samp{Basic} will be used as the default name.
#' @param params A list of query parameters to be sent with the API call.
#' @param headers A named character vector of HTTP headers, e.g., \code{c(Accept
#'   = "application/vnd.github.v3+json")}.
#' @rdname rest_api
#' @export
rest_api_raw = function(root, endpoint, token = '', params = list(), headers = NULL)  {
  if (is.null(names(token))) names(token) = 'Basic'
  endpoint = sub('^/?', '/', endpoint)  # make sure it has a leading /
  url2 = if ('headers' %in% names(formals(url))) url else stop2(
    "The url() function does not support the 'headers' argument. Please upgrade R (>= 3.6.0)."
  )
  con = url2(
    paste0(root, endpoint, query_params(.list = params)), encoding = 'UTF-8',
    headers = c(
      headers, if (token != '') c(Authorization = sprintf('%s %s', names(token), token))
    )
  )
  on.exit(close(con), add = TRUE)
  raw_string(suppressWarnings(read_utf8(con)))
}

has_fun = function(name, envir = baseenv()) exists(name, envir, inherits = FALSE)

# for R < 4.0
if (!has_fun('startsWith')) startsWith = function(x, s) {
  substring(x, 1, nchar(s)) == s
}
if (!has_fun('endsWith')) endsWith = function(x, s) {
  n = nchar(x)
  substring(x, n - nchar(s) + 1, n) == s
}
if (!has_fun('strrep')) strrep = function(x, times) {
  mapply(function(x, n) paste(rep(x, n), collapse = ''), x, times, USE.NAMES = FALSE)
}

#' Test if a character vector consists of blank strings
#'
#' Return a logical vector indicating if elements of a character vector are
#' blank (white spaces or empty strings).
#' @param x A character vector.
#' @return `TRUE` for blank elements, or `FALSE` otherwise.
#' @export
#' @examples
#' xfun::is_blank('')
#' xfun::is_blank('abc')
#' xfun::is_blank(c('', '  ', '\n\t'))
#' xfun::is_blank(c('', ' ', 'abc'))
is_blank = function(x) grepl('^\\s*$', x)

#' Convert numbers to English words
#'
#' This can be helpful when writing reports with \pkg{knitr}/\pkg{rmarkdown} if
#' we want to print numbers as English words in the output. The function `n2w()`
#' is an alias of `numbers_to_words()`.
#' @param x A numeric vector. The absolute values should be less than `1e15`.
#' @param cap Whether to capitalize the first letter of the word. This can be
#'   useful when the word is at the beginning of a sentence. Default is `FALSE`.
#' @param hyphen Whether to insert hyphen (-) when the number is between 21 and
#'   99 (except 30, 40, etc.).
#' @param and Whether to insert `and` between hundreds and tens, e.g., write 110
#'   as \dQuote{one hundred and ten} if `TRUE` instead of \dQuote{one hundred
#'   ten}.
#' @return A character vector.
#' @author Daijiang Li
#' @export
#' @examples library(xfun)
#' n2w(0, cap = TRUE)
#' n2w(0:121, and = TRUE)
#' n2w(1e6)
#' n2w(1e11+12345678)
#' n2w(-987654321)
#' n2w(1e15-1)
#' n2w(123.456)
#' n2w(123.45678901)
#' n2w(123.456789098765)
numbers_to_words = function(x, cap = FALSE, hyphen = TRUE, and = FALSE) {
  if (!is.numeric(x)) stop('The input is not numeric.')
  if (any(abs(x) >= 1e15)) stop('The absolute value must be less than 1e15.')
  opts = options(scipen = 15, OutDec = '.')  # avoid scientific notation
  on.exit(options(opts), add = TRUE)

  zero_to_19 = c(
    'zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten',
    'eleven', 'twelve', paste0(c('thir', 'four', 'fif', 'six', 'seven', 'eigh', 'nine'), 'teen')
  )
  names(zero_to_19) = as.character(0:19)
  tens = c('twenty', 'thirty', 'forty', 'fifty', 'sixty', 'seventy', 'eighty', 'ninety')
  names(tens) = as.character(seq(20, 90, 10))
  marks = c('', 'thousand,', 'million,', 'billion,', 'trillion,')

  convert_1 = function(x_c) zero_to_19[x_c]  # 0 - 9

  # 10 - 99
  convert_2 = function(x_c) {
    x_cs = strsplit(x_c, split = '')[[1]]
    if (x_cs[1] == 1) return(zero_to_19[x_c])  # 10 - 19
    if (x_cs[2] == 0) return(tens[x_c])  # 20, 30, 40, ...
    # 21, 22, etc.
    paste(tens[as.integer(x_cs[1]) - 1], convert_1(x_cs[2]), sep = if (hyphen) '-' else ' ')
  }

  # 100 - 999
  convert_3 = function(x_c) {
    x_cs = strsplit(x_c, split = '')[[1]]
    n_hundreds = paste(convert_1(x_cs[1]), 'hundred', sep = ' ')
    out = if (x_cs[2] == '0') {
      if (x_cs[3] == '0') return(n_hundreds)  # x00
      convert_1(x_cs[3])  # x0x
    } else {
      convert_2(paste(x_cs[2:3], collapse = ''))  # xxx
    }
    paste(n_hundreds, out, sep = if (and) ' and ' else ' ')
  }

  convert_le3 = function(x_c) {
    x_c = gsub('^0+', '', x_c) # avoid something like 000, 001, 010; but also remove 0
    n = nchar(x_c)
    if (n == 0) return('')
    if (n == 1) return(convert_1(x_c))
    if (n == 2) return(convert_2(x_c))
    if (n == 3) return(convert_3(x_c))
  }

  convert_one = function(x) {
    minus = if (x >= 0) '' else {
      x = abs(x); 'minus '
    }
    if (x == 0) {
      out = 'zero'  # because convert_le3 removed all 0s
    } else {
      x_marks = strsplit(format(floor(x), big.mark = ','), split = ',')[[1]]  # e.g. 123,456,789
      out = vapply(x_marks, convert_le3, character(1))  # group by 3 digits
      x_marks2 = marks[length(x_marks):1]  # units?
      x_marks2[which(out == '')] = ''  # e.g. 4,000,123, 000, remove millions
      out = paste(out, x_marks2, sep = ' ', collapse = ' ')  # zip together
    }
    out = paste0(minus, out)
    out = gsub('^ *|,? *$', '', out)  # trim heading/trailing space
    out = gsub(' {2,}', ' ', out)  # remove multiple spaces
    if (cap) out = sub('^([a-z])', '\\U\\1', out, perl = TRUE)
    if (x - floor(x) > 0) {
      frac = sub('^[0-9]+[.]', '', as.character(x))
      frac = convert_1(strsplit(frac, '')[[1]])
      out = paste(c(out, 'point', frac), collapse = ' ')
    }
    out
  }

  if (length(x) > 1) vapply(x, convert_one, character(1)) else convert_one(x)
}

#' @export
#' @rdname numbers_to_words
n2w = numbers_to_words

#' Evaluate an expression after forcing the decimal point to be a dot
#'
#' Sometimes it is necessary to use the dot character as the decimal separator.
#' In R, this could be affected by two settings: the global option
#' `options(OutDec)` and the `LC_NUMERIC` locale. This function sets the former
#' to `.` and the latter to `C` before evaluating an expression, such as
#' coercing a number to character.
#' @param x An expression.
#' @export
#' @return The value of `x`.
#' @examples
#' opts = options(OutDec = ',')
#' as.character(1.234)  # using ',' as the decimal separator
#' print(1.234)  # same
#' xfun::decimal_dot(as.character(1.234))  # using dot
#' xfun::decimal_dot(print(1.234))  # using dot
#' options(opts)
decimal_dot = function(x) {
  opts = options(OutDec = '.'); on.exit(options(opts), add = TRUE)
  lcn = Sys.getlocale('LC_NUMERIC')
  if (lcn != 'C') {
    Sys.setlocale('LC_NUMERIC', 'C')
    on.exit(suppressWarnings(Sys.setlocale('LC_NUMERIC', lcn)), add = TRUE)
  }
  x
}

# create a URL query string from named parameters
query_params = function(..., .list = list()) {
  x = if (missing(.list)) list(...) else .list
  x = paste(names(x), x, sep = '=', collapse = '&')
  if (x != '') paste0('?', x) else x
}

#' Split a character vector by line breaks
#'
#' Call `unlist(strsplit(x, '\n'))` on the character vector `x` and
#' make sure it works in a few edge cases: `split_lines('')` returns
#' `''` instead of `character(0)` (which is the returned value of
#' `strsplit('', '\n')`); `split_lines('a\n')` returns `c('a',
#' '')` instead of `c('a')` (which is the returned value of
#' `strsplit('a\n', '\n')`.
#' @param x A character vector.
#' @return All elements of the character vector are split by `'\n'` into
#'   lines.
#' @export
#' @examples xfun::split_lines(c('a', 'b\nc'))
split_lines = function(x) {
  if (length(grep('\n', x)) == 0L) return(x)
  x = gsub('\n$', '\n\n', x)
  x[x == ''] = '\n'
  unlist(strsplit(x, '\r?\n'))
}

#' Split source lines into complete expressions
#'
#' Parse the lines of code one by one to find complete expressions in the code,
#' and put them in a list.
#' @param x A character vector of R source code.
#' @param merge_comments Whether to merge consecutive lines of comments as a
#'   single expression to be combined with the next non-comment expression (if
#'   any).
#' @param line_number Whether to store the line numbers of each expression in
#'   the returned value.
#' @return A list of character vectors, and each vector contains a complete R
#'   expression, with an attribute `lines` indicating the starting and ending
#'   line numbers of the expression if the argument `line_number = TRUE`.
#' @export
#' @examples
#' code = c('# comment 1', '# comment 2', 'if (TRUE) {', '1 + 1', '}', 'print(1:5)')
#' xfun::split_source(code)
#' xfun::split_source(code, merge_comments = TRUE)
split_source = function(x, merge_comments = FALSE, line_number = FALSE) {
  if ((n <- length(x)) < 1) return(list(x))
  i1 = i2 = 1L
  res = list()
  add_source = function(x) {
    res[[length(res) + 1]] <<- if (line_number) structure(x, lines = c(i1, i2)) else x
  }
  while (i2 <= n) {
    piece = x[i1:i2]
    if ((!merge_comments || (!all(grepl('^#', piece)) || i2 == n)) && valid_syntax(piece)) {
      add_source(piece)
      i1 = i2 + 1L # start from the next line
    }
    i2 = i2 + 1L
  }
  if (i1 <= n) parse(text = piece)  # must be an error there
  res
}

#' Check if the syntax of the code is valid
#'
#' Try to [parse()] the code and see if an error occurs.
#' @param code A character vector of R source code.
#' @param silent Whether to suppress the error message when the code is not
#'   valid.
#' @return `TRUE` if the code could be parsed, otherwise `FALSE`.
#' @export
#' @examples xfun::valid_syntax('1+1')
#' xfun::valid_syntax('1+')
#' xfun::valid_syntax(c('if(T){1+1}', 'else {2+2}'), silent = FALSE)
valid_syntax = function(code, silent = TRUE) {
  !inherits(suppressWarnings(try(parse_only(code), silent = silent)), 'try-error')
}

#' Bump version numbers
#'
#' Increase the last digit of version numbers, e.g., from `0.1` to
#' `0.2`, or `7.23.9` to `7.23.10`.
#' @param x A vector of version numbers (of the class `"numeric_version"`),
#'   or values that can be coerced to version numbers via
#'   `as.numeric_version()`.
#' @return A vector of new version numbers.
#' @export
#' @examples xfun::bump_version(c('0.1', '91.2.14'))
bump_version = function(x) {
  x = as.numeric_version(x)
  for (i in seq_along(x)) {
    v = x[i]
    n = length(unclass(v)[[1]])
    v[[1, n]] = v[[1, n]] + 1  # bump the last digit
    x[i] = v
  }
  x
}

#' Fix pairs of characters in a file
#'
#' For example, the curly braces may be wrong (the opening and closing braces
#' are swapped for some reason).
#' @param x A character vector (by default, read from `file`).
#' @param file Path to a text file.
#' @param chars A vector of characters of length 2. By default, it is a pair of
#'   curly double quotes.
#' @references <https://d.cosx.org/d/420794/5>
#' @noRd
#' @examples
#' files = list.files('.', '[.]R?md$', recursive = TRUE, full.names = TRUE)
#' for (f in files) {
#'   pair_chars(file = f)
#'   # curly single quotes
#'   pair_chars(file = f, chars = c('\U2018', '\U2019'))
#' }
pair_chars = function(x = read_utf8(file), file, chars = c('\U201c', '\U201d')) {
  if (length(chars) != 2) stop("'chars' must be of length 2 (i.e., a pair of characters)")
  is_file = !missing(file)
  r = paste(c('[', chars, ']'), collapse = '')
  k = gregexpr(r, x)
  m = regmatches(x, k)
  for (i in seq_along(m)) {
    n = length(m[[i]])
    if (n %% 2 != 0) {
      warning(
        'The characters do not appear in pairs in the text (',
        'line: ', i, if (is_file) c('; file: ', file), '):\n', x[i], '\n'
      )
      next
    }
    m[[i]] = rep(chars, length.out = n)
  }
  x2 = x
  regmatches(x, k) = m
  if (is_file) {
    if (!identical(x, x2)) write_utf8(x, file)
    invisible(x)
  } else x
}

#' Generate ID strings
#'
#' Substitute certain (by default, non-alphanumeric) characters with dashes and
#' remove extra dashes at both ends to generate ID strings. This function is
#' intended for generating IDs for HTML elements, so HTML tags in the input text
#' will be removed first.
#' @param x A character vector.
#' @param exclude A (Perl) regular expression to detect characters to be
#'   replaced by dashes. By default, non-alphanumeric characters are replaced.
#' @return A character vector of IDs.
#' @export
#' @examples
#' x = c('Hello world 123!', 'a  &b*^##c 456')
#' xfun::alnum_id(x)
#' xfun::alnum_id(x, '[^[:alpha:]]+')  # only keep alphabetical chars
#' # when text contains HTML tags
#' xfun::alnum_id('<h1>Hello <strong>world</strong>!')
alnum_id = function(x, exclude = '[^[:alnum:]]+') {
  x = strip_html(x)
  tolower(gsub('^-+|-+$', '', gsub(exclude, '-', x, perl = TRUE)))
}

#' Strip HTML tags
#'
#' Remove HTML tags and comments from text.
#' @param x A character vector.
#' @return A character vector with HTML tags and comments stripped off.
#' @export
#' @examples
#' xfun::strip_html('<a href="#">Hello <!-- comment -->world!</a>')
strip_html = function (x) {
  x = gsub('<!--.*?-->', '', x)
  x = gsub('<[^>]+>', '', x)
  x
}

# escape special HTML characters
escape_html = function (x) {
  x = gsubf('&', '&amp;', x)
  x = gsubf('<', '&lt;', x)
  x = gsubf('>', '&gt;', x)
  x = gsubf('"', '&quot;', x)
  x
}

one_string = function(x, ...) paste(x, ..., collapse = '\n')

library(testit)

assert('prose_index() works', {
  x = c('a', '```', 'b', '```', 'c')
  out = c(1L, 5L)
  (prose_index(x) %==% out)

  x = c('a', '````', '```r', '1+1', '```', '````', 'c')
  out = c(1L, 7L)
  (prose_index(x) %==% out)

  x = c('a', '``', 'b', '``', 'c')
  out = seq_along(x)
  (prose_index(x) %==% out)

  # a character vector of length zero
  x = character()
  out = integer()
  (prose_index(x) %==% out)

  # one backbrick
  x = c('`', 'a', '`')
  out = seq_along(x)
  (prose_index(x) %==% out)

  # two backbrick
  x = c('``', 'a', '``')
  out = seq_along(x)
  (prose_index(x) %==% out)

  # no code fences
  x = c('a', 'b')
  out = c(1L, 2L)
  (prose_index(x) %==% out)

  # two code fences
  x = c('```', 'b', '```', '```', 'd', '```')
  out = integer()
  (prose_index(x) %==% out)

  # code fences commented out
  x = c('```', 'b', '```', '<!--```', 'd', '```-->')
  (prose_index(x) %==% 4:6)

  # if the code fences are not balanced
  x = c('a', '```', 'b', '``', 'c')
  out = seq_along(x)
  (has_warning(prose_index(x)))
  (prose_index(x) %==% out)
})


assert('protect_math() puts inline math expressions in backticks', {
  (protect_math('$x$') %==% '`\\(x\\)`')
  (protect_math('$x$', '===') %==% '`===\\(x\\)===`')
  (protect_math('hi $x$ a') %==% 'hi `\\(x\\)` a')
  (protect_math('$ a $') %==% '$ a $')
  (protect_math(' $a$') %==% ' `\\(a\\)`')
  (protect_math('$ x$') %==% '$ x$')
  (protect_math('$x $') %==% '$x $')
  (protect_math('b$a$') %==% 'b$a$')  # no space before $; unlikely to be math
  (protect_math('`$a$`') %==% '`$a$`')
  (protect_math('hi $x$9') %==% 'hi $x$9')
  (protect_math('$500 $600') %==% '$500 $600')

  (protect_math('$$a$$') %==% '`$$a$$`')
  (protect_math('$$a$') %==% '$$a$')
  (protect_math('hi $$\alpha$$') %==% 'hi `$$\alpha$$`')
  (protect_math('hi $$\alpha $$') %==% 'hi $$\alpha $$')
  (protect_math('hi $$ \alpha$$') %==% 'hi $$ \alpha$$')
  (protect_math('hi $$\alpha$$ and $$ \alpha$$') %==% 'hi `$$\alpha$$` and $$ \alpha$$')
})

library(testit)

assert('is_blank() tests if a vector is blank', {
  (is_blank(''))
  (is_blank(NULL))
  (is_blank(c('', '   ', '\n', '\t')))
  (!is_blank('abc'))
  (!is_blank(c('', 'abc')))
})

assert('n2w converts numbers to words', {
  (n2w(0) %==% 'zero')
  # cap capitalizes the first letter
  (n2w(0, cap = TRUE) %==% 'Zero')
  # hyphen adds '-' for 21-99 (except 30, 40, ...)
  (n2w(21, cap = TRUE, hyphen = TRUE) %==% 'Twenty-one')
  (n2w(21, cap = TRUE, hyphen = FALSE) %==% 'Twenty one')
  # x can be negative integers
  (n2w(-21, cap = TRUE, hyphen = TRUE) %==% 'Minus twenty-one')
  (n2w(-21, cap = TRUE, hyphen = FALSE) %==% 'Minus twenty one')
  # and controls whether to have 'add' between hundreds and double digits
  (n2w(121, and = FALSE) %==% 'one hundred twenty-one')
  (n2w(121, and = TRUE) %==% 'one hundred and twenty-one')
  # x can be a vector with length > 1
  (n2w(c(10, 13, 99, 1e6)) %==% c('ten', 'thirteen', 'ninety-nine', 'one million'))
  # the number should be less than 1e15
  (has_error(n2w(1e15)))
})

assert('split_lines() splits a character vector into lines', {
  (split_lines('a') %==% 'a')
  (split_lines('') %==% '')
  (split_lines(NULL) %==% NULL)
  (split_lines('a\n') %==% c('a', ''))
  (split_lines(c('a', 'b\nc')) %==% c('a', 'b', 'c'))
  (split_lines(c('', '\n')) %==% c('', '', ''))
  (split_lines('a\nb') %==% c('a', 'b'))
  (split_lines('a\nb\n\n') %==% c('a', 'b', '', ''))
  (split_lines(c('a\nb', '', ' ', 'c')) %==% c('a', 'b', '', ' ', 'c'))
})

assert('split_source() puts lines of the same expression into a list element', {
  (split_source('1+1') %==% list('1+1'))
  (split_source(c('1+1+', '1')) %==% list(c('1+1+', '1')))
  (split_source(c('1+1+', '1', 'TRUE')) %==% list(c('1+1+', '1'), 'TRUE'))
})

assert('split_source() should signal an error for incomplete code', {
  (has_error(split_source('1+1+')))
  (has_error(split_source(c('1+1', '1+1+'))))
})

assert('valid_syntax() tells if a code fragment is syntactically valid', {
  (valid_syntax('1+1'))
  (!valid_syntax('1+1+'))
  (valid_syntax('if(TRUE)1'))
  (!valid_syntax(c('if(T){', 'F')))
  (valid_syntax(c('if(T){', 'F}')))
})

assert('alnum_id() generates ID strings', {
  x = c('Hello world 123!', 'a  &b*^##c 456')
  (alnum_id(x) %==% c('hello-world-123', 'a-b-c-456'))
  (alnum_id(x, '[^[:alpha:]]+') %==% c('hello-world', 'a-b-c'))
})

#' Type a character vector into the RStudio source editor
#'
#' Use the \pkg{rstudioapi} package to insert characters one by one into the
#' RStudio source editor, as if they were typed by a human.
#' @param x A character vector.
#' @param pause A function to return a number in seconds to pause after typing
#'   each character.
#' @param mistake The probability of making random mistakes when typing the next
#'   character. A random mistake is a random string typed into the editor and
#'   deleted immediately.
#' @param save The probability of saving the document after typing each
#'   character. Note that If a document is not opened from a file, it will never
#'   be saved.
#' @export
#' @import stats
#' @examples library(xfun)
#' if (loadable('rstudioapi') && rstudioapi::isAvailable()) {
#'   rstudio_type('Hello, RStudio! xfun::rstudio_type() looks pretty cool!',
#'     pause = function() runif(1, 0, .5), mistake = .1)
#' }
rstudio_type = function(x, pause = function() .1, mistake = 0, save = 0) {
  get_ctx = function() rstudioapi::getSourceEditorContext()
  ctx = get_ctx()
  if (is.null(id <- ctx$id)) {
    message('Please make sure an RStudio editor tab is open')
    return()
  }

  save_it = function(prob = 1) {
    if (ctx$path == '' || (rbinom(1, 1, prob) == 0)) return()
    ctx = get_ctx()
    # in case a new line is automatically added at the end when saving the doc
    on.exit(rstudioapi::setSelectionRanges(ctx$selection[[1]]$range, id), add = TRUE)
    rstudioapi::documentSave(id)
  }

  type_one = function(x) {
    rstudioapi::insertText(text = x, id = id)
    Sys.sleep(pause())
  }

  type_mistake = function() {
    n = sample(1:10, 1)
    x = sample(ascii_chars, n, replace = TRUE)
    for (i in x) type_one(i)
    Sys.sleep(.5)
    ctx = rstudioapi::getSourceEditorContext()
    r = ctx$selection[[1]]$range
    r$start[2] = r$start[2] - n
    rstudioapi::modifyRange(r, '', id)
    Sys.sleep(.5)
  }

  x = paste(x, collapse = '\n')
  for (i in unlist(strsplit(x, ''))) {
    type_one(i); save_it(save)
    if (runif(1) < mistake) type_mistake()
  }

  save_it(as.integer(save > 0))  # if prob is non-zero, save it finally

  invisible()
}

ascii_chars = intToUtf8(32:126, TRUE)

#' Get data from a REST API
#'
#' Read data from a REST API and optionally with an authorization token in the
#' request header. The function `rest_api_raw()` returns the raw text of
#' the response, and `rest_api()` will parse the response with
#' `jsonlite::fromJSON()` (assuming that the response is in the JSON
#' format).
#'
#' These functions are simple wrappers based on [url()] and
#' [read_utf8()]. Specifically, the `headers` argument is
#' passed to `url()`, and `read_utf8()` will send a \samp{GET} request
#' to the API server. This means these functions only support the \samp{GET}
#' method. If you need to use other HTTP methods (such as \samp{POST}), you have
#' to use other packages such as \pkg{curl} and \pkg{httr}.
#' @param ... Arguments to be passed to `rest_api_raw()`.
#' @return A character vector (the raw JSON response) or an R object parsed from
#'   the JSON text.
#' @export
#' @examplesIf interactive()
#' # a normal GET request
#' xfun::rest_api('https://httpbin.org', '/get')
#' xfun::rest_api_raw('https://httpbin.org', '/get')
#'
#' # send the request with an auth header
#' xfun::rest_api('https://httpbin.org', '/headers', 'OPEN SESAME!')
#'
#' # with query parameters
#' xfun::rest_api('https://httpbin.org', '/response-headers', params = list(foo = 'bar'))
#'
#' # get the rate limit info from Github
#' xfun::github_api('/rate_limit')
rest_api = function(...) {
  res = rest_api_raw(...)
  jsonlite::fromJSON(res, simplifyVector = FALSE)
}

#' @param root The API root URL.
#' @param endpoint The API endpoint.
#' @param token A named character string (e.g., `c(token = "xxxx")`), which
#'   will be used to create an authorization header of the form
#'   \samp{Authorization: NAME TOKEN} for the API call, where \samp{NAME} is the
#'   name of the string and \samp{TOKEN} is the string. If the string does not
#'   have a name, \samp{Basic} will be used as the default name.
#' @param params A list of query parameters to be sent with the API call.
#' @param headers A named character vector of HTTP headers, e.g., `c(Accept
#'   = "application/vnd.github.v3+json")`.
#' @rdname rest_api
#' @export
rest_api_raw = function(root, endpoint, token = '', params = list(), headers = NULL)  {
  if (is.null(names(token))) names(token) = 'Basic'
  endpoint = sub('^/?', '/', endpoint)  # make sure it has a leading /
  url2 = if ('headers' %in% names(formals(url))) url else stop2(
    "The url() function does not support the 'headers' argument. Please upgrade R (>= 3.6.0)."
  )
  con = url2(
    paste0(root, endpoint, query_params(.list = params)), encoding = 'UTF-8',
    headers = c(
      headers, if (token != '') c(Authorization = sprintf('%s %s', names(token), token))
    )
  )
  on.exit(close(con), add = TRUE)
  raw_string(suppressWarnings(read_utf8(con)))
}

#' Test for types of operating systems
#'
#' Functions based on `.Platform$OS.type` and `Sys.info()` to test if
#' the current operating system is Windows, macOS, Unix, or Linux.
#' @rdname os
#' @export
#' @examples
#' library(xfun)
#' # only one of the following statements should be true
#' is_windows()
#' is_unix() && is_macos()
#' is_linux()
#' # In newer Macs, CPU can be either Intel or Apple
#' is_arm64()  # TRUE on Apple silicone machines
is_windows = function() .Platform$OS.type == 'windows'

#' @rdname os
#' @export
is_unix = function() .Platform$OS.type == 'unix'

#' @rdname os
#' @export
is_macos = function() unname(Sys.info()['sysname'] == 'Darwin')

#' @rdname os
#' @export
is_linux = function() unname(Sys.info()['sysname'] == 'Linux')

#' @rdname os
#' @export
is_arm64 = function() Sys.info()[['machine']] %in% c('arm64', 'aarch64')

# open a file (optionally in an editor) or directory
open_path = function(x, edit = !dir_exists(x), ...) {
  if (edit) {
    tryCatch(rstudioapi::navigateToFile(x, ...), error = function(e) file.edit(x))
  } else {
    if (is_windows()) shell.exec(x) else {
      system2(if (is_macos()) 'open' else 'xdg-open', shQuote(x))
    }
  }
}

#' A simple JSON serializer
#'
#' A JSON serializer that only works on a limited types of R data (`NULL`,
#' lists, logical scalars, character/numeric vectors). A character string of the
#' class `JS_EVAL` is treated as raw JavaScript, so will not be quoted. The
#' function `json_vector()` converts an atomic R vector to JSON.
#' @param x An R object.
#' @export
#' @return A character string.
#' @seealso The \pkg{jsonlite} package provides a full JSON serializer.
#' @examples library(xfun)
#' tojson(NULL); tojson(1:10); tojson(TRUE); tojson(FALSE)
#' cat(tojson(list(a = 1, b = list(c = 1:3, d = 'abc'))))
#' cat(tojson(list(c('a', 'b'), 1:5, TRUE)))
#'
#' # the class JS_EVAL is originally from htmlwidgets::JS()
#' JS = function(x) structure(x, class = 'JS_EVAL')
#' cat(tojson(list(a = 1:5, b = JS('function() {return true;}'))))
tojson = function(x) {
  if (is.null(x)) return('null')
  if (is.logical(x)) {
    if (length(x) != 1 || any(is.na(x)))
      stop('Logical values of length > 1 and NA are not supported')
    return(tolower(as.character(x)))
  }
  if (is.character(x) && inherits(x, 'JS_EVAL')) return(paste(x, collapse = '\n'))
  if (is.character(x) || is.numeric(x)) {
    return(json_vector(x, length(x) != 1 || inherits(x, 'AsIs'), is.character(x)))
  }
  if (is.list(x)) {
    if (length(x) == 0) return('{}')
    return(if (is.null(names(x))) {
      json_vector(unlist(lapply(x, tojson)), TRUE, quote = FALSE)
    } else {
      nms = quote_string(names(x))
      paste0('{\n', paste(nms, unlist(lapply(x, tojson)), sep = ': ', collapse = ',\n'), '\n}')
    })
  }
  stop('The class of x is not supported: ', paste(class(x), collapse = ', '))
}

#' @param to_array Whether to convert a vector to a JSON array (use `[]`).
#' @param quote Whether to double quote the elements.
#' @rdname tojson
#' @export
json_vector = function(x, to_array = FALSE, quote = TRUE) {
  if (quote) {
    x = quote_string(x)
    x = gsub('\n', '\\\\n', x)
    x = gsub('\b', '\\\\b', x)
    x = gsub('\f', '\\\\f', x)
    x = gsub('\r', '\\\\r', x)
    x = gsub('\t', '\\\\t', x)
  }
  if (to_array) paste0('[', paste(x, collapse = ', '), ']') else x
}

# escape \ and " in strings, and quote them
quote_string = function(x) {
  x = gsub('(["\\])', "\\\\\\1", x)
  if (length(x)) x = paste0('"', x, '"')
  x
}

#' Run \code{system2()} and mark its character output as UTF-8 if appropriate
#'
#' This is a wrapper function based on \code{system2()}. If \code{system2()}
#' returns character output (e.g., with the argument \code{stdout = TRUE}),
#' check if the output is encoded in UTF-8. If it is, mark it with UTF-8
#' explicitly.
#' @param ... Passed to \code{\link{system2}()}.
#' @return The value returned by \code{system2()}.
#' @export
#' @examplesIf interactive()
#' a = shQuote(c('-e', 'print(intToUtf8(c(20320, 22909)))'))
#' x2 = system2('Rscript', a, stdout = TRUE)
#' Encoding(x2)  # unknown
#'
#' x3 = xfun::system3('Rscript', a, stdout = TRUE)
#' # encoding of x3 should be UTF-8 if the current locale is UTF-8
#' !l10n_info()[['UTF-8']] || Encoding(x3) == 'UTF-8'  # should be TRUE
system3 = function(...) {
  res = system2(...)
  if (is.character(res)) {
    if (all(is_utf8(res))) Encoding(res) = 'UTF-8'
  }
  if (is.integer(res) && res == 0) invisible(res) else res
}

#' Run OptiPNG on all PNG files under a directory
#'
#' Call the command \command{optipng} via \code{system2()} to optimize all PNG
#' files under a directory.
#' @param dir Path to a directory.
#' @param files Alternatively, you can choose the specific files to optimize.
#' @param ... Arguments to be passed to \code{system2()}.
#' @references OptiPNG: \url{https://optipng.sourceforge.net}.
#' @export
optipng = function(dir = '.', files = all_files('[.]png$', dir), ...) {
  if (Sys.which('optipng') != '') for (f in files) system2('optipng', shQuote(f), ...)
}

#' Run the commands \command{Rscript} and \command{R CMD}
#'
#' Wrapper functions to run the commands \command{Rscript} and \command{R CMD}.
#' @param args A character vector of command-line arguments.
#' @param ... Other arguments to be passed to \code{\link{system2}()}.
#' @export
#' @return A value returned by \code{system2()}.
#' @examples library(xfun)
#' Rscript(c('-e', '1+1'))
#' Rcmd(c('build', '--help'))
Rscript = function(args, ...) {
  # unset R_TESTS for the new R session: https://stackoverflow.com/a/27994299
  if (is_R_CMD_check()) {
    v = set_envvar(c(R_TESTS = NA)); on.exit(set_envvar(v), add = TRUE)
  }
  system2(file.path(R.home('bin'), 'Rscript'), args, ...)
}

#' @rdname Rscript
#' @export
Rcmd = function(args, ...) {
  system2(file.path(R.home('bin'), 'R'), c('CMD', args), ...)
}

#' Call a function in a new R session via \code{Rscript()}
#'
#' Save the argument values of a function in a temporary RDS file, open a new R
#' session via \code{\link{Rscript}()}, read the argument values, call the
#' function, and read the returned value back to the current R session.
#' @param fun A function, or a character string that can be parsed and evaluated
#'   to a function.
#' @param args A list of argument values.
#' @param options A character vector of options to passed to
#'   \code{\link{Rscript}}, e.g., \code{"--vanilla"}.
#' @param ...,wait Arguments to be passed to \code{\link{system2}()}.
#' @param fail The desired error message when an error occurred in calling the
#'   function.
#' @export
#' @return The returned value of the function in the new R session.
#' @examples factorial(10)
#' # should return the same value
#' xfun::Rscript_call('factorial', list(10))
#'
#' # the first argument can be either a character string or a function
#' xfun::Rscript_call(factorial, list(10))
#'
#' # Run Rscript starting a vanilla R session
#' xfun::Rscript_call(factorial, list(10), options = c("--vanilla"))
Rscript_call = function(
  fun, args = list(), options = NULL, ..., wait = TRUE,
  fail = sprintf("Failed to run '%s' in a new R session.", deparse(substitute(fun))[1])
) {
  f = replicate(2, tempfile(fileext = '.rds'))
  on.exit(unlink(if (wait) f else f[2]), add = TRUE)
  saveRDS(list(fun, args), f[1])
  Rscript(
    c(options, shQuote(c(pkg_file('scripts', 'call-fun.R'), f)))
    ,..., wait = wait
  )
  if (wait) if (file_exists(f[2])) readRDS(f[2]) else stop(fail, call. = FALSE)
}

# call a function in a background process
Rscript_bg = function(fun, args = list(), timeout = 10) {
  pid = tempfile()  # to store the process ID of the new R session
  saveRDS(NULL, pid)

  Rscript_call(function() {
    saveRDS(Sys.getpid(), pid)
    # remove this pid file when the function finishes
    on.exit(unlink(pid), add = TRUE)
    do.call(fun, args)
  }, wait = FALSE)

  id = NULL  # read the above process ID into this R session
  res = list(pid = id, is_alive = function() FALSE)

  # check if the pid file still exists; if not, the process has ended
  if (!file_exists(pid)) return(res)

  t0 = Sys.time()
  while (difftime(Sys.time(), t0, units = 'secs') < timeout) {
    Sys.sleep(.1)
    if (!file_exists(pid)) return(res)
    if (length(id <- readRDS(pid)) == 1) break
  }
  if (length(id) == 0) stop(
    'Failed to launch the background process in ', timeout, ' seconds (timeout).'
  )

  list(pid = id, is_alive = function() file_exists(pid))
}

#' Kill a process and (optionally) all its child processes
#'
#' Run the command \command{taskkill /f /pid} on Windows and \command{kill} on
#' Unix, respectively, to kill a process.
#' @param pid The process ID.
#' @param recursive Whether to kill the child processes of the process.
#' @param ... Arguments to be passed to \code{\link{system2}()} to run the
#'   command to kill the process.
#' @return The status code returned from \code{system2()}.
#' @export
proc_kill = function(pid, recursive = TRUE, ...) {
  if (is_windows()) {
    system2('taskkill', c(if (recursive) '/t', '/f', '/pid', pid), ...)
  } else {
    system2('kill', c(pid, if (recursive) child_pids(pid)), ...)
  }
}

# obtain pids of all child processes (recursively)
child_pids = function(id) {
  x = system2('sh', shQuote(c(pkg_file('scripts', 'child-pids.sh'), id)), stdout = TRUE)
  grep('^[0-9]+$', x, value = TRUE)
}

powershell = function(command) {
  if (Sys.which('powershell') == '') return()
  command = paste(command, collapse = ' ')
  system2('powershell', c('-Command', shQuote(command)), stdout = TRUE)
}

# start a background process via the PowerShell cmdlet and return its pid
ps_process = function(command, args = character(), verbose = FALSE) {
  powershell(c(
    'echo (Start-Process', '-FilePath', shQuote(command), '-ArgumentList',
    ps_quote(args), '-PassThru', '-WindowStyle',
    sprintf('%s).ID', if (verbose) 'Normal' else 'Hidden')
  ))
}

# quote PowerShell arguments properly
ps_quote = function(x) {
  x = gsub('"', '""', x)  # '""' mean a literal '"'
  # if an argument contains a space, surround it with escaped double quotes `"`"
  i = grep('\\s', x)
  x[i] = sprintf('`"%s`"', x[i])
  sprintf('"%s"', paste(x, collapse = ' '))
}

#' Start a background process
#'
#' Start a background process using the PowerShell cmdlet \command{Start-Process
#' -PassThru} on Windows or the ampersand \command{&} on Unix, and return the
#' process ID.
#' @param command,args The system command and its arguments. They do not need to
#'   be quoted, since they will be quoted via \code{\link{shQuote}()}
#'   internally.
#' @param verbose If \code{FALSE}, suppress the output from \verb{stdout} (and
#'   also \verb{stderr} on Windows). The default value of this argument can be
#'   set via a global option, e.g., \code{options(xfun.bg_process.verbose =
#'   TRUE)}.
#' @return The process ID as a character string.
#' @note On Windows, if PowerShell is not available, try to use
#'   \code{\link{system2}(wait = FALSE)} to start the background process
#'   instead. The process ID will be identified from the output of the command
#'   \command{tasklist}. This method of looking for the process ID may not be
#'   reliable. If the search is not successful in 30 seconds, it will throw an
#'   error (timeout). If a longer time is needed, you may set
#'   \code{options(xfun.bg_process.timeout)} to a larger value, but it should be
#'   very rare that a process cannot be started in 30 seconds. When you reach
#'   the timeout, it is more likely that the command actually failed.
#' @export
#' @seealso \code{\link{proc_kill}()} to kill a process.
bg_process = function(
  command, args = character(), verbose = getOption('xfun.bg_process.verbose', FALSE)
) {
  throw_error = function(...) stop(
    'Failed to run the command', ..., ' in the background: ',
    paste(shQuote(c(command, args)), collapse = ' '), call. = FALSE
  )

  # check the possible pid returned from system2()
  check_pid = function(res) {
    if (is.null(res)) return(res)
    if (!is.null(attr(res, 'status'))) throw_error()
    if (length(res) == 1 && grepl('^[0-9]+$', res)) return(res)
    throw_error()
  }

  if (is_windows()) {
    # first try 'Start-Process -PassThrough' to start a background process; if
    # PowerShell is unavailable, fall back to system2(wait = FALSE), and the
    # method to find out the pid is not 100% reliable
    if (length(pid <- check_pid(ps_process(command, args, verbose))) == 1) return(pid)

    message(
      'It seems you do not have PowerShell installed. The process ID may be inaccurate.'
    )
    # format of task list: hugo.exe    4592 Console      1     35,188 K
    tasklist = function() system2('tasklist', stdout = TRUE)
    pid1 = tasklist()
    system2(command, shQuote(args), wait = FALSE)

    get_pid = function() {
      # make sure the command points to an actual executable (e.g., resolve 'R'
      # to 'R.exe')
      if (!file_exists(command)) {
        if (Sys.which(command) != '') command = Sys.which(command)
      }
      cmd = basename(command)

      pid2 = setdiff(tasklist(), pid1)
      # the process's info should start with the command name
      pid2 = pid2[substr(pid2, 1, nchar(cmd)) == cmd]
      if (length(pid2) == 0) return()
      m = regexec('\\s+([0-9]+)\\s+', pid2)
      for (v in regmatches(pid2, m)) if (length(v) >= 2) return(v[2])
    }

    t0 = Sys.time(); id = NULL; timeout = getOption('xfun.bg_process.timeout', 30)
    while (difftime(Sys.time(), t0, units = 'secs') < timeout) {
      if (length(id <- get_pid()) > 0) break
    }

    if (length(id) > 0) return(id)

    system2(command, args, timeout = timeout)  # see what the error is
    throw_error(' in ', timeout, ' second(s)')
  } else {
    pid = tempfile(); on.exit(unlink(pid), add = TRUE)
    code = paste(c(
      shQuote(c(command, args)), if (!verbose) '> /dev/null', '& echo $! >', shQuote(pid)
    ), collapse = ' ')
    system2('sh', c('-c', shQuote(code)))
    return(check_pid(readLines(pid)))
  }
}

#' Upload to an FTP server via \command{curl}
#'
#' The function \code{upload_ftp()} runs the command \command{curl -T file
#' server} to upload a file to an FTP server if the system command
#' \command{curl} is available, otherwise it uses the R package \pkg{curl}. The
#' function \code{upload_win_builder()} uses \code{upload_ftp()} to upload
#' packages to the win-builder server.
#'
#' These functions were written mainly to save package developers the trouble of
#' going to the win-builder web page and uploading packages there manually.
#' @param file Path to a local file.
#' @param server The address of the FTP server. For \code{upload_win_builder()},
#'   \code{server = 'https'} means uploading to
#'   \code{'https://win-builder.r-project.org/upload.aspx'}.
#' @param dir The remote directory to which the file should be uploaded.
#' @param version The R version(s) on win-builder.
#' @return Status code returned from \code{\link{system2}()} or
#'   \code{curl::curl_fetch_memory()}.
#' @export
upload_ftp = function(file, server, dir = '') {
  if (dir != '') dir = gsub('/*$', '/', dir)
  server = paste0(server, dir)
  if (Sys.which('curl') == '') {
    curl::curl_upload(file, server)$status_code
  } else {
    system2('curl', shQuote(c('-T', file, server)))
  }
}

#' @param solaris Whether to also upload the package to the Rhub server to check
#'   it on Solaris.
#' @rdname upload_ftp
#' @export
upload_win_builder = function(
  file = pkg_build(), version = c("R-devel", "R-release", "R-oldrelease"),
  server = c('ftp', 'https'), solaris = pkg_available('rhub')
) {
  if (missing(file)) on.exit(file.remove(file), add = TRUE)
  if (system2('git', 'status', stderr = FALSE) == 0) system2('git', 'pull')
  server = server[1]
  server = switch(
    server,
    'ftp'   = paste0(server, '://win-builder.r-project.org/'),
    'https' = paste0(server, '://win-builder.r-project.org/upload.aspx'),
    server
  )
  res = if (grepl('^ftp://', server)) {
    lapply(version, upload_ftp, file = file, server = server)
  } else {
    vers = c('R-devel' = 2, 'R-release' = 1, 'R-oldrelease' = 3)
    params = list(
      FileUpload = file,
      Button = 'Upload File',
      # perhaps we should read these tokens dynamically from
      # https://win-builder.r-project.org/upload.aspx
      `__VIEWSTATE` = '/wEPDwULLTE0OTY5NTg0MTUPZBYCAgIPFgIeB2VuY3R5cGUFE211bHRpcGFydC9mb3JtLWRhdGFkZFHMrNH6JjHTyJ00T0dAADGf4oa0',
      `__VIEWSTATEGENERATOR` = '69164837',
      `__EVENTVALIDATION` = '/wEWBQKksYbrBgKM54rGBgK7q7GGCAKF2fXbAwLWlM+bAqR2dARbCNfKVu0vDawqWYgB5kKI'
    )
    lapply(version, function(i) {
      names(params)[1:2] = paste0(names(params)[1:2], vers[i])
      if (Sys.which('curl') == '') {
        h = curl::new_handle()
        params[[1]] = curl::form_file(params[[1]])
        curl::handle_setform(h, .list = params)
        curl::curl_fetch_memory(server, h)$status_code
      } else {
        params[1] = paste0('@', params[1])
        system2('curl', shQuote(c(
          rbind('-F', paste(names(params), params, sep = '=')),
          server
        )), stdout = FALSE)
      }
    })
  }

  if (solaris) rhub::check_on_solaris(
    file, check_args = '--no-manual', show_status = FALSE,
    env_vars = c(`_R_CHECK_FORCE_SUGGESTS_` = 'false')
  )

  setNames(unlist(res), version)
}

#' Strict lists
#'
#' A strict list is essentially a normal \code{\link{list}()} but it does not
#' allow partial matching with \code{$}.
#'
#' To me, partial matching is often more annoying and surprising than
#' convenient. It can lead to bugs that are very hard to discover, and I have
#' been bitten by it many times. When I write \code{x$name}, I always mean
#' precisely \code{name}. You should use a modern code editor to autocomplete
#' the \code{name} if it is too long to type, instead of using partial names.
#' @param ... Objects (list elements), possibly named. Ignored in the
#'   \code{print()} method.
#' @export
#' @return Both \code{strict_list()} and \code{as_strict_list()} return a list
#'   with the class \code{xfun_strict_list}. Whereas \code{as_strict_list()}
#'   attempts to coerce its argument \code{x} to a list if necessary,
#'   \code{strict_list()} just wraps its argument \code{...} in a list, i.e., it
#'   will add another list level regardless if \code{...} already is of type
#'   list.
#' @examples library(xfun)
#' (z = strict_list(aaa = 'I am aaa', b = 1:5))
#' z$a  # NULL!
#' z$aaa  # I am aaa
#' z$b
#' z$c = 'create a new element'
#'
#' z2 = unclass(z)  # a normal list
#' z2$a  # partial matching
#'
#' z3 = as_strict_list(z2) # a strict list again
#' z3$a  # NULL again!
strict_list = function(...) {
  as_strict_list(list(...))
}

# https://twitter.com/xieyihui/status/782462926862954496

#' @param x For \code{as_strict_list()}, the object to be coerced to a strict
#'   list.
#'
#'   For \code{print()}, a strict list.
#' @rdname strict_list
#' @export
as_strict_list = function(x) {
  structure(as.list(x), class = 'xfun_strict_list')
}

#' @param name The name (a character string) of the list element.
#' @rdname strict_list
#' @export
`$.xfun_strict_list` = function(x, name) x[[name]]

#' @rdname strict_list
#' @export
print.xfun_strict_list = function(x, ...) {
  print(unclass(x))
}

#' Print a character vector in its raw form
#'
#' The function \code{raw_string()} assigns the class \code{xfun_raw_string} to
#' the character vector, and the corresponding printing function
#' \code{print.xfun_raw_string()} uses \code{cat(x, sep = '\n')} to write the
#' character vector to the console, which will suppress the leading indices
#' (such as \code{[1]}) and double quotes, and it may be easier to read the
#' characters in the raw form (especially when there are escape sequences).
#' @param x For \code{raw_string()}, a character vector. For the print method,
#'   the \code{raw_string()} object.
#' @export
#' @examples library(xfun)
#' raw_string(head(LETTERS))
#' raw_string(c('a "b"', 'hello\tworld!'))
raw_string = function(x) {
  if (is.null(x)) x = as.character(x)
  class(x) = c('xfun_raw_string', class(x))
  x
}

#' @param ... Other arguments (currently ignored).
#' @rdname raw_string
#' @export
print.xfun_raw_string = function(x, ...) {
  if (length(x)) cat(x, sep = '\n')
  invisible(x)
}

#' Cache the execution of an expression in memory or on disk
#'
#' Caching is based on the assumption that if the input does not change, the
#' output will not change. After an expression is executed for the first time,
#' its result will be saved (either in memory or on disk). The next run will be
#' skipped and the previously saved result will be loaded directly if all
#' external inputs of the expression remain the same, otherwise the cache will
#' be invalidated and the expression will be re-executed.
#'
#' Arguments supported in `...` include:
#'
#' - `vars`: Names of local variables (which are created inside the expression).
#' By default, local variables are automatically detected from the expression
#' via [codetools::findLocalsList()]. Locally created variables are cached along
#' with the value of the expression.
#'
#' - `hash` and `extra`: R objects to be used to determine if cache should be
#' loaded or invalidated. If (the MD5 hash of) the objects is not changed, the
#' cache is loaded, otherwise the cache is invalidated and rebuilt. By default,
#' `hash` is a list of values of global variables in the expression (i.e.,
#' variables created outside the expression). Global variables are automatically
#' detected by [codetools::findGlobals()]. You can provide a vector of names to
#' override the automatic detection if you want some specific global variables
#' to affect caching, or the automatic detection is not reliable. You can also
#' provide additional information via the `extra` argument. For example, if the
#' expression reads an external file `foo.csv`, and you want the cache to be
#' invalidated after the file is modified, you may use `extra =
#' file.mtime("foo.csv")`.
#'
#' - `keep`: By default, only one copy of the cache corresponding to an `id`
#' under `path` is kept, and all other copies for this `id` is automatically
#' purged. If `TRUE`, all copies of the cache are kept. If `FALSE`, all copies
#' are removed, which means the cache is *always* invalidated, and can be useful
#' to force re-executing the expression.
#'
#' - `rw`: A list of functions to read/write the cache files. The list is of the
#' form `list(name = 'xxx', load = function(file) {}, save = function(x, file)
#' {})`. By default, [readRDS()] and [saveRDS()] are used. This argument can
#' also take a character string to use some built-in read/write methods.
#' Currently available methods include `rds` (the default), `raw` (using
#' [serialize()] and [unserialize()]), and `qs` (using [qs::qread()] and
#' [qs::qsave()]). The `rds` and `raw` methods only use base R functions (the
#' `rds` method generates smaller files because it uses compression, but is
#' often slower than the `raw` method, which does not use compression). The `qs`
#' method requires the \pkg{qs} package, which can be much faster than base R
#' methods and also supports compression.
#' @param expr An R expression to be cached.
#' @param path The path to save the cache. The special value `":memory:"` means
#'   in-memory caching. If it is intended to be a directory path, please make
#'   sure to add a trailing slash.
#' @param id A stable and unique string identifier for the expression to be used
#'   to identify a unique copy of cache for the current expression from all
#'   cache files (or in-memory elements). If not provided, an MD5 digest of the
#'   [deparse]d expression will be used, which means if the expression does not
#'   change (changes in comments or white spaces do not matter), the `id` will
#'   remain the same. This may not be a good default is two identical
#'   expressions are cached under the same `path`, because they could overwrite
#'   each other's cache when one expression's cache is invalidated, which may or
#'   may not be what you want. If you do not want that to happen, you need to
#'   manually provide an `id`.
#' @param ... More arguments to control the behavior of caching (see
#'   \sQuote{Details}).
#' @return If the cache is found, the cached value of the expression will be
#'   loaded and returned (other local variables will also be lazy-loaded into
#'   the current environment as a side-effect). If cache does not exist, the
#'   expression is executed and its value is returned.
#' @export
#' @examples
#' # the first run takes about 1 second
#' y1 = xfun::cache_exec({
#'   x = rnorm(1e5)
#'   Sys.sleep(1)
#'   x
#' }, path = ':memory:', id = 'sim-norm')
#'
#' # the second run takes almost no time
#' y2 = xfun::cache_exec({
#'   # comments won't affect caching
#'   x = rnorm(1e5)
#'   Sys.sleep(1)
#'   x
#' }, path = ':memory:', id = 'sim-norm')
#'
#' # y1, y2, and x should be identical
#' stopifnot(identical(y1, y2), identical(y1, x))
cache_exec = function(expr, path = 'cache/', id = NULL, ...) {
  use_cache = FALSE
  ret = cache_code(
    substitute(expr), parent.frame(), use_cache <- TRUE,
    list(path = path, id = id, ...)
  )
  if (use_cache) ret else expr
}

# a dictionary containing names and hashes of global variables
.cache_dict = new.env(parent = emptyenv())

# an in-memory database containing results from previous run, values of local
# variables, and their hashes, of the form environment(*__id__* = list(results,
# values, hashes))
.cache_db = new.env(parent = emptyenv())

cache_code = function(
  code, envir = parent.frame(), found = NULL, config = list(
    path = NULL, vars = NULL, hash = NULL, extra = NULL, keep = NULL, id = NULL,
    rw = NULL
  )
) {
  dict = .cache_dict
  if (!is.character(path <- config$path)) {
    # when caching is not enabled, we should clean up hashes for local variables
    # to make sure the next cached chunk will get the up-to-date hash
    vars = intersect(config$vars %||% find_locals(code), ls_all(dict))
    rm_vars(vars, dict)
    return()
  }

  # functions to read/write cache files
  if (is.null(rw <- config$rw)) rw = 'rds'
  if (is.character(rw)) rw = io_methods[[rw]]
  if (!is.character(rw$name) || length(rw$name) != 1) stop(
    'The cache method must have a name (as a character string).'
  )

  hash = config$hash %||% find_globals(code, envir)
  # get the values of hash variables (unless hash is I(character()))
  if (is.character(hash) && !inherits(hash, 'AsIs')) {
    # if a variable exists in the hash dictionary, use its hash (32 chars) in
    # the dict instead of its actual value, because the value may be large and
    # slow to serialize or compute md5 checksum
    hash = lapply(hash, function(x) dict[[x]] %||% get(x, envir))
  }
  # "normalized" code that doesn't rely on comments or white spaces
  code2 = deparse(if (is.character(code)) parse_only(code) else code)
  hash = md5_one(c(hash, config$extra, list(code2)))

  mem_cache = path == ':memory:'  # in-memory or disk cache

  # by default, the ID is a checksum of the code, which may not be a good idea
  # (e.g., two identical code fragments could be executed one after another
  # but they should not share the same cache); for litedown, id is chunk label
  if (is.null(id <- config$id)) id = md5_one(code2)
  id = paste0(path, if (mem_cache) '__', id, if (mem_cache) '__')

  file_pattern = function(x) sprintf('^[0-9a-z]{32}[.]%s([.][0-9]+)?$', x)
  # try to retrieve cache from memory (the dictionary) or disk
  cached = if (mem_cache) {
    db = .cache_db
    id_names = function(x) {
      x = ls_all(db)
      x[startsWith(x, id) & nchar(x) == nchar(id) + 32]
    }
    hits = id_names(db)
    # clean up all cache for this id, i.e., remove :memory:__id__hash
    if (base::isFALSE(config$keep)) rm_vars(hits, db)
    id = paste0(id, hash)
    id %in% ls_all(db)
  } else {
    db = list.files(id, full.names = TRUE)
    hits = grepl(file_pattern('[[:alnum:]]+'), basename(db))
    if (base::isFALSE(config$keep)) file.remove(db[hits])
    id = file.path(id, paste0(hash, '.', rw$name))
    file_exists(id)
  }
  # return now if cache is found
  if (cached) {
    # I could implement lazy-loading here but I'm not sure if it's worth it
    ret = if (mem_cache) db[[id]] else rw$load(id)
    list2env(ret$hashes, dict)  # update the hash dictionary
    # lazy-load cached objects into envir
    .mapply(function(name, file) {
      delayedAssign(name, if (mem_cache) ret$values[[name]] else rw$load(file), assign.env = envir)
    }, list(names(ret$hashes), sprintf('%s.%d', id, seq_along(ret$hashes))), NULL)
    found
    return(ret$result)
  }

  # clean up other versions of cache before saving a new version
  if (!isTRUE(config$keep)) {
    if (mem_cache) rm_vars(hits, db) else {
      # clean cache for the current method only; don't touch other methods' cache
      hits = grepl(file_pattern(rw$name), basename(db))
      file.remove(db[hits])
    }
  }
  vars = config$vars %||% find_locals(code)
  # inject an on.exit() call to the parent function to save its returned value;
  # note that returnValue() requires R >= 3.2.0 released on 2015-04-16, which I
  # hope is a reasonable requirement (anyone using R from a decade ago?)
  exit_call(function() {
    # we can't tell if returnValue() is from a successful call of a function or
    # not, so we pass a default that is very unlikely to be the returned value
    # on success; if we get this value, it's very likely that errors occurred
    void_return = new.env()  # should be unlikely for two new.env() to be identical
    res = returnValue(void_return)
    if (identical(res, void_return)) return()
    # save the checksums of objects to be used for a future run
    v1 = ls_all(envir)
    v2 = setdiff(vars, v1)
    if (length(v2)) warning(
      'Variable(s) not found in the environment: ', paste(v2, collapse = ', ')
    )
    hashes = as.list(do.call(md5, vals <- mget(vars, envir)))
    ret = list(result = res, hashes = hashes)
    if (mem_cache) {
      ret$values = vals
      db[[id]] = ret
    } else {
      dir_create(dirname(id))
      rw$save(ret, id)  # result and variable hashes are saved to the main file
      # variable values are saved in separate files of the form HASH.i; the i-th
      # file is for the i-th variable
      .mapply(rw$save, list(vals, sprintf('%s.%d', id, seq_along(vals))), NULL)
    }
    list2env(hashes, dict)
  })
}

# functions to load/save cache files
io_methods = list(
  raw = list(
    name = 'raw',
    load = function(...) unserialize(read_bin(...)),
    save = function(x, file, ...) {
      s = serialize(x, NULL, xdr = FALSE, ...)
      writeBin(s, file)
    }
  ),
  rds = list(
    name = 'rds',
    load = function(...) readRDS(...),
    save = function(x, file, ...) saveRDS(x, file, ...)
  ),
  qs = list(
    name = 'qs',
    load = function(...) qs::qread(...),
    save = function(x, file, ...) qs::qsave(x, file, ...)
  )
)

#' Cache the value of an R expression to an RDS file
#'
#' Save the value of an expression to a cache file (of the RDS format). Next
#' time the value is loaded from the file if it exists.
#'
#' Note that the `file` argument does not provide the full cache filename. The
#' actual name of the cache file is of the form \file{BASENAME_HASH.rds}, where
#' \file{BASENAME} is the base name provided via the \file{file} argument (e.g.,
#' if `file = 'foo.rds'`, `BASENAME` would be \file{foo}), and \file{HASH} is
#' the MD5 hash (also called the \sQuote{checksum}) calculated from the R code
#' provided to the `expr` argument and the value of the `hash` argument, which
#' means when the code or the `hash` argument changes, the \file{HASH} string
#' may also change, and the old cache will be invalidated (if it exists). If you
#' want to find the cache file, look for \file{.rds} files that contain 32
#' hexadecimal digits (consisting of 0-9 and a-z) at the end of the filename.
#'
#' The possible ways to invalidate the cache are: 1) change the code in `expr`
#' argument; 2) delete the cache file manually or automatically through the
#' argument `rerun = TRUE`; and 3) change the value of the `hash` argument. The
#' first two ways should be obvious. For the third way, it makes it possible to
#' automatically invalidate the cache based on changes in certain R objects. For
#' example, when you run `cache_rds({ x + y })`, you may want to invalidate the
#' cache to rerun `{ x + y }` when the value of `x` or `y` has been changed, and
#' you can tell `cache_rds()` to do so by `cache_rds({ x + y }, hash = list(x,
#' y))`. The value of the argument `hash` is expected to be a list, but it can
#' also take a special value, `"auto"`, which means `cache_rds(expr)` will try
#' to automatically figure out the global variables in `expr`, return a list of
#' their values, and use this list as the actual value of `hash`. This behavior
#' is most likely to be what you really want: if the code in `expr` uses an
#' external global variable, you may want to invalidate the cache if the value
#' of the global variable has changed. Here a \dQuote{global variable} means a
#' variable not created locally in `expr`, e.g., for `cache_rds({ x <- 1; x + y
#' })`, `x` is a local variable, and `y` is (most likely to be) a global
#' variable, so changes in `y` should invalidate the cache. However, you know
#' your own code the best. If you want to be completely sure when to invalidate
#' the cache, you can always provide a list of objects explicitly rather than
#' relying on `hash = "auto"`.
#'
#' By default (the argument `clean = TRUE`), old cache files will be
#' automatically cleaned up. Sometimes you may want to use `clean = FALSE` (set
#' the R global option `options(xfun.cache_rds.clean = FALSE)` if you want
#' `FALSE` to be the default). For example, you may not have decided which
#' version of code to use, and you can keep the cache of both versions with
#' `clean = FALSE`, so when you switch between the two versions of code, it will
#' still be fast to run the code.
#' @param expr An R expression.
#' @param rerun Whether to delete the RDS file, rerun the expression, and save
#'   the result again (i.e., invalidate the cache if it exists).
#' @param file The *base* (see Details) cache filename under the directory
#'   specified by the `dir` argument. If not specified and this function is
#'   called inside a code chunk of a \pkg{knitr} document (e.g., an R Markdown
#'   document), the default is the current chunk label plus the extension
#'   \file{.rds}.
#' @param dir The path of the RDS file is partially determined by `paste0(dir,
#'   file)`. If not specified and the \pkg{knitr} package is available, the
#'   default value of `dir` is the \pkg{knitr} chunk option `cache.path` (so if
#'   you are compiling a \pkg{knitr} document, you do not need to provide this
#'   `dir` argument explicitly), otherwise the default is \file{cache/}. If you
#'   do not want to provide a `dir` but simply a valid path to the `file`
#'   argument, you may use `dir = ""`.
#' @param hash A `list` object that contributes to the MD5 hash of the cache
#'   filename (see Details). It can also take a special character value
#'   `"auto"`. Other types of objects are ignored.
#' @param clean Whether to clean up the old cache files automatically when
#'   `expr` has changed.
#' @param ... Other arguments to be passed to [saveRDS()].
#' @note Changes in the code in the `expr` argument do not necessarily always
#'   invalidate the cache, if the changed code is [`parse`]`d` to the same
#'   expression as the previous version of the code. For example, if you have
#'   run `cache_rds({Sys.sleep(5);1+1})` before, running `cache_rds({ Sys.sleep(
#'   5 ) ; 1 + 1 })` will use the cache, because the two expressions are
#'   essentially the same (they only differ in white spaces). Usually you can
#'   add/delete white spaces or comments to your code in `expr` without
#'   invalidating the cache. See the package vignette `vignette('xfun', package
#'   = 'xfun')` for more examples.
#'
#'   When this function is called in a code chunk of a \pkg{knitr} document, you
#'   may not want to provide the filename or directory of the cache file,
#'   because they have reasonable defaults.
#'
#'   Side-effects (such as plots or printed output) will not be cached. The
#'   cache only stores the last value of the expression in `expr`.
#' @return If the cache file does not exist, run the expression and save the
#'   result to the file, otherwise read the cache file and return the value.
#' @seealso [cache_exec()], which is more flexible (e.g., it supports in-memory
#'   caching and different read/write methods for cache files).
#' @export
#' @examples
#' f = tempfile()  # the cache file
#' compute = function(...) {
#'   res = xfun::cache_rds({
#'     Sys.sleep(1)
#'     1:10
#'   }, file = f, dir = '', ...)
#'   res
#' }
#' compute()  # takes one second
#' compute()  # returns 1:10 immediately
#' compute()  # fast again
#' compute(rerun = TRUE)  # one second to rerun
#' compute()
#' unlink(paste0(f, '_*.rds'))
cache_rds = function(
  expr = {}, rerun = FALSE, file = 'cache.rds', dir = 'cache/',
  hash = NULL, clean = getOption('xfun.cache_rds.clean', TRUE), ...
) {
  if (loadable('knitr')) {
    if (missing(file) && !is.null(lab <- knitr::opts_current$get('label')))
      file = paste0(lab, '.rds')
    if (missing(dir) && !is.null(d <- knitr::opts_current$get('cache.path')))
      dir = d
  }
  path = paste0(dir, file)
  if (!grepl(r <- '([.]rds)$', path)) path = paste0(path, '.rds')
  code = deparse(substitute(expr))
  md5  = md5_obj(code)
  if (identical(hash, 'auto')) hash = global_vars(code, parent.frame(2))
  if (is.list(hash)) md5 = md5_obj(c(md5, md5_obj(hash)))
  path = sub(r, paste0('_', md5, '\\1'), path)
  if (rerun) unlink(path)
  if (clean) clean_cache(path)
  if (file_exists(path)) readRDS(path) else {
    obj = expr  # lazy evaluation
    dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
    saveRDS(obj, path, ...)
    obj
  }
}

# an old hash function for cache_rds() only
md5_obj = function(x) md5_one(x, function(x, f) {
  if (is.character(x)) writeLines(x, f) else saveRDS(x, f)
})

#' Calculate the MD5 checksums of R objects
#'
#' [Serialize][serialize()] an object to a temporary file, calculate the
#' checksum via [tools::md5sum()], and delete the file.
#' @param ... Any number of R objects.
#' @return A character vector of the checksums of objects passed to `md5()`. If
#'   the arguments are named, the results will also be named.
#' @export
#' @examples
#' x1 = 1; x2 = 1:10; x3 = seq(1, 10); x4 = iris; x5 = paste
#' (m = xfun::md5(x1, x2, x3, x4, x5))
#' stopifnot(m[2] == m[3])  # x2 and x3 should be identical
#'
#' xfun::md5(x1 = x1, x2 = x2)  # named arguments
md5 = function(...) {
  obj = list(...)
  res = unlist(lapply(obj, md5_one))
  names(res) = names(obj)
  res
}

md5_one = function(x, write = serialize_bin) {
  # we have wished for years that tools::md5sum() could just accept raw bytes
  # (HenrikBengtsson/Wishlist-for-R#21; I've also asked an R core member in
  # person in 2019) so we don't have to use the ugly tmp file hack below
  f = tempfile(); on.exit(unlink(f), add = TRUE)
  write(x, f)
  unname(tools::md5sum(f))
}

serialize_bin = function(x, f) {
  s = serialize(x, NULL, xdr = FALSE)
  writeBin(tail(s, -14), f)  # the first 14 bytes contain version info, etc
}

# clean up old cache files (those with the same base names as the new cache
# file, e.g., if the new file is FOO_0123abc...z.rds, then FOO_9876def...x.rds
# should be deleted)
clean_cache = function(path) {
  olds = list.files(dirname(path), '_[0-9a-f]{32}[.]rds$', full.names = TRUE)
  olds = c(olds, path)  # `path` may not exist; make sure it is in target paths
  base = basename(olds)
  keep = basename(path) == base  # keep this file (will cache to this file)
  base = substr(base, 1, nchar(base) - 37)  # 37 = 1 (_) + 32 (md5 sum) + 4 (.rds)
  unlink(olds[(base == base[keep][1]) & !keep])
}

# analyze code and find out global variables used from an environment
find_globals = function(code, envir = parent.frame()) {
  if (is.language(code)) {
    fun = function() {}
    body(fun) = code
  } else {
    fun = eval(parse_only(c('function(){', code, '}')), baseenv())
  }
  obj = codetools::findGlobals(fun)
  intersect(obj, ls_all(envir, TRUE))
}

# ls() all objects in the envir and if recursive, all parent environments until
# globalenv() or emptyenv()
ls_all = function(envir, recursive = FALSE) {
  x = ls(envir, all.names = TRUE)
  if (!recursive) return(x)
  while (TRUE) {
    # in theory, we shouldn't stop at global env but should keep recursion, but
    # in practice, objects in the parent environments of globalenv() (often
    # package namespaces) are unlikely to change and won't affect cache
    if (identical(envir, emptyenv()) || identical(envir, .GlobalEnv)) break
    envir = parent.env(envir)
    x = c(x, ls(envir, all.names = TRUE))
  }
  unique(x)
}

# find local variables in code (those getting assigned in the code)
find_locals = function(code) {
  code = if (is.language(code)) as.expression(code) else parse_only(code)
  codetools::findLocalsList(code)
}

# return a list of values of global variables in code
global_vars = function(code, envir) {
  if (length(vars <- find_globals(code, envir)) > 0)
    mget(vars, envir, inherits = TRUE)
}

#' Download a file from a URL and cache it on disk
#'
#' This object provides methods to download files and cache them on disk.
#' @format A list of methods:
#'
#' - `$get(url, type, handler)` downloads a URL, caches it, and returns the file
#'   content according to the value of `type` (possible values: `"text"` means
#'   the text content; `"base64"` means the base64 encoded data; `"raw"` means
#'   the raw binary content; `"auto"` is the default and means the type is
#'   determined by the content type in the URL headers). Optionally a `handler`
#'   function can be applied to the content.
#' - `$list()` gives the list of cache files.
#' - `$summary()` gives a summary of existing cache files.
#' - `$remove(url, type)` removes a single cache file.
#' - `$purge()` deletes all cache files.
#' @export
#' @examplesIf interactive()
#' # the first time it may take a few seconds
#' x1 = xfun::download_cache$get('https://www.r-project.org/')
#' head(x1)
#'
#' # now you can get the cached content
#' x2 = xfun::download_cache$get('https://www.r-project.org/')
#' identical(x1, x2)  # TRUE
#'
#' # a binary file
#' x3 = xfun::download_cache$get('https://yihui.org/images/logo.png', 'raw')
#' length(x3)
#'
#' # show a summary
#' xfun::download_cache$summary()
#' # remove a specific cache file
#' xfun::download_cache$remove('https://yihui.org/images/logo.png', 'raw')
#' # remove all cache files
#' xfun::download_cache$purge()
download_cache = local({
  pre = 'url'  # file prefix
  c_dir = function() {
    getOption('xfun.cache.dir', tools::R_user_dir('xfun', 'cache'))
  }
  c_file = function(url, type) {
    file.path(c_dir(), sprintf('%s_%s_%s.rds', pre, type, md5_one(url)))
  }
  read = function(url, type) {
    if (length(f <- c_file(url, type)) && file.exists(f)) readRDS(f)
  }
  write = function(url, type, data) {
    if (length(f <- c_file(url, type))) {
      dir_create(dirname(f))
      saveRDS(data, f)
    }
  }
  list_cache = function() {
    d = c_dir()
    list.files(d, sprintf('^%s_.+[.]rds$', pre), full.names = TRUE)
  }
  list(
    get = function(url, type = c('auto', 'text', 'base64', 'raw'), handler = NULL) {
      type = type[1]
      if (!is.null(x <- read(url, type))) return(x[[url]])
      if ((auto <- type == 'auto')) type = if (length(grep(
        '^content-type:\\s+(text/.+|[^;]+;\\s+charset=utf-8)\\s*$',
        curlGetHeaders(url), ignore.case = TRUE
      ))) 'text' else 'raw'
      dir_create(d <- tempfile())
      on.exit(unlink(d, recursive = TRUE), add = TRUE)
      x = in_dir(d, {
        o = url_filename(url)
        download_file(url, o)
        switch(
          type, text = read_utf8(o), base64 = base64_uri(o), raw = read_bin(o)
        )
      })
      if (is.function(handler)) x = handler(x)
      write(url, if (auto) 'auto' else type, setNames(list(x), url))
      x
    },
    summary = function() {
      f = list_cache()
      if (length(f) == 0) return(invisible())
      t = gsub('^url_([^_]+)_.+$', '\\1', basename(f))
      u = vapply(f, function(x) names(readRDS(x)), character(1))
      s = file.size(f)
      if (length(f) > 1) message('Total size: ', format_bytes(sum(s)))
      d = data.frame(url = u, type = t, size = s, size_h = format_bytes(s))
      rownames(d) = NULL
      unname(split(d, seq_len(nrow(d))))
    },
    list = list_cache,
    remove = function(url, type = 'auto') file.remove(c_file(url, type)),
    purge = function() {
      f = list_cache()
      s = file.size(f)
      i = file.remove(f)
      message(sprintf(
        "Purged %d cache file(s) from '%s' (%s)",
        sum(i), c_dir(), format_bytes(sum(s[i]))
      ))
    }
  )
})

#' Encode/decode data into/from base64 encoding.
#'
#' The function \code{base64_encode()} encodes a file or a raw vector into the
#' base64 encoding. The function \code{base64_decode()} decodes data from the
#' base64 encoding.
#' @param x For \code{base64_encode()}, a raw vector. If not raw, it is assumed
#'   to be a file or a connection to be read via \code{readBin()}. For
#'   \code{base64_decode()}, a string.
#' @param from If provided (and \code{x} is not provided), a connection or file
#'   to be read via \code{readChar()}, and the result will be passed to the
#'   argument \code{x}.
#' @return \code{base64_encode()} returns a character string.
#'   \code{base64_decode()} returns a raw vector.
#' @useDynLib xfun, .registration = TRUE
#' @export
#' @examples xfun::base64_encode(as.raw(1:10))
#' logo = xfun:::R_logo()
#' xfun::base64_encode(logo)
base64_encode = function(x) {
  if (!is.raw(x)) x = read_bin(x)
  .Call('base64_enc', x, PACKAGE = 'xfun')
}

#' @export
#' @rdname base64_encode
#' @examples xfun::base64_decode("AQIDBAUGBwgJCg==")
base64_decode = function(x, from = NA) {
  if (!is.na(from)) {
    if (!missing(x)) stop("Please provide either 'x' or 'from', but not both.")
    x = readChar(from, file.size(from), TRUE)
  }
  if (!is.character(x) || length(x) != 1) stop("'x' must be a single character string.")
  .Call('base64_dec', x, PACKAGE = 'xfun')
}

# an R implementation of base64 encoding by Wush Wu moved from knitr (of
# historic interest only): https://github.com/yihui/knitr/pull/324
base64_encode_r = function(x) {
  if (!is.raw(x)) x = read_bin(x)
  chars = c(LETTERS, letters, 0:9, '+', '/')
  n = length(s <- as.integer(x))
  res = rep(NA, (n + 2) / 3 * 4)
  i = 0L  # index of res vector
  j = 1L  # index of base64_table
  while (n > 2L) {
    res[i <- i + 1L] = chars[s[j] %/% 4L + 1L]
    res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
    res[i <- i + 1L] = chars[4L * (s[j + 1L] %% 16) + s[j + 2L] %/% 64L + 1L]
    res[i <- i + 1L] = chars[s[j + 2L] %% 64L + 1L]
    j = j + 3L
    n = n - 3L
  }
  if (n) {
    res[i <- i + 1L] = chars[s[j] %/% 4L + 1L]
    if (n > 1L) {
      res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
      res[i <- i + 1L] = chars[4L * (s[j + 1L] %% 16) + 1L]
      res[i <- i + 1L] = '='
    } else {
      res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + 1L]
      res[i <- i + 1L] = '='
      res[i <- i + 1L] = '='
    }
  }
  paste(res[!is.na(res)], collapse = '')
}

#' Generate the Data URI for a file
#'
#' Encode the file in the base64 encoding, and add the media type. The data URI
#' can be used to embed data in HTML documents, e.g., in the \code{src}
#' attribute of the \verb{<img />} tag.
#' @param x A file path.
#' @param type The MIME type of the file, e.g., \code{"image/png"} for a PNG
#'   image file.
#' @return A string of the form \verb{data:<media type>;base64,<data>}.
#' @note By default, this function requires the \pkg{mime} package to determine
#'   the MIME type of the file.
#' @export
#' @examples
#' logo = xfun:::R_logo()
#' img = htmltools::img(src = xfun::base64_uri(logo), alt = 'R logo')
#' if (interactive()) htmltools::browsable(img)
base64_uri = function(x, type = mime::guess_type(x)) {
  if (missing(type)) type = guess_type(x)
  paste0("data:", type, ";base64,", base64_encode(x))
}

# a limited version of mime::guess_type()
guess_type = function(x, use_mime = loadable('mime')) {
  if (use_mime) return(mime::guess_type(x))
  res = mimemap[tolower(file_ext(x))]
  if (any(i <- is.na(res))) {
    warning(
      'Cannot determine the MIME type(s) of ', paste(x[i], collapse = ', '),
      '. You may try to install the "mime" package or report an issue to ',
      packageDescription('xfun')$BugReports, '.'
    )
    res[i] = 'application/octet-stream'
  }
  unname(res)
}

# a comprehensive version is mime::mimemap (can extend it upon user request)
mimemap = c(
  css = 'text/css', csv = 'text/csv', gif = 'image/gif', jpeg = 'image/jpeg',
  jpg = 'image/jpeg', js = 'application/javascript', png = 'image/png',
  svg = 'image/svg+xml', ttf = 'application/font-sfnt',
  woff = 'application/font-woff', woff2 = 'application/octet-stream'
)

#' Generate BibTeX bibliography databases for R packages
#'
#' Call [utils::citation()] and [utils::toBibtex()] to create bib entries for R
#' packages and write them in a file. It can facilitate the auto-generation of
#' bibliography databases for R packages, and it is easy to regenerate all the
#' citations after updating R packages.
#'
#' For a package, the keyword `R-pkgname` is used for its bib item, where
#' `pkgname` is the name of the package. Citation entries specified in the
#' \file{CITATION} file of the package are also included. The main purpose of
#' this function is to automate the generation of the package citation
#' information because it often changes (e.g., the author, year, package
#' version, and so on).
#'
#' There are at least two different uses for the URL in a reference list. You
#' might want to tell users where to go for more information. In that case, use
#' the default `packageURL = TRUE`, and the first URL listed in the
#' \file{DESCRIPTION} file will be used. Be careful: some authors don't put the
#' most relevant URL first. Alternatively, you might want to identify exactly
#' which version of the package was used in the document. If it was installed
#' from CRAN or some other repositories, the version number identifies it, and
#' `packageURL = FALSE` will use the repository URL (as used by
#' [utils::citation()]).
#' @param x Package names. Packages which are not installed are ignored.
#' @param file The (\file{.bib}) file to write. By default, or if `NULL`, output
#'   is written to the R console.
#' @param tweak Whether to fix some known problems in the citations, especially
#'   non-standard format of author names.
#' @param width Width of lines in bibliography entries. If `NULL`, lines will
#'   not be wrapped.
#' @param prefix Prefix string for keys in BibTeX entries; by default, it is
#'   `R-` unless `option('xfun.bib.prefix')` has been set to another string.
#' @param lib.loc A vector of path names of R libraries.
#' @param packageURL Use the `URL` field from the \file{DESCRIPTION} file. See
#'   Details below.
#' @param date The way to include a date of the form `(retrieved YYYY-mm-dd)` in
#'   the `note` field of bib entries. For `date = 'none'`, no date is included.
#'   For `date = 'package'`, the [packageDate()] of the package is included. For
#'   `date = 'today'`, [Sys.Date()] is included. Alternatively, you can provide
#'   a date directly, e.g., `date = '2025-04-01'.`
#' @return A list containing the citations. Citations are also written to the
#'   `file` as a side effect.
#' @note Some packages on CRAN do not have standard bib entries, which was once
#'   reported by Michael Friendly at
#'   \url{https://stat.ethz.ch/pipermail/r-devel/2010-November/058977.html}. I
#'   find this a pain, and there are no easy solutions except contacting package
#'   authors to modify their DESCRIPTION files. The argument `tweak` has
#'   provided hacks to deal with known packages with non-standard bib entries;
#'   `tweak = TRUE` is by no means intended to hide or modify the original
#'   citation information. It is just due to the loose requirements on package
#'   authors for the DESCRIPTION file. On one hand, I apologize if it really
#'   mangles the information about certain packages; on the other, I strongly
#'   recommend package authors to consider the `Authors@@R` field (see the
#'   manual _Writing R Extensions_) to make it easier for other people to cite R
#'   packages.
#' @export
#' @author Yihui Xie and Michael Friendly
#' @examples
#' pkg_bib(c('base', 'MASS', 'xfun'))
#' pkg_bib('cluster', prefix = 'R-pkg-')  # a different prefix
#' pkg_bib('xfun', date = 'package')
pkg_bib = function(
  x = .packages(), file = '', tweak = TRUE, width = NULL,
  prefix = getOption('xfun.bib.prefix', 'R-'), lib.loc = NULL,
  packageURL = TRUE, date = c('none', 'package', 'today')
) {
  system.file = function(...) base::system.file(..., lib.loc = lib.loc)
  citation = function(...) utils::citation(..., lib.loc = lib.loc)
  date = date[1]
  x = x[nzchar(x)] # remove possible empty string
  idx = mapply(system.file, package = x) == ''
  if (any(idx)) {
    warning('package(s) ', paste(x[idx], collapse = ', '), ' not found')
    x = x[!idx]
  }
  # no need to write bib for packages in base R other than `base` itself
  x = setdiff(x, setdiff(base_pkgs(), 'base'))
  x = sort(x)
  bib = sapply(x, function(pkg) {
    meta = packageDescription(pkg, lib.loc)
    # don't use the citation() URL if the package has provided its own URL
    cite = citation(pkg, auto = if (is.null(meta$URL)) meta else {
      if (packageURL) meta$Repository = meta$RemoteType = NULL
      # use the first URL in case the package provided multiple URLs
      meta$URL = sub('[, \t\n].*', '', meta$URL)
      meta
    })
    if (tweak) {
      # e.g. gpairs has "gpairs: " in the title
      cite$title = gsub(sprintf('^(%s: )(\\1)', pkg), '\\1', cite$title)
    }
    date = format(switch(
      date, none = '', package = packageDate(pkg, lib.loc), today = Sys.Date(), date
    ))
    if (!is.na(date) && date != '') cite$note = paste0(cite$note, ' (retrieved ', date, ')')
    entry = toBibtex(cite)
    entry[1] = sub('\\{,$', sprintf('{%s%s,', prefix, pkg), entry[1])
    entry
  }, simplify = FALSE)
  if (tweak) {
    # packages with non-standard author fields; use maintainer instead
    pkgs = c(
      "ash", "cat", "CircStats", "Fahrmeir", "flashClust", "leaps", "mapproj",
      "maps", "pbivnorm", "quadprog", "R2WinBUGS", "Sleuth2", "sm"
    )
    for (i in intersect(pkgs, x)) {
      message('tweaking ', i)
      maintainer = packageDescription(i, lib.loc, fields = 'Maintainer')
      bib[[i]]['author'] = sprintf('  author = {%s},', gsub('\\s+<.*>\\s*$', '', maintainer))
    }
    bib = lapply(bib, function(b) {
      b['author'] = sub('Duncan Temple Lang', 'Duncan {Temple Lang}', b['author'])
      # remove the ugly single quotes required by CRAN policy
      b['title'] = gsub("(^|\\W)'([^']+)'(\\W|$)", '\\1\\2\\3', b['title'])
      # keep the first URL if multiple are provided
      if (!is.na(b['note'])) b['note'] = gsub(
        '(^.*?https?://.*?),\\s+https?://.*?(},\\s*)$', '\\1\\2', b['note']
      )
      if (!('year' %in% names(b))) b['year'] = .this.year
      b
    })
  }
  # also read citation entries from the CITATION file if provided
  bib2 = lapply(x, function(pkg) {
    if (pkg == 'base') return()
    if (system.file('CITATION', package = pkg) == '') return()
    cites = citation(pkg, auto = FALSE)
    cites = Filter(x = cites, function(cite) {
      # exclude entries identical to citation(pkg, auto = TRUE)
      !isTRUE(grepl('R package version', cite$note))
    })
    s = make_unique(unlist(lapply(cites, function(cite) {
      if (is.null(cite$year)) format(Sys.Date(), '%Y') else cite$year
    })))
    mapply(cites, s, FUN = function(cite, suffix) {
      # the entry is likely to be the same as citation(pkg, auto = TRUE)
      if (isTRUE(grepl('R package version', cite$note))) return()
      entry = toBibtex(cite)
      entry[1] = sub('\\{,$', sprintf('{%s%s,', pkg, suffix), entry[1])
      entry
    }, SIMPLIFY = FALSE)
  })
  bib = c(bib, unlist(bib2, recursive = FALSE))
  bib = lapply(bib, function(b) {
    idx = which(names(b) == '')
    if (!is.null(width)) b[-idx] = str_wrap(b[-idx], width, 2, 4)
    lines = c(b[idx[1L]], b[-idx], b[idx[2L]], '')
    if (tweak) {
      # e.g. KernSmooth and spam has & in the title and the journal, respectively
      lines = gsub('(?<!\\\\)&', '\\\\&', lines, perl = TRUE)
    }
    structure(lines, class = 'Bibtex')
  })
  if (!is.null(file) && length(x)) write_utf8(unlist(bib), file)
  invisible(bib)
}

.this.year = sprintf('  year = {%s},', format(Sys.Date(), '%Y'))

# c(1, 1, 1, 2, 3, 3) -> c(1a, 1b, 1c, 2a, 3a, 3b)
make_unique = function(x) {
  if (length(x) == 0) return(x)
  x2 = make.unique(x)
  if (all(i <- x2 == x)) return(x)
  x2[i] = paste0(x2[i], '.0')
  i = as.numeric(sub('.*[.]([0-9]+)$', '\\1', x2)) + 1
  s = letters[i]
  s = ifelse(is.na(s), i, s)
  paste0(x, s)
}

#' An alternative to sessionInfo() to print session information
#'
#' This function tweaks the output of \code{\link{sessionInfo}()}: (1) It adds
#' the RStudio version information if running in the RStudio IDE; (2) It removes
#' the information about matrix products, BLAS, and LAPACK; (3) It removes the
#' names of base R packages; (4) It prints out package versions in a single
#' group, and does not differentiate between loaded and attached packages.
#'
#' It also allows you to only print out the versions of specified packages (via
#' the \code{packages} argument) and optionally their recursive dependencies.
#' For these specified packages (if provided), if a function
#' \code{xfun_session_info()} exists in a package, it will be called and
#' expected to return a character vector to be appended to the output of
#' \code{session_info()}. This provides a mechanism for other packages to inject
#' more information into the \code{session_info} output. For example,
#' \pkg{rmarkdown} (>= 1.20.2) has a function \code{xfun_session_info()} that
#' returns the version of Pandoc, which can be very useful information for
#' diagnostics.
#' @param packages A character vector of package names, of which the versions
#'   will be printed. If not specified, it means all loaded and attached
#'   packages in the current R session.
#' @param dependencies Whether to print out the versions of the recursive
#'   dependencies of packages.
#' @return A character vector of the session information marked as
#'   \code{\link{raw_string}()}.
#' @export
#' @examplesIf interactive()
#' xfun::session_info()
#' if (xfun::loadable('MASS')) xfun::session_info('MASS')
session_info = function(packages = NULL, dependencies = TRUE) {
  res = sessionInfo()
  res$matprod = res$BLAS = res$LAPACK = NULL
  if (loadable('rstudioapi') && rstudioapi::isAvailable()) {
    res$running = paste0(res$running, ', RStudio ', rstudioapi::getVersion())
  }

  tweak_info = function(obj, extra = NULL) {
    res = capture.output(print(obj))
    i = grep('^(attached base packages|Matrix products):\\s*$', res, ignore.case = TRUE)
    if (length(i)) res = res[-c(i, i + 1)]
    res = gsubi('^\\s*locale:\\s*$', 'Locale:', res)
    res = gsub('^\\s*\\[[0-9]+]\\s*', '  ', res)  # remove vector indices like [1]
    res = gsubi('^\\s*other attached packages:\\s*$', 'Package version:', res)
    # print the locale info on a single line if possible
    if (length(i <- which(res == 'Locale:')) == 1 && res[i + 2] == '') {
      res[i] = paste(res[i], gsub('\\s*/\\s*', ' / ', gsub('^\\s+', '', res[i + 1])))
      res = res[-(i + 1)]
    }
    raw_string(c(res, extra))
  }

  version_info = function(pkgs) {
    res = lapply(pkgs, function(p) {
      list(Version = as.character(packageVersion(p)), Package = p)
    })
    as.list(setNames(res, pkgs))
  }

  res$basePkgs = raw_string(list())
  info = c(res$otherPkgs, res$loadedOnly)
  packages = setdiff(packages, '')  # remove empty strings (#65)
  if (length(packages) > 0) {
    info = info[intersect(names(info), packages)]
    info = c(info, version_info(setdiff(packages, names(info))))
  }
  res$loadedOnly = NULL
  if (dependencies) {
    deps = pkg_dep(names(info), installed.packages(), recursive = TRUE)
    deps = sort(setdiff(deps, names(info)))
    info = c(info, version_info(deps))
  }
  if (length(packages) > 0 || dependencies) info = info[sort(names(info))]
  res$otherPkgs = info
  extra = unlist(lapply(packages, function(p) tryCatch(
    c('', getFromNamespace('xfun_session_info', p)()), error = function(e) NULL)
  ))

  tweak_info(res, extra)
}

#' Perform a task once in an R session
#'
#' Perform a task once in an R session, e.g., emit a message or warning. Then
#' give users an optional hint on how not to perform this task at all.
#' @param task Any R code expression to be evaluated once to perform a task,
#'   e.g., \code{warning('Danger!')} or \code{message('Today is ', Sys.Date())}.
#' @param option An R option name. This name should be as unique as possible in
#'   \code{\link{options}()}. After the task has been successfully performed,
#'   this option will be set to \code{FALSE} in the current R session, to
#'   prevent the task from being performed again the next time when
#'   \code{do_once()} is called.
#' @param hint A character vector to provide a hint to users on how not to
#'   perform the task or see the message again in the current R session. Set
#'   \code{hint = ""} if you do not want to provide the hint.
#' @return The value returned by the \code{task}, invisibly.
#' @export
#' @examples
#' do_once(message("Today's date is ", Sys.Date()), "xfun.date.reminder")
#' # if you run it again, it will not emit the message again
#' do_once(message("Today's date is ", Sys.Date()), "xfun.date.reminder")
#'
#' do_once({Sys.sleep(2); 1 + 1}, "xfun.task.1plus1")
#' do_once({Sys.sleep(2); 1 + 1}, "xfun.task.1plus1")
do_once = function(task, option, hint = c(
  'You will not see this message again in this R session.',
  'If you never want to see this message,',
  sprintf('you may set options(%s = FALSE) in your .Rprofile.', option)
)) {
  if (isFALSE(getOption(option))) return(invisible())
  task
  hint = paste(hint, collapse = ' ')
  if (hint != '') message(hint)
  options(setNames(list(FALSE), option))
  invisible(task)
}

#' Get the tags of GitHub releases of a repository
#'
#' Use the GitHub API ([github_api()]) to obtain the tags of the
#' releases.
#' @param repo The repository name of the form `user/repo`, e.g.,
#'   `"yihui/xfun"`.
#' @param tag A tag as a character string. If provided, it will be returned if
#'   the tag exists. If `tag = "latest"`, the tag of the latest release is
#'   returned.
#' @param pattern A regular expression to match the tags.
#' @param use_jsonlite Whether to use \pkg{jsonlite} to parse the releases info.
#' @export
#' @return A character vector of (GIT) tags.
#' @examplesIf interactive()
#' xfun::github_releases('yihui/litedown')
#' xfun::github_releases('gohugoio/hugo', 'latest')
github_releases = function(
  repo, tag = '', pattern = 'v[0-9.]+', use_jsonlite = loadable('jsonlite')
) {
  if (tag != '') return(github_releases2(repo, tag, pattern))

  i = 1; v = character()
  repeat {
    res = github_api(
      sprintf('/repos/%s/tags', repo), NULL, list(per_page = 100, page = i),
      raw = !use_jsonlite
    )
    v2 = unlist(if (use_jsonlite) {
      lapply(res, `[[`, 'name')
    } else {
      reg_match('\\{"name":"([^"]+)",', res)
    })
    if (length(v2) == 0) break
    v = c(v, v2)
    if (length(v2) < 100) break  # not enough items for the next page
    i = i + 1
  }
  grep(sprintf('^%s$', pattern), unique(v), value = TRUE)
}

# extract the matched elements in the n-th pair of () in the regex
reg_match = function(p, x, n = 1, ...) {
  # TODO: gregexec was added in R 4.1.0; remove this workaround when we don't
  # need to support R < 4.1.0
  v = 'gregexec' %in% ls(baseenv())
  m = (if (v) base::gregexec else base::gregexpr)(p, x, ...)
  lapply(regmatches(x, m), function(x) {
    if (v) x[n + 1, ] else gsub(p, paste0('\\', n), x)
  })
}

# the fallback method to retrieve release tags (read HTML source)
github_releases2 = function(repo, tag = '', pattern = '[^"&]+') {
  read = function() suppressWarnings(
    read_utf8(sprintf('https://github.com/%s/releases/%s', repo, tag))
  )
  h = if (tag == '') read() else tryCatch(read(), error = function(e) '')
  r = sprintf('^.*?%s/releases/tag/(%s)".*', repo, pattern)
  v = unique(grep_sub(r, '\\1', h))
  # if a tag is specified, use the 1st value because other tags may be mentioned
  # in the current tag page
  if (tag == '') v else head(v, 1)
}

#' @details `github_api()` is a wrapper function based on
#'   `rest_api_raw()` to obtain data from the GitHub API:
#'   <https://docs.github.com/en/rest>. You can provide a personal access
#'   token (PAT) via the `token` argument, or via one of the environment
#'   variables \var{GITHUB_PAT}, \var{GITHUB_TOKEN}, \var{GH_TOKEN}. A PAT
#'   allows for a much higher rate limit in API calls. Without a token, you can
#'   only make 60 calls in an hour.
#' @param raw Whether to return the raw response or parse the response with
#'   \pkg{jsonlite}.
#' @rdname rest_api
#' @export
github_api = function(
  endpoint, token = '', params = list(), headers = NULL, raw = !loadable('jsonlite')
) {
  token = c(token, github_token())
  token = if (length(token <- token[token != ''])) token[1] else ''
  names(token) = 'token'
  error = TRUE
  on.exit(if (error) github_token(NA, token))
  res = rest_api_raw('https://api.github.com', endpoint, token, params, headers)
  error = FALSE
  if (raw) res else jsonlite::fromJSON(res, FALSE)
}

# the environment variables to check for a GitHub personal access token
github_envs = c('GITHUB_PAT', 'GITHUB_TOKEN', 'GH_TOKEN')

github_token = function(error = FALSE, token = unname(Sys.getenv(github_envs))) {
  if (length(token <- token[token != ''])) return(token[1])
  msg = c(
    'You may need to save a GitHub personal access token in one of the ',
    'environment variables: ', paste(github_envs, collapse = ', ')
  )
  if (is.na(error)) message(msg) else if (error) stop(msg)
  ''
}

git = function(...) {
  if (Sys.which('git') == '') stop('git is not available')
  # R's HOME var is different from the system's HOME on Windows:
  # https://github.com/yihui/crandalf/issues/24
  if (is_windows()) {
    env = set_envvar(c(HOME = Sys.getenv('USERPROFILE')))
    on.exit(set_envvar(env), add = TRUE)
  }
  system3('git', ...)
}

git_co = function(args = NULL, ...) {
  git(c('checkout', args), ...)
}

git_test_branch = function() {
  if (length(d <- git(c('diff', '--name-only'), stdout = TRUE))) stop(
    'The current branch has changes not stated for commit:\n',
    paste(d, collapse = '\n')
  )
}

# pull from the current branch if it is a git repository and has a current branch
git_pull_current <- function() {
  is_repo = git(c('rev-parse', '--is-inside-work-tree'), stdout = FALSE, stderr = FALSE)
  if (is_repo == 0) {
    branch = git(c('branch', '--show-current'), stdout = TRUE)
    if (length(branch) == 1 && branch != '') {
      if (length(git('diff', stdout = TRUE))) {
        git('stash'); on.exit(git(c('stash', 'pop')), add = TRUE)
      }
      git(c('pull', 'origin', branch))
    }
  }
}

gh = function(...) {
  if (Sys.which('gh') == '') stop('GitHub CLI not found: https://cli.github.com')
  system3('gh', ...)
}

gh_run = function(..., repo = NA) {
  gh(c(if (!is.na(repo)) c('-R', repo), 'run', ...), stdout = TRUE)
}

# Detect the sprintf template for NEWS.md headers from existing content.
# Falls back to '# CHANGES IN %s VERSION %s' when format cannot be inferred.
news_tpl = function(news, pkg, ver) {
  h1 = grep('^# ', news, value = TRUE)[1]
  tpl = sub(sprintf('\\b%s\\b', ver), '%s', sub(sprintf('\\b%s\\b', pkg), '%s', h1))
  if (grepl('%s.*%s', tpl)) tpl else '# CHANGES IN %s VERSION %s'
}

#' Perform post-CRAN-release routine tasks
#'
#' After a CRAN release is accepted, commit and tag the release, bump the
#' package version to the next patch dev version, prepend a new header to
#' \file{NEWS.md}, push to the remote repository, and create a GitHub release
#' with the news items of this version.
#'
#' The version number `X.Y` or `X.Y.0` is extracted from the `Version` field in
#' \file{DESCRIPTION}. For example, if the current version is `0.58` (or
#' `0.58.0`), the repo will be tagged as `v0.58` (or `v0.58.0`). The patch
#' version is then bumped to `0.58.1`, and a new header for the next minor
#' version `0.59` (or `0.59.0`) is prepended to \file{NEWS.md}.
#' @export
post_release = function() {
  desc = read.dcf('DESCRIPTION')
  pkg = desc[, 'Package']
  ver = desc[, 'Version']

  vers = release_versions(ver)
  ver_next = vers$next_ver
  ver_patch = vers$patch

  # Step 1: commit all current changes (only if Version was modified)
  ver_in_head = system3(
    'git', c('show', 'HEAD:DESCRIPTION'), stdout = TRUE, stderr = FALSE
  )
  ver_in_head = trimws(grep_sub('^Version:\\s+(.*)', '\\1', ver_in_head))
  if (!identical(ver_in_head, ver)) {
    message('Committing CRAN release v', ver, ' ...')
    git(c('commit', '-a', '-m', sprintf('CRAN release v%s', ver)))
  }

  # Step 2: tag
  message('Tagging v', ver, ' ...')
  git(c('tag', sprintf('v%s', ver)))

  # Step 3: bump patch version and prepend NEWS.md
  process_file('DESCRIPTION', function(x) {
    ver_line = grep('^Version:\\s+', x)
    x[ver_line] = sprintf('Version: %s', ver_patch)
    x
  })

  news = trimws(read_utf8('NEWS.md'))
  tpl = news_tpl(news, pkg, ver)
  news_hdr = function(v) sprintf(tpl, pkg, v)

  process_file('NEWS.md', function(x) c(paste0(news_hdr(ver_next), '\n\n'), x))

  hdr = news_hdr(ver)
  i = which(hdr == news)
  if (length(i) == 0) stop('Cannot find the header "', hdr, '" in NEWS.md')

  message('Committing the start of the next version ...')
  git(c('commit', '-a', '-m', 'start the next version'))

  # Step 4: push with tags
  message('Pushing to remote ...')
  git(c('push', '--tags', 'origin', 'HEAD'))

  # Step 5: create GitHub release
  github_token(TRUE)
  j = grep('^# ', news)
  j = j[j > i]
  end = if (length(j) > 0) j[1] - 1 else length(news)
  items = strip_blank(news[(i + 1):end])

  f = tempfile()
  writeLines(items, f)
  on.exit(unlink(f), add = TRUE)

  gh(c(
    'release', 'create', sprintf('v%s', ver),
    '--title', sprintf('%s %s', pkg, ver), '--notes-file', f)
  )

  invisible()
}

library(testit)

# test yaml_load() error handler (requires the yaml package)
if (loadable('yaml')) assert('yaml_load() shows helpful error message on invalid YAML', {
  invalid_yaml = 'a: [\nunclosed bracket'
  msgs = c()
  tryCatch(
    withCallingHandlers(
      yaml_load(invalid_yaml),
      message = function(m) {
        msgs <<- c(msgs, conditionMessage(m)); invokeRestart('muffleMessage')
      }
    ),
    error = function(e) NULL
  )
  (gsub('^Failed to parse YAML:.*', '', msgs) %==% '')
})

#' Read YAML data
#'
#' If the \pkg{yaml} package is installed, use [yaml::yaml.load()] to read the
#' data. If not, use a simple parser instead, which only supports a limited
#' number of data types (see \dQuote{Examples}). In particular, it does not
#' support values that span across multiple lines (such as multi-line text).
#' @param x A character vector of YAML data.
#' @param ...,handlers Arguments to be passed to [yaml::yaml.load()].
#' @param envir The environment in which R expressions in YAML are evaluated. To
#'   disable the evaluation, use `envir = FALSE`.
#' @param use_yaml Whether to use the \pkg{yaml} package.
#' @return An R object (typically a list).
#' @note R expressions in YAML will be returned as [expression]s when they are
#'   not evaluated. This is different with [yaml::yaml.load()], which returns
#'   character strings for expressions.
#' @export
#' @examples
#' # test the simple parser without using the yaml package
#' read_yaml = function(...) xfun::yaml_load(..., use_yaml = FALSE)
#' read_yaml('a: 1')
#' read_yaml('a: 1\nb: "foo"\nc: null')
#' read_yaml('a:\n  b: false\n  c: true\n  d: 1.234\ne: bar')
#' read_yaml('a: !expr paste(1:10, collapse = ", ")')
#' read_yaml('a: [1, 3, 4, 2]')
#' read_yaml('a: [1, "abc", 4, 2]')
#' read_yaml('a: ["foo", "bar"]')
#' read_yaml('a: [true, false, true]')
#' # the other form of array is not supported
#' read_yaml('a:\n  - b\n  - c')
#' # and you must use the yaml package
#' if (loadable('yaml')) yaml_load('a:\n  - b\n  - c')
yaml_load = function(
  x, ..., handlers = NULL, envir = parent.frame(), use_yaml = loadable('yaml')
) {
  if (use_yaml) return(handle_error(
    yaml::yaml.load(x, eval.expr = FALSE, handlers = yaml_handlers(handlers, envir), ...),
    function(loc) {
      s = geterrmessage()
      r = 'line (\\d+), column (\\d+)'
      m = regmatches(s, regexec(r, s, perl = TRUE))[[1]]
      if (length(m) < 3) return()
      m = as.integer(m[-1])  # c(row, col)
      if (loc != '') loc = paste(' at lines', loc)
      c(
        sprintf('Failed to parse YAML%s:', loc), '',
        append(x, paste0(strrep(' ', m[2]), '^~~~~~'), m[1]), ''
      )
    }
  ))
  # the below simple parser is quite limited
  res = list()
  r = '^( *)([^ ]+?):($|\\s+.*)'
  x = split_lines(x)
  x = x[grep(r, x)]
  x = x[grep('^\\s*#', x, invert = TRUE)]  # comments
  if (length(x) == 0) return(res)
  lvl = gsub(r, '\\1', x)  # indentation level
  key = gsub(r, '\\2', x)
  val = gsub('^\\s*|\\s*$', '', gsub(r, '\\3', x))
  keys = NULL
  for (i in seq_along(x)) {
    keys = c(head(keys, nchar(lvl[i])/2), key[i])
    v = if (is_blank(val[i])) list() else yaml_value(val[i], envir)
    # special treatment of NULL (to preserve a key with a null value)
    if (is.null(v)) {
      if (length(keys) <= 1) res[keys] = list(v) else {
        res[[head(keys, -1)]][tail(keys, 1)] = list(v)
      }
    } else res[[keys]] = v
  }
  res
}

# only support logical, numeric, character values (both scalar and [] arrays),
# and R expressions starting with !r/!expr
yaml_value = function(x, envir = parent.frame()) {
  v = tolower(x)
  if (v == 'null') return()
  if (grepl('^true|false$', v)) return(as.logical(x))
  if (grepl(r <- '^\\s*\\[(.*)\\]\\s*$', v)) {
    v = gsub(r, '\\1', v)
    if (is_blank(v)) return(list())
    v = unname(unlist(read.csv(text = v, header = FALSE)))
    if (is.numeric(v)) return(v)
    v = gsub('^ ', '', v)  # [a, b] -> ["a", " b"] -> ["a", "b"]
    return(if (all(grepl('^true|false$', v))) as.logical(v) else v)
  }
  if (grepl('^[0-9.e+-]', v)) {
    v = suppressWarnings(as.numeric(v))
    if (!is.na(v)) return(if ((v2 <- as.integer(v)) == v) v2 else v)
  }
  x = gsub('^["\']|["\']$', '', x)  # remove optional quotes for strings
  if (grepl(r <- '^!(r|expr) (.+)$', x)) {
    x = yaml_expr(gsub(r, '\\2', x), envir)
  }
  x
}

yaml_expr = function(x, envir) {
  x = parse_only(x)
  if (is.environment(envir)) x = eval(x, envir)
  x
}

# add !r and !expr handlers to support a custom eval environment, which
# yaml::yaml.load() doesn't support by default (vubiostat/r-yaml#54)
yaml_handlers = function(h, envir) {
  h = as.list(h)
  f = function(x) yaml_expr(x, envir)
  for (i in c('r', 'expr')) if (is.null(h[[i]])) h[[i]] = f
  h
}

#' Partition the YAML metadata and the body in a document
#'
#' Split a document into the YAML metadata (which starts with `---` in the
#' beginning of the document) and the body. The YAML metadata will be parsed.
#' @param x A character vector of the document content.
#' @param ... Arguments to be passed to `yaml_load()`.
#' @export
#' @return A list of components `yaml` (the parsed YAML data), `lines` (starting
#'   and ending line numbers of YAML), and `body` (a character vector of the
#'   body text). If YAML metadata does not exist in the document, the components
#'   `yaml` and `lines` will be missing.
#' @examples
#' xfun::yaml_body(c('---', 'title: Hello', 'output: markdown::html_document', '---', '', 'Content.'))
yaml_body = function(x, ...) {
  n = length(x)
  res = if (length(i <- locate_yaml(x)) == 0) {
    list(body = x)
  } else list(
    yaml = x[i[1]:i[2]], body = c(rep('', i[2]), tail(x, n - i[2])), lines = i
  )
  if ((n <- length(res$yaml)) >= 2) {
    res['yaml'] = list(yaml_load(res$yaml[-c(1, n)], ...))
  }
  res
}

# find lines of YAML
locate_yaml = function(x) {
  i = grep('^---\\s*$', x)
  if (length(i) > 1 && all(is_blank(x[seq_len(i[1] - 1)]))) i[1:2]
}

has_fun = function(name, envir = baseenv()) exists(name, envir, inherits = FALSE)

# for R < 4.0
if (!has_fun('isFALSE')) isFALSE = function(x) {
  is.logical(x) && length(x) == 1 && !is.na(x) && !x
}
if (!has_fun('startsWith')) startsWith = function(x, s) {
  substring(x, 1, nchar(s)) == s
}
if (!has_fun('endsWith')) endsWith = function(x, s) {
  n = nchar(x)
  substring(x, n - nchar(s) + 1, n) == s
}
if (!has_fun('strrep')) strrep = function(x, times) {
  mapply(function(x, n) paste(rep(x, n), collapse = ''), x, times, USE.NAMES = FALSE)
}
if (!has_fun('packageDate', asNamespace('utils'))) packageDate = function(...) {
  as.Date(packageDescription(..., fields = 'Date/Publication'))
}

library(testit)

if ('CRAN_package_db' %in% getNamespaceExports('tools')) assert("pkg_maintainers() works", {
  x = pkg_maintainers("xfun")
  (length(x) > 0)
  # %==% not working for some reason
  (x %==% c(xfun = packageDescription('xfun', fields = 'Maintainer')))
})

library(testit)

record2 = function(...) unlist(record(...))

assert('record() works', {
  (record2('quote(x)') %==% c('quote(x)', 'x'))
  (record2('quote(x+y)') %==% c('quote(x+y)', 'x + y'))
})

assert('record() handles raw_string objects', {
  s = raw_string(c('line 1', 'line 2\nwith newline', 'line 3 "with quotes"'))
  rec = record('s')
  (as.character(rec[[2]]) %==% as.character(s))
})

if (loadable('data.table')) assert('record() works with data.table', {
  x = c('d = data.table::as.data.table(mtcars)', 'd[, mpg := 99]')
  rec = record(x, envir = globalenv())
  (length(rec) %==% 1L)  # only source code, no output printed
  (as.character(rec[[1]]) %==% x)
})

assert('record() avoids recording empty plots', {
  rec = record('library(grid); g <- pointsGrob(); x <- convertUnit(g$x, "mm")')
  (length(rec) %==% 1L)  # only source code, no plot generated
})

assert('record() can selectively keep plots', {
  f = function(i = TRUE) record(c('plot(1)', 'plot(2)'), dev.keep = i)
  cl = function(x) sub('^record_', '', sapply(x, class))
  (cl(f()) %==% rep(c('source', 'plot'), 2))  # keep all plots
  (cl(f(-1)) %==% c('source', 'plot'))  # remove 1st plot
  (cl(f(c(FALSE, TRUE))) %==% c('source', 'plot'))  # ditto
  (cl(f('last')) %==% c('source', 'plot'))  # keep the last plot only
  (cl(f(-2)) %==% c('source', 'plot', 'source'))  # remove 2nd plot
  (cl(f(FALSE)) %==% 'source')  # remove all plots
})

# the option try.outFile was introduced in R 3.4.0
opts = options(try.outFile = NULL)
if (getRversion() >= '3.4.0') assert('record() can capture try() messages', {
  rec = record('try(stop("asdf qwer zxcv"))')
  (length(rec) %==% 2L)
  (grepl('asdf qwer zxcv', rec[[2]]))
})
options(opts)

assert('record() returns empty result for NULL code', {
  res = record(NULL)
  (inherits(res, 'xfun_record_results'))
  (length(res) %==% 0L)
})

assert('record() captures messages', {
  res = record('message("hello")', message = TRUE)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  ('message' %in% cls)
  (any(grepl('hello', unlist(res))))
})

assert('record() suppresses messages when message = FALSE', {
  res = record('message("hidden")', message = FALSE)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  (!('message' %in% cls))
})

assert('record() captures warnings', {
  res = record('warning("oops")', warning = TRUE)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  ('warning' %in% cls)
  (any(grepl('oops', unlist(res))))
})

assert('record() captures errors when error = TRUE', {
  res = record('stop("boom")', error = TRUE)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  ('error' %in% cls)
  (any(grepl('boom', unlist(res))))
})

assert('record() handles syntax errors gracefully', {
  res = record('1 +', error = TRUE)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  ('source' %in% cls)
  ('error' %in% cls)
})

assert('record() with verbose = 2 prints invisible values', {
  res = record('invisible(42)', verbose = 2)
  (any(grepl('42', unlist(res))))
})

assert('record() with verbose = 1 prints last invisible value', {
  res = record(c('invisible(1)', 'invisible(2)'), verbose = 1)
  (any(grepl('2', unlist(res))))
})

assert('record() captures text output', {
  res = record('cat("hello world")')
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  ('output' %in% cls)
  (any(grepl('hello world', unlist(res))))
})

assert('format.xfun_record_results() works for text output', {
  res = record(c('1 + 1', 'cat("hi")'))
  txt = format(res, to = 'text')
  (is.character(txt))
  (any(grepl('1 \\+ 1', txt)))
  (any(grepl('hi', txt)))
})

assert('format.xfun_record_results() works for markdown output', {
  res = record(c('1 + 1'))
  md = format(res, to = 'markdown')
  (is.character(md))
  (any(grepl('```', md)))
})

assert('format.xfun_record_results() works for html output', {
  res = record(c('1 + 1'))
  html = format(res, to = 'html')
  (is.character(html))
  (any(grepl('<pre', html)))
  (any(grepl('</code></pre>', html)))
})

assert('format.xfun_record_results() html with template', {
  res = record('1 + 1')
  html = format(res, to = 'html', template = TRUE)
  (any(grepl('<html', html) | grepl('<!DOCTYPE', html)))
})

assert('print.xfun_record_results() works in non-interactive mode', {
  res = record('1 + 1')
  out = capture.output(print(res, browse = FALSE))
  (length(out) > 0L)
})

assert('record() with multiple messages in one expression', {
  res = record('{message("a"); message("b")}', message = TRUE)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  ('message' %in% cls)
})

assert('new_record() creates record objects with correct class', {
  r = new_record('hello', 'output')
  (inherits(r, 'record_output'))
  (r %==% structure('hello', class = 'record_output'))
})

assert('record_print.default() captures print output', {
  out = record_print(1:3)
  (is.character(out))
  (any(grepl('1 2 3', out)))
})

assert('format() markdown output includes fenced blocks for messages/warnings', {
  res = record(c('message("hi")', 'warning("yo")'), message = TRUE, warning = TRUE)
  md = format(res, to = 'markdown')
  (any(grepl('```', md)))
  (any(grepl('[.]message', md)))
  (any(grepl('[.]warning', md)))
})

assert('format() html output for messages and errors', {
  res = record('stop("boom")', error = TRUE)
  html = format(res, to = 'html')
  (any(grepl('class="error"', html)))
  (any(grepl('boom', html)))
})

assert('record() with error = NA halts on error', {
  (has_error(record('stop("fail")', error = NA)))
})

assert('record() with message = NA does not capture messages', {
  res = record('message("pass through")', message = NA)
  cls = sapply(res, function(x) sub('record_', '', class(x)))
  (!('message' %in% cls))
})

# add a border to an image via ImageMagick
add_border = function(input, pixels = 1, color = 'black', output) {
  input = normalizePath(input)
  if (missing(output))
    output = paste0(sans_ext(input), '-output.', file_ext(input))
  system2('convert', shQuote(c(
    input, '-shave', paste(pixels, pixels, sep = 'x'), '-bordercolor', color,
    '-border', pixels, output)
  ))
  optipng(dirname(output))
}

#' Use the Tinify API to compress PNG and JPEG images
#'
#' Compress PNG/JPEG images with \samp{api.tinify.com}, and download the
#' compressed images. These functions require R packages \pkg{curl} and
#' \pkg{jsonlite}. `tinify_dir()` is a wrapper function of `tinify()` to
#' compress images under a directory.
#'
#' You are recommended to set the API key in \file{.Rprofile} or
#' \file{.Renviron}. After that, the only required argument of this function is
#' `input`. If the original images can be overwritten by the compressed images,
#' you may either use `output = identity`, or set the value of the `history`
#' argument in \file{.Rprofile} or \file{.Renviron}.
#' @param input A vector of input paths of images.
#' @param output A vector of output paths or a function that takes `input` and
#'   returns a vector of output paths (e.g., `output = `[`identity`] means
#'   `output = input`). By default, if the `history` argument is not a provided,
#'   `output` is `input` with a suffix `-min` (e.g., when `input = 'foo.png'`,
#'   `output = 'foo-min.png'`), otherwise `output` is the same as `input`, which
#'   means the original image files will be overwritten.
#' @param quiet Whether to suppress detailed information about the compression,
#'   which is of the form \samp{input.png (10 Kb) ==> output.png (5 Kb, 50\%);
#'   compression count: 42}. The percentage after `output.png` stands for
#'   the compression ratio, and the compression count shows the number of
#'   compressions used for the current month.
#' @param force Whether to compress an image again when it appears to have been
#'   compressed before. This argument only makes sense when the `history`
#'   argument is provided.
#' @param key The Tinify API key. It can be set via either the global option
#'   `xfun.tinify.key` or the environment variable `R_XFUN_TINIFY_KEY` (see
#'   [env_option()]).
#' @param history Path to a history file to record the MD5 checksum of
#'   compressed images. If the checksum of an expected output image exists in
#'   this file and `force = FALSE`, the compression will be skipped. This can
#'   help you avoid unnecessary API calls.
#' @return The output file paths.
#' @references Tinify API: <https://tinypng.com/developers>.
#' @seealso The \pkg{tinieR} package (<https://github.com/jmablog/tinieR/>) is a
#'   more comprehensive implementation of the Tinify API, whereas
#'   `xfun::tinify()` has only implemented the feature of shrinking images.
#' @export
#' @examplesIf interactive()
#' f = xfun:::R_logo('jpg$')
#' xfun::tinify(f)  # remember to set the API key before trying this
tinify = function(
  input, output, quiet = FALSE, force = FALSE,
  key = env_option('xfun.tinify.key'),
  history = env_option('xfun.tinify.history')
) {
  if (!(is.character(key) && length(key) == 1 && key != '')) stop(
    "The value of the 'key' argument must be a single non-empty character string."
  )
  if (length(input) == 0) return(invisible(input))
  if (any(i <- !file_exists(input))) stop(
    'Input file(s) not found: ', paste(input[i], collapse = ', ')
  )
  if (missing(output)) {
    output = if (is.character(history)) input else {
      paste0(sans_ext(input), '-min.', file_ext(input))
    }
  } else if (is.function(output)) output = output(input)

  # avoid optimizing the input image if its md5 checksum exists in history
  save_history = function(file) {
    if (!is.character(history) || history == '') return()
    dir_create(dirname(history))
    cat(paste0(tools::md5sum(file), '\n'), file = history, append = TRUE)
  }
  test_history = function(file) {
    is.character(history) && all(file_exists(c(history, file))) &&
      (tools::md5sum(file) %in% readLines(history))
  }

  auth = paste('Authorization: Basic', base64_encode(charToRaw(paste0('api:', key))))

  mapply(input, output, FUN = function(i, o) {
    if (!force && test_history(o)) {
      if (!quiet) message(
        'The image ', o, ' has been compressed before. ',
        'To compress it again, call tinify() with force = TRUE.'
      )
      return()
    }
    if (grepl('[.]png$', i, ignore.case = TRUE))
      optipng(files = i, stdout = if (quiet) FALSE else '')
    res = curl::curl_upload(i, 'https://api.tinify.com/shrink', httpheader = auth, verbose = FALSE)
    cnt = curl::parse_headers_list(res$headers)[['compression-count']]
    res = jsonlite::fromJSON(rawToChar(res$content))
    if (!is.character(u <- res$output$url)) stop2(
      "Failed to shrink '", i, "'", sprintf(': %s (%s)', res$error, res$message)
    )
    if (!quiet) message(sprintf(
      '%s (%s) ==> %s (%s, %.01f%%); compression count: %s',
      i, format_bytes(res$input$size), o, format_bytes(res$output$size),
      res$output$ratio * 100, if (length(cnt)) cnt else NA
    ))
    # back up the original image and restore it if download failed
    if (i == o) {
      b = paste0(i, '~')
      file.rename(i, b)
      on.exit(if (file_exists(o)) file.remove(b) else file.rename(b, i), add = TRUE)
    }
    curl::curl_download(u, o)
    save_history(o)
  })

  invisible(output)
}

#' @param dir A directory under which all \file{.png}, \file{.jpeg}, and
#'   \file{.webp} files are to be compressed.
#' @param ... Arguments passed to [tinify()].
#' @rdname tinify
#' @export
tinify_dir = function(dir = '.', ...) {
  tinify(all_files('[.](png|jpe?g|webp)$', dir), ...)
}

#' Shrink images to a maximum width
#'
#' Use [magick::image_resize()] to shrink an
#' image if its width is larger than the value specified by the argument
#' `width`, and optionally call [tinify()] to compress it.
#' @param width The desired maximum width of images.
#' @param dir The directory of images.
#' @param files A vector of image file paths. By default, this is all
#'   \file{.png}, \file{.jpeg}, and \file{.webp} images under `dir`.
#' @param tinify Whether to compress images using [tinify()].
#' @export
#' @examples
#' f = xfun:::all_files('[.](png|jpe?g)$', R.home('doc'))
#' file.copy(f, tempdir())
#' f = file.path(tempdir(), basename(f))
#' magick::image_info(magick::image_read(f))  # some widths are larger than 300
#' xfun::shrink_images(300, files = f)
#' magick::image_info(magick::image_read(f))  # all widths <= 300 now
#' file.remove(f)
shrink_images = function(
  width = 800, dir = '.', files = all_files('[.](png|jpe?g|webp)$', dir),
  tinify = FALSE
) {
  for (f in files) {
    x = magick::image_read(f)
    if (magick::image_info(x)$width <= width) next
    x = magick::image_resize(x, sprintf('%dx', width))
    magick::image_write(x, f)
  }
  if (tinify) tinify(files, identity)
}

#' Upload an image to imgur.com
#'
#' This function uses the \pkg{curl} package or the system command `curl`
#' (whichever is available) to upload a image to <https://imgur.com>.
#'
#' One application is to upload local image files to Imgur when knitting a
#' document with \pkg{knitr}: you can set the `knitr::opts_knit$set(upload.fun =
#' xfun::upload_imgur`, so the output document does not need local image files
#' any more, and it is ready to be published online.
#' @param file Path to the image file to be uploaded.
#' @param key Client ID for Imgur. It can be set via either the global option
#'   `xfun.upload_imgur.key` or the environment variable
#'   `R_XFUN_UPLOAD_IMGUR_KEY` (see [xfun::env_option()]). If neither is set,
#'   this uses a client ID registered by Yihui Xie.
#' @param use_curl Whether to use the R package \pkg{curl} to upload the image.
#'   If `FALSE`, the system command `curl` will be used.
#' @param include_xml Whether to include the XML response in the returned value.
#' @return A character string of the link to the image. If `include_xml = TRUE`,
#'   this string carries an attribute named `XML`, which is the XML response
#'   from Imgur (it will be parsed by \pkg{xml2} if available). See Imgur API in
#'   the references.
#' @author Yihui Xie, adapted from the \pkg{imguR} package by Aaron Statham
#' @note Please register your own Imgur application to get your client ID; you
#'   can certainly use mine, but this ID is in the public domain so everyone has
#'   access to all images associated to it.
#' @references A demo: <https://yihui.org/knitr/demo/upload/>
#' @export
#' @examples \dontrun{
#' f = tempfile(fileext = '.png')
#' png(f); plot(rnorm(100), main = R.version.string); dev.off()
#'
#' res = imgur_upload(f, include_xml = TRUE)
#' res  # link to original URL of the image
#' attr(res, 'XML')  # all information
#' if (interactive()) browseURL(res)
#'
#' # to use your own key
#' options(xfun.upload_imgur.key = 'your imgur key')
#' }
upload_imgur = function(
  file, key = env_option('xfun.upload_imgur.key', '9f3460e67f308f6'),
  use_curl = loadable('curl'), include_xml = FALSE
) {
  if (!is.character(key)) stop('The Imgur API Key must be a character string!')
  api = 'https://api.imgur.com/3/image.xml'
  hdr = paste('Authorization: Client-ID', key)
  if (use_curl) {
    h = curl::new_handle(httpheader = hdr)
    curl::handle_setform(h, image = curl::form_file(file))
    res = curl::curl_fetch_memory(api, h)$content
  } else {
    file = path.expand(file)
    res = system2(
      'curl', shQuote(c('-H', hdr, '-F', paste0('image=@', file), '-s', api)),
      stdout = TRUE
    )
    res = one_string(res)
    if (res == '') stop('Failed to upload ', file)
  }
  if (loadable('xml2')) {
    res = xml2::as_list(xml2::read_xml(res))
    link = res[[1]]$link[[1]]
  } else {
    if (is.raw(res)) res = rawToChar(res)
    link = grep_sub('.*<link>([^<]+)</link>.*', '\\1', res)
  }
  if (length(link) != 1) stop(
    'Failed to upload ', file, sprintf(' (reason: %s)', if (is.character(res)) {
      grep_sub('.*<error>([^<]+)</error>.*', '\\1', res)
    } else res[[1]]$error[[1]])
  )
  if (include_xml) structure(link, XML = res) else link
}

#' Try to use the system native encoding to represent a character vector
#'
#' Apply `enc2native()` to the character vector, and check if `enc2utf8()` can
#' convert it back without a loss. If it does, return `enc2native(x)`, otherwise
#' return the original vector with a warning.
#' @param x A character vector.
#' @note On platforms that supports UTF-8 as the native encoding
#'   ([l10n_info()]`[['UTF-8']]` returns `TRUE`), the conversion will be
#'   skipped.
#' @export
#' @examples
#' library(xfun)
#' s = intToUtf8(c(20320, 22909))
#' Encoding(s)
#'
#' s2 = native_encode(s)
#' Encoding(s2)
native_encode = function(x) {
  if (isTRUE(l10n_info()[['UTF-8']])) return(x)
  if (identical(enc2utf8(x2 <- enc2native(x)), x)) return(x2)
  warning('The character vector cannot be represented in the native encoding')
  x
}

#' Check if a character vector consists of entirely ASCII characters
#'
#' Converts the encoding of a character vector to `'ascii'`, and check if
#' the result is `NA`.
#' @param x A character vector.
#' @return A logical vector indicating whether each element of the character
#'   vector is ASCII.
#' @export
#' @examples library(xfun)
#' is_ascii(letters)  # yes
#' is_ascii(intToUtf8(8212))  # no
is_ascii = function(x) {
  out = iconv(x, to = 'ascii') == x
  out[is.na(out)] = FALSE
  out[is.na(x)] = NA
  out
}

#' Create a local web application
#'
#' An experimental function to create a local web application based on R's
#' internal `httpd` server (which is primarily for running R's dynamic help
#' system).
#' @param name The app name (a character string, and each app should have a
#'   unique name).
#' @param handler A function that takes the HTTP request information (the first
#'   argument is the requested path) and returns a response.
#' @param open Whether to open the app, or a function to open the app URL.
#' @param ports A vector of ports to try for starting the server.
#' @return The app URL of the form `http://127.0.0.1:port/custom/name/`.
#' @note This function is not based on base R's public API, and is possible to
#'   break in the future, which is also why the documentation here is terse.
#'   Please avoid creating public-facing web apps with it. You may consider
#'   packages like \pkg{httpuv} and \pkg{Rserve} for production web apps.
#' @export
new_app = function(name, handler, open = interactive(), ports = 4321 + 1:10) {
  if (is.null(getOption('help.ports'))) {
    options(help.ports = ports); on.exit(options(help.ports = NULL))
  }
  port = suppressMessages(tools::startDynamicHelp(NA))
  url  = sprintf('http://127.0.0.1:%d/custom/%s/', port, name)
  wd = getwd()  # always run handler under the original working directory
  h = function(path, ...) {
    path = sub(paste0('^/custom/', name, '/'), '', path)
    if (path == '') path = '.'
    in_dir(wd, handler(path, ...))
  }
  # to anyone who sees the dirty assign() here, please close your eyes and walk
  # away as quickly as possible; thanks!
  assign(name, h, envir = getFromNamespace('.httpd.handlers.env', 'tools'))
  if (isTRUE(open)) open = getOption('viewer', getOption('browser'))
  if (is.function(open)) open(url)
  invisible(url)
}

#' Get data from a REST API
#'
#' Read data from a REST API and optionally with an authorization token in the
#' request header. The function `rest_api_raw()` returns the raw text of
#' the response, and `rest_api()` will parse the response with
#' `jsonlite::fromJSON()` (assuming that the response is in the JSON
#' format).
#'
#' These functions are simple wrappers based on [url()] and
#' [read_utf8()]. Specifically, the `headers` argument is
#' passed to `url()`, and `read_utf8()` will send a \samp{GET} request
#' to the API server. This means these functions only support the \samp{GET}
#' method. If you need to use other HTTP methods (such as \samp{POST}), you have
#' to use other packages such as \pkg{curl} and \pkg{httr}.
#' @param ... Arguments to be passed to `rest_api_raw()`.
#' @return A character vector (the raw JSON response) or an R object parsed from
#'   the JSON text.
#' @export
#' @examplesIf interactive()
#' # a normal GET request
#' xfun::rest_api('https://httpbin.org', '/get')
#' xfun::rest_api_raw('https://httpbin.org', '/get')
#'
#' # send the request with an auth header
#' xfun::rest_api('https://httpbin.org', '/headers', 'OPEN SESAME!')
#'
#' # with query parameters
#' xfun::rest_api('https://httpbin.org', '/response-headers', params = list(foo = 'bar'))
#'
#' # get the rate limit info from GitHub
#' xfun::github_api('/rate_limit')
rest_api = function(...) {
  res = rest_api_raw(...)
  jsonlite::fromJSON(res, simplifyVector = FALSE)
}

#' @param root The API root URL.
#' @param endpoint The API endpoint.
#' @param token A named character string (e.g., `c(token = "xxxx")`), which
#'   will be used to create an authorization header of the form
#'   \samp{Authorization: NAME TOKEN} for the API call, where \samp{NAME} is the
#'   name of the string and \samp{TOKEN} is the string. If the string does not
#'   have a name, \samp{Basic} will be used as the default name.
#' @param params A list of query parameters to be sent with the API call.
#' @param headers A named character vector of HTTP headers, e.g., `c(Accept
#'   = "application/vnd.github.v3+json")`.
#' @rdname rest_api
#' @export
rest_api_raw = function(root, endpoint, token = '', params = list(), headers = NULL)  {
  if (is.null(names(token))) names(token) = 'Basic'
  endpoint = sub('^/?', '/', endpoint)  # make sure it has a leading /
  url2 = if ('headers' %in% names(formals(url))) url else stop2(
    "The url() function does not support the 'headers' argument. Please upgrade R (>= 3.6.0)."
  )
  con = url2(
    paste0(root, endpoint, query_params(.list = params)), encoding = 'UTF-8',
    headers = c(
      headers, if (token != '') c(Authorization = sprintf('%s %s', names(token), token))
    )
  )
  on.exit(close(con), add = TRUE)
  raw_string(suppressWarnings(read_utf8(con)))
}

`%|%` = function(x, y) if (length(x)) x else y
if (getRversion() < '4.4.0') `%||%` = function(x, y) if (is.null(x)) y else x

stop2 = function(...) stop(..., call. = FALSE)

warning2 = function(...) warning(..., call. = FALSE)

#' Obtain an attribute of an object without partial matching
#'
#' An abbreviation of [`base::attr`]`(exact = TRUE)`.
#' @param ... Passed to [base::attr()] (without the `exact` argument).
#' @export
#' @examples
#' z = structure(list(a = 1), foo = 2)
#' base::attr(z, 'f')  # 2
#' xfun::attr(z, 'f')  # NULL
#' xfun::attr(z, 'foo')  # 2
attr = function(...) base::attr(..., exact = TRUE)

#' Set environment variables
#'
#' Set environment variables from a named character vector, and return the old
#' values of the variables, so they could be restored later.
#'
#' The motivation of this function is that [Sys.setenv()] does not
#' return the old values of the environment variables, so it is not
#' straightforward to restore the variables later.
#' @param vars A named character vector of the form `c(VARIABLE = VALUE)`.
#'   If any value is `NA`, this function will try to unset the variable.
#' @return Old values of the variables (if not set, `NA`).
#' @export
#' @examples
#' vars = xfun::set_envvar(c(FOO = '1234'))
#' Sys.getenv('FOO')
#' xfun::set_envvar(vars)
#' Sys.getenv('FOO')
set_envvar = function(vars) {
  if (is.null(nms <- names(vars)) || any(nms == '')) stop(
    "The 'vars' argument must take a named character vector."
  )
  vals = Sys.getenv(nms, NA, names = TRUE)
  i = is.na(vars)
  suppressWarnings(Sys.unsetenv(nms[i]))
  if (length(vars <- vars[!i])) do.call(Sys.setenv, as.list(vars))
  invisible(vals)
}

#' Retrieve a global option from both `options()` and environment variables
#'
#' If the option exists in [options()], use its value. If not, query the
#' environment variable with the name `R_NAME` where `NAME` is the capitalized
#' option name with dots substituted by underscores. For example, for an option
#' `xfun.foo`, first we try `getOption('xfun.foo')`; if it does not exist, we
#' check the environment variable `R_XFUN_FOO`.
#'
#' This provides two possible ways, whichever is more convenient, for users to
#' set an option. For example, global options can be set in the [.Rprofile]
#' file, and environment variables can be set in the [.Renviron] file.
#' @param name The option name.
#' @param default The default value if the option is not found in [options()] or
#'   environment variables.
#' @return The option value.
#' @export
#' @examples
#' xfun::env_option('xfun.test.option')  # NULL
#'
#' Sys.setenv(R_XFUN_TEST_OPTION = '1234')
#' xfun::env_option('xfun.test.option')  # 1234
#'
#' options(xfun.test.option = TRUE)
#' xfun::env_option('xfun.test.option')  # TRUE (from options())
#' options(xfun.test.option = NULL)  # reset the option
#' xfun::env_option('xfun.test.option')  # 1234 (from env var)
#'
#' Sys.unsetenv('R_XFUN_TEST_OPTION')
#' xfun::env_option('xfun.test.option')  # NULL again
#'
#' xfun::env_option('xfun.test.option', FALSE)  # use default
env_option = function(name, default = NULL) {
  if (name %in% names(.Options)) return(.Options[[name]])
  name = toupper(paste0('R_', gsub('[.]', '_', name)))
  envs = Sys.getenv()
  if (name %in% names(envs)) envs[[name]] else default
}

#' Call `on.exit()` in a parent function
#'
#' The function [on.exit()] is often used to perform tasks when the
#' current function exits. This `exit_call()` function allows calling a
#' function when a parent function exits (thinking of it as inserting an
#' `on.exit()` call into the parent function).
#' @param fun A function to be called when the parent function exits.
#' @param n The parent frame number. For `n = 1`, `exit_call(fun)` is
#'   the same as `on.exit(fun())`; `n = 2` means adding
#'   `on.exit(fun())` in the parent function; `n = 3` means the
#'   grandparent, etc.
#' @param ... Other arguments to be passed to `on.exit()`.
#' @references This function was inspired by Kevin Ushey:
#'   <https://yihui.org/en/2017/12/on-exit-parent/>
#' @export
#' @examples
#' f = function(x) {
#'   print(x)
#'   xfun::exit_call(function() print('The parent function is exiting!'))
#' }
#' g = function(y) {
#'   f(y)
#'   print('f() has been called!')
#' }
#' g('An argument of g()!')
exit_call = function(fun, n = 2, ...) {
  do.call(
    on.exit, list(substitute(fun(), list(fun = fun)), add = TRUE, ...),
    envir = parent.frame(n)
  )
}

#' Evaluate an expression under a specified working directory
#'
#' Change the working directory, evaluate the expression, and restore the
#' working directory.
#' @param dir Path to a directory.
#' @param expr An R expression.
#' @export
#' @examples
#' library(xfun)
#' in_dir(tempdir(), {print(getwd()); list.files()})
in_dir = function(dir, expr) {
  owd = setwd(dir); on.exit(setwd(owd))
  expr
}

#' Test if an object is `FALSE`
#'
#' For R versions lower than 3.5.0, this function is a simple abbreviation of
#' `identical(x, FALSE)`. For higher R versions, this function calls
#' `base::isFALSE()`.
#' @param x An R object.
#' @note This function will be deprecated in the future. We recommend that you
#'   use [base::isFALSE()] instead. If you have to support R versions lower
#'   than 3.5.0, you may use `identical(x, FALSE)`, but please note that it is
#'   not equivalent to `base::isFALSE()`.
#' @export
#' @keywords internal
isFALSE = function(x) {
  pkgs = tools::dependsOnPkgs('xfun', dependencies = 'all', recursive = FALSE)
  pkgs = intersect(pkgs, sys.packages())
  vers = sapply(pkgs, function(p) as.character(packageVersion(p)))
  if ('isFALSE' %in% ls(baseenv())) stop(
    'The function xfun::isFALSE() has been deprecated. Please ',
    if (length(vers)) {
      c('update the possibly outdated package(s): ', paste(pkgs, vers, sep = ' ', collapse = ', '), '. ')
    } else {
      'consider using base::isFALSE(x) or identical(x, FALSE) instead. '
    },
    'You may see https://yihui.org/en/2023/02/xfun-isfalse/ for more info.'
  )
  identical(x, FALSE)
}

# try to get the names of packages for all functions on the call stack
sys.packages = function() {
  unique(unlist(lapply(seq_along(sys.calls()), function(i) {
    environment(sys.function(i))$.packageName
  })))
}

#' Parse R code and do not keep the source
#'
#' An abbreviation of `parse(keep.source = FALSE)`.
#' @param code A character vector of the R source code.
#' @export
#' @return R [expression()]s.
#' @examples library(xfun)
#' parse_only('1+1'); parse_only(c('y~x', '1:5 # a comment'))
#' parse_only(character(0))
parse_only = function(code) {
  if (length(code) == 0) return(expression())
  parse(text = code, keep.source = FALSE)
}

#' Try to evaluate an expression silently
#'
#' An abbreviation of `try(silent = TRUE)`.
#' @param expr An R expression.
#' @export
#' @examples library(xfun)
#' z = try_silent(stop('Wrong!'))
#' inherits(z, 'try-error')
try_silent = function(expr) try(expr, silent = TRUE)

#' Try an expression and see if it throws an error
#'
#' Use [tryCatch()] to check if an expression throws an error.
#' @inheritParams try_silent
#' @return `TRUE` (error) or `FALSE` (success).
#' @export
#' @examples
#' xfun::try_error(stop('foo'))  # TRUE
#' xfun::try_error(1:10)  # FALSE
try_error = function(expr) {
  err = FALSE
  tryCatch(expr, error = function(e) err <<- TRUE)
  err
}

#' Retry calling a function for a number of times
#'
#' If the function returns an error, retry it for the specified number of
#' times, with a pause between attempts.
#'
#' One application of this function is to download a web resource. Since the
#' download might fail sometimes, you may want to retry it for a few more times.
#' @param fun A function.
#' @param ... Arguments to be passed to the function.
#' @param .times The number of times.
#' @param .pause The number of seconds to wait before the next attempt.
#' @export
#' @examplesIf interactive()
#' # read the GitHub releases info of the repo yihui/xfun
#' xfun::retry(xfun::github_releases, 'yihui/xfun')
retry = function(fun, ..., .times = 3, .pause = 5) {
  for (i in seq_len(.times)) {
    if (!inherits(res <- tryCatch(fun(...), error = identity), 'error'))
      return(res)
    Sys.sleep(.pause)
  }
  stop(res$message, call. = FALSE)
}

gsubi = function(...) gsub(..., ignore.case = TRUE)
gsubf = function(...) gsub(..., fixed = TRUE)

#' Turn the output of [str()] into a tree diagram
#'
#' The super useful function `str()` uses \verb{..} to indicate the level
#' of sub-elements of an object, which may be difficult to read. This function
#' uses vertical pipes to connect all sub-elements on the same level, so it is
#' clearer which elements belong to the same parent element in an object with a
#' nested structure (such as a nested list).
#' @param ... Arguments to be passed to [str()] (note that the
#'   `comp.str` is hardcoded inside this function, and it is the only
#'   argument that you cannot customize).
#' @return A character string as a [raw_string()].
#' @export
#' @examples fit = lsfit(1:9, 1:9)
#' str(fit)
#' xfun::tree(fit)
#'
#' fit = lm(dist ~ speed, data = cars)
#' str(fit)
#' xfun::tree(fit)
#'
#' # some trivial examples
#' xfun::tree(1:10)
#' xfun::tree(iris)
tree = function(...) {
  x = capture.output(str(..., comp.str = '$ '))
  r = '^([^$-]+[$-] )(.*)$'
  x1 = gsub(r, '\\1', x)
  x2 = gsub(r, '\\2', x)
  x1 = gsub('[.][.]', '  ', x1)
  x1 = gsub('[$] $', '|-', x1)
  x1 = connect_pipes(x1)
  x3 = paste(x1, x2, sep = '')
  i = !grepl(r, x)
  x3[i] = x[i]
  raw_string(x3)
}

# for a tree diagram, connect the pipes on the same level, e.g., change

# |- ..
#   |- ..
#
#   |- ..

# to

# |- ..
#   |- ..
#   |
#   |- ..

# this task is not complicated, but just boring nested for-loops
connect_pipes = function(x) {
  ns = nchar(x); n = max(ns); m = length(x)
  if (n < 2 || m < 3) return(x)
  A = matrix('', nrow = m, ncol = n)
  x = strsplit(x, '')
  for (i in seq_len(m)) {
    A[i, seq_len(ns[i])] = x[[i]]
  }
  k = NULL
  for (j in seq_len(n - 1)) {
    for (i in seq_len(m - 2)) {
      if (!all(A[i, j + 0:1] == c('|', '-'))) next
      for (l in (i + 1):m) {
        cells = A[l, j + 0:1]
        if (all(cells == ' ')) {
          if (l == m) {
            k = NULL; break
          } else k = c(k, l)
        } else if (all(cells == c('|', '-'))) {
          break
        } else {
          k = NULL; break
        }
      }
      if (length(k) > 0) A[k, j] = '|'
      k = NULL
    }
  }
  apply(A, 1, paste, collapse = '')
}

pkg_file = function(...) system.file(..., package = 'xfun', mustWork = TRUE)

#' Format numbers of bytes using a specified unit
#'
#' Call the S3 method `format.object_size()` to format numbers of bytes.
#' @param x A numeric vector (each element represents a number of bytes).
#' @param units,... Passed to [`format()`][format.object_size].
#' @return A character vector.
#' @export
#' @examples
#' xfun::format_bytes(c(1, 1024, 2000, 1e6, 2e8))
#' xfun::format_bytes(c(1, 1024, 2000, 1e6, 2e8), units = 'KB')
format_bytes = function(x, units = 'auto', ...) {
  vapply(x, function(b) {
    format(structure(b, class = 'object_size'), units = units, ...)
  }, character(1))
}

# get the function name of the parent call
func_name = function(which = 1) {
  x = sys.call(which)[[1]]
  deparse(x)[1]
}

# evaluate an expression with an error handler; originally this was for knitr to
# output error location but can also be useful for other applications
handle_error = function(
  expr, handler, label = '', fun = getOption('xfun.handle_error.loc_fun')
) {
  on.exit(if (!ok) {
    loc = if (is.function(fun)) trimws(fun(label)) else ''
    m = handler(loc)
    message(one_string(m))
  })
  ok = FALSE
  expr  # evaluate now
  ok = TRUE
  expr  # won't be evaluated again
}

# a shorthand for rm(list =, envir =)
rm_vars = function(x, envir, ...) rm(list = x, envir = envir, ...)

#' Run \command{R CMD check} on the reverse dependencies of a package
#'
#' Install the source package, figure out the reverse dependencies on CRAN,
#' download all of their source packages, and run \command{R CMD check} on them
#' in parallel.
#'
#' Everything occurs under the current working directory, and you are
#' recommended to call this function under a designated directory, especially
#' when the number of reverse dependencies is large, because all source packages
#' will be downloaded to this directory, and all \file{*.Rcheck} directories
#' will be generated under this directory, too.
#'
#' If a source tarball of the expected version has been downloaded before (under
#' the \file{tarball} directory), it will not be downloaded again (to save time
#' and bandwidth).
#'
#' After a package has been checked, the associated \file{*.Rcheck} directory
#' will be deleted if the check was successful (no warnings or errors or notes),
#' which means if you see a \file{*.Rcheck} directory, it means the check
#' failed, and you need to take a look at the log files under that directory.
#'
#' The time to finish the check is recorded for each package. As the check goes
#' on, the total remaining time will be roughly estimated via `n *
#' mean(times)`, where `n` is the number of packages remaining to be
#' checked, and `times` is a vector of elapsed time of packages that have
#' been checked.
#'
#' If a check on a reverse dependency failed, its \file{*.Rcheck} directory will
#' be renamed to \file{*.Rcheck2}, and another check will be run against the
#' CRAN version of the package unless `options(xfun.rev_check.compare =
#' FALSE)` is set. If the logs of the two checks are the same, it means no new
#' problems were introduced in the package, and you can probably ignore this
#' particular reverse dependency. The function `compare_Rcheck()` can be
#' used to create a summary of all the differences in the check logs under
#' \file{*.Rcheck} and \file{*.Rcheck2}. This will be done automatically if
#' `options(xfun.rev_check.summary = TRUE)` has been set.
#'
#' A recommended workflow is to use a special directory to run
#' `rev_check()`, set the global [options()]
#' `xfun.rev_check.src_dir` and `repos` in the R startup (see
#' `?`[`Startup`]) profile file `.Rprofile` under this directory,
#' and (optionally) set `R_LIBS_USER` in \file{.Renviron} to use a special
#' library path (so that your usual library will not be cluttered). Then run
#' `xfun::rev_check(pkg)` once, investigate and fix the problems or (if you
#' believe it was not your fault) ignore broken packages in the file
#' \file{00ignore}, and run `xfun::rev_check(pkg)` again to recheck the
#' failed packages. Repeat this process until all \file{*.Rcheck} directories
#' are gone.
#'
#' As an example, I set `options(repos = c(CRAN =
#' 'https://cran.rstudio.com'), xfun.rev_check.src_dir = '~/Dropbox/repo')` in
#' \file{.Rprofile}, and `R_LIBS_USER=~/R-tmp` in \file{.Renviron}. Then I
#' can run, for example, `xfun::rev_check('knitr')` repeatedly under a
#' special directory \file{~/Downloads/revcheck}. Reverse dependencies and their
#' dependencies will be installed to \file{~/R-tmp}, and \pkg{knitr} will be
#' installed from \file{~/Dropbox/repo/kintr}.
#' @param pkg The package name.
#' @param which Which types of reverse dependencies to check. See
#'   [tools::package_dependencies()] for possible values. The
#'   special value `'hard'` means the hard dependencies, i.e.,
#'   `c('Depends', 'Imports', 'LinkingTo')`.
#' @param recheck A vector of package names to be (re)checked. If not provided
#'   and there are any \file{*.Rcheck} directories left by certain packages
#'   (this often means these packages failed the last time), `recheck` will
#'   be these packages; if there are no \file{*.Rcheck} directories but a text
#'   file \file{recheck} exists, `recheck` will be the character vector
#'   read from this file. This provides a way for you to manually specify the
#'   packages to be checked. If there are no packages to be rechecked, all
#'   reverse dependencies will be checked.
#' @param ignore A vector of package names to be ignored in \command{R CMD
#'   check}. If this argument is missing and a file \file{00ignore} exists, the
#'   file will be read as a character vector and passed to this argument.
#' @param update Whether to update all packages before the check.
#' @param src The path of the source package directory.
#' @param src_dir The parent directory of the source package directory. This can
#'   be set in a global option if all your source packages are under a common
#'   parent directory.
#' @param timeout Timeout in seconds for \command{R CMD check} to check each
#'   package. The (approximate) total time can be limited by the global option
#'   `xfun.rev_check.timeout_total`.
#' @return A named numeric vector with the names being package names of reverse
#'   dependencies; `0` indicates check success, `1` indicates failure,
#'   and `2` indicates that a package was not checked due to global
#'   timeout.
#' @seealso `devtools::revdep_check()` is more sophisticated, but currently
#'   has a few major issues that affect me: (1) It always deletes the
#'   \file{*.Rcheck} directories
#'   (<https://github.com/r-lib/devtools/issues/1395>), which makes it
#'   difficult to know more information about the failures; (2) It does not
#'   fully install the source package before checking its reverse dependencies
#'   (<https://github.com/r-lib/devtools/pull/1397>); (3) I feel it is
#'   fairly difficult to iterate the check (ignore the successful packages and
#'   only check the failed packages); by comparison, `xfun::rev_check()`
#'   only requires you to run a short command repeatedly (failed packages are
#'   indicated by the existing \file{*.Rcheck} directories, and automatically
#'   checked again the next time).
#'
#'   `xfun::rev_check()` borrowed a very nice feature from
#'   `devtools::revdep_check()`: estimating and displaying the remaining
#'   time. This is particularly useful for packages with huge numbers of reverse
#'   dependencies.
#' @export
rev_check = function(
  pkg, which = 'all', recheck = NULL, ignore = NULL, update = TRUE,
  timeout = getOption('xfun.rev_check.timeout', 15 * 60),
  src = file.path(src_dir, pkg), src_dir = getOption('xfun.rev_check.src_dir')
) {
  if (length(src) != 1 || !dir_exists(src)) stop(
    'The package source dir (the "src" argument) must be an existing directory'
  )

  message('Installing the source package ', src)
  install_dir(path.expand(src))

  db = available.packages(type = 'source')

  # install packages that are not loadable (testing in parallel)
  p_install = function(pkgs) {
    pkgs_up = NULL
    if (update) {
      message('Updating all R packages...')
      pkgs_up = intersect(old.packages(checkBuilt = TRUE)[, 'Package'], pkgs)
      pkg_install(pkgs_up)
    }
    pkgs = setdiff(pkgs, pkgs_up)  # don't install pkgs that were just updated
    print(system.time(
      pkg_install(unlist(plapply(pkgs, function(p) if (!loadable(p, new_session = TRUE)) p)))
    ))
  }

  unlink('*.Rcheck2', recursive = TRUE)
  if (missing(recheck)) {
    dirs = list.files('.', '.+[.]Rcheck$')
    pkgs = gsub('.Rcheck$', '', dirs)
    recheck = if (length(pkgs) == 0 && file_exists('recheck')) {
      scan('recheck', 'character')
    } else pkgs
  }
  pkgs = if (length(recheck)) {
    p_install(pkg_dep(recheck, db, which = 'all'))
    recheck
  } else {
    res = check_deps(pkg, db, which)
    message('Installing dependencies of reverse dependencies')
    res$install = setdiff(res$install, ignore_deps())
    print(system.time(p_install(res$install)))
    res$check
  }
  pkgs = intersect(pkgs, rownames(db))  # make sure the pkgs are on CRAN

  lib_cran = './library-cran'
  on.exit(unlink(lib_cran, recursive = TRUE), add = TRUE)
  dir.create(lib_cran, showWarnings = FALSE)
  pkg_install(pkg, lib = lib_cran)  # the CRAN version of the package

  f = tempfile('check-done', fileext = '.rds')
  l = tempfile('check-lock'); on.exit(unlink(c(f, l)), add = TRUE)
  n = length(pkgs)
  if (n == 0) {
    message('No reverse dependencies to be checked for the package ', pkg); return()
  }

  if (missing(ignore) && file_exists('00ignore')) ignore = scan('00ignore', 'character')
  if (length(ignore)) {
    message('Ignoring packages: ', paste(ignore, collapse = ' '))
    unlink(sprintf('%s.Rcheck', ignore), recursive = TRUE)
    pkgs = setdiff(pkgs, ignore)
    if ((n <- length(pkgs)) == 0) {
      message('No packages left to be checked'); return()
    }
  }

  message('Downloading tarballs')
  tars = download_tarball(pkgs, db, dir = 'tarball')
  tars = setNames(tars, pkgs)

  t0 = Sys.time()
  tt = getOption('xfun.rev_check.timeout_total', Inf)
  message('Checking ', n, ' packages: ', paste(pkgs, collapse = ' '))

  res = plapply(pkgs, function(p) {
    d = sprintf('%s.Rcheck', p)
    if (!p %in% rownames(db)) {
      message('Checking ', p, ' (aborted since it is no longer on CRAN')
      unlink(d, recursive = TRUE)
      return()
    }

    timing = function() {
      # in case two packages finish at exactly the same time
      while (file_exists(l)) Sys.sleep(.1)
      file.create(l); on.exit(unlink(l), add = TRUE)
      done = c(if (file_exists(f)) readRDS(f), p)
      saveRDS(done, f)
      n2 = length(setdiff(pkgs, done))  # remaining packages
      t1 = Sys.time(); t2 = Sys.time() + n2 * (t1 - t0) / (n - n2)
      message(
        'Packages remaining: ', n2, '/', n, '; Expect to finish at ', t2,
        ' (', format(round(difftime(t2, t1))), ')'
      )
      # 0 (FALSE): success; 1: failure
      setNames(as.integer(dir_exists(d)), p)
    }

    if (!file_exists(z <- tars[p])) {
      dir.create(d, showWarnings = FALSE)
      return(timing())
    }

    # timeout; package not checked
    if (difftime(Sys.time(), t0, units = 'secs') > tt) {
      return(setNames(2L, p))
    }

    check_it = function(args = NULL, ...) {
      system2(
        file.path(R.home('bin'), 'R'),
        c(args, 'CMD', 'check', '--no-manual', shQuote(z)),
        stdout = FALSE, stderr = FALSE, timeout = timeout, ...
      )
    }
    check_it()

    if (!clean_Rcheck(d)) {
      if (!dir_exists(d)) {dir.create(d); return(timing())}
      # try to install missing LaTeX packages for vignettes if possible, then recheck
      vigs = list.files(
        file.path(d, 'vign_test', p, 'vignettes'), '[.](Rnw|Rmd)$',
        ignore.case = TRUE, full.names = TRUE
      )
      pkg_load2('tinytex')
      if (length(vigs) && any(file_exists(with_ext(vigs, 'log')))) {
        if (tinytex::is_tinytex()) for (vig in vigs) in_dir(dirname(vig), {
          Rscript(shQuote(c('-e', 'if (grepl("[.]Rnw$", f <- commandArgs(T), ignore.case = T)) knitr::knit2pdf(f) else rmarkdown::render(f)', basename(vig))))
        })
        check_it()
        if (clean_Rcheck(d)) return(timing())
      }
      # if there are still missing LaTeX packages, install them and recheck
      l0 = tinytex::tl_pkgs()
      lapply(
        list.files(d, '[.]log$', full.names = TRUE, recursive = TRUE),
        tinytex::parse_install, quiet = TRUE
      )
      if (!identical(l0, tinytex::tl_pkgs())) {
        check_it()
        if (clean_Rcheck(d)) return(timing())
      }
      # clean up the check log, and recheck with the current CRAN version of pkg
      cleanup = function() in_dir(d, {
        clean_log()
        # so that I can easily preview it in the Finder on macOS
        file_exists('00install.out') && file.rename('00install.out', '00install.log')
      })
      # ignore vignettes that failed to build for unknown reasons
      cleanup()
      if (clean_Rcheck(d)) return(timing())
      # whether to check the package against the CRAN version?
      if (!getOption('xfun.rev_check.compare', TRUE)) return(timing())
      file.rename(d, d2 <- paste0(d, '2'))
      check_it('--no-environ', env = tweak_r_libs(lib_cran))
      if (!dir_exists(d)) file.rename(d2, d) else {
        cleanup()
        if (identical_logs(c(d, d2))) unlink(c(d, d2), recursive = TRUE)
      }
    }
    timing()
  })
  if (getOption('xfun.rev_check.summary', FALSE)) {
    html = compare_Rcheck(); if (isTRUE(grepl('[.]html$', html))) browseURL(html)
  }
  unlist(res)
}

# remove the OK lines in the check log
clean_log = function() {
  if (!file_exists(l <- '00check.log')) return()
  x = grep('^[*].+OK$', read_utf8(l), invert = TRUE, value = TRUE)
  # don't want diffs in random tempdir/tempfile paths when comparing check logs
  x[grep(dirname(tempdir()), x, fixed = TRUE)] = 'RANDOM TEMPDIR/TEMPFILE PATH DELETED'
  # delete the download progress
  x = grep('^\\s*\\[\\d+%] Downloaded \\d+ bytes...\\s*$', x, invert = TRUE, value = TRUE)
  # delete lines of the form "address 0x1067143eb, cause 'illegal opcode'"
  x = grep("address 0x[[:xdigit:]]+, cause '[^']+'", x, invert = TRUE, value = TRUE)
  x = recheck_vig(x)
  x = tail(x, -2)
  writeLines(x, l)  # remove the first 2 lines (log dir name and R version)
  x
}

# sometimes R CMD check fails to build vignettes for unknown reasons; try to
# recheck the package in this case
recheck_vig = function(x) {
  if (!any(i1 <- (x == '* checking re-building of vignette outputs ... WARNING')))
    return(x)

  i1 = which(i1)[1]
  i2 = which(x == 'Execution halted')
  i2 = i2[i2 > i1]
  if (length(i2) == 0) return(x)

  i3 = grep('^[*] checking ', x)  # next checking item
  i3 = i3[i3 > i1]
  if (length(i3)) {
    i2 = i2[i2 < i3[1]]  # 'Execution halted' needs to appear before next '* checking'
    if (length(i2) == 0) return(x)
  }

  # if no explicit errors were found in processing vignettes (except pandoc
  # error), remove the relevant log
  i2 = tail(i2, 1)
  if (length(grep('pandoc document conversion failed with error', x[i1:i2])) > 0 ||
      length(grep('Error: processing vignette .+ failed with diagnostics:', x[i1:i2])) == 0)
    x = x[-(i1:i2)]
  x
}

# are the check logs identical under a series of *.Rcheck directories?
identical_logs = function(dirs) {
  if (length(dirs) < 2) return(FALSE)
  if (!all(file_exists(logs <- file.path(dirs, '00check.log')))) return(FALSE)
  x = read_utf8(logs[1])
  for (i in 2:length(dirs)) if (!identical(x, read_utf8(logs[i]))) return(FALSE)
  TRUE
}

# delete files/dirs that are usually not helpful
clean_Rcheck2 = function(dir = '.') {
  owd = setwd(dir); on.exit(setwd(owd), add = TRUE)
  ds = list.files('.', '.+[.]Rcheck$')
  for (d in c(ds, paste0(ds, '2'))) {
    f1 = list.files(d, full.names = TRUE)
    f2 = file.path(d, c('00_pkg_src', '00check.log', '00install.log'), fsep = '/')
    unlink(setdiff(f1, f2), recursive = TRUE)
  }
}

# add a new library path to R_LIBS_USER
tweak_r_libs = function(new) {
  x = read_all(existing_files(c('~/.Renviron', '.Renviron')))
  x = grep('^\\s*#', x, invert = TRUE, value = TRUE)
  x = gsub('^\\s+|\\s+$', '', x)
  x = x[x != '']
  i = grep('^R_LIBS_USER=.+', x)
  if (length(i)) {
    x[i[1]] = sub('(="?)', path_sep('\\1', new), x[i[1]])
    x
  } else {
    v = Sys.getenv('R_LIBS_USER')
    v = if (v == '') new else path_sep(new, v)
    c(paste0('R_LIBS_USER=', v), x)
  }
}

# separate paths by the path separator on a specific platform
path_sep = function(...) paste(..., sep = .Platform$path.sep)

# a shorthand of tools::package_dependencies()
pkg_dep = function(x, ...) {
  if (length(x)) unique(unlist(tools::package_dependencies(x, ...)))
}

# calculate the packages required to check a package
check_deps = function(x, db = available.packages(), which = 'all') {
  if (identical(which, 'hard')) which = c('Depends', 'Imports', 'LinkingTo')
  x0 = db[, 'Package']  # all available packages
  # packages that reverse depend on me
  x1 = pkg_dep(x, db, which, reverse = TRUE)
  x1 = intersect(x1, x0)
  # only check a sample of soft reverse dependencies (useful if there are too many)
  if (identical(which, 'all') && (n <- getOption('xfun.rev_check.sample', 100)) >= 0) {
    x2 = pkg_dep(x, db, c('Suggests', 'Enhances'), reverse = TRUE)
    x2 = intersect(x2, x0)
    if (n < length(x2)) x1 = c(setdiff(x1, x2), sample(x2, n))
  }
  # to R CMD check x1, I have to install all their dependencies
  x2 = pkg_dep(x1, db, 'all')
  # and for those dependencies, I have to install the default dependencies
  x3 = pkg_dep(x2, db, recursive = TRUE)
  list(check = x1, install = intersect(c(x1, x2, x3), x0))
}

#' Submit check jobs to crandalf
#'
#' Check the reverse dependencies of a package using the crandalf service:
#' <https://github.com/yihui/crandalf>. If the number of reverse
#' dependencies is large, they will be split into batches and pushed to crandalf
#' one by one.
#'
#' Due to the time limit of a single job on GitHub Actions (6 hours), you will
#' have to split the large number of reverse dependencies into batches and check
#' them sequentially on GitHub (at most 5 jobs in parallel). The function
#' `crandalf_check()` does this automatically when necessary. It requires
#' the \command{git} command to be available.
#'
#' The function `crandalf_results()` fetches check results from GitHub
#' after all checks are completed, merge the results, and show a full summary of
#' check results. It requires `gh` (GitHub CLI:
#' <https://cli.github.com/manual/>) to be installed and you also need to
#' authenticate with your GitHub account beforehand.
#' @param pkg The package name of which the reverse dependencies are to be
#'   checked.
#' @param size The number of reverse dependencies to be checked in each job.
#' @param jobs The number of jobs to run in GitHub Actions (by default, all jobs
#'   are submitted, but you can choose to submit the first few jobs).
#' @param which The type of dependencies (see [rev_check()]).
#' @export
crandalf_check = function(pkg, size = 400, jobs = Inf, which = 'all') {
  git_test_branch()
  git_co('main')
  on.exit(git_co('main'), add = TRUE)
  git_test_branch()

  # do everything inside the check-pkg branch
  b = paste0('check-', pkg)
  if (git_co(b, stderr = FALSE) != 0) {
    git_co(c('-b', b))
    file.create('recheck')
    git(c('add', 'recheck'))
    git(c('commit', '-m', shQuote(paste('Revcheck', pkg))))
    git('push')
    message(
      'Please create a pull request from the branch ', b,
      ' on GitHub and re-run xfun::crandalf_check("', pkg, '").'
    )
    return(invisible())
  }
  git(c('merge', 'main'))

  x = check_deps(pkg, which = which)$check
  n = length(x)
  if (n <= size) {
    message('No need to split ', n, ' reverse dependencies into batches of size ', size, '.')
    if (any(grepl('Your branch is ahead of ', git('status', stdout = TRUE)))) {
      git('push')
    } else if (Sys.which('gh') != '') {
      gh(c('workflow', 'run', 'rev-check.yaml', '--ref', b))
      message('Triggering rev-check.yaml job against ', b, ' branch in crandalf repo on GitHub.')
    } else {
      message('Remember to re-run the last job for the package ', pkg, ' on GitHub.')
    }
    return(invisible())
  }

  b = ceiling(n/size)
  i = rep(seq_len(b), each = size)[seq_len(n)]
  k = 1
  # use an id in the commit so that I know which jobs are for the same pkg
  id = format(Sys.time(), '%Y%m%d%H%M')
  for (p in head(split(x, i), jobs)) {
    message('Batch ', k)
    writeLines(p, 'recheck')
    git(c('add', 'recheck'))
    git(c('commit', '-m', shQuote(paste(
      c(id, 'checking:', head(p, 3), '...'), collapse = ' '
    ))))
    git('push')
    Sys.sleep(10)
    k = k + 1
  }
}

#' @param repo The crandalf repo on GitHub (of the form `user/repo` such as
#'   `"yihui/crandalf"`). Usually you do not need to specify it, unless you
#'   are not calling this function inside the crandalf project, because
#'   \command{gh} should be able to figure out the repo automatically.
#' @param limit The maximum of records for \command{gh run list} to retrieve.
#'   You only need a larger number if the check results are very early in the
#'   GitHub Action history.
#' @param wait Number of seconds to wait if not all jobs have been completed on
#'   GitHub. By default, this function checks the status every 5 minutes until
#'   all jobs are completed. Set `wait` to 0 to disable waiting (and throw
#'   an error immediately when any jobs are not completed).
#' @rdname crandalf_check
#' @export
crandalf_results = function(pkg, repo = NA, limit = 200, wait = 5 * 60) {
  res = crandalf_jobs(pkg, repo, limit)
  if (NROW(res) == 0) {
    stop('Did not find check results for ', pkg, ' from GitHub Actions.')
  }
  if (any(res[, 1] != 'completed')) {
    if (wait <= 0) stop('Please wait till all jobs have been completed on GitHub Actions.')
    status = NULL
    repeat {
      res = crandalf_jobs(pkg, repo, limit)
      if (all(res[, 1] == 'completed')) break
      if (is.null(status) || !identical(status, table(res[, 1]))) {
        status = table(res[, 1])
        timestamp()
        print(status)
      }
      Sys.sleep(wait)
    }
  }
  ids = grep_sub('^(\\d+) checking: .+', '\\1', res[, 3])
  i = if (length(ids) > 0) grep(sprintf('^%s checking: ', ids[1]), res[, 3]) else {
    head(which(res[, 2] == 'failure'), 1)
  }
  res = res[i, , drop = FALSE]
  res = res[res[, 2] == 'failure', , drop = FALSE]
  if (NROW(res) == 0) {
    stop('Did not find any failed results on GitHub Actions.')
  }
  for (i in seq_len(nrow(res))) {
    message('Downloading check results (', i, '/', nrow(res), ')')
    gh_run('download', res[i, 7], '-D', tempfile('crandalf-', '.'), repo = repo)
  }
  if (interactive()) browseURL(crandalf_merge(pkg))
}

# retrieve the first N jobs info
crandalf_jobs = function(pkg, repo = NA, limit = 200) {
  res = gh_run('list', '-L', limit, '-w', 'rev-check', repo = repo)
  res = res[grep(paste0('rev-check\tcheck-', pkg), res)]
  do.call(rbind, strsplit(res, '\t'))
}

crandalf_merge = function(pkg) {
  unlink(list.files('.', '[.]Rcheck2?$'), recursive = TRUE)
  x1 = x2 = x3 = NULL
  f1 = '00check_diffs.html'; f3 = 'latex.txt'
  for (d in list.files('.', '^crandalf-.+')) {
    if (!dir_exists(d)) next
    p = file.path(d, 'macOS-rev-check-results')
    if (file_exists(f <- file.path(p, f1))) {
      x = read_utf8(f)
      x1 = if (length(x1) == 0) x else {
        i1 = grep('<body>', x)[1]
        i2 = tail(grep('</body>', x), 1)
        i3 = tail(grep('</body>', x1), 1)
        append(x1, x[(i1 + 1):(i2 - 1)], i3 - 1)
      }
      file.remove(f)
    }
    if (file_exists(f <- file.path(p, 'recheck2'))) {
      x2 = c(x2, read_utf8(f))
      file.remove(f)
    }
    cs = list.files(p, '[.]Rcheck[2]?$', full.names = TRUE)
    file.rename(cs, basename(cs))
    if (file_exists(f <- file.path(p, f3))) {
      x3 = c(x3, read_utf8(f))
      file.remove(f)
    }
    unlink(d, recursive = TRUE)
  }
  write_utf8(x1, f1)  # the full summary

  # store newly detected missing latex packages in latex.txt and commit/push
  git_co('main')
  append_unique(x3, f3)
  find_missing_latex()
  git(c('commit', '-m', shQuote('add more latex packages'), f3))
  git('push')

  git_co(paste0('check-', pkg))
  r = '[.]Rcheck2$'
  write_utf8(sort(unique(c(x2, gsub(r, '', list.files('.', r))))), 'recheck')
  f1
}

# mclapply() with a different default for mc.cores and disable prescheduling
plapply = function(X, FUN, ...) {
  parallel::mclapply(
    X, FUN, ..., mc.cores = getOption('mc.cores', parallel::detectCores()),
    mc.preschedule = FALSE
  )
}

# download the source package from CRAN
download_tarball = function(p, db = available.packages(type = 'source'), dir = '.', retry = 3) {
  if (!dir_exists(dir)) dir.create(dir, recursive = TRUE)
  z = file.path(dir, sprintf('%s_%s.tar.gz', p, db[p, 'Version']))
  mapply(function(p, z) {
    # remove other versions of the package tarball
    unlink(setdiff(list.files(dir, sprintf('^%s_.+.tar.gz', p), full.names = TRUE), z))
    for (i in seq_len(retry)) {
      if (file_exists(z)) break
      try(download.file(paste(db[p, 'Repository'], basename(z), sep = '/'), z, mode = 'wb'))
    }
  }, p, z, SIMPLIFY = FALSE)
  z
}

# clean up *.Rcheck if there are no warnings, errors, or notes in the log
clean_Rcheck = function(dir, log = read_utf8(file.path(dir, '00check.log'))) {
  # do not check the status line
  if (length(grep('^Status: ', tail(log, 1)))) log = head(log, -1)
  if (length(grep('(WARNING|ERROR|NOTE)$', log)) == 0 ||
      length(grep('[*] checking whether package .+ can be installed ... ERROR', log)))
    unlink(dir, recursive = TRUE)
  !dir_exists(dir)
}

#' @rdname rev_check
#' @param status_only If `TRUE`, only compare the final statuses of the
#'   checks (the last line of \file{00check.log}), and delete \file{*.Rcheck}
#'   and \file{*.Rcheck2} if the statuses are identical, otherwise write out the
#'   full diffs of the logs. If `FALSE`, compare the full logs under
#'   \file{*.Rcheck} and \file{*.Rcheck2}.
#' @param output The output Markdown file to which the diffs in check logs will
#'   be written. If the \pkg{markdown} package is available, the Markdown file
#'   will be converted to HTML, so you can see the diffs more clearly.
#' @export
compare_Rcheck = function(status_only = TRUE, output = '00check_diffs.md') {
  if (length(dirs <- list.files('.', '.+[.]Rcheck$')) == 0) {
    # clean up the `recheck` file
    if (file_exists('recheck')) writeLines(character(), 'recheck')
    return()
  }
  d2 = function(d) c(d, paste0(d, '2'))
  logs = function(d) file.path(d2(d), '00check.log')
  res = NULL
  if (!status_only && Sys.which('diff') == '')
    warning("The command 'diff' is not available; will not calculate exact diffs in logs.")
  for (d in dirs) {
    f = existing_files(logs(d))
    if (status_only && length(f) == 2) {
      status_line = function(file) {
        x = tail(read_utf8(file), 1)
        if (grepl('^Status: ', x)) x else {
          warning('The last line of ', file, ' is not the status.')
          NULL
        }
      }
      # if the check with current CRAN version of package also failed, or the
      # two statues are the same, chances are we are good to go
      s1 = status_line(f[1])
      if (length(grep('Status: .*\\d+ ERROR', s1)) || identical(s1, status_line(f[2]))) {
        unlink(d2(d), recursive = TRUE); next
      }
    }
    res = c(
      res, paste('##', p <- sans_ext(d)), '',
      sprintf('[CRAN version](https://cran.rstudio.com/package=%s) (-) vs current version (+):\n', p),
      '```diff', file_diff(f), '```', ''
    )
    if (length(res2 <- cran_check_page(p, NULL))) res = c(
      res, 'CRAN check logs:\n\n```', head_tail(unique(unlist(strsplit(res2, '\n')))), '```\n'
    )
  }
  if (length(res) == 0) return()
  res = c('---', 'title: xfun::rev_check() results', '---', '', res)
  xfun::write_utf8(res, output)
  if (!loadable('litedown')) return(output)
  html_file = litedown::mark(output, text = res)
  if (!getOption('xfun.rev_check.keep_md', FALSE)) unlink(output)
  html_file
}

# keep the first and last n elements in x, and omit the middle
head_tail = function(x, n = 10) {
  if (length(x) <= 2 * n) return(x)
  c(head(x, n), '....', tail(x, n))
}

# compute the diffs of two files; if diffs too large, dedup them
file_diff = function(files, len = 200, use_diff = Sys.which('diff') != '') {
  n = length(files)
  if (n == 0) return()
  if (n == 1) {
    f = tempfile(); on.exit(unlink(f), add = TRUE); file.create(f)
    files = c(f, files)
  }
  d = if (use_diff) {
    suppressWarnings(system2('diff', shQuote(files), stdout = TRUE))
  } else {
    c(paste('<', read_utf8(files[1])), '---', paste('>', read_utf8(files[2])))
  }
  if (length(d) >= len) d = unique(d)
  gsub('^>', '+', gsub('^<', '-', d))
}

# specify a list of package names to be ignored when installing all dependencies
ignore_deps = function() {
  if (file_exists('00ignore_deps')) scan('00ignore_deps', 'character')
}

# download a check summary of a package from CRAN
cran_check_page = function(pkg, con = '00check-cran.log') {
  u = sprintf('https://cran.rstudio.com/web/checks/check_results_%s.html', pkg)
  x = read_utf8(u)
  if (length(i <- grep('Check Details', x, ignore.case = TRUE)) == 0) return()
  x = x[i[1]:length(x)]
  x = gsub('<[^>]+>', '', x)
  x = gsub('&nbsp;', ' ', x)
  x = gsub('&gt;', '>', x)
  x = gsub('&lt;', '<', x)
  x = gsub('\\s+', ' ', x)
  x = paste(trimws(x), collapse = '\n')
  x = gsub('\n\n+', '\n\n', x)
  if (length(con) == 1) writeLines(x, con) else x
}

# download CRAN check summaries of all failed packages
cran_check_pages = function() {
  dirs = list.files('.', '[.]Rcheck$')
  for (d in dirs) {
    if (dir_exists(d)) in_dir(d, cran_check_page(gsub('[.]Rcheck$', '', d)))
  }
}

# parse the check log for missing LaTeX packages and install them
find_missing_latex = function() {
  dirs = list.files('.', '[.]Rcheck2?$')
  pkgs = NULL
  for (d in dirs) {
    if (dir_exists(d)) pkgs = c(pkgs, in_dir(
      d, tinytex::parse_packages('00check.log', quiet = c(TRUE, FALSE, FALSE))
    ))
  }
  pkgs = unique(pkgs)
  if (file_exists(f <- 'latex.txt')) append_unique(pkgs, f)
  pkgs
}

library(testit)

assert("tojson() works", {
  (tojson(NULL) %==% "null")
  (tojson(list()) %==% "{}")
  (has_error(tojson(Sys.Date())))
  (has_error(tojson(NA)))
  (tojson(NA_character_) %==% '"NA"')
  (tojson(1:10) %==% "[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]")
  (tojson(TRUE) %==% "true")
  (tojson(FALSE) %==% "false")

  x = list(a = 1, b = list(c = 1:3, d = "abc"))
  out = '{\n"a": 1,\n"b": {\n"c": [1, 2, 3],\n"d": "abc"\n}\n}'
  (tojson(x) %==% out)

  x = list(c("a", "b"), 1:5, TRUE)
  out = '[["a", "b"], [1, 2, 3, 4, 5], true]'
  (tojson(x) %==% out)

  (tojson(list('"a b"' = 'quotes "\'')) %==% '{\n"\\"a b\\"": "quotes \\"\'"\n}')

  JS = function(x) structure(x, class = "JS_EVAL")
  x = list(a = 1:5, b = JS("function() {return true;}"))
  out = '{\n"a": [1, 2, 3, 4, 5],\n"b": function() {return true;}\n}'
  (tojson(x) %==% out)
})

library(testit)

assert('browser_dom() executes JS, returns rendered DOM, and strips scripts', {
  f = tempfile(fileext = '.html')
  on.exit(unlink(f), add = TRUE)
  writeLines(c(
    '<html><body><div id="out"></div>',
    '<script>document.getElementById("out").textContent = 1 + 1;</script>',
    '</body></html>'
  ), f)
  (unclass(browser_dom(f)) %==% '<html><head></head><body><div id="out">2</div></body></html>')
})

assert('browser_dom() saves to output file', {
  f = tempfile(fileext = '.html')
  out = tempfile(fileext = '.html')
  on.exit(unlink(c(f, out)), add = TRUE)
  writeLines(c(
    '<html><body><p>static</p>',
    '<script>\ndocument.body.insertAdjacentHTML("beforeend", "<p>" + (2+3) + "</p>")</script>',
    '</body></html>'
  ), f)
  res = browser_dom(f, out)
  (res %==% out)
  (unclass(file_string(out)) %==% '<html><head></head><body><p>static</p><p>5</p>\n\n</body></html>')
})

assert('browser_dom(fragment = TRUE) returns body content only', {
  f = tempfile(fileext = '.html')
  on.exit(unlink(f), add = TRUE)
  writeLines('<html><head><title>test</title></head>\n<body>\n <p>content\n</p>\t\n</body></html>', f)
  res = browser_dom(f, fragment = TRUE)
  (unclass(res) %==% '<p>content\n</p>')
})

library(testit)

assert('invalid_utf8() works and respects NA', {
  (invalid_utf8(character()) %==% integer())
  x = 'fa\xE7ile'; Encoding(x) = 'latin1'
  (invalid_utf8(c('aaa', x, NA_character_, '', '\u4e2d\u6587')) %==% 2L)
})

assert('read/write_utf8() works', {
  ascii_txt = c('aa', 'bb', 'cc')
  utf8_txt = c('\u4e2d\u6587', '\u5927\u5bb6\u597d', '\u65e9\u996d')
  latin_txt = local({x = 'fa\xE7ile'; Encoding(x) = 'latin1'; x})
  mixed_txt = c(ascii_txt, latin_txt, utf8_txt)

  write_utf8(ascii_txt, con = (ascii_file = tempfile()))
  write_utf8(utf8_txt, con = (utf8_file = tempfile()))
  write_utf8(latin_txt, con = (latin_file = tempfile()))
  write_utf8(mixed_txt, con = (mixed_file = tempfile()))

  (read_utf8(ascii_file) %==% ascii_txt)
  (read_utf8(utf8_file) %==% utf8_txt)
  (read_utf8(latin_file) %==% latin_txt) # identical will not compare Encoding
  (Encoding(read_utf8(latin_file)) %==% 'UTF-8')
  (read_utf8(mixed_file) %==% mixed_txt)
  (Encoding(read_utf8(mixed_file)) %==% c(rep('unknown', 3), rep('UTF-8', 4)))

  mixed_file2 = tempfile()
  local({
    opts = options(encoding = 'native.enc'); on.exit(options(opts), add = TRUE)
    writeLines(mixed_txt, con = mixed_file2, useBytes = TRUE)
  })
  (suppressWarnings(read_utf8(mixed_file2)[4] != mixed_txt[4]))
  has_warning(read_utf8(mixed_file2))
  has_error(read_utf8(mixed_file2, error = TRUE))
})

assert('empty files produce character() via file_string()', {
  tmp = tempfile()
  writeLines(character(), tmp)
  (file_string(tmp) %==% raw_string(character()))
})

assert('file_string() concatenates non-empty file lines with newlines', {
  f = tempfile()
  write_utf8(c('a', 'b', 'c'), f)
  (file_string(f) %==% raw_string('a\nb\nc'))
  unlink(f)
})

assert('grep_sub() matches elements and do substitution on them', {
  (grep_sub('a([b]+)c', 'a\\U\\1c', c('abc', 'abbbc', 'addc', '123'), perl = TRUE) %==%
     c('aBc', 'aBBBc'))
})

assert('read_bin() reads binary files', {
  f = tempfile()
  writeBin(as.raw(1:10), f)
  (read_bin(f) %==% as.raw(1:10))
  unlink(f)
})

assert('append_utf8() appends content to a file and sorts it by default', {
  f = tempfile()
  write_utf8(c('b', 'a'), f)
  append_utf8('c', f)
  (read_utf8(f) %==% c('a', 'b', 'c'))
})

assert('append_unique() appends only lines not already in the file and sorts', {
  f = tempfile()
  write_utf8(c('b', 'a'), f)
  append_unique(c('a', 'c'), f)  # 'a' is already there; only 'c' is new
  (read_utf8(f) %==% c('a', 'b', 'c'))
})

assert('process_file() applies a function to file contents', {
  f = tempfile()
  write_utf8('Hello World', f)
  process_file(f, function(x) gsub('World', 'R', x))
  (read_utf8(f) %==% 'Hello R')
  unlink(f)
})

assert('process_file() returns transformed result when file is missing', {
  (process_file(fun = toupper, x = 'hello') %==% 'HELLO')
})

assert('sort_file() sorts lines in a file', {
  f = tempfile()
  write_utf8(c('b', 'a', 'c'), f)
  sort_file(f)
  (read_utf8(f) %==% c('a', 'b', 'c'))
  unlink(f)
})

assert('gsub_file() replaces patterns in a file', {
  f = tempfile()
  writeLines(c('hello', 'world'), f)
  gsub_file(f, 'world', 'R', fixed = TRUE)
  (readLines(f) %==% c('hello', 'R'))
  unlink(f)
})

assert('gsub_file() skips file silently when rw_error=FALSE and read fails', {
  # a file with invalid UTF-8 content cannot be read; with rw_error=FALSE it is silently skipped
  f = tempfile()
  writeBin(as.raw(c(0x68, 0x65, 0x6c, 0x6c, 0xc0, 0x80)), f)  # invalid UTF-8 bytes
  result = gsub_file(f, 'a', 'b', rw_error = FALSE)
  (is.null(result))
  unlink(f)
})

assert('read_all() reads and concatenates files', {
  f1 = tempfile(); f2 = tempfile()
  write_utf8(c('a', 'b'), f1)
  write_utf8(c('c', 'd'), f2)
  res = read_all(c(f1, f2))
  (as.character(res) %==% c('a', 'b', 'c', 'd'))
  # with before/after: 1 before + 2 content + 1 after = 4 per file, 8 total
  res2 = read_all(c(f1, f2), before = function(f) paste('#', f), after = '')
  (length(res2) %==% 8L)
  unlink(c(f1, f2))
})

assert('gsub_dir() and gsub_ext() replace patterns in files within a dir', {
  d = tempfile()
  dir.create(d)
  f1 = file.path(d, 'a.txt')
  f2 = file.path(d, 'b.R')
  writeLines('hello world', f1)
  writeLines('foo bar', f2)
  gsub_dir('world', 'R', dir = d, fixed = TRUE)
  (readLines(f1) %==% 'hello R')
  (readLines(f2) %==% 'foo bar')  # unchanged
  # gsub_ext only processes files with specified extension
  gsub_ext('R', 'bar', 'BAR', dir = d, fixed = TRUE)
  (readLines(f2) %==% 'foo BAR')
  # filter by mimetype: only text files
  gsub_dir('R', 'REPLACED', dir = d, fixed = TRUE, mimetype = '^text/')
  (readLines(f1) %==% 'hello REPLACED')
  unlink(d, recursive = TRUE)
})

assert('lazy_save() and lazy_load() round-trip objects', {
  d = tempdir()
  path = file.path(d, 'lazy-test-')
  x = 1:10
  y = list(a = 'hello')
  lazy_save(c('x', 'y'), path)
  # verify index file exists
  (file.exists(paste0(path, '0.rds')))
  # load into new env
  e = new.env(parent = emptyenv())
  lazy_load(path, envir = e)
  (e$x %==% x)
  (e$y %==% y)
  unlink(list.files(d, pattern = '^lazy-test-', full.names = TRUE))
})

assert('write_utf8() handles NULL input', {
  f = tempfile()
  write_utf8(NULL, f)
  (read_utf8(f) %==% character(0))
  unlink(f)
})

assert('gsub_file() with rw_error=FALSE does not stop on read errors', {
  # a non-existent file should return invisible with rw_error=FALSE
  f = tempfile()  # doesn't exist
  result = gsub_file(f, 'a', 'b', rw_error = FALSE)
  (is.null(result))
})

assert('io_method() returns a list method as-is', {
  m = list(name = 'custom', save = saveRDS, load = readRDS)
  (io_method(m, '.') %==% m)
})

assert('write_utf8() to empty string prints to stdout', {
  out = capture.output(write_utf8(c('hello', 'world'), ''))
  (out %==% c('hello', 'world'))
})

assert('append_utf8() with sort = FALSE does not sort', {
  f = tempfile()
  write_utf8(c('b', 'a'), f)
  append_utf8('c', f, sort = FALSE)
  (read_utf8(f) %==% c('b', 'a', 'c'))
  unlink(f)
})

assert('read_all() with constant before/after works', {
  f1 = tempfile(); f2 = tempfile()
  write_utf8('line1', f1)
  write_utf8('line2', f2)
  res = read_all(c(f1, f2), before = '/*', after = '*/')
  (as.character(res) %==% c('/*', 'line1', '*/', '/*', 'line2', '*/'))
  unlink(c(f1, f2))
})

assert('process_file() does not write when content is unchanged', {
  f = tempfile()
  write_utf8('hello', f)
  mtime1 = file.mtime(f)
  Sys.sleep(0.1)
  process_file(f, identity)
  (file.mtime(f) %==% mtime1)
  unlink(f)
})

assert('gsub_file() errors on unreadable file with rw_error = TRUE', {
  (has_error(gsub_file(tempfile(), 'a', 'b', rw_error = TRUE)))
})

assert('gsub_files() processes multiple files', {
  f1 = tempfile(); f2 = tempfile()
  writeLines('hello world', f1)
  writeLines('world hello', f2)
  gsub_files(c(f1, f2), 'world', 'R', fixed = TRUE)
  (readLines(f1) %==% 'hello R')
  (readLines(f2) %==% 'R hello')
  unlink(c(f1, f2))
})

#' Test if a character vector consists of blank strings
#'
#' Return a logical vector indicating if elements of a character vector are
#' blank (white spaces or empty strings).
#' @param x A character vector.
#' @return `TRUE` for blank elements, or `FALSE` otherwise.
#' @export
#' @examples
#' xfun::is_blank('')
#' xfun::is_blank('abc')
#' xfun::is_blank(c('', '  ', '\n\t'))
#' xfun::is_blank(c('', ' ', 'abc'))
is_blank = function(x) grepl('^\\s*$', x)

# remote blank elements from both ends
strip_blank = function(x) {
  if (!(N <- length(x))) return(x)
  r = rle(is_blank(x))
  l = r$lengths; v = r$values; n = length(v)
  i = c(if (v[1]) seq_len(l[1]), if (n > 1 && v[n]) N - seq_len(l[n]) + 1)
  if (length(i)) x[-i] else x
}

#' Convert numbers to English words
#'
#' This can be helpful when writing reports with \pkg{knitr}/\pkg{rmarkdown} if
#' we want to print numbers as English words in the output. The function `n2w()`
#' is an alias of `numbers_to_words()`.
#' @param x A numeric vector. The absolute values should be less than `1e15`.
#' @param cap Whether to capitalize the first letter of the word. This can be
#'   useful when the word is at the beginning of a sentence. Default is `FALSE`.
#' @param hyphen Whether to insert hyphen (-) when the number is between 21 and
#'   99 (except 30, 40, etc.).
#' @param and Whether to insert `and` between hundreds and tens, e.g., write 110
#'   as \dQuote{one hundred and ten} if `TRUE` instead of \dQuote{one hundred
#'   ten}.
#' @return A character vector.
#' @author Daijiang Li
#' @export
#' @examples library(xfun)
#' n2w(0, cap = TRUE)
#' n2w(0:121, and = TRUE)
#' n2w(1e6)
#' n2w(1e11+12345678)
#' n2w(-987654321)
#' n2w(1e15-1)
#' n2w(123.456)
#' n2w(123.45678901)
#' n2w(123.456789098765)
numbers_to_words = function(x, cap = FALSE, hyphen = TRUE, and = FALSE) {
  if (!is.numeric(x)) stop('The input is not numeric.')
  if (any(abs(x) >= 1e15)) stop('The absolute value must be less than 1e15.')
  opts = options(scipen = 15, OutDec = '.')  # avoid scientific notation
  on.exit(options(opts), add = TRUE)

  zero_to_19 = c(
    'zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven', 'eight', 'nine', 'ten',
    'eleven', 'twelve', paste0(c('thir', 'four', 'fif', 'six', 'seven', 'eigh', 'nine'), 'teen')
  )
  names(zero_to_19) = as.character(0:19)
  tens = c('twenty', 'thirty', 'forty', 'fifty', 'sixty', 'seventy', 'eighty', 'ninety')
  names(tens) = as.character(seq(20, 90, 10))
  marks = c('', 'thousand,', 'million,', 'billion,', 'trillion,')

  convert_1 = function(x_c) zero_to_19[x_c]  # 0 - 9

  # 10 - 99
  convert_2 = function(x_c) {
    x_cs = strsplit(x_c, split = '')[[1]]
    if (x_cs[1] == 1) return(zero_to_19[x_c])  # 10 - 19
    if (x_cs[2] == 0) return(tens[x_c])  # 20, 30, 40, ...
    # 21, 22, etc.
    paste(tens[as.integer(x_cs[1]) - 1], convert_1(x_cs[2]), sep = if (hyphen) '-' else ' ')
  }

  # 100 - 999
  convert_3 = function(x_c) {
    x_cs = strsplit(x_c, split = '')[[1]]
    n_hundreds = paste(convert_1(x_cs[1]), 'hundred', sep = ' ')
    out = if (x_cs[2] == '0') {
      if (x_cs[3] == '0') return(n_hundreds)  # x00
      convert_1(x_cs[3])  # x0x
    } else {
      convert_2(paste(x_cs[2:3], collapse = ''))  # xxx
    }
    paste(n_hundreds, out, sep = if (and) ' and ' else ' ')
  }

  convert_le3 = function(x_c) {
    x_c = gsub('^0+', '', x_c) # avoid something like 000, 001, 010; but also remove 0
    n = nchar(x_c)
    if (n == 0) return('')
    if (n == 1) return(convert_1(x_c))
    if (n == 2) return(convert_2(x_c))
    if (n == 3) return(convert_3(x_c))
  }

  convert_one = function(x) {
    minus = if (x >= 0) '' else {
      x = abs(x); 'minus '
    }
    if (x == 0) {
      out = 'zero'  # because convert_le3 removed all 0s
    } else {
      x_marks = strsplit(format(floor(x), big.mark = ','), split = ',')[[1]]  # e.g. 123,456,789
      out = vapply(x_marks, convert_le3, character(1))  # group by 3 digits
      x_marks2 = marks[length(x_marks):1]  # units?
      x_marks2[which(out == '')] = ''  # e.g. 4,000,123, 000, remove millions
      out = paste(out, x_marks2, sep = ' ', collapse = ' ')  # zip together
    }
    out = paste0(minus, out)
    out = gsub('^ *|,? *$', '', out)  # trim heading/trailing space
    out = gsub(' {2,}', ' ', out)  # remove multiple spaces
    if (cap) out = sub('^([a-z])', '\\U\\1', out, perl = TRUE)
    if (x - floor(x) > 0) {
      frac = sub('^[0-9]+[.]', '', as.character(x))
      frac = convert_1(strsplit(frac, '')[[1]])
      out = paste(c(out, 'point', frac), collapse = ' ')
    }
    out
  }

  if (length(x) > 1) vapply(x, convert_one, character(1)) else convert_one(x)
}

#' @export
#' @rdname numbers_to_words
n2w = numbers_to_words

#' Join multiple words into a single string
#'
#' If `words` is of length 2, the first word and second word are joined by the
#' `and` string; if `and` is blank, `sep` is used. When the length is greater
#' than 2, `sep` is used to separate all words, and the `and` string is
#' prepended to the last word.
#' @param words A character vector.
#' @param sep Separator to be inserted between words.
#' @param and Character string to be prepended to the last word.
#' @param before,after A character string to be added before/after each word.
#' @param oxford_comma Whether to insert the separator between the last two
#'   elements in the list.
#' @return A character string marked by [raw_string()].
#' @export
#' @examples join_words('a'); join_words(c('a', 'b'))
#' join_words(c('a', 'b', 'c'))
#' join_words(c('a', 'b', 'c'), sep = ' / ', and = '')
#' join_words(c('a', 'b', 'c'), and = '')
#' join_words(c('a', 'b', 'c'), before = '"', after = '"')
#' join_words(c('a', 'b', 'c'), before = '"', after = '"', oxford_comma = FALSE)
join_words = function(
  words, sep = ', ', and = ' and ', before = '', after = before, oxford_comma = TRUE
) {
  n = length(words)
  if (n > 0) {
    words = paste0(before, words, after)
    if (n == 2) {
      words = paste(words, collapse = if (is_blank(and)) sep else and)
    } else if (n > 2) {
      if (oxford_comma && grepl('^ ', and) && grepl(' $', sep)) and = gsub('^ ', '', and)
      words[n] = paste0(and, words[n])
      # combine the last two words directly without the comma
      if (!oxford_comma) {
        words[n - 1] = paste0(words[n - 1:0], collapse = '')
        words = words[-n]
      }
      words = paste(words, collapse = sep)
    }
  }
  raw_string(words)
}

#' Evaluate an expression after forcing the decimal point to be a dot
#'
#' Sometimes it is necessary to use the dot character as the decimal separator.
#' In R, this could be affected by two settings: the global option
#' `options(OutDec)` and the `LC_NUMERIC` locale. This function sets the former
#' to `.` and the latter to `C` before evaluating an expression, such as
#' coercing a number to character.
#' @param x An expression.
#' @export
#' @return The value of `x`.
#' @examples
#' opts = options(OutDec = ',')
#' as.character(1.234)  # using ',' as the decimal separator
#' print(1.234)  # same
#' xfun::decimal_dot(as.character(1.234))  # using dot
#' xfun::decimal_dot(print(1.234))  # using dot
#' options(opts)
decimal_dot = function(x) {
  opts = options(OutDec = '.'); on.exit(options(opts), add = TRUE)
  lcn = Sys.getlocale('LC_NUMERIC')
  if (lcn != 'C') {
    Sys.setlocale('LC_NUMERIC', 'C')
    on.exit(suppressWarnings(Sys.setlocale('LC_NUMERIC', lcn)), add = TRUE)
  }
  x
}

# create a URL query string from named parameters
query_params = function(..., .list = list()) {
  x = if (missing(.list)) list(...) else .list
  x = paste(names(x), x, sep = '=', collapse = '&')
  if (x != '') paste0('?', x) else x
}

#' Wrap character vectors
#'
#' A wrapper function to make [strwrap()] return a character vector of the same
#' length as the input vector; each element of the output vector is a string
#' formed by concatenating wrapped strings by `\n`.
#' @param ... Arguments passed to [strwrap()].
#' @return A character vector.
#' @export
#' @examples
#' x = sample(c(letters, ' '), 200, TRUE, c(rep(.5/26, 26), .5))
#' x = rep(paste(x, collapse = ''), 2)
#' strwrap(x, width = 30)
#' xfun::str_wrap(x, width = 30)  # same length as x
str_wrap = function(...) {
  res = strwrap(..., simplify = FALSE)
  unlist(lapply(res, one_string))
}

#' Split a character vector by line breaks
#'
#' Call `unlist(strsplit(x, '\n'))` on the character vector `x` and
#' make sure it works in a few edge cases: `split_lines('')` returns
#' `''` instead of `character(0)` (which is the returned value of
#' `strsplit('', '\n')`); `split_lines('a\n')` returns `c('a',
#' '')` instead of `c('a')` (which is the returned value of
#' `strsplit('a\n', '\n')`.
#' @param x A character vector.
#' @return All elements of the character vector are split by `'\n'` into
#'   lines.
#' @export
#' @examples xfun::split_lines(c('a', 'b\nc'))
split_lines = function(x) {
  if (length(grep('\n', x)) == 0L) return(x)
  x = gsub('\n$', '\n\n', x)
  x[x == ''] = '\n'
  unlist(strsplit(x, '\r?\n'))
}

#' Split source lines into complete expressions
#'
#' Parse the lines of code one by one to find complete expressions in the code,
#' and put them in a list.
#' @param x A character vector of R source code.
#' @param merge_comments Whether to merge consecutive lines of comments as a
#'   single expression to be combined with the next non-comment expression (if
#'   any).
#' @param line_number Whether to store the line numbers of each expression in
#'   the returned value.
#' @return A list of character vectors, and each vector contains a complete R
#'   expression, with an attribute `lines` indicating the starting and ending
#'   line numbers of the expression if the argument `line_number = TRUE`.
#' @export
#' @examples
#' code = c('# comment 1', '# comment 2', 'if (TRUE) {', '1 + 1', '}', 'print(1:5)')
#' xfun::split_source(code)
#' xfun::split_source(code, merge_comments = TRUE)
split_source = function(x, merge_comments = FALSE, line_number = FALSE) {
  if ((n <- length(x)) < 1) return(list(x))
  i1 = i2 = 1L
  res = list()
  add_source = function(x) {
    res[[length(res) + 1]] <<- if (line_number) structure(x, lines = c(i1, i2)) else x
  }
  while (i2 <= n) {
    piece = x[i1:i2]
    if ((!merge_comments || (!all(grepl('^#', piece)) || i2 == n)) && valid_syntax(piece)) {
      add_source(piece)
      i1 = i2 + 1L # start from the next line
    }
    i2 = i2 + 1L
  }
  if (i1 <= n) parse(text = piece)  # must be an error there
  res
}

#' Check if the syntax of the code is valid
#'
#' Try to [parse()] the code and see if an error occurs.
#' @param code A character vector of R source code.
#' @param silent Whether to suppress the error message when the code is not
#'   valid.
#' @return `TRUE` if the code could be parsed, otherwise `FALSE`.
#' @export
#' @examples xfun::valid_syntax('1+1')
#' xfun::valid_syntax('1+')
#' xfun::valid_syntax(c('if(T){1+1}', 'else {2+2}'), silent = FALSE)
valid_syntax = function(code, silent = TRUE) {
  !inherits(suppressWarnings(try(parse_only(code), silent = silent)), 'try-error')
}

#' Bump version numbers
#'
#' Increase the last digit of version numbers, e.g., from `0.1` to
#' `0.2`, or `7.23.9` to `7.23.10`.
#' @param x A vector of version numbers (of the class `"numeric_version"`),
#'   or values that can be coerced to version numbers via
#'   `as.numeric_version()`.
#' @return A vector of new version numbers.
#' @export
#' @examples xfun::bump_version(c('0.1', '91.2.14'))
bump_version = function(x) {
  x = as.numeric_version(x)
  for (i in seq_along(x)) {
    v = x[i]
    n = length(unclass(v)[[1]])
    v[[1, n]] = v[[1, n]] + 1  # bump the last digit
    x[i] = v
  }
  x
}

#' Fix pairs of characters in a file
#'
#' For example, the curly braces may be wrong (the opening and closing braces
#' are swapped for some reason).
#' @param x A character vector (by default, read from `file`).
#' @param file Path to a text file.
#' @param chars A vector of characters of length 2. By default, it is a pair of
#'   curly double quotes.
#' @references <https://d.cosx.org/d/420794/5>
#' @noRd
#' @examples
#' files = list.files('.', '[.]R?md$', recursive = TRUE, full.names = TRUE)
#' for (f in files) {
#'   pair_chars(file = f)
#'   # curly single quotes
#'   pair_chars(file = f, chars = c('\U2018', '\U2019'))
#' }
pair_chars = function(x = read_utf8(file), file, chars = c('\U201c', '\U201d')) {
  if (length(chars) != 2) stop("'chars' must be of length 2 (i.e., a pair of characters)")
  is_file = !missing(file)
  r = paste(c('[', chars, ']'), collapse = '')
  k = gregexpr(r, x)
  m = regmatches(x, k)
  for (i in seq_along(m)) {
    n = length(m[[i]])
    if (n %% 2 != 0) {
      warning(
        'The characters do not appear in pairs in the text (',
        'line: ', i, if (is_file) c('; file: ', file), '):\n', x[i], '\n'
      )
      next
    }
    m[[i]] = rep(chars, length.out = n)
  }
  x2 = x
  regmatches(x, k) = m
  if (is_file) {
    if (!identical(x, x2)) write_utf8(x, file)
    invisible(x)
  } else x
}

#' Generate ID strings
#'
#' Substitute certain (by default, non-alphanumeric) characters with dashes and
#' remove extra dashes at both ends to generate ID strings. This function is
#' intended for generating IDs for HTML elements, so HTML tags in the input text
#' will be removed first.
#' @param x A character vector.
#' @param exclude A (Perl) regular expression to detect characters to be
#'   replaced by dashes. By default, non-alphanumeric characters are replaced.
#' @return A character vector of IDs.
#' @export
#' @examples
#' x = c('Hello world 123!', 'a  &b*^##c 456')
#' xfun::alnum_id(x)
#' xfun::alnum_id(x, '[^[:alpha:]]+')  # only keep alphabetical chars
#' # when text contains HTML tags
#' xfun::alnum_id('<h1>Hello <strong>world</strong>!')
alnum_id = function(x, exclude = '[^[:alnum:]]+') {
  x = strip_html(x)
  tolower(gsub('^-+|-+$', '', gsub(exclude, '-', x, perl = TRUE)))
}

#' Strip HTML tags
#'
#' Remove HTML tags and comments from text.
#' @param x A character vector.
#' @return A character vector with HTML tags and comments stripped off.
#' @export
#' @examples
#' xfun::strip_html('<a href="#">Hello <!-- comment -->world!</a>')
strip_html = function(x) {
  x = gsub('<!--.*?-->', '', x)
  x = gsub('<[^>]+>', '', x)
  x
}

# no need to close these tags
.void_tags = c(
  'area', 'base', 'br', 'col', 'command', 'embed', 'hr', 'img', 'input',
  'keygen', 'link', 'meta', 'param', 'source', 'track', 'wbr'
)

.html_class2 = c(.html_class <- c('xfun_html', 'html'), 'xfun_raw_string')

#' Tools for HTML tags
#'
#' Given a tag name, generate an HTML tag with optional attributes and content.
#' [html_tag()] can be viewed as a simplified version of `htmltools::tags`,
#' [html_value()] adds classes on the value so that it will be treated as raw
#' HTML (not escaped by `html_tag()`), [html_escape()] escapes special
#' characters in HTML, and `html_view()` launches a browser or viewer to view
#' the HTML content.
#' @param .name The tag name.
#' @param .content The content between opening and closing tags. Ignored for
#'   void tags such as `<img>`. Special characters such as `&`, `<`, and `>`
#'   will be escaped unless the value was generated from [html_value()]. The
#'   content can be either a character vector or a list. If it is a list, it may
#'   contain both normal text and HTML content.
#' @param .attrs A named list of attributes.
#' @param ... For `html_tag()`, named arguments as an alternative way to provide
#'   attributes. For `html_view()`, other arguments to be passed to [new_app()].
#' @return A character string.
#' @export
#' @examples
#' xfun::html_tag('a', '<R Project>', href = 'https://www.r-project.org', target = '_blank')
#' xfun::html_tag('br')
#' xfun::html_tag('a', xfun::html_tag('strong', 'R Project'), href = '#')
#' xfun::html_tag('a', list('<text>', xfun::html_tag('b', 'R Project')), href = '#')
html_tag = function(.name, .content = NULL, .attrs = NULL, ...) {
  if (is.null(.attrs)) .attrs = list(...)
  if (length(.attrs) && (is.null(nm <- names(.attrs)) || any(nm == '')))
    stop('Tag attributes must be named values')
  x1 = c('<', .name)
  x2 = if (length(.attrs)) .mapply(function(a, v) {
    if (is.null(v)) a else sprintf('%s="%s"', a, html_escape(v, TRUE))
  }, list(nm, .attrs), NULL)
  x2 = paste(unlist(x2), collapse = ' ')
  x3 = if (.name %in% .void_tags) ' />' else {
    c('>', one_string(html_content(.content)), '</', .name, '>')
  }
  x = c(x1, if (x2 != '') c(' ', x2), x3)
  x = paste(x, collapse = '')
  html_value(x)
}

# recursively resolve HTML content if it's a list containing both HTML and regular text
html_content = function(x) {
  if (is.list(x)) unlist(lapply(x, html_content)) else {
    if (length(x) && !inherits(x, .html_class)) x = html_escape(x)
    x
  }
}

#' @param x A character vector to be treated as raw HTML content for
#'   `html_value()`, escaped for `html_escape()`, and viewed for `html_view()`.
#' @rdname html_tag
#' @export
html_value = function(x) structure(x, class = .html_class2, lang = '.html')

#' @param attr Whether to escape `"`, `\r`, and `\n` (which should be escaped
#'   for tag attributes).
#' @rdname html_tag
#' @export
#' @examples
#' xfun::html_escape('" quotes " & brackets < >')
#' xfun::html_escape('" & < > \r \n', attr = TRUE)
html_escape = function(x, attr = FALSE) {
  x = gsubf('&', '&amp;', x)
  x = gsubf('<', '&lt;', x)
  x = gsubf('>', '&gt;', x)
  # for attributes, we still need to escape more characters
  if (attr) {
    x = gsubf('"', '&quot;', x)
    x = gsubf('\r', '&#13;', x)
    x = gsubf('\n', '&#10;', x)
  }
  x
}

#' @rdname html_tag
#' @export
html_view = function(x, ...) {
  new_app('xfun-html', function(path, ...) {
    if (dir_exists(path)) list(payload = if (path == '.') x else path) else {
      list(file = normalizePath(path), `content-type` = mime_type(path))
    }
  }, ...)
}

one_string = function(x, ...) paste(x, ..., collapse = '\n')

# en/decrypt a string via a character map (old and new must be 16 unique hex chars)
.crypt = function(x, old, new) {
  x2 = chartr(old, new, as.character(charToRaw(x)))
  rawToChar(as.raw(strtoi(x2, 16L)))
}
encrypt = function(x, key) .crypt(x, '0123456789abcdef', key)
decrypt = function(x, key) .crypt(x, key, '0123456789abcdef')

library(testit)

yaml = '
a: 1
b: [1, 2, 3]
c: true
d:
  e: !expr 1+1
  f: null
'

if (loadable('yaml')) assert('yaml_load() works with the yaml package', {
  (yaml_load(yaml) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (yaml_load(yaml, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
  # this test won't work for yaml::yaml.load() but works for yaml_load()
  f = function() {
    foo = 1:10
    yaml_load('a: !expr head(foo, 4)')
  }
  (f() %==% list(a = 1:4))
})

assert('yaml_load() works without the yaml package', {
  (yaml_load(yaml, use_yaml = FALSE) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (yaml_load(yaml, envir = FALSE, use_yaml = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
  # this test won't work for yaml::yaml.load() but works for yaml_load()
  f = function() {
    foo = 1:10
    yaml_load('a: !expr head(foo, 4)', use_yaml = FALSE)
  }
  (f() %==% list(a = 1:4))
})

#' Find the indices of lines in Markdown that are prose (not code blocks)
#'
#' Filter out the indices of lines between code block fences such as \verb{```}
#' (could be three or four or more backticks).
#' @param x A character vector of text in Markdown.
#' @param warn Whether to emit a warning when code fences are not balanced.
#' @note If the code fences are not balanced (e.g., a starting fence without an
#'   ending fence), this function will treat all lines as prose.
#' @return An integer vector of indices of lines that are prose in Markdown.
#' @export
#' @examples library(xfun)
#' prose_index(c('a', '```', 'b', '```', 'c'))
#' prose_index(c('a', '````', '```r', '1+1', '```', '````', 'c'))
prose_index = function(x, warn = TRUE) {
  idx = NULL
  # if raw HTML <pre></pre> exists, it should be treated as code block
  inside_pre = if (length(p1 <- grep('<pre>', x))) {
    p2 = grep('</pre>', x)
    if (length(p1) == length(p2)) {
      idx = rbind(p1, p2)
      function(i) any(i > p1 & i < p2)
    }
  }
  r = '^(\\s*```+).*'; s = ''
  # shouldn't match ``` ``text``` ```, which is inline code, not code block
  i1 = grepl(r, x); i2 = !grepl('^\\s*```+\\s+`', x); i3 = !grepl('-->\\s*$', x)
  for (i in which(i1 & i2 & i3)) {
    if (is.function(inside_pre) && inside_pre(i)) next
    if (s == '') {
      s = gsub(r, '\\1', x[i]); idx = c(idx, i); next
    }
    # look for the next line with the same amount of backticks (end of block)
    if (grepl(paste0('^', s), x[i])) {
      idx = c(idx, i); s = ''
    }
  }
  xi = seq_along(x); n = length(idx)
  if (n == 0) return(xi)
  if (n %% 2 != 0) {
    if (warn) warning('Code fences are not balanced')
    # treat all lines as prose
    return(xi)
  }
  idx2 = matrix(idx, nrow = 2)
  idx2 = unlist(mapply(seq, idx2[1, ], idx2[2, ], SIMPLIFY = FALSE))
  xi[-idx2]
}

#' Protect math expressions in pairs of backticks in Markdown
#'
#' For Markdown renderers that do not support LaTeX math, we need to protect
#' math expressions as verbatim code (in a pair of backticks), because some
#' characters in the math expressions may be interpreted as Markdown syntax
#' (e.g., a pair of underscores may make text italic). This function detects
#' math expressions in Markdown (by heuristics), and wrap them in backticks.
#'
#' Expressions in pairs of dollar signs or double dollar signs are treated as
#' math, if there are no spaces after the starting dollar sign, or before the
#' ending dollar sign. There should be spaces before the starting dollar sign,
#' unless the math expression starts from the very beginning of a line. For a
#' pair of single dollar signs, the ending dollar sign should not be followed by
#' a number. With these assumptions, there should not be too many false
#' positives when detecing math expressions.
#'
#' Besides, LaTeX environments (\verb{\begin{*}} and \verb{\end{*}}) are also
#' protected in backticks.
#' @param x A character vector of text in Markdown.
#' @param token A character string to wrap math expressions at both ends. This
#'   can be a unique token so that math expressions can be reliably identified
#'   and restored after the Markdown text is converted.
#' @return A character vector with math expressions in backticks.
#' @note If you are using Pandoc or the \pkg{rmarkdown} package, there is no
#'   need to use this function, because Pandoc's Markdown can recognize math
#'   expressions.
#' @export
#' @examples library(xfun)
#' protect_math(c('hi $a+b$', 'hello $$\\alpha$$', 'no math here: $x is $10 dollars'))
#' protect_math(c('hi $$', '\\begin{equation}', 'x + y = z', '\\end{equation}'))
#' protect_math('$a+b$', '===')
protect_math = function(x, token = '') {
  i = prose_index(x)
  if (length(i)) x[i] = escape_math(x[i], token)
  x
}

escape_math = function(x, token = '') {
  # replace $x$ with `\(x\)` (protect inline math in <code></code>)
  m = gregexpr('(?<=^|[\\s])[$](?! )[^$]+?(?<! )[$](?![$0123456789])', x, perl = TRUE)
  regmatches(x, m) = lapply(regmatches(x, m), function(z) {
    if (length(z) == 0) return(z)
    z = sub('^[$]', paste0('`', token, '\\\\('), z)
    z = sub('[$]$', paste0('\\\\)', token, '`'), z)
    z
  })
  # replace $$x$$ with `$$x$$` (protect display math)
  m = gregexpr('(?<=^|[\\s])[$][$](?! )[^$]+?(?<! )[$][$]', x, perl = TRUE)
  regmatches(x, m) = lapply(regmatches(x, m), function(z) {
    if (length(z) == 0) return(z)
    paste0('`', token, z, token, '`')
  })
  # now, if there are still lines starting and ending with $$, they might be
  # math expressions of display style spanning multiple lines, e.g.,
  # $$\alpha +
  # \beta$$
  # we assume that $$ can only appear once on one line
  i = vapply(gregexpr('[$]', x), length, integer(1)) == 2
  if (any(i)) {
    x[i] = gsub('^(\\s*)([$][$][^ ]+)', paste0('\\1`', token, '\\2'), x[i], perl = TRUE)
    x[i] = gsub('([^ ][$][$])$', paste0('\\1', token, '`'), x[i], perl = TRUE)
  }
  # equation environments (\begin and \end must match)
  i1 = grep('^\\\\begin\\{[^}]+\\}$', x)
  i2 = grep('^\\\\end\\{[^}]+\\}$', x)
  if (length(i1) == length(i2)) {
    # TODO: do not protect inner environments in case of nested environments (#57)
    x[i1] = paste0('`', token, x[i1])
    x[i2] = paste0(x[i2], token, '`')
  }
  x
}

#' Create a fenced block in Markdown
#'
#' Wrap content with fence delimiters such as backticks (code blocks) or colons
#' (fenced Div). Optionally the fenced block can have attributes. The function
#' `fenced_div()` is a shorthand of `fenced_block(char = ':')`.
#' @param x A character vector of the block content.
#' @param attrs A vector of block attributes.
#' @param fence The fence string, e.g., `:::` or ```` ``` ````. This will be
#'   generated from the `char` argument by default.
#' @param char The fence character to be used to generate the fence string by
#'   default.
#' @return `fenced_block()` returns a character vector that contains both the
#'   fences and content.
#' @export
#' @examples
#' # code block with class 'r' and ID 'foo'
#' xfun::fenced_block('1+1', c('.r', '#foo'))
#' # fenced Div
#' xfun::fenced_block('This is a **Div**.', char = ':')
fenced_block = function(x, attrs = NULL, fence = make_fence(x, char), char = '`') {
  a = block_attr(attrs)
  c('', paste0(fence, a), x, fence)
}

#' @param ... Arguments to be passed to `fenced_block()`.
#' @rdname fenced_block
#' @export
fenced_div = function(...) fenced_block(..., char = ':')

#' @return `make_fence()` returns a character string. If the block content
#'   contains `N` fence characters (e.g., backticks), use `N + 1` characters as
#'   the fence.
#' @rdname fenced_block
#' @export
#' @examples
#' # three backticks by default
#' xfun::make_fence('1+1')
#' # needs five backticks for the fences because content has four
#' xfun::make_fence(c('````r', '1+1', '````'))
make_fence = function(x, char = '`') {
  f = strrep(char, 3)
  while (any(grepl(f, x, fixed = TRUE))) f = paste0(f, char)
  f
}

# concatenate block attributes for fenced blocks
block_attr = function(attrs) {
  a = paste(attrs, collapse = ' ')
  if (grepl('[ .=#]', a)) a = paste0('{', a, '}')
  if (a == '') a else paste0(' ', a)
}

#' Embed a file, multiple files, or directory on an HTML page
#'
#' For a file, first encode it into base64 data (a character string). Then
#' generate a hyperlink of the form \samp{<a href="base64 data"
#' download="filename">Download filename</a>}. The file can be downloaded when
#' the link is clicked in modern web browsers. For a directory, it will be
#' compressed as a zip archive first, and the zip file is passed to
#' `embed_file()`. For multiple files, they are also compressed to a zip file
#' first.
#'
#' These functions can be called in R code chunks in R Markdown documents with
#' HTML output formats. You may embed an arbitrary file or directory in the HTML
#' output file, so that readers of the HTML page can download it from the
#' browser. A common use case is to embed data files for readers to download.
#' @param path Path to the file(s) or directory.
#' @param name The default filename to use when downloading the file. Note that
#'   for `embed_dir()`, only the base name (of the zip filename) will be used.
#' @param text The text for the hyperlink.
#' @param ... For `embed_file()`, additional arguments to be passed to
#'   [html_tag()] (e.g., `class = 'foo'`). For `embed_dir()` and
#'   `embed_files()`, arguments passed to `embed_file()`.
#' @note Windows users may need to install Rtools to obtain the \command{zip}
#'   command to use `embed_dir()` and `embed_files()`.
#'
#'   Currently Internet Explorer does not support downloading embedded files
#'   (<https://caniuse.com/download>). Chrome has a 2MB limit on the file size.
#' @return An HTML tag \samp{<a>} with the appropriate attributes.
#' @export
#' @examples
#' logo = xfun:::R_logo()
#' link = xfun::embed_file(logo, text = 'Download R logo')
#' link
#' if (interactive()) xfun::html_view(link)
embed_file = function(path, name = basename(path), text = paste('Download', name), ...) {
  h = base64_uri(path)
  html_tag('a', text, href = h, download = name, ...)
}

#' @rdname embed_file
#' @export
embed_dir = function(path, name = paste0(normalize_path(path), '.zip'), ...) {
  name  = gsub('/', '', basename(name))
  in_dir(path, {
    name = file.path(tempdir(), name); on.exit(file.remove(name), add = TRUE)
    zip(name, '.'); embed_file(name, ...)
  })
}

#' @rdname embed_file
#' @export
embed_files = function(path, name = with_ext(basename(path[1]), '.zip'), ...) {
  name = file.path(tempdir(), basename(name))
  on.exit(file.remove(name), add = TRUE)
  zip(name, path)
  embed_file(name, ...)
}

zip = function(name, ...) {
  if (utils::zip(name, ...) != 0) stop('Failed to create the zip archive ', name)
  invisible(0)
}

#' Generate a simple Markdown pipe table
#'
#' A minimal Markdown table generator using the pipe `|` as column separators.
#'
#' The default argument values can be set via global options with the prefix
#' `xfun.md_table.`, e.g., `options(xfun.md_table.digits 2, xfun.md_table.na =
#' 'n/a')`.
#' @param x A 2-dimensional object (e.g., a matrix or data frame).
#' @param digits The number of decimal places to be passed to [round()]. It can
#'   be a integer vector of the same length as the number of columns in `x` to
#'   round columns separately. The default is `3`.
#' @param na A character string to represent `NA` values. The default is an
#'   empty string.
#' @param newline A character string to substitute `\n` in `x` (because pipe
#'   tables do not support line breaks in cells). The default is a space.
#' @param limit The maximum number of rows to show in the table. If it is
#'   smaller than the number of rows, the data in the middle will be omitted. If
#'   it is of length 2, the second number will be used to limit the number of
#'   columns. Zero and negative values are ignored.
#' @return A character vector.
#' @seealso [knitr::kable()] (which supports more features)
#' @export
#' @examples
#' xfun::md_table(head(iris))
#' xfun::md_table(mtcars, limit = c(10, 6))
md_table = function(x, digits = NULL, na = NULL, newline = NULL, limit = NULL) {
  if (length(d <- dim(x)) != 2)
    stop('xfun::md_table() only supports 2-dimensional objects.')
  if (d[2] == 0) return(character())
  if (is.null(digits))
    digits = getOption('xfun.md_table.digits', min(getOption('digits'), 3))
  digits = rep(digits, d[2])  # recycle for all columns
  num = logical(d[2])  # numeric columns
  for (j in seq_len(d[2])) if (is.numeric(x[, j])) {
    num[j] = TRUE
    x[, j] = round(x[, j], digits[j])
  }
  is_na = is.na(x)
  x = as.matrix(format(x))
  if (any(is_na)) x[is_na] = na %||% getOption('xfun.md_table.na', '')
  rn = rownames(x)
  # ignore empty and automatic row names (integers)
  if (!all(grepl('^[0-9]*$', rn))) {
    x = cbind(' ' = rn, x)
    num = c(FALSE, num)
  }
  # get first and last limit/2 rows/cols in N rows/cols
  if (length(limit <- limit %||% getOption('xfun.md_table.limit'))) {
    # subset rows
    l1 = limit[1]
    if (l1 > 0 && l1 < d[1]) {
      n1 = round(l1/2); n2 = l1 - n1
      x = rbind(head(x, n1), '&vellip;', tail(x, n2))
    }
    # subset columns
    if (length(limit) >= 2 && (l2 <- limit[2]) > 0 && l2 < d[2]) {
      n1 = round(l2/2); n2 = l2 - n1
      x = cbind(
        x[, seq_len(n1), drop = FALSE],
        `...` = '...',
        x[, d[2] - seq_len(n2) + 1, drop = FALSE]
      )
      num = c(head(num, n1), FALSE, tail(num, n2))
    }
    d = dim(x)
    if (d[2] == 0) return(character())
  }
  cn = colnames(x) %||% rep(' ', d[2])  # table header
  a = ifelse(num, '--:', '---')  # alignment
  a[cn == '...'] = ':-:'
  x = rbind(cn, a, x)
  d = dim(x)
  x = gsub('|', '&#124;', x, fixed = TRUE)
  dim(x) = d
  res = do.call(function(...) paste(..., sep = '|'), as.data.frame(x))
  res = gsub('\n', newline %||% getOption('xfun.md_table.newline', ' '), res, fixed = TRUE)
  paste0('|', res, '|')
}

#' A simple JSON serializer
#'
#' A JSON serializer that only works on a limited types of R data (`NULL`,
#' lists, arrays, logical/character/numeric/date/time vectors). Other types of
#' data will be coerced to character. A character string of the class
#' `JS_LITERAL` is treated as raw JavaScript, so will not be quoted. The
#' function `json_vector()` converts an atomic R vector to JSON.
#'
#' Both `NULL` and `NA` are converted to `null`. Named lists are converted to
#' objects of the form `{key1: value1, key2: value2, ...}`. Unnamed lists are
#' converted to arrays of the form `[[value1], [value2], ...]`. The same rules
#' apply to data frames since technically they are also lists. However, please
#' note that unnamed data frames (i.e., without column names) will be converted
#' to an array with each _row_ as an array element, whereas named data frames
#' will have each _column_ as an individual element. For matrices, the JSON
#' array will have each row as an individual element, and names are discarded.
#'
#' Dates and times are coerced to character using UTC as the timezone, and
#' represented via the JavaScript expression `new Date(value)` (which is not
#' standard JSON but practically more useful).
#' @param x An R object.
#' @export
#' @return A character string.
#' @seealso The \pkg{jsonlite} package provides a full JSON serializer.
#' @examples library(xfun)
#' tojson(NULL); tojson(1:10); tojson(TRUE); tojson(FALSE)
#' tojson(list(a = 1, b = list(c = 1:3, d = 'abc')))
#' tojson(list(c('a', 'b'), 1:5, TRUE, Sys.Date() + 1:3))
#' tojson(head(iris))  # each column is in an element
#' tojson(unname(head(iris)))  # each row is in an element
#' tojson(matrix(1:12, 3))
#'
#' # literal JS code
#' JS = function(x) structure(x, class = 'JS_LITERAL')
#' tojson(list(a = 1:5, b = JS('function() {return true;}')))
tojson = function(x) {
  if (inherits(x, 'json')) return(x)
  res = structure(.tojson(x), class = 'json')
  raw_string(res, lang = '.json')
}

.tojson = function(x, n = 1) {
  make_array = function(..., braces = c('[', ']')) {
    inner = paste0(strrep('  ', n), ..., collapse = ',\n')
    paste0(braces[1], '\n', inner, '\n', strrep('  ', n - 1), braces[2])
  }
  if (is.null(x)) 'null' else if (is.array(x)) {
    make_array(apply(x, 1, .tojson, n + 1))
  } else if (is.list(x)) {
    if (length(x) == 0) return('{}')
    # output unnamed data frames by rows instead of columns
    nms = names(x)
    by_row = is.data.frame(x) && is.null(nms)
    cols = unlist(lapply(x, function(z) {
      if (by_row) json_atomic(z, FALSE) else .tojson(z, n + 1)
    }))
    if (is.null(nms)) {
      if (by_row) {
        dim(cols) = dim(x)
        cols = apply(cols, 1, json_vector, TRUE, FALSE)
      }
      make_array(cols)
    } else {
      make_array(quote_string(nms), ': ', cols, braces = c('{', '}'))
    }
  } else if (is.character(x) && inherits(x, c('JS_LITERAL', 'JS_EVAL'))) {
    paste(x, collapse = '\n')
  } else json_atomic(x)
}

json_atomic = function(x, to_array = NA) {
  use_quote = !(is.numeric(x) || is.logical(x))
  asis = inherits(x, 'AsIs')
  if (is.factor(x)) x = as.character(x)
  if (is.logical(x)) x = tolower(as.character(x))
  if (inherits(x, c('Date', 'POSIXct', 'POSIXt'))) {
    x = sprintf('new Date("%s")', format(x, tz = 'UTC'))
    use_quote = FALSE
  }
  if (is.na(to_array)) to_array = length(x) != 1 || asis
  json_vector(x, to_array, use_quote)
}

#' @param to_array Whether to convert a vector to a JSON array (use `[]`).
#' @param quote Whether to double quote the elements.
#' @rdname tojson
#' @export
json_vector = function(x, to_array = FALSE, quote = TRUE) {
  i = is.na(x)
  if (quote) {
    x = quote_string(x)
    x = gsub('\n', '\\\\n', x)
    x = gsub('\b', '\\\\b', x)
    x = gsub('\f', '\\\\f', x)
    x = gsub('\r', '\\\\r', x)
    x = gsub('\t', '\\\\t', x)
  }
  x[i] = 'null'
  if (to_array) paste0('[', paste(x, collapse = ', '), ']') else x
}

# escape \ and " in strings, and quote them
quote_string = function(x) {
  x = gsub('(["\\])', "\\\\\\1", x)
  if (length(x)) x = paste0('"', x, '"')
  x
}

library(testit)

assert('cache_exec() with disk path caches and retrieves', {
  d = tempfile(); dir.create(d)
  path = paste0(d, '/')
  r1 = cache_exec({Sys.time()}, path = path, id = 'disk-test')
  Sys.sleep(0.01)
  r2 = cache_exec({Sys.time()}, path = path, id = 'disk-test')
  (r1 %==% r2)
  unlink(d, recursive = TRUE)
})

assert('cache_exec() invalidates when hash changes', {
  x = 1
  r1 = cache_exec({x + 1}, path = ':memory:', id = 'hash-change-test')
  (r1 %==% 2)
  x = 10
  r2 = cache_exec({x + 1}, path = ':memory:', id = 'hash-change-test')
  (r2 %==% 11)
})

assert('cache_exec() with keep = FALSE always re-evaluates', {
  r1 = cache_exec({runif(1)}, path = ':memory:', id = 'keep-false-test', keep = FALSE)
  r2 = cache_exec({runif(1)}, path = ':memory:', id = 'keep-false-test', keep = FALSE)
  (r1 != r2)
})

assert('cache_rds() with hash = "auto" detects global variables', {
  d = tempfile(); dir.create(d)
  x = 100
  r1 = cache_rds(x * 2, file = 'auto-hash.rds', dir = paste0(d, '/'), hash = 'auto')
  (r1 %==% 200)
  r2 = cache_rds(x * 2, file = 'auto-hash.rds', dir = paste0(d, '/'), hash = 'auto')
  (r2 %==% 200)
  unlink(d, recursive = TRUE)
})

assert('cache_rds() with hash as list invalidates on change', {
  d = tempfile(); dir.create(d)
  r1 = cache_rds(1 + 1, file = 'hash-list.rds', dir = paste0(d, '/'), hash = list(a = 1))
  (r1 %==% 2)
  r2 = cache_rds(1 + 1, file = 'hash-list.rds', dir = paste0(d, '/'), hash = list(a = 2))
  (r2 %==% 2)
  # clean = TRUE removes old cache, so only one file exists
  (length(list.files(d, pattern = '[.]rds$')) %==% 1L)
  unlink(d, recursive = TRUE)
})

assert('cache_rds() errors on unsupported file extension', {
  d = tempfile(); dir.create(d)
  (has_error(cache_rds(1, file = 'test.xyz', dir = paste0(d, '/'))))
  unlink(d, recursive = TRUE)
})

assert('cache_dir() returns a character string', {
  d = cache_dir()
  (is.character(d))
  (nchar(d) > 0)
})

assert('find_locals() finds assigned variables', {
  (find_locals('x = 1; y = 2') %==% c('x', 'y'))
  (find_locals('for (i in 1:10) {}') %==% 'i')
  (find_locals('x = 1 +') %==% character(0))
})

assert('find_globals() handles language input', {
  e = new.env(parent = emptyenv())
  e$z = 99
  (find_globals(quote(y <- z + 1), e) %==% 'z')
})

assert('md5() returns consistent checksums', {
  x1 = 1; x2 = 1:10; x3 = seq(1, 10)
  m = md5(x1, x2, x3)
  (length(m) %==% 3L)
  (is.character(m))
  (nchar(m) %==% c(32L, 32L, 32L))
  # x2 and x3 are identical sequences
  (m[2] %==% m[3])
  # named arguments
  mn = md5(a = x1, b = x2)
  (names(mn) %==% c('a', 'b'))
})

assert('md5_one() produces a 32-character hex string', {
  h = md5_one(42)
  (is.character(h) && nchar(h) %==% 32L)
  # same object same hash
  (md5_one(42) %==% md5_one(42))
  # different objects different hashes
  (md5_one(42) != md5_one(43))
})

assert('cache_exec() with vars argument lazy-loads locals on cache hit', {
  # first run: populates cache
  r1 = cache_exec({
    local_var = 123
    local_var
  }, path = ':memory:', id = 'vars-lazy-test', vars = 'local_var')
  (r1 %==% 123)
  # second run: hits cache and lazy-loads local_var into the calling env
  r2 = cache_exec({
    local_var = 123
    local_var
  }, path = ':memory:', id = 'vars-lazy-test', vars = 'local_var')
  (r2 %==% 123)
  # local_var should have been lazy-loaded into the current environment
  (local_var %==% 123)
})

assert('cache_code() cleans up hash dict when path is NULL', {
  # seed the dict with a fake hash for 'x', then call cache_code with path = NULL
  assign('x', 'fakehash', envir = .cache_dict)
  (exists('x', envir = .cache_dict, inherits = FALSE))
  cache_code(quote(x <- 1), parent.frame(), config = list(path = NULL, vars = 'x'))
  # the stale entry should have been removed
  (!exists('x', envir = .cache_dict, inherits = FALSE))
})

assert('find_globals() returns character(0) when no globals exist', {
  e = new.env(parent = emptyenv())
  (find_globals('y = 1 + 1', e) %==% character(0))
})

library(testit)

pkgs = c('KernSmooth', 'xfun', 'knitr')
pkg_bib(pkgs, tempfile(), tweak = FALSE)

assert('& is escaped in title when pkg_bib(tweak = TRUE)', {
  (length(grep(' & ', grep(
    '^\\s*title =', capture.output(pkg_bib(pkgs, tweak = TRUE)), value = TRUE
  ))) %==% 0L)
})

assert('pkg_bib() warns on missing packages', {
  has_warning(pkg_bib(c('xfun', '__nonexistent_pkg__'), tempfile()))
})

assert('make_unique() converts repeated values to unique suffixes', {
  # already unique → returned unchanged
  (make_unique(c('a', 'b', 'c')) %==% c('a', 'b', 'c'))
  # empty input
  (make_unique(character(0)) %==% character(0))
  # duplicate values get letter suffixes (a, b, c, ...)
  res = make_unique(c('2020', '2020', '2020'))
  (res %==% c('2020a', '2020b', '2020c'))
  # more than 26 duplicates: numeric suffix fallback
  res2 = make_unique(rep('x', 28))
  (res2[27] %==% 'x27')
  (res2[28] %==% 'x28')
})

#' Cache the execution of an expression in memory or on disk
#'
#' Caching is based on the assumption that if the input does not change, the
#' output will not change. After an expression is executed for the first time,
#' its result will be saved (either in memory or on disk). The next run will be
#' skipped and the previously saved result will be loaded directly if all
#' external inputs of the expression remain the same, otherwise the cache will
#' be invalidated and the expression will be re-executed.
#'
#' Arguments supported in `...` include:
#'
#' - `vars`: Names of local variables (which are created inside the expression).
#' By default, local variables are automatically detected from the expression
#' via [find_locals()]. Locally created variables are cached along with the
#' value of the expression.
#'
#' - `hash` and `extra`: R objects to be used to determine if cache should be
#' loaded or invalidated. If (the MD5 hash of) the objects is not changed, the
#' cache is loaded, otherwise the cache is invalidated and rebuilt. By default,
#' `hash` is a list of values of global variables in the expression (i.e.,
#' variables created outside the expression). Global variables are automatically
#' detected by [find_globals()]. You can provide a vector of names to override
#' the automatic detection if you want some specific global variables to affect
#' caching, or the automatic detection is not reliable. You can also provide
#' additional information via the `extra` argument. For example, if the
#' expression reads an external file `foo.csv`, and you want the cache to be
#' invalidated after the file is modified, you may use `extra =
#' file.mtime("foo.csv")`.
#'
#' - `keep`: By default, only one copy of the cache corresponding to an `id`
#' under `path` is kept, and all other copies for this `id` is automatically
#' purged. If `TRUE`, all copies of the cache are kept. If `FALSE`, all copies
#' are removed, which means the cache is *always* invalidated, and can be useful
#' to force re-executing the expression.
#'
#' - `rw`: A list of functions to read/write the cache files. The list is of the
#' form `list(name = 'xxx', load = function(file) {}, save = function(x, file)
#' {})`. By default, [readRDS()] and [saveRDS()] are used. This argument can
#' also take a character string to use some built-in read/write methods.
#' Currently available methods include `rds` (the default), `raw` (using
#' [serialize()] and [unserialize()]), `qs2` (using [qs2::qs_read()] and
#' [qs2::qs_save()]), and `qdata` (using [qs2::qd_read()] and [qs2::qd_save()]).
#' The `rds` and `raw` methods only use base R functions (the `rds` method
#' generates smaller files because it uses compression, but is often slower than
#' the `raw` method, which does not use compression). The `qs2` and `qdata`
#' methods require the \pkg{qs2} package, which can be much faster than base R
#' methods and also supports compression.
#' @param expr An R expression to be cached.
#' @param path The path to save the cache. The special value `":memory:"` means
#'   in-memory caching. If it is intended to be a directory path, please make
#'   sure to add a trailing slash.
#' @param id A stable and unique string identifier for the expression to be used
#'   to identify a unique copy of cache for the current expression from all
#'   cache files (or in-memory elements). If not provided, an MD5 digest of the
#'   [deparse]d expression will be used, which means if the expression does not
#'   change (changes in comments or white spaces do not matter), the `id` will
#'   remain the same. This may not be a good default is two identical
#'   expressions are cached under the same `path`, because they could overwrite
#'   each other's cache when one expression's cache is invalidated, which may or
#'   may not be what you want. If you do not want that to happen, you need to
#'   manually provide an `id`.
#' @param ... More arguments to control the behavior of caching (see
#'   \sQuote{Details}).
#' @return If the cache is found, the cached value of the expression will be
#'   loaded and returned (other local variables will also be lazy-loaded into
#'   the current environment as a side-effect). If cache does not exist, the
#'   expression is executed and its value is returned.
#' @references See <https://yihui.org/litedown/#sec:option-cache> for how it
#'   works and an application in \pkg{litedown}.
#' @export
#' @examples
#' # the first run takes about 1 second
#' y1 = xfun::cache_exec({
#'   x = rnorm(1e5)
#'   Sys.sleep(1)
#'   x
#' }, path = ':memory:', id = 'sim-norm')
#'
#' # the second run takes almost no time
#' y2 = xfun::cache_exec({
#'   # comments won't affect caching
#'   x = rnorm(1e5)
#'   Sys.sleep(1)
#'   x
#' }, path = ':memory:', id = 'sim-norm')
#'
#' # y1, y2, and x should be identical
#' stopifnot(identical(y1, y2), identical(y1, x))
cache_exec = function(expr, path = 'cache/', id = NULL, ...) {
  use_cache = FALSE
  ret = cache_code(
    substitute(expr), parent.frame(), use_cache <- TRUE,
    list(path = path, id = id, ...)
  )
  if (use_cache) ret else expr
}

# a dictionary containing names and hashes of global variables
.cache_dict = new.env(parent = emptyenv())

# an in-memory database containing results from previous run, values of local
# variables, and their hashes, of the form environment(*__id__* = list(results,
# values, hashes))
.cache_db = new.env(parent = emptyenv())

cache_code = function(
  code, envir = parent.frame(), found = NULL, config = list(
    path = NULL, vars = NULL, hash = NULL, extra = NULL, keep = NULL, id = NULL,
    rw = NULL
  )
) {
  dict = .cache_dict
  if (!is.character(path <- config$path)) {
    # when caching is not enabled, we should clean up hashes for local variables
    # to make sure the next cached chunk will get the up-to-date hash
    vars = intersect(config$vars %||% find_locals(code), ls_all(dict))
    rm_vars(vars, dict)
    return()
  }

  # functions to read/write cache files
  if (is.null(rw <- config$rw)) rw = 'rds'
  if (is.character(rw)) rw = io_methods[[rw]]
  if (!is.character(rw$name) || length(rw$name) != 1) stop(
    'The cache method must have a name (as a character string).'
  )

  hash = config$hash %||% find_globals(code, envir)
  # get the values of hash variables (unless hash is I(character()))
  if (is.character(hash) && !inherits(hash, 'AsIs')) {
    # if a variable exists in the hash dictionary, use its hash (32 chars) in
    # the dict instead of its actual value, because the value may be large and
    # slow to serialize or compute md5 checksum
    hash = lapply(hash, function(x) dict[[x]] %||% get(x, envir))
  }
  # "normalized" code that doesn't rely on comments or white spaces
  code2 = deparse(if (is.character(code)) parse_only(code) else code)
  hash = md5_one(c(hash, config$extra, list(code2)))

  mem_cache = path == ':memory:'  # in-memory or disk cache

  # by default, the ID is a checksum of the code, which may not be a good idea
  # (e.g., two identical code fragments could be executed one after another
  # but they should not share the same cache); for litedown, id is chunk label
  if (is.null(id <- config$id)) id = md5_one(code2)
  id = paste0(path, if (mem_cache) '__', id, if (mem_cache) '__')

  file_pattern = function(x) sprintf('^[0-9a-z]{32}[.]%s([.][0-9]+)?$', x)
  # try to retrieve cache from memory (the dictionary) or disk
  cached = if (mem_cache) {
    db = .cache_db
    id_names = function(x) {
      x = ls_all(db)
      x[startsWith(x, id) & nchar(x) == nchar(id) + 32]
    }
    hits = id_names(db)
    # clean up all cache for this id, i.e., remove :memory:__id__hash
    if (isFALSE(config$keep)) rm_vars(hits, db)
    id = paste0(id, hash)
    id %in% ls_all(db)
  } else {
    db = list.files(id, full.names = TRUE)
    hits = grepl(file_pattern('[[:alnum:]]+'), basename(db))
    if (isFALSE(config$keep)) file.remove(db[hits])
    id = file.path(id, paste0(hash, '.', rw$name))
    file_exists(id)
  }
  # return now if cache is found
  if (cached) {
    # I could implement lazy-loading here but I'm not sure if it's worth it
    ret = if (mem_cache) db[[id]] else rw$load(id)
    list2env(ret$hashes, dict)  # update the hash dictionary
    # lazy-load cached objects into envir
    .mapply(function(name, file) {
      delayedAssign(name, if (mem_cache) ret$values[[name]] else rw$load(file), assign.env = envir)
    }, list(names(ret$hashes), sprintf('%s.%d', id, seq_along(ret$hashes))), NULL)
    found
    return(ret$result)
  }

  # clean up other versions of cache before saving a new version
  if (!isTRUE(config$keep)) {
    if (mem_cache) rm_vars(hits, db) else {
      # clean cache for the current method only; don't touch other methods' cache
      hits = grepl(file_pattern(rw$name), basename(db))
      file.remove(db[hits])
    }
  }
  vars = config$vars %||% find_locals(code)
  # inject an on.exit() call to the parent function to save its returned value;
  # note that returnValue() requires R >= 3.2.0 released on 2015-04-16, which I
  # hope is a reasonable requirement (anyone using R from a decade ago?)
  exit_call(function() {
    # we can't tell if returnValue() is from a successful call of a function or
    # not, so we pass a default that is very unlikely to be the returned value
    # on success; if we get this value, it's very likely that errors occurred
    void_return = new.env()  # should be unlikely for two new.env() to be identical
    res = returnValue(void_return)
    if (identical(res, void_return)) return()
    # save the checksums of objects to be used for a future run
    v1 = ls_all(envir)
    v2 = setdiff(vars, v1)
    if (length(v2)) warning(
      'Variable(s) not found in the environment: ', paste(v2, collapse = ', ')
    )
    hashes = as.list(do.call(md5, vals <- mget(vars, envir)))
    ret = list(result = res, hashes = hashes)
    if (mem_cache) {
      ret$values = vals
      db[[id]] = ret
    } else {
      dir_create(dirname(id))
      rw$save(ret, id)  # result and variable hashes are saved to the main file
      # variable values are saved in separate files of the form HASH.i; the i-th
      # file is for the i-th variable
      .mapply(rw$save, list(vals, sprintf('%s.%d', id, seq_along(vals))), NULL)
    }
    list2env(hashes, dict)
  })
}

# functions to load/save cache files
io_methods = list(
  raw = list(
    name = 'raw',
    load = function(...) unserialize(read_bin(...)),
    save = function(x, file, ...) {
      s = serialize(x, NULL, xdr = FALSE, ...)
      writeBin(s, file)
    }
  ),
  rds = list(
    name = 'rds',
    load = function(...) readRDS(...),
    save = function(x, file, ...) saveRDS(x, file, ...)
  ),
  qs = list(
    name = 'qs',
    load = function(...) qs::qread(...),
    save = function(x, file, ...) qs::qsave(x, file, ...)
  ),
  qs2 = list(
    name = 'qs2',
    load = function(...) qs2::qs_read(...),
    save = function(x, file, ...) qs2::qs_save(x, file, ...)
  ),
  qdata = list(
    name = 'qdata',
    load = function(...) qs2::qd_read(...),
    save = function(x, file, ...) qs2::qd_save(x, file, ...)
  )
)

#' Cache the value of an R expression to an RDS file
#'
#' Save the value of an expression to a cache file. Next time the value is
#' loaded from the file if it exists. Please consider using [cache_exec()]
#' instead, which is more flexible and intelligent.
#'
#' Note that the `file` argument does not provide the full cache filename. The
#' actual name of the cache file is of the form \file{BASENAME_HASH.rds}, where
#' \file{BASENAME} is the base name provided via the \file{file} argument (e.g.,
#' if `file = 'foo.rds'`, `BASENAME` would be \file{foo}), and \file{HASH} is
#' the MD5 hash (also called the \sQuote{checksum}) calculated from the R code
#' provided to the `expr` argument and the value of the `hash` argument, which
#' means when the code or the `hash` argument changes, the \file{HASH} string
#' may also change, and the old cache will be invalidated (if it exists). If you
#' want to find the cache file, look for \file{.rds} files that contain 32
#' hexadecimal digits (consisting of 0-9 and a-z) at the end of the filename.
#'
#' The file can be other formats such as `.qs2`. See the `rw` argument of
#' [cache_exec()].
#'
#' The possible ways to invalidate the cache are: 1) change the code in `expr`
#' argument; 2) delete the cache file manually or automatically through the
#' argument `rerun = TRUE`; and 3) change the value of the `hash` argument. The
#' first two ways should be obvious. For the third way, it makes it possible to
#' automatically invalidate the cache based on changes in certain R objects. For
#' example, when you run `cache_rds({ x + y })`, you may want to invalidate the
#' cache to rerun `{ x + y }` when the value of `x` or `y` has been changed, and
#' you can tell `cache_rds()` to do so by `cache_rds({ x + y }, hash = list(x,
#' y))`. The value of the argument `hash` is expected to be a list, but it can
#' also take a special value, `"auto"`, which means `cache_rds(expr)` will try
#' to automatically figure out the global variables in `expr`, return a list of
#' their values, and use this list as the actual value of `hash`. This behavior
#' is most likely to be what you really want: if the code in `expr` uses an
#' external global variable, you may want to invalidate the cache if the value
#' of the global variable has changed. Here a \dQuote{global variable} means a
#' variable not created locally in `expr`, e.g., for `cache_rds({ x <- 1; x + y
#' })`, `x` is a local variable, and `y` is (most likely to be) a global
#' variable, so changes in `y` should invalidate the cache. However, you know
#' your own code the best. If you want to be completely sure when to invalidate
#' the cache, you can always provide a list of objects explicitly rather than
#' relying on `hash = "auto"`.
#'
#' By default (the argument `clean = TRUE`), old cache files will be
#' automatically cleaned up. Sometimes you may want to use `clean = FALSE` (set
#' the R global option `options(xfun.cache_rds.clean = FALSE)` if you want
#' `FALSE` to be the default). For example, you may not have decided which
#' version of code to use, and you can keep the cache of both versions with
#' `clean = FALSE`, so when you switch between the two versions of code, it will
#' still be fast to run the code.
#' @param expr An R expression.
#' @param rerun Whether to delete the RDS file, rerun the expression, and save
#'   the result again (i.e., invalidate the cache if it exists).
#' @param file The *base* (see Details) cache filename under the directory
#'   specified by the `dir` argument. If not specified and this function is
#'   called inside a code chunk of a \pkg{knitr} document (e.g., an R Markdown
#'   document), the default is the current chunk label plus the extension
#'   \file{.rds}. The file can also be other formats (see Details).
#' @param dir The path of the RDS file is partially determined by `paste0(dir,
#'   file)`. If not specified and the \pkg{knitr} package is available, the
#'   default value of `dir` is the \pkg{knitr} chunk option `cache.path` (so if
#'   you are compiling a \pkg{knitr} document, you do not need to provide this
#'   `dir` argument explicitly), otherwise the default is \file{cache/}. If you
#'   do not want to provide a `dir` but simply a valid path to the `file`
#'   argument, you may use `dir = ""`.
#' @param hash A `list` object that contributes to the MD5 hash of the cache
#'   filename (see Details). It can also take a special character value
#'   `"auto"`. Other types of objects are ignored.
#' @param clean Whether to clean up the old cache files automatically when
#'   `expr` has changed.
#' @param ... Other arguments to be passed to the saving method such as
#'   [saveRDS()].
#' @note Changes in the code in the `expr` argument do not necessarily always
#'   invalidate the cache, if the changed code is [`parse`]`d` to the same
#'   expression as the previous version of the code. For example, if you have
#'   run `cache_rds({Sys.sleep(5);1+1})` before, running `cache_rds({ Sys.sleep(
#'   5 ) ; 1 + 1 })` will use the cache, because the two expressions are
#'   essentially the same (they only differ in white spaces). Usually you can
#'   add/delete white spaces or comments to your code in `expr` without
#'   invalidating the cache. See the package vignette `vignette('xfun', package
#'   = 'xfun')` for more examples.
#'
#'   When this function is called in a code chunk of a \pkg{knitr} document, you
#'   may not want to provide the filename or directory of the cache file,
#'   because they have reasonable defaults.
#'
#'   Side-effects (such as plots or printed output) will not be cached. The
#'   cache only stores the last value of the expression in `expr`.
#'
#'   The hash is calculated by saving the hash object to a temporary file via
#'   `saveRDS()` and getting the MD5 digest of the file. Since the file
#'   generated by `saveRDS()` may be platform-dependent, the cache may not be
#'   platform-agnostic. In other words, the cache saved on one platform may be
#'   invalidated on another platform even if nothing in the hash object has
#'   changed. By comparison, [cache_exec()] doesn't have this problem.
#' @return If the cache file does not exist, run the expression and save the
#'   result to the file, otherwise read the cache file and return the value.
#' @seealso [cache_exec()], which is more flexible (e.g., it supports in-memory
#'   caching and different read/write methods for cache files).
#' @export
#' @keywords internal
#' @examples
#' f = tempfile()  # the cache file
#' compute = function(...) {
#'   res = xfun::cache_rds({
#'     Sys.sleep(1)
#'     1:10
#'   }, file = f, dir = '', ...)
#'   res
#' }
#' compute()  # takes one second
#' compute()  # returns 1:10 immediately
#' compute()  # fast again
#' compute(rerun = TRUE)  # one second to rerun
#' compute()
#' unlink(paste0(f, '_*.rds'))
cache_rds = function(
  expr = {}, rerun = FALSE, file = 'cache.rds', dir = 'cache/',
  hash = NULL, clean = getOption('xfun.cache_rds.clean', TRUE), ...
) {
  if (loadable('knitr')) {
    if (!is.null(lab <- knitr::opts_current$get('label'))) {
      if (missing(file)) file = '.rds'
      if (startsWith(file, '.')) file = paste0(lab, file)
    }
    if (missing(dir) && !is.null(d <- knitr::opts_current$get('cache.path')))
      dir = d
  }
  path = paste0(dir, file)
  if (file_ext(path) == '') path = paste0(path, '.rds')
  code = deparse(substitute(expr))
  md5  = md5_obj(code)
  if (identical(hash, 'auto')) hash = global_vars(code, parent.frame(2))
  if (is.list(hash)) md5 = md5_obj(c(md5, md5_obj(hash)))
  ext = file_ext(path)
  if (!ext %in% names(io_methods))
    stop('The file extension ', ext, ' is not supported.')
  path = paste0(sans_ext(path), '_', md5, '.', ext)
  if (rerun) unlink(path)
  if (clean) clean_cache(path, ext)
  m = io_methods[[ext]]
  if (file_exists(path)) m$load(path) else {
    obj = expr  # lazy evaluation
    dir_create(dirname(path))
    m$save(obj, path, ...)
    obj
  }
}

# an old hash function for cache_rds() only
md5_obj = function(x) {
  f = tempfile(); on.exit(unlink(f), add = TRUE)
  if (is.character(x)) writeLines(x, f) else saveRDS(x, f)
  unname(tools::md5sum(f))
}

#' Calculate the MD5 checksums of R objects
#'
#' [Serialize][serialize()] an object and calculate the checksum via
#' [tools::md5sum()]. If `tools::md5sum()` does not have the argument `bytes`,
#' the object will be first serialized to a temporary file, which will be
#' deleted after the checksum is calculated, otherwise the raw bytes of the
#' object will be passed to the `bytes` argument directly (which will be
#' faster than writing to a temporary file).
#' @param ... Any number of R objects.
#' @return A character vector of the checksums of objects passed to `md5()`. If
#'   the arguments are named, the results will also be named.
#' @export
#' @examples
#' x1 = 1; x2 = 1:10; x3 = seq(1, 10); x4 = iris; x5 = paste
#' (m = xfun::md5(x1, x2, x3, x4, x5))
#' stopifnot(m[2] == m[3])  # x2 and x3 should be identical
#'
#' xfun::md5(x1 = x1, x2 = x2)  # named arguments
md5 = function(...) {
  obj = list(...)
  res = unlist(lapply(obj, md5_one))
  names(res) = names(obj)
  res
}

md5_one = function(x) {
  # no need to write to a file if md5sum() has the 'bytes' arg (R > 4.4.1)
  m = tools::md5sum
  if ('bytes' %in% names(formals(m))) f = NULL else {
    f = tempfile(); on.exit(unlink(f), add = TRUE)
  }
  s = serialize(x, NULL, xdr = FALSE)
  s = tail(s, -14)  # the first 14 bytes contain version info, etc
  if (is.null(f)) m(bytes = s) else {
    writeBin(s, f); unname(m(f))
  }
}

# clean up old cache files (those with the same base names as the new cache
# file, e.g., if the new file is FOO_0123abc...z.rds, then FOO_9876def...x.rds
# should be deleted)
clean_cache = function(path, ext) {
  olds = list.files(dirname(path), sprintf('_[0-9a-f]{32}[.]%s$', ext), full.names = TRUE)
  olds = c(olds, path)  # `path` may not exist; make sure it is in target paths
  base = basename(olds)
  keep = basename(path) == base  # keep this file (will cache to this file)
  base = substr(base, 1, nchar(base) - 34 - nchar(ext))  # 33 = 1 (_) + 32 (md5 sum) + 1 (.)
  unlink(olds[(base == base[keep][1]) & !keep])
}

#' Find global/local variables in R code
#'
#' Use [codetools::findGlobals()] and [codetools::findLocalsList()] to find
#' global and local variables in a piece of code. Global variables are defined
#' outside the code, and local variables are created inside the code.
#' @param code Either a character vector of R source code, or an R expression.
#' @param envir The global environment in which global variables are to be
#'   found.
#' @return A character vector of the variable names. If the source code contains
#'   syntax errors, an empty character vector will be returned.
#' @note Due to the flexibility of creating and getting variables in R, these
#'   functions are not guaranteed to find all possible variables in the code
#'   (e.g., when the code is hidden behind `eval()`).
#' @export
#' @examples
#' x = 2
#' xfun::find_globals('y = x + 1')
#' xfun::find_globals("y = get('x') + 1")  # x is not recognized
#' xfun::find_globals('y = zzz + 1')  # zzz doesn't exist
#'
#' xfun::find_locals('y = x + 1')
#' xfun::find_locals("assign('y', x + 1)")  # it works
#' xfun::find_locals("assign('y', x + 1, new.env())")  # still smart
#' xfun::find_locals("eval(parse(text = 'y = x + 1'))")  # no way
find_globals = function(code, envir = parent.frame()) {
  fun = function() {}
  if (is.language(code)) {
    body(fun) = code
  } else {
    fun = eval(parse2(c('function(){', code, '}'), fun), baseenv())
  }
  obj = codetools::findGlobals(fun)
  intersect(obj, ls_all(envir, TRUE))
}

# ls() all objects in the envir and if recursive, all parent environments until
# globalenv() or emptyenv()
ls_all = function(envir, recursive = FALSE) {
  x = ls(envir, all.names = TRUE)
  if (!recursive) return(x)
  while (TRUE) {
    # in theory, we shouldn't stop at global env but should keep recursion, but
    # in practice, objects in the parent environments of globalenv() (often
    # package namespaces) are unlikely to change and won't affect cache
    if (identical(envir, emptyenv()) || identical(envir, .GlobalEnv)) break
    envir = parent.env(envir)
    x = c(x, ls(envir, all.names = TRUE))
  }
  unique(x)
}

#' @rdname find_globals
#' @export
find_locals = function(code) {
  code = if (is.language(code)) as.expression(code) else parse2(code)
  codetools::findLocalsList(code)
}

# if code cannot be parsed, return an empty expression
parse2 = function(code, fallback = expression()) {
  tryCatch(parse_only(code), error = function(e) fallback)
}

# return a list of values of global variables in code
global_vars = function(code, envir) {
  if (length(vars <- find_globals(code, envir)) > 0)
    mget(vars, envir, inherits = TRUE)
}

#' Download a file from a URL and cache it on disk
#'
#' This object provides methods to download files and cache them on disk.
#' @format A list of methods:
#'
#' - `$get(url, type, handler)` downloads a URL, caches it, and returns the file
#'   content according to the value of `type` (possible values: `"text"` means
#'   the text content; `"base64"` means the base64 encoded data; `"raw"` means
#'   the raw binary content; `"auto"` is the default and means the type is
#'   determined by the content type in the URL headers). Optionally a `handler`
#'   function can be applied to the content.
#' - `$list()` gives the list of cache files.
#' - `$summary()` gives a summary of existing cache files.
#' - `$remove(url, type)` removes a single cache file.
#' - `$purge()` deletes all cache files.
#' @export
#' @examplesIf interactive()
#' # the first time it may take a few seconds
#' x1 = xfun::download_cache$get('https://www.r-project.org/')
#' head(x1)
#'
#' # now you can get the cached content
#' x2 = xfun::download_cache$get('https://www.r-project.org/')
#' identical(x1, x2)  # TRUE
#'
#' # a binary file
#' x3 = xfun::download_cache$get('https://yihui.org/images/logo.png', 'raw')
#' length(x3)
#'
#' # show a summary
#' xfun::download_cache$summary()
#' # remove a specific cache file
#' xfun::download_cache$remove('https://yihui.org/images/logo.png', 'raw')
#' # remove all cache files
#' xfun::download_cache$purge()
download_cache = local({
  pre = 'url'  # file prefix
  c_file = function(url, type) {
    file.path(cache_dir(), sprintf('%s_%s_%s.rds', pre, type, md5_one(url)))
  }
  read = function(url, type) {
    if (length(f <- c_file(url, type)) && file.exists(f)) readRDS(f)
  }
  write = function(url, type, data) {
    if (length(f <- c_file(url, type))) {
      dir_create(dirname(f))
      saveRDS(data, f)
    }
  }
  list_cache = function() {
    d = cache_dir()
    list.files(d, sprintf('^%s_.+[.]rds$', pre), full.names = TRUE)
  }
  list(
    get = function(url, type = c('auto', 'text', 'base64', 'raw'), handler = NULL) {
      type = type[1]
      if (!is.null(x <- read(url, type))) return(x[[url]])
      if ((auto <- type == 'auto')) type = if (length(grep(
        '^content-type:\\s+(text/.+|[^;]+;\\s+charset=utf-8)\\s*$',
        curlGetHeaders(url), ignore.case = TRUE
      ))) 'text' else 'raw'
      dir_create(d <- tempfile())
      on.exit(unlink(d, recursive = TRUE), add = TRUE)
      x = in_dir(d, {
        o = url_filename(url)
        download_file(url, o)
        switch(
          type, text = read_utf8(o), base64 = base64_uri(o), raw = read_bin(o)
        )
      })
      if (is.function(handler)) x = handler(x)
      write(url, if (auto) 'auto' else type, setNames(list(x), url))
      x
    },
    summary = function() {
      f = list_cache()
      if (length(f) == 0) return(invisible())
      t = gsub('^url_([^_]+)_.+$', '\\1', basename(f))
      u = vapply(f, function(x) names(readRDS(x)), character(1))
      s = file.size(f)
      if (length(f) > 1) message('Total size: ', format_bytes(sum(s)))
      d = data.frame(url = u, type = t, size = s, size_h = format_bytes(s))
      rownames(d) = NULL
      unname(split(d, seq_len(nrow(d))))
    },
    list = list_cache,
    remove = function(url, type = 'auto') file.remove(c_file(url, type)),
    purge = function() {
      f = list_cache()
      s = file.size(f)
      i = file.remove(f)
      message(sprintf(
        "Purged %d cache file(s) from '%s' (%s)",
        sum(i), cache_dir(), format_bytes(sum(s[i]))
      ))
    }
  )
})

cache_dir = if (getRversion() >= '4.0.0') function() {
  getOption('xfun.cache.dir', tools::R_user_dir('xfun', 'cache'))
} else function() {
  getOption('xfun.cache.dir', file.path(tempdir(), 'xfun', 'cache'))
}

#' Use the Tinify API to compress PNG and JPEG images
#'
#' Compress PNG/JPEG images with \samp{api.tinify.com}, and download the
#' compressed images. These functions require R packages \pkg{curl} and
#' \pkg{jsonlite}. `tinify_dir()` is a wrapper function of `tinify()` to
#' compress images under a directory.
#'
#' You are recommended to set the API key in \file{.Rprofile} or
#' \file{.Renviron}. After that, the only required argument of this function is
#' `input`. If the original images can be overwritten by the compressed images,
#' you may either use `output = identity`, or set the value of the `history`
#' argument in \file{.Rprofile} or \file{.Renviron}.
#' @param input A vector of input paths of images.
#' @param output A vector of output paths or a function that takes `input` and
#'   returns a vector of output paths (e.g., `output = `[`identity`] means
#'   `output = input`). By default, if the `history` argument is not a provided,
#'   `output` is `input` with a suffix `-min` (e.g., when `input = 'foo.png'`,
#'   `output = 'foo-min.png'`), otherwise `output` is the same as `input`, which
#'   means the original image files will be overwritten.
#' @param quiet Whether to suppress detailed information about the compression,
#'   which is of the form \samp{input.png (10 Kb) ==> output.png (5 Kb, 50\%);
#'   compression count: 42}. The percentage after `output.png` stands for
#'   the compression ratio, and the compression count shows the number of
#'   compressions used for the current month.
#' @param force Whether to compress an image again when it appears to have been
#'   compressed before. This argument only makes sense when the `history`
#'   argument is provided.
#' @param key The Tinify API key. It can be set via either the global option
#'   `xfun.tinify.key` or the environment variable `R_XFUN_TINIFY_KEY` (see
#'   [env_option()]).
#' @param history Path to a history file to record the MD5 checksum of
#'   compressed images. If the checksum of an expected output image exists in
#'   this file and `force = FALSE`, the compression will be skipped. This can
#'   help you avoid unnecessary API calls.
#' @return The output file paths.
#' @references Tinify API: <https://tinypng.com/developers>.
#' @seealso The \pkg{tinieR} package (<https://github.com/jmablog/tinieR/>) is a
#'   more comprehensive implementation of the Tinify API, whereas
#'   `xfun::tinify()` has only implemented the feature of shrinking images.
#' @export
#' @examples
#' \dontrun{
#' f = xfun:::R_logo('jpg$')
#' xfun::tinify(f)  # remember to set the API key before trying this
#' }
tinify = function(
  input, output, quiet = FALSE, force = FALSE,
  key = env_option('xfun.tinify.key'),
  history = env_option('xfun.tinify.history')
) {
  if (!(is.character(key) && length(key) == 1 && key != '')) stop(
    "The value of the 'key' argument must be a single non-empty character string."
  )
  if (length(input) == 0) return(invisible(input))
  if (any(i <- !file_exists(input))) stop(
    'Input file(s) not found: ', paste(input[i], collapse = ', ')
  )
  if (missing(output)) {
    output = if (is.character(history)) input else {
      paste0(sans_ext(input), '-min.', file_ext(input))
    }
  } else if (is.function(output)) output = output(input)

  # avoid optimizing the input image if its md5 checksum exists in history
  save_history = function(file) {
    if (!is.character(history) || history == '') return()
    dir_create(dirname(history))
    cat(paste0(tools::md5sum(file), '\n'), file = history, append = TRUE)
  }
  test_history = function(file) {
    is.character(history) && all(file_exists(c(history, file))) &&
      (tools::md5sum(file) %in% readLines(history))
  }

  auth = paste('Authorization: Basic', base64_encode(charToRaw(paste0('api:', key))))

  mapply(input, output, FUN = function(i, o) {
    if (!force && test_history(o)) {
      if (!quiet) message(
        'The image ', o, ' has been compressed before. ',
        'To compress it again, call tinify() with force = TRUE.'
      )
      return()
    }
    if (grepl('[.]png$', i, ignore.case = TRUE))
      optipng(files = i, stdout = if (quiet) FALSE else '')
    res = curl::curl_upload(i, 'https://api.tinify.com/shrink', httpheader = auth, verbose = FALSE)
    cnt = curl::parse_headers_list(res$headers)[['compression-count']]
    res = jsonlite::fromJSON(rawToChar(res$content))
    if (!is.character(u <- res$output$url)) stop2(
      "Failed to shrink '", i, "'", sprintf(': %s (%s)', res$error, res$message)
    )
    if (!quiet) message(sprintf(
      '%s (%s) ==> %s (%s, %.01f%%); compression count: %s',
      i, format_bytes(res$input$size), o, format_bytes(res$output$size),
      res$output$ratio * 100, if (length(cnt)) cnt else NA
    ))
    # back up the original image and restore it if download failed
    if (i == o) {
      b = paste0(i, '~')
      file.rename(i, b)
      on.exit(if (file_exists(o)) file.remove(b) else file.rename(b, i), add = TRUE)
    }
    curl::curl_download(u, o)
    save_history(o)
  })

  invisible(output)
}

#' @param dir A directory under which all \file{.png}, \file{.jpeg}, and
#'   \file{.webp} files are to be compressed.
#' @param ... Arguments passed to [tinify()].
#' @rdname tinify
#' @export
tinify_dir = function(dir = '.', ...) {
  tinify(all_files('[.](png|jpe?g|webp)$', dir), ...)
}

#' Shrink images to a maximum width
#'
#' Use [magick::image_resize()] to shrink an
#' image if its width is larger than the value specified by the argument
#' `width`, and optionally call [tinify()] to compress it.
#' @param width The desired maximum width of images.
#' @param dir The directory of images.
#' @param files A vector of image file paths. By default, this is all
#'   \file{.png}, \file{.jpeg}, and \file{.webp} images under `dir`.
#' @param tinify Whether to compress images using [tinify()].
#' @export
#' @examplesIf xfun::loadable('magick')
#' f = xfun:::all_files('[.](png|jpe?g)$', R.home('doc'))
#' file.copy(f, tempdir())
#' f = file.path(tempdir(), basename(f))
#' magick::image_info(magick::image_read(f))  # some widths are larger than 300
#' xfun::shrink_images(300, files = f)
#' magick::image_info(magick::image_read(f))  # all widths <= 300 now
#' file.remove(f)
shrink_images = function(
  width = 800, dir = '.', files = all_files('[.](png|jpe?g|webp)$', dir),
  tinify = FALSE
) {
  for (f in files) {
    x = magick::image_read(f)
    if (magick::image_info(x)$width <= width) next
    x = magick::image_resize(x, sprintf('%dx', width))
    magick::image_write(x, f)
  }
  if (tinify) tinify(files, identity)
}

#' Upload an image to imgur.com
#'
#' This function uses the \pkg{curl} package or the system command `curl`
#' (whichever is available) to upload a image to <https://imgur.com>.
#'
#' One application is to upload local image files to Imgur when knitting a
#' document with \pkg{knitr}: you can set the `knitr::opts_knit$set(upload.fun =
#' xfun::upload_imgur`, so the output document does not need local image files
#' any more, and it is ready to be published online.
#' @param file Path to the image file to be uploaded.
#' @param key Client ID for Imgur. It can be set via either the global option
#'   `xfun.upload_imgur.key` or the environment variable
#'   `R_XFUN_UPLOAD_IMGUR_KEY` (see [env_option()]). If neither is set, this
#'   uses a client ID registered by Yihui Xie.
#' @param use_curl Whether to use the R package \pkg{curl} to upload the image.
#'   If `FALSE`, the system command `curl` will be used.
#' @param include_xml Whether to include the XML response in the returned value.
#' @return A character string of the link to the image. If `include_xml = TRUE`,
#'   this string carries an attribute named `XML`, which is the XML response
#'   from Imgur (it will be parsed by \pkg{xml2} if available). See Imgur API in
#'   the references.
#' @author Yihui Xie, adapted from the \pkg{imguR} package by Aaron Statham
#' @note Please register your own Imgur application to get your client ID. You
#'   can certainly use mine, but this ID is in the public domain and it might
#'   reach Imgur's rate limit if too many people are using it in the same time
#'   period or someone is trying to upload too many images at once.
#' @references A demo: <https://yihui.org/knitr/demo/upload/>
#' @export
#' @examples \dontrun{
#' f = tempfile(fileext = '.png')
#' png(f); plot(rnorm(100), main = R.version.string); dev.off()
#'
#' res = imgur_upload(f, include_xml = TRUE)
#' res  # link to original URL of the image
#' xfun::attr2(res, 'XML')  # all information
#' if (interactive()) browseURL(res)
#'
#' # to use your own key
#' options(xfun.upload_imgur.key = 'your imgur key')
#' }
upload_imgur = function(
  file, key = env_option('xfun.upload_imgur.key'), use_curl = loadable('curl'),
  include_xml = FALSE
) {
  if (is.null(key) || key == '' || key == '9f3460e67f308f6') {
    i = floor(as.numeric(Sys.time()) %% 3) + 1
    key = c("qyv(sq y 't)ws(", 'q (t"vxs%t "wr(', "xpyvq%vp'rwx)yq")[i]
    key = decrypt(key, '1987052436fedcba')
  }
  api = 'https://api.imgur.com/3/image.xml'
  hdr = paste('Authorization: Client-ID', key)
  if (use_curl) {
    h = curl::new_handle(httpheader = hdr)
    curl::handle_setform(h, image = curl::form_file(file))
    res = curl::curl_fetch_memory(api, h)$content
  } else {
    file = path.expand(file)
    res = system2(
      'curl', shQuote(c('-H', hdr, '-F', paste0('image=@', file), '-s', api)),
      stdout = TRUE
    )
    res = one_string(res)
    if (res == '') stop('Failed to upload ', file)
  }
  if (loadable('xml2')) {
    res = xml2::as_list(xml2::read_xml(res))
    link = res[[1]]$link[[1]]
  } else {
    if (is.raw(res)) res = rawToChar(res)
    link = grep_sub('.*<link>([^<]+)</link>.*', '\\1', res)
  }
  if (length(link) != 1) stop(
    'Failed to upload ', file, sprintf(' (reason: %s)', if (is.character(res)) {
      grep_sub('.*<error>([^<]+)</error>.*', '\\1', res)
    } else res[[c('data', 'error', 'message')]][[1]])
  )
  if (include_xml) structure(link, XML = res) else link
}

`%|%` = function(x, y) if (length(x)) x else y
if (getRversion() < '4.4.0') `%||%` = function(x, y) if (is.null(x)) y else x

stop2 = function(...) stop(..., call. = FALSE)

warning2 = function(...) warning(..., call. = FALSE)

#' Obtain an attribute of an object without partial matching
#'
#' An abbreviation of [`base::attr`]`(exact = TRUE)`.
#' @param ... Passed to [base::attr()] (without the `exact` argument).
#' @export
#' @examples
#' z = structure(list(a = 1), foo = 2)
#' base::attr(z, 'f')  # 2
#' xfun::attr2(z, 'f')  # NULL
#' xfun::attr2(z, 'foo')  # 2
attr2 = function(...) base::attr(..., exact = TRUE)

#' @rdname attr2
#' @export
#' @note `xfun::attr()` is deprecated. Please use `xfun::attr2()` instead.
attr = function(...) {
  .Deprecated('xfun::attr2()', old = 'xfun::attr()')
  attr2(...)
}

#' Set environment variables
#'
#' Set environment variables from a named character vector, and return the old
#' values of the variables, so they could be restored later.
#'
#' The motivation of this function is that [Sys.setenv()] does not
#' return the old values of the environment variables, so it is not
#' straightforward to restore the variables later.
#' @param vars A named character vector of the form `c(VARIABLE = VALUE)`.
#'   If any value is `NA`, this function will try to unset the variable.
#' @return Old values of the variables (if not set, `NA`).
#' @export
#' @examples
#' vars = xfun::set_envvar(c(FOO = '1234'))
#' Sys.getenv('FOO')
#' xfun::set_envvar(vars)
#' Sys.getenv('FOO')
set_envvar = function(vars) {
  if (is.null(nms <- names(vars)) || any(nms == '')) stop(
    "The 'vars' argument must take a named character vector."
  )
  vals = Sys.getenv(nms, NA, names = TRUE)
  i = is.na(vars)
  suppressWarnings(Sys.unsetenv(nms[i]))
  if (length(vars <- vars[!i])) do.call(Sys.setenv, as.list(vars))
  invisible(vals)
}

#' Retrieve a global option from both `options()` and environment variables
#'
#' If the option exists in [options()], use its value. If not, query the
#' environment variable with the name `R_NAME` where `NAME` is the capitalized
#' option name with dots substituted by underscores. For example, for an option
#' `xfun.foo`, first we try `getOption('xfun.foo')`; if it does not exist, we
#' check the environment variable `R_XFUN_FOO`.
#'
#' This provides two possible ways, whichever is more convenient, for users to
#' set an option. For example, global options can be set in the [.Rprofile]
#' file, and environment variables can be set in the [.Renviron] file.
#' @param name The option name.
#' @param default The default value if the option is not found in [options()] or
#'   environment variables.
#' @return The option value.
#' @export
#' @examples
#' xfun::env_option('xfun.test.option')  # NULL
#'
#' Sys.setenv(R_XFUN_TEST_OPTION = '1234')
#' xfun::env_option('xfun.test.option')  # 1234
#'
#' options(xfun.test.option = TRUE)
#' xfun::env_option('xfun.test.option')  # TRUE (from options())
#' options(xfun.test.option = NULL)  # reset the option
#' xfun::env_option('xfun.test.option')  # 1234 (from env var)
#'
#' Sys.unsetenv('R_XFUN_TEST_OPTION')
#' xfun::env_option('xfun.test.option')  # NULL again
#'
#' xfun::env_option('xfun.test.option', FALSE)  # use default
env_option = function(name, default = NULL) {
  if (name %in% names(.Options)) return(.Options[[name]])
  name = toupper(paste0('R_', gsub('[.]', '_', name)))
  envs = Sys.getenv()
  if (name %in% names(envs)) envs[[name]] else default
}

#' Call `on.exit()` in a parent function
#'
#' The function [on.exit()] is often used to perform tasks when the
#' current function exits. This `exit_call()` function allows calling a
#' function when a parent function exits (thinking of it as inserting an
#' `on.exit()` call into the parent function).
#' @param fun A function to be called when the parent function exits.
#' @param n The parent frame number. For `n = 1`, `exit_call(fun)` is
#'   the same as `on.exit(fun())`; `n = 2` means adding
#'   `on.exit(fun())` in the parent function; `n = 3` means the
#'   grandparent, etc.
#' @param ... Other arguments to be passed to `on.exit()`.
#' @references This function was inspired by Kevin Ushey:
#'   <https://yihui.org/en/2017/12/on-exit-parent/>
#' @export
#' @examples
#' f = function(x) {
#'   print(x)
#'   xfun::exit_call(function() print('The parent function is exiting!'))
#' }
#' g = function(y) {
#'   f(y)
#'   print('f() has been called!')
#' }
#' g('An argument of g()!')
exit_call = function(fun, n = 2, ...) {
  do.call(
    on.exit, list(substitute(fun(), list(fun = fun)), add = TRUE, ...),
    envir = parent.frame(n)
  )
}

#' Evaluate an expression under a specified working directory
#'
#' Change the working directory, evaluate the expression, and restore the
#' working directory.
#' @param dir Path to a directory.
#' @param expr An R expression.
#' @export
#' @examples
#' library(xfun)
#' in_dir(tempdir(), {print(getwd()); list.files()})
in_dir = function(dir, expr) {
  owd = setwd(dir); on.exit(setwd(owd))
  expr
}

#' Parse R code and do not keep the source
#'
#' An abbreviation of `parse(keep.source = FALSE)`.
#' @param code A character vector of the R source code.
#' @export
#' @return R [expression()]s.
#' @examples library(xfun)
#' parse_only('1+1'); parse_only(c('y~x', '1:5 # a comment'))
#' parse_only(character(0))
parse_only = function(code) {
  if (length(code) == 0) return(expression())
  parse(text = code, keep.source = FALSE)
}

#' Try to evaluate an expression silently
#'
#' An abbreviation of `try(silent = TRUE)`.
#' @param expr An R expression.
#' @export
#' @examples library(xfun)
#' z = try_silent(stop('Wrong!'))
#' inherits(z, 'try-error')
try_silent = function(expr) try(expr, silent = TRUE)

#' Try an expression and see if it throws an error
#'
#' Use [tryCatch()] to check if an expression throws an error.
#' @inheritParams try_silent
#' @return `TRUE` (error) or `FALSE` (success).
#' @export
#' @examples
#' xfun::try_error(stop('foo'))  # TRUE
#' xfun::try_error(1:10)  # FALSE
try_error = function(expr) {
  err = FALSE
  tryCatch(expr, error = function(e) err <<- TRUE)
  err
}

#' Retry calling a function for a number of times
#'
#' If the function returns an error, retry it for the specified number of
#' times, with a pause between attempts.
#'
#' One application of this function is to download a web resource. Since the
#' download might fail sometimes, you may want to retry it for a few more times.
#' @param fun A function.
#' @param ... Arguments to be passed to the function.
#' @param .times The number of times.
#' @param .pause The number of seconds to wait before the next attempt.
#' @export
#' @examplesIf interactive()
#' # read the GitHub releases info of the repo yihui/xfun
#' xfun::retry(xfun::github_releases, 'yihui/xfun')
retry = function(fun, ..., .times = 3, .pause = 5) {
  for (i in seq_len(.times)) {
    if (!inherits(res <- tryCatch(fun(...), error = identity), 'error'))
      return(res)
    Sys.sleep(.pause)
  }
  stop(res$message, call. = FALSE)
}

gsubi = function(...) gsub(..., ignore.case = TRUE)
gsubf = function(...) gsub(..., fixed = TRUE)

#' Turn the output of [str()] into a tree diagram
#'
#' The super useful function `str()` uses \verb{..} to indicate the level
#' of sub-elements of an object, which may be difficult to read. This function
#' uses vertical pipes to connect all sub-elements on the same level, so it is
#' clearer which elements belong to the same parent element in an object with a
#' nested structure (such as a nested list).
#' @param ... Arguments to be passed to [str()] (note that the
#'   `comp.str` is hardcoded inside this function, and it is the only
#'   argument that you cannot customize).
#' @return A character string as a [raw_string()].
#' @export
#' @examples fit = lsfit(1:9, 1:9)
#' str(fit)
#' xfun::tree(fit)
#'
#' fit = lm(dist ~ speed, data = cars)
#' str(fit)
#' xfun::tree(fit)
#'
#' # some trivial examples
#' xfun::tree(1:10)
#' xfun::tree(iris)
tree = function(...) {
  x = capture.output(str(..., comp.str = '$ '))
  r = '^([^$-]+[$-] )(.*)$'
  x1 = gsub(r, '\\1', x)
  x2 = gsub(r, '\\2', x)
  x1 = gsub('[.][.]', '  ', x1)
  x1 = gsub('[$] $', '|-', x1)
  x1 = connect_pipes(x1)
  x3 = paste(x1, x2, sep = '')
  i = !grepl(r, x)
  x3[i] = x[i]
  raw_string(x3)
}

# for a tree diagram, connect the pipes on the same level, e.g., change

# |- ..
#   |- ..
#
#   |- ..

# to

# |- ..
#   |- ..
#   |
#   |- ..

# this task is not complicated, but just boring nested for-loops
connect_pipes = function(x) {
  ns = nchar(x); n = max(ns); m = length(x)
  if (n < 2 || m < 3) return(x)
  A = matrix('', nrow = m, ncol = n)
  x = strsplit(x, '')
  for (i in seq_len(m)) {
    A[i, seq_len(ns[i])] = x[[i]]
  }
  k = NULL
  for (j in seq_len(n - 1)) {
    for (i in seq_len(m - 2)) {
      if (!all(A[i, j + 0:1] == c('|', '-'))) next
      for (l in (i + 1):m) {
        cells = A[l, j + 0:1]
        if (all(cells == ' ')) {
          if (l == m) {
            k = NULL; break
          } else k = c(k, l)
        } else if (all(cells == c('|', '-'))) {
          break
        } else {
          k = NULL; break
        }
      }
      if (length(k) > 0) A[k, j] = '|'
      k = NULL
    }
  }
  apply(A, 1, paste, collapse = '')
}

pkg_file = function(...) system.file(..., package = 'xfun', mustWork = TRUE)

#' Format numbers of bytes using a specified unit
#'
#' Call the S3 method `format.object_size()` to format numbers of bytes.
#' @param x A numeric vector (each element represents a number of bytes).
#' @param units,... Passed to [`format()`][format.object_size].
#' @return A character vector.
#' @export
#' @examples
#' xfun::format_bytes(c(1, 1024, 2000, 1e6, 2e8))
#' xfun::format_bytes(c(1, 1024, 2000, 1e6, 2e8), units = 'KB')
format_bytes = function(x, units = 'auto', ...) {
  vapply(x, function(b) {
    format(structure(b, class = 'object_size'), units = units, ...)
  }, character(1))
}

# get the function name of the parent call
func_name = function(which = 1) {
  x = sys.call(which)[[1]]
  deparse(x)[1]
}

# evaluate an expression with an error handler; originally this was for knitr to
# output error location but can also be useful for other applications
handle_error = function(
  expr, handler, label = '', fun = getOption('xfun.handle_error.loc_fun')
) {
  on.exit(if (!ok) {
    loc = if (is.function(fun)) trimws(fun(label)) else ''
    m = handler(loc)
    message(one_string(m))
  })
  ok = FALSE
  expr  # evaluate now
  ok = TRUE
  expr  # won't be evaluated again
}

# a shorthand for rm(list =, envir =)
rm_vars = function(x, envir, ...) rm(list = x, envir = envir, ...)

# functions extracted from knitr and to be reused in other packages like litedown

#' Parse comma-separated chunk options
#'
#' For \pkg{knitr} and R Markdown documents, code chunk options can be written
#' using the comma-separated syntax (e.g., `opt1=value1, opt2=value2`). This
#' function parses these options and returns a list. If an option is not named,
#' it will be treated as the chunk label.
#' @param x The chunk options as a string.
#' @return A list of chunk options.
#' @export
#' @examples
#' xfun::csv_options('foo, eval=TRUE, fig.width=5, echo=if (TRUE) FALSE')
csv_options = function(x) {
  x = one_string(x)
  res = handle_error(
    eval(parse_only(paste('alist(', quote_label(x), ')'))),
    function(loc) {
      if (loc != '') loc = paste(' at lines', loc)
      c(
        sprintf('Invalid syntax for chunk options%s:\n', loc), x,
        '\nPlease see documentation at https://yihui.org/knitr/options/.\n'
      )
    }
  )
  idx = which(names(res) == '')  # which option is not named?
  # remove empty options
  j = NULL
  for (i in idx) if (identical(res[[i]], alist(,)[[1]])) j = c(j, i)
  if (length(j)) res[j] = NULL
  idx = if (is.null(names(res)) && length(res) == 1L) 1L else which(names(res) == '')
  if ((n <- length(idx)) > 1L || (length(res) > 1L && is.null(names(res)))) stop(
    'Invalid chunk options: ', x,
    "\n\nAll options must be of the form 'tag=value' except for the chunk label."
  )
  if (is.null(res$label)) {
    if (n == 0L) res$label = '' else names(res)[idx] = 'label'
  }
  if (!is.character(res$label))
    res$label = gsub(' ', '', as.character(as.expression(res$label)))
  if (res$label == '') res$label = NULL
  res
}

# quote the chunk label if necessary
quote_label = function(x) {
  x = gsub('^\\s*,?', '', x)
  if (grepl('^\\s*[^\'"](,|\\s*$)', x)) {
    # <<a,b=1>>= ---> <<'a',b=1>>=
    x = gsub('^\\s*([^\'"])(,|\\s*$)', "'\\1'\\2", x)
  } else if (grepl('^\\s*[^\'"](,|[^=]*(,|\\s*$))', x)) {
    # <<abc,b=1>>= ---> <<'abc',b=1>>=
    x = gsub('^\\s*([^\'"][^=]*)(,|\\s*$)', "'\\1'\\2", x)
  }
  x
}

# comment characters for various languages
comment_chars = list(
  `#` = c('awk', 'bash', 'coffee', 'gawk', 'julia', 'octave', 'perl', 'powershell', 'python', 'r', 'ruby', 'sed', 'stan'),
  '//' = c('asy', 'cc', 'csharp', 'd3', 'dot', 'fsharp', 'go', 'groovy', 'java', 'js', 'node', 'ojs', 'Rcpp', 'sass', 'scss', 'scala'),
  `%%` = c('mermaid'),
  `%` = c('matlab', 'tikz'),
  `/* */` = c('c', 'css'),
  `* ;` = c('sas'),
  `--` = c('haskell', 'lua', 'mysql', 'psql', 'sql'),
  `!` = c('fortran', 'fortran95'),
  `*` = c('stata')
)
# reshape it using the language name as the index, i.e., from list(char = lang)
# to list(lang = char)
comment_chars = local({
  res = list(apl = '\u235D')
  for (i in names(comment_chars)) {
    chars = comment_chars[[i]]
    res = c(res, setNames(rep(list(strsplit(i, ' ')[[1]]), length(chars)), chars))
  }
  res[order(names(res))]
})

get_option_comment = function(engine) {
  char = comment_chars[[engine]] %||% '#'
  s1 = paste0(char[[1]], '| ')
  s2 = ifelse(length(char) > 1, char[[2]], '')
  list(start = s1, end = s2)
}

#' Divide chunk options from the code chunk body
#'
#' Chunk options can be written in special comments (e.g., after `#|` for R code
#' chunks) inside a code chunk. This function partitions these options from the
#' chunk body.
#' @param engine The name of the language engine (to determine the appropriate
#'   comment character).
#' @param code A character vector (lines of code).
#' @return A list with the following items:
#'
#' - `options`: The parsed options (if there are any) as a list.
#' - `src`: The part of the input that contains the options.
#' - `code`: The part of the input that contains the code.
#'
#' @note Chunk options must be written on _continuous_ lines (i.e., all lines
#'   must start with the special comment prefix such as `#|`) at the beginning
#'   of the chunk body.
#' @export
#' @examples
#' # parse yaml-like items
#' yaml_like = c("#| label: mine", "#| echo: true", "#| fig.width: 8", "#| foo: bar", "1 + 1")
#' writeLines(yaml_like)
#' xfun::divide_chunk("r", yaml_like)
#'
#' # parse CSV syntax
#' csv_like = c("#| mine, echo = TRUE, fig.width = 8, foo = 'bar'", "1 + 1")
#' writeLines(csv_like)
#' xfun::divide_chunk("r", csv_like)
divide_chunk = function(engine, code) {
  res = list(options = NULL, src = NULL, code = code)
  # mask out empty blocks
  if (length(code) == 0) return(res)

  opt_comment = get_option_comment(engine)
  s1 = opt_comment$start
  s2 = opt_comment$end

  # check for option comments
  i1 = startsWith(code, s1)
  i2 = endsWith(trimws(code, 'right'), s2)
  # if "commentChar| " is not found, try "#| " instead
  if (!i1[1] && !identical(s1, '#|')) {
    s1 = '#| '; s2 = ''
    i1 = startsWith(code, s1); i2 = TRUE
  }
  m = i1 & i2

  # has to have at least one matched line at the beginning
  if (!m[[1]]) return(res)

  # divide into yaml and code
  if (all(m)) {
    src = code
    code = NULL
  } else {
    src = head(code, which.min(m) - 1)
    code = tail(code, -length(src))
  }

  # trim right
  if (any(i2)) src = trimws(src, 'right')

  # extract meta from comments, then parse it
  meta = substr(src, nchar(s1) + 1, nchar(src) - nchar(s2))
  # see if the metadata looks like YAML or CSV
  if (grepl('^[^ :]+:($|\\s)', meta[1])) {
    meta = yaml_load(meta, envir = FALSE)
    if (!is.list(meta) || length(names(meta)) == 0) {
      warning('Invalid YAML option format in chunk: \n', one_string(meta), '\n')
      meta = list()
    }
  } else {
    meta = csv_options(meta)
  }

  # normalize field name 'id' to 'label' if provided
  meta$label = unlist(meta[c('label', 'id')])[[1]]
  meta$id = NULL

  # extract code
  if (length(code) > 0 && is_blank(code[[1]])) {
    code = code[-1]
    src = c(src, '')
  }

  list(options = meta, src = src, code = code)
}

#' Read YAML data
#'
#' If the \pkg{yaml} package is installed, use [yaml::yaml.load()] to read the
#' data. If not, use the simple parser [taml_load()] instead.
#' @param x A character vector of YAML data.
#' @param ...,handlers Arguments to be passed to [yaml::yaml.load()].
#' @param envir The environment in which R expressions in YAML are evaluated. To
#'   disable the evaluation, use `envir = FALSE`.
#' @param use_yaml Whether to use the \pkg{yaml} package.
#' @return An R object (typically a list).
#' @note R expressions in YAML will be returned as [expression]s when they are
#'   not evaluated. This is different with [yaml::yaml.load()], which returns
#'   character strings for expressions.
#' @export
#' @examples
#' yaml_load('a: 1')
#' yaml_load('a: 1', use_yaml = FALSE)
yaml_load = function(
  x, ..., handlers = NULL, envir = parent.frame(), use_yaml = loadable('yaml')
) {
  if (!use_yaml) taml_load(x) else handle_error(
    yaml::yaml.load(x, eval.expr = FALSE, handlers = yaml_handlers(handlers, envir), ...),
    function(loc) {
      s = geterrmessage()
      r = 'line (\\d+), column (\\d+)'
      m = regmatches(s, regexec(r, s, perl = TRUE))[[1]]
      if (length(m) < 3) return()
      m = as.integer(m[-1])  # c(row, col)
      if (loc != '') loc = paste(' at lines', loc)
      c(
        sprintf('Failed to parse YAML%s:', loc), '',
        append(x, paste0(strrep(' ', m[2]), '^~~~~~'), m[1]), ''
      )
    }
  )
}

#' A simple YAML reader and writer
#'
#' TAML is a tiny subset of YAML. See
#' <https://yihui.org/litedown/#sec:yaml-syntax> for its specifications.
#' @param x For `taml_load()`, a character vector of the YAML content. For
#'   `taml_save()`, a list to be converted to YAML.
#' @param path A file path to read from or write to.
#' @inheritParams yaml_load
#' @return `taml_load()` and `taml_file()` return a list; if `path = NULL`,
#'   `taml_save()` returns a character vector, otherwise the vector is written
#'   to the file specified by the `path`.
#' @export
#' @examples
#' (res = taml_load('a: 1'))
#' taml_save(res)
#'
#' (res = taml_load('a: 1\nb: "foo"\nc: null'))
#' taml_save(res)
#'
#' (res = taml_load('a:\n  b: false\n  c: true\n  d: 1.234\ne: bar'))
#' taml_save(res)
#' taml_save(res, indent = '\t')
#'
#' taml_load('a: !expr paste(1:10, collapse = ", ")')
#' taml_load('a: [1, 3, 4, 2]')
#' taml_load('a: [1, "abc", 4, 2]')
#' taml_load('a: ["foo", "bar"]')
#' taml_load('a: [true, false, true]')
#' # the other form of array is not supported
#' taml_load('a:\n  - b\n  - c')
#' # and you must use the yaml package
#' if (loadable('yaml')) yaml_load('a:\n  - b\n  - c')
taml_load = function(x, envir = parent.frame()) {
  res = list()
  r = '^(\\s*)(.+?):($|\\s+.*)'
  x = split_lines(x)
  x = x[grep(r, x)]
  x = x[grep('^\\s*#', x, invert = TRUE)]  # comments
  if (length(x) == 0) return(res)
  lvl = gsub(r, '\\1', x)  # indentation level
  lvl = indent_level(lvl)
  key = gsub(r, '\\2', x)
  val = gsub('^\\s*|\\s*$', '', gsub(r, '\\3', x))
  keys = NULL
  for (i in seq_along(x)) {
    keys = c(head(keys, lvl[i]), key[i])
    v = if (is_blank(val[i])) list() else yaml_value(val[i], envir)
    # special treatment of NULL (to preserve a key with a null value)
    if (is.null(v)) {
      if (length(keys) <= 1) res[keys] = list(v) else {
        res[[head(keys, -1)]][tail(keys, 1)] = list(v)
      }
    } else res[[keys]] = v
  }
  res
}

#' @rdname taml_load
#' @export
taml_file = function(path) taml_load(read_utf8(path))

#' @param indent A character string to indent sub-lists by one level.
#' @rdname taml_load
#' @export
taml_save = function(x, path = NULL, indent = '  ') {
  res = .taml_save(x, indent)
  if (is.null(path)) raw_string(res, lang = '.yaml') else write_utf8(res, path)
}

.taml_save = function(x, indent = '  ', level = 1) {
  if (is.list(x)) {
    nms = names(x)
    if (is.null(nms)) {
      str(x); stop('Lists must be named')
    }
    val = lapply(x, function(z) {
      if (!is.list(z)) .taml_save(z) else {
        .indent = strrep(indent, level)
        one_string(c('', paste0(.indent, .taml_save(z, indent, level + 1))))
      }
    })
    paste(nms, val, sep = ': ')
  } else {
    asis = inherits(x, 'AsIs')
    if (is.null(x)) return('null')
    if (is.logical(x)) x = tolower(x) else if (is.expression(x)) {
      x = unlist(lapply(x, deparse))
      x = paste('!expr', paste(x, collapse = '\\n'))
    } else if (is.numeric(x) || is.character(x)) {
      if (!is.numeric(x) && length(x)) x = paste0('"', x, '"')
    } else {
      str(x); stop("Unsupported data type ('", class(x)[1], "')")
    }
    if (length(x) == 1 && !asis) x else paste0('[', paste(x, collapse = ', '), ']')
  }
}

indent_level = function(x) {
  N = nchar(x); n = N[N > 0]
  if (length(n) == 0) N else ceiling(N / min(n))
}

# only support logical, numeric, character values (both scalar and [] arrays),
# and R expressions starting with !r/!expr
yaml_value = function(x, envir = parent.frame()) {
  v = tolower(x)
  if (v == 'null') return()
  if (yaml_bool(v)) return(as.logical(x))
  if (grepl(r <- '^\\s*\\[(.*)\\]\\s*$', v)) {
    v = gsub(r, '\\1', x)
    if (is_blank(v)) return()
    v = unname(unlist(read.csv(text = v, header = FALSE)))
    if (is.numeric(v)) return(v)
    v = gsub('^ ', '', v)  # [a, b] -> ["a", " b"] -> ["a", "b"]
    return(if (yaml_bool(v)) as.logical(v) else yaml_unquote(v))
  }
  if (grepl('^[0-9.e+-]', v)) {
    v = suppressWarnings(as.numeric(v))
    if (!is.na(v)) return(if ((v2 <- as.integer(v)) == v) v2 else v)
  }
  if (grepl(r <- '^!(r|expr) (.+)$', x)) {
    yaml_expr(gsub(r, '\\2', x), envir)
  } else yaml_unquote(x)
}

# convert true/false/na to boolean
yaml_bool = function(x) all(grepl('^true|false|na$', x))
# remove optional quotes for strings
yaml_unquote = function(x) gsub('^["\']|["\']$', '', x)

yaml_expr = function(x, envir) {
  x = parse_only(x)
  if (is.environment(envir)) x = eval(x, envir)
  x
}

# add !r and !expr handlers to support a custom eval environment, which
# yaml::yaml.load() doesn't support by default (vubiostat/r-yaml#54)
yaml_handlers = function(h, envir) {
  h = as.list(h)
  f = function(x) yaml_expr(x, envir)
  for (i in c('r', 'expr')) if (is.null(h[[i]])) h[[i]] = f
  h
}

#' Partition the YAML metadata and the body in a document
#'
#' Split a document into the YAML metadata (which starts with `---` in the
#' beginning of the document) and the body.
#' @param x A character vector of the document content.
#' @param ... Arguments to be passed to `yaml_load()`.
#' @param parse Whether to parse the YAML data.
#' @export
#' @return A list of components `yaml` (the YAML data), `lines` (starting and
#'   ending line numbers of YAML), and `body` (a character vector of the body
#'   text). If YAML metadata does not exist in the document, the components
#'   `yaml` and `lines` will be missing.
#' @examples
#' xfun::yaml_body(c('---', 'title: Hello', 'output: litedown::html_format', '---', '', 'Content.'))
yaml_body = function(x, ..., parse = TRUE) {
  n = length(x)
  res = if (length(i <- locate_yaml(x)) == 0) {
    list(body = x)
  } else list(
    yaml = x[i[1]:i[2]], body = c(rep('', i[2]), tail(x, n - i[2])), lines = i
  )
  if ((n <- length(res$yaml)) >= 2) {
    res$yaml = res$yaml[-c(1, n)]
    if (parse) res['yaml'] = list(yaml_load(res$yaml, ...))
  }
  res
}

# find lines of YAML
locate_yaml = function(x) {
  i = grep('^---\\s*$', x)
  if (length(i) > 1 && all(is_blank(x[seq_len(i[1] - 1)]))) i[1:2]
}

#' Run R code and record the results
#'
#' Run R code and capture various types of output, including text output, plots,
#' messages, warnings, and errors.
#' @param code A character vector of R source code.
#' @param dev A graphics device. It can be a function name, a function, or a
#'   character string that can be evaluated to a function to open a graphics
#'   device.
#' @param dev.path A base file path for plots. Actual plot filenames will be
#'   this base path plus incremental suffixes. For example, if `dev.path =
#'   "foo"`, the plot files will be `foo-1.png`, `foo-2.png`, and so on. If
#'   `dev.path` is not character (e.g., `FALSE`), plots will not be recorded.
#' @param dev.ext The file extension for plot files. By default, it will be
#'   inferred from the first argument of the device function if possible.
#' @param dev.args Extra arguments to be passed to the device. The default
#'   arguments are `list(units = 'in', onefile = FALSE, width = 7, height = 7,
#'   res = 96)`. If any of these arguments is not present in the device
#'   function, it will be dropped.
#' @param message,warning,error If `TRUE`, record and store messages / warnings
#'   / errors in the output. If `FALSE`, suppress them. If `NA`, do not process
#'   them (messages will be emitted to the console, and errors will halt the
#'   execution).
#' @param cache A list of options for caching. See the `path`, `id`, and `...`
#'   arguments of [cache_exec()].
#' @param print A (typically S3) function that takes the value of an expression
#'   in the code as input and returns output. The default is [record_print()].
#' @param print.args A list of arguments for the `print` function. By default,
#'   the whole list is not passed directly to the function, but only an element
#'   in the list with a name identical to the first class name of the returned
#'   value of the expression, e.g., `list(data.frame = list(digits = 3), matrix
#'   = list())`. This makes it possible to apply different print arguments to
#'   objects of different classes. If the whole list is intended to be passed to
#'   the print function directly, wrap the list in [I()].
#' @param verbose `2` means to always print the value of each expression in the
#'   code, no matter if the value is [invisible()] or not; `1` means to always
#'   print the value of the last expression; `0` means no special handling
#'   (i.e., print only when the value is visible).
#' @param envir An environment in which the code is evaluated.
#' @return `record()` returns a list of the class `xfun_record_results` that
#'   contains elements with these possible classes: `record_source` (source
#'   code), `record_output` (text output), `record_plot` (plot file paths),
#'   `record_message` (messages), `record_warning` (warnings), and
#'   `record_error` (errors, only when the argument `error = TRUE`).
#' @import grDevices
#' @export
#' @examples
#' code = c('# a warning test', '1:2 + 1:3', 'par(mar = c(4, 4, 1, .2))', 'barplot(5:1, col = 2:6, horiz = TRUE)', 'head(iris)', "sunflowerplot(iris[, 3:4], seg.col = 'purple')", "if (TRUE) {\n  message('Hello, xfun::record()!')\n}", '# throw an error', "1 + 'a'")
#' res = xfun::record(code, dev.args = list(width = 9, height = 6.75), error = TRUE)
#' xfun::tree(res)
#' format(res)
#' # find and clean up plot files
#' plots = Filter(function(x) inherits(x, 'record_plot'), res)
#' file.remove(unlist(plots))
record = function(
  code = NULL, dev = 'png', dev.path = 'xfun-record', dev.ext = dev_ext(dev),
  dev.args = list(), message = TRUE, warning = TRUE, error = NA, cache = list(),
  print = record_print, print.args = list(),
  verbose = getOption('xfun.record.verbose', 0), envir = parent.frame()
) {
  new_result = function(x = list()) structure(x, class = 'xfun_record_results')
  res = new_result()
  if (length(code) == 0) return(res)

  # use cached results if cache exists
  use_cache = FALSE
  ret = cache_code(code, envir, use_cache <- TRUE, cache)
  if (use_cache) return(ret)

  code = split_lines(code)

  add_result = function(x, type = NULL, pos = length(res), insert = TRUE) {
    # default type is output
    if (is.null(type) && !inherits(x, .record_classes)) type = 'output'
    # insert a whole element or append to an existing element in res
    if (!insert) x = c(res[[pos]], x)
    el = if (is.null(type)) x else new_record(x, type)
    N = length(res)
    if (insert) {
      if (N == pos) res[[N + 1]] <<- el else res <<- append(res, el, pos)
    } else {
      res[[pos]] <<- el
    }
  }

  # dev.off() will set the current device to next instead of previous device,
  # but we often want to restore to the previous instead
  dev_cur = dev.cur()
  dev_reset = function() if (dev_cur != 1 && dev_cur %in% dev.list()) dev.set(dev_cur)

  # look for all possible ${dev.path}-%d.${dev.ext} files created by the device
  get_plots = function() {
    files = sprintf(dev.path, seq_along(list.files(dirname(dev.path))))
    files[file_exists(files)]
  }
  # open a graphics device
  dev_num = if (is.character(dev.path)) {
    if (!is.function(dev)) dev = tryCatch(match.fun(dev), error = function(e) {
      eval(parse(text = dev), envir = envir)
    })
    # normalize \ to / and remove the leading ./
    if (is_windows()) dev.path = gsubf('\\', '/', dev.path)
    dev.path = sub('^[.]/+', '', dev.path)
    # add extension
    dev.path = with_ext(paste0(dev.path, if (!grepl('/$', dev.path)) '-', '%d'), dev.ext)
    dir_create(dirname(dev.path))
    # clean up existing plots before opening the device
    if (any(i <- !file.remove(old_plots <- get_plots()))) stop(
      'Failed to delete existing plot file(s): ',
      paste("'", old_plots[i], "'", collapse = ', ')
    )
    dev_open(dev, dev.path, dev.args)
  }
  dev_old = dev.list()  # keep track of current devices
  dev_off = function() {
    if (length(dev_num) && dev_num %in% dev.list()) dev.off(dev_num)
    dev_reset()
    dev_num <<- NULL  # prevent dev.off() again by accident
  }
  on.exit(dev_off(), add = TRUE)

  # check if new plots are generated
  handle_plot = local({
    # don't record plots if no device was opened
    if (length(dev_num) == 0) return(function(...) {})
    old_files = character(); old_times = NULL  # previous plot files and file mtimes
    old_plot = recordPlot()
    function(last = FALSE) {
      # if dev.list() has changed, no longer record graphics, except for the last plot
      if (!last) {
        if (!(length(dev_num) && identical(dev.list(), dev_old))) return()
        dev.set(dev_num)
      }
      files = get_plots()
      # on Windows, an empty plot file is created upon opening a device
      files = files[file.size(files) > 0]

      if (!last) {
        new_plot = recordPlot()
        if (!identical(old_plot, new_plot)) {
          # add a placeholder for new plots, which may not have been created by
          # the device until the next new plot is drawn
          add_result(character(), 'plot')
          old_plot <<- new_plot
        }
      }

      if (length(files) == 0) return()

      # old files that have been modified should be removed from existing records
      k = file.mtime(old_files) > old_times
      if (any(k)) for (i in seq_along(res)) {
        if (inherits(res[[i]], 'record_plot'))
          res[[i]] <<- new_record(setdiff(res[[i]], old_files[k]), 'plot')
      }
      old_files = old_files[!k]

      plots = setdiff(files, old_files)
      old_files <<- files
      old_times <<- file.mtime(files)
      if ((n <- length(plots)) == 0) return()

      # indices of plots in results
      i = which(vapply(res, inherits, logical(1), 'record_plot'))
      N = length(i)
      # the last plot should always be appended to the last plot block
      if (last) {
        # N = 0 means the code didn't produce any plots
        if (N == 0) file.remove(plots) else add_result(plots, 'plot', i[N], FALSE)
        return()
      }

      # for the newly generated plots, append the first one (which should be the
      # last plot of a previous code chunk, since it will be created only when
      # the next code chunk produces a new plot) to the previous plot block; the
      # rest are for the current block
      if (N > 1) {
        add_result(plots[1], 'plot', i[N - 1], FALSE)
        if (n > 1) add_result(plots[2:n], 'plot', i[N], FALSE)
      } else if (N == 1) {
        # if there exists only one plot block, add plots to that block
        add_result(plots, 'plot', i[N], FALSE)
      }
    }
  })

  handle = if (is.na(error)) identity else try_silent
  # split code into individual expressions
  codes = handle(split_source(code, merge_comments = TRUE, line_number = TRUE))
  # code may contain syntax errors
  if (is_error(codes)) {
    add_result(code, 'source'); add_result(attr(codes, 'condition')$message, 'error')
    return(new_result(res))
  }

  handle_message = function(type, add = TRUE) {
    mf = sub('^(.)', 'muffle\\U\\1', type, perl = TRUE)
    function(e) {
      if (is.na(add)) return()
      if (isTRUE(add)) add_result(e$message, type)
      if (type %in% c('message', 'warning')) invokeRestart(mf)
    }
  }
  handle_m = handle_message('message', message)
  handle_w = handle_message('warning', warning)
  handle_e = handle_message('error', error)

  # don't use withCallingHandlers() if message/warning/error are all NA
  handle_eval = function(expr) {
    handle(if (is.na(message) && is.na(warning) && is.na(error)) expr else {
      withCallingHandlers(
        expr, message = handle_m, warning = handle_w, error = handle_e
      )
    })
  }
  # a simplified version of capture.output()
  handle_output = function(expr) {
    out = NULL
    con = textConnection('out', 'w', local = TRUE)
    on.exit(close(con))
    sink(con); on.exit({ sink(); close(con) })
    expr  # lazy evaluation
    on.exit()  # if no error occurred, clear up previous on-exit calls
    sink()
    close(con)
    if (length(out)) add_result(out, 'output')
    expr
  }
  n = length(codes)
  for (i in seq_len(n)) {
    add_result(code <- codes[[i]], 'source')
    expr = parse_only(code)
    if (length(expr) == 0) next
    # evaluate the code and capture output
    out = handle_output(handle_eval(withVisible(eval(expr, envir))))
    # verbose = 1: always print the last value; verbose = 2: print all values
    if (verbose == 2 || (verbose == 1 && i == n)) {
      # make invisible values visible unless they are NULL (which is often not useful)
      if (!is_error(out) && !out$visible && !is.null(out$value)) out$visible = TRUE
    }
    # print value (via record_print()) if visible
    if (!is_error(out) && out$visible) {
      if (is.null(print)) print = record_print
      p_args = print.args
      # index print args by first class of out unless args are wrapped in I()
      if (!inherits(p_args, 'AsIs')) p_args = p_args[[class(out$value)[1]]]
      out = handle_eval(do.call(print, c(list(out$value), p_args)))
      if (length(out) && !is_error(out)) {
        if (is.list(out)) lapply(out, add_result) else add_result(out)
      }
    }
    handle_plot()
  }
  # shut off the device to write out the last plot if there exists one
  dev_off()
  handle_plot(TRUE)

  # remove empty blocks
  res = Filter(length, res)
  res = merge_record(res)

  new_result(res)
}

# merge neighbor elements of the same class
merge_record = function(x) {
  n = length(x)
  if (n <= 1) return(x)
  k = NULL
  for (i in 2:n) {
    r1 = x[[i - 1]]; c1 = class(r1); r2 = x[[i]]; c2 = class(r2)
    if (!identical(c1, c2)) next
    # concatenate messages to a single string (consider message(appendLF = FALSE))
    x[[i]] = if ('record_message' %in% c1) paste(c(r1, r2), collapse = '') else c(r1, r2)
    attributes(x[[i]]) = attributes(r1)
    k = c(k, i - 1)
  }
  if (length(k)) x = x[-k]
  x
}

#' Print methods for `record()`
#'
#' An S3 generic function to be called to print visible values in code when the
#' code is recorded by [record()]. It is similar to [knitr::knit_print()]. By
#' default, it captures the normal [print()] output and returns the result as a
#' character vector. The `knitr_kable` method is for printing [knitr::kable()]
#' output. Users and package authors can define other S3 methods to extend this
#' function.
#' @param x For `record_print()`, the value to be printed. For `new_record()`, a
#'   character vector to be included in the printed results.
#' @param ... Other arguments to be passed to `record_print()` methods.
#' @return A `record_print()` method should return a character vector or a list
#'   of character vectors. The original classes of the vector will be discarded,
#'   and the vector will be treated as console output by default (i.e.,
#'   `new_record(class = "output")`). If it should be another type of output,
#'   wrap the vector in [new_record()] and specify a class name.
#' @export
record_print = function(x, ...) {
  UseMethod('record_print')
}

#' @rdname record_print
#' @export
record_print.default = function(x, ...) {
  # the default print method is just print()/show()
  capture.output(if (isS4(x)) methods::show(x, ...) else print(x, ...))
}

#' @rdname record_print
#' @export
record_print.record_asis = function(x, ...) {
  x
}

#' @param class A class name. Possible values are `xfun:::.record_cls`.
#' @rdname record_print
#' @export
new_record = function(x, class) structure(x, class = paste0('record_', class))

# all possible classes for record() results at the moment
.record_cls = c('source', 'output', 'message', 'warning', 'error', 'plot', 'asis')
.record_classes = paste0('record_', .record_cls)

dev_open = function(dev, file, args) {
  m = names(formals(dev))
  a = list(units = 'in', onefile = FALSE, width = 8, height = 8, res = 84)
  for (i in names(a)) {
    if (i %in% m && is.null(args[[i]])) args[[i]] = a[[i]]
  }
  do.call(dev, c(list(file), args))
  dev.control('enable')
  dev.cur()
}

# infer the filename extension from a device's first argument
dev_ext = function(dev) {
  # the first argument could be a string or an expression
  if (!is.character(name <- formals(dev)[[1]])) name = gsub('"', '', deparse(name))
  file_ext(name)
}

is_error = function(x) inherits(x, 'try-error')

#' @param to The output format (text or html).
#' @param encode For HTML output, whether to base64 encode plots.
#' @param template For HTML output, whether to embed the formatted results in an
#'   HTML template. Alternatively, this argument can take a file path, i.e.,
#'   path to an HTML template that contains the variable `$body$`. If `TRUE`,
#'   the default template in this package will be used
#'   (`xfun:::pkg_file('resources', 'record.html')`).
#' @rdname record
#' @exportS3Method
#' @return The `format()` method returns a character vector of plain-text output
#'   or HTML code for displaying the results.
format.xfun_record_results = function(
    x, to = c('text', 'html'), encode = FALSE, template = FALSE, ...
) {
  if (to[1] == 'text') {
    res = unlist(lapply(x, function(z) {
      if (!inherits(z, c('record_source', 'record_asis'))) z = paste('#>', z)
      gsub('\n*$', '\n', one_string(z))
    }))
    return(raw_string(res))
  }
  res = unlist(lapply(x, function(z) {
    cls = sub('^record_', '', class(z))
    if (cls == 'asis') z else if (cls == 'plot') {
      sprintf(
        '<p class="%s"><img src="%s" alt="A plot recorded by xfun::record()" /></p>',
        cls, if (encode) vapply(z, base64_uri, '') else URLencode(z)
      )
    } else {
      paste0(
        sprintf(
          '<pre class="%s"><code>',
          if (cls == 'source') paste0('language-r" data-start="', attr(z, 'lines')[1]) else cls
        ),
        escape_html(one_string(z)), '</code></pre>'
      )
    }
  }))
  if (isTRUE(template)) template = pkg_file('resources', 'record.html')
  if (is.character(template)) {
    res = sub('$body$', one_string(res), read_utf8(template), fixed = TRUE)
    if (length(x) > 0) res = sub('$title$', make_title(x[[1]]), res, fixed = TRUE)
  }
  raw_string(res)
}

# generate a title from the first line of a character vector
make_title = function(x) {
  if (length(x) == 0) return('')
  x = gsub('^#+\\s*', '', x[1])  # remove possible comment chars
  x = gsub('\\s*[.]*$', '... | ', x)  # add ... to the end
  escape_html(x)
}

#' @param x An object returned by `record()`.
#' @param browse Whether to browse the results on an HTML page.
#' @param ... Currently ignored.
#' @rdname record
#' @exportS3Method
#' @return The `print()` method prints the results as plain text or HTML to the
#'   console or displays the HTML page.
print.xfun_record_results = function(
  x, browse = interactive(), to = if (browse) 'html' else 'text', template = TRUE, ...
) {
  res = format(x, to, encode = browse, template = template, ...)
  if (browse && to == 'html') {
    viewer = getOption('viewer', utils::browseURL)
    f = tempfile('record-', fileext = '.html')
    write_utf8(res, f)
    viewer(f)
  } else {
    cat(res, sep = '\n')
  }
  invisible(x)
}

#' Attach or load packages, and automatically install missing packages if
#' requested
#'
#' `pkg_attach()` is a vectorized version of [library()] over
#' the `package` argument to attach multiple packages in a single function
#' call. `pkg_load()` is a vectorized version of
#' [requireNamespace()] to load packages (without attaching them).
#' The functions `pkg_attach2()` and `pkg_load2()` are wrappers of
#' `pkg_attach(install = TRUE)` and `pkg_load(install = TRUE)`,
#' respectively. `loadable()` is an abbreviation of
#' `requireNamespace(quietly = TRUE)`. `pkg_available()` tests if a
#' package with a minimal version is available.
#'
#' These are convenience functions that aim to solve these common problems: (1)
#' We often need to attach or load multiple packages, and it is tedious to type
#' several `library()` calls; (2) We are likely to want to install the
#' packages when attaching/loading them but they have not been installed.
#' @param ... Package names (character vectors, and must always be quoted).
#' @param install Whether to automatically install packages that are not
#'   available using [install.packages()]. Besides `TRUE` and
#'   `FALSE`, the value of this argument can also be a function to install
#'   packages (`install = TRUE` is equivalent to `install =
#'   install.packages`), or a character string `"pak"` (equivalent to
#'   `install = pak::pkg_install`, which requires the \pkg{pak} package).
#'   You are recommended to set a CRAN mirror in the global option `repos`
#'   via [options()] if you want to automatically install packages.
#' @param message Whether to show the package startup messages (if any startup
#'   messages are provided in a package).
#' @return `pkg_attach()` returns `NULL` invisibly. `pkg_load()`
#'   returns a logical vector, indicating whether the packages can be loaded.
#' @seealso `pkg_attach2()` is similar to `pacman::p_load()`, but does
#'   not allow non-standard evaluation (NSE) of the `...` argument, i.e.,
#'   you must pass a real character vector of package names to it, and all names
#'   must be quoted. Allowing NSE adds too much complexity with too little gain
#'   (the only gain is that it saves your effort in typing two quotes).
#' @import utils
#' @export
#' @examples library(xfun)
#' pkg_attach('stats', 'graphics')
#' # pkg_attach2('servr')  # automatically install servr if it is not installed
#'
#' (pkg_load('stats', 'graphics'))
pkg_attach = function(
  ..., install = FALSE, message = getOption('xfun.pkg_attach.message', TRUE)
) {
  if (!message) library = function(...) {
    suppressPackageStartupMessages(base::library(...))
  }
  for (i in c(...)) {
    if (!identical(install, FALSE) && !loadable(i)) pkg_install(i, install)
    library(i, character.only = TRUE)
  }
}

#' @param error Whether to signal an error when certain packages cannot be loaded.
#' @rdname pkg_attach
#' @export
pkg_load = function(..., error = TRUE, install = FALSE) {
  n = length(pkg <- c(...)); res = logical(n)
  if (n == 0) return(invisible(res))
  for (i in seq_len(n)) {
    res[i] = loadable(p <- pkg[i])
    if (!identical(install, FALSE) && !res[i]) {
      pkg_install(p, install); res[i] = loadable(p)
    }
  }
  if (error && any(!res)) stop('Package(s) not loadable: ', paste(pkg[!res], collapse = ' '))
  invisible(res)
}

#' @param pkg A single package name.
#' @param strict If `TRUE`, use [requireNamespace()] to test if
#'   a package is loadable; otherwise only check if the package is in
#'   [`.packages`]`(TRUE)` (this does not really load the package, so it
#'   is less rigorous but on the other hand, it can keep the current R session
#'   clean).
#' @param new_session Whether to test if a package is loadable in a new R
#'   session. Note that `new_session = TRUE` implies `strict = TRUE`.
#' @rdname pkg_attach
#' @export
loadable = function(pkg, strict = TRUE, new_session = FALSE) {
  if (length(pkg) != 1L) stop("'pkg' must be a character vector of length one")
  if (new_session) {
    Rscript(c('-e', shQuote(sprintf('library("%s")', pkg))), stdout = FALSE, stderr = FALSE) == 0
  } else {
    if (strict) {
      suppressPackageStartupMessages(requireNamespace(pkg, quietly = TRUE))
    } else pkg %in% .packages(TRUE)
  }
}

#' @param version A minimal version number. If `NULL`, only test if a
#'   package is available and do not check its version.
#' @rdname pkg_attach
#' @export
pkg_available = function(pkg, version = NULL) {
  loadable(pkg) && (is.null(version) || packageVersion(pkg) >= version)
}

#' @rdname pkg_attach
#' @export
pkg_attach2 = function(...) pkg_attach(..., install = TRUE)

#' @rdname pkg_attach
#' @export
pkg_load2 = function(...) pkg_load(..., install = TRUE)

pkg_require = function(pkgs, which = length(sys.calls()) - 1) {
  f = func_name(which)
  for (p in pkgs) if (!loadable(p)) stop2(
    "The '", p, "' package is required by the function '", f, "' but not available.",
    if (is_R_CMD_check()) c(
      " If you are developing an R package, you need to declare the dependency on '",
      p, "' in the DESCRIPTION file (e.g., in 'Imports')."
    )
  )
}

# update all packages in libraries that are writable by the current user
pkg_update = function(...) {
  libs = .libPaths()
  libs = libs[file.access(libs, 2) >= 0]
  for (l in libs) update.packages(l, ask = FALSE, checkBuilt = TRUE, ...)
}

# allow users to specify a custom install.packages() function via the global
# option xfun.install.packages
pkg_install = function(pkgs, install = TRUE, ...) {
  if (length(pkgs) == 0) return()
  # in case the CRAN repo is not set up
  repos = getOption('repos')
  if (length(repos) == 0 || identical(repos, c(CRAN = '@CRAN@'))) {
    opts = options(repos = c(CRAN = 'https://cran.rstudio.com'))
    on.exit(options(opts), add = TRUE)
  }
  if (length(pkgs) > 1)
    message('Installing ', length(pkgs), ' packages: ', paste(pkgs, collapse = ' '))
  if (isTRUE(install)) install = getOption(
    'xfun.install.packages',
    if (is.na(Sys.getenv('RENV_PROJECT', NA)) || !loadable('renv')) install.packages else {
      function(pkgs, lib = NULL, ...) renv::install(pkgs, library = lib, ...)
    }
  )
  if (identical(install, 'pak')) install = pak::pkg_install
  retry(install, pkgs, ..., .pause = 0)
}

#' Find out broken packages and reinstall them
#'
#' If a package is broken (i.e., not [loadable()]), reinstall it.
#'
#' Installed R packages could be broken for several reasons. One common reason
#' is that you have upgraded R to a newer `x.y` version, e.g., from `4.0.5` to
#' `4.1.0`, in which case you need to reinstall previously installed packages.
#' @param reinstall Whether to reinstall the broken packages, or only list their
#'   names.
#' @return A character vector of names of broken package.
#' @export
broken_packages = function(reinstall = TRUE) {
  libs = .libPaths()
  pkgs = unlist(lapply(libs, function(lib) {
    p = unlist(lapply(.packages(TRUE, lib), function(p) {
      if (!loadable(p, new_session = TRUE)) p
    }))
    if (length(p) && reinstall) {
      remove.packages(p, lib); pkg_install(p, lib = lib)
    }
    p
  }))
  if(reinstall) invisible(pkgs) else pkgs
}

# remove (binary) packages that were built with a previous major-minor version of R
check_built = function(dir = '.', dry_run = TRUE) {
  ext = if (is_macos()) 'tgz' else if (is_windows()) 'zip' else 'tar.gz'
  r =  paste0('_[-.0-9]+[.]', ext, '$')
  pkgs = list.files(dir, r, full.names = TRUE)
  meta = file.path(dir, 'PACKAGES')
  info = if (file_exists(meta)) read.dcf(meta)
  extract = if (grepl('gz$', ext)) untar else unzip
  for (f in pkgs) {
    d = file.path(gsub(r, '', basename(f)), 'DESCRIPTION')
    extract(f, d)
    if (is.na(b <- read.dcf(d, 'Built')[1, 1])) next
    unlink(dirname(d), recursive = TRUE)
    v = as.numeric_version(gsub('^\\s*R ([^;]+);.*', '\\1', b))
    if (major_minor_smaller(v, getRversion())) {
      message('The package ', f, ' was built with R ', v)
      if (!dry_run) file.remove(f)
    }
  }
  if (!is.null(info) && !dry_run) tools::write_PACKAGES(dir)
}

# is one version smaller than the other in major.minor? e.g., 4.1.0 is smaller
# than 4.2.0, but not smaller than 4.1.1
major_minor_smaller = function(v1, v2) {
  v1 = unclass(v1)[[1]]
  v2 = unclass(v2)[[1]]
  if (length(v1) < 3 || length(v2) < 3) return(TRUE)  # should return NA
  v1[1] < v2[1] || v1[2] < v2[2]
}

#' Install a source package from a directory
#'
#' Run \command{R CMD build} to build a tarball from a source directory, and run
#' \command{R CMD INSTALL} to install it.
#' @param pkg The package source directory.
#' @param build Whether to build a tarball from the source directory. If
#'   `FALSE`, run \command{R CMD INSTALL} on the directory directly (note
#'   that vignettes will not be automatically built).
#' @param build_opts The options for \command{R CMD build}.
#' @param install_opts The options for \command{R CMD INSTALL}.
#' @export
#' @return Invisible status from \command{R CMD INSTALL}.
install_dir = function(pkg, build = TRUE, build_opts = NULL, install_opts = NULL) {
  if (build) {
    pkg = pkg_build(pkg, build_opts)
    on.exit(unlink(pkg), add = TRUE)
  }
  res = Rcmd(c('INSTALL', install_opts, pkg))
  if (res != 0) stop('Failed to install the package ', pkg)
  invisible(res)
}

pkg_build = function(dir = '.', opts = NULL) {
  desc = file.path(dir, 'DESCRIPTION')
  pv = read.dcf(desc, fields = c('Package', 'Version'))
  # delete existing tarballs
  unlink(sprintf('%s_*.tar.gz', pv[1, 1]))
  Rcmd(c('build', opts, shQuote(dir)))
  pkg = sprintf('%s_%s.tar.gz', pv[1, 1], pv[1, 2])
  if (!file_exists(pkg)) stop('Failed to build the package ', pkg)
  pkg
}

# query the Homebrew dependencies of an R package
brew_dep = function(pkg) {
  u = sprintf('https://sysreqs.r-hub.io/pkg/%s/osx-x86_64-clang', pkg)
  x = retry(readLines, u, warn = FALSE)
  x = gsub('^\\s*\\[|\\]\\s*$', '', x)
  x = unlist(strsplit(gsub('"', '', x), '[, ]+'))
  x = setdiff(x, 'null')
  if (length(x))
    message('Package ', pkg, ' requires Homebrew packages: ', paste(x, collapse = ' '))
  x
}
brew_deps = function(pkgs) {
  if (length(pkgs) == 0) return()
  deps = pkg_brew_deps()
  unlist(lapply(pkgs, function(p) {
    if (is.null(deps[[p]])) brew_dep(p) else deps[[p]]
  }))
}

pkg_brew_deps = function() {
  con = url('https://macos.rbind.io/bin/macosx/sysreqsdb.rds')
  on.exit(close(con), add = TRUE)
  readRDS(con)
}

install_brew_deps = function(pkg = .packages(TRUE)) {
  inst = installed.packages()
  pkg = intersect(pkg, pkg_needs_compilation(inst))
  deps = pkg_brew_deps()
  deps = deps[c(pkg, pkg_dep(pkg, inst, recursive = TRUE))]
  deps = paste(na.omit(unique(unlist(deps))), collapse = ' ')
  if (deps != '') system(paste('brew install', deps))
}

pkg_needs_compilation = function(db = installed.packages()) {
  pkgs = unname(db[tolower(db[, 'NeedsCompilation']) == 'yes', 'Package'])
  pkgs[!is.na(pkgs)]
}

#' An alias of `remotes::install_github()`
#'
#' This alias is to make autocomplete faster via `xfun::install_github`, because
#' most `remotes::install_*` functions are never what I want. I only use
#' `install_github` and it is inconvenient to autocomplete it, e.g.
#' `install_git` always comes before `install_github`, but I never use it. In
#' RStudio, I only need to type `xfun::ig` to get `xfun::install_github`.
#' @param ... Arguments to be passed to [remotes::install_github()].
#' @export
install_github = function(...) remotes::install_github(...)

# Remove packages not installed from CRAN
reinstall_from_cran = function(dry_run = TRUE, skip_github = TRUE) {
  r = paste(c('Repository', if (skip_github) 'GithubRepo'), collapse = '|')
  r = paste0('^(', r, '): ')
  for (lib in .libPaths()) {
    pkgs = .packages(TRUE, lib)
    pkgs = setdiff(pkgs, c('xfun', 'rstudio', base_pkgs()))
    for (p in pkgs) {
      desc = read_utf8(system.file('DESCRIPTION', package = p, lib.loc = lib))
      if (!any(grepl(r, desc))) {
        if (dry_run) message(p, ': ', lib) else install.packages(p, lib = lib)
      }
    }
  }
}

#' Convert package news to the Markdown format
#'
#' Read the package news with [news()], convert the result to
#' Markdown, and write to an output file (e.g., \file{NEWS.md}). Each package
#' version appears in a first-level header, each category (e.g., \samp{NEW
#' FEATURES} or \samp{BUG FIXES}) is in a second-level header, and the news
#' items are written into bullet lists.
#' @param package,... Arguments to be passed to [news()].
#' @param output The output file path.
#' @param category Whether to keep the category names.
#' @return If `output = NA`, returns the Markdown content as a character
#'   vector, otherwise the content is written to the output file.
#' @export
#' @examplesIf interactive()
#' # news for the current version of R
#' xfun::news2md('R', Version == getRversion(), output = NA)
news2md = function(package, ..., output = 'NEWS.md', category = TRUE) {
  db = news(package = package, ...)
  k = db[, 'Category']
  db[is.na(k), 'Category'] = ''  # replace NA category with ''
  res = unlist(lapply(unique(db[, 'Version']), function(v) {
    d1 = db[db[, 'Version'] == v, ]
    res = unlist(lapply(unique(d1[, 'Category']), function(k) {
      txt = d1[d1[, 'Category'] == k, 'Text']
      txt = txt[txt != '']
      if (k == '' && length(txt) == 0) return()
      txt = gsub('\n *', ' ', txt)
      c(if (category && k != '') paste('##', k), if (length(txt)) paste('-', txt))
    }))
    if (is.na(dt <- d1[1, 'Date'])) dt = '' else dt = paste0(' (', dt, ')')
    c(sprintf('# CHANGES IN %s VERSION %s%s', package, v, dt), res)
  }))
  res = c(rbind(res, ''))  # add a blank line after each line
  if (is.na(output)) raw_string(res) else write_utf8(res, output)
}

#' Get base R package names
#'
#' Return names of packages from [installed.packages()] of which the
#' priority is `"base"`.
#' @return A character vector of base R package names.
#' @export
#' @examplesIf interactive()
#' xfun::base_pkgs()
base_pkgs = function() rownames(installed.packages(.Library, priority = 'base'))

# update one package (from source by default)
pkg_update_one = function(pkg, type = 'source') {
  opts = options(repos = c(CRAN = 'https://cran.r-project.org'))
  on.exit(options(opts), add = TRUE)
  if (is.null(pkgs <- old.packages(type = type)) || !pkg %in% rownames(pkgs)) return()
  install.packages(pkg, pkgs[pkg, 'LibPath'], type = type, INSTALL_opts = '--no-staged-install')
  NULL
}

#' Pseudo-random numbers on `[0, 1)` based on a linear congruential generator
#'
#' Generate pseudo-random numbers from \eqn{X_{i+1} = (aX_i + c) \bmod{m}}{X[i +
#' 1] = (a * X[i] + c) %% m}, where \eqn{X_1}{X[1]} is the initial value (seed).
#' @param n The desired length of the sequence.
#' @param a,c,m Parameters for the generator (see References for the default
#'   values).
#' @param seed The seed. The default is the system time when this function is
#'   called for the first time in the current session. For subsequent calls, the
#'   last \eqn{X_i}{X[i]} of the previous call will be used as the default seed.
#' @return Random numbers on `[0, 1)` (i.e., \eqn{X/m} instead of \eqn{X}). Note
#'   the unit interval is open on the right (excluding 1).
#' @note All argument values must be smaller than \eqn{2^{64}}{2^64} as they
#'   will be coerced to 64-bit integers.
#' @references Steele, Guy L. Jr.; Vigna, Sebastiano (2022). "Computationally
#'   easy, spectrally good multipliers for congruential pseudorandom number
#'   generators". _Software: Practice and Experience._ 52 (2): 443–458.
#' @export
#' @examples
#' rand_unit(10)
#' rand_unit(10, seed = 0)
#' rand_unit(10, seed = 0)  # identical results
#' rand_unit(10, seed = Sys.getpid())
rand_unit = local({
  .seed = NULL
  function(n = 1, a = 0xd9f5, c = 0, m = 4294967296, seed = NULL) {
    .seed <<- seed %||% .seed %||% as.numeric(Sys.time())
    x = .Call('rand_lcg', n, .seed, a, c, m, PACKAGE = 'xfun')
    if (n > 0) .seed <<- x[n]
    x / m
  }
})

# n random integers from 1 to M
rand_n = function(n, M) floor(rand_unit(n) * M) + 1

#' Read / write files encoded in UTF-8
#'
#' Read or write files, assuming they are encoded in UTF-8. `read_utf8()` is
#' roughly `readLines(encoding = 'UTF-8')` (a warning will be issued if non-UTF8
#' lines are found), and `write_utf8()` calls `writeLines(enc2utf8(text),
#' useBytes = TRUE)`.
#'
#' The function `append_utf8()` appends UTF-8 content to a file or connection
#' based on `read_utf8()` and `write_utf8()`, and optionally sort the content.
#' The function `append_unique()` appends unique lines to a file or connection.
#' @param con A connection or a file path.
#' @param error Whether to signal an error when non-UTF8 characters are detected
#'   (if `FALSE`, only a warning message is issued).
#' @param text A character vector (will be converted to UTF-8 via [enc2utf8()]).
#' @param ... Other arguments passed to [writeLines()] (except `useBytes`, which
#'   is `TRUE` in `write_utf8()`).
#' @return `read_utf8()` returns a character vector of the file content;
#'   `write_utf8()` returns the `con` argument (invisibly).
#' @export
read_utf8 = function(con, error = FALSE) {
  # users may have set options(encoding = 'UTF-8'), which usually won't help but
  # will bring more trouble than good, so we reset this option temporarily
  opts = options(encoding = 'native.enc'); on.exit(options(opts), add = TRUE)
  x = readLines(con, encoding = 'UTF-8', warn = FALSE)
  i = invalid_utf8(x)
  n = length(i)
  if (n > 0) (if (error) stop else warning)(
    if (is.character(con)) c('The file ', con, ' is not encoded in UTF-8. '),
    'These lines contain invalid UTF-8 characters: ',
    paste(c(head(i), if (n > 6) '...'), collapse = ', ')
  )
  x
}

#' @rdname read_utf8
#' @export
write_utf8 = function(text, con, ...) {
  if (is.null(text)) text = character(0)
  if (identical(con, '')) {
    cat(text, sep = '\n', file = con)
  } else {
    # prevent re-encoding the text in the file() connection in writeLines()
    # https://kevinushey.github.io/blog/2018/02/21/string-encoding-and-r/
    opts = options(encoding = 'native.enc'); on.exit(options(opts), add = TRUE)
    writeLines(enc2utf8(text), con, ..., useBytes = TRUE)
  }
  invisible(con)
}

#' @param sort Logical (`FALSE` means not to sort the content) or a
#'   function to sort the content; `TRUE` is equivalent to
#'   `base::sort`.
#' @rdname read_utf8
#' @export
append_utf8 = function(text, con, sort = TRUE) {
  x = read_utf8(con, error = TRUE)
  x = c(x, text)
  if (is.logical(sort)) sort = if (sort) base::sort else identity
  if (is.function(sort)) x = sort(x)
  write_utf8(x, con)
}

#' @rdname read_utf8
#' @export
append_unique = function(text, con, sort = function(x) base::sort(unique(x))) {
  append_utf8(text, con, sort)
}

# which lines are invalid UTF-8
invalid_utf8 = function(x) {
  which(!is_utf8(x))
}

test_utf8 = function(x) {
  is.na(x) | !is.na(iconv(x, 'UTF-8', 'UTF-8'))
}

# validUTF8() was added to base R 3.3.0
is_utf8 = function(x) {
  if ('validUTF8' %in% ls(baseenv())) validUTF8(x) else test_utf8(x)
}

#' Read a text file and concatenate the lines by `'\n'`
#'
#' The source code of this function should be self-explanatory.
#' @param file Path to a text file (should be encoded in UTF-8).
#' @return A character string of text lines concatenated by `'\n'`.
#' @export
#' @examples
#' xfun::file_string(system.file('DESCRIPTION', package = 'xfun'))
file_string = function(file) {
  x = read_utf8(file)
  # paste converts 0-length character() into 1-length ""
  if (length(x)) x = paste(x, collapse = '\n')
  raw_string(x)
}

#' Read all records of a binary file as a raw vector by default
#'
#' This is a wrapper function of [readBin()] with default arguments
#' `what = "raw"` and `n = `[`file.size`]`(file)`, which means it
#' will read the full content of a binary file as a raw vector by default.
#' @param file,what,n,... Arguments to be passed to `readBin()`.
#' @return A vector returned from `readBin()`.
#' @export
#' @examples
#' f = tempfile()
#' cat('abc', file = f)
#' xfun::read_bin(f)
#' unlink(f)
read_bin = function(file, what = 'raw', n = file.info(file)$size, ...) {
  readBin(file, what, n, ...)
}

#' Read all text files and concatenate their content
#'
#' Read files one by one, and optionally add text before/after the content. Then
#' combine all content into one character vector.
#' @param files A vector of file paths.
#' @param before,after A function that takes one file path as the input and
#'   returns values to be added before or after the content of the file.
#'   Alternatively, they can be constant values to be added.
#' @return A character vector.
#' @export
#' @examples
#' # two files in this package
#' fs = system.file('scripts', c('call-fun.R', 'child-pids.sh'), package = 'xfun')
#' xfun::read_all(fs)
#'
#' # add file paths before file content and an empty line after content
#' xfun::read_all(fs, before = function(f) paste('#-----', f, '-----'), after = '')
#'
#' # add constants
#' xfun::read_all(fs, before = '/*', after = c('*/', ''))
read_all = function(files, before = function(f) NULL, after = function(f) NULL) {
  b = before; a = after
  x = unlist(lapply(files, function(f) {
    c(if (is.function(b)) b(f) else b, read_utf8(f), if (is.function(a)) a(f) else a)
  }))
  raw_string(x)
}

#' Read a text file, process the text with a function, and write the text back
#'
#' Read a text file with the UTF-8 encoding, apply a function to the text, and
#' write back to the original file if the processed text is different with the
#' original input.
#'
#' `sort_file()` is an application of `process_file()`, with the processing
#' function being [sort()], i.e., it sorts the text lines in a file and write
#' back the sorted text.
#' @param file Path to a text file.
#' @param fun A function to process the text.
#' @param x The content of the file.
#' @param ... Arguments to be passed to `process_file()`.
#' @return If `file` is provided, invisible `NULL` (the file is updated as a
#'   side effect), otherwise the processed content (as a character vector).
#' @export
#' @examples f = tempfile()
#' xfun::write_utf8('Hello World', f)
#' xfun::process_file(f, function(x) gsub('World', 'woRld', x))
#' xfun::read_utf8(f)  # see if it has been updated
#' file.remove(f)
process_file = function(file, fun = identity, x = read_utf8(file)) {
  x2 = fun(x)
  if (missing(file)) x2 else {
    if ((length(x2) != length(x)) || !all(x2 == x)) write_utf8(x2, file)
  }
}

#' @rdname process_file
#' @export
sort_file = function(..., fun = sort) {
  process_file(fun = fun, ...)
}

#' Search and replace strings in files
#'
#' These functions provide the "file" version of [gsub()], i.e.,
#' they perform searching and replacement in files via `gsub()`.
#' @param file Path of a single file.
#' @param ... For `gsub_file()`, arguments passed to `gsub()`. For
#'   other functions, arguments passed to `gsub_file()`. Note that the
#'   argument `x` of `gsub()` is the content of the file.
#' @param rw_error Whether to signal an error if the file cannot be read or
#'   written. If `FALSE`, the file will be ignored (with a warning).
#' @param files A vector of file paths.
#' @param dir Path to a directory (all files under this directory will be
#'   replaced).
#' @param recursive Whether to find files recursively under a directory.
#' @param ext A vector of filename extensions (without the leading periods).
#' @param mimetype A regular expression to filter files based on their MIME
#'   types, e.g., `'^text/'` for plain text files. This requires the
#'   \pkg{mime} package.
#' @note These functions perform in-place replacement, i.e., the files will be
#'   overwritten. Make sure you backup your files in advance, or use version
#'   control!
#' @export
#' @examples library(xfun)
#' f = tempfile()
#' writeLines(c('hello', 'world'), f)
#' gsub_file(f, 'world', 'woRld', fixed = TRUE)
#' readLines(f)
gsub_file = function(file, ..., rw_error = TRUE) {
  if (!(file.access(file, 2) == 0 && file.access(file, 4) == 0)) {
    (if (rw_error) stop else warning)('Unable to read or write to ', file)
    if (!rw_error) return(invisible())
  }
  x1 = tryCatch(read_utf8(file, error = TRUE), error = function(e) if (rw_error) stop(e))
  if (is.null(x1)) return(invisible())
  x2 = gsub(x = x1, ...)
  if (!identical(x1, x2)) write_utf8(x2, file)
}

#' @rdname gsub_file
#' @export
gsub_files = function(files, ...) {
  for (f in files) gsub_file(f, ...)
}

#' @rdname gsub_file
#' @export
gsub_dir = function(..., dir = '.', recursive = TRUE, ext = NULL, mimetype = '.*') {
  files = list.files(dir, full.names = TRUE, recursive = recursive)
  if (length(ext)) files = files[file_ext(files) %in% ext]
  if (mimetype != '.*') files = files[grep(mimetype, mime::guess_type(files))]
  gsub_files(files, ...)
}

#' @rdname gsub_file
#' @export
gsub_ext = function(ext, ..., dir = '.', recursive = TRUE) {
  gsub_dir(..., dir = dir, recursive = recursive, ext = ext)
}

#' Perform replacement with `gsub()` on elements matched from `grep()`
#'
#' This function is a shorthand of `gsub(pattern, replacement,
#' grep(pattern, x, value = TRUE))`.
#' @param pattern,replacement,x,... Passed to [grep()] and
#'   `gsub()`.
#' @return A character vector.
#' @export
#' @examples # find elements that matches 'a[b]+c' and capitalize 'b' with perl regex
#' xfun::grep_sub('a([b]+)c', 'a\\U\\1c', c('abc', 'abbbc', 'addc', '123'), perl = TRUE)
grep_sub = function(pattern, replacement, x, ...) {
  x = grep(pattern, x, value = TRUE, ...)
  gsub(pattern, replacement, x, ...)
}

#' Try various methods to download a file
#'
#' Try all possible methods in [download.file()] (e.g.,
#' `libcurl`, `curl`, `wget`, and `wininet`) and see if any
#' method can succeed. The reason to enumerate all methods is that sometimes the
#' default method does not work, e.g.,
#' <https://stat.ethz.ch/pipermail/r-devel/2016-June/072852.html>.
#' @param url The URL of the file.
#' @param output Path to the output file. By default, it is determined by
#'   [url_filename()].
#' @param ... Other arguments to be passed to [download.file()]
#'   (except `method`).
#' @param .error An error message to signal when the download fails.
#' @note To allow downloading large files, the `timeout` option in
#'   [options()] will be temporarily set to one hour (3600 seconds)
#'   inside this function when this option has the default value of 60 seconds.
#'   If you want a different `timeout` value, you may set it via
#'   `options(timeout = N)`, where `N` is the number of seconds (not
#'   60).
#' @return The integer code `0` for success, or an error if none of the
#'   methods work.
#' @export
download_file = function(
  url, output = url_filename(url), ...,
  .error = 'No download method works (auto/wininet/wget/curl/lynx)'
) {
  if (getOption('timeout') == 60L) {
    opts = options(timeout = 3600)  # one hour
    on.exit(options(opts), add = TRUE)
  }
  download = function(method = 'auto') suppressWarnings({
    tryCatch(download.file(url, output, ..., method = method), error = function(e) 1L)
  })
  for (method in c(if (is_windows()) 'wininet', 'libcurl', 'auto')) {
    if (download(method = method) == 0) return(0L)
  }

  # check for libcurl/curl/wget/lynx, call download.file with appropriate method
  if (Sys.which('curl') != '') {
    # curl needs to add a -L option to follow redirects
    opts2 = if (is.null(getOption('download.file.extra')))
      options(download.file.extra = c('-L', '--fail'))
    res = download(method = 'curl')
    options(opts2)
    if (res == 0) return(res)
  }
  if (Sys.which('wget') != '') {
    if ((res <- download(method = 'wget')) == 0) return(res)
  }
  if (Sys.which('lynx') != '') {
    if ((res <- download(method = 'lynx')) == 0) return(res)
  }

  stop(.error)
}

#' Test if a URL is accessible
#'
#' Try to send a `HEAD` request to a URL using
#' [curlGetHeaders()] or the \pkg{curl} package, and see if it
#' returns a successful status code.
#' @param x A URL as a character string.
#' @param use_curl Whether to use the \pkg{curl} package or the
#'   `curlGetHeaders()` function in base R to send the request to the URL.
#'   By default, \pkg{curl} will be used when base R does not have the
#'   \command{libcurl} capability (which should be rare).
#' @param ... Arguments to be passed to `curlGetHeaders()`.
#' @return `TRUE` or `FALSE`.
#' @export
#' @examples xfun::url_accessible('https://yihui.org')
url_accessible = function(x, use_curl = !capabilities('libcurl'), ...) {
  try_status = function(code) tryCatch(code < 400, error = function(e) FALSE)
  if (use_curl) {
    h = curl::new_handle()
    curl::handle_setopt(h, customrequest = 'HEAD', nobody = TRUE)
    try_status(curl::curl_fetch_memory(x, h)$status_code)
  } else {
    # use curlGetHeaders() instead
    try_status(attr(curlGetHeaders(x, ...), 'status'))
  }
}

#' Generate a message with `cat()`
#'
#' This function is similar to [message()], and the difference is
#' that `msg_cat()` uses [cat()] to write out the message,
#' which is sent to [stdout()] instead of [stderr()]. The
#' message can be suppressed by [suppressMessages()].
#' @param ... Character strings of messages, which will be concatenated into one
#'   string via `paste(c(...), collapse = '')`.
#' @note By default, a newline will not be appended to the message. If you need
#'   a newline, you have to explicitly add it to the message (see
#'   \sQuote{Examples}).
#' @return Invisible `NULL`, with the side-effect of printing the message.
#' @seealso This function was inspired by `rlang::inform()`.
#' @export
#' @examples
#' {
#' # a message without a newline at the end
#' xfun::msg_cat('Hello world!')
#' # add a newline at the end
#' xfun::msg_cat(' This message appears right after the previous one.\n')
#' }
#' suppressMessages(xfun::msg_cat('Hello world!'))
msg_cat = function(...) {
  x = paste(c(...), collapse = '')
  withRestarts({
    signalCondition(simpleMessage(x))
    cat(x)
  }, muffleMessage = function() invisible(NULL))
}

#' Save objects to files and lazy-load them
#'
#' The function [lazy_save()] saves objects to files with incremental integer
#' names (e.g., the first object is saved to `1.rds`, and the second object is
#' saved to `2.rds`, etc.). The function [lazy_load()] lazy-load objects from
#' files saved via [lazy_save()], i.e., a file will not be read until the object
#' is used.
#' @param list A character vector of object names. This list will be written to
#'   an index file with `0` as the base name (e.g., `0.rds`).
#' @param path The path to write files to / read files from.
#' @param method The file save/load method. It can be a string (e.g., `rds`,
#'   `raw`, or `qs`) or a list. See the `rw` argument of [cache_exec()]. By
#'   default, it is automatically detected by checking the existence of the
#'   index file (e.g., `0.rds`, `0.raw`, or `0.qs`).
#' @param envir The environment to [get] or [assign] objects.
#' @return [lazy_save()] returns invisible `NULL`; [lazy_load()] returns the
#'   object names invisibly.
#' @seealso [delayedAssign()]
#' @export
lazy_save = function(list = NULL, path = './', method = 'auto', envir = parent.frame()) {
  m = io_method(method, path)
  idx = lazy_idx(path, m$name)
  m$save(list, idx)
  for (i in seq_along(list)) {
    v = get(list[[i]], envir, inherits = FALSE)
    m$save(v, sprintf('%s%d.%s', path, i, m$name))
  }
}

#' @rdname lazy_save
#' @export
lazy_load = function(path = './', method = 'auto', envir = parent.frame()) {
  m = io_method(method, path)
  idx = lazy_idx(path, m$name)
  vars = m$load(idx)
  .mapply(function(path, name) {
    delayedAssign(name, m$load(path), assign.env = envir)
  }, list(sprintf('%s%d.%s', path, seq_along(vars), m$name), vars), NULL)
  invisible(vars)
}

# use 0.ext as the index file to store the object names
lazy_idx = function(path, ext) paste0(path, '0.', ext)

io_method = function(method, path) {
  if (is.character(method)) {
    if (method == 'auto') {
      for (m in io_methods) if (file_exists(lazy_idx(path, m$name))) {
        method = m$name; break
      }
    }
    if (method == 'auto') method = 'rds'
  }
  if (is.list(method)) method else io_methods[[method]]
}

library(testit)

assert('do_once() executes the task exactly once', {
  opt = 'xfun.test.do_once.exec'
  options(setNames(list(NULL), opt))  # ensure clean state
  # use an env to share a counter with the nested expression
  e = new.env(parent = emptyenv()); e$count = 0L
  suppressMessages(do_once({ e$count = e$count + 1L }, opt))
  (e$count %==% 1L)
  # second call is skipped
  suppressMessages(do_once({ e$count = e$count + 1L }, opt))
  (e$count %==% 1L)
  options(setNames(list(NULL), opt))
})

assert('do_once() returns the task value invisibly', {
  opt = 'xfun.test.do_once.ret'
  options(setNames(list(NULL), opt))
  val = suppressMessages(do_once({
    v = 42L; v
  }, opt))
  (val %==% 42L)
  options(setNames(list(NULL), opt))
})

assert('do_once() emits a hint message and sets option to FALSE', {
  opt = 'xfun.test.do_once.hint'
  options(setNames(list(NULL), opt))
  msgs = character()
  withCallingHandlers(
    do_once(invisible(NULL), opt),
    message = function(m) {
      msgs <<- c(msgs, conditionMessage(m))
      invokeRestart('muffleMessage')
    }
  )
  # the hint message should mention the option name
  (any(grepl(opt, msgs, fixed = TRUE)))
  # option is now FALSE so subsequent calls are no-ops
  (identical(getOption(opt), FALSE))
  options(setNames(list(NULL), opt))
})

assert('do_once() is a no-op when option is already FALSE', {
  opt = 'xfun.test.do_once.noop'
  options(setNames(list(FALSE), opt))
  e = new.env(parent = emptyenv()); e$count = 0L
  res = do_once({ e$count = e$count + 1L; 99L }, opt)
  (e$count %==% 0L)
  (is.null(res))
  options(setNames(list(NULL), opt))
})

assert('do_once() with hint = "" emits no extra message', {
  opt = 'xfun.test.do_once.nohint'
  options(setNames(list(NULL), opt))
  msgs = character()
  withCallingHandlers(
    do_once(invisible(NULL), opt, hint = ''),
    message = function(m) {
      msgs <<- c(msgs, conditionMessage(m))
      invokeRestart('muffleMessage')
    }
  )
  (length(msgs) %==% 0L)
  options(setNames(list(NULL), opt))
})

# retrieve the release dates of packages
cran_pkg_dates = function(full = FALSE, maintainer = 'Yihui Xie') {
  info = tools::CRAN_package_db()
  pkgs = info[grep(maintainer, info$Maintainer), 'Package']
  info = setNames(vector('list', length(pkgs)), pkgs)
  for (p in pkgs) {
    message('Processing ', p)
    x = readLines(u <- sprintf('https://cran.rstudio.com/web/packages/%s/', p))
    i = which(x == '<td>Published:</td>')
    if (length(i) == 0) stop('Cannot find the publishing date from ', u)
    d = as.Date(gsub('</?td>', '', x[i[1] + 1]))
    x = try_silent(suppressWarnings(readLines(
      u <- sprintf('https://cran.r-project.org/src/contrib/Archive/%s/', p)
    )))
    if (inherits(x, 'try-error')) {
      info[[p]] = d; next
    }
    r = '.+</td><td align="right">(\\d{4,}-\\d{2}-\\d{2}) .+'
    d = c(d, as.Date(gsub(r, '\\1', grep(r, x, value = TRUE))))
    info[[p]] = sort(d, decreasing = TRUE)
  }
  if (full) info else sort(do.call(c, lapply(info, `[`, 1)), decreasing = TRUE)
}

# return packages that haven't been updated for X days, and can be updated on CRAN
cran_updatable = function(days = 90, maintainer = 'Yihui Xie') {
  info = cran_pkg_dates(TRUE, maintainer)
  flag = unlist(lapply(info, function(d) {
    sum(d > Sys.Date() - 180) < 6 && d[1] < Sys.Date() - days
  }))
  if (length(pkgs <- names(which(flag))) == 0) return(pkgs)
  # look into DESCRIPTION in GitHub repos and see if new version has been pushed
  info = tools::CRAN_package_db()
  info = info[info$Package %in% pkgs, , drop = FALSE]
  pkgs = info$Package
  for (i in seq_len(nrow(info))) {
    b = grep_sub('^(https://github.com/[^/]+/[^/]+)/issues$', '\\1', info$BugReports[i])
    if (length(b) != 1) next
    f = tempfile()
    u = paste0(b, '/raw/HEAD/DESCRIPTION')
    if (is.null(tryCatch(download.file(u, f, quiet = TRUE), error = function(e) NULL))) next
    d = read.dcf(f)
    file.remove(f)
    if (!'Version' %in% colnames(d)) next
    if (as.numeric_version(d[, 'Version']) <= paste0(info$Version[i], '.1')) {
      pkgs = setdiff(pkgs, info$Package[i])
      message('Skipped package ', info$Package[i], ' ', d[, 'Version'], ' (no new version).')
    } else {
      message('Package can be updated: ', b)
    }
  }
  pkgs
}


#' Some utility functions for checking packages
#'
#' Miscellaneous utility functions to obtain information about the package
#' checking environment.
#' @export
#' @keywords internal
is_R_CMD_check = function() {
  !is.na(check_package_name()) || tolower(Sys.getenv('_R_CHECK_LICENSE_')) == 'true'
}

#' @rdname is_R_CMD_check
#' @export
is_CRAN_incoming = function() {
  isTRUE(as.logical(Sys.getenv('_R_CHECK_CRAN_INCOMING_REMOTE_')))
}

#' @rdname is_R_CMD_check
#' @export
check_package_name = function() {
  Sys.getenv('_R_CHECK_PACKAGE_NAME_', NA)
}

# is R CMD check running on a package that has a version lower or equal to `version`?
#' @rdname is_R_CMD_check
#' @export
check_old_package = function(name, version) {
  if (is.na(pkg <- check_package_name()) || !(pkg %in% name)) return(FALSE)
  tryCatch(packageVersion(pkg) <= version[pkg == name], error = function(e) FALSE)
}

# return package maintainers (with email addresses)
pkg_maintainers = function(pkgs) {
  info = tools::CRAN_package_db()
  info = info[match(pkgs, info$Package), c('Package', 'Maintainer')]
  setNames(info$Maintainer, info$Package)
}

#' Submit a source package to CRAN
#'
#' Build a source package and submit it to CRAN with the \pkg{curl} package.
#' @param file The path to the source package tarball. By default, the current
#'   working directory is treated as the package root directory, and
#'   automatically built into a tarball, which is deleted after submission. This
#'   means you should run `xfun::submit_cran()` in the root directory of a
#'   package project, unless you want to pass a path explicitly to the
#'   `file` argument.
#' @param comment Submission comments for CRAN. By default, if a file
#'   \file{cran-comments.md} exists, its content will be read and used as the
#'   comment.
#' @seealso `devtools::submit_cran()` does the same job, with a few more
#'   dependencies in addition to \pkg{curl} (such as \pkg{cli});
#'   `xfun::submit_cran()` only depends on \pkg{curl}.
#' @export
submit_cran = function(file = pkg_build(), comment = '') {
  # if the tarball is automatically created, delete it after submission
  if (missing(file)) on.exit(file.remove(file), add = TRUE)

  # read the maintainer's name/email
  dir_create(d <- tempfile())
  on.exit(unlink(d, recursive = TRUE), add = TRUE)
  desc = file.path(gsub('_.*', '', basename(file)), 'DESCRIPTION')
  untar(file, desc, exdir = d)
  info = read.dcf(file.path(d, desc), fields = 'Maintainer')[1, 1]
  info = unlist(strsplit(info, '( <|>)'))

  # read submission comments from cran-comments.md if exists
  if (missing(comment) && file_exists(f <- 'cran-comments.md')) {
    comment = file_string(f)
  }
  params = list(
    uploaded_file = curl::form_file(file), name = info[1], email = info[2],
    upload = 'Upload package'
  )
  params$comment = if (length(comment)) comment
  server = 'https://xmpalantir.wu.ac.at/cransubmit/index2.php'

  # submit the form
  h = curl::new_handle()
  curl::handle_setform(h, .list = params)
  res = curl::curl_fetch_memory(server, h)

  # find the pkg_id from the response page
  id = grep_sub(
    '(.*<input name="pkg_id" type="hidden" value=")([^"]+)(".*)', '\\2',
    rawToChar(res$content)
  )
  if (length(id) != 1) stop('Failed to submit ', file, ' to CRAN')

  # skip the review and submit directly
  h = curl::new_handle()
  curl::handle_setform(h, .list = list(pkg_id = id, submit = 'Submit package'))
  res = curl::curl_fetch_memory(server, h)
  if (grepl('>Step 3<', rawToChar(res$content))) message(
    'The package has been submitted. Please confirm the submission in email: ',
    params$email
  ) else message('The submission may be unsuccessful.')
}

library(testit)

assert('session_info() strips BLAS/LAPACK/matrix-products and base-packages noise', {
  si = session_info()
  (!any(grepl('BLAS', si, fixed = TRUE)))
  (!any(grepl('LAPACK', si, fixed = TRUE)))
  (!any(grepl('Matrix products', si, fixed = TRUE)))
  (!any(grepl('attached base packages', si, ignore.case = TRUE)))
  # no consecutive blank lines
  blank = !nzchar(trimws(si))
  (!any(blank[-1] & blank[-length(blank)]))
})

assert('session_info() renames "other attached packages" and filters by package name', {
  si_all = session_info()
  (!any(grepl('other attached packages', si_all, ignore.case = TRUE)))
  (any(grepl('Package version:', si_all, fixed = TRUE)))
  # packages= restricts output to the named package
  si = session_info('xfun', dependencies = FALSE)
  ver = as.character(packageVersion('xfun'))
  (any(grepl(paste0('xfun_', ver), si, fixed = TRUE)))
  # empty strings in packages= are silently ignored
  si2 = session_info(c('xfun', ''), dependencies = FALSE)
  (identical(si, si2))
})

library(testit)

assert('file_ext() and sans_ext() work', {
  p = c('abc.doc', 'def123.tex#', 'path/to/foo.Rmd', 'backup.ppt~', 'pkg.tar.xz')
  (file_ext(p) %==% c('doc', 'tex#', 'Rmd', 'ppt~', 'tar.xz'))
  (sans_ext(p) %==% c('abc', 'def123', 'path/to/foo', 'backup', 'pkg'))
  (file_ext(c('foo.bar.gz', 'foo', 'file.nb.html')) %==% c('gz', '', 'nb.html'))
  # some special extensions
  p = c('abc.c++', 'def.c--', 'ghi.e##', 'jkl.FB2K-COMPONENT', 'mno.WITNESS_CAMPAIGN', 'pqr.H!')
  (file_ext(p, '-+!_#') %==% c('c++', 'c--', 'e##', 'FB2K-COMPONENT', 'WITNESS_CAMPAIGN', 'H!'))
  # by default, these extensions are not recognized
  (file_ext(p) %==% character(length(p)))
})

assert('with_ext() works for corner cases', {
  (with_ext(character(), 'abc') %==% character())
  (with_ext('abc', character()) %==% 'abc')
  (with_ext(NA_character_, 'abc') %==% NA_character_)
  (has_error(with_ext('abc', NA_character_)))
  (with_ext('abc', c('d', 'e')) %==% c('abc.d', 'abc.e'))
  (has_error(with_ext(c('a', 'b'), c('d', 'e', 'f'))))
  (with_ext(c('a', 'b'), c('d', 'e')) %==% c('a.d', 'b.e'))
  (with_ext(c('a', 'b'), c('d')) %==% c('a.d', 'b.d'))
  (with_ext(c('a', 'b', 'c'), c('', '.d', 'e.e')) %==% c('a', 'b.d', 'c.e.e'))
})

assert('same_path() works', {
  (is.na(same_path('~/foo', NA_character_)))
  (is.na(same_path(NA_character_, '~/foo')))
  (same_path('~/foo', file.path(Sys.getenv('HOME'), 'foo')))
  (!same_path(tempdir(), 'foo'))
})

assert('normalize_path() works', {
  f1 = tempfile()
  writeLines('test symlink', f1)
  f2 = paste0(f1, '~')
  res = file.symlink(f1, f2)  # this may fail (on Windows), i.e., res = FALSE
  # resolve symlink by default
  (!res || basename(normalize_path(f2)) %==% basename(f1))
  # do not resolve symlink
  (!res || basename(normalize_path(f2, resolve_symlink = FALSE)) %==% basename(f2))
  # resolve_symlink = FALSE should work with inputs like . and ..
  (normalize_path(c('.', '..'), resolve_symlink = FALSE) %==% normalize_path(c('.', '..')))
})

assert('url_filename() returns the file names in URLs', {
  (url_filename('https://yihui.org/images/logo.png') %==% 'logo.png')
  (url_filename(c(
    'https://yihui.org/index.html',
    'https://yihui.org/index.html?foo=bar',
    'https://yihui.org/index.html#about'
  )) %==% rep('index.html', 3))
})

assert('is_abs_path() recognizes absolute paths on Windows and *nix', {
  (!is_abs_path('abc/def'))
  (is_abs_path(if (.Platform$OS.type == 'windows') {
    c('D:\\abc', '\\\\netdrive\\somewhere')
  } else '/abc/def'))
})

assert('del_empty_dir() correctly deletes empty dirs', {
  # do nothing is NULL
  (del_empty_dir(NULL) %==% NULL)
  # remove if empty
  dir.create(temp_dir <- tempfile())
  del_empty_dir(temp_dir)
  (!dir_exists(temp_dir))
  # do not remove if not empty
  dir.create(temp_dir <- tempfile())
  writeLines('test', tempfile(tmpdir = temp_dir))
  (del_empty_dir(temp_dir) %==% NULL)
  (dir_exists(temp_dir))
  unlink(temp_dir, recursive = TRUE)
})

assert('mark_dirs add trailing / when necessary', {
  local({
    dir.create(tmp_dir <- tempfile())
    tmp_dir_slash = paste0(tmp_dir, '/')
    file.create(tmp_file <- tempfile(tmpdir = tmp_dir))
    (mark_dirs(c(tmp_dir, tmp_file)) %==% c(tmp_dir_slash, tmp_file))
    (mark_dirs(c(tmp_dir_slash, tmp_file)) %==% c(tmp_dir_slash, tmp_file))
    unlink(tmp_dir, recursive = TRUE)
  })
})

assert('relative_path() works', {
  (relative_path(c('foo/bar.txt', 'foo/baz.txt'), 'foo/') %==% c('bar.txt', 'baz.txt'))
  (relative_path('foo/bar.txt', 'foo') %==% 'bar.txt')
})

assert('proj_root() works', {
  # detect .Rproj root
  dir.create(tmp_dir <- tempfile())
  tmp_dir_slash <- paste0(tmp_dir, '/')
  file.create(f1 <- file.path(tmp_dir, 'test.Rproj'))
  writeLines(c('Version: 1.2.3', 'test: 321'), f1)

  (same_path(proj_root(tmp_dir), tmp_dir) %==% TRUE)
  unlink(f1)

  # detect package root
  file.create(f2 <- file.path(tmp_dir, 'DESCRIPTION'))
  writeLines(c('Package: abc', 'test: 321'), f2)
  dir.create(tmp_dir_child <- tempfile(tmpdir = tmp_dir))

  (same_path(proj_root(tmp_dir_child), tmp_dir) %==% TRUE)
  unlink(tmp_dir, recursive = TRUE)
})

assert('file_rename() works', {
  # work in temp dir
  dir.create(tmp_dir <- tempfile())
  owd = setwd(tmp_dir)

  # empty dir is not moved but deleted
  dir.create('dest')
  dir.create('empty')
  file_rename('empty', 'dest')
  (dir.exists(c('empty', 'dest')) %==% c(FALSE, TRUE))

  # files are moved correctly
  dir.create('filled')
  dummy_files = c('dummy1', 'dummy2')
  file.create(file.path('filled', dummy_files))
  file_rename('filled', 'dest1')
  (dir.exists('filled') %==% FALSE)
  (list.files('dest1') %==% dummy_files)

  # rename multiple dirs
  dir.create('dest2')
  file_rename(c('dest1', 'dest2'), c('dest3', 'dest4'))
  (dir.exists(sprintf('dest%d', 1:4)) %==% c(FALSE, FALSE, TRUE, TRUE))

  # rename files
  file_rename(file.path('dest3', dummy_files), dummy_files)
  (length(list.files('dest3'))  %==% 0L)
  (file.exists(dummy_files) %==% c(TRUE, TRUE))

  # remove temp dir
  setwd(owd)
  unlink(tmp_dir, recursive = TRUE)
})

assert('rename_seq() returns renamed file patterns', {
  dir.create(tmp_dir <- tempfile())
  owd = setwd(tmp_dir)
  # no matching files -> returns character(0)
  (rename_seq() %==% character(0))
  # create some files matching the pattern
  file.create(c('01-foo.Rmd', '02-bar.Rmd', '03-baz.Rmd'))
  res = rename_seq()
  (length(res) %==% 3L)
  (inherits(res, 'xfun_rename_seq'))
  # print method should work without error
  capture.output(print(res))
  # dry_run = FALSE actually renames
  file.create(c('01-x.Rmd', '02-y.Rmd'))
  rename_seq(pattern = '^[0-9]+-[xy][.]Rmd$', dry_run = FALSE)
  (all(file.exists(c('1-x.Rmd', '2-y.Rmd'))))
  setwd(owd)
  unlink(tmp_dir, recursive = TRUE)
})

assert('existing_files() with first = TRUE works', {
  f = tempfile(); writeLines('test', f)
  (existing_files(f, first = TRUE) %==% f)
  unlink(f)
  (has_error(existing_files(c('nonexistent1', 'nonexistent2'), first = TRUE)))
  (existing_files(c('nonexistent1', 'nonexistent2'), first = TRUE, error = FALSE) %==% character(0))
  (has_error(existing_files('nonexistent', first = TRUE, error = 'Custom error msg')))
})

assert('is_web_path() detects web paths', {
  (is_web_path(c('https://www.r-project.org', 'http://example.com',
    'ftp://files.example.com', 'ftps://secure.example.com')))
  (!is_web_path(c('www.r-project.org', '/local/path', 'relative/path')))
})

assert('is_rel_path() is the negation of is_abs_path()', {
  (is_rel_path('foo.txt'))
  (!is_rel_path(tempdir()))
})

assert('with_ext() removes extensions when ext is empty string', {
  (with_ext(c('a.txt', 'b.R'), '') %==% c('a', 'b'))
})

assert('relative_path() edge cases', {
  (has_error(relative_path('/tmp/some/file.txt', '/different/dir', use.. = FALSE)))
  d = normalize_path(tempdir())
  (relative_path(d, d) %==% '.')
})

assert('relative_path() with .. works across sibling directories', {
  d = tempfile(); dir.create(d)
  a = file.path(d, 'a'); dir.create(a)
  b = file.path(d, 'b'); dir.create(b)
  f = file.path(a, 'file.txt'); file.create(f)
  res = relative_path(f, b)
  (grepl('\\.\\.', res))
  unlink(d, recursive = TRUE)
})

assert('R_logo() finds an existing logo', {
  logo = R_logo()
  (length(logo) %==% 1L)
  (file.exists(logo))
})

assert('all_files() lists files recursively', {
  d = tempfile(); dir.create(d)
  sub = file.path(d, 'sub'); dir.create(sub)
  file.create(file.path(d, 'a.txt'))
  file.create(file.path(sub, 'b.R'))
  res = all_files('[.]R$', dir = d)
  (length(res) %==% 1L)
  (grepl('b.R$', res))
  unlink(d, recursive = TRUE)
})

assert('magic_path() finds a file recursively in subdirs', {
  d = tempfile(); dir.create(d)
  sub = file.path(d, 'a', 'b'); dir.create(sub, recursive = TRUE)
  f = file.path(sub, 'target.txt'); writeLines('found', f)
  res = magic_path('target.txt', root = d, relative = FALSE)
  (same_path(res, f))
  unlink(d, recursive = TRUE)
})

assert('magic_path() errors when file not found and error = TRUE', {
  d = tempfile(); dir.create(d)
  (has_error(magic_path('nonexistent.txt', root = d, error = TRUE)))
  unlink(d, recursive = TRUE)
})

assert('magic_path() returns input path when not found and error = FALSE', {
  d = tempfile(); dir.create(d)
  (magic_path('nonexistent.txt', root = d, error = FALSE) %==% 'nonexistent.txt')
  unlink(d, recursive = TRUE)
})

assert('magic_path() errors or warns on multiple matches', {
  d = tempfile(); dir.create(d)
  sub1 = file.path(d, 'a'); dir.create(sub1)
  sub2 = file.path(d, 'b'); dir.create(sub2)
  file.create(file.path(sub1, 'dup.txt'))
  file.create(file.path(sub2, 'dup.txt'))
  (has_error(magic_path('dup.txt', root = d, error = TRUE)))
  res = suppressMessages(magic_path('dup.txt', root = d, error = FALSE))
  (file.exists(res))
  unlink(d, recursive = TRUE)
})

assert('from_root() errors when root is NULL', {
  (has_error(from_root('foo.R', root = NULL)))
})

assert('from_root() returns relative path from project root', {
  d = tempfile(); dir.create(d)
  writeLines('Package: testpkg', file.path(d, 'DESCRIPTION'))
  sub = file.path(d, 'R'); dir.create(sub)
  file.create(file.path(sub, 'code.R'))
  owd = setwd(sub)
  res = from_root('R', 'code.R', root = d)
  (res %==% 'code.R')
  setwd(owd)
  unlink(d, recursive = TRUE)
})

#' Find the indices of lines in Markdown that are prose (not code blocks)
#'
#' Filter out the indices of lines between code block fences such as \verb{```}
#' (could be three or four or more backticks).
#' @param x A character vector of text in Markdown.
#' @param warn Whether to emit a warning when code fences are not balanced.
#' @note If the code fences are not balanced (e.g., a starting fence without an
#'   ending fence), this function will treat all lines as prose.
#' @return An integer vector of indices of lines that are prose in Markdown.
#' @export
#' @examples library(xfun)
#' prose_index(c('a', '```', 'b', '```', 'c'))
#' prose_index(c('a', '````', '```r', '1+1', '```', '````', 'c'))
prose_index = function(x, warn = TRUE) {
  xi = seq_along(x); n = length(idx <- code_lines_regex(x))
  if (n == 0) return(xi)
  if (n %% 2 != 0) {
    if (warn) warning('Code fences are not balanced')
    # treat all lines as prose
    return(xi)
  }
  idx2 = matrix(idx, nrow = 2)
  idx2 = unlist(mapply(seq, idx2[1, ], idx2[2, ], SIMPLIFY = FALSE))
  xi[-idx2]
}

# find starting and ending lines of code blocks via regex (may not be accurate)
code_lines_regex = function(x) {
  idx = NULL
  # if raw HTML <pre></pre> exists, it should be treated as code block
  inside_pre = if (length(p1 <- grep('<pre>', x))) {
    p2 = grep('</pre>', x)
    if (length(p1) == length(p2)) {
      idx = rbind(p1, p2)
      function(i) any(i > p1 & i < p2)
    }
  }
  r = '^(\\s*```+).*'; s = ''
  # shouldn't match ``` ``text``` ```, which is inline code, not code block
  i1 = grepl(r, x); i2 = !grepl('^\\s*```+\\s+`', x); i3 = !grepl('-->\\s*$', x)
  for (i in which(i1 & i2 & i3)) {
    if (is.function(inside_pre) && inside_pre(i)) next
    if (s == '') {
      s = gsub(r, '\\1', x[i]); idx = c(idx, i); next
    }
    # look for the next line with the same amount of backticks (end of block)
    if (grepl(paste0('^', s), x[i])) {
      idx = c(idx, i); s = ''
    }
  }
  idx
}

# find lines via commonmark (accurate but gregexpr()/substring() are slow)
code_lines_cmark = function(x) {
  xml = commonmark::markdown_xml(x, sourcepos = TRUE)
  r = '(?<=<code_block sourcepos=")(\\d+):\\d+-(\\d+):\\d+(?=")'
  m = gregexpr(r, xml, perl = TRUE)[[1]]
  if (all(m < 0)) return()
  s = attr2(m, 'capture.start'); l = attr2(m, 'capture.length')
  as.integer(substring(xml, s, s + l - 1))
}

#' Protect math expressions in pairs of backticks in Markdown
#'
#' For Markdown renderers that do not support LaTeX math, we need to protect
#' math expressions as verbatim code (in a pair of backticks), because some
#' characters in the math expressions may be interpreted as Markdown syntax
#' (e.g., a pair of underscores may make text italic). This function detects
#' math expressions in Markdown (by heuristics), and wrap them in backticks.
#'
#' Expressions in pairs of dollar signs or double dollar signs are treated as
#' math, if there are no spaces after the starting dollar sign, or before the
#' ending dollar sign. There should be a space or `(` before the starting dollar
#' sign, unless the math expression starts from the very beginning of a line.
#' For a pair of single dollar signs, the ending dollar sign should not be
#' followed by a number, and the inner math expression should not be wrapped in
#' backticks. With these assumptions, there should not be too many false
#' positives when detecing math expressions.
#'
#' Besides, LaTeX environments (\verb{\begin{*}} and \verb{\end{*}}) are also
#' protected in backticks.
#' @param x A character vector of text in Markdown.
#' @param token A character string to wrap math expressions at both ends. This
#'   can be a unique token so that math expressions can be reliably identified
#'   and restored after the Markdown text is converted.
#' @param use_block Whether to use code blocks (```` ```md-math ````) to protect
#'   `$$ $$` expressions that span across multiple lines. This is necessary when
#'   a certain line in the math expression starts with a special character that
#'   can accidentally start a new element (e.g., a leading `+` may start a
#'   bullet list). Only code blocks can prevent this case.
#' @return A character vector with math expressions in backticks.
#' @note If you are using Pandoc or the \pkg{rmarkdown} package, there is no
#'   need to use this function, because Pandoc's Markdown can recognize math
#'   expressions.
#' @export
#' @examples library(xfun)
#' protect_math(c('hi $a+b$', 'hello $$\\alpha$$', 'no math here: $x is $10 dollars'))
#' protect_math(c('hi $$', '\\begin{equation}', 'x + y = z', '\\end{equation}'))
#' protect_math('$a+b$', '===')
protect_math = function(x, token = '', use_block = FALSE) {
  i = prose_index(x)
  if (length(i)) x[i] = escape_math(x[i], token, use_block)
  x
}

escape_math = function(x, token = '', use_block = FALSE) {
  # replace $x$ with `\(x\)` (protect inline math in <code></code>)
  m = gregexpr('(?<=^|[\\s(])[$](?![ `])[^$]+?(?<![ `])[$](?![$0123456789])', x, perl = TRUE)
  regmatches(x, m) = lapply(regmatches(x, m), function(z) {
    if (length(z) == 0) return(z)
    z = sub('^[$]', paste0('`', token, '\\\\('), z)
    z = sub('[$]$', paste0('\\\\)', token, '`'), z)
    z
  })
  # replace $$x$$ with `$$x$$` (protect display math)
  m = gregexpr('(?<=^|[\\s(])[$][$](?! )[^$]+?(?<! )[$][$]', x, perl = TRUE)
  regmatches(x, m) = lapply(regmatches(x, m), function(z) {
    if (length(z) == 0) return(z)
    paste0('`', token, z, token, '`')
  })
  # now, if there are still lines starting and ending with $$, they might be
  # math expressions of display style spanning multiple lines, e.g.,
  # $$\alpha +
  # \beta$$
  # we assume that $$ can only appear once on one line
  i = vapply(gregexpr('[$]', x), length, integer(1)) == 2
  if (any(i)) {
    r1 = sprintf('\\1%s\\2', if (use_block) {
      paste(c('```{.md-math', if (token != '') c(' .', token), '}\n\\1'), collapse = '')
    } else paste0('`', token))
    x[i] = gsub('^(\\s*)([$][$][^ ]+)', r1, x[i], perl = TRUE)
    r2 = if (use_block) '\\1\\2\n\\1```' else paste0('\\1\\2', token, '`')
    x[i] = gsub('^(\\s*)(.*?[^ ][$][$])$', r2, x[i], perl = TRUE)
  }
  # equation environments (\begin and \end must match)
  i1 = grep('^\\\\begin\\{[^}]+\\}$', x)
  i2 = grep('^\\\\end\\{[^}]+\\}$', x)
  if (length(i1) == length(i2)) {
    # TODO: do not protect inner environments in case of nested environments (#57)
    x[i1] = paste0('`', token, x[i1])
    x[i2] = paste0(x[i2], token, '`')
  }
  x
}

#' Create a fenced block in Markdown
#'
#' Wrap content with fence delimiters such as backticks (code blocks) or colons
#' (fenced Div). Optionally the fenced block can have attributes. The function
#' `fenced_div()` is a shorthand of `fenced_block(char = ':')`.
#' @param x A character vector of the block content.
#' @param attrs A vector of block attributes.
#' @param fence The fence string, e.g., `:::` or ```` ``` ````. This will be
#'   generated from the `char` argument by default.
#' @param char The fence character to be used to generate the fence string by
#'   default.
#' @return `fenced_block()` returns a character vector that contains both the
#'   fences and content.
#' @export
#' @examples
#' # code block with class 'r' and ID 'foo'
#' xfun::fenced_block('1+1', c('.r', '#foo'))
#' # fenced Div
#' xfun::fenced_block('This is a **Div**.', char = ':')
fenced_block = function(x, attrs = NULL, fence = make_fence(x, char), char = '`') {
  a = block_attr(attrs)
  c('', paste0(fence, a), x, fence)
}

#' @param ... Arguments to be passed to `fenced_block()`.
#' @rdname fenced_block
#' @export
fenced_div = function(...) fenced_block(..., char = ':')

#' @param start The number of characters to start searching `x` with. If the
#'   string of this number of characters is found, add one more character, and
#'   repeat the search.
#' @return `make_fence()` returns a character string. If the block content
#'   contains `N` fence characters (e.g., backticks), use `N + 1` characters as
#'   the fence.
#' @rdname fenced_block
#' @export
#' @examples
#' # three backticks by default
#' xfun::make_fence('1+1')
#' # needs five backticks for the fences because content has four
#' xfun::make_fence(c('````r', '1+1', '````'))
make_fence = function(x, char = '`', start = 3) {
  f = strrep(char, start)
  while (any(grepl(f, x, fixed = TRUE))) f = paste0(f, char)
  f
}

# concatenate block attributes for fenced blocks
block_attr = function(attrs) {
  a = paste(attrs, collapse = ' ')
  if (grepl('[ .=#]', a)) a = paste0('{', a, '}')
  if (a == '') a else paste0(' ', a)
}

#' Embed a file, multiple files, or directory on an HTML page
#'
#' For a file, first encode it into base64 data (a character string). Then
#' generate a hyperlink of the form \samp{<a href="base64 data"
#' download="filename">Download filename</a>}. The file can be downloaded when
#' the link is clicked in modern web browsers. For a directory, it will be
#' compressed as a zip archive first, and the zip file is passed to
#' `embed_file()`. For multiple files, they are also compressed to a zip file
#' first.
#'
#' These functions can be called in R code chunks in R Markdown documents with
#' HTML output formats. You may embed an arbitrary file or directory in the HTML
#' output file, so that readers of the HTML page can download it from the
#' browser. A common use case is to embed data files for readers to download.
#' @param path Path to the file(s) or directory.
#' @param name The default filename to use when downloading the file. Note that
#'   for `embed_dir()`, only the base name (of the zip filename) will be used.
#' @param text The text for the hyperlink.
#' @param ... For `embed_file()`, additional arguments to be passed to
#'   [html_tag()] (e.g., `class = 'foo'`). For `embed_dir()` and
#'   `embed_files()`, arguments passed to `embed_file()`.
#' @note Windows users may need to install Rtools to obtain the \command{zip}
#'   command to use `embed_dir()` and `embed_files()`.
#'
#'   Internet Explorer does not support downloading embedded files. Chrome has a
#'   2MB limit on the file size.
#' @return An HTML tag \samp{<a>} with the appropriate attributes.
#' @export
#' @examples
#' logo = xfun:::R_logo()
#' link = xfun::embed_file(logo, text = 'Download R logo')
#' link
#' if (interactive()) xfun::html_view(link)
embed_file = function(path, name = basename(path), text = paste('Download', name), ...) {
  h = base64_uri(path)
  html_tag('a', text, href = h, download = name, ...)
}

#' @rdname embed_file
#' @export
embed_dir = function(path, name = paste0(normalize_path(path), '.zip'), ...) {
  name  = gsub('/', '', basename(name))
  in_dir(path, {
    name = file.path(tempdir(), name); on.exit(file.remove(name), add = TRUE)
    zip(name, '.'); embed_file(name, ...)
  })
}

#' @rdname embed_file
#' @export
embed_files = function(path, name = with_ext(basename(path[1]), '.zip'), ...) {
  name = file.path(tempdir(), basename(name))
  on.exit(file.remove(name), add = TRUE)
  zip(name, path)
  embed_file(name, ...)
}

zip = function(name, ...) {
  if (utils::zip(name, ...) != 0) stop('Failed to create the zip archive ', name)
  invisible(0)
}

#' Generate a simple Markdown pipe table
#'
#' A minimal Markdown table generator using the pipe `|` as column separators.
#'
#' The default argument values can be set via global options with the prefix
#' `xfun.md_table.`, e.g., `options(xfun.md_table.digits 2, xfun.md_table.na =
#' 'n/a')`.
#' @param x A 2-dimensional object (e.g., a matrix or data frame).
#' @param digits The number of decimal places to be passed to [round()]. It can
#'   be a integer vector of the same length as the number of columns in `x` to
#'   round columns separately. The default is `3`.
#' @param na A character string to represent `NA` values. The default is an
#'   empty string.
#' @param newline A character string to substitute `\n` in `x` (because pipe
#'   tables do not support line breaks in cells). The default is a space.
#' @param limit The maximum number of rows to show in the table. If it is
#'   smaller than the number of rows, the data in the middle will be omitted. If
#'   it is of length 2, the second number will be used to limit the number of
#'   columns. Zero and negative values are ignored.
#' @return A character vector.
#' @seealso [knitr::kable()] (which supports more features)
#' @export
#' @examples
#' xfun::md_table(head(iris))
#' xfun::md_table(mtcars, limit = c(10, 6))
md_table = function(x, digits = NULL, na = NULL, newline = NULL, limit = NULL) {
  if (length(d <- dim(x)) != 2)
    stop('xfun::md_table() only supports 2-dimensional objects.')
  if (d[2] == 0) return(character())
  if (is.null(digits))
    digits = getOption('xfun.md_table.digits', min(getOption('digits'), 3))
  digits = rep(digits, d[2])  # recycle for all columns
  num = logical(d[2])  # numeric columns
  for (j in seq_len(d[2])) if (is.numeric(x[, j])) {
    num[j] = TRUE
    x[, j] = round(x[, j], digits[j])
  }
  is_na = is.na(x)
  x = as.matrix(format(x))
  if (any(is_na)) x[is_na] = na %||% getOption('xfun.md_table.na', '')
  rn = rownames(x)
  # ignore empty and automatic row names (integers)
  if (!all(grepl('^[0-9]*$', rn))) {
    x = cbind(' ' = rn, x)
    num = c(FALSE, num)
  }
  # get first and last limit/2 rows/cols in N rows/cols
  if (length(limit <- limit %||% getOption('xfun.md_table.limit'))) {
    # subset rows
    l1 = limit[1]
    if (l1 > 0 && l1 < d[1]) {
      n1 = round(l1/2); n2 = l1 - n1
      x = rbind(head(x, n1), '&vellip;', tail(x, n2))
    }
    # subset columns
    if (length(limit) >= 2 && (l2 <- limit[2]) > 0 && l2 < d[2]) {
      n1 = round(l2/2); n2 = l2 - n1
      x = cbind(
        x[, seq_len(n1), drop = FALSE],
        `...` = '...',
        x[, d[2] - seq_len(n2) + 1, drop = FALSE]
      )
      num = c(head(num, n1), FALSE, tail(num, n2))
    }
    d = dim(x)
    if (d[2] == 0) return(character())
  }
  cn = colnames(x) %||% rep(' ', d[2])  # table header
  a = ifelse(num, '--:', '---')  # alignment
  a[cn == '...'] = ':-:'
  x = rbind(cn, a, x)
  d = dim(x)
  x = gsub('|', '\\|', x, fixed = TRUE)
  dim(x) = d
  res = do.call(function(...) paste(..., sep = '|'), as.data.frame(x))
  res = gsub('\n', newline %||% getOption('xfun.md_table.newline', ' '), res, fixed = TRUE)
  paste0('|', res, '|')
}

#' Represent a (recursive) list with (nested) tabsets
#'
#' The tab titles are names of list members, and the tab content contains the
#' values of list members. If a list member is also a list, it will be
#' represented recursively with a child tabset.
#' @param x A list.
#' @param value A function to print the value of a list member. By default,
#'   [str()] is used to print the structure of the value. You may also use
#'   [dput()] to output the full value, but it may be slow when the size of the
#'   value is too big.
#' @return A character vector of Markdown that can be rendered to HTML with
#'   [litedown::mark()].
#' @export
#' @examples
#' xfun::tabset(iris)
#' xfun::tabset(iris, dput)
#' xfun::tabset(iris, print)
#'
#' # a deeply nested list
#' plot(1:10)
#' p = recordPlot()
#' xfun::tabset(p)
tabset = function(x, value = str) {
  obj = paste(deparse(substitute(x)), collapse = ' ')
  md_viewable(.tabset(x, value), meta = list(
    css = c('@default', '@tabsets'), js = '@tabsets',
    title = paste0('`xfun::tabset(', obj, ')`')
  ))
}

.tabset = function(x, value) {
  if (is.list(x)) {
    idx = seq_along(x)
    nms = names(x) %||% sprintf('[[%d]]', idx)
    res = NULL
    for (i in idx) {
      con = paste0('  ', .tabset(unclass(x)[[i]], value))  # indent content
      res = c(res, paste0('- `', nms[i], '`'), '', con)
    }
    # attach attributes to the last tab if necessary
    if (is.list(att <- attributes(x))) {
      att$names = NULL  # names have been displayed
      if (length(att)) res = c(
        res, '', '- `attr(*)`', '', paste0('  ', .tabset(att, value))
      )
    }
    c('::: tabset', res, ':::')
  } else {
    c('```r', capture.output(value(x)), '```')
  }
}

md_viewable = function(x, ...) {
  structure(x, class = c('xfun_md_viewable', 'record_asis'), ...)
}

#' @export
print.xfun_md_viewable = function(x, ...) {
  if (loadable('litedown')) {
    html = litedown::mark(text = c('---', '---', '', x), meta = attr2(x, 'meta'), ...)
    html_view(html)
  } else {
    warning("Cannot preview the result since the 'litedown' package is not available.")
    raw_string(x)
  }
}

library(testit)

assert('find_globals() identifies global variables', {
  env = list2env(
    list(qwer = 1, g = 2, y = 3, d = 4, z = 5), parent = emptyenv()
  )
  # nothing from outside environment
  (find_globals('x=1', env) %==% character(0))
  # qwer must be created outside somewhere
  (find_globals('a=1; b=a; d=qwer', env) %==% 'qwer')
  (find_globals('a=function(){f=2;g}', env) %==% 'g')
  # y was assigned locally in z, but there is another y outside from nowhere
  (find_globals('z=function(){y=1};y', env) %==% 'y')
  # more complicated cases: operators, subscripts, ...
  (find_globals(c('a=1%*%1%o%2 %in% d', 'b=d%%10+3%/%2-z[1:3]'), env) %==% c('d', 'z'))
})

assert('global_vars() returns values of global variables in code', {
  e = new.env(parent = emptyenv())
  e$x = 42
  res = global_vars('y = x + 1', e)
  (is.list(res))
  ('x' %in% names(res))
  (res$x %==% 42)
})

assert('md5_obj() computes md5 hash for a character vector or object', {
  h1 = md5_obj(c('a', 'b'))
  h2 = md5_obj(c('a', 'b'))
  h3 = md5_obj(c('a', 'c'))
  (h1 %==% h2)
  (h1 != h3)
  (nchar(h1) %==% 32L)
})

assert('clean_cache() removes old cache files with same base name', {
  d = tempfile()
  dir.create(d)
  base = 'mycode'
  ext = 'rds'
  old_hash = paste(rep('a', 32), collapse = '')
  new_hash = paste(rep('b', 32), collapse = '')
  old_file = file.path(d, sprintf('%s_%s.%s', base, old_hash, ext))
  new_file = file.path(d, sprintf('%s_%s.%s', base, new_hash, ext))
  writeLines('old', old_file)
  clean_cache(new_file, ext)  # should remove old_file
  (!file.exists(old_file))
  unlink(d, recursive = TRUE)
})

assert('cache_rds() caches an expression to disk', {
  d = tempfile()
  dir.create(d)
  f = 'mycode.rds'
  # first call evaluates and caches
  r1 = cache_rds(1 + 1, file = f, dir = paste0(d, '/'))
  (r1 %==% 2)
  # second call loads from cache
  r2 = cache_rds(1 + 1, file = f, dir = paste0(d, '/'))
  (r2 %==% 2)
  # rerun = TRUE forces re-evaluation
  r3 = cache_rds(1 + 1, file = f, dir = paste0(d, '/'), rerun = TRUE)
  (r3 %==% 2)
  unlink(d, recursive = TRUE)
})

#' Manipulate filename extensions
#'
#' Functions to obtain (`file_ext()`), remove (`sans_ext()`), and
#' change (`with_ext()`) extensions in filenames.
#'
#' `file_ext()` is similar to [tools::file_ext()], and
#' `sans_ext()` is similar to [tools::file_path_sans_ext()].
#' The main differences are that they treat `tar.(gz|bz2|xz)` and
#' `nb.html` as extensions (but functions in the \pkg{tools} package
#' doesn't allow double extensions by default), and allow characters `~`
#' and `#` to be present at the end of a filename.
#' @param x A character of file paths.
#' @param extra Extra characters to be allowed in the extensions. By default,
#'   only alphanumeric characters are allowed (and also some special cases in
#'   \sQuote{Details}). If other characters should be allowed, they can be
#'   specified in a character string, e.g., `"-+!_#"`.
#' @export
#' @return A character vector of the same length as `x`.
#' @examples library(xfun)
#' p = c('abc.doc', 'def123.tex', 'path/to/foo.Rmd', 'backup.ppt~', 'pkg.tar.xz')
#' file_ext(p); sans_ext(p); with_ext(p, '.txt')
#' with_ext(p, c('.ppt', '.sty', '.Rnw', 'doc', 'zip')); with_ext(p, 'html')
#'
#' # allow for more characters in extensions
#' p = c('a.c++', 'b.c--', 'c.e##')
#' file_ext(p)  # -/+/# not recognized by default
#' file_ext(p, extra = '-+#')
file_ext = function(x, extra = '') {
  ext = character(length(x))
  i = grep(r <- reg_path(extra), x)
  ext[i] = sub(r, '\\3', x[i])
  ext
}

#' @rdname file_ext
#' @export
sans_ext = function(x, extra = '') {
  sub(reg_path(extra), '\\1', x)
}

#' @param ext A vector of new extensions. It must be either of length 1, or the
#'   same length as `x`.
#' @rdname file_ext
#' @export
with_ext = function(x, ext, extra = '') {
  if (anyNA(ext)) stop("NA is not allowed in 'ext'")
  n1 = length(x); n2 = length(ext)
  if (n1 * n2 == 0) return(x)
  i = !grepl('^[.]', ext) & ext != ''
  ext[i] = paste0('.', ext[i])

  if (all(ext == '')) ext = ''
  r = sub('[$]$', '?$', reg_ext(extra))  # make extensions in 'x' optional
  if (length(ext) == 1) return(sub(r, ext, x))

  if (n1 > 1 && n1 != n2) stop("'ext' must be of the same length as 'x'")
  mapply(sub, r, ext, x, USE.NAMES = FALSE)
}

# regex to extract base path and extension from a file path
reg_ext  = function(extra = '') {
  sprintf('([.](([%s[:alnum:]]+|tar[.](gz|bz2|xz)|nb[.]html)[~#]?))$', extra)
}
reg_path = function(...) paste0('^(.*?)', reg_ext(...))

#' Normalize paths
#'
#' A wrapper function of `normalizePath()` with different defaults.
#' @param x,winslash,must_work Arguments passed to
#'   [normalizePath()].
#' @param resolve_symlink Whether to resolve symbolic links.
#' @export
#' @examples library(xfun)
#' normalize_path('~')
normalize_path = function(x, winslash = '/', must_work = FALSE, resolve_symlink = TRUE) {
  if (!resolve_symlink) {
    # apply the trick on all files on Windows since Sys.readlink() doesn't work
    # and we can't know which files are symlinks
    i = if (is_windows()) file_test('-f', x) else is_symlink(x)
    b = basename(x[i])
    x[i] = dirname(x[i])  # normalize the dirs of symlinks instead
  }
  res = normalizePath(x, winslash = winslash, mustWork = must_work)
  if (is_windows()) res[is.na(x)] = NA
  if (!resolve_symlink) {
    res[i] = file.path(res[i], b, fsep = winslash)
  }
  res
}

is_symlink = function(x) {
  !is.na(y <- Sys.readlink(x)) & (y != '')
}

#' Test if two paths are the same after they are normalized
#'
#' Compare two paths after normalizing them with the same separator (`/`).
#' @param p1,p2 Two vectors of paths.
#' @param ... Arguments to be passed to [normalize_path()].
#' @export
#' @examples library(xfun)
#' same_path('~/foo', file.path(Sys.getenv('HOME'), 'foo'))
same_path = function(p1, p2, ...) {
  normalize_path(p1, ...) == normalize_path(p2, ...)
}

#' Find file paths that exist
#'
#' This is a shorthand of `x[file.exists(x)]`, and optionally returns the
#' first existing file path.
#' @param x A vector of file paths.
#' @param first Whether to return the first existing path. If `TRUE` and no
#'   specified files exist, it will signal an error unless the argument
#'   `error = FALSE`.
#' @param error Whether to throw an error when `first = TRUE` but no files
#'   exist. It can also take a character value, which will be used as the error
#'   message.
#' @return A vector of existing file paths.
#' @export
#' @examples
#' xfun::existing_files(c('foo.txt', system.file('DESCRIPTION', package='xfun')))
existing_files = function(x, first = FALSE, error = TRUE) {
  x = x[file_exists(x)]
  if (!first) return(x)
  x = head(x, 1)
  if (length(x) != 1 && !identical(error, FALSE)) {
    if (isTRUE(error)) error = 'None of the specified files exist.'
    stop(error, call. = FALSE)
  }
  x
}

#' Return the (possible) root directory of a project
#'
#' Given a path of a file (or dir) in a potential project (e.g., an R package or
#' an RStudio project), return the path to the project root directory.
#'
#' The search for the root directory is performed by a series of tests,
#' currently including looking for a \file{DESCRIPTION} file that contains
#' `Package: *` (which usually indicates an R package), and a
#' \file{*.Rproj} file that contains `Version: *` (which usually indicates
#' an RStudio project). If files with the expected patterns are not found in the
#' initial directory, the search will be performed recursively in upper-level
#' directories.
#' @param path The initial path to start the search. If it is a file path, its
#'   parent directory will be used.
#' @param rules A matrix of character strings of two columns: the first column
#'   contains regular expressions to look for filenames that match the patterns,
#'   and the second column contains regular expressions to match the content of
#'   the matched files. The regular expression can be an empty string, meaning
#'   that it will match anything.
#' @return Path to the root directory if found, otherwise `NULL`.
#' @export
#' @note This function was inspired by the \pkg{rprojroot} package, but is much
#'   less sophisticated. It is a rather simple function designed to be used in
#'   some of packages that I maintain, and may not meet the need of general
#'   users until this note is removed in the future (which should be unlikely).
#'   If you are sure that you are working on the types of projects mentioned in
#'   the \sQuote{Details} section, this function may be helpful to you,
#'   otherwise please consider using \pkg{rprojroot} instead.
proj_root = function(path = './', rules = root_rules) {
  path = normalize_path(path)
  dir = if (dir_exists(path)) path else dirname(path)
  if (same_path(dir, file.path(dir, '..'))) return()
  if (is.null(dim(rules))) dim(rules) = c(1, length(rules))
  for (i in seq_len(nrow(rules))) {
    file = rules[i, 1]; pattern = rules[i, 2]
    for (f in list.files(dir, file, full.names = TRUE)) {
      if (pattern == '' || length(grep(pattern, read_utf8(f)))) return(dir)
    }
  }
  proj_root(dirname(dir), rules)
}

#' @rdname proj_root
#' @export
root_rules = matrix(c(
  '^DESCRIPTION$', '^Package: ',
  '.+[.]Rproj$',   '^Version: '
), ncol = 2, byrow = TRUE, dimnames = list(NULL, c('file', 'pattern')))

#' Get the relative path of a path relative to a directory
#'
#' Given a directory, return the relative path that is relative to this
#' directory. For example, the path \file{foo/bar.txt} relative to the directory
#' \file{foo/} is \file{bar.txt}, and the path \file{/a/b/c.txt} relative to
#' \file{/d/e/} is \file{../../a/b/c.txt}.
#' @param dir Path to a directory.
#' @param x A vector of paths to be converted to relative paths.
#' @param use.. Whether to use double-dots (\file{..}) in the relative path. A
#'   double-dot indicates the parent directory (starting from the directory
#'   provided by the `dir` argument).
#' @param error Whether to signal an error if a path cannot be converted to a
#'   relative path.
#' @return A vector of relative paths if the conversion succeeded; otherwise the
#'   original paths when `error = FALSE`, and an error when `error =
#'   TRUE`.
#' @export
#' @examples
#' xfun::relative_path('foo/bar.txt', 'foo/')
#' xfun::relative_path('foo/bar/a.txt', 'foo/haha/')
#' xfun::relative_path(getwd())
relative_path = function(x, dir = '.', use.. = TRUE, error = TRUE) {
  res = x
  for (i in seq_along(x)) res[i] = relative_path_one(x[i], dir, use.., error)
  res
}

relative_path_one = function(x, dir, use.., error) {
  # on Windows, if a relative path doesn't exist, normalizePath() will use
  # getwd() as its parent dir; however, normalizePath() just returns the
  # relative path on *nix, and we have to assume it's relative to getwd()
  abs_path = function(p) {
    if (!file.exists(p) && is_unix() && is_rel_path(p)) p = file.path(getwd(), p)
    normalize_path(p)
  }
  p = abs_path(x); n1 = nchar(p)
  if ((n1 <- nchar(p)) == 0) return(x)  # not sure what you mean
  d = abs_path(dir); n2 = nchar(d)
  if (is_sub_path(p, d, n2)) {
    p2 = get_subpath(p, n1, n2)
    if (p2 == '') p2 = '.'  # if the subpath is empty, it means the current dir
    return(p2)
  }
  if (!use..) {
    if (error) stop("When use.. = FALSE, the path 'x' must be under the 'dir'")
    return(x)
  }
  s = '../'; d1 = d
  while (!is_sub_path(p, d2 <- dirname(d1))) {
    if (same_path(d1, d2)) {
      if (error) stop(
        "The path 'x' cannot be converted to a relative path to 'dir'. ",
        "Perhaps they are on different volumes of the disk."
      )
      return(x)
    }
    s = paste0('../', s)
    d1 = d2  # go to one level up
  }
  paste0(s, get_subpath(p, n1, nchar(d2)))
}

#' Test if a path is a subpath of a dir
#'
#' Check if the path starts with the dir path.
#' @inheritParams is_abs_path
#' @param dir A vector of directory paths.
#' @param n The length of `dir` paths.
#' @return A logical vector.
#' @note You may want to normalize the values of the `x` and `dir` arguments
#'   first (with [xfun::normalize_path()]), to make sure the path separators
#'   are consistent.
#' @export
#' @examples
#' xfun::is_sub_path('a/b/c.txt', 'a/b')  # TRUE
#' xfun::is_sub_path('a/b/c.txt', 'd/b')  # FALSE
#' xfun::is_sub_path('a/b/c.txt', 'a\\b')  # FALSE (even on Windows)
is_sub_path = function(x, dir, n = nchar(dir)) substr(x, 1, n) == dir

# remove the first n2 characters and the possible / from the path
get_subpath = function(p, n1, n2) {
  p = substr(p, n2 + 1, n1)
  sub('^/', '', p)
}

#' Test if paths are relative or absolute
#'
#' On Unix, check if the paths start with \file{/} or \file{~} (if they do, they
#' are absolute paths). On Windows, check if a path remains the same (via
#' [xfun::same_path()]) if it is prepended with \file{./} (if it does, it is a
#' relative path).
#' @param x A vector of paths.
#' @return A logical vector.
#' @export
#' @examples
#' xfun::is_abs_path(c('C:/foo', 'foo.txt', '/Users/john/', tempdir()))
#' xfun::is_rel_path(c('C:/foo', 'foo.txt', '/Users/john/', tempdir()))
is_abs_path = function(x) {
  if (is_unix()) grepl('^[/~]', x) else !same_path(x, file.path('.', x))
}

#' @rdname is_abs_path
#' @export
is_rel_path = function(x) !is_abs_path(x)

#' Test if a path is a web path
#'
#' Check if a path starts with \file{http://} or \file{https://} or
#' \file{ftp://} or \file{ftps://}.
#' @inheritParams is_abs_path
#' @return A logical vector.
#' @export
#' @examples
#' xfun::is_web_path('https://www.r-project.org')  # TRUE
#' xfun::is_web_path('www.r-project.org')  # FALSE
is_web_path = function(x) {
  grepl('^(f|ht)tps?://', x)
}

#' Get the relative path of a path in a project relative to the current working
#' directory
#'
#' First compose an absolute path using the project root directory and the
#' relative path components, i.e., [`file.path`]`(root, ...)`. Then
#' convert it to a relative path with [relative_path()], which is
#' relative to the current working directory.
#'
#' This function was inspired by `here::here()`, and the major difference
#' is that it returns a relative path by default, which is more portable.
#' @param ... A character vector of path components *relative to the root
#'   directory of the project*.
#' @param root The root directory of the project.
#' @param error Whether to signal an error if the path cannot be converted to a
#'   relative path.
#' @return A relative path, or an error when the project root directory cannot
#'   be determined or the conversion failed and `error = TRUE`.
#' @export
#' @examples
#' \dontrun{
#' xfun::from_root('data', 'mtcars.csv')
#' }
from_root = function(..., root = proj_root(), error = TRUE) {
  if (is.null(root)) stop('Cannot determine the root directory of the current project.')
  p = file.path(root, ..., fsep = '/')
  relative_path(p, error = error)
}

#' Find a file or directory under a root directory
#'
#' Given a path, try to find it recursively under a root directory. The input
#' path can be an incomplete path, e.g., it can be a base filename, and
#' `magic_path()` will try to find this file under subdirectories.
#' @param ... A character vector of path components.
#' @param root The root directory under which to search for the path. If
#'   `NULL`, the current working directory is used.
#' @param relative Whether to return a relative path.
#' @param error Whether to signal an error if the path is not found, or multiple
#'   paths are found.
#' @param message Whether to emit a message when multiple paths are found and
#'   `error = FALSE`.
#' @param n_dirs The number of subdirectories to recursively search. The
#'   recursive search may be time-consuming when there are a large number of
#'   subdirectories under the root directory. If you really want to search for
#'   all subdirectories, you may try `n_dirs = Inf`.
#' @return The path found under the root directory, or an error when `error
#'   = TRUE` and the path is not found (or multiple paths are found).
#' @export
#' @examples
#' \dontrun{
#' xfun::magic_path('mtcars.csv')  # find any file that has the base name mtcars.csv
#' }
magic_path = function(
  ..., root = proj_root(), relative = TRUE, error = TRUE,
  message = getOption('xfun.magic_path.message', TRUE),
  n_dirs = getOption('xfun.magic_path.n_dirs', 10000)
) {
  if (file.exists(p <- file.path(...))) return(p)
  if (is.null(root)) root = getwd()
  nd = 0
  # find a path 'f' recursively under a directory 'd'
  find_it = function(f, d) {
    if (nd > n_dirs) {
      if (error) stop(
        'Failed to find the path under ', n_dirs, ' subdirectories. If you want ',
        'to search for the path in more subdirectories, increase the value of ',
        "the 'n_dirs' argument of magic_path()."
      )
      return(p)
    }
    ds = list.files(d, full.names = TRUE)
    ds = ds[dir_exists(ds)]
    if ((n1 <- length(ds)) == 0) return()
    nd <<- nd + n1
    fs = file.path(ds, f)
    fs = fs[file.exists(fs)]
    if ((n2 <- length(fs)) == 1) return(fs)
    if (n2 > 1) {
      msg = c(
        'Found more than one path containing the input path "', f, '":\n\n',
        paste('*', fs, collapse = '\n')
      )
      if (error) stop(msg)
      if (message) base::message(msg, '\n\nReturned the first one.')
      return(fs[1])
    }
    # look into subdirectories one by one
    for (i in seq_len(n1)) {
      fs = find_it(f, file.path(ds[i]))
      if (length(fs)) return(fs)
    }
  }
  f = find_it(p, root)
  if (is.null(f)) {
    if (error) stop('Could not find the path "', p, '" in any subdirectories.')
    p
  } else {
    if (relative) relative_path(f, error = error) else f
  }
}

#' Test the existence of files and directories
#'
#' These are wrapper functions of [`utils::file_test()]` to test the
#' existence of directories and files. Note that `file_exists()` only tests
#' files but not directories, which is the main difference between
#' [file.exists()] in base R. If you use are using the R version
#' 3.2.0 or above, `dir_exists()` is the same as [dir.exists()]
#' in base R.
#' @param x A vector of paths.
#' @export
#' @return A logical vector.
dir_exists = function(x) file_test('-d', x)

#' @rdname dir_exists
#' @export
file_exists = function(x) file_test('-f', x)

#' Create a directory recursively by default
#'
#' First check if a directory exists. If it does, return `TRUE`, otherwise
#' create it with [`dir.create`]`(recursive = TRUE)` by default.
#' @param x A path name.
#' @param recursive Whether to create all directory components in the path.
#' @param ... Other arguments to be passed to [dir.create()].
#' @return A logical value indicating if the directory either exists or is
#'   successfully created.
#' @export
dir_create = function(x, recursive = TRUE, ...) {
  dir_exists(x) || dir.create(x, recursive = recursive, ...)
}

#' Rename files with a sequential numeric prefix
#'
#' Rename a series of files and add an incremental numeric prefix to the
#' filenames. For example, files \file{a.txt}, \file{b.txt}, and \file{c.txt}
#' can be renamed to \file{1-a.txt}, \file{2-b.txt}, and \file{3-c.txt}.
#' @param pattern A regular expression for [list.files()] to obtain
#'   the files to be renamed. For example, to rename `.jpeg` files, use
#'   `pattern = "[.]jpeg$"`.
#' @param format The format for the numeric prefix. This is passed to
#'   [sprintf()]. The default format is `"\%0Nd"` where `N
#'   = floor(log10(n)) + 1` and `n` is the number of files, which means the
#'   prefix may be padded with zeros. For example, if there are 150 files to be
#'   renamed, the format will be `"\%03d"` and the prefixes will be
#'   `001`, `002`, ..., `150`.
#' @param replace Whether to remove existing numeric prefixes in filenames.
#' @param start The starting number for the prefix (it can start from 0).
#' @param dry_run Whether to not really rename files. To be safe, the default is
#'   `TRUE`. If you have looked at the new filenames and are sure the new
#'   names are what you want, you may rerun `rename_seq()` with
#'   `dry_run = FALSE` to actually rename files.
#' @return A named character vector. The names are original filenames, and the
#'   vector itself is the new filenames.
#' @export
#' @examples xfun::rename_seq()
#' xfun::rename_seq('[.](jpeg|png)$', format = '%04d')
rename_seq = function(
  pattern = '^[0-9]+-.+[.]Rmd$', format = 'auto', replace = TRUE, start = 1,
  dry_run = TRUE
) {
  n = length(files <- list.files('.', pattern))
  if (n == 0) return(files)

  files2 = if (replace) sub('^[0-9]+-*', '', files) else files
  if (format == 'auto') format = paste0('%0', floor(log10(n)) + 1, 'd')
  files2 = paste(sprintf(format, seq_len(n) + start - 1), files2, sep = '-')

  if (!dry_run) file.rename(files, files2)
  structure(setNames(files2, files), class = 'xfun_rename_seq')
}

#' @export
print.xfun_rename_seq = function(x, ...) {
  x = unclass(x)
  tab = data.frame(original = names(x), ' ' = '->', new = unname(x), check.names = FALSE)
  if (loadable('knitr')) tab = knitr::kable(tab, 'simple')
  print(tab)
}

# return path to R's svg logo if it exists, otherwise return the jpg logo; or
# specify a regex to match the logo path, e.g., ext = 'jpg$'
R_logo = function(ext = NULL, all = FALSE) {
  x = file.path(R.home('doc'), 'html', c('Rlogo.svg', 'logo.jpg'))
  if (!is.null(ext)) x = grep(ext, x, value = TRUE)
  existing_files(x, first = !all)
}

#' Extract filenames from a URLs
#'
#' Get the base names of URLs via [basename()], and remove the
#' possible query parameters or hash from the names.
#' @param x A character vector of URLs.
#' @param default The default filename when it cannot be determined from the
#'   URL, e.g., when the URL ends with a slash.
#' @return A character vector of filenames at the end of URLs.
#' @export
#' @examples
#' xfun::url_filename('https://yihui.org/images/logo.png')
#' xfun::url_filename('https://yihui.org/index.html')
#' xfun::url_filename('https://yihui.org/index.html?foo=bar')
#' xfun::url_filename('https://yihui.org/index.html#about')
#' xfun::url_filename('https://yihui.org')
#' xfun::url_filename('https://yihui.org/')
url_filename = function(x, default = 'index.html') {
  # protocol shouldn't be treated as dir name, and query/hash should be removed
  x = gsub('^https?://|[?#].*$', '', x)
  f = basename(x)
  ifelse(grepl('/$', x) | x == f, default, f)
}

#' Delete an empty directory
#'
#' Use `list.file()` to check if there are any files or subdirectories
#' under a directory. If not, delete this empty directory.
#' @param dir Path to a directory. If `NULL` or the directory does not
#'   exist, no action will be performed.
#' @export
del_empty_dir = function(dir) {
  if (is.null(dir) || !dir_exists(dir)) return()
  files = list.files(dir, all.files = TRUE, no.. = TRUE)
  if (length(files) == 0) unlink(dir, recursive = TRUE)
}

#' Mark some paths as directories
#'
#' Add a trailing backlash to a file path if this is a directory. This is useful
#' in messages to the console for example to quickly identify directories from
#' files.
#'
#' If `x` is a vector of relative paths, directory test is done with path
#' relative to the current working dir. Use [xfun::in_dir()] or use absolute
#' paths.
#'
#' @param x Character vector of paths to files and directories.
#' @examples
#' mark_dirs(list.files(find.package("xfun"), full.names = TRUE))
#' @export
mark_dirs = function(x) {
  i = dir_exists(x) & !grepl("/$", x)
  x[i] = paste0(x[i], "/")
  x
}

# change list.files()'s default argument values
all_files = function(
  pattern = NULL, dir = '.', ignore.case = TRUE, full.names = TRUE,
  recursive = TRUE, ...
) {
  list.files(
    dir, pattern, ignore.case = ignore.case, full.names = full.names,
    recursive = recursive, no.. = TRUE, ...
  )
}

library(testit)

assert('clean_log() removes OK lines, current time, and other noise', {
  d = tempfile(); dir.create(d)
  owd = setwd(d); on.exit({ setwd(owd); unlink(d, recursive = TRUE) }, add = TRUE)

  writeLines(c(
    'log dir line',
    'R version line',
    '* using R version 4.4.0 ... OK',
    '* current time: 2026-07-07 14:49:46 UTC',
    '* checking foo ... WARNING',
    'some detail'
  ), '00check.log')

  x = clean_log()
  # first 2 lines removed, OK line removed, current time line removed
  (!any(grepl('OK$', x)))
  (!any(grepl('current time', x)))
  (x %==% c('* checking foo ... WARNING', 'some detail'))
})

#' Create or stop a local web application
#'
#' Create a local web application based on R's internal httpd.
#'
#' `new_app()` has two modes:
#'
#' - `name = ''`: start an app behind a lightweight proxy at
#'   `http://host:PORT/`, forwarding to R's internal httpd path
#'   `/custom/xfun:PORT/`.
#' - `name != ''`: register a legacy app directly on R's internal httpd at
#'   `http://127.0.0.1:BACKEND/custom/name/` (compatible with older versions).
#'
#' `stop_app()` deregisters one or more running apps.
#' @param name App name (a character string). Use `name = ''` to serve the app
#'   at the root URL `http://host:PORT/` via the proxy. If `name` is non-empty,
#'   the app is served directly by R's internal httpd at `/custom/name/`. Each
#'   name must be unique; calling `new_app()` with an existing name replaces
#'   that app. For `stop_app()`, a character vector of app names to stop;
#'   defaults to all running apps.
#' @param handler A function with signature `function(path, query, post,
#'   headers)` that handles HTTP requests and returns a response list.
#' @param open Whether to open the app URL in a browser, or a function to open
#'   it. In non-interactive sessions this also controls whether the call
#'   blocks: passing `open = FALSE` explicitly suppresses both browser-opening
#'   and blocking.
#' @param host Bind address for the proxy (`"127.0.0.1"` or `"0.0.0.0"`). Only
#'   used when `name = ''`.
#' @param port Candidate proxy ports when `name = ''`. The first available port
#'   is selected (falling back to [random_port()] if needed). Ignored when
#'   `name != ''`.
#' @return `new_app()` returns the app URL invisibly. `stop_app()` returns
#'   nothing.
#' @export
new_app = function(
  name, handler, open = interactive(), host = '127.0.0.1', port = NULL
) {
  backend = .httpd_port()

  # Replace any existing app with the same name.
  if (name %in% names(.proxy$apps)) stop_app(name)

  if (name == '') {
    port = .find_proxy_port(port, names(.proxy$help))
    slot = proxy_start(as.integer(port), as.integer(backend), host = host)
    key = paste0('xfun:', port)
    assign(key, .make_app_handler(paste0('/custom/xfun:', port), handler, getwd()), envir = .httpd_env())
    .proxy$apps[[name]] = list(type = 'proxy', slot = slot, key = key, port = port)
    url_host = if (host == '0.0.0.0') '127.0.0.1' else host
    url = sprintf('http://%s:%d/', url_host, port)
  } else {
    key = as.character(name)
    assign(key, .make_app_handler(paste0('/custom/', name), handler, getwd()), envir = .httpd_env())
    .proxy$apps[[name]] = list(type = 'httpd', key = key)
    url = sprintf('http://127.0.0.1:%d/custom/%s/', backend, name)
  }

  if (isTRUE(open)) open = getOption('viewer', browseURL)
  if (is.function(open)) open(url)

  if (!interactive() && !isFALSE(open)) {
    on.exit(stop_app(name), add = TRUE)
    message('Serving at ', url, ' (press Ctrl+C to stop)')
    tryCatch(
      while (TRUE) Sys.sleep(1),
      interrupt = function(e) invisible(NULL)
    )
  }
  invisible(url)
}

#' @rdname new_app
#' @export
stop_app = function(name = names(.proxy$apps)) {
  if (!missing(name)) name = intersect(name, names(.proxy$apps))
  for (n in name) {
    idx = match(n, names(.proxy$apps))
    app = .proxy$apps[[idx]]
    rm_vars(app$key, .httpd_env())
    .proxy$apps[[idx]] = NULL
    if (app$type == 'proxy') proxy_stop(app$slot)
  }
}

# start an internal passthrough proxy for R's help pages.
help_proxy = function(port = NULL, host = '0.0.0.0') {
  p1 = .httpd_port()
  p2 = .find_proxy_port(port)
  old = .proxy$help[[as.character(p2)]]
  if (!is.null(old)) proxy_stop(old)
  slot = proxy_start(as.integer(p2), as.integer(p1), TRUE, host)
  .proxy$help[[as.character(p2)]] = slot
  u = sprintf('http://127.0.0.1:%d/doc/html/index.html', p2)
  message('Proxy started at ', u)
  invisible(u)
}

#' Find a random available TCP port
#'
#' Find an available TCP port, starting with `port`, then sampling from
#' 3000--8000 (excluding ports known to be blocked by Chrome).
#' @param port Default port to try first.
#' @param n Number of additional random ports to try.
#' @param exclude Integer vector of ports to exclude from the search.
#' @param error Whether to signal an error (default) or return `NULL` when no
#'   port is found.
#' @return An integer port number. When `error = TRUE` (default), signals an
#'   error if no port is found; when `error = FALSE`, returns `NULL`.
#' @export
random_port = function(port = 4321L, n = 20L, exclude = NULL, error = TRUE) {
  # exclude ports considered unsafe by Chrome http://superuser.com/a/188070
  unsafe = c(3659, 4045, 5060, 5061, 6000, 6566, 6665:6669, 6697)
  ports  = sample(setdiff(3000:8000, unsafe), n)
  ports  = setdiff(c(port, ports), exclude)
  for (p in ports) if (.port_available(p)) return(p)
  if (error) stop2("Cannot find an available TCP port")
}

# ---- internal helpers -------------------------------------------------------

# Start R's internal httpd; return its port (or -1 on failure).
.httpd_port = function() {
  if (is.null(getOption('help.ports'))) {
    options(help.ports = 4321 + 1:30); on.exit(options(help.ports = NULL))
  }
  p = suppressMessages(tools::startDynamicHelp(NA))
  if (p <= 0L) stop2("Failed to start R's internal httpd.")
  p
}

# Environment where R's httpd looks up /custom/* handlers.
.httpd_env = function() getFromNamespace('.httpd.handlers.env', 'tools')

# Internal proxy state.
.proxy = new.env(parent = emptyenv())
.proxy$apps = list()  # app name → list(type, key, slot?)
.proxy$help = list()  # help proxy port → slot index

# Find an available proxy port without starting the proxy.
.proxy_app_ports = function() {
  vapply(Filter(function(x) x$type == 'proxy', .proxy$apps), `[[`, integer(1), 'port')
}

.find_proxy_port = function(candidates = NULL, exclude = integer()) {
  if (is.null(candidates)) candidates = 4321 + 1:30
  exclude = unique(c(as.integer(exclude), .proxy_app_ports(), as.integer(names(.proxy$help))))
  candidates = setdiff(candidates, exclude)
  for (p in candidates) if (.port_available(p)) return(as.integer(p))
  if (!is.null(p <- random_port(error = FALSE, exclude = c(candidates, exclude))))
    return(p)
  stop2(
    "No available proxy port found (tried candidates ",
    paste(unique(candidates), collapse = ', '),
    " and random fallback ports)."
  )
}

# Start a proxy instance on `port` forwarding to `backend_port`.
# passthrough = TRUE: forward all request paths verbatim (use this to expose
#   the full R httpd server, e.g. on host = "0.0.0.0" for LAN access).
# passthrough = FALSE (default): rewrite /path to /custom/xfun:PORT/path.
# host: bind address; "0.0.0.0" to listen on all interfaces.
proxy_start = function(port, backend_port, passthrough = FALSE, host = '127.0.0.1') {
  port         = as.integer(port)
  backend_port = as.integer(backend_port)
  passthrough  = isTRUE(passthrough)
  host         = as.character(host)[1]

  bg = Rscript_bg(.proxy_run, list(
    port = port, backend_port = backend_port, passthrough = passthrough, host = host
  ))
  if (!.proxy_wait_ready(port, host)) {
    if (isTRUE(bg$is_alive())) try(proc_kill(bg$pid), silent = TRUE)
    stop2("Failed to start proxy on port ", port, ".")
  }
  paste0('r:', bg$pid)
}

# Stop the proxy instance identified by its slot.
proxy_stop = function(slot) {
  if (is.character(slot) && length(slot) == 1L && startsWith(slot, 'r:')) {
    pid = sub('^r:', '', slot)
    if (grepl('^[0-9]+$', pid)) try_silent(proc_kill(as.integer(pid)))
  }
}

# Coerce body/headers-like argument to a raw vector.
.as_raw = function(x) {
  if (is.null(x) || length(x) == 0L) raw(0) else {
    if (is.raw(x)) x else charToRaw(x[[1L]])
  }
}

# Check if a TCP port is truly available: nothing must be listening on it, and
# we must be able to bind a server socket.  The serverSocket() bind alone is
# insufficient because some servers (e.g. httpuv) use SO_REUSEADDR/SO_REUSEPORT,
# allowing a second bind to succeed even when they already hold the port.
# Probing with a client connection detects any active listener regardless.
.port_available = function(port) {
  port = as.integer(port)
  if (is.na(port)) return(FALSE)
  # If a client can connect, something is already listening → not available.
  con = suppressWarnings(tryCatch(
    socketConnection('127.0.0.1', port = port, open = 'r+b', blocking = FALSE, timeout = 0.5),
    error = function(e) NULL
  ))
  if (!is.null(con)) { try_silent(close(con)); return(FALSE) }
  # Confirm we can actually bind the port as a server.
  server_socket = base::serverSocket
  s = tryCatch(server_socket(port), error = function(e) NULL)
  if (is.null(s)) FALSE else { close(s); TRUE }
}

# Build an R httpd handler closure.
.make_app_handler = function(prefix, fn, wd) {
  function(path, query = NULL, body = NULL, headers = NULL) {
    real = if (startsWith(path, prefix)) substring(path, nchar(prefix) + 1L) else path
    real = sub('^/', '', real)
    if (real == '') real = '.'
    q = if (is.null(query)) .parse_xfun_query(headers) else query
    in_dir(wd, fn(real, q, .as_raw(body), .as_raw(headers)))
  }
}

.stop_help_proxy = function(port = names(.proxy$help)) {
  if (!missing(port)) port = intersect(as.character(port), names(.proxy$help))
  for (p in port) {
    proxy_stop(.proxy$help[[p]])
    .proxy$help[[p]] = NULL
  }
}

.proxy_connect_host = function(host) {
  host = as.character(host)[1]
  # server may bind all interfaces (0.0.0.0), but local readiness checks should
  # use loopback for a deterministic self-connection target.
  if (host == '0.0.0.0') '127.0.0.1' else host
}

# Poll until the proxy port is accepting connections (or give up after timeout).
.proxy_wait_ready = function(port, host, tries = 100L, delay = 0.1) {
  host = .proxy_connect_host(host)
  for (i in seq_len(tries)) {
    con = tryCatch(
      socketConnection(host = host, port = port, open = 'r+b', blocking = TRUE, timeout = 1),
      error = function(e) NULL
    )
    if (!is.null(con)) {
      close(con)
      return(TRUE)
    }
    Sys.sleep(delay)
  }
  FALSE
}

# Main proxy loop running in a background R process.
.proxy_run = function(port, backend_port, passthrough = FALSE, host = '127.0.0.1') {
  server_socket = base::serverSocket
  socket_accept = base::socketAccept

  listener = server_socket(port)
  on.exit(close(listener), add = TRUE)

  repeat {
    con = tryCatch(
      socket_accept(listener, blocking = FALSE, open = 'r+b', timeout = 1),
      error = function(e) NULL
    )
    if (is.null(con)) next
    try_silent(.proxy_forward(con, port, backend_port, passthrough))
    try_silent(close(con))
  }
}

# Read a client request, forward it to backend httpd, and relay the response.
.proxy_forward = function(client, port, backend_port, passthrough = FALSE) {
  req = .proxy_read_request(client)
  if (is.null(req)) return()

  path  = req$path
  path2 = if (isTRUE(passthrough)) path else sprintf('/custom/xfun:%d%s', port, path)
  q_pos = regexpr('\\?', path)[1L]
  query = if (q_pos > 0L) substring(path, q_pos + 1L) else ''

  backend = tryCatch(
    socketConnection(host = '127.0.0.1', port = backend_port, open = 'r+b', blocking = FALSE, timeout = 5),
    error = function(e) NULL
  )
  if (is.null(backend)) return()
  on.exit(try(close(backend), silent = TRUE), add = TRUE)

  headers = req$headers
  if (length(headers)) {
    keep = !grepl('^(connection|keep-alive|proxy-connection):', headers, ignore.case = TRUE)
    headers = headers[keep]
  }
  if (nzchar(query)) headers = c(headers, paste0('X-Xfun-Query: ', query))
  txt = paste(c(sprintf('%s %s HTTP/1.0', req$method, path2), headers, '', ''), collapse = '\r\n')
  writeBin(charToRaw(txt), backend)
  if (length(req$body)) writeBin(req$body, backend)
  flush(backend)

  buf = raw(0); clen = NA_integer_; sep_end = NA_integer_
  for (i in seq_len(200L)) {
    Sys.sleep(0.05)
    ready = tryCatch(socketSelect(list(backend), write = FALSE, timeout = 0), error = function(e) FALSE)
    if (!isTRUE(ready[1L])) next
    x = tryCatch(readBin(backend, raw(), 65536L), error = function(e) raw(0))
    if (!length(x)) break
    buf = c(buf, x)
    if (is.na(sep_end)) {
      s = tryCatch(rawToChar(buf), error = function(e) '')
      m = regexpr('\r\n\r\n|\n\n', s, perl = TRUE)
      if (m[1L] > 0L) {
        sep_end = m[1L] + attr(m, 'match.length') - 1L
        cl_m = regexpr('(?i)content-length:[ \t]*([0-9]+)', s, perl = TRUE)
        if (cl_m[1L] > 0L)
          clen = as.integer(trimws(sub('(?i).*:', '', regmatches(s, cl_m), perl = TRUE)))
      }
    }
    if (!is.na(sep_end) && !is.na(clen) && (length(buf) - sep_end) >= clen) break
  }
  if (length(buf)) { writeBin(buf, client); flush(client) }
}

# Parse one HTTP request from a non-blocking socket connection.
# Uses socketSelect() polling so readBin() returns immediately with whatever
# bytes are available without blocking for the full buffer size.
.proxy_read_request = function(con) {
  buf = raw(0); head_end = 0L; sep_len = 0L
  for (i in seq_len(200L)) {  # up to 200 * 0.05 s = 10 s
    Sys.sleep(0.05)
    ready = tryCatch(socketSelect(list(con), write = FALSE, timeout = 0), error = function(e) FALSE)
    if (!isTRUE(ready[1L])) next
    x = tryCatch(readBin(con, raw(), 65536L), error = function(e) raw(0))
    if (!length(x)) return(NULL)
    buf = c(buf, x)
    s = tryCatch(rawToChar(buf), error = function(e) '')
    m = regexpr('\r\n\r\n|\n\n', s, perl = TRUE)
    if (m[1L] > 0L) { head_end = m[1L]; sep_len = attr(m, 'match.length'); break }
    if (length(buf) >= 65536L) return(NULL)
  }
  if (head_end == 0L) return(NULL)

  hs = substring(rawToChar(buf), 1L, head_end + sep_len - 1L)
  hs = sub('\r\n\r\n$|\n\n$', '', hs, perl = TRUE)
  lines = strsplit(hs, '\r\n|\n', perl = TRUE)[[1L]]
  if (!length(lines)) return(NULL)

  req = strsplit(lines[[1L]], ' ', fixed = TRUE)[[1L]]
  req = req[nzchar(req)]
  if (length(req) < 2L) return(NULL)
  method = req[[1L]]; path = req[[2L]]; headers = lines[-1L]

  clen = 0L
  if (length(headers)) {
    i = grep('^content-length:', headers, ignore.case = TRUE)
    if (length(i)) {
      v = trimws(sub('^[^:]*:', '', headers[[i[1L]]]))
      if (grepl('^[0-9]+$', v)) clen = as.integer(v)
    }
    if (is.na(clen) || clen < 0L) clen = 0L
  }
  body_start = head_end + sep_len
  body0 = if (body_start <= length(buf)) buf[body_start:length(buf)] else raw(0)
  body = if (length(body0) < clen)
    c(body0, tryCatch(readBin(con, raw(), clen - length(body0)), error = function(e) raw(0)))
  else body0[seq_len(clen)]
  list(method = method, path = path, headers = headers, body = body)
}

# Parse the X-Xfun-Query header (added by the proxy) into a named, URL-decoded
# character vector. Falls back to character(0) when absent.
.parse_xfun_query = function(headers) {
  if (is.null(headers) || length(headers) == 0L) return(character(0))
  hs = if (is.raw(headers)) rawToChar(headers) else paste(headers, collapse = '\n')
  # Find the X-Xfun-Query line.
  qs = ''
  for (line in strsplit(hs, '\n', fixed = TRUE)[[1L]]) {
    line = sub('\r$', '', line)
    if (startsWith(line, 'X-Xfun-Query: ')) {
      qs = substring(line, nchar('X-Xfun-Query: ') + 1L)
      break
    }
  }
  if (qs == '') return(character(0))
  pairs = Filter(nzchar, strsplit(qs, '&', fixed = TRUE)[[1L]])
  if (!length(pairs)) return(character(0))
  keys = vals = character(length(pairs))
  for (i in seq_along(pairs)) {
    kv = strsplit(pairs[[i]], '=', fixed = TRUE)[[1L]]
    keys[[i]] = utils::URLdecode(kv[[1L]])
    vals[[i]] = if (length(kv) >= 2L)
      utils::URLdecode(paste(kv[-1L], collapse = '=')) else ''
  }
  setNames(vals, keys)
}

#' Get data from a REST API
#'
#' Read data from a REST API and optionally with an authorization token in the
#' request header. The function `rest_api_raw()` returns the raw text of
#' the response, and `rest_api()` will parse the response with
#' `jsonlite::fromJSON()` (assuming that the response is in the JSON
#' format).
#'
#' These functions are simple wrappers based on [url()] and
#' [read_utf8()]. Specifically, the `headers` argument is
#' passed to `url()`, and `read_utf8()` will send a \samp{GET} request
#' to the API server. This means these functions only support the \samp{GET}
#' method. If you need to use other HTTP methods (such as \samp{POST}), you have
#' to use other packages such as \pkg{curl} and \pkg{httr}.
#' @param ... Arguments to be passed to `rest_api_raw()`.
#' @return A character vector (the raw JSON response) or an R object parsed from
#'   the JSON text.
#' @export
#' @examplesIf interactive()
#' # a normal GET request
#' xfun::rest_api('https://mockhttp.org', '/get')
#' xfun::rest_api_raw('https://mockhttp.org', '/get')
#'
#' # send the request with an auth header
#' xfun::rest_api('https://mockhttp.org', '/headers', 'OPEN SESAME!')
#'
#' # with query parameters
#' xfun::rest_api('https://mockhttp.org', '/response-headers', params = list(foo = 'bar'))
#'
#' # get the rate limit info from GitHub
#' xfun::github_api('/rate_limit')
rest_api = function(...) {
  res = rest_api_raw(...)
  jsonlite::fromJSON(res, simplifyVector = FALSE)
}

#' @param root The API root URL.
#' @param endpoint The API endpoint.
#' @param token A named character string (e.g., `c(token = "xxxx")`), which
#'   will be used to create an authorization header of the form
#'   \samp{Authorization: NAME TOKEN} for the API call, where \samp{NAME} is the
#'   name of the string and \samp{TOKEN} is the string. If the string does not
#'   have a name, \samp{Basic} will be used as the default name.
#' @param params A list of query parameters to be sent with the API call.
#' @param headers A named character vector of HTTP headers, e.g., `c(Accept
#'   = "application/vnd.github.v3+json")`.
#' @rdname rest_api
#' @export
rest_api_raw = function(root, endpoint, token = '', params = list(), headers = NULL)  {
  if (is.null(names(token))) names(token) = 'Basic'
  endpoint = sub('^/?', '/', endpoint)  # make sure it has a leading /
  url2 = if ('headers' %in% names(formals(url))) url else stop2(
    "The url() function does not support the 'headers' argument. Please upgrade R (>= 3.6.0)."
  )
  con = url2(
    paste0(root, endpoint, query_params(.list = params)), encoding = 'UTF-8',
    headers = c(
      headers, if (token != '') c(Authorization = sprintf('%s %s', names(token), token))
    )
  )
  on.exit(close(con), add = TRUE)
  raw_string(suppressWarnings(read_utf8(con)))
}

library(testit)

assert("tojson() works", {
  (.tojson(NULL) %==% "null")
  (.tojson(list()) %==% "{}")
  (.tojson(NA) %==% 'null')
  (.tojson(NA_character_) %==% 'null')
  (.tojson(1:10) %==% "[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]")
  (.tojson(TRUE) %==% "true")
  (.tojson(FALSE) %==% "false")

  x = list(a = 1, b = list(c = 1:3, d = "abc"))
  out = '{\n  "a": 1,\n  "b": {\n    "c": [1, 2, 3],\n    "d": "abc"\n  }\n}'
  (.tojson(x) %==% out)

  x = list(c("a", "b"), 1:5, TRUE)
  out = '[\n  ["a", "b"],\n  [1, 2, 3, 4, 5],\n  true\n]'
  (.tojson(x) %==% out)

  (.tojson(list('"a b"' = 'quotes "\'')) %==% '{\n  "\\"a b\\"": "quotes \\"\'"\n}')

  # data frames
  (.tojson(data.frame()) %==% '{}')
  # by column for named data frames
  x = data.frame(a = 1:3, b = c('x', 'y', 'z'))
  (.tojson(x) %==% '{\n  "a": [1, 2, 3],\n  "b": ["x", "y", "z"]\n}')
  # length 1 columns must be arrays, too
  x = data.frame(a = 1, b = 'x')
  (.tojson(x) %==% '{\n  "a": [1],\n  "b": ["x"]\n}')
  # by row for unnamed data frames
  x = unname(data.frame(a = 1:3, b = c('x', 'y', 'z')))
  (.tojson(x) %==% '[\n  [1, "x"],\n  [2, "y"],\n  [3, "z"]\n]')

  x = list(a = 1:5, b = js("function() {return true;}"))
  out = '{\n  "a": [1, 2, 3, 4, 5],\n  "b": function() {return true;}\n}'
  (.tojson(x) %==% out)

  res = tojson(list(NULL, 1:10, TRUE, FALSE))
  # passing a json object through tojson() should return it unchanged
  (tojson(res) %==% res)
})

assert('json_vector() converts atomic vectors to JSON', {
  (json_vector(c('a', 'b'), to_array = TRUE) %==% '["a", "b"]')
  (json_vector(c('a', 'b'), to_array = FALSE) %==% c('"a"', '"b"'))
  (json_vector(1:3, to_array = TRUE, quote = FALSE) %==% '[1, 2, 3]')
  # NA becomes null
  (json_vector(c('a', NA_character_), to_array = TRUE) %==% '["a", null]')
})

assert('json_atomic() handles Date, POSIXct, and factor', {
  d = as.Date('2024-01-15')
  (json_atomic(d) %==% 'new Date("2024-01-15")')
  f = factor(c('a', 'b', 'a'))
  (json_atomic(f) %==% '["a", "b", "a"]')
})

assert('.tojson() handles arrays', {
  m = matrix(1:4, nrow = 2)
  out = .tojson(m)
  (grepl('\\[', out))  # should produce nested arrays
})

#' Encode/decode data into/from base64 encoding.
#'
#' The function `base64_encode()` encodes a file or a raw vector into the
#' base64 encoding. The function `base64_decode()` decodes data from the
#' base64 encoding.
#' @param x For `base64_encode()`, a raw vector. If not raw, it is assumed
#'   to be a file or a connection to be read via `readBin()`. For
#'   `base64_decode()`, a string.
#' @param from If provided (and `x` is not provided), a connection or file
#'   to be read via `readChar()`, and the result will be passed to the
#'   argument `x`.
#' @return `base64_encode()` returns a character string.
#'   `base64_decode()` returns a raw vector.
#' @useDynLib xfun, .registration = TRUE
#' @export
#' @examples xfun::base64_encode(as.raw(1:10))
#' logo = xfun:::R_logo()
#' xfun::base64_encode(logo)
base64_encode = function(x) {
  if (!is.raw(x)) x = read_bin(x)
  .Call('base64_enc', x, PACKAGE = 'xfun')
}

#' @export
#' @rdname base64_encode
#' @examples xfun::base64_decode("AQIDBAUGBwgJCg==")
base64_decode = function(x, from = NA) {
  if (!is.na(from)) {
    if (!missing(x)) stop("Please provide either 'x' or 'from', but not both.")
    x = readChar(from, file.size(from), TRUE)
  }
  if (!is.character(x) || length(x) != 1) stop("'x' must be a single character string.")
  .Call('base64_dec', x, PACKAGE = 'xfun')
}

# an R implementation of base64 encoding by Wush Wu moved from knitr (of
# historic interest only): https://github.com/yihui/knitr/pull/324
base64_encode_r = function(x) {
  if (!is.raw(x)) x = read_bin(x)
  chars = c(LETTERS, letters, 0:9, '+', '/')
  n = length(s <- as.integer(x))
  res = rep(NA, (n + 2) / 3 * 4)
  i = 0L  # index of res vector
  j = 1L  # index of base64_table
  while (n > 2L) {
    res[i <- i + 1L] = chars[s[j] %/% 4L + 1L]
    res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
    res[i <- i + 1L] = chars[4L * (s[j + 1L] %% 16) + s[j + 2L] %/% 64L + 1L]
    res[i <- i + 1L] = chars[s[j + 2L] %% 64L + 1L]
    j = j + 3L
    n = n - 3L
  }
  if (n) {
    res[i <- i + 1L] = chars[s[j] %/% 4L + 1L]
    if (n > 1L) {
      res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
      res[i <- i + 1L] = chars[4L * (s[j + 1L] %% 16) + 1L]
      res[i <- i + 1L] = '='
    } else {
      res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + 1L]
      res[i <- i + 1L] = '='
      res[i <- i + 1L] = '='
    }
  }
  paste(res[!is.na(res)], collapse = '')
}

#' Generate the Data URI for a file
#'
#' Encode the file in the base64 encoding, and add the media type. The data URI
#' can be used to embed data in HTML documents, e.g., in the `src` attribute of
#' the `<img />` tag.
#' @param x A file path.
#' @param type The MIME type of the file, e.g., `"image/png"` for a PNG image
#'   file.
#' @return A string of the form `data:<media type>;base64,<data>`.
#' @export
#' @examples
#' logo = xfun:::R_logo()
#' img = xfun::html_tag('img', src = xfun::base64_uri(logo), alt = "R logo")
#' if (interactive()) xfun::html_view(img)
base64_uri = function(x, type = mime_type(x)) {
  paste0("data:", type, ";base64,", base64_encode(x))
}

#' Get the MIME types of files
#'
#' If the \pkg{mime} package is installed, call [mime::guess_type()], otherwise
#' use the system command `file --mime-type` to obtain the MIME type of a file.
#' Typically, the `file` command exists on *nix. On Windows, the command should
#' exist if Cygwin or Rtools is installed. If it is not found, .NET's
#' `MimeMapping` class will be used instead (which requires the .NET framework).
#' @param x A vector of file paths.
#' @param use_mime Whether to use the \pkg{mime} package.
#' @param empty The MIME type for files without extensions (e.g., `Makefile`).
#'   If `NA`, the type will be obtained from system command. This argument is
#'   used only for `use_mime = FALSE`.
#' @return A character vector of MIME types.
#' @note When querying the MIME type via the system command, the result will be
#'   cached to `xfun:::cache_dir()`. This will make future queries much faster,
#'   since running the command in real time can be a little slow.
#' @export
#' @examplesIf tolower(Sys.getenv('CI')) == 'true'
#' f = list.files(R.home('doc'), full.names = TRUE)
#' mime_type(f)
#' mime_type(f, FALSE)  # don't use mime
#' mime_type(f, FALSE, NA)  # run command for files without extension
mime_type = function(x, use_mime = loadable('mime'), empty = 'text/plain') {
  if (use_mime) return(mime::guess_type(x))
  ext = tolower(file_ext(x))
  res = character(n <- length(x))
  # try to read from cache since running command in real time can be slow
  p = file.path(cache_dir(), 'mime.rds')
  db = mimemap
  for (i in seq_len(n)) {
    e = ext[i]
    m = if (e == '') empty else .mime_type(x[i], e)
    if (is.na(m) && is.na(m <- db[e])) {
      if (file_exists(p)) db = readRDS(p)
      m = db[e]
    }
    if (is.na(m)) {
      m = .mime_cmd(x[i])
      if (grepl('^[a-zA-Z]+/[-a-zA-Z0-9.+]+$', m)) {
        if (e != '') {
          db[e] = m; if (dir_create(dirname(p))) saveRDS(db, p)
        }
      } else warning("'", m, "' does not appear to be a valid MIME type for '", x[i], "'")
    }
    res[i] = m
  }
  res
}

# try to get the MIME type from tools:::mime_type()
.mime_type = function(x, ext = file_ext(x)) {
  if (ext != '' && is.function(f <- asNamespace('tools')$mime_type)) {
    m = f(x, ext)
    if (m != 'text/plain') return(m)
  }
  NA
}

# MIME type from command line `powershell` or `file`
.mime_cmd = function(x, use_file = !is_windows() || Sys.which('file') != '') {
  if (!file.exists(x <- path.expand(x))) stop("The file '", x, "' does not exist.")
  if (use_file) {
    cmd = 'file'
    arg = c('--mime-type', '-b')
  } else {
    # IT said "thou shall not include .ps1 in source package", so I use .txt;
    # then Windoz said "thou shall not execute .txt", so I copy .txt to .ps1;
    # How much of our life has been wasted on making these idiots happy...
    ps1 = tempfile(fileext = '.ps1'); on.exit(unlink(ps1), add = TRUE)
    file.copy(pkg_file('scripts', 'mime-type.txt'), ps1)
    cmd = 'powershell'
    arg = c('-ExecutionPolicy', 'Bypass', '-File', shQuote(ps1))
  }
  if (Sys.which(cmd) == '') stop("The '", cmd, "' command is not found")
  system2(cmd, c(arg, shQuote(x)), stdout = TRUE)[1]
}

# a comprehensive version is mime::mimemap
mimemap = c(
  css = 'text/css', csv = 'text/csv', gif = 'image/gif', jpeg = 'image/jpeg',
  jpg = 'image/jpeg', js = 'text/javascript', png = 'image/png',
  r = 'text/plain', rmd = 'text/x-markdown', svg = 'image/svg+xml',
  ttf = 'font/ttf', woff = 'font/woff', woff2 = 'font/woff2'
)

# TODO: remove this after new release of litedown
guess_type = function(x) mime_type(x)

#' Run R code and record the results
#'
#' Run R code and capture various types of output, including text output, plots,
#' messages, warnings, and errors.
#' @param code A character vector of R source code.
#' @param dev A graphics device. It can be a function name, a function, or a
#'   character string that can be evaluated to a function to open a graphics
#'   device.
#' @param dev.path A base file path for plots. Actual plot filenames will be
#'   this base path plus incremental suffixes. For example, if `dev.path =
#'   "foo"`, the plot files will be `foo-1.png`, `foo-2.png`, and so on. If
#'   `dev.path` is not character (e.g., `FALSE`), plots will not be recorded.
#' @param dev.ext The file extension for plot files. By default, it will be
#'   inferred from the first argument of the device function if possible.
#' @param dev.args Extra arguments to be passed to the device. The default
#'   arguments are `list(units = 'in', onefile = FALSE, width = 7, height = 7,
#'   res = 96)`. If any of these arguments is not present in the device
#'   function, it will be dropped.
#' @param dev.keep Indices of plots to be kept. The indices can be either
#'   numeric (positive or negative integers) or logical. Negative integers and
#'   false values will remove the corresponding plots. For example, if the code
#'   generated 3 plots, `dev.keep = c(1, 3)`, `-2`, and `c(T, F, T)` are
#'   equivalent ways to remove the second plot. The special value `dev.keep =
#'   'last'` means to keep the last plot only.
#' @param message,warning,error If `TRUE`, record and store messages / warnings
#'   / errors in the output. If `FALSE`, suppress them. If `NA`, do not process
#'   them (messages will be emitted to the console, and errors will halt the
#'   execution).
#' @param cache A list of options for caching. See the `path`, `id`, and `...`
#'   arguments of [cache_exec()].
#' @param print A (typically S3) function that takes the value of an expression
#'   in the code as input and returns output. The default is [record_print()].
#'   If a non-function value (e.g., `NA`) is passed to this argument, [print()]
#'   (or [show()] for S4 objects) will be used.
#' @param print.args A list of arguments for the `print` function. By default,
#'   the whole list is not passed directly to the function, but only an element
#'   in the list with a name identical to the first class name of the returned
#'   value of the expression, e.g., `list(data.frame = list(digits = 3), matrix
#'   = list())`. This makes it possible to apply different print arguments to
#'   objects of different classes. If the whole list is intended to be passed to
#'   the print function directly, wrap the list in [I()].
#' @param verbose `2` means to always print the value of each expression in the
#'   code, no matter if the value is [invisible()] or not; `1` means to always
#'   print the value of the last expression; `0` means no special handling
#'   (i.e., print only when the value is visible).
#' @param envir An environment in which the code is evaluated.
#' @return `record()` returns a list of the class `xfun_record_results` that
#'   contains elements with these possible classes: `record_source` (source
#'   code), `record_output` (text output), `record_plot` (plot file paths),
#'   `record_message` (messages), `record_warning` (warnings), and
#'   `record_error` (errors, only when the argument `error = TRUE`).
#' @import grDevices
#' @export
#' @examples
#' code = c('# a warning test', '1:2 + 1:3', 'par(mar = c(4, 4, 1, .2))', 'barplot(5:1, col = 2:6, horiz = TRUE)', 'head(iris)', "sunflowerplot(iris[, 3:4], seg.col = 'purple')", "if (TRUE) {\n  message('Hello, xfun::record()!')\n}", '# throw an error', "1 + 'a'")
#' res = xfun::record(code, dev.args = list(width = 9, height = 6.75), error = TRUE)
#' xfun::tree(res)
#' format(res)
#' # find and clean up plot files
#' plots = Filter(function(x) inherits(x, 'record_plot'), res)
#' file.remove(unlist(plots))
record = function(
  code = NULL, dev = 'png', dev.path = 'xfun-record', dev.ext = dev_ext(dev),
  dev.args = list(), dev.keep = TRUE, message = TRUE, warning = TRUE, error = NA,
  cache = list(), print = record_print, print.args = list(),
  verbose = getOption('xfun.record.verbose', 0), envir = parent.frame()
) {
  new_result = function(x = list()) structure(x, class = 'xfun_record_results')
  res = new_result()
  if (length(code) == 0) return(res)

  # use cached results if cache exists
  use_cache = FALSE
  ret = cache_code(code, envir, use_cache <- TRUE, cache)
  if (use_cache) return(ret)

  code = split_lines(code)

  add_result = function(x, type = NULL, pos = length(res), insert = TRUE) {
    # default type is output
    if (is.null(type) && !inherits(x, .record_classes)) type = 'output'
    # insert a whole element or append to an existing element in res
    if (!insert) x = c(res[[pos]], x)
    el = if (is.null(type)) x else new_record(x, type)
    N = length(res)
    if (insert) {
      if (N == pos) res[[N + 1]] <<- el else res <<- append(res, el, pos)
    } else {
      res[[pos]] <<- el
    }
  }

  # dev.off() will set the current device to next instead of previous device,
  # but we often want to restore to the previous instead
  dev_cur = dev.cur()
  dev_reset = function() if (dev_cur != 1 && dev_cur %in% dev.list()) dev.set(dev_cur)

  # look for all possible ${dev.path}-%d.${dev.ext} files created by the device
  get_plots = function() {
    files = sprintf(dev.path, seq_along(list.files(dirname(dev.path))))
    files[file_exists(files)]
  }
  # open a graphics device
  dev_num = if (is.character(dev.path)) {
    if (!is.function(dev)) dev = tryCatch(match.fun(dev), error = function(e) {
      eval(parse(text = dev), envir = envir)
    })
    # normalize \ to / and remove the leading ./
    if (is_windows()) dev.path = gsubf('\\', '/', dev.path)
    dev.path = sub('^[.]/+', '', dev.path)
    # add extension
    dev.path = with_ext(paste0(dev.path, if (!grepl('/$', dev.path)) '-', '%d'), dev.ext)
    dir_create(dirname(dev.path))
    # clean up existing plots before opening the device
    if (any(i <- !file.remove(old_plots <- get_plots()))) stop(
      'Failed to delete existing plot file(s): ',
      paste("'", old_plots[i], "'", collapse = ', ')
    )
    dev_open(dev, dev.path, dev.args)
  }
  dev_old = dev.list()  # keep track of current devices
  dev_off = function() {
    if (length(dev_num) && dev_num %in% dev.list()) dev.off(dev_num)
    dev_reset()
    dev_num <<- NULL  # prevent dev.off() again by accident
  }
  on.exit(dev_off(), add = TRUE)

  # check if new plots are generated
  handle_plot = local({
    # don't record plots if no device was opened
    if (length(dev_num) == 0) return(function(...) {})
    old_files = character(); old_times = NULL  # previous plot files and file mtimes
    old_plot = recordPlot()
    function(last = FALSE) {
      # if dev.list() has changed, no longer record graphics, except for the last plot
      if (!last) {
        if (!(length(dev_num) && identical(dev.list(), dev_old))) return()
        dev.set(dev_num)
      }
      files = get_plots()
      # on Windows, an empty plot file is created upon opening a device
      files = files[file.size(files) > 0]

      if (!last) {
        new_plot = recordPlot()
        if (!identical(old_plot, new_plot)) {
          # add a placeholder for new plots, which may not have been created by
          # the device until the next new plot is drawn
          add_result(character(), 'plot')
          old_plot <<- new_plot
        }
      }

      if (length(files) == 0) return()

      # old files that have been modified should be removed from existing records
      k = file.mtime(old_files) > old_times
      if (any(k)) for (i in seq_along(res)) {
        if (inherits(res[[i]], 'record_plot'))
          res[[i]] <<- new_record(setdiff(res[[i]], old_files[k]), 'plot')
      }
      old_files = old_files[!k]

      plots = setdiff(files, old_files)
      old_files <<- files
      old_times <<- file.mtime(files)
      if ((n <- length(plots)) == 0) return()
      # if an empty plot is recorded, remove it (e.g., yihui/litedown#96)
      if (n == 1 && empty_plot(old_plot)) {
        file.remove(plots); old_files <<- character(); old_times <<- NULL
        return()
      }

      # indices of plots in results
      i = which(vapply(res, inherits, logical(1), 'record_plot'))
      N = length(i)
      # the last plot should always be appended to the last plot block
      if (last) {
        # N = 0 means the code didn't produce any plots
        if (N == 0) file.remove(plots) else add_result(plots, 'plot', i[N], FALSE)
        return()
      }

      # for the newly generated plots, append the first one (which should be the
      # last plot of a previous code chunk, since it will be created only when
      # the next code chunk produces a new plot) to the previous plot block; the
      # rest are for the current block
      if (N > 1) {
        add_result(plots[1], 'plot', i[N - 1], FALSE)
        if (n > 1) add_result(plots[2:n], 'plot', i[N], FALSE)
      } else if (N == 1) {
        # if there exists only one plot block, add plots to that block
        add_result(plots, 'plot', i[N], FALSE)
      }
    }
  })

  handle = if (is.na(error)) identity else try_silent
  # split code into individual expressions
  codes = handle(split_source(code, merge_comments = TRUE, line_number = TRUE))
  # code may contain syntax errors
  if (is_error(codes)) {
    add_result(code, 'source'); add_result(attr2(codes, 'condition')$message, 'error')
    return(new_result(res))
  }

  # trim error message and remove the uninformative part (yihui/litedown#109)
  trim_error = function(e) {
    x = sub('\\s+$', '', as.character(e))
    sub(' in eval\\(expr, envir(, enclos)?\\):', ':', x)
  }
  handle_message = function(type, add = TRUE) {
    err = type == 'error'
    mf = if (!err) sub('^(.)', 'muffle\\U\\1', type, perl = TRUE)
    function(e) {
      if (is.na(add)) return()
      if (isTRUE(add)) add_result(if (err) trim_error(e) else e$message, type)
      if (!err) invokeRestart(mf)
    }
  }
  handle_m = handle_message('message', message)
  handle_w = handle_message('warning', warning)
  handle_e = handle_message('error', error)

  # don't use withCallingHandlers() if message/warning/error are all NA
  handle_eval = function(expr) {
    handle(if (is.na(message) && is.na(warning) && is.na(error)) expr else {
      withCallingHandlers(
        expr, message = handle_m, warning = handle_w, error = handle_e
      )
    })
  }
  # a simplified version of capture.output()
  handle_output = function(expr) {
    out = NULL
    con = textConnection('out', 'w', local = TRUE)
    on.exit(close(con))
    # capture try() messages
    opts = if (is.null(getOption('try.outFile'))) options(try.outFile = con)
    sink(con); on.exit({ sink(); close(con); options(opts) })
    expr  # lazy evaluation
    on.exit()  # if no error occurred, clear up previous on-exit calls
    sink(); close(con); options(opts)
    if (length(out)) add_result(out, 'output')
    expr
  }
  # disable ANSI sequences in text output
  opts = options(cli.num_colors = 1); on.exit(options(opts), add = TRUE)
  n = length(codes)
  for (i in seq_len(n)) {
    add_result(code <- codes[[i]], 'source')
    expr = parse_only(code)
    if (length(expr) == 0) next
    # evaluate the code and capture output
    out = handle_output(handle_eval(withVisible(eval(expr, envir))))
    # verbose = 1: always print the last value; verbose = 2: print all values
    if (verbose == 2 || (verbose == 1 && i == n)) {
      # make invisible values visible unless they are NULL (which is often not useful)
      if (!is_error(out) && !out$visible && !is.null(out$value)) out$visible = TRUE
    }
    # print value (via record_print()) if visible
    if (!is_error(out) && out$visible) {
      val = out$value
      # if val is data.table, check if it needs to be printed
      if (inherits(val, 'data.table') && 'data.table' %in% loadedNamespaces() &&
          !data.table::shouldPrint(val)) next
      if (is.null(print)) print = record_print
      if (!is.function(print)) print = record_print.default
      p_args = print.args
      # index print args by first class of out unless args are wrapped in I()
      if (!inherits(p_args, 'AsIs')) p_args = p_args[[class(val)[1]]]
      out = handle_eval(if (length(p_args) == 0) print(val) else do.call(
        print, c(list(val), p_args), quote = inherits(val, c('name', 'call'))
      ))
      if (length(out) && !is_error(out)) {
        if (is.list(out)) lapply(out, add_result) else add_result(out)
      }
    }
    handle_plot()
  }
  # shut off the device to write out the last plot if there exists one
  dev_off()
  handle_plot(TRUE)

  # filter plots
  if (!isTRUE(dev.keep)) {
    p = get_plots()  # all plots
    if (identical(dev.keep, 'last')) dev.keep = length(p)
    d = setdiff(p, p[dev.keep])  # plots to be deleted
    if (length(d)) {
      for (i in seq_along(res)) {
        if (!inherits(r <- res[[i]], 'record_plot') || all(is.na(k <- match(d, r)))) next
        r2 = r[-k[!is.na(k)]]  # remove plots matching d
        attributes(r2) = attributes(r)  # preserve class
        res[[i]] = r2
      }
      file.remove(d)
    }
  }

  # remove empty blocks
  res = Filter(length, res)
  res = merge_record(res)

  new_result(res)
}

# merge neighbor elements of the same class
merge_record = function(x) {
  n = length(x)
  if (n <= 1) return(x)
  k = NULL
  for (i in 2:n) {
    r1 = x[[i - 1]]; c1 = class(r1); r2 = x[[i]]; c2 = class(r2)
    if (!identical(c1, c2)) next
    # concatenate messages to a single string (consider message(appendLF = FALSE))
    x[[i]] = if ('record_message' %in% c1) paste(c(r1, r2), collapse = '') else c(r1, r2)
    attributes(x[[i]]) = attributes(r1)
    k = c(k, i - 1)
  }
  if (length(k)) x = x[-k]
  x
}

#' Print methods for `record()`
#'
#' An S3 generic function to be called to print visible values in code when the
#' code is recorded by [record()]. It is similar to [knitr::knit_print()]. By
#' default, it captures the normal [print()] output and returns the result as a
#' character vector. The `knitr_kable` method is for printing [knitr::kable()]
#' output. Users and package authors can define other S3 methods to extend this
#' function.
#' @param x For `record_print()`, the value to be printed. For `new_record()`, a
#'   character vector to be included in the printed results.
#' @param ... Other arguments to be passed to `record_print()` methods.
#' @return A `record_print()` method should return a character vector or a list
#'   of character vectors. The original classes of the vector will be discarded,
#'   and the vector will be treated as console output by default (i.e.,
#'   `new_record(class = "output")`). If it should be another type of output,
#'   wrap the vector in [new_record()] and specify a class name.
#' @export
record_print = function(x, ...) {
  UseMethod('record_print')
}

#' @rdname record_print
#' @export
record_print.default = function(x, ...) {
  # the default print method is just print()/show()
  capture.output(if (isS4(x)) methods::show(x, ...) else print(x, ...))
}

#' @rdname record_print
#' @export
record_print.record_asis = function(x, ...) {
  x
}

#' @param class A class name. Possible values are `xfun:::.record_cls`.
#' @rdname record_print
#' @export
new_record = function(x, class) structure(x, class = paste0('record_', class))

# all possible classes for record() results at the moment
.record_cls = c('source', 'output', 'message', 'warning', 'error', 'plot', 'asis')
.record_classes = paste0('record_', .record_cls)

# default device arguments
dev_args = list(units = 'in', onefile = FALSE, width = 8, height = 8, res = 84)

dev_open = function(dev, file, args) {
  m = names(formals(dev))
  for (i in names(dev_args)) {
    if (i %in% m && is.null(args[[i]])) args[[i]] = dev_args[[i]]
  }
  n = length(dev.list())
  do.call(dev, c(list(file), args))
  if (length(dev.list()) <= n) stop('Failed to open device for recording plots.')
  dev.control('enable')
  dev.cur()
}

# infer the filename extension from a device's first argument
dev_ext = function(dev) {
  # the first argument could be a string or an expression
  if (!is.character(name <- formals(dev)[[1]])) name = gsub('"', '', deparse(name))
  file_ext(name)
}

# test if a recorded plot is empty
empty_plot = function(p) {
  xs = lapply(p[[1]], function(x) x[[2]][[1]])
  for (x in xs) {
    if ('name' %in% names(x)) {
      # base graphics
      if (!x$name %in% empty_calls) return(FALSE)
    } else if (is.call(x)) {
      # grid graphics
      if (!identical(as.character(x[[1]]), 'requireNamespace')) return(FALSE)
    }
  }
  TRUE
}

empty_calls = c(
  'C_clip', 'C_layout', 'C_par', 'C_plot_window', 'C_strHeight', 'C_strWidth',
  'palette', 'palette2'
)

is_error = function(x) inherits(x, 'try-error')

#' @param to The output format (text, markdown, or html).
#' @param encode For HTML output, whether to base64 encode plots.
#' @param template For HTML output, whether to embed the formatted results in an
#'   HTML template. Alternatively, this argument can take a file path, i.e.,
#'   path to an HTML template that contains the variable `$body$`. If `TRUE`,
#'   the default template in this package will be used
#'   (`xfun:::pkg_file('resources', 'record.html')`).
#' @rdname record
#' @exportS3Method
#' @return The `format()` method returns a character vector of plain-text output
#'   or HTML code for displaying the results.
format.xfun_record_results = function(
    x, to = c('text', 'markdown', 'html'), encode = FALSE, template = FALSE, ...
) {
  alt = 'A plot recorded by xfun::record()'
  if (to[1] != 'html') {
    is_md = to[1] == 'markdown'
    res = unlist(lapply(x, function(z) {
      if (length(z) == 0) return()
      cls_all = sub('^record_', '', class(z)); cls = cls_all[1]
      if (cls != 'asis') {
        if (is_md && cls == 'plot') {
          z = sprintf('![%s](<%s>)', alt, z)
        } else {
          o = attr2(z, 'opts'); a = o$attr
          z = gsub('^(\\s*\n)+|\n\\s*$', '', one_string(z))  # trim blank lines
          if (cls != 'source') z = paste0(o$comment %||% '#> ', split_lines(z))
          if (is_md) {
            cls_all = if (any(c('message', 'warning', 'error') %in% cls_all)) {
              c('plain', cls_all)
            } else if (cls == 'source') {
              replace(cls_all, cls_all == 'source', 'r')
            } else {
              if (length(a)) a = c(sub('^[.]', '', a), 'plain')
              setdiff(c(a, cls_all), 'output')
            }
            z = fenced_block(z, sprintf('.%s', cls_all))
          }
        }
      }
      z = one_string(z)
      if (!is_md) z = gsub('\n*$', '\n', z)
      z
    }))
    return(raw_string(res, lang = if (is_md) '.md'))
  }
  res = unlist(lapply(x, function(z) {
    cls = sub('^record_', '', class(z))
    if (cls == 'asis') z else if (cls == 'plot') {
      sprintf(
        '<p class="%s"><img src="%s" alt="%s" /></p>',
        cls, if (encode) vapply(z, base64_uri, '') else URLencode(z), alt
      )
    } else {
      paste0(
        sprintf(
          '<pre class="%s"><code>',
          if (cls == 'source') paste0('language-r" data-start="', attr2(z, 'lines')[1]) else cls
        ),
        html_escape(one_string(z)), '</code></pre>'
      )
    }
  }))
  if (isTRUE(template)) template = pkg_file('resources', 'record.html')
  if (is.character(template)) {
    res = sub('$body$', one_string(res), read_utf8(template), fixed = TRUE)
    if (length(x) > 0) res = sub('$title$', make_title(x[[1]]), res, fixed = TRUE)
  }
  raw_string(res, lang = '.html')
}

# generate a title from the first line of a character vector
make_title = function(x) {
  if (length(x) == 0) return('')
  x = gsub('^#+\\s*', '', x[1])  # remove possible comment chars
  x = gsub('\\s*[.]*$', '... | ', x)  # add ... to the end
  html_escape(x)
}

#' @param x An object returned by `record()`.
#' @param browse Whether to browse the results on an HTML page.
#' @param ... Currently ignored.
#' @rdname record
#' @exportS3Method
#' @return The `print()` method prints the results as plain text or HTML to the
#'   console or displays the HTML page.
print.xfun_record_results = function(
  x, browse = interactive(), to = if (browse) 'html' else 'text', template = TRUE, ...
) {
  res = format(x, to, encode = browse, template = template, ...)
  if (browse && to == 'html') {
    viewer = getOption('viewer', utils::browseURL)
    f = tempfile('record-', fileext = '.html')
    write_utf8(res, f)
    viewer(f)
  } else {
    cat(res, sep = '\n')
  }
  invisible(x)
}

library(testit)

assert('attr2() is strict', {
  z = structure(list(a = 1), foo = 2)
  (attr2(z, 'foo') %==% 2)
  (attr2(z, 'f') %==% NULL)
})

assert('in_dir() preserves the working directory', {
  owd = getwd()
  in_dir('.', setwd(tempdir()))
  (same_path(owd, getwd()))
})

assert('set_envvar() sets and restores environment variables', {
  old = set_envvar(c(XFUN_TEST_VAR_123 = 'hello'))
  (Sys.getenv('XFUN_TEST_VAR_123') %==% 'hello')
  set_envvar(old)
  (is.na(old['XFUN_TEST_VAR_123']))
  (Sys.getenv('XFUN_TEST_VAR_123') %==% '')
})

assert('env_option() reads from options() and env vars', {
  # clean state
  options(xfun.test.opt = NULL)
  Sys.unsetenv('R_XFUN_TEST_OPT')
  (is.null(env_option('xfun.test.opt')))
  (env_option('xfun.test.opt', 42) %==% 42)

  Sys.setenv(R_XFUN_TEST_OPT = 'from_env')
  (env_option('xfun.test.opt') %==% 'from_env')

  options(xfun.test.opt = 'from_opt')
  (env_option('xfun.test.opt') %==% 'from_opt')

  options(xfun.test.opt = NULL)
  Sys.unsetenv('R_XFUN_TEST_OPT')
})

assert('parse_only() parses R code without keeping source', {
  (parse_only(character(0)) %==% expression())
  (parse_only('1+1') %==% expression(1+1))
  (parse_only(c('y~x', '1:5')) %==% expression(y~x, 1:5))
})

assert('try_silent() suppresses errors', {
  z = try_silent(stop('oops'))
  (inherits(z, 'try-error'))
  (try_silent(1 + 1) %==% 2)
})

assert('try_error() returns TRUE on error, FALSE otherwise', {
  (try_error(stop('foo')))
  (!try_error(1:10))
})

assert('bump_version() increments the last digit', {
  (bump_version('0.1') %==% as.numeric_version('0.2'))
  (bump_version('7.23.9') %==% as.numeric_version('7.23.10'))
  (bump_version(c('0.1', '91.2.14')) %==% as.numeric_version(c('0.2', '91.2.15')))
})

assert('strip_html() removes HTML tags and comments', {
  (strip_html('<a href="#">Hello <!-- comment -->world!</a>') %==% 'Hello world!')
  (strip_html('<b>bold</b> and <i>italic</i>') %==% 'bold and italic')
  (strip_html('no tags') %==% 'no tags')
})

assert('html_escape() escapes special HTML characters', {
  (html_escape('& < >') %==% '&amp; &lt; &gt;')
  (html_escape('" & < > \r \n') %==% '" &amp; &lt; &gt; \r \n')
  (html_escape('" & < > \r \n', attr = TRUE) %==% '&quot; &amp; &lt; &gt; &#13; &#10;')
  (html_escape('plain text') %==% 'plain text')
})

assert('html_tag() generates correct HTML', {
  (inherits(html_tag('br'), 'xfun_html'))
  (c(html_tag('br')) %==% '<br />')
  (c(html_tag('p', 'hello')) %==% '<p>hello</p>')
  (c(html_tag('a', 'click', href = 'https://example.com')) %==% '<a href="https://example.com">click</a>')
  (c(html_tag('p', '<b>bold</b>')) %==% '<p>&lt;b&gt;bold&lt;/b&gt;</p>')
  # html_value content should not be escaped
  (c(html_tag('p', html_value('<b>bold</b>'))) %==% '<p><b>bold</b></p>')
  (has_error(html_tag('p', .attrs = list(1))))
  # NULL attribute value: only the attribute name is emitted
  (c(html_tag('input', type = 'checkbox', checked = NULL)) %==% '<input type="checkbox" checked />')
})

assert('format_bytes() formats byte counts', {
  res = format_bytes(c(1, 1024))
  (res %==% c('1 bytes', '1 Kb'))
})

assert('msg_cat() prints messages suppressibly', {
  out = capture.output(suppressMessages(msg_cat('hello')))
  (length(out) %==% 0L)
  # msg_cat() signals a message condition with the correct content
  msg = tryCatch(msg_cat('world'), message = function(m) conditionMessage(m))
  (msg %==% 'world')
})

assert('%|% returns x if non-empty, otherwise y', {
  (('hello' %|% 'world') %==% 'hello')
  ((character(0) %|% 'fallback') %==% 'fallback')
  ((NULL %|% 'fallback') %==% 'fallback')
  ((1:3 %|% 99) %==% 1:3)
})

assert('stop2() and warning2() do not include the call', {
  err = tryCatch(stop2('test error'), error = identity)
  (conditionMessage(err) %==% 'test error')
  (is.null(err$call))
  wrn = tryCatch(warning2('test warn'), warning = identity)
  (conditionMessage(wrn) %==% 'test warn')
  (is.null(wrn$call))
})

assert('gsubi() performs case-insensitive substitution', {
  (gsubi('hello', 'Hi', 'Hello World') %==% 'Hi World')
  (gsubi('world', 'R', 'hello WORLD') %==% 'hello R')
})

assert('retry() retries a failing function', {
  i = 0L
  # succeeds on the 2nd call
  res = retry(function() { i <<- i + 1L; if (i < 2L) stop('fail'); i }, .times = 3, .pause = 0)
  (res %==% 2L)
  # fails all .times times and stops
  (has_error(retry(function() stop('always'), .times = 2, .pause = 0)))
})

assert('handle_error() calls handler and shows message on error', {
  msgs = c()
  tryCatch(
    withCallingHandlers(
      handle_error(stop('oops'), function(loc) c('caught', loc)),
      message = function(m) {
        msgs <<- c(msgs, conditionMessage(m)); invokeRestart('muffleMessage')
      }
    ),
    error = function(e) NULL
  )
  (any(grepl('caught', msgs)))
})

assert('is_windows/unix/macos/linux/arm64 return logical(1)', {
  (is.logical(is_windows()) && length(is_windows()) == 1L)
  (is.logical(is_unix()) && length(is_unix()) == 1L)
  (is.logical(is_macos()) && length(is_macos()) == 1L)
  (is.logical(is_linux()) && length(is_linux()) == 1L)
  (is.logical(is_arm64()) && length(is_arm64()) == 1L)
  # exactly one of windows / (unix && macos) / (unix && linux) / other unix is true
  (is_windows() || is_unix())
})

assert('set_envvar() errors on unnamed or partially unnamed vars', {
  (has_error(set_envvar(c(1, 2))))        # unnamed
  (has_error(set_envvar(c(A = '1', '')))  # partial names
  )
})

assert('tree() produces a character tree diagram', {
  x = list(a = 1, b = list(c = 2, d = 3))
  out = tree(x)
  (inherits(out, 'xfun_raw_string'))
  (is.character(out))
  (length(out) > 0)
})

assert('connect_pipes() fills vertical pipes in tree diagrams', {
  # early return for m < 3 (too few rows)
  (connect_pipes(c('|- a', '  ')) %==% c('|- a', '  '))
  # pipe connection: blank rows between two |-, first column gets filled with '|'
  x = c('|- a', '  ', '  ', '|- c')
  res = connect_pipes(x)
  (substr(res[2], 1, 1) %==% '|')
  (substr(res[3], 1, 1) %==% '|')
  # k is cleared when a non-matching cell is found (line 272 path)
  x2 = c('|- a', '  ', 'xy', '|- c')
  res2 = connect_pipes(x2)
  (substr(res2[2], 1, 1) %==% ' ')
  # k is cleared when last row reached with all spaces (line 267 path)
  x3 = c('|- a', '  ', '  ', '  ')
  res3 = connect_pipes(x3)
  (substr(res3[2], 1, 1) %==% ' ')
})

assert('func_name() returns the name of the calling function', {
  f = function() func_name(sys.nframe())
  (f() %==% 'f')
})

#' Serve a static directory
#'
#' Serve a directory of files via [new_app()]. If `index.html` exists under the
#' directory, it is shown; otherwise a directory listing is shown. This is
#' similar to `servr::httd()` and Python's `http.server` module.
#' @param dir The directory to serve.
#' @param name The app name passed to [new_app()]. Use `''` to serve the
#'   directory at the root URL via the proxy.
#' @param ... Other arguments passed to [new_app()].
#' @return The app URL, returned invisibly.
#' @export
#' @examplesIf interactive()
#' xfun::serve_dir(tempdir())
serve_dir = function(dir = '.', name = 'xfun-dir', ...) {
  dir = normalizePath(dir, mustWork = TRUE)
  handler = function(path, query = NULL, post = NULL, headers = NULL) {
    in_dir(dir, .resolve_path_to_httpd(path))
  }
  new_app(name, handler, ...)
}

# Resolve a request path under the current working directory and return a
# neutral intermediate response. This is the shared core used by both
# `xfun::serve_dir()` (mapped to R's internal httpd shape) and
# `servr:::serve_dir()` (mapped to the httpuv shape). The `action` field is
# one of:
#   'file'         — list(action, path = <abs>, mime)
#   'payload'      — list(action, body = <text>, mime, status)
#   'add-slash'    — caller redirects to its own request URL + '/'
#   '404-redirect' — caller redirects to '/404.html' (if such a file exists)
resolve_path = function(path) {
  if (path == '' || path == '/') path = '.'
  if (file_test('-d', path)) {
    # ensure a trailing slash so relative links inside the dir resolve correctly
    if (path != '.' && !grepl('/$', path)) return(list(action = 'add-slash'))
    idx = file.path(path, 'index.html')
    if (file_exists(idx))
      return(list(action = 'file', path = normalizePath(idx), mime = mime_type(idx)))
    title = html_escape(path)
    info = file.info(list.files(path, all.files = TRUE, full.names = TRUE))
    body = one_string(html_doc(
      c(sprintf('<h1>Index of %s</h1>', title), fileinfo_table(info)), title
    ))
    return(list(action = 'payload', body = body, mime = 'text/html', status = 200L))
  }
  if (!file_exists(path)) {
    if (path == 'favicon.ico') return(list(
      action = 'file', path = file.path(R.home('doc'), 'html', 'favicon.ico'),
      mime = 'image/x-icon'
    ))
    # redirect to a custom 404.html if it exists and the request looks like a page
    if (file_exists('404.html') && grepl('(/|[.]html)$', path, ignore.case = TRUE))
      return(list(action = '404-redirect'))
    return(list(
      action = 'payload', body = paste('Not found:', path),
      mime = 'text/plain', status = 404L
    ))
  }
  list(action = 'file', path = normalizePath(path), mime = mime_type(path))
}

# Map the neutral response to R's internal httpd response shape.
.resolve_path_to_httpd = function(path) {
  r = resolve_path(path)
  redir = function(dest, status) list(
    payload = sprintf('Redirect to <a href="%s">%s</a>', dest, dest),
    `content-type` = 'text/html', header = paste0('Location: ', dest),
    `status code` = as.integer(status)
  )
  switch(
    r$action,
    file = list(file = r$path, `content-type` = r$mime),
    payload = list(payload = r$body, `content-type` = r$mime, `status code` = r$status),
    `add-slash` = redir(sprintf('%s/', sub('^/?', '/', path)), 301L),
    `404-redirect` = redir('/404.html', 302L)
  )
}

# Render a file.info() data frame as an HTML directory listing.
fileinfo_table = function(info) {
  info = info[order(info$isdir, decreasing = TRUE), ]
  d = info$isdir; i = !is.na(d)
  x1 = paste0(basename(rownames(info)), ifelse(d & i, '/', ''))
  x1 = html_escape(x1)
  x1[i] = sprintf('<a href="%s">%s</a>', x1[i], x1[i])
  x2 = paste(format(info$size, scientific = FALSE, big.mark = ','), 'B')
  x2[is.na(info$size) | d] = ''
  x3 = as.character(info$mtime); x3[is.na(x3)] = ''
  c('<table>', '<thead><tr>',
    sprintf('<th>%s</th>', c('Name', 'Size', 'Date Modified')),
    '</tr></thead>',
    apply(cbind(
      '<tr>', sprintf('<td>%s</td>', x1),
      sprintf('<td align="right">%s</td>', x2),
      sprintf('<td>%s</td>', x3), '</tr>'
    ), 1, paste, collapse = ''),
    '</table>')
}

# Wrap an HTML body in a minimal HTML document with the given title.
html_doc = function(body, title = NULL) {
  c('<!DOCTYPE html>', '<html>',
    '<head>', sprintf('<title>%s</title>', title), '</head>',
    '<body>', body, '</body>', '</html>')
}

library(testit)

yaml = '
a: 1
b: [1, 2, 3]
c: true
d:
  e: !expr 1+1
  f: null
'

yaml_long_indent = '
a: 1
b: [1, 2, 3]
c: true
d:
    e: !expr 1+1
    f: null
'


yaml_tabs = '
a: 1
b: [1, 2, 3]
c: true
d:
\te: !expr 1+1
\tf: null
'

# see https://github.com/yihui/xfun/issues/94
yaml_mre = '
output:
    latex:
        latex_engine: pdflatex
        options:
            toc: true
'


if (loadable('yaml')) assert('yaml_load() works with the yaml package', {
  (yaml_load(yaml) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (yaml_load(yaml, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
  f = function() {
    foo = 1:10
    yaml_load('a: !expr head(foo, 4)')
  }
  (f() %==% list(a = 1:4))
})

assert('taml_load() works', {
  (taml_load(yaml) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (taml_load(yaml, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
  f = function() {
    foo = 1:10
    taml_load('a: !expr head(foo, 4)')
  }
  (f() %==% list(a = 1:4))
  res = list(a = c('A', 'B', 'C D'))
  (taml_load('a: [A, B, C D]') %==% res)
  (taml_load('a: ["A", \'B\', "C D"]') %==% res)
  (taml_load("a: ['A', 'B', 'C D']") %==% res)
})

assert('taml_load() works with variable indent', {
  (taml_load(yaml_long_indent) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (taml_load(yaml_long_indent, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
})

assert('taml_load() works with tabs indent', {
  (taml_load(yaml_tabs) %==% list(a = 1L, b = 1:3, c = TRUE, d = list(e = 2, f = NULL)))
  (taml_load(yaml_tabs, envir = FALSE)[[c('d', 'e')]] %==% expression(1 + 1))
})

assert('taml_load() works with non-uniform indent (varying indent sizes)', {
  x = 'runs:
  steps:
      name: Install R
      with:
        use-public-rspm: true'
  res = taml_load(x)
  (res %==% list(runs = list(steps = list(
    name = 'Install R', with = list(`use-public-rspm` = TRUE)
  ))))
})

assert('taml_load() ignores array items (- ) and their indented children', {
  # simple block sequence: items with no colon (already filtered by regex)
  (taml_load('a:\n  - b\n  - c') %==% list(a = list()))
  # block sequence items with sub-keys must also be ignored
  x = 'runs:\n  using: composite\n  steps:\n    - name: foo\n      uses: bar\n      with:\n        key: val\n    - name: baz\n'
  res = taml_load(x)
  (res %==% list(runs = list(using = 'composite', steps = list())))
  # action.yml-style: inputs are kept, steps array items are dropped
  action_yml = '
name: Build site
description: Build a site.
inputs:
  site-dir:
    description: Directory of the site source
    default: site
runs:
  using: composite
  steps:
    - name: Install R
      uses: r-lib/actions/setup-r@HEAD
      with:
        use-public-rspm: true
    - name: Build site
      shell: bash
      run: Rscript -e "litedown::fuse_site()"
'
  res = taml_load(action_yml)
  (res$name %==% 'Build site')
  (res$inputs[['site-dir']]$default %==% 'site')
  (res$runs$using %==% 'composite')
  (res$runs$steps %==% list())  # array items ignored
})

assert('yaml_load() works with a complex output', {
  expected = list(
    output = list(
      latex = list(
        latex_engine = "pdflatex", options = list(toc = TRUE)
      )
    )
  )

  (yaml_load(yaml_mre) %==% expected)
  (yaml_load(yaml_mre, use_yaml = FALSE) %==% expected)
})

assert('taml_load() handles empty input', {
  (taml_load('') %==% list())
  (taml_load(character(0)) %==% list())
  (taml_load('# comment only') %==% list())
})

assert('taml_load() handles null value in nested key', {
  # tests null value handling in nested keys
  x = 'a:\n  b: null\n  c: 1'
  res = taml_load(x)
  (is.null(res[['a']][['b']]))
  (res[['a']][['c']] %==% 1L)
  # top-level null
  (taml_load('a: null') %==% list(a = NULL))
})

assert('yaml_value() returns float for non-integer numerics', {
  res = taml_load('a: 1.5')
  (res %==% list(a = 1.5))
  (is.double(res$a))
  # empty array [] returns NULL (yaml_value line 156)
  (taml_load('a: []') %==% list(a = NULL))
})

assert('taml_file() reads YAML from a file', {
  f = tempfile()
  writeLines('a: 1\nb: foo', f)
  res = taml_file(f)
  (res %==% list(a = 1L, b = 'foo'))
  unlink(f)
})

assert('taml_save() converts a list to YAML', {
  res = list(a = 1L, b = 'foo', c = TRUE, d = list(e = 2L, f = NULL))
  yaml_str = as.character(taml_save(res))
  (any(grepl('a: 1', yaml_str)))
  (any(grepl('b: "foo"', yaml_str)))
  (any(grepl('c: true', yaml_str)))
  (any(grepl('f: null', yaml_str)))
  # round-trip
  (taml_load(yaml_str) %==% res)
})

assert('taml_save() supports custom indent', {
  res = list(a = list(b = 1L))
  yaml_str = as.character(taml_save(res, indent = '\t'))
  (grepl('\t', yaml_str))
})

assert('taml_save() errors on unsupported types', {
  (has_error(taml_save(list(a = list(1, 2)))))  # unnamed sub-list
  (has_error(taml_save(list(a = raw(1)))))  # unsupported raw type
})

assert('taml_save() handles expressions and arrays', {
  # expression value
  yaml_expr = as.character(taml_save(list(a = expression(1 + 1))))
  (any(grepl('!expr', yaml_expr)))
  # vector value -> array syntax
  yaml_arr = as.character(taml_save(list(a = c(1L, 2L, 3L))))
  (any(grepl('^a: \\[', yaml_arr)))
  # AsIs scalar -> still array
  yaml_asis = as.character(taml_save(list(a = I(1L))))
  (any(grepl('^a: \\[', yaml_asis)))
})

assert('taml_save() writes to a file when path is provided', {
  res = list(a = 1L, b = 'foo')
  f = tempfile()
  taml_save(res, path = f)
  (file.exists(f))
  (taml_load(read_utf8(f)) %==% res)
  unlink(f)
})

assert('yaml_body() splits YAML front matter from body', {
  doc = c('---', 'title: Hello', 'output: html', '---', '', 'Content here.')
  res = yaml_body(doc)
  (res$yaml %==% list(title = 'Hello', output = 'html'))
  (res$body %==% c('', '', '', '', '', 'Content here.'))
  (res$lines %==% c(1L, 4L))
  # no YAML
  doc2 = c('# heading', 'text')
  res2 = yaml_body(doc2)
  (is.null(res2$yaml))
  (res2$body %==% doc2)
  # yaml_body with parse=FALSE
  res3 = yaml_body(doc, parse = FALSE)
  (res3$yaml %==% c('title: Hello', 'output: html'))
})

library(testit)

assert('rand_unit() returns values in [0, 1)', {
  x = rand_unit(100, seed = 42)
  (length(x) %==% 100L)
  (x >= 0)
  (x < 1)
})

assert('rand_unit() produces reproducible results with the same seed', {
  x1 = rand_unit(10, seed = 0)
  x2 = rand_unit(10, seed = 0)
  (x1 %==% x2)
})

assert('rand_unit() returns 0 values for n = 0', {
  (length(rand_unit(0, seed = 1)) %==% 0L)
})

assert('rand_n() returns integers in [1, M]', {
  x = rand_n(100, 6)
  (length(x) %==% 100L)
  (x >= 1)
  (x <= 6)
})

assert('md5() computes checksums', {
  (is.character(md5(1:10)))
  (nchar(md5(1:10)) %==% 32L)
  # x2 and x3 should be identical since they are the same integer sequence
  x2 = 1:10; x3 = seq(1, 10)
  (md5(x2) %==% md5(x3))
  # named arguments produce named results
  res = md5(a = 1, b = 'hello')
  (names(res) %==% c('a', 'b'))
})

assert('find_locals() identifies locally created variables', {
  (find_locals('y = x + 1') %==% 'y')
  ("y" %in% find_locals("assign('y', x + 1)"))
})

assert('cache_exec() returns correct result', {
  r = cache_exec({2 + 2}, path = ':memory:', id = 'xfun-test-result')
  (r %==% 4)
})

assert('cache_exec() uses cache on second call', {
  id = 'xfun-test-cache-hit'
  r1 = cache_exec({Sys.time()}, path = ':memory:', id = id)
  Sys.sleep(0.01)  # ensure time would differ if re-evaluated
  r2 = cache_exec({Sys.time()}, path = ':memory:', id = id)
  # r2 should come from cache, so they should be identical
  (r1 %==% r2)
})

# functions extracted from knitr and to be reused in other packages like litedown

#' Parse comma-separated chunk options
#'
#' For \pkg{knitr} and R Markdown documents, code chunk options can be written
#' using the comma-separated syntax (e.g., `opt1=value1, opt2=value2`). This
#' function parses these options and returns a list. If an option is not named,
#' it will be treated as the chunk label.
#' @param x The chunk options as a string.
#' @return A list of chunk options.
#' @export
#' @examples
#' xfun::csv_options('foo, eval=TRUE, fig.width=5, echo=if (TRUE) FALSE')
csv_options = function(x) {
  x = one_string(x)
  res = handle_error(
    eval(parse_only(paste('alist(', quote_label(x), ')'))),
    function(loc) {
      if (loc != '') loc = paste(' at lines', loc)
      c(
        sprintf('Invalid syntax for chunk options%s:\n', loc), x,
        '\nPlease see documentation at https://yihui.org/knitr/options/.\n'
      )
    }
  )
  idx = which(names(res) == '')  # which option is not named?
  # remove empty options
  j = NULL
  for (i in idx) if (identical(res[[i]], alist(,)[[1]])) j = c(j, i)
  if (length(j)) res[j] = NULL
  idx = if (is.null(names(res)) && length(res) == 1L) 1L else which(names(res) == '')
  if ((n <- length(idx)) > 1L || (length(res) > 1L && is.null(names(res)))) stop(
    'Invalid chunk options: ', x,
    "\n\nAll options must be of the form 'tag=value' except for the chunk label."
  )
  if (is.null(res$label)) {
    if (n == 0L) res$label = '' else names(res)[idx] = 'label'
  }
  if (!is.character(res$label))
    res$label = gsub(' ', '', as.character(as.expression(res$label)))
  if (res$label == '') res$label = NULL
  res
}

# quote the chunk label if necessary
quote_label = function(x) {
  x = gsub('^\\s*,?', '', x)
  if (grepl('^\\s*[^\'"](,|\\s*$)', x)) {
    # <<a,b=1>>= ---> <<'a',b=1>>=
    x = gsub('^\\s*([^\'"])(,|\\s*$)', "'\\1'\\2", x)
  } else if (grepl('^\\s*[^\'"](,|[^=]*(,|\\s*$))', x)) {
    # <<abc,b=1>>= ---> <<'abc',b=1>>=
    x = gsub('^\\s*([^\'"][^=]*)(,|\\s*$)', "'\\1'\\2", x)
  }
  x
}

# comment characters for various languages
comment_chars = list(
  `#` = c('awk', 'bash', 'coffee', 'gawk', 'julia', 'octave', 'perl', 'powershell', 'python', 'r', 'ruby', 'sed', 'stan'),
  '//' = c('asy', 'cc', 'csharp', 'd3', 'dot', 'fsharp', 'go', 'groovy', 'java', 'js', 'node', 'ojs', 'Rcpp', 'sass', 'scss', 'scala'),
  `%%` = c('mermaid'),
  `%` = c('matlab', 'tikz'),
  `/* */` = c('c', 'css'),
  `* ;` = c('sas'),
  `--` = c('haskell', 'lua', 'mysql', 'psql', 'sql'),
  `!` = c('fortran', 'fortran95'),
  `*` = c('stata')
)
# reshape it using the language name as the index, i.e., from list(char = lang)
# to list(lang = char)
comment_chars = local({
  res = list(apl = '\u235D')
  for (i in names(comment_chars)) {
    chars = comment_chars[[i]]
    res = c(res, setNames(rep(list(strsplit(i, ' ')[[1]]), length(chars)), chars))
  }
  res[order(names(res))]
})

get_option_comment = function(engine) {
  char = comment_chars[[engine]] %||% '#'
  s1 = paste0(char[[1]], '| ')
  s2 = ifelse(length(char) > 1, char[[2]], '')
  list(start = s1, end = s2)
}

#' Divide chunk options from the code chunk body
#'
#' Chunk options can be written in special comments (e.g., after `#|` for R code
#' chunks) inside a code chunk. This function partitions these options from the
#' chunk body.
#' @param engine The name of the language engine (to determine the appropriate
#'   comment character).
#' @param code A character vector (lines of code).
#' @param strict If `FALSE`, allow chunk options to be written after `#|` even
#'   if `#` is not the comment character of the engine (e.g., when `engine =
#'   'js'`), otherwise throw an error if `#|` is detected but `#` is not the
#'   comment character.
#' @param ... Arguments to be passed to [yaml_load()].
#' @return A list with the following items:
#'
#' - `options`: The parsed options (if there are any) as a list.
#' - `src`: The part of the input that contains the options.
#' - `code`: The part of the input that contains the code.
#'
#' @note Chunk options must be written on _continuous_ lines (i.e., all lines
#'   must start with the special comment prefix such as `#|`) at the beginning
#'   of the chunk body.
#' @export
#' @examples
#' # parse yaml-like items
#' yaml_like = c("#| label: mine", "#| echo: true", "#| fig.width: 8", "#| foo: bar", "1 + 1")
#' writeLines(yaml_like)
#' xfun::divide_chunk("r", yaml_like)
#'
#' # parse CSV syntax
#' csv_like = c("#| mine, echo = TRUE, fig.width = 8, foo = 'bar'", "1 + 1")
#' writeLines(csv_like)
#' xfun::divide_chunk("r", csv_like)
divide_chunk = function(engine, code, strict = FALSE, ...) {
  res = list(options = NULL, src = NULL, code = code)
  # mask out empty blocks
  if (length(code) == 0) return(res)

  opt_comment = get_option_comment(engine)
  s1 = opt_comment$start
  s2 = opt_comment$end

  # check for option comments
  i1 = startsWith(code, s1)
  # if "commentChar| " is not found, try "#| " instead
  if (!i1[1] && s1 != '#| ') {
    i1 = startsWith(code, '#| ')
    if (strict && i1[1]) stop2(
      "The chunk options should start with '", s1, "' instead of '#| '",
      if (s2 != '') c(", and end with '", s2, "'"), '.'
    )
    s1 = '#| '; s2 = ''
  }
  # must have at least one matched line at the beginning
  if (!i1[[1]]) return(res)

  # end of the pipe comment block
  n2 = if (s2 == '') {
    i2 = TRUE
    if (all(i1)) length(code) else which.min(i1) - 1  # which.min() finds first FALSE
  } else {
    i2 = endsWith(trimws(code, 'right'), s2)
    if (i2[1]) which.min(i2) - 1 else which.max(i2)
  }

  # divide into yaml and code
  i = 1:n2; src = code[i]; code = code[-i]

  # extract meta from comments, then parse it
  c1 = nchar(s1) * i1[i] + 1
  c2 = nchar(src) - if (s2 == '') 0 else nchar(s2) * i2[i]
  meta = substr(src, c1, c2)
  # see if the metadata looks like YAML or CSV
  if (grepl('^[^ :]+:($|\\s)', meta[1])) {
    meta = yaml_load(meta, envir = FALSE, ...)
    if (!is.list(meta) || length(names(meta)) == 0) {
      warning('Invalid YAML option format in chunk: \n', one_string(meta), '\n')
      meta = list()
    }
  } else {
    meta = csv_options(meta)
  }

  # normalize field name 'id' to 'label' if provided
  meta$label = unlist(meta[c('label', 'id')])[[1]]
  meta$id = NULL

  # extract code
  if (length(code) > 0 && is_blank(code[[1]])) {
    code = code[-1]
    src = c(src, '')
  }

  list(options = meta, src = src, code = code)
}

# change chunk option `tag=value` to `tag = value`, i.e., add spaces around =
chunk_spaces = function(dir = '.', pattern = '[.]Rmd$') {
  for (f in list.files(dir, pattern, full.names = TRUE, recursive = TRUE))
    process_file(f, function(x) {
      i = grep('^\\s*(> )*(#[|] |```+\\s*\\{)', x)
      x[i] = gsub('(?<! )=', ' =', x[i], perl = TRUE)
      x[i] = gsub('=(?! )', '= ', x[i], perl = TRUE)
      x
    })
}

#' Attach or load packages, and automatically install missing packages if
#' requested
#'
#' `pkg_attach()` is a vectorized version of [library()] over
#' the `package` argument to attach multiple packages in a single function
#' call. `pkg_load()` is a vectorized version of
#' [requireNamespace()] to load packages (without attaching them).
#' The functions `pkg_attach2()` and `pkg_load2()` are wrappers of
#' `pkg_attach(install = TRUE)` and `pkg_load(install = TRUE)`,
#' respectively. `loadable()` is an abbreviation of
#' `requireNamespace(quietly = TRUE)`. `pkg_available()` tests if a
#' package with a minimal version is available.
#'
#' These are convenience functions that aim to solve these common problems: (1)
#' We often need to attach or load multiple packages, and it is tedious to type
#' several `library()` calls; (2) We are likely to want to install the
#' packages when attaching/loading them but they have not been installed.
#' @param ... Package names (character vectors, and must always be quoted).
#' @param install Whether to automatically install packages that are not
#'   available using [install.packages()]. Besides `TRUE` and
#'   `FALSE`, the value of this argument can also be a function to install
#'   packages (`install = TRUE` is equivalent to `install =
#'   install.packages`), or a character string `"pak"` (equivalent to
#'   `install = pak::pkg_install`, which requires the \pkg{pak} package).
#'   You are recommended to set a CRAN mirror in the global option `repos`
#'   via [options()] if you want to automatically install packages.
#' @param message Whether to show the package startup messages (if any startup
#'   messages are provided in a package).
#' @return `pkg_attach()` returns `NULL` invisibly. `pkg_load()`
#'   returns a logical vector, indicating whether the packages can be loaded.
#' @seealso `pkg_attach2()` is similar to `pacman::p_load()`, but does
#'   not allow non-standard evaluation (NSE) of the `...` argument, i.e.,
#'   you must pass a real character vector of package names to it, and all names
#'   must be quoted. Allowing NSE adds too much complexity with too little gain
#'   (the only gain is that it saves your effort in typing two quotes).
#' @import utils
#' @export
#' @examples library(xfun)
#' pkg_attach('stats', 'graphics')
#' # pkg_attach2('servr')  # automatically install servr if it is not installed
#'
#' (pkg_load('stats', 'graphics'))
pkg_attach = function(
  ..., install = FALSE, message = getOption('xfun.pkg_attach.message', TRUE)
) {
  if (!message) library = function(...) {
    suppressPackageStartupMessages(base::library(...))
  }
  for (i in c(...)) {
    if (!identical(install, FALSE) && !loadable(i)) pkg_install(i, install)
    library(i, character.only = TRUE)
  }
}

#' @param error Whether to signal an error when certain packages cannot be loaded.
#' @rdname pkg_attach
#' @export
pkg_load = function(..., error = TRUE, install = FALSE) {
  n = length(pkg <- c(...)); res = logical(n)
  if (n == 0) return(invisible(res))
  for (i in seq_len(n)) {
    res[i] = loadable(p <- pkg[i])
    if (!identical(install, FALSE) && !res[i]) {
      pkg_install(p, install); res[i] = loadable(p)
    }
  }
  if (error && any(!res)) stop('Package(s) not loadable: ', paste(pkg[!res], collapse = ' '))
  invisible(res)
}

#' @param pkg A single package name.
#' @param strict If `TRUE`, use [requireNamespace()] to test if
#'   a package is loadable; otherwise only check if the package is in
#'   [`.packages`]`(TRUE)` (this does not really load the package, so it
#'   is less rigorous but on the other hand, it can keep the current R session
#'   clean).
#' @param new_session Whether to test if a package is loadable in a new R
#'   session. Note that `new_session = TRUE` implies `strict = TRUE`.
#' @rdname pkg_attach
#' @export
loadable = function(pkg, strict = TRUE, new_session = FALSE) {
  if (length(pkg) != 1L) stop("'pkg' must be a character vector of length one")
  if (new_session) {
    Rscript(c('-e', shQuote(sprintf('library("%s")', pkg))), stdout = FALSE, stderr = FALSE) == 0
  } else {
    if (strict) {
      suppressPackageStartupMessages(requireNamespace(pkg, quietly = TRUE))
    } else pkg %in% .packages(TRUE)
  }
}

#' @param version A minimal version number. If `NULL`, only test if a
#'   package is available and do not check its version.
#' @rdname pkg_attach
#' @export
pkg_available = function(pkg, version = NULL) {
  loadable(pkg) && (is.null(version) || packageVersion(pkg) >= version)
}

#' @rdname pkg_attach
#' @export
pkg_attach2 = function(...) pkg_attach(..., install = TRUE)

#' @rdname pkg_attach
#' @export
pkg_load2 = function(...) pkg_load(..., install = TRUE)

pkg_require = function(pkgs, which = length(sys.calls()) - 1) {
  f = func_name(which)
  for (p in pkgs) if (!loadable(p)) stop2(
    "The '", p, "' package is required by the function '", f, "' but not available.",
    if (is_R_CMD_check()) c(
      " If you are developing an R package, you need to declare the dependency on '",
      p, "' in the DESCRIPTION file (e.g., in 'Imports')."
    )
  )
}

# update all packages in libraries that are writable by the current user
pkg_update = function(...) {
  libs = .libPaths()
  libs = libs[file.access(libs, 2) >= 0]
  for (l in libs) update.packages(l, ask = FALSE, checkBuilt = TRUE, ...)
}

# allow users to specify a custom install.packages() function via the global
# option xfun.install.packages
pkg_install = function(pkgs, install = TRUE, ...) {
  if (length(pkgs) == 0) return(invisible())
  # in case the CRAN repo is not set up
  repos = getOption('repos')
  if (length(repos) == 0 || identical(repos, c(CRAN = '@CRAN@'))) {
    opts = options(repos = c(CRAN = 'https://cloud.r-project.org'))
    on.exit(options(opts), add = TRUE)
  }
  if (length(pkgs) > 1)
    message('Installing ', length(pkgs), ' packages: ', paste(pkgs, collapse = ' '))
  if (isTRUE(install)) install = getOption(
    'xfun.install.packages',
    if (is.na(Sys.getenv('RENV_PROJECT', NA)) || !loadable('renv')) install.packages else {
      function(pkgs, lib = NULL, ...) getFromNamespace('install', 'renv')(pkgs, library = lib, ...)
    }
  )
  if (identical(install, 'pak')) install = pak::pkg_install
  invisible(retry(install, pkgs, ..., .pause = 0))
}

#' Find out broken packages and reinstall them
#'
#' If a package is broken (i.e., not [loadable()]), reinstall it.
#'
#' Installed R packages could be broken for several reasons. One common reason
#' is that you have upgraded R to a newer `x.y` version, e.g., from `4.0.5` to
#' `4.1.0`, in which case you need to reinstall previously installed packages.
#' @param reinstall Whether to reinstall the broken packages, or only list their
#'   names.
#' @return A character vector of names of broken package.
#' @export
broken_packages = function(reinstall = TRUE) {
  libs = .libPaths()
  pkgs = unlist(lapply(libs, function(lib) {
    p = unlist(lapply(.packages(TRUE, lib), function(p) {
      if (!loadable(p, new_session = TRUE)) p
    }))
    if (length(p) && reinstall) {
      remove.packages(p, lib); pkg_install(p, lib = lib)
    }
    p
  }))
  if(reinstall) invisible(pkgs) else pkgs
}

#' Install a source package from a directory
#'
#' Run \command{R CMD build} to build a tarball from a source directory, and run
#' \command{R CMD INSTALL} to install it.
#' @param pkg The package source directory.
#' @param build Whether to build a tarball from the source directory. If
#'   `FALSE`, run \command{R CMD INSTALL} on the directory directly (note
#'   that vignettes will not be automatically built).
#' @param build_opts The options for \command{R CMD build}.
#' @param install_opts The options for \command{R CMD INSTALL}.
#' @export
#' @return Invisible status from \command{R CMD INSTALL}.
install_dir = function(pkg = '.', build = TRUE, build_opts = NULL, install_opts = NULL) {
  if (build) {
    pkg = pkg_build(pkg, build_opts)
    on.exit(unlink(pkg), add = TRUE)
  }
  res = Rcmd(c('INSTALL', install_opts, pkg))
  if (res != 0) stop('Failed to install the package ', pkg)
  invisible(res)
}

pkg_build = function(dir = '.', opts = NULL) {
  desc = file.path(dir, 'DESCRIPTION')
  pv = read.dcf(desc, fields = c('Package', 'Version'))
  # delete existing tarballs
  unlink(sprintf('%s_*.tar.gz', pv[1, 1]))
  Rcmd(c('build', opts, shQuote(dir)))
  pkg = sprintf('%s_%s.tar.gz', pv[1, 1], pv[1, 2])
  if (!file_exists(pkg)) stop('Failed to build the package ', pkg)
  pkg
}

# install dependencies given a package DESCRIPTION
install_deps = function(dir = '.') {
  d = read.dcf(file.path(dir, 'DESCRIPTION'))
  d = d[1, intersect(colnames(d), c('Depends', 'Imports', 'Suggests', 'LinkingTo'))]
  d = unlist(strsplit(d, ',\\s*'), use.names = FALSE)
  d = gsub('\\(.*\\)', '', d)
  d = gsub('^\\s+|\\s+$', '', d)
  d = setdiff(d, c('R', base_pkgs()))
  if (length(d)) {
    db = available.packages()
    d2 = NULL
    for (p in intersect(d, rownames(db))) d2 = c(d2, tryCatch(
      if (packageVersion(p) < db[p, 'Version']) p,
      error = function(e) p
    ))
    pkg_install(d2)
  }
}

#' An alias of `remotes::install_github()`
#'
#' This alias is to make autocomplete faster via `xfun::install_github`, because
#' most `remotes::install_*` functions are never what I want. I only use
#' `install_github` and it is inconvenient to autocomplete it, e.g.
#' `install_git` always comes before `install_github`, but I never use it. In
#' RStudio, I only need to type `xfun::ig` to get `xfun::install_github`.
#' @param ... Arguments to be passed to [remotes::install_github()].
#' @export
install_github = function(...) remotes::install_github(...)

# Remove packages not installed from CRAN
reinstall_from_cran = function(dry_run = TRUE, skip_github = TRUE) {
  r = paste(c('Repository', if (skip_github) 'GithubRepo'), collapse = '|')
  r = paste0('^(', r, '): ')
  for (lib in .libPaths()) {
    pkgs = .packages(TRUE, lib)
    pkgs = setdiff(pkgs, c('xfun', 'rstudio', base_pkgs()))
    for (p in pkgs) {
      desc = read_utf8(system.file('DESCRIPTION', package = p, lib.loc = lib))
      if (!any(grepl(r, desc))) {
        if (dry_run) message(p, ': ', lib) else install.packages(p, lib = lib)
      }
    }
  }
}

#' Convert package news to the Markdown format
#'
#' Read the package news with [news()], convert the result to
#' Markdown, and write to an output file (e.g., \file{NEWS.md}). Each package
#' version appears in a first-level header, each category (e.g., \samp{NEW
#' FEATURES} or \samp{BUG FIXES}) is in a second-level header, and the news
#' items are written into bullet lists.
#' @param package,... Arguments to be passed to [news()].
#' @param output The output file path.
#' @param category Whether to keep the category names.
#' @return If `output = NA`, returns the Markdown content as a character
#'   vector, otherwise the content is written to the output file.
#' @export
#' @examplesIf interactive()
#' # news for the current version of R
#' xfun::news2md('R', Version == getRversion(), output = NA)
news2md = function(package, ..., output = 'NEWS.md', category = TRUE) {
  db = news(package = package, ...)
  k = db[, 'Category']
  db[is.na(k), 'Category'] = ''  # replace NA category with ''
  res = unlist(lapply(unique(db[, 'Version']), function(v) {
    d1 = db[db[, 'Version'] == v, ]
    res = unlist(lapply(unique(d1[, 'Category']), function(k) {
      txt = d1[d1[, 'Category'] == k, 'Text']
      txt = txt[txt != '']
      if (k == '' && length(txt) == 0) return()
      txt = gsub('\n *', ' ', txt)
      c(if (category && k != '') paste('##', k), if (length(txt)) paste('-', txt))
    }))
    if (is.na(dt <- d1[1, 'Date'])) dt = '' else dt = paste0(' (', dt, ')')
    c(sprintf('# CHANGES IN %s VERSION %s%s', package, v, dt), res)
  }))
  res = c(rbind(res, ''))  # add a blank line after each line
  if (is.na(output)) raw_string(res) else write_utf8(res, output)
}

#' Get base R package names
#'
#' Return base R package names.
#' @return A character vector of base R package names.
#' @export
#' @examples
#' xfun::base_pkgs()
base_pkgs = function() {
  if (is.function(f <- asNamespace('tools')$standard_package_names)) f()[['base']] else c(
    'base', 'compiler', 'datasets', 'graphics', 'grDevices', 'grid', 'methods',
    'parallel', 'splines', 'stats', 'stats4', 'tcltk', 'tools', 'utils'
  )
}

library(testit)

assert("loadable() works", {
  (loadable("base"))
  # (loadable("base", new_session = TRUE))
  (loadable("base", strict = FALSE))
  (has_error(loadable(c("base", "base2"))))
  (has_error(loadable(character())))
  (!loadable("#base"))
  # (!loadable("#base", new_session = TRUE))
  (!loadable("#base", strict = FALSE))
})

assert('pkg_load() loads packages and returns logical vector', {
  res = pkg_load('base', 'stats', error = FALSE)
  (res %==% c(TRUE, TRUE))
  res2 = pkg_load('#nonexistent#pkg', error = FALSE)
  (!res2)
})

assert('pkg_attach() attaches packages', {
  pkg_attach('stats')
  ('package:stats' %in% search())
})

assert('pkg_available() checks if a package with a version is available', {
  (pkg_available('base'))
  (pkg_available('base', '1.0.0'))
  (!pkg_available('#nonexistent#pkg'))
})

assert('base_pkgs() returns base package names', {
  pkgs = base_pkgs()
  (is.character(pkgs))
  (c('base', 'stats', 'utils', 'methods', 'tools') %in% pkgs)
})

assert('pkg_load() with error = TRUE stops on missing package', {
  (has_error(pkg_load('#nonexistent#pkg', error = TRUE)))
})

if (getRversion() >= '3.3.0') assert('pkg_attach() with message = FALSE suppresses messages', {
  out = capture.output(type = 'message', pkg_attach('stats', message = FALSE))
  (length(out) %==% 0L)
})

assert('loadable() with strict = FALSE checks .packages(TRUE)', {
  (loadable('base', strict = FALSE))
  (!loadable('#nonexistent#', strict = FALSE))
})

assert('pkg_available() with version constraint works', {
  (pkg_available('base', '0.0.1'))
  (!pkg_available('xfun', '999999.999.999'))
})

#' Run `system2()` and mark its character output as UTF-8 if appropriate
#'
#' This is a wrapper function based on `system2()`. If `system2()`
#' returns character output (e.g., with the argument `stdout = TRUE`),
#' check if the output is encoded in UTF-8. If it is, mark it with UTF-8
#' explicitly.
#' @param ... Passed to [system2()].
#' @return The value returned by `system2()`.
#' @export
#' @examplesIf interactive()
#' a = shQuote(c('-e', 'print(intToUtf8(c(20320, 22909)))'))
#' x2 = system2('Rscript', a, stdout = TRUE)
#' Encoding(x2)  # unknown
#'
#' x3 = xfun::system3('Rscript', a, stdout = TRUE)
#' # encoding of x3 should be UTF-8 if the current locale is UTF-8
#' !l10n_info()[['UTF-8']] || Encoding(x3) == 'UTF-8'  # should be TRUE
system3 = function(...) {
  res = system2(...)
  if (is.character(res)) {
    if (all(is_utf8(res))) Encoding(res) = 'UTF-8'
  }
  if (is.integer(res) && res == 0) invisible(res) else res
}

#' Run OptiPNG on all PNG files under a directory
#'
#' Call the command \command{optipng} via `system2()` to optimize all PNG
#' files under a directory.
#' @param dir Path to a directory.
#' @param files Alternatively, you can choose the specific files to optimize.
#' @param ... Arguments to be passed to `system2()`.
#' @references OptiPNG: <https://optipng.sourceforge.net>.
#' @export
optipng = function(dir = '.', files = all_files('[.]png$', dir), ...) {
  if (Sys.which('optipng') != '') for (f in files) system2('optipng', shQuote(f), ...)
}

#' Run the commands \command{Rscript} and \command{R CMD}
#'
#' Wrapper functions to run the commands \command{Rscript} and \command{R CMD}.
#' @param args A character vector of command-line arguments.
#' @param ... Other arguments to be passed to [system2()].
#' @export
#' @return A value returned by `system2()`.
#' @examples library(xfun)
#' Rscript(c('-e', '1+1'))
#' Rcmd(c('build', '--help'))
Rscript = function(args, ...) {
  # unset R_TESTS for the new R session: https://stackoverflow.com/a/27994299
  if (is_R_CMD_check()) {
    v = set_envvar(c(R_TESTS = NA)); on.exit(set_envvar(v), add = TRUE)
  }
  system2(file.path(R.home('bin'), 'Rscript'), args, ...)
}

#' @rdname Rscript
#' @export
Rcmd = function(args, ...) {
  system2(file.path(R.home('bin'), 'R'), c('CMD', args), ...)
}

#' Call a function in a new R session via `Rscript()`
#'
#' Save the argument values of a function in a temporary RDS file, open a new R
#' session via [Rscript()], read the argument values, call the function, and
#' read the returned value back to the current R session.
#' @param fun A function, or a character string that can be parsed and evaluated
#'   to a function.
#' @param args A list of argument values.
#' @param options A character vector of options to passed to [Rscript()], e.g.,
#'   `"--vanilla"`.
#' @param ...,wait Arguments to be passed to [system2()].
#' @param fail The desired error message when an error occurred in calling the
#'   function. If the actual error message during running the function is
#'   available, it will be appended to this message.
#' @export
#' @return If `wait = TRUE`, the returned value of the function in the new R
#'   session. If `wait = FALSE`, three file paths will be returned: the first
#'   one stores `fun` and `args` (as a list), the second one is supposed to
#'   store the returned value of the function, and the third one stores the
#'   possible error message.
#' @examples factorial(10)
#' # should return the same value
#' xfun::Rscript_call('factorial', list(10))
#'
#' # the first argument can be either a character string or a function
#' xfun::Rscript_call(factorial, list(10))
#'
#' # Run Rscript starting a vanilla R session
#' xfun::Rscript_call(factorial, list(10), options = c("--vanilla"))
Rscript_call = function(
  fun, args = list(), options = NULL, ..., wait = TRUE,
  fail = sprintf("Failed to run '%s' in a new R session", deparse(substitute(fun))[1])
) {
  f = replicate(3, tempfile(fileext = '.rds'))
  on.exit(if (wait) unlink(f), add = TRUE)
  saveRDS(list(fun, args), f[1])
  res = Rscript(
    c(options, shQuote(c(pkg_file('scripts', 'call-fun.R'), f))), ..., wait = wait
  )
  if (wait) {
    if (res == 0) readRDS(f[2]) else stop(
      fail, if (file_exists(f[3])) c(': ', readRDS(f[3])) else '.', call. = FALSE
    )
  } else f
}

# call a function in a background process
Rscript_bg = function(fun, args = list(), timeout = 10) {
  pid = tempfile()  # to store the process ID of the new R session
  saveRDS(NULL, pid)

  Rscript_call(function() {
    saveRDS(Sys.getpid(), pid)
    # remove this pid file when the function finishes
    on.exit(unlink(pid), add = TRUE)
    do.call(fun, args)
  }, wait = FALSE)

  id = NULL  # read the above process ID into this R session
  res = list(pid = id, is_alive = function() FALSE)

  # check if the pid file still exists; if not, the process has ended
  if (!file_exists(pid)) return(res)

  t0 = Sys.time()
  while (difftime(Sys.time(), t0, units = 'secs') < timeout) {
    Sys.sleep(.1)
    if (!file_exists(pid)) return(res)
    if (length(id <- readRDS(pid)) == 1) break
  }
  if (length(id) == 0) stop(
    'Failed to launch the background process in ', timeout, ' seconds (timeout).'
  )

  list(pid = id, is_alive = function() file_exists(pid))
}

#' Kill a process and (optionally) all its child processes
#'
#' Run the command \command{taskkill /f /pid} on Windows and \command{kill} on
#' Unix, respectively, to kill a process.
#' @param pid The process ID.
#' @param recursive Whether to kill the child processes of the process.
#' @param ... Arguments to be passed to [system2()] to run the
#'   command to kill the process.
#' @return The status code returned from `system2()`.
#' @export
proc_kill = function(pid, recursive = TRUE, ...) {
  if (is_windows()) {
    system2('taskkill', c(if (recursive) '/t', '/f', '/pid', pid), ...)
  } else {
    system2('kill', c(pid, if (recursive) child_pids(pid)), ...)
  }
}

# obtain pids of all child processes (recursively)
child_pids = function(id) {
  x = system2('sh', shQuote(c(pkg_file('scripts', 'child-pids.sh'), id)), stdout = TRUE)
  grep('^[0-9]+$', x, value = TRUE)
}

powershell = function(command) {
  if (Sys.which('powershell') == '') return()
  command = paste(command, collapse = ' ')
  system2('powershell', c('-Command', shQuote(command)), stdout = TRUE)
}

# start a background process via the PowerShell cmdlet and return its pid
ps_process = function(command, args = character(), verbose = FALSE) {
  powershell(c(
    'echo (Start-Process', '-FilePath', shQuote(command), '-ArgumentList',
    ps_quote(args), '-PassThru', '-WindowStyle',
    sprintf('%s).ID', if (verbose) 'Normal' else 'Hidden')
  ))
}

# quote PowerShell arguments properly
ps_quote = function(x) {
  x = gsub('"', '""', x)  # '""' mean a literal '"'
  # if an argument contains a space, surround it with escaped double quotes `"`"
  i = grep('\\s', x)
  x[i] = sprintf('`"%s`"', x[i])
  sprintf('"%s"', paste(x, collapse = ' '))
}

#' Start a background process
#'
#' Start a background process using the PowerShell cmdlet
#' \command{Start-Process-PassThru} on Windows or the ampersand \command{&} on
#' Unix, and return the process ID.
#' @param command,args The system command and its arguments. They do not need to
#'   be quoted, since they will be quoted via [shQuote()] internally.
#' @param verbose If `FALSE`, suppress the output from `stdout` (and also
#'   `stderr` on Windows). The default value of this argument can be set via a
#'   global option, e.g., `options(xfun.bg_process.verbose = TRUE)`.
#' @return The process ID as a character string.
#' @note On Windows, if PowerShell is not available, try to use
#'   [`system2`]`(wait = FALSE)` to start the background process instead. The
#'   process ID will be identified from the output of the command
#'   \command{tasklist}. This method of looking for the process ID may not be
#'   reliable. If the search is not successful in 30 seconds, it will throw an
#'   error (timeout). If a longer time is needed, you may set
#'   `options(xfun.bg_process.timeout)` to a larger value, but it should be very
#'   rare that a process cannot be started in 30 seconds. When you reach the
#'   timeout, it is more likely that the command actually failed.
#' @export
#' @seealso [proc_kill()] to kill a process.
bg_process = function(
  command, args = character(), verbose = getOption('xfun.bg_process.verbose', FALSE)
) {
  throw_error = function(...) stop(
    'Failed to run the command', ..., ' in the background: ',
    paste(shQuote(c(command, args)), collapse = ' '), call. = FALSE
  )

  # check the possible pid returned from system2()
  check_pid = function(res) {
    if (is.null(res)) return(res)
    if (!is.null(attr2(res, 'status'))) throw_error()
    if (length(res) == 1 && grepl('^[0-9]+$', res)) return(res)
    throw_error()
  }

  if (is_windows()) {
    # first try 'Start-Process -PassThrough' to start a background process; if
    # PowerShell is unavailable, fall back to system2(wait = FALSE), and the
    # method to find out the pid is not 100% reliable
    if (length(pid <- check_pid(ps_process(command, args, verbose))) == 1) return(pid)

    message(
      'It seems you do not have PowerShell installed. The process ID may be inaccurate.'
    )
    # format of task list: hugo.exe    4592 Console      1     35,188 K
    tasklist = function() system2('tasklist', stdout = TRUE)
    pid1 = tasklist()
    system2(command, shQuote(args), wait = FALSE)

    get_pid = function() {
      # make sure the command points to an actual executable (e.g., resolve 'R'
      # to 'R.exe')
      if (!file_exists(command)) {
        if (Sys.which(command) != '') command = Sys.which(command)
      }
      cmd = basename(command)

      pid2 = setdiff(tasklist(), pid1)
      # the process's info should start with the command name
      pid2 = pid2[substr(pid2, 1, nchar(cmd)) == cmd]
      if (length(pid2) == 0) return()
      m = regexec('\\s+([0-9]+)\\s+', pid2)
      for (v in regmatches(pid2, m)) if (length(v) >= 2) return(v[2])
    }

    t0 = Sys.time(); id = NULL; timeout = getOption('xfun.bg_process.timeout', 30)
    while (difftime(Sys.time(), t0, units = 'secs') < timeout) {
      if (length(id <- get_pid()) > 0) break
    }

    if (length(id) > 0) return(id)

    system2(command, args, timeout = timeout)  # see what the error is
    throw_error(' in ', timeout, ' second(s)')
  } else {
    pid = tempfile(); on.exit(unlink(pid), add = TRUE)
    code = paste(c(
      shQuote(c(command, args)), if (!verbose) '> /dev/null', '& echo $! >', shQuote(pid)
    ), collapse = ' ')
    system2('sh', c('-c', shQuote(code)))
    return(check_pid(readLines(pid)))
  }
}

#' Upload to an FTP server via \command{curl}
#'
#' The function `upload_ftp()` runs the command \command{curl -T file
#' server} to upload a file to an FTP server if the system command
#' \command{curl} is available, otherwise it uses the R package \pkg{curl}. The
#' function `upload_win_builder()` uses `upload_ftp()` to upload
#' packages to the win-builder server.
#'
#' These functions were written mainly to save package developers the trouble of
#' going to the win-builder web page and uploading packages there manually.
#' @param file Path to a local file.
#' @param server The address of the FTP server. For `upload_win_builder()`,
#'   `server = 'https'` means uploading to
#'   `'https://win-builder.r-project.org/upload.aspx'`.
#' @param dir The remote directory to which the file should be uploaded.
#' @param version The R version(s) on win-builder.
#' @return Status code returned from [system2()] or
#'   `curl::curl_fetch_memory()`.
#' @export
upload_ftp = function(file, server, dir = '') {
  if (dir != '') dir = gsub('/*$', '/', dir)
  server = paste0(server, dir)
  if (Sys.which('curl') == '') {
    curl::curl_upload(file, server)$status_code
  } else {
    system2('curl', shQuote(c('-T', file, server)))
  }
}

#' @rdname upload_ftp
#' @export
upload_win_builder = function(
  file = pkg_build(), version = c("R-devel", "R-release", "R-oldrelease"),
  server = c('ftp', 'https')
) {
  if (missing(file)) on.exit(file.remove(file), add = TRUE)
  if (system2('git', 'status', stderr = FALSE) == 0) system2('git', 'pull')
  server = server[1]
  server = switch(
    server,
    'ftp'   = paste0(server, '://win-builder.r-project.org/'),
    'https' = paste0(server, '://win-builder.r-project.org/upload.aspx'),
    server
  )
  res = if (grepl('^ftp://', server)) {
    lapply(version, upload_ftp, file = file, server = server)
  } else {
    vers = c('R-devel' = 2, 'R-release' = 1, 'R-oldrelease' = 3)
    params = list(
      FileUpload = file,
      Button = 'Upload File',
      # perhaps we should read these tokens dynamically from
      # https://win-builder.r-project.org/upload.aspx
      `__VIEWSTATE` = '/wEPDwULLTE0OTY5NTg0MTUPZBYCAgIPFgIeB2VuY3R5cGUFE211bHRpcGFydC9mb3JtLWRhdGFkZFHMrNH6JjHTyJ00T0dAADGf4oa0',
      `__VIEWSTATEGENERATOR` = '69164837',
      `__EVENTVALIDATION` = '/wEWBQKksYbrBgKM54rGBgK7q7GGCAKF2fXbAwLWlM+bAqR2dARbCNfKVu0vDawqWYgB5kKI'
    )
    lapply(version, function(i) {
      names(params)[1:2] = paste0(names(params)[1:2], vers[i])
      if (Sys.which('curl') == '') {
        h = curl::new_handle()
        params[[1]] = curl::form_file(params[[1]])
        curl::handle_setform(h, .list = params)
        curl::curl_fetch_memory(server, h)$status_code
      } else {
        params[1] = paste0('@', params[1])
        system2('curl', shQuote(c(
          rbind('-F', paste(names(params), params, sep = '=')),
          server
        )), stdout = FALSE)
      }
    })
  }

  setNames(unlist(res), version)
}

#' Read / write files encoded in UTF-8
#'
#' Read or write files, assuming they are encoded in UTF-8. `read_utf8()` is
#' roughly `readLines(encoding = 'UTF-8')` (a warning will be issued if non-UTF8
#' lines are found), and `write_utf8()` calls `writeLines(enc2utf8(text),
#' useBytes = TRUE)`.
#'
#' The function `append_utf8()` appends UTF-8 content to a file or connection
#' based on `read_utf8()` and `write_utf8()`, and optionally sort the content.
#' The function `append_unique()` appends unique lines to a file or connection.
#' @param con A connection or a file path.
#' @param error Whether to signal an error when non-UTF8 characters are detected
#'   (if `FALSE`, only a warning message is issued).
#' @param text A character vector (will be converted to UTF-8 via [enc2utf8()]).
#' @param ... Other arguments passed to [writeLines()] (except `useBytes`, which
#'   is `TRUE` in `write_utf8()`).
#' @return `read_utf8()` returns a character vector of the file content;
#'   `write_utf8()` returns the `con` argument (invisibly).
#' @export
read_utf8 = function(con, error = FALSE) {
  # users may have set options(encoding = 'UTF-8'), which usually won't help but
  # will bring more trouble than good, so we reset this option temporarily
  opts = options(encoding = 'native.enc'); on.exit(options(opts), add = TRUE)
  x = readLines(con, encoding = 'UTF-8', warn = FALSE)
  i = invalid_utf8(x)
  n = length(i)
  if (n > 0) (if (error) stop else warning)(
    if (is.character(con)) c('The file ', con, ' is not encoded in UTF-8. '),
    'These lines contain invalid UTF-8 characters: ',
    paste(c(head(i), if (n > 6) '...'), collapse = ', ')
  )
  x
}

#' @rdname read_utf8
#' @export
write_utf8 = function(text, con, ...) {
  if (is.null(text)) text = character(0)
  if (identical(con, '')) {
    cat(text, sep = '\n', file = con)
  } else {
    # prevent re-encoding the text in the file() connection in writeLines()
    # https://kevinushey.github.io/blog/2018/02/21/string-encoding-and-r/
    opts = options(encoding = 'native.enc'); on.exit(options(opts), add = TRUE)
    writeLines(enc2utf8(text), con, ..., useBytes = TRUE)
  }
  invisible(con)
}

#' @param sort Logical (`FALSE` means not to sort the content) or a
#'   function to sort the content; `TRUE` is equivalent to
#'   `base::sort`.
#' @rdname read_utf8
#' @export
append_utf8 = function(text, con, sort = TRUE) {
  x = read_utf8(con, error = TRUE)
  x = c(x, text)
  if (is.logical(sort)) sort = if (sort) base::sort else identity
  if (is.function(sort)) x = sort(x)
  write_utf8(x, con)
}

#' @rdname read_utf8
#' @export
append_unique = function(text, con, sort = function(x) base::sort(unique(x))) {
  append_utf8(text, con, sort)
}

# which lines are invalid UTF-8
invalid_utf8 = function(x) {
  which(!is_utf8(x))
}

test_utf8 = function(x) {
  is.na(x) | !is.na(iconv(x, 'UTF-8', 'UTF-8'))
}

# validUTF8() was added to base R 3.3.0
is_utf8 = function(x) {
  if ('validUTF8' %in% ls(baseenv())) validUTF8(x) else test_utf8(x)
}

#' Read a text file and concatenate the lines by `'\n'`
#'
#' The source code of this function should be self-explanatory.
#' @param file Path to a text file (should be encoded in UTF-8).
#' @return A character string of text lines concatenated by `'\n'`.
#' @export
#' @examples
#' xfun::file_string(system.file('DESCRIPTION', package = 'xfun'))
file_string = function(file) {
  x = read_utf8(file)
  # paste converts 0-length character() into 1-length ""
  if (length(x)) x = paste(x, collapse = '\n')
  raw_string(x)
}

#' Read all records of a binary file as a raw vector by default
#'
#' This is a wrapper function of [readBin()] with default arguments
#' `what = "raw"` and `n = `[`file.size`]`(file)`, which means it
#' will read the full content of a binary file as a raw vector by default.
#' @param file,what,n,... Arguments to be passed to `readBin()`.
#' @return A vector returned from `readBin()`.
#' @export
#' @examples
#' f = tempfile()
#' cat('abc', file = f)
#' xfun::read_bin(f)
#' unlink(f)
read_bin = function(file, what = 'raw', n = file.info(file)$size, ...) {
  readBin(file, what, n, ...)
}

#' Read all text files and concatenate their content
#'
#' Read files one by one, and optionally add text before/after the content. Then
#' combine all content into one character vector.
#' @param files A vector of file paths.
#' @param before,after A function that takes one file path and its content as
#'   the input and returns values to be added before or after the content of the
#'   file. Alternatively, they can be constant values to be added.
#' @return A character vector.
#' @export
#' @examples
#' # two files in this package
#' fs = system.file('scripts', c('call-fun.R', 'child-pids.sh'), package = 'xfun')
#' xfun::read_all(fs)
#'
#' # add file paths before file content and an empty line after content
#' xfun::read_all(fs, before = function(f) paste('#-----', f, '-----'), after = '')
#'
#' # add constants
#' xfun::read_all(fs, before = '/*', after = c('*/', ''))
read_all = function(files, before = function(f, x) NULL, after = function(f, x) NULL) {
  call_fun = function(fun, x, ...) {
    if (is.function(fun)) {
      if (length(formals(fun)) > 1) fun(x, ...) else fun(x)
    } else fun
  }
  x = unlist(lapply(files, function(f) {
    t = read_utf8(f)
    c(call_fun(before, f, t), t, call_fun(after, f, t))
  }))
  raw_string(x)
}

#' Read a text file, process the text with a function, and write the text back
#'
#' Read a text file with the UTF-8 encoding, apply a function to the text, and
#' write back to the original file if the processed text is different with the
#' original input.
#'
#' `sort_file()` is an application of `process_file()`, with the processing
#' function being [sort()], i.e., it sorts the text lines in a file and write
#' back the sorted text.
#' @param file Path to a text file.
#' @param fun A function to process the text.
#' @param x The content of the file.
#' @param ... Arguments to be passed to `process_file()`.
#' @return If `file` is provided, invisible `NULL` (the file is updated as a
#'   side effect), otherwise the processed content (as a character vector).
#' @export
#' @examples f = tempfile()
#' xfun::write_utf8('Hello World', f)
#' xfun::process_file(f, function(x) gsub('World', 'woRld', x))
#' xfun::read_utf8(f)  # see if it has been updated
#' file.remove(f)
process_file = function(file, fun = identity, x = read_utf8(file)) {
  x2 = fun(x)
  if (missing(file)) x2 else {
    if ((length(x2) != length(x)) || !all(x2 == x)) write_utf8(x2, file)
  }
}

#' @rdname process_file
#' @export
sort_file = function(..., fun = sort) {
  process_file(fun = fun, ...)
}

#' Search and replace strings in files
#'
#' These functions provide the "file" version of [gsub()], i.e.,
#' they perform searching and replacement in files via `gsub()`.
#' @param file Path of a single file.
#' @param ... For `gsub_file()`, arguments passed to `gsub()`. For
#'   other functions, arguments passed to `gsub_file()`. Note that the
#'   argument `x` of `gsub()` is the content of the file.
#' @param rw_error Whether to signal an error if the file cannot be read or
#'   written. If `FALSE`, the file will be ignored (with a warning).
#' @param files A vector of file paths.
#' @param dir Path to a directory (all files under this directory will be
#'   replaced).
#' @param recursive Whether to find files recursively under a directory.
#' @param ext A vector of filename extensions (without the leading periods).
#' @param mimetype A regular expression to filter files based on their MIME
#'   types, e.g., `'^text/'` for plain text files.
#' @note These functions perform in-place replacement, i.e., the files will be
#'   overwritten. Make sure you backup your files in advance, or use version
#'   control!
#' @export
#' @examples library(xfun)
#' f = tempfile()
#' writeLines(c('hello', 'world'), f)
#' gsub_file(f, 'world', 'woRld', fixed = TRUE)
#' readLines(f)
gsub_file = function(file, ..., rw_error = TRUE) {
  if (!(file.access(file, 2) == 0 && file.access(file, 4) == 0)) {
    (if (rw_error) stop else warning)('Unable to read or write to ', file)
    if (!rw_error) return(invisible())
  }
  x1 = tryCatch(read_utf8(file, error = TRUE), error = function(e) if (rw_error) stop(e))
  if (is.null(x1)) return(invisible())
  x2 = gsub(x = x1, ...)
  if (!identical(x1, x2)) write_utf8(x2, file)
}

#' @rdname gsub_file
#' @export
gsub_files = function(files, ...) {
  for (f in files) gsub_file(f, ...)
}

#' @rdname gsub_file
#' @export
gsub_dir = function(..., dir = '.', recursive = TRUE, ext = NULL, mimetype = '.*') {
  files = list.files(dir, full.names = TRUE, recursive = recursive)
  if (length(ext)) files = files[file_ext(files) %in% ext]
  if (mimetype != '.*') files = files[grep(mimetype, mime_type(files))]
  gsub_files(files, ...)
}

#' @rdname gsub_file
#' @export
gsub_ext = function(ext, ..., dir = '.', recursive = TRUE) {
  gsub_dir(..., dir = dir, recursive = recursive, ext = ext)
}

#' Perform replacement with `gsub()` on elements matched from `grep()`
#'
#' This function is a shorthand of `gsub(pattern, replacement,
#' grep(pattern, x, value = TRUE))`.
#' @param pattern,replacement,x,... Passed to [grep()] and
#'   `gsub()`.
#' @return A character vector.
#' @export
#' @examples # find elements that matches 'a[b]+c' and capitalize 'b' with perl regex
#' xfun::grep_sub('a([b]+)c', 'a\\U\\1c', c('abc', 'abbbc', 'addc', '123'), perl = TRUE)
grep_sub = function(pattern, replacement, x, ...) {
  x = grep(pattern, x, value = TRUE, ...)
  gsub(pattern, replacement, x, ...)
}

#' Try various methods to download a file
#'
#' Try all possible methods in [download.file()] (e.g., `libcurl`, `curl`,
#' `wget`, and `wininet`) and see if any method can succeed. The reason to
#' enumerate all methods is that sometimes the default method does not work,
#' e.g., <https://stat.ethz.ch/pipermail/r-devel/2016-June/072852.html>.
#' @param url The URL of the file.
#' @param output Path to the output file. By default, it is determined by
#'   [url_filename()].
#' @param ... Other arguments to be passed to [download.file()] (except
#'   `method`).
#' @param .error An error message to signal when the download fails.
#' @note To allow downloading large files, the `timeout` option in [options()]
#'   will be temporarily set to one hour (3600 seconds) inside this function
#'   when this option has the default value of 60 seconds. If you want a
#'   different `timeout` value, you may set it via `options(timeout = N)`, where
#'   `N` is the number of seconds (not 60).
#' @return The `output` file path if the download succeeded, or an error if none
#'   of the download methods worked.
#' @export
download_file = function(
  url, output = url_filename(url), ...,
  .error = 'No download method works (auto/wininet/wget/curl/lynx)'
) {
  if (getOption('timeout') == 60L) {
    opts = options(timeout = 3600)  # one hour
    on.exit(options(opts), add = TRUE)
  }
  dir_create(dirname(output))
  download = function(method = 'auto') suppressWarnings({
    # curl needs to add a -L option to follow redirects
    if (method == 'curl' && is.null(getOption('download.file.extra'))) {
      opts2 = options(download.file.extra = c('-L', '--fail'))
      on.exit(options(opts2), add = TRUE)
    }
    tryCatch(download.file(url, output, ..., method = method), error = function(e) 1L)
  })
  for (method in c('libcurl', 'auto', if (is_windows()) 'wininet')) {
    if (download(method) == 0) return(output)
  }
  # check for libcurl/curl/wget/lynx, call download.file with appropriate method
  for (method in c('curl', 'wget', 'lynx')) {
    if (Sys.which(method) != '' && download(method) == 0) return(output)
  }
  stop(.error)
}

#' Test if a URL is accessible
#'
#' Try to send a `HEAD` request to a URL using
#' [curlGetHeaders()] or the \pkg{curl} package, and see if it
#' returns a successful status code.
#' @param x A URL as a character string.
#' @param use_curl Whether to use the \pkg{curl} package or the
#'   `curlGetHeaders()` function in base R to send the request to the URL.
#'   By default, \pkg{curl} will be used when base R does not have the
#'   \command{libcurl} capability (which should be rare).
#' @param ... Arguments to be passed to `curlGetHeaders()`.
#' @return `TRUE` or `FALSE`.
#' @export
#' @examples xfun::url_accessible('https://yihui.org')
url_accessible = function(x, use_curl = !capabilities('libcurl'), ...) {
  try_status = function(code) tryCatch(code < 400, error = function(e) FALSE)
  if (use_curl) {
    h = curl::new_handle()
    curl::handle_setopt(h, customrequest = 'HEAD', nobody = TRUE)
    try_status(curl::curl_fetch_memory(x, h)$status_code)
  } else {
    # use curlGetHeaders() instead
    try_status(attr2(curlGetHeaders(x, ...), 'status'))
  }
}

#' Get the final destination of a URL
#'
#' If a URL is redirected, query the new location via [curlGetHeaders()],
#' otherwise return the original URL.
#' @param x A character vector of URLs.
#' @param force By default, the query is cached in the current R session. To
#'   bypass the cache, use `force = TRUE`.
#' @return The final destination(s). If the URL returns a status code greater
#'   than or equal to 400, it will throw an error.
#' @export
#' @examples
#' u = 'https://tinytex.yihui.org'  # redirected to https://yihui.org/tinytex/
#' if (url_accessible(u)) url_destination(u)
url_destination = function(x, force = FALSE) {
  unlist(lapply(x, .url_destination, force))
}

.url_destination = local({
  db = list()
  function(x, force = FALSE) {
    if (!force && is.character(db[[x]])) return(db[[x]])
    h = curlGetHeaders(x)
    u = grep_sub('^location:\\s+(.+?)[\r\n]*$', '\\1', h, ignore.case = TRUE)
    db[[x]] <<- if (length(u)) tail(u, 1) else {
      if ((s <- attr2(h, 'status')) >= 400) stop('URL ', x, ' returned status code ', s)
      x
    }
  }
})

#' Generate a message with `cat()`
#'
#' This function is similar to [message()], and the difference is
#' that `msg_cat()` uses [cat()] to write out the message,
#' which is sent to [stdout()] instead of [stderr()]. The
#' message can be suppressed by [suppressMessages()].
#' @param ... Character strings of messages, which will be concatenated into one
#'   string via `paste(c(...), collapse = '')`.
#' @note By default, a newline will not be appended to the message. If you need
#'   a newline, you have to explicitly add it to the message (see
#'   \sQuote{Examples}).
#' @return Invisible `NULL`, with the side-effect of printing the message.
#' @seealso This function was inspired by `rlang::inform()`.
#' @export
#' @examples
#' {
#' # a message without a newline at the end
#' xfun::msg_cat('Hello world!')
#' # add a newline at the end
#' xfun::msg_cat(' This message appears right after the previous one.\n')
#' }
#' suppressMessages(xfun::msg_cat('Hello world!'))
msg_cat = function(...) {
  x = paste(c(...), collapse = '')
  withRestarts({
    signalCondition(simpleMessage(x))
    cat(x)
  }, muffleMessage = function() invisible(NULL))
}

#' Save objects to files and lazy-load them
#'
#' The function [lazy_save()] saves objects to files with incremental integer
#' names (e.g., the first object is saved to `1.rds`, and the second object is
#' saved to `2.rds`, etc.). The function [lazy_load()] lazy-load objects from
#' files saved via [lazy_save()], i.e., a file will not be read until the object
#' is used.
#' @param list A character vector of object names. This list will be written to
#'   an index file with `0` as the base name (e.g., `0.rds`).
#' @param path The path to write files to / read files from.
#' @param method The file save/load method. It can be a string (e.g., `rds`,
#'   `raw`, or `qs`) or a list. See the `rw` argument of [cache_exec()]. By
#'   default, it is automatically detected by checking the existence of the
#'   index file (e.g., `0.rds`, `0.raw`, or `0.qs`).
#' @param envir The environment to [get] or [assign] objects.
#' @return [lazy_save()] returns invisible `NULL`; [lazy_load()] returns the
#'   object names invisibly.
#' @seealso [delayedAssign()]
#' @export
lazy_save = function(list = NULL, path = './', method = 'auto', envir = parent.frame()) {
  m = io_method(method, path)
  idx = lazy_idx(path, m$name)
  m$save(list, idx)
  for (i in seq_along(list)) {
    v = get(list[[i]], envir, inherits = FALSE)
    m$save(v, sprintf('%s%d.%s', path, i, m$name))
  }
}

#' @rdname lazy_save
#' @export
lazy_load = function(path = './', method = 'auto', envir = parent.frame()) {
  m = io_method(method, path)
  idx = lazy_idx(path, m$name)
  vars = m$load(idx)
  .mapply(function(path, name) {
    delayedAssign(name, m$load(path), assign.env = envir)
  }, list(sprintf('%s%d.%s', path, seq_along(vars), m$name), vars), NULL)
  invisible(vars)
}

# use 0.ext as the index file to store the object names
lazy_idx = function(path, ext) paste0(path, '0.', ext)

io_method = function(method, path) {
  if (is.character(method)) {
    if (method == 'auto') {
      for (m in io_methods) if (file_exists(lazy_idx(path, m$name))) {
        method = m$name; break
      }
    }
    if (method == 'auto') method = 'rds'
  }
  if (is.list(method)) method else io_methods[[method]]
}

#' Strict lists
#'
#' A strict list is essentially a normal [list()] but it does not
#' allow partial matching with `$`.
#'
#' To me, partial matching is often more annoying and surprising than
#' convenient. It can lead to bugs that are very hard to discover, and I have
#' been bitten by it many times. When I write `x$name`, I always mean
#' precisely `name`. You should use a modern code editor to autocomplete
#' the `name` if it is too long to type, instead of using partial names.
#' @param ... Objects (list elements), possibly named. Ignored in the
#'   `print()` method.
#' @export
#' @return Both `strict_list()` and `as_strict_list()` return a list
#'   with the class `xfun_strict_list`. Whereas `as_strict_list()`
#'   attempts to coerce its argument `x` to a list if necessary,
#'   `strict_list()` just wraps its argument `...` in a list, i.e., it
#'   will add another list level regardless if `...` already is of type
#'   list.
#' @examples library(xfun)
#' (z = strict_list(aaa = 'I am aaa', b = 1:5))
#' z$a  # NULL!
#' z$aaa  # I am aaa
#' z$b
#' z$c = 'create a new element'
#'
#' z2 = unclass(z)  # a normal list
#' z2$a  # partial matching
#'
#' z3 = as_strict_list(z2) # a strict list again
#' z3$a  # NULL again!
strict_list = function(...) {
  as_strict_list(list(...))
}

# https://twitter.com/xieyihui/status/782462926862954496

#' @param x For `as_strict_list()`, the object to be coerced to a strict
#'   list.
#'
#'   For `print()`, a strict list.
#' @rdname strict_list
#' @export
as_strict_list = function(x) {
  structure(as.list(x), class = 'xfun_strict_list')
}

#' @param name The name (a character string) of the list element.
#' @rdname strict_list
#' @export
`$.xfun_strict_list` = function(x, name) x[[name]]

#' @rdname strict_list
#' @export
print.xfun_strict_list = function(x, ...) {
  print(unclass(x))
}

#' Print a character vector in its raw form
#'
#' The function `raw_string()` assigns the class `xfun_raw_string` to
#' the character vector, and the corresponding printing function
#' `print.xfun_raw_string()` uses `cat(x, sep = '\n')` to write the
#' character vector to the console, which will suppress the leading indices
#' (such as `[1]`) and double quotes, and it may be easier to read the
#' characters in the raw form (especially when there are escape sequences).
#' @param x For `raw_string()`, a character vector. For the print method,
#'   the `raw_string()` object.
#' @param ... Other attributes for `x`.
#' @export
#' @examples library(xfun)
#' raw_string(head(LETTERS))
#' raw_string(c('a "b"', 'hello\tworld!'))
raw_string = function(x, ...) {
  if (is.null(x)) x = as.character(x)
  class(x) = c('xfun_raw_string', class(x))
  structure(x, ...)
}

#' @param ... Other arguments (currently ignored).
#' @rdname raw_string
#' @export
print.xfun_raw_string = function(x, ...) {
  if (length(x)) cat(x, sep = '\n')
  invisible(x)
}

# provide default chunk options to litedown
#' @export
record_print.xfun_raw_string = function(x, ...) {
  cls = intersect(c(class(x), 'record_output'), .record_classes)[1]
  structure(x, class = cls, opts = list(comment = '', attr = attr2(x, 'lang')))
}

#' Encode/decode data into/from base64 encoding.
#'
#' The function `base64_encode()` encodes a file or a raw vector into the
#' base64 encoding. The function `base64_decode()` decodes data from the
#' base64 encoding.
#' @param x For `base64_encode()`, a raw vector. If not raw, it is assumed
#'   to be a file or a connection to be read via `readBin()`. For
#'   `base64_decode()`, a string.
#' @param from If provided (and `x` is not provided), a connection or file
#'   to be read via `readChar()`, and the result will be passed to the
#'   argument `x`.
#' @return `base64_encode()` returns a character string.
#'   `base64_decode()` returns a raw vector.
#' @useDynLib xfun, .registration = TRUE
#' @export
#' @examples xfun::base64_encode(as.raw(1:10))
#' logo = xfun:::R_logo()
#' xfun::base64_encode(logo)
base64_encode = function(x) {
  if (!is.raw(x)) x = read_bin(x)
  .Call('base64_enc', x, PACKAGE = 'xfun')
}

#' @export
#' @rdname base64_encode
#' @examples xfun::base64_decode("AQIDBAUGBwgJCg==")
base64_decode = function(x, from = NA) {
  if (!is.na(from)) {
    if (!missing(x)) stop("Please provide either 'x' or 'from', but not both.")
    x = readChar(from, file.size(from), TRUE)
  }
  if (!is.character(x) || length(x) != 1) stop("'x' must be a single character string.")
  .Call('base64_dec', x, PACKAGE = 'xfun')
}

# an R implementation of base64 encoding by Wush Wu moved from knitr (of
# historic interest only): https://github.com/yihui/knitr/pull/324
base64_encode_r = function(x) {
  if (!is.raw(x)) x = read_bin(x)
  chars = c(LETTERS, letters, 0:9, '+', '/')
  n = length(s <- as.integer(x))
  res = rep(NA, (n + 2) / 3 * 4)
  i = 0L  # index of res vector
  j = 1L  # index of base64_table
  while (n > 2L) {
    res[i <- i + 1L] = chars[s[j] %/% 4L + 1L]
    res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
    res[i <- i + 1L] = chars[4L * (s[j + 1L] %% 16) + s[j + 2L] %/% 64L + 1L]
    res[i <- i + 1L] = chars[s[j + 2L] %% 64L + 1L]
    j = j + 3L
    n = n - 3L
  }
  if (n) {
    res[i <- i + 1L] = chars[s[j] %/% 4L + 1L]
    if (n > 1L) {
      res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
      res[i <- i + 1L] = chars[4L * (s[j + 1L] %% 16) + 1L]
      res[i <- i + 1L] = '='
    } else {
      res[i <- i + 1L] = chars[16 * (s[j] %% 4L) + 1L]
      res[i <- i + 1L] = '='
      res[i <- i + 1L] = '='
    }
  }
  paste(res[!is.na(res)], collapse = '')
}

#' Generate the Data URI for a file
#'
#' Encode the file in the base64 encoding, and add the media type. The data URI
#' can be used to embed data in HTML documents, e.g., in the `src` attribute of
#' the `<img />` tag.
#' @param x A file path.
#' @param type The MIME type of the file, e.g., `"image/png"` for a PNG image
#'   file.
#' @return A string of the form `data:<media type>;base64,<data>`.
#' @export
#' @examples
#' logo = xfun:::R_logo()
#' img = xfun::html_tag('img', src = xfun::base64_uri(logo), alt = "R logo")
#' if (interactive()) xfun::html_view(img)
base64_uri = function(x, type = mime_type(x)) {
  paste0("data:", type, ";base64,", base64_encode(x))
}

#' Get the MIME types of files
#'
#' If the \pkg{mime} package is installed, call [mime::guess_type()], otherwise
#' use the system command `file --mime-type` to obtain the MIME type of a file.
#' Typically, the `file` command exists on *nix. On Windows, the command should
#' exist if Cygwin or Rtools is installed. If it is not found, .NET's
#' `MimeMapping` class will be used instead (which requires the .NET framework).
#' @param x A vector of file paths.
#' @param use_mime Whether to use the \pkg{mime} package.
#' @param empty The MIME type for files without extensions (e.g., `Makefile`).
#'   If `NA`, the type will be obtained from system command. This argument is
#'   used only for `use_mime = FALSE`.
#' @return A character vector of MIME types.
#' @note When querying the MIME type via the system command, the result will be
#'   cached to `xfun:::cache_dir()`. This will make future queries much faster,
#'   since running the command in real time can be a little slow.
#' @export
#' @examplesIf tolower(Sys.getenv('CI')) == 'true'
#' f = list.files(R.home('doc'), full.names = TRUE)
#' mime_type(f)
#' mime_type(f, FALSE)  # don't use mime
#' mime_type(f, FALSE, NA)  # run command for files without extension
mime_type = function(x, use_mime = loadable('mime'), empty = 'text/plain') {
  if (use_mime) return(mime::guess_type(x))
  ext = tolower(file_ext(x))
  res = character(n <- length(x))
  # try to read from cache since running command in real time can be slow
  p = file.path(cache_dir(), 'mime.rds')
  db = mimemap
  for (i in seq_len(n)) {
    e = ext[i]
    m = if (e == '') empty else .mime_type(x[i], e)
    if (is.na(m) && is.na(m <- db[e])) {
      if (file_exists(p)) db = readRDS(p)
      m = db[e]
    }
    if (is.na(m)) {
      m = .mime_cmd(x[i])
      if (grepl('^[a-zA-Z]+/[-a-zA-Z0-9.+]+$', m)) {
        if (e != '') {
          db[e] = m; if (dir_create(dirname(p))) saveRDS(db, p)
        }
      } else warning("'", m, "' does not appear to be a valid MIME type for '", x[i], "'")
    }
    res[i] = m
  }
  res
}

# try to get the MIME type from tools:::mime_type()
.mime_type = function(x, ext = file_ext(x)) {
  if (ext != '' && is.function(f <- asNamespace('tools')$mime_type)) {
    m = f(x, ext)
    if (m != 'text/plain') return(m)
  }
  NA
}

# MIME type from command line `powershell` or `file`
.mime_cmd = function(x, use_file = !is_windows() || Sys.which('file') != '') {
  if (!file.exists(x <- path.expand(x))) stop("The file '", x, "' does not exist.")
  if (use_file) {
    cmd = 'file'
    arg = c('--mime-type', '-b')
  } else {
    # IT said "thou shall not include .ps1 in source package", so I use .txt;
    # then Windoz said "thou shall not execute .txt", so I copy .txt to .ps1;
    # How much of our life has been wasted on making these idiots happy...
    ps1 = tempfile(fileext = '.ps1'); on.exit(unlink(ps1), add = TRUE)
    file.copy(pkg_file('scripts', 'mime-type.txt'), ps1)
    cmd = 'powershell'
    arg = c('-ExecutionPolicy', 'Bypass', '-File', shQuote(ps1))
  }
  if (Sys.which(cmd) == '') stop("The '", cmd, "' command is not found")
  system2(cmd, c(arg, shQuote(x)), stdout = TRUE)[1]
}

# a comprehensive version is mime::mimemap
mimemap = c(
  css = 'text/css', csv = 'text/csv', gif = 'image/gif', jpeg = 'image/jpeg',
  jpg = 'image/jpeg', js = 'text/javascript', png = 'image/png',
  r = 'text/plain', rmd = 'text/x-markdown', svg = 'image/svg+xml',
  ttf = 'font/ttf', woff = 'font/woff', woff2 = 'font/woff2'
)

library(testit)

assert('csv_options() parses chunk options to a list', {
  (csv_options('a-s-d,b=TRUE,c=def') %==% alist(label = 'a-s-d', b = TRUE, c = def))
  (csv_options('a,opt=c(1,3,5)') %==% alist(label = 'a', opt = c(1, 3, 5)))
  (csv_options('label="xx",opt=zz') %==% alist(label = 'xx', opt = zz))
  (csv_options('label=foo') %==% alist(label = 'foo'))
  (csv_options('a,b=2,c="qwer",asdf="efg"') %==%
      alist(label = 'a', b = 2, c = 'qwer', asdf = 'efg'))
  (csv_options('2a') %==% alist(label = '2a'))
  (csv_options('abc-function,fig.path="foo/bar-"') %==%
      alist(label = 'abc-function', fig.path = "foo/bar-"))
  (has_error(csv_options('a,b')))
  (has_error(csv_options('a,b,c=qwer')))
  # empty options string: returns an empty list (no label)
  (length(csv_options('')) %==% 0L)
  # trailing comma: empty option is removed, only named args kept
  res = csv_options('a, b=1,')
  (res$label %==% 'a')
  (res$b %==% 1)
  (length(res) %==% 2L)
  # invalid R syntax in options triggers error with a message
  (has_error(csv_options('a=1+')))
})

assert('divide_chunk() parses YAML-style chunk options', {
  yaml_like = c('#| label: mine', '#| echo: true', '#| fig.width: 8', '1 + 1')
  res = divide_chunk('r', yaml_like, use_yaml = FALSE)
  (res$options$label %==% 'mine')
  (isTRUE(res$options$echo))
  (res$options$fig.width %==% 8L)
  (res$code %==% '1 + 1')
})

assert('divide_chunk() parses CSV-style chunk options', {
  csv_like = c("#| mine, echo = TRUE, fig.width = 8", "1 + 1")
  res = divide_chunk('r', csv_like)
  (res$options$label %==% 'mine')
  (isTRUE(res$options$echo))
  (res$options$fig.width %==% 8)
  (res$code %==% '1 + 1')
})

assert('divide_chunk() returns empty options for empty code', {
  res = divide_chunk('r', character(0))
  (is.null(res$options))
  (length(res$code) %==% 0L)
})

assert('divide_chunk() returns code unchanged when no option comments', {
  code = c('x = 1', 'y = 2')
  res = divide_chunk('r', code)
  (is.null(res$options))
  (res$code %==% code)
})

assert('divide_chunk() falls back to #| for non-R engines', {
  # js uses //| but falls back to #| when strict=FALSE
  js_like = c('#| echo: true', 'console.log(1)')
  res = divide_chunk('js', js_like, use_yaml = FALSE)
  (isTRUE(res$options$echo))
  (res$code %==% 'console.log(1)')
  # with strict=TRUE, #| for a non-# engine is an error
  (has_error(divide_chunk('js', js_like, strict = TRUE, use_yaml = FALSE)))
})

assert('divide_chunk() handles block-comment engines (s2 != empty)', {
  # C engine uses /*| */ style: s2 != ''
  c_like = c('/*| echo: true */', 'int x = 1;')
  res = divide_chunk('c', c_like, use_yaml = FALSE)
  (isTRUE(res$options$echo))
  (res$code %==% 'int x = 1;')
})

assert('divide_chunk() strips leading blank line from code', {
  # a blank line immediately after the options block is removed from code
  yaml_like = c('#| echo: true', '', '1 + 1')
  res = divide_chunk('r', yaml_like, use_yaml = FALSE)
  (isTRUE(res$options$echo))
  (res$code %==% '1 + 1')  # blank line stripped
  (length(res$src) %==% 2L)  # blank line moved to src
})

assert('divide_chunk() when all lines are option comments', {
  # all lines match #| so n2 = length(code)
  all_opts = c('#| echo: true', '#| eval: false')
  res = divide_chunk('r', all_opts, use_yaml = FALSE)
  (isTRUE(res$options$echo))
  (isFALSE(res$options$eval))
  (length(res$code) %==% 0L)
})

assert('divide_chunk() normalizes id field to label', {
  yaml_like = c('#| id: my-chunk', '1 + 1')
  res = divide_chunk('r', yaml_like, use_yaml = FALSE)
  (res$options$label %==% 'my-chunk')
  (is.null(res$options$id))
})

#' An alternative to sessionInfo() to print session information
#'
#' This function tweaks the output of [sessionInfo()]: (1) It adds
#' the RStudio version information if running in the RStudio IDE; (2) It removes
#' the information about matrix products, BLAS, and LAPACK; (3) It removes the
#' names of base R packages; (4) It prints out package versions in a single
#' group, and does not differentiate between loaded and attached packages.
#'
#' It also allows you to only print out the versions of specified packages (via
#' the `packages` argument) and optionally their recursive dependencies.
#' For these specified packages (if provided), if a function
#' `xfun_session_info()` exists in a package, it will be called and
#' expected to return a character vector to be appended to the output of
#' `session_info()`. This provides a mechanism for other packages to inject
#' more information into the `session_info` output. For example,
#' \pkg{rmarkdown} (>= 1.20.2) has a function `xfun_session_info()` that
#' returns the version of Pandoc, which can be very useful information for
#' diagnostics.
#' @param packages A character vector of package names, of which the versions
#'   will be printed. If not specified, it means all loaded and attached
#'   packages in the current R session.
#' @param dependencies Whether to print out the versions of the recursive
#'   dependencies of packages.
#' @return A character vector of the session information marked as
#'   [raw_string()].
#' @export
#' @examplesIf interactive()
#' xfun::session_info()
#' if (xfun::loadable('MASS')) xfun::session_info('MASS')
session_info = function(packages = NULL, dependencies = TRUE) {
  res = sessionInfo()
  res$matprod = res$BLAS = res$LAPACK = NULL
  res$running = paste(c(res$running, if (Sys.getenv('POSITRON') == '1') {
    c(', Positron ', Sys.getenv('POSITRON_VERSION'))
  } else if (loadable('rstudioapi') && rstudioapi::isAvailable()) {
    c(', RStudio ', as.character(rstudioapi::getVersion()))
  }), collapse = '')

  tweak_info = function(obj, extra = NULL) {
    res = capture.output(print(obj, tzone = FALSE))
    i = grep('^(attached base packages|Matrix products):\\s*$', res, ignore.case = TRUE)
    if (length(i)) res = res[-c(i, i + 1)]
    res = gsubi('^\\s*locale:\\s*$', 'Locale:', res)
    res = gsub('^\\s*\\[[0-9]+]\\s*', '  ', res)  # remove vector indices like [1]
    res = gsubi('^\\s*other attached packages:\\s*$', 'Package version:', res)
    # print the locale info on a single line if possible
    if (length(i <- which(res == 'Locale:')) == 1 && res[i + 2] == '') {
      res[i] = paste(res[i], gsub('\\s*/\\s*', ' / ', gsub('^\\s+', '', res[i + 1])))
      res = res[-(i + 1)]
    }
    # remove extra blank lines
    if ((n <- length(res)) > 1) {
      i = is_blank(res)
      res = res[!c(FALSE, i[1:(n-1)] & i[2:n])]
    }
    raw_string(c(res, extra))
  }

  version_info = function(pkgs) {
    res = lapply(pkgs, function(p) {
      list(Version = as.character(packageVersion(p)), Package = p)
    })
    as.list(setNames(res, pkgs))
  }

  res$basePkgs = raw_string(list())
  info = c(res$otherPkgs, res$loadedOnly)
  packages = setdiff(packages, '')  # remove empty strings (#65)
  if (length(packages) > 0) {
    info = info[intersect(names(info), packages)]
    info = c(info, version_info(setdiff(packages, names(info))))
  }
  res$loadedOnly = NULL
  if (dependencies) {
    deps = pkg_dep(names(info), installed.packages(), recursive = TRUE)
    deps = sort(setdiff(deps, names(info)))
    info = c(info, version_info(deps))
  }
  if (length(packages) > 0 || dependencies) info = info[sort(names(info))]
  res$otherPkgs = info
  extra = unlist(lapply(packages, function(p) tryCatch(
    c('', getFromNamespace('xfun_session_info', p)()), error = function(e) NULL)
  ))

  tweak_info(res, extra)
}

#' Perform a task once in an R session
#'
#' Perform a task once in an R session, e.g., emit a message or warning. Then
#' give users an optional hint on how not to perform this task at all.
#' @param task Any R code expression to be evaluated once to perform a task,
#'   e.g., `warning('Danger!')` or `message('Today is ', Sys.Date())`.
#' @param option An R option name. This name should be as unique as possible in
#'   [options()]. After the task has been successfully performed,
#'   this option will be set to `FALSE` in the current R session, to
#'   prevent the task from being performed again the next time when
#'   `do_once()` is called.
#' @param hint A character vector to provide a hint to users on how not to
#'   perform the task or see the message again in the current R session. Set
#'   `hint = ""` if you do not want to provide the hint.
#' @return The value returned by the `task`, invisibly.
#' @export
#' @examples
#' do_once(message("Today's date is ", Sys.Date()), "xfun.date.reminder")
#' # if you run it again, it will not emit the message again
#' do_once(message("Today's date is ", Sys.Date()), "xfun.date.reminder")
#'
#' do_once({Sys.sleep(2); 1 + 1}, "xfun.task.1plus1")
#' do_once({Sys.sleep(2); 1 + 1}, "xfun.task.1plus1")
do_once = function(task, option, hint = c(
  'You will not see this message again in this R session.',
  'If you never want to see this message,',
  sprintf('you may set options(%s = FALSE) in your .Rprofile.', option)
)) {
  if (identical(getOption(option), FALSE)) return(invisible())
  task
  hint = paste(hint, collapse = ' ')
  if (hint != '') message(hint)
  options(setNames(list(FALSE), option))
  invisible(task)
}

library(testit)

assert('prose_index() works', {
  x = c('a', '```', 'b', '```', 'c')
  (prose_index(x) %==% c(1L, 5L))

  x = c('a', '````', '```r', '1+1', '```', '````', 'c')
  (prose_index(x) %==% c(1L, 7L))

  x = c('a', '``', 'b', '``', 'c')
  (prose_index(x) %==% seq_along(x))

  # a character vector of length zero
  (prose_index(character()) %==% integer())

  # one backbrick
  x = c('`', 'a', '`')
  (prose_index(x) %==% seq_along(x))

  # two backbrick
  x = c('``', 'a', '``')
  (prose_index(x) %==% seq_along(x))

  # no code fences
  x = c('a', 'b')
  (prose_index(x) %==% seq_along(x))

  # two code fences
  x = c('```', 'b', '```', '```', 'd', '```')
  (prose_index(x) %==% integer())

  # code fences commented out
  x = c('```', 'b', '```', '<!--```', 'd', '```-->')
  (prose_index(x) %==% 4:6)

  # if the code fences are not balanced
  x = c('a', '```', 'b', '``', 'c')
  (has_warning(prose_index(x)))
  (prose_index(x) %==% seq_along(x))

  # recognize ``` ``text`` ```
  x = c('a', '``` md', 'b', '```', '``` ``text`` ```')
  (prose_index(x) %==% c(1L, 5L))

  # <pre> should also be treated as code blocks
  x = c('<pre><code>```', '```', '```', '</code></pre>')
  (prose_index(x) %==% integer())
})


assert('protect_math() puts inline math expressions in backticks', {
  (protect_math('$x$') %==% '`\\(x\\)`')
  (protect_math('$x$', '===') %==% '`===\\(x\\)===`')
  (protect_math('($x$)') %==% '(`\\(x\\)`)')
  (protect_math(' $x$') %==% ' `\\(x\\)`')
  (protect_math('.$x$') %==% '.$x$')
  (protect_math('$`x`$') %==% '$`x`$')
  (protect_math('hi $x$ a') %==% 'hi `\\(x\\)` a')
  (protect_math('$ a $') %==% '$ a $')
  (protect_math(' $a$') %==% ' `\\(a\\)`')
  (protect_math('$ x$') %==% '$ x$')
  (protect_math('$x $') %==% '$x $')
  (protect_math('b$a$') %==% 'b$a$')  # no space before $; unlikely to be math
  (protect_math('`$a$`') %==% '`$a$`')
  (protect_math('hi $x$9') %==% 'hi $x$9')
  (protect_math('$500 $600') %==% '$500 $600')

  (protect_math('$$a$$') %==% '`$$a$$`')
  (protect_math('$$a$') %==% '$$a$')
  (protect_math('($$a$$)') %==% '(`$$a$$`)')
  (protect_math(' $$a$$') %==% ' `$$a$$`')
  (protect_math('.$$a$$') %==% '.$$a$$')
  (protect_math('hi $$\alpha$$') %==% 'hi `$$\alpha$$`')
  (protect_math('hi $$\alpha $$') %==% 'hi $$\alpha $$')
  (protect_math('hi $$ \alpha$$') %==% 'hi $$ \alpha$$')
  (protect_math('hi $$\alpha$$ and $$ \alpha$$') %==% 'hi `$$\alpha$$` and $$ \alpha$$')
  (protect_math(c('$$a', 'b$$')) %==% c('`$$a', 'b$$`'))
  (protect_math(c('$$a', 'b$$'), 'asdf') %==% c('`asdf$$a', 'b$$asdf`'))
  (protect_math(c('$$a', 'b$$'), use_block = TRUE) %==% c('```{.md-math}\n$$a', 'b$$\n```'))
  (protect_math(c('$$a', 'b$$'), 'asdf', use_block = TRUE) %==% c('```{.md-math .asdf}\n$$a', 'b$$\n```'))
  (protect_math(c('  $$a', '  b$$'), use_block = TRUE) %==% c('  ```{.md-math}\n  $$a', '  b$$\n  ```'))
})


assert('block_attr(x) turns a character vector into Pandoc attributes', {
  (block_attr(NULL) %==% '')
  (block_attr('.a') %==% ' {.a}')
  (block_attr('.a b="11"') %==% ' {.a b="11"}')
  (block_attr(c('.a', 'b="11"')) %==% ' {.a b="11"}')
})

assert('make_fence() uses the right number of fence characters', {
  (make_fence('1+1') %==% '```')
  (make_fence(c('1+1', '`````')) %==% '``````')
  (make_fence(':::') %==% '```')
  (make_fence(':::', ':') %==% '::::')
})

assert('fenced_block() wraps content inside fences', {
  (fenced_block('1+1') %==% c('', '```', '1+1', '```'))
  (fenced_block('<i>info</i>', '=html') %==% c('', '``` {=html}', '<i>info</i>', '```'))
  (fenced_block('1+1', c('.lang', '#id', 'foo="BAR"')) %==% c('', '``` {.lang #id foo="BAR"}', '1+1', '```'))
})

assert('fenced_div() uses colons as fence character', {
  res = fenced_div('content')
  (res %==% c('', ':::', 'content', ':::'))
})

assert('md_table() generates a Markdown pipe table', {
  # basic usage
  x = data.frame(a = 1:3, b = c('x', 'y', 'z'), stringsAsFactors = FALSE)
  res = md_table(x)
  (length(res) %==% 5L)  # header + separator + 3 rows
  (grepl('^\\|', res[1]))  # starts with pipe
  (grepl('\\|$', res[1]))  # ends with pipe
  # numeric column is right-aligned
  (grepl('--:', res[2]))

  # empty columns
  (md_table(data.frame()) %==% character())

  # error for non-2D objects
  (has_error(md_table(1:5)))

  # NA values in table
  x2 = data.frame(a = c(1, NA, 3), b = c('x', 'y', NA))
  res2 = md_table(x2)
  (any(grepl('NA', res2) | grepl('', res2)))  # NA shown as empty by default

  # row names (non-numeric) are included as first column
  x3 = data.frame(a = 1:2, row.names = c('r1', 'r2'))
  res3 = md_table(x3)
  (grepl('r1', res3[3]))  # first data row (after header + separator) contains r1

  # limit parameter
  x4 = data.frame(a = 1:10)
  res4 = md_table(x4, limit = 4)
  (length(res4) < 12L)  # fewer rows than full table
})

assert('fenced_div() uses colons and supports attributes', {
  (fenced_div('text', '.note') %==% c('', '::: {.note}', 'text', ':::'))
})

assert('md_table() with newlines, NA, digits, and pipes', {
  x = data.frame(a = 'line1\nline2', stringsAsFactors = FALSE)
  (!any(grepl('\n', md_table(x))))
  x = data.frame(a = c(1, NA))
  (any(grepl('N/A', md_table(x, na = 'N/A'))))
  x = data.frame(a = 3.14159265)
  res = md_table(x, digits = 2)
  (any(grepl('3.14', res)))
  (!any(grepl('3.1415', res)))
  x = data.frame(a = 'a|b', stringsAsFactors = FALSE)
  (any(grepl('\\\\[|]', md_table(x))))
})

assert('md_table() escapes special Markdown characters', {
  x = data.frame(a = '<TaskRegr:mtcars[0]>', stringsAsFactors = FALSE)
  res = md_table(x, escape = TRUE)
  (any(grepl('\\\\<', res)))
  (any(grepl('\\\\\\[', res)))
  # escape by column name
  x = data.frame(a = '*bold*', b = '*bold*', stringsAsFactors = FALSE)
  res = md_table(x, escape = 'a')
  (any(grepl('\\\\[*]', res)))
  # column b should not be escaped
  (grepl('\\*bold\\*\\|$', res[3]))
  # escape = FALSE (default) does not escape
  (!any(grepl('\\\\<', md_table(data.frame(a = '<x>'), escape = FALSE))))
})

assert('md_table() with column limit', {
  x = as.data.frame(matrix(1:20, nrow = 2))
  (any(grepl('[.][.][.]', md_table(x, limit = c(0, 4)))))
})

assert('md_table() with row limit shows ellipsis', {
  (any(grepl('vellip', md_table(data.frame(a = 1:20), limit = 4))))
})

assert('protect_math() does not touch code blocks', {
  (protect_math(c('```', '$x$', '```')) %==% c('```', '$x$', '```'))
})

assert('protect_math() handles \\begin/\\end environments', {
  x = c('\\begin{equation}', 'x = y', '\\end{equation}')
  res = protect_math(x)
  (grepl('^`', res[1]))
  (grepl('`$', res[3]))
})

assert('block_attr() works', {
  (block_attr('#myid') %==% ' {#myid}')
  (block_attr('lang') %==% ' lang')
})

assert('embed_file() produces an anchor tag with base64 data', {
  f = tempfile(fileext = '.txt'); writeLines('hello', f)
  link = embed_file(f)
  (inherits(link, 'xfun_html'))
  (grepl('<a ', link) && grepl('download=', link) && grepl('data:text/plain;base64,', link))
  unlink(f)
})

assert('embed_files() compresses multiple files into a zip link', {
  d = tempfile(); dir.create(d)
  f1 = file.path(d, 'a.txt'); writeLines('aaa', f1)
  f2 = file.path(d, 'b.txt'); writeLines('bbb', f2)
  link = embed_files(c(f1, f2))
  (inherits(link, 'xfun_html'))
  (grepl('\\.zip', link))
  unlink(d, recursive = TRUE)
})

assert('embed_dir() compresses a directory into a zip link', {
  d = tempfile(); dir.create(d)
  writeLines('aaa', file.path(d, 'file.txt'))
  link = embed_dir(d)
  (inherits(link, 'xfun_html'))
  (grepl('\\.zip', link))
  unlink(d, recursive = TRUE)
})

#' Print a web page to PDF/PNG/JPEG
#'
#' Print a web page via a headless browser such as Chromium or Google Chrome.
#' @param url Path or URL to the page to be printed.
#' @param output An output filename. If only an extension is provided, the
#'   filename will be inferred from `url` via [url_filename()]. Only `.pdf`,
#'   `.png`, and `.jpeg` are supported.
#' @param args Command-line arguments to be passed to the headless browser. The
#'   default arguments can be found in `xfun:::browser_args()`. You may pass
#'   additional arguments on top of these via, e.g., `c('default',
#'   '--no-pdf-header-footer')`, or completely override the default arguments by
#'   providing a vector of other arguments.
#' @param window_size The browser window size when printing to PNG/JPEG.
#' @param browser Path to the web browser. By default, the browser is found via
#'   `xfun:::find_browser()`. If it cannot be found, you may set the global
#'   option `options(xfun.browser = )` or environment variable `R_XFUN_BROWSER`
#'   to the path of the browser, or simply pass the path to the `browser`
#'   argument.
#' @return The `output` path if the web page is successfully printed.
#' @export
#' @examplesIf interactive()
#' xfun::browser_print('https://www.r-project.org')
browser_print = function(
  url, output = '.pdf', args = 'default', window_size = c(1280, 1024),
  browser = env_option('xfun.browser', find_browser())
) {
  if (!file.exists(browser)) browser = Sys.which(browser)
  if (!utils::file_test('-x', browser)) stop('The browser is not executable: ', browser)
  if (sans_ext(output) == '') output = with_ext(url_filename(url), output)
  to_pdf = tolower(file_ext(output)) == 'pdf'
  if (!to_pdf && !grepl('--window-size=', args))
    args = c(args, paste0('--window-size=', paste(window_size, collapse = ',')))
  if ('default' %in% args) {
    args = setdiff(c(
      args, browser_args(),
      sprintf('--%s="%s"', if (to_pdf) 'print-to-pdf' else 'screenshot', normalize_path(output)),
      shQuote(url)
    ), 'default')
  }
  if (system2(browser, args, stderr = FALSE) == 0) output else stop('Failed to print to ', output)
}

browser_args = function() c(
  proxy_args(), if (is_windows()) '--no-sandbox', '--headless',
  '--no-first-run', '--no-default-browser-check', '--hide-scrollbars'
)

find_browser = function() {
  switch(
    .Platform$OS.type,
    windows = {
      res = unlist(lapply(c('ChromeHTML', 'MSEdgeHTM'), function(x) {
        res = tryCatch({
          unlist(utils::readRegistry(sprintf('%s\\shell\\open\\command', x), 'HCR'))
        }, error = function(e) '')
        res = unlist(strsplit(res, '"'))
        res = head(res[file.exists(res)], 1)
      }))
      if (length(res) == 0) stop(
        'Cannot find Google Chrome or Edge automatically from the Windows Registry Hive. ',
        "Please pass the full path of chrome.exe or msedge.exe to the 'browser' argument ",
        "or to the environment variable 'R_XFUN_BROWSER'."
      )
      res[1]
    },
    unix = {
      for (i in c(
        '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome',
        'google-chrome', 'chromium-browser', 'chromium', 'google-chrome-stable')) {
        if ((res <- Sys.which(i)) != '') break
      }
      if (res == '') stop('Cannot find Chromium or Google Chrome')
      res
    },
    stop('Your platform is not supported')
  )
}

proxy_args = function() {
  x = Sys.getenv(c('https_proxy', 'HTTPS_PROXY', 'http_proxy', 'HTTP_PROXY'))
  x = x[x != '']
  if (length(x) == 0) return()
  c(
    paste0('--proxy-server=', x[1]),
    paste0('--proxy-bypass-list=', paste(no_proxy(), collapse = ';'))
  )
}

no_proxy = function() {
  x = do.call(c, strsplit(Sys.getenv(c('no_proxy', 'NO_PROXY')), '[,;]'))
  x = c('localhost', '127.0.0.1', x)
  unique(x)
}

library(testit)

assert('release_versions() handles X.Y and X.Y.0 formats', {
  r = release_versions('0.59')
  (r$next_ver %==% '0.60')
  (r$patch %==% '0.59.1')

  r = release_versions('1.2.0')
  (r$next_ver %==% '1.3.0')
  (r$patch %==% '1.2.1')

  r = release_versions('10.0')
  (r$next_ver %==% '10.1')
  (r$patch %==% '10.0.1')

  (has_error(release_versions('1.2.3')))
  (has_error(release_versions('1')))
  (has_error(release_versions('1.2.1')))
})

assert('news_tpl() detects header format and falls back correctly', {
  hdr = function(news, pkg, ver, v) sprintf(news_tpl(news, pkg, ver), pkg, v)

  (hdr(c('# CHANGES IN foo VERSION 1.0', '', 'text'), 'foo', '1.0', '1.1') %==% '# CHANGES IN foo VERSION 1.1')
  (hdr(c('# foo 1.0', '', 'text'), 'foo', '1.0', '1.1') %==% '# foo 1.1')
  (hdr(c('# foo (version 1.0)', '', 'text'), 'foo', '1.0', '1.1') %==% '# foo (version 1.1)')

  # pkg name must match as whole word (not substring of another word)
  (hdr(c('# CHANGES in GE 1.1'), 'GE', '1.1', '1.2') %==% '# CHANGES in GE 1.2')

  # fallback: empty NEWS or unrecognized heading
  (hdr(character(0), 'foo', '1.0', '1.1') %==% '# CHANGES IN foo VERSION 1.1')
  (hdr(c('# Some unrelated heading'), 'foo', '1.0', '1.1') %==% '# CHANGES IN foo VERSION 1.1')
})

assert('github_token() returns the token from environment variables', {
  # save original env vars
  orig = Sys.getenv(c('GITHUB_PAT', 'GITHUB_TOKEN', 'GH_TOKEN'), names = TRUE)
  on.exit({
    do.call(Sys.setenv, as.list(orig))
  }, add = TRUE)
  # unset all three
  Sys.unsetenv(c('GITHUB_PAT', 'GITHUB_TOKEN', 'GH_TOKEN'))
  (github_token() %==% '')

  Sys.setenv(GITHUB_PAT = 'pat_token')
  (github_token() %==% 'pat_token')

  Sys.unsetenv('GITHUB_PAT')
  Sys.setenv(GITHUB_TOKEN = 'token_val')
  (github_token() %==% 'token_val')

  Sys.unsetenv('GITHUB_TOKEN')
  Sys.setenv(GH_TOKEN = 'gh_token_val')
  (github_token() %==% 'gh_token_val')

  # priority: GITHUB_PAT > GITHUB_TOKEN > GH_TOKEN
  Sys.setenv(GITHUB_PAT = 'pat_1st', GITHUB_TOKEN = 'token_2nd', GH_TOKEN = 'gh_3rd')
  (github_token() %==% 'pat_1st')
})

library(testit)

assert('html_doc() wraps content in an HTML document', {
  x = html_doc('<p>hi</p>', 'My Title')
  (any(grepl('<!DOCTYPE html>', x)))
  (any(grepl('<title>My Title</title>', x)))
  (any(grepl('<p>hi</p>', x)))
})

assert('fileinfo_table() builds a directory listing table', {
  d = tempfile(); dir.create(d)
  on.exit(unlink(d, recursive = TRUE))
  writeLines('a', file.path(d, 'a.txt'))
  dir.create(file.path(d, 'sub'))
  info = file.info(list.files(d, all.files = TRUE, full.names = TRUE))
  tbl = fileinfo_table(info)
  (any(grepl('<table>', tbl)))
  (any(grepl('a.txt', tbl)))
  # directories get a trailing slash
  (any(grepl('sub/', tbl)))
})

assert('resolve_path() classifies files, dirs, redirects, and 404s', {
  d = tempfile(); dir.create(d)
  on.exit(unlink(d, recursive = TRUE))
  writeLines('hello', file.path(d, 'a.txt'))
  writeLines('<h1>Index</h1>', file.path(d, 'index.html'))
  dir.create(file.path(d, 'sub'))
  writeLines('sub-content', file.path(d, 'sub', 'b.txt'))

  in_dir(d, {
    # root with index.html
    r = resolve_path('.')
    (r$action %==% 'file')
    (basename(r$path) %==% 'index.html')

    # explicit file
    r = resolve_path('a.txt')
    (r$action %==% 'file')
    (basename(r$path) %==% 'a.txt')

    # directory without trailing slash → caller redirects
    r = resolve_path('sub')
    (r$action %==% 'add-slash')

    # directory listing (trailing slash, no index.html)
    r = resolve_path('sub/')
    (r$action %==% 'payload')
    (r$mime %==% 'text/html')
    (grepl('Index of', r$body))
    (grepl('b.txt', r$body))

    # missing file gives 404 payload
    r = resolve_path('missing.txt')
    (r$action %==% 'payload')
    (r$status %==% 404L)
    (grepl('Not found', r$body))

    # favicon.ico falls back to R's built-in
    r = resolve_path('favicon.ico')
    (r$action %==% 'file')
    (basename(r$path) %==% 'favicon.ico')
    (r$mime %==% 'image/x-icon')

    # custom 404.html → caller redirects
    writeLines('<h1>404</h1>', '404.html')
    r = resolve_path('nope.html')
    (r$action %==% '404-redirect')
  })
})

assert('.resolve_path_to_httpd() maps to R httpd response shape', {
  d = tempfile(); dir.create(d)
  on.exit(unlink(d, recursive = TRUE))
  writeLines('hi', file.path(d, 'a.txt'))
  dir.create(file.path(d, 'sub'))

  in_dir(d, {
    # file response
    r = .resolve_path_to_httpd('a.txt')
    (basename(r$file) %==% 'a.txt')

    # add-slash → 301 with Location header
    r = .resolve_path_to_httpd('sub')
    (r$`status code` %==% 301L)
    (grepl('^Location: /sub/$', r$header))

    # 404 payload
    r = .resolve_path_to_httpd('missing.txt')
    (r$`status code` %==% 404L)
    (r$`content-type` %==% 'text/plain')

    # 404.html redirect
    writeLines('x', '404.html')
    r = .resolve_path_to_httpd('nope.html')
    (r$`status code` %==% 302L)
    (r$header %==% 'Location: /404.html')
  })
})

#' Manipulate filename extensions
#'
#' Functions to obtain (`file_ext()`), remove (`sans_ext()`), and
#' change (`with_ext()`) extensions in filenames.
#'
#' `file_ext()` is similar to [tools::file_ext()], and
#' `sans_ext()` is similar to [tools::file_path_sans_ext()].
#' The main differences are that they treat `tar.(gz|bz2|xz)` and
#' `nb.html` as extensions (but functions in the \pkg{tools} package
#' doesn't allow double extensions by default), and allow characters `~`
#' and `#` to be present at the end of a filename.
#' @param x A character of file paths.
#' @param extra Extra characters to be allowed in the extensions. By default,
#'   only alphanumeric characters are allowed (and also some special cases in
#'   \sQuote{Details}). If other characters should be allowed, they can be
#'   specified in a character string, e.g., `"-+!_#"`.
#' @export
#' @return A character vector of the same length as `x`.
#' @examples library(xfun)
#' p = c('abc.doc', 'def123.tex', 'path/to/foo.Rmd', 'backup.ppt~', 'pkg.tar.xz')
#' file_ext(p); sans_ext(p); with_ext(p, '.txt')
#' with_ext(p, c('.ppt', '.sty', '.Rnw', 'doc', 'zip')); with_ext(p, 'html')
#'
#' # allow for more characters in extensions
#' p = c('a.c++', 'b.c--', 'c.e##')
#' file_ext(p)  # -/+/# not recognized by default
#' file_ext(p, extra = '-+#')
file_ext = function(x, extra = '') {
  ext = character(length(x))
  i = grep(r <- reg_path(extra), x)
  ext[i] = sub(r, '\\3', x[i])
  ext
}

#' @rdname file_ext
#' @export
sans_ext = function(x, extra = '') {
  sub(reg_path(extra), '\\1', x)
}

#' @param ext A vector of new extensions. It must be either of length 1, or the
#'   same length as `x`.
#' @rdname file_ext
#' @export
with_ext = function(x, ext, extra = '') {
  if (anyNA(ext)) stop("NA is not allowed in 'ext'")
  n1 = length(x); n2 = length(ext)
  if (n1 * n2 == 0) return(x)
  i = !grepl('^[.]', ext) & ext != ''
  ext[i] = paste0('.', ext[i])

  if (all(ext == '')) ext = ''
  r = sub('[$]$', '?$', reg_ext(extra))  # make extensions in 'x' optional
  if (length(ext) == 1) return(sub(r, ext, x))

  if (n1 > 1 && n1 != n2) stop("'ext' must be of the same length as 'x'")
  mapply(sub, r, ext, x, USE.NAMES = FALSE)
}

# regex to extract base path and extension from a file path
reg_ext  = function(extra = '') {
  sprintf('([.](([%s[:alnum:]]+|tar[.](gz|bz2|xz)|nb[.]html)[~#]?))$', extra)
}
reg_path = function(...) paste0('^(.*?)', reg_ext(...))

#' Normalize paths
#'
#' A wrapper function of `normalizePath()` with different defaults.
#' @param x,winslash,must_work Arguments passed to
#'   [normalizePath()].
#' @param resolve_symlink Whether to resolve symbolic links.
#' @export
#' @examples library(xfun)
#' normalize_path('~')
normalize_path = function(x, winslash = '/', must_work = FALSE, resolve_symlink = TRUE) {
  if (!resolve_symlink) {
    # apply the trick on all files on Windows since Sys.readlink() doesn't work
    # and we can't know which files are symlinks
    i = if (is_windows()) file_test('-f', x) else is_symlink(x)
    b = basename(x[i])
    x[i] = dirname(x[i])  # normalize the dirs of symlinks instead
  }
  res = normalizePath(x, winslash = winslash, mustWork = must_work)
  if (is_windows()) res[is.na(x)] = NA
  if (!resolve_symlink) {
    res[i] = file.path(res[i], b, fsep = winslash)
  }
  res
}

is_symlink = function(x) {
  !is.na(y <- Sys.readlink(x)) & (y != '')
}

#' Test if two paths are the same after they are normalized
#'
#' Compare two paths after normalizing them with the same separator (`/`).
#' @param p1,p2 Two vectors of paths.
#' @param ... Arguments to be passed to [normalize_path()].
#' @export
#' @examples library(xfun)
#' same_path('~/foo', file.path(Sys.getenv('HOME'), 'foo'))
same_path = function(p1, p2, ...) {
  normalize_path(p1, ...) == normalize_path(p2, ...)
}

#' Find file paths that exist
#'
#' This is a shorthand of `x[file.exists(x)]`, and optionally returns the
#' first existing file path.
#' @param x A vector of file paths.
#' @param first Whether to return the first existing path. If `TRUE` and no
#'   specified files exist, it will signal an error unless the argument
#'   `error = FALSE`.
#' @param error Whether to throw an error when `first = TRUE` but no files
#'   exist. It can also take a character value, which will be used as the error
#'   message.
#' @return A vector of existing file paths.
#' @export
#' @examples
#' xfun::existing_files(c('foo.txt', system.file('DESCRIPTION', package='xfun')))
existing_files = function(x, first = FALSE, error = TRUE) {
  x = x[file_exists(x)]
  if (!first) return(x)
  x = head(x, 1)
  if (length(x) != 1 && !identical(error, FALSE)) {
    if (isTRUE(error)) error = 'None of the specified files exist.'
    stop(error, call. = FALSE)
  }
  x
}

#' Return the (possible) root directory of a project
#'
#' Given a path of a file (or dir) in a potential project (e.g., an R package or
#' an RStudio project), return the path to the project root directory.
#'
#' The search for the root directory is performed by a series of tests,
#' currently including looking for a \file{DESCRIPTION} file that contains
#' `Package: *` (which usually indicates an R package), and a
#' \file{*.Rproj} file that contains `Version: *` (which usually indicates
#' an RStudio project). If files with the expected patterns are not found in the
#' initial directory, the search will be performed recursively in upper-level
#' directories.
#' @param path The initial path to start the search. If it is a file path, its
#'   parent directory will be used.
#' @param rules A matrix of character strings of two columns: the first column
#'   contains regular expressions to look for filenames that match the patterns,
#'   and the second column contains regular expressions to match the content of
#'   the matched files. The regular expression can be an empty string, meaning
#'   that it will match anything.
#' @return Path to the root directory if found, otherwise `NULL`.
#' @export
#' @note This function was inspired by the \pkg{rprojroot} package, but is much
#'   less sophisticated. It is a rather simple function designed to be used in
#'   some of packages that I maintain, and may not meet the need of general
#'   users until this note is removed in the future (which should be unlikely).
#'   If you are sure that you are working on the types of projects mentioned in
#'   the \sQuote{Details} section, this function may be helpful to you,
#'   otherwise please consider using \pkg{rprojroot} instead.
proj_root = function(path = './', rules = root_rules) {
  path = normalize_path(path)
  dir = if (dir_exists(path)) path else dirname(path)
  if (same_path(dir, file.path(dir, '..'))) return()
  if (is.null(dim(rules))) dim(rules) = c(1, length(rules))
  for (i in seq_len(nrow(rules))) {
    file = rules[i, 1]; pattern = rules[i, 2]
    for (f in list.files(dir, file, full.names = TRUE)) {
      if (pattern == '' || length(grep(pattern, read_utf8(f)))) return(dir)
    }
  }
  proj_root(dirname(dir), rules)
}

#' @rdname proj_root
#' @export
root_rules = matrix(c(
  '^DESCRIPTION$', '^Package: ',
  '.+[.]Rproj$',   '^Version: '
), ncol = 2, byrow = TRUE, dimnames = list(NULL, c('file', 'pattern')))

#' Get the relative path of a path relative to a directory
#'
#' Given a directory, return the relative path that is relative to this
#' directory. For example, the path \file{foo/bar.txt} relative to the directory
#' \file{foo/} is \file{bar.txt}, and the path \file{/a/b/c.txt} relative to
#' \file{/d/e/} is \file{../../a/b/c.txt}.
#' @param dir Path to a directory.
#' @param x A vector of paths to be converted to relative paths.
#' @param use.. Whether to use double-dots (\file{..}) in the relative path. A
#'   double-dot indicates the parent directory (starting from the directory
#'   provided by the `dir` argument).
#' @param error Whether to signal an error if a path cannot be converted to a
#'   relative path.
#' @return A vector of relative paths if the conversion succeeded; otherwise the
#'   original paths when `error = FALSE`, and an error when `error =
#'   TRUE`.
#' @export
#' @examples
#' xfun::relative_path('foo/bar.txt', 'foo/')
#' xfun::relative_path('foo/bar/a.txt', 'foo/haha/')
#' xfun::relative_path(getwd())
relative_path = function(x, dir = '.', use.. = TRUE, error = TRUE) {
  res = x
  for (i in seq_along(x)) res[i] = relative_path_one(x[i], dir, use.., error)
  res
}

relative_path_one = function(x, dir, use.., error) {
  # on Windows, if a relative path doesn't exist, normalizePath() will use
  # getwd() as its parent dir; however, normalizePath() just returns the
  # relative path on *nix, and we have to assume it's relative to getwd()
  abs_path = function(p) {
    if (!file.exists(p) && is_unix() && is_rel_path(p)) p = file.path(getwd(), p)
    normalize_path(p)
  }
  p = abs_path(x); n1 = nchar(p)
  if ((n1 <- nchar(p)) == 0) return(x)  # not sure what you mean
  d = abs_path(dir); n2 = nchar(d)
  if (is_sub_path(p, d, n2)) {
    p2 = get_subpath(p, n1, n2)
    if (p2 == '') p2 = '.'  # if the subpath is empty, it means the current dir
    return(p2)
  }
  if (!use..) {
    if (error) stop("When use.. = FALSE, the path 'x' must be under the 'dir'")
    return(x)
  }
  s = '../'; d1 = d
  while (!is_sub_path(p, d2 <- dirname(d1))) {
    if (same_path(d1, d2)) {
      if (error) stop(
        "The path 'x' cannot be converted to a relative path to 'dir'. ",
        "Perhaps they are on different volumes of the disk."
      )
      return(x)
    }
    s = paste0('../', s)
    d1 = d2  # go to one level up
  }
  paste0(s, get_subpath(p, n1, nchar(d2)))
}

#' Test if a path is a subpath of a dir
#'
#' Check if the path starts with the dir path.
#' @inheritParams is_abs_path
#' @param dir A vector of directory paths.
#' @param n The length of `dir` paths.
#' @return A logical vector.
#' @note You may want to normalize the values of the `x` and `dir` arguments
#'   first (with [xfun::normalize_path()]), to make sure the path separators
#'   are consistent.
#' @export
#' @examples
#' xfun::is_sub_path('a/b/c.txt', 'a/b')  # TRUE
#' xfun::is_sub_path('a/b/c.txt', 'd/b')  # FALSE
#' xfun::is_sub_path('a/b/c.txt', 'a\\b')  # FALSE (even on Windows)
is_sub_path = function(x, dir, n = nchar(dir)) substr(x, 1, n) == dir

# remove the first n2 characters and the possible / from the path
get_subpath = function(p, n1, n2) {
  p = substr(p, n2 + 1, n1)
  sub('^/', '', p)
}

#' Test if paths are relative or absolute
#'
#' On Unix, check if the paths start with \file{/} or \file{~} (if they do, they
#' are absolute paths). On Windows, check if a path remains the same (via
#' [xfun::same_path()]) if it is prepended with \file{./} (if it does, it is a
#' relative path).
#' @param x A vector of paths.
#' @return A logical vector.
#' @export
#' @examples
#' xfun::is_abs_path(c('C:/foo', 'foo.txt', '/Users/john/', tempdir()))
#' xfun::is_rel_path(c('C:/foo', 'foo.txt', '/Users/john/', tempdir()))
is_abs_path = function(x) {
  if (is_unix()) grepl('^[/~]', x) else !same_path(x, file.path('.', x))
}

#' @rdname is_abs_path
#' @export
is_rel_path = function(x) !is_abs_path(x)

#' Test if a path is a web path
#'
#' Check if a path starts with \file{http://} or \file{https://} or
#' \file{ftp://} or \file{ftps://}.
#' @inheritParams is_abs_path
#' @return A logical vector.
#' @export
#' @examples
#' xfun::is_web_path('https://www.r-project.org')  # TRUE
#' xfun::is_web_path('www.r-project.org')  # FALSE
is_web_path = function(x) {
  grepl('^(f|ht)tps?://', x)
}

#' Get the relative path of a path in a project relative to the current working
#' directory
#'
#' First compose an absolute path using the project root directory and the
#' relative path components, i.e., [`file.path`]`(root, ...)`. Then
#' convert it to a relative path with [relative_path()], which is
#' relative to the current working directory.
#'
#' This function was inspired by `here::here()`, and the major difference
#' is that it returns a relative path by default, which is more portable.
#' @param ... A character vector of path components *relative to the root
#'   directory of the project*.
#' @param root The root directory of the project.
#' @param error Whether to signal an error if the path cannot be converted to a
#'   relative path.
#' @return A relative path, or an error when the project root directory cannot
#'   be determined or the conversion failed and `error = TRUE`.
#' @export
#' @examples
#' \dontrun{
#' xfun::from_root('data', 'mtcars.csv')
#' }
from_root = function(..., root = proj_root(), error = TRUE) {
  if (is.null(root)) stop('Cannot determine the root directory of the current project.')
  p = file.path(root, ..., fsep = '/')
  relative_path(p, error = error)
}

#' Find a file or directory under a root directory
#'
#' Given a path, try to find it recursively under a root directory. The input
#' path can be an incomplete path, e.g., it can be a base filename, and
#' `magic_path()` will try to find this file under subdirectories.
#' @param ... A character vector of path components.
#' @param root The root directory under which to search for the path. If
#'   `NULL`, the current working directory is used.
#' @param relative Whether to return a relative path.
#' @param error Whether to signal an error if the path is not found, or multiple
#'   paths are found.
#' @param message Whether to emit a message when multiple paths are found and
#'   `error = FALSE`.
#' @param n_dirs The number of subdirectories to recursively search. The
#'   recursive search may be time-consuming when there are a large number of
#'   subdirectories under the root directory. If you really want to search for
#'   all subdirectories, you may try `n_dirs = Inf`.
#' @return The path found under the root directory, or an error when `error
#'   = TRUE` and the path is not found (or multiple paths are found).
#' @export
#' @examples
#' \dontrun{
#' xfun::magic_path('mtcars.csv')  # find any file that has the base name mtcars.csv
#' }
magic_path = function(
  ..., root = proj_root(), relative = TRUE, error = TRUE,
  message = getOption('xfun.magic_path.message', TRUE),
  n_dirs = getOption('xfun.magic_path.n_dirs', 10000)
) {
  if (file.exists(p <- file.path(...))) return(p)
  if (is.null(root)) root = getwd()
  nd = 0
  # find a path 'f' recursively under a directory 'd'
  find_it = function(f, d) {
    if (nd > n_dirs) {
      if (error) stop(
        'Failed to find the path under ', n_dirs, ' subdirectories. If you want ',
        'to search for the path in more subdirectories, increase the value of ',
        "the 'n_dirs' argument of magic_path()."
      )
      return(p)
    }
    ds = list.files(d, full.names = TRUE)
    ds = ds[dir_exists(ds)]
    if ((n1 <- length(ds)) == 0) return()
    nd <<- nd + n1
    fs = file.path(ds, f)
    fs = fs[file.exists(fs)]
    if ((n2 <- length(fs)) == 1) return(fs)
    if (n2 > 1) {
      msg = c(
        'Found more than one path containing the input path "', f, '":\n\n',
        paste('*', fs, collapse = '\n')
      )
      if (error) stop(msg)
      if (message) base::message(msg, '\n\nReturned the first one.')
      return(fs[1])
    }
    # look into subdirectories one by one
    for (i in seq_len(n1)) {
      fs = find_it(f, file.path(ds[i]))
      if (length(fs)) return(fs)
    }
  }
  f = find_it(p, root)
  if (is.null(f)) {
    if (error) stop('Could not find the path "', p, '" in any subdirectories.')
    p
  } else {
    if (relative) relative_path(f, error = error) else f
  }
}

#' Test the existence of files and directories
#'
#' These are wrapper functions of [`utils::file_test()]` to test the
#' existence of directories and files. Note that `file_exists()` only tests
#' files but not directories, which is the main difference between
#' [file.exists()] in base R. If you use are using the R version
#' 3.2.0 or above, `dir_exists()` is the same as [dir.exists()]
#' in base R.
#' @param x A vector of paths.
#' @export
#' @return A logical vector.
dir_exists = function(x) file_test('-d', x)

#' @rdname dir_exists
#' @export
file_exists = function(x) file_test('-f', x)

#' Create a directory recursively by default
#'
#' First check if a directory exists. If it does, return `TRUE`, otherwise
#' create it with [`dir.create`]`(recursive = TRUE)` by default.
#' @param x A path name.
#' @param recursive Whether to create all directory components in the path.
#' @param ... Other arguments to be passed to [dir.create()].
#' @return A logical value indicating if the directory either exists or is
#'   successfully created.
#' @export
dir_create = function(x, recursive = TRUE, ...) {
  dir_exists(x) || dir.create(x, recursive = recursive, ...)
}

#' Rename files with a sequential numeric prefix
#'
#' Rename a series of files and add an incremental numeric prefix to the
#' filenames. For example, files \file{a.txt}, \file{b.txt}, and \file{c.txt}
#' can be renamed to \file{1-a.txt}, \file{2-b.txt}, and \file{3-c.txt}.
#' @param pattern A regular expression for [list.files()] to obtain
#'   the files to be renamed. For example, to rename `.jpeg` files, use
#'   `pattern = "[.]jpeg$"`.
#' @param format The format for the numeric prefix. This is passed to
#'   [sprintf()]. The default format is `"\%0Nd"` where `N
#'   = floor(log10(n)) + 1` and `n` is the number of files, which means the
#'   prefix may be padded with zeros. For example, if there are 150 files to be
#'   renamed, the format will be `"\%03d"` and the prefixes will be
#'   `001`, `002`, ..., `150`.
#' @param replace Whether to remove existing numeric prefixes in filenames.
#' @param start The starting number for the prefix (it can start from 0).
#' @param dry_run Whether to not really rename files. To be safe, the default is
#'   `TRUE`. If you have looked at the new filenames and are sure the new
#'   names are what you want, you may rerun `rename_seq()` with
#'   `dry_run = FALSE` to actually rename files.
#' @return A named character vector. The names are original filenames, and the
#'   vector itself is the new filenames.
#' @export
#' @examples xfun::rename_seq()
#' xfun::rename_seq('[.](jpeg|png)$', format = '%04d')
rename_seq = function(
  pattern = '^[0-9]+-.+[.]Rmd$', format = 'auto', replace = TRUE, start = 1,
  dry_run = TRUE
) {
  n = length(files <- list.files('.', pattern))
  if (n == 0) return(files)

  files2 = if (replace) sub('^[0-9]+-*', '', files) else files
  if (format == 'auto') format = paste0('%0', floor(log10(n)) + 1, 'd')
  files2 = paste(sprintf(format, seq_len(n) + start - 1), files2, sep = '-')

  if (!dry_run) file.rename(files, files2)
  structure(setNames(files2, files), class = 'xfun_rename_seq')
}

#' @export
print.xfun_rename_seq = function(x, ...) {
  x = unclass(x)
  tab = data.frame(original = names(x), ' ' = '->', new = unname(x), check.names = FALSE)
  if (loadable('knitr')) tab = knitr::kable(tab, 'simple')
  print(tab)
}

#' Rename files and directories
#'
#' First try [file.rename()]. If it fails (e.g., renaming a file from one volume
#' to another on disk is likely to fail), try [file.copy()] instead, and clean
#' up the original files if the copy succeeds.
#' @param from,to Original and target paths, respectively.
#' @return A logical vector (`TRUE` for success and `FALSE` for failure).
#' @export
file_rename = function(from, to) {
  for (d in dirname(to)) dir_create(d)  # make sure target directories exist

  res = suppressWarnings(file.rename(from, to))

  for (i in which(!res)) {
    f1 = from[i]; f2 = to[i]
    if (dir_exists(f1)) {
      if (res[i] <- dir_create(f2) && all(file.copy(
        list.files(f1, full.names = TRUE, all.files = TRUE), f2, recursive =  TRUE
      ))) unlink(f1, recursive = TRUE)
    } else {
      if (res[i] <- file.copy(f1, f2, overwrite = TRUE)) file.remove(f1)
    }
  }
  res
}

# return path to R's svg logo if it exists, otherwise return the jpg logo; or
# specify a regex to match the logo path, e.g., ext = 'jpg$'
R_logo = function(ext = NULL, all = FALSE) {
  x = file.path(R.home('doc'), 'html', c('Rlogo.svg', 'logo.jpg'))
  if (!is.null(ext)) x = grep(ext, x, value = TRUE)
  existing_files(x, first = !all)
}

#' Extract filenames from a URLs
#'
#' Get the base names of URLs via [basename()], and remove the
#' possible query parameters or hash from the names.
#' @param x A character vector of URLs.
#' @param default The default filename when it cannot be determined from the
#'   URL, e.g., when the URL ends with a slash.
#' @return A character vector of filenames at the end of URLs.
#' @export
#' @examples
#' xfun::url_filename('https://yihui.org/images/logo.png')
#' xfun::url_filename('https://yihui.org/index.html')
#' xfun::url_filename('https://yihui.org/index.html?foo=bar')
#' xfun::url_filename('https://yihui.org/index.html#about')
#' xfun::url_filename('https://yihui.org')
#' xfun::url_filename('https://yihui.org/')
url_filename = function(x, default = 'index.html') {
  # protocol shouldn't be treated as dir name, and query/hash should be removed
  x = gsub('^https?://|[?#].*$', '', x)
  f = basename(x)
  ifelse(grepl('/$', x) | x == f, default, f)
}

#' Delete an empty directory
#'
#' Use `list.file()` to check if there are any files or subdirectories
#' under a directory. If not, delete this empty directory.
#' @param dir Path to a directory. If `NULL` or the directory does not
#'   exist, no action will be performed.
#' @export
del_empty_dir = function(dir) {
  if (is.null(dir) || !dir_exists(dir)) return()
  files = list.files(dir, all.files = TRUE, no.. = TRUE)
  if (length(files) == 0) unlink(dir, recursive = TRUE)
}

#' Mark some paths as directories
#'
#' Add a trailing backlash to a file path if this is a directory. This is useful
#' in messages to the console for example to quickly identify directories from
#' files.
#'
#' If `x` is a vector of relative paths, directory test is done with path
#' relative to the current working dir. Use [xfun::in_dir()] or use absolute
#' paths.
#'
#' @param x Character vector of paths to files and directories.
#' @examples
#' mark_dirs(list.files(find.package("xfun"), full.names = TRUE))
#' @export
mark_dirs = function(x) {
  i = dir_exists(x) & !grepl("/$", x)
  x[i] = paste0(x[i], "/")
  x
}

# change list.files()'s default argument values
all_files = function(
  pattern = NULL, dir = '.', ignore.case = TRUE, full.names = TRUE,
  recursive = TRUE, ...
) {
  list.files(
    dir, pattern, ignore.case = ignore.case, full.names = full.names,
    recursive = recursive, no.. = TRUE, ...
  )
}

library(testit)

# Generate the exact response body for given request parameters.
# Used in both the test handler (to produce the response) and the assertions
# (to compute the expected value), so %==% catches any proxy/routing mismatch.
make_body = function(path, foo = '', post = '', method = '', x_custom = 'MISSING') {
  paste0(
    'path=', path, '\n',
    'foo=', foo, '\n',
    'post=', post, '\n',
    'method=', method, '\n',
    'x_custom=', x_custom
  )
}

# Extract the HTTP response body (strip status line + headers).
# Works on the raw bytes returned by http_request() (handles both \r\n\r\n and \n\n).
http_body = function(resp) {
  if (is.null(resp)) return(NULL)
  for (sep in c('\r\n\r\n', '\n\n')) {
    m = regexpr(sep, resp, fixed = TRUE)
    if (m[1L] > 0L) return(substring(resp, m[1L] + nchar(sep)))
  }
  resp
}

# Send a raw HTTP/1.0 request and return the full response as a string.
# Use non-blocking I/O plus Content-Length checks to avoid partial-response flakes.
http_request = function(host, port, method, path, body = NULL, extra_headers = '') {
  sock = tryCatch(
    socketConnection(host, port = port, open = 'r+b', blocking = FALSE, timeout = 5),
    error = function(e) NULL
  )
  if (is.null(sock)) return(NULL)
  on.exit(try(close(sock), silent = TRUE), add = TRUE)
  clen = if (is.null(body)) 0L else nchar(body, type = 'bytes')
  req = paste0(
    method, ' ', path, ' HTTP/1.0\r\n',
    'Host: ', host, '\r\n',
    if (clen > 0) paste0('Content-Length: ', clen, '\r\n') else '',
    extra_headers,
    '\r\n',
    if (!is.null(body)) body else ''
  )
  writeBin(req, sock)
  buf = raw(0L)
  # 120 * 0.1s = up to 12 seconds for slow CI environments.
  for (i in seq_len(120L)) {
    Sys.sleep(0.1)
    ready = tryCatch(socketSelect(list(sock), write = FALSE, timeout = 0), error = function(e) FALSE)
    if (!isTRUE(ready[1L])) next
    chunk = tryCatch(readBin(sock, raw(), n = 65536L), error = function(e) NULL)
    if (is.null(chunk)) next
    if (length(chunk) == 0L) break
    buf = c(buf, chunk)
    s = rawToChar(buf)
    sep = regexpr('\r\n\r\n|\n\n', s, perl = TRUE)
    if (sep[1L] < 0L) next
    body_start = sep[1L] + attr(sep, 'match.length')
    cl = regmatches(s, regexpr('(?i)Content-Length: *(\\d+)', s, perl = TRUE))
    if (length(cl) == 0L || !nzchar(cl)) next
    expected = as.integer(sub('(?i).*Content-Length: *', '', cl, perl = TRUE))
    if (nchar(s) - body_start + 1L >= expected) break
  }
  if (length(buf) == 0L) NULL else rawToChar(buf)
}

# Retry a request until a complete response with a non-empty body is received.
http_request_full = function(host, port, method, path, body = NULL, extra_headers = '', tries = 30L) {
  resp = NULL
  for (i in seq_len(tries)) {
    resp = http_request(host, port, method, path, body, extra_headers)
    has_sep = !is.null(resp) && grepl('\r\n\r\n|\n\n', resp, perl = TRUE)
    b = http_body(resp)
    if (has_sep && !is.null(b) && nzchar(b)) break
    Sys.sleep(0.1)
  }
  resp
}

url_port = function(url) as.integer(sub('^https?://[^:]+:([0-9]+)/.*$', '\\1', url))

if (
  has_fun('serverSocket') && has_fun('socketAccept') &&
  !is.null(port <- random_port(error = FALSE))
) {

  # Run the proxy-backed app in a separate R process so requests are sent from
  # this R session to reduce random timing issues in CI.
  pid_file = tempfile(fileext = '.rds')
  saveRDS(NULL, pid_file)
  Rscript_call(function(make_body, port, pid_file, libs) {
    .libPaths(libs)
    saveRDS(Sys.getpid(), pid_file)
    on.exit(unlink(pid_file), add = TRUE)
    xfun::new_app(
      '',
      function(path, query, post, headers) {
        h          = if (length(headers) > 0L) rawToChar(headers) else ''
        foo        = if ('foo' %in% names(query)) query[['foo']] else ''
        post_str   = if (length(post) > 0L) rawToChar(post) else ''
        method_str = {
          m = regmatches(h, regexpr('Request-Method: [^\r\n]+', h))
          if (length(m) > 0L) sub('Request-Method: ', '', m) else ''
        }
        x_custom_str = {
          m = regmatches(h, regexpr('X-Custom: [^\r\n]+', h))
          if (length(m) > 0L) sub('X-Custom: ', '', m) else 'MISSING'
        }
        list(payload = make_body(path, foo, post_str, method_str, x_custom_str),
             'content-type' = 'text/plain')
      },
      open = NA,
      port = port
    )
  }, list(make_body = make_body, port = port, pid_file = pid_file, libs = .libPaths()), wait = FALSE)
  pid = NULL
  for (i in seq_len(100L)) {
    Sys.sleep(0.1)
    if (!file.exists(pid_file)) break
    pid = tryCatch(readRDS(pid_file), error = function(e) NULL)
    if (length(pid) == 1L) break
  }
  on.exit(if (length(pid) == 1L) try(proc_kill(pid), silent = TRUE), add = TRUE)

  # Wait for the background app to become reachable.
  for (i in seq_len(50L)) {
    resp = http_request_full('127.0.0.1', port, 'GET', '/hello')
    if (!is.null(resp) && grepl('200 OK', resp, fixed = TRUE)) break
    Sys.sleep(0.1)
  }

  assert('proxy root app responds to a GET request', {
    (!is.null(resp))
    (grepl('200 OK', resp, fixed = TRUE))
    (http_body(resp) %==% make_body('hello', method = 'GET'))
  })

  assert('proxy query parameters are URL-decoded', {
    resp = http_request_full('127.0.0.1', port, 'GET', '/page?foo=bar%20baz')
    (!is.null(resp))
    (http_body(resp) %==% make_body('page', foo = 'bar baz', method = 'GET'))
  })

  assert('proxy POST body is forwarded correctly', {
    resp = http_request_full('127.0.0.1', port, 'POST', '/submit', body = 'hello=world')
    (!is.null(resp))
    (http_body(resp) %==% make_body('submit', post = 'hello=world', method = 'POST'))
  })

  assert('proxy request headers reach the handler', {
    resp = http_request_full('127.0.0.1', port, 'GET', '/hdr',
                             extra_headers = 'X-Custom: test-value\r\n')
    (!is.null(resp))
    (http_body(resp) %==% make_body('hdr', method = 'GET', x_custom = 'test-value'))
  })

  assert('.find_proxy_port() excludes ports used by new_app(name = \"\")', {
    old_apps = .proxy$apps
    old_help = .proxy$help
    on.exit({
      .proxy$apps = old_apps
      .proxy$help = old_help
    }, add = TRUE)
    .proxy$apps[['']] = list(type = 'proxy', port = 4322L, slot = 0L, key = 'xfun:4322')
    p = .find_proxy_port(4322:4325)
    (p != 4322L)
  })

  assert('.find_proxy_port() excludes ports used by help_proxy()', {
    old_apps = .proxy$apps
    old_help = .proxy$help
    on.exit({
      .proxy$apps = old_apps
      .proxy$help = old_help
    }, add = TRUE)
    .proxy$help[['4323']] = 1L
    p = .find_proxy_port(4323:4326)
    (p != 4323L)
  })

  url = new_app(
    'test',
    function(path, query, post, headers) {
      h          = if (length(headers) > 0L) rawToChar(headers) else ''
      foo        = if ('foo' %in% names(query)) query[['foo']] else ''
      post_str   = if (length(post) > 0L) rawToChar(post) else ''
      method_str = {
        m = regmatches(h, regexpr('Request-Method: [^\r\n]+', h))
        if (length(m) > 0L) sub('Request-Method: ', '', m) else ''
      }
      list(payload = make_body(path, foo, post_str, method_str), 'content-type' = 'text/plain')
    },
    open = FALSE
  )
  on.exit(stop_app('test'), add = TRUE)

  assert('new_app(name != \"\") uses legacy /custom/name/ path', {
    m = regexec('^http://127\\.0\\.0\\.1:([0-9]+)/custom/test/$', url)
    cap = regmatches(url, m)[[1]]
    (length(cap) == 2L)
    (!is.null(get0('test', .httpd_env(), inherits = FALSE)))
  })

  assert('stop_app() deregisters legacy app handler', {
    stop_app('test')
    (is.null(.proxy$apps[['test']]))
    (is.null(get0('test', .httpd_env(), inherits = FALSE)))
  })

  assert('html_view() uses default name "xfun-html"', {
    html_view('<p>hi</p>', open = FALSE)
    on.exit(stop_app('xfun-html'), add = TRUE)
    (!is.null(.proxy$apps[['xfun-html']]))
    (!is.null(get0('xfun-html', .httpd_env(), inherits = FALSE)))
  })

  assert('html_view() accepts a custom name', {
    html_view('<p>hi</p>', name = 'my-app', open = FALSE)
    on.exit(stop_app('my-app'), add = TRUE)
    (!is.null(.proxy$apps[['my-app']]))
    (!is.null(get0('my-app', .httpd_env(), inherits = FALSE)))
  })

  d = tempfile(); dir.create(d)
  writeLines('hello', file.path(d, 'a.txt'))
  url = serve_dir(d, name = 'serve-dir-test', open = FALSE)
  on.exit({ stop_app('serve-dir-test'); unlink(d, recursive = TRUE) }, add = TRUE)

  assert('serve_dir() registers an app and returns its URL', {
    (grepl('/custom/serve-dir-test/$', url))
    (!is.null(.proxy$apps[['serve-dir-test']]))
    handler = get0('serve-dir-test', .httpd_env(), inherits = FALSE)
    (!is.null(handler))
    # request for the served file
    r = handler('/custom/serve-dir-test/a.txt')
    (basename(r$file) %==% 'a.txt')
    # request for the directory listing
    r = handler('/custom/serve-dir-test/')
    (r$`content-type` %==% 'text/html')
    (grepl('a.txt', r$payload))
  })
}

library(testit)

assert('is_blank() tests if a vector is blank', {
  (is_blank(''))
  (is_blank(NULL) %==% logical(0))
  (is_blank(c('', '   ', '\n', '\t')))
  (!is_blank('abc'))
  (is_blank(c('', 'abc')) %==% c(TRUE, FALSE))
})

assert('n2w converts numbers to words', {
  (n2w(0) %==% 'zero')
  # cap capitalizes the first letter
  (n2w(0, cap = TRUE) %==% 'Zero')
  # hyphen adds '-' for 21-99 (except 30, 40, ...)
  (n2w(21, cap = TRUE, hyphen = TRUE) %==% 'Twenty-one')
  (n2w(21, cap = TRUE, hyphen = FALSE) %==% 'Twenty one')
  # x can be negative integers
  (n2w(-21, cap = TRUE, hyphen = TRUE) %==% 'Minus twenty-one')
  (n2w(-21, cap = TRUE, hyphen = FALSE) %==% 'Minus twenty one')
  # and controls whether to have 'add' between hundreds and double digits
  (n2w(121, and = FALSE) %==% 'one hundred twenty-one')
  (n2w(121, and = TRUE) %==% 'one hundred and twenty-one')
  # x can be a vector with length > 1
  (n2w(c(10, 13, 99, 1e6)) %==% c('ten', 'thirteen', 'ninety-nine', 'one million'))
  # the number should be less than 1e15
  (has_error(n2w(1e15)))
})

assert('join_words() joins multiple words into a single string', {
  jw = function(...) unclass(join_words(...))
  (jw(NULL) %==% character())
  (jw(c('a')) %==% 'a')
  (jw(c('a', 'b')) %==% 'a and b')
  (jw(c('a', 'b'), and = "") %==% 'a, b')
  (jw(c('a', 'b', 'c')) %==% 'a, b, and c')
  (jw(c('a', 'b', 'c'), and = '') %==% 'a, b, c')
  (jw(c('a', 'b', 'c'), ' / ', '') %==% 'a / b / c')
  (jw(c('a', 'b', 'c'), before = '"') %==% '"a", "b", and "c"')
  (jw(c('a', 'b', 'c'), before = '``', after = "''") %==% "``a'', ``b'', and ``c''")
  (jw(c('a', 'b', 'c'), before = '``', after = "''", oxford_comma = FALSE) %==% "``a'', ``b'' and ``c''")
  rm(list = 'jw')
})

assert('split_lines() splits a character vector into lines', {
  (split_lines('a') %==% 'a')
  (split_lines('') %==% '')
  (split_lines(NULL) %==% NULL)
  (split_lines('a\n') %==% c('a', ''))
  (split_lines(c('a', 'b\nc')) %==% c('a', 'b', 'c'))
  (split_lines(c('', '\n')) %==% c('', '', ''))
  (split_lines('a\nb') %==% c('a', 'b'))
  (split_lines('a\nb\n\n') %==% c('a', 'b', '', ''))
  (split_lines(c('a\nb', '', ' ', 'c')) %==% c('a', 'b', '', ' ', 'c'))
})

assert('split_source() puts lines of the same expression into a list element', {
  (split_source('1+1') %==% list('1+1'))
  (split_source(c('1+1+', '1')) %==% list(c('1+1+', '1')))
  (split_source(c('1+1+', '1', 'TRUE')) %==% list(c('1+1+', '1'), 'TRUE'))
  x = c('# a', '# b', '1', '# c', 'if (T)', 'F')
  (split_source(x) %==% c(as.list(x[1:4]), list(x[5:6])))
  (split_source(x, merge_comments = TRUE) %==% list(x[1:3], x[4:6]))
  x = c(x, '1+1')
  (split_source(x, TRUE, line_number = TRUE)[2:3] %==% list(
    structure(x[4:6], lines = c(4L, 6L)),
    structure(x[7], lines = c(7L, 7L))
  ))
})

assert('split_source() returns list with input for empty input', {
  # empty input (length < 1) should return list(x) immediately
  (split_source(character(0)) %==% list(character(0)))
})

assert('split_source() should signal an error for incomplete code', {
  (has_error(split_source('1+1+')))
  (has_error(split_source(c('1+1', '1+1+'))))
})

assert('valid_syntax() tells if a code fragment is syntactically valid', {
  (valid_syntax('1+1'))
  (!valid_syntax('1+1+'))
  (valid_syntax('if(TRUE)1'))
  (!valid_syntax(c('if(T){', 'F')))
  (valid_syntax(c('if(T){', 'F}')))
})

assert('alnum_id() generates ID strings', {
  x = c('Hello world 123!', 'a  &b*^##c 456')
  (alnum_id(x) %==% c('hello-world-123', 'a-b-c-456'))
  (alnum_id(x, '[^[:alpha:]]+') %==% c('hello-world', 'a-b-c'))
})

assert('str_wrap() wraps text and returns same-length output', {
  x = c('hello world foo bar baz', 'another long string here')
  res = str_wrap(x, width = 10)
  (length(res) %==% length(x))
  (is.character(res))
  # each element should contain newlines when wrapping was needed
  (grepl('\n', res[1]))
})

assert('decimal_dot() forces dot as the decimal separator', {
  old = options(OutDec = ',')
  r = decimal_dot(as.character(1.234))
  options(old)  # restore OutDec
  (r %==% '1.234')
})

assert('strip_blank() removes blank elements from both ends', {
  (strip_blank(character(0)) %==% character(0))
  (strip_blank(c('', 'a', 'b', '')) %==% c('a', 'b'))
  (strip_blank(c('', '  ', 'a', '', 'b', '', '')) %==% c('a', '', 'b'))
  (strip_blank(c('a', 'b')) %==% c('a', 'b'))
  (strip_blank(c('', '')) %==% character(0))
})

assert('query_params() creates URL query strings', {
  (query_params() %==% '')
  (query_params(a = 1, b = 'foo') %==% '?a=1&b=foo')
  (query_params(.list = list(x = 'bar')) %==% '?x=bar')
})

assert('pair_chars() checks balanced paired characters', {
  # balanced quotes
  x = c('He said \u201chello\u201d.', 'No quotes here.')
  (pair_chars(x) %==% x)
  # unbalanced quotes produce a warning
  (has_warning(pair_chars(c('\u201chello.'))))
  # wrong chars length errors
  (has_error(pair_chars('text', chars = c('\u201c'))))
  # file mode: file content unchanged (quotes are already balanced)
  f = tempfile()
  writeLines(c('\u201chello\u201d', '\u201cworld\u201d'), f)
  pair_chars(file = f)
  (read_utf8(f) %==% c('\u201chello\u201d', '\u201cworld\u201d'))
  # file mode: content changed (mismatched quotes get replaced)
  f2 = tempfile()
  writeLines(c('\u201chello\u201d', '\u2018world\u2018'), f2)  # second line has wrong closing quote
  has_warning(pair_chars(file = f2))
  unlink(c(f, f2))
})

assert('html_content() resolves HTML content recursively', {
  # list with mixed html and plain text
  result = html_content(list('hello', html_value('<b>world</b>')))
  (result %==% c(html_escape('hello'), '<b>world</b>'))
})

assert('numbers_to_words() handles more cases', {
  # non-numeric input errors
  (has_error(n2w('abc')))
  # value >= 1e15 errors
  (has_error(n2w(1e15)))
  # 10-19 range (x_cs[1] == 1)
  (n2w(10) %==% 'ten')
  (n2w(15) %==% 'fifteen')
  (n2w(19) %==% 'nineteen')
  # x00 pattern (100, 200, ...) and x0x (101, 201, ...)
  (n2w(100) %==% 'one hundred')
  (n2w(200) %==% 'two hundred')
  (n2w(101) %==% 'one hundred one')
  # float with decimal part
  (n2w(1.5) %==% 'one point five')
  (n2w(3.14) %==% 'three point one four')
  # round tens like 30, 40, 50 (x_cs[2] == 0 branch in convert_2)
  (n2w(30) %==% 'thirty')
  (n2w(40) %==% 'forty')
  (n2w(50) %==% 'fifty')
})

assert('encrypt() and decrypt() are inverses', {
  key = 'fedcba9876543210'
  x = 'hello'
  (decrypt(encrypt(x, key), key) %==% x)
})

assert('n2w() handles large numbers and trillion range', {
  (n2w(1e12) %==% 'one trillion')
  (n2w(1e9) %==% 'one billion')
  (n2w(1e6) %==% 'one million')
  (n2w(1000) %==% 'one thousand')
  (n2w(1e14) %==% 'one hundred trillion')
  # number just below the limit
  (is.character(n2w(1e15 - 1)))
})

assert('n2w() handles vectors and zero correctly', {
  (n2w(0:3) %==% c('zero', 'one', 'two', 'three'))
})

assert('join_words() with length-1 input and oxford_comma edge cases', {
  jw = function(...) unclass(join_words(...))
  # single word
  (jw('hello') %==% 'hello')
  # oxford comma with sep not ending in space
  (jw(c('a', 'b', 'c'), sep = ', ', and = ' and ', oxford_comma = TRUE) %==% 'a, b, and c')
  (jw(c('a', 'b', 'c'), sep = ', ', and = ' and ', oxford_comma = FALSE) %==% 'a, b and c')
  rm(list = 'jw')
})

assert('split_lines() handles \\r\\n line endings', {
  # CRLF should be split the same as LF
  (split_lines('a\r\nb') %==% c('a', 'b'))
  (split_lines('a\r\nb\r\n') %==% c('a', 'b', ''))
})

assert('alnum_id() strips HTML tags before generating ID', {
  (alnum_id('<h1>Hello <strong>world</strong>!') %==% 'hello-world')
})

assert('strip_html() handles multiline-like patterns and nested tags', {
  (strip_html('<!-- comment --><b>text</b>') %==% 'text')
  # nested tags
  (strip_html('<div><p>hello</p></div>') %==% 'hello')
})

---
title: An Introduction to xfun
subtitle: A Collection of Miscellaneous Functions
author: "Yihui Xie"
date: "`{r} Sys.Date()`"
slug: xfun
githubEditURL: https://github.com/yihui/xfun/edit/main/vignettes/xfun.Rmd
output:
  html:
    meta:
      css: ["@default", "@article"]
      js: ["@sidenotes"]
    options:
      toc: true
      number_sections: true
---

<!--
%\VignetteIndexEntry{An Introduction to xfun}
%\VignetteEngine{litedown::vignette}
-->

```{r setup, include=FALSE}
library(xfun)
```

After writing about 20 R packages, I found I had accumulated several utility functions that I used across different packages, so I decided to extract them into a separate package. Previously I had been using the evil triple-colon `:::` to access these internal utility functions. Now with **xfun**, these functions have been exported, and more importantly, documented. It should be better to use them under the sun instead of in the dark.

This page shows examples of a subset of functions in this package. For a full list of functions, see the help page `help(package = 'xfun')`. The source package is available on GitHub: https://github.com/yihui/xfun.

## No more partial matching for lists!

I have been bitten many times by partial matching in lists, e.g., when I want `x$a` but the element `a` does not exist in the list `x`, it returns the value `x$abc` if `abc` exists in `x`. A strict list is a list for which the partial matching of the `$` operator is disabled. The functions `xfun::strict_list()` and `xfun::as_strict_list()` are the equivalents to `base::list()` and `base::as.list()` respectively which always return as strict list, e.g.,

```{r}
library(xfun)
(z = strict_list(aaa = "I am aaa", b = 1:5))
z$a  # NULL (strict matching)
z$aaa  # I am aaa
z$b
z$c = "you can create a new element"

z2 = unclass(z)  # a normal list
z2$a  # partial matching

z3 = as_strict_list(z2)  # a strict list again
z3$a  # NULL (strict matching) again!
```

Similarly, the default partial matching in `attr()` can be annoying, too. The function `xfun::attr2()` is simply a shorthand of `attr(..., exact = TRUE)`.

I want it, or I do not want. There is no "I probably want".

## Output character vectors for human eyes

When R prints a character vector, your eyes may be distracted by the indices like `[1]`, double quotes, and escape sequences. To see a character vector in its "raw" form, you can use `cat(..., sep = '\n')`. The function `raw_string()` marks a character vector as "raw", and the corresponding printing function will call `cat(sep = '\n')` to print the character vector to the console.

```{r comment=''}
library(xfun)
raw_string(head(LETTERS))
(x = c("a \"b\"", "hello\tworld!"))
raw_string(x)  # this is more likely to be what you want to see
```

## Print the content of a text file

I have used `paste(readLines('foo'), collapse = '\n')` many times before I decided to write a simple wrapper function `xfun::file_string()`. This function also makes use of `raw_string()`, so you can see the content of a file in the console as a side-effect, e.g.,

```{r comment=''}
f = system.file("LICENSE", package = "xfun")
xfun::file_string(f)
as.character(xfun::file_string(f))  # essentially a character string
```

## Get the data URI of a file

Files can be encoded into base64 strings via `base64_uri()`. This is a common technique to embed arbitrary files in HTML documents (which is [what `xfun::embed_file()` does](https://bookdown.org/yihui/rmarkdown-cookbook/embed-file.html) and it is based on `base64_uri()`).

```{r}
f = system.file("LICENSE", package = "xfun")
xfun::base64_uri(f)
```

## Match strings and do substitutions

After typing the code `x = grep(pattern, x, value = TRUE); gsub(pattern, '\\1', x)` many times, I combined them into a single function `xfun::grep_sub()`.

```{r}
xfun::grep_sub('a([b]+)c', 'a\\U\\1c', c('abc', 'abbbc', 'addc', '123'), perl = TRUE)
```

## Search and replace strings in files

I can never remember how to properly use `grep` or `sed` to search and replace strings in multiple files. My favorite IDE, RStudio, has not provided this feature yet (you can only search and replace in the currently opened file). Therefore I did a quick and dirty implementation in R, including functions `gsub_files()`, `gsub_dir()`, and `gsub_ext()`, to search and replace strings in multiple files under a directory. Note that the files are assumed to be encoded in UTF-8. If you do not use UTF-8, we cannot be friends. Seriously.

All functions are based on `gsub_file()`, which performs searching and replacing in a single file, e.g.,

```{r comment=''}
library(xfun)
f = tempfile()
writeLines(c("hello", "world"), f)
gsub_file(f, "world", "woRld", fixed = TRUE)
file_string(f)
```

The function `gsub_dir()` is very flexible: you can limit the list of files by MIME types, or extensions. For example, if you want to do substitution in text files, you may use `gsub_dir(..., mimetype = '^text/')`.

The function `process_file()` is a more general way to process files. Basically it reads a file, process the content with a function that you pass to it, and writes back the text, e.g.,

```{r, comment=''}
process_file(f, function(x) {
  rep(x, 3)  # repeat the content 3 times
})
file_string(f)
```

**WARNING**: Before using these functions, make sure that you have backed up your files, or version control your files. The files will be modified in-place. If you do not back up or use version control, there is no chance to regret.

## Manipulate filename extensions

Functions `file_ext()` and `sans_ext()` are based on functions in **tools**. The function `with_ext()` adds or replaces extensions of filenames, and it is vectorized.

```{r}
library(xfun)
p = c("abc.doc", "def123.tex", "path/to/foo.Rmd")
file_ext(p)
sans_ext(p)
with_ext(p, ".txt")
with_ext(p, c(".ppt", ".sty", ".Rnw"))
with_ext(p, "html")
```

## Find files (in a project) without the pain of thinking about absolute/relative paths

The function `proj_root()` was inspired by the **rprojroot** package, and tries to find the root directory of a project. Currently it only supports R package projects and RStudio projects by default. It is much less sophisticated than **rprojroot**.

The function `from_root()` was inspired by `here::here()`, but returns a relative path (relative to the project's root directory found by `proj_root()`) instead of an absolute path. For example, `xfun::from_root('data', 'cars.csv')` in a code chunk of `docs/foo.Rmd` will return `../data/cars.csv` when `docs/` and `data/` directories are under the root directory of a project.

```
root/
  |-- data/
  |   |-- cars.csv
  |
  |-- docs/
      |-- foo.Rmd
```

If file paths are too much pain for you to think about, you can just pass an incomplete path to the function `magic_path()`, and it will try to find the actual path recursively under subdirectories of a root directory. For example, you may only provide a base filename, and `magic_path()` will look for this file under subdirectories and return the actual path if it is found. By default, it returns a relative path, which is relative to the current working directory. With the above example, `xfun::magic_path('cars.csv')` in a code chunk of `docs/foo.Rmd` will return `../data/cars.csv`, if `cars.csv` is a unique filename in the project. You can freely move it to any folders of this project, and `magic_path()` will still find it. If you are not using a project to manage files, `magic_path()` will look for the file under subdirectories of the current working directory.

## Types of operating systems

The series of functions `is_linux()`, `is_macos()`, `is_unix()`, and `is_windows()` test the types of the OS, using the information from `.Platform` and `Sys.info()`, e.g.,

```{r}
xfun::is_macos()
xfun::is_unix()
xfun::is_linux()
xfun::is_windows()
```

## Loading and attaching packages

Oftentimes I see users attach a series of packages in the beginning of their scripts by repeating `library()` multiple times. This could be easily vectorized, and the function `xfun::pkg_attach()` does this job. For example,

```{r eval=FALSE}
library(testit)
library(parallel)
library(tinytex)
library(mime)
```

is equivalent to

```{r eval=FALSE}
xfun::pkg_attach(c('testit', 'parallel', 'tinytex', 'mime'))
```

I also see scripts that contain code to install a package if it is not available, e.g.,

```{r eval=FALSE}
if (!requireNamespace('tinytex')) install.packages('tinytex')
library(tinytex)
```

This could be done via

```{r eval=FALSE}
xfun::pkg_attach2('tinytex')
```

The function `pkg_attach2()` is a shorthand of `pkg_attach(..., install = TRUE)`, which means if a package is not available, install it. This function can also deal with multiple packages.

The function `loadable()` tests if a package is loadable.

## Read/write files in UTF-8

Functions `read_utf8()` and `write_utf8()` can be used to read/write files in UTF-8. They are simple wrappers of `readLines()` and `writeLines()`.

## Convert numbers to English words

The function `numbers_to_words()` (or `n2w()` for short) converts numbers to English words.

```{r}
n2w(0, cap = TRUE)
n2w(seq(0, 121, 11), and = TRUE)
n2w(1e+06)
n2w(1e+11 + 12345678)
n2w(-987654321)
n2w(1e+15 - 1)
```

## Cache an R expression to disk or in memory

Since **xfun** v0.44, you are recommended to use the function [`cache_exec()`](https://git.yihui.org/xfun/manual.html#sec:man-cache_exec), which provides a simple yet flexible caching mechanism. See https://yihui.org/litedown/#sec:option-cache for how it works. Previously, `cache_rds()` was mentioned here but it is no longer recommended (see [#100](https://github.com/yihui/xfun/issues/100)).

## Check reverse dependencies of a package

Running `R CMD check` on the reverse dependencies of **knitr** and **rmarkdown** is my least favorite thing in developing R packages, because the numbers of their reverse dependencies are huge. The function `rev_check()` reflects some of my past experience in this process. I think I have automated it as much as possible, and made it as easy as possible to discover possible new problems introduced by the current version of the package (compared to the CRAN version). Finally I can just sit back and let it run.

## Input a character vector into the RStudio source editor

The function `rstudio_type()` inputs characters in the RStudio source editor as if they were typed by a human. I came up with the idea when preparing my talk for rstudio::conf 2018 ([see this post](https://yihui.org/en/2018/03/blogdown-video-rstudio-conf/) for more details).

## Print session information

Since I have never been fully satisfied by the output of `sessionInfo()`, I tweaked it to make it more useful in my use cases. For example, it is rarely useful to print out the names of base R packages, or information about the matrix products / BLAS / LAPACK. Oftentimes I want additional information in the session information, such as the Pandoc version when **rmarkdown** is used. The function `session_info()` tweaks the output of `sessionInfo()`, and makes it possible for other packages to append information in the output of `session_info()`.

You can choose to print out the versions of only the packages you specify, e.g.,

```{r}
xfun::session_info(c('xfun', 'litedown', 'tinytex'), dependencies = FALSE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{same_path}
\alias{same_path}
\title{Test if two paths are the same after they are normalized}
\usage{
same_path(p1, p2, ...)
}
\arguments{
\item{p1, p2}{Two vectors of paths.}

\item{...}{Arguments to be passed to \code{\link[=normalize_path]{normalize_path()}}.}
}
\description{
Compare two paths after normalizing them with the same separator (\code{/}).
}
\examples{
library(xfun)
same_path("~/foo", file.path(Sys.getenv("HOME"), "foo"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{same_path}
\alias{same_path}
\title{Test if two paths are the same after they are normalized}
\usage{
same_path(p1, p2, ...)
}
\arguments{
\item{p1, p2}{Two vectors of paths.}

\item{...}{Arguments to be passed to \code{\link{normalize_path}()}.}
}
\description{
Compare two paths after normalizing them with the same separator (\code{/}).
}
\examples{
library(xfun)
same_path("~/foo", file.path(Sys.getenv("HOME"), "foo"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{system3}
\alias{system3}
\title{Run \code{system2()} and mark its character output as UTF-8 if appropriate}
\usage{
system3(...)
}
\arguments{
\item{...}{Passed to \code{\link[=system2]{system2()}}.}
}
\value{
The value returned by \code{system2()}.
}
\description{
This is a wrapper function based on \code{system2()}. If \code{system2()}
returns character output (e.g., with the argument \code{stdout = TRUE}),
check if the output is encoded in UTF-8. If it is, mark it with UTF-8
explicitly.
}
\examples{\dontshow{if (interactive()) \{}
a = shQuote(c("-e", "print(intToUtf8(c(20320, 22909)))"))
x2 = system2("Rscript", a, stdout = TRUE)
Encoding(x2)  # unknown

x3 = xfun::system3("Rscript", a, stdout = TRUE)
# encoding of x3 should be UTF-8 if the current locale is UTF-8
!l10n_info()[["UTF-8"]] || Encoding(x3) == "UTF-8"  # should be TRUE
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{system3}
\alias{system3}
\title{Run \code{system2()} and mark its character output as UTF-8 if appropriate}
\usage{
system3(...)
}
\arguments{
\item{...}{Passed to \code{\link{system2}()}.}
}
\value{
The value returned by \code{system2()}.
}
\description{
This is a wrapper function based on \code{system2()}. If \code{system2()}
returns character output (e.g., with the argument \code{stdout = TRUE}),
check if the output is encoded in UTF-8. If it is, mark it with UTF-8
explicitly.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
a = shQuote(c("-e", "print(intToUtf8(c(20320, 22909)))"))
x2 = system2("Rscript", a, stdout = TRUE)
Encoding(x2)  # unknown

x3 = xfun::system3("Rscript", a, stdout = TRUE)
# encoding of x3 should be UTF-8 if the current locale is UTF-8
!l10n_info()[["UTF-8"]] || Encoding(x3) == "UTF-8"  # should be TRUE
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/record.R
\name{record_print}
\alias{record_print}
\alias{record_print.default}
\alias{record_print.record_asis}
\alias{new_record}
\title{Print methods for \code{record()}}
\usage{
record_print(x, ...)

\method{record_print}{default}(x, ...)

\method{record_print}{record_asis}(x, ...)

new_record(x, class)
}
\arguments{
\item{x}{For \code{record_print()}, the value to be printed. For \code{new_record()}, a
character vector to be included in the printed results.}

\item{...}{Other arguments to be passed to \code{record_print()} methods.}

\item{class}{A class name. Possible values are \code{xfun:::.record_cls}.}
}
\value{
A \code{record_print()} method should return a character vector or a list
of character vectors. The original classes of the vector will be discarded,
and the vector will be treated as console output by default (i.e.,
\code{new_record(class = "output")}). If it should be another type of output,
wrap the vector in \code{\link[=new_record]{new_record()}} and specify a class name.
}
\description{
An S3 generic function to be called to print visible values in code when the
code is recorded by \code{\link[=record]{record()}}. It is similar to \code{\link[knitr:knit_print]{knitr::knit_print()}}. By
default, it captures the normal \code{\link[=print]{print()}} output and returns the result as a
character vector. The \code{knitr_kable} method is for printing \code{\link[knitr:kable]{knitr::kable()}}
output. Users and package authors can define other S3 methods to extend this
function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-structure.R
\name{raw_string}
\alias{raw_string}
\alias{print.xfun_raw_string}
\title{Print a character vector in its raw form}
\usage{
raw_string(x)

\method{print}{xfun_raw_string}(x, ...)
}
\arguments{
\item{x}{For \code{raw_string()}, a character vector. For the print method,
the \code{raw_string()} object.}

\item{...}{Other arguments (currently ignored).}
}
\description{
The function \code{raw_string()} assigns the class \code{xfun_raw_string} to
the character vector, and the corresponding printing function
\code{print.xfun_raw_string()} uses \code{cat(x, sep = '\n')} to write the
character vector to the console, which will suppress the leading indices
(such as \code{[1]}) and double quotes, and it may be easier to read the
characters in the raw form (especially when there are escape sequences).
}
\examples{
library(xfun)
raw_string(head(LETTERS))
raw_string(c("a \"b\"", "hello\tworld!"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{del_empty_dir}
\alias{del_empty_dir}
\title{Delete an empty directory}
\usage{
del_empty_dir(dir)
}
\arguments{
\item{dir}{Path to a directory. If \code{NULL} or the directory does not
exist, no action will be performed.}
}
\description{
Use \code{list.file()} to check if there are any files or subdirectories
under a directory. If not, delete this empty directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-structure.R
\name{raw_string}
\alias{raw_string}
\alias{print.xfun_raw_string}
\title{Print a character vector in its raw form}
\usage{
raw_string(x, ...)

\method{print}{xfun_raw_string}(x, ...)
}
\arguments{
\item{x}{For \code{raw_string()}, a character vector. For the print method,
the \code{raw_string()} object.}

\item{...}{Other arguments (currently ignored).}
}
\description{
The function \code{raw_string()} assigns the class \code{xfun_raw_string} to
the character vector, and the corresponding printing function
\code{print.xfun_raw_string()} uses \code{cat(x, sep = '\\n')} to write the
character vector to the console, which will suppress the leading indices
(such as \verb{[1]}) and double quotes, and it may be easier to read the
characters in the raw form (especially when there are escape sequences).
}
\examples{
library(xfun)
raw_string(head(LETTERS))
raw_string(c("a \"b\"", "hello\tworld!"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/browser.R
\name{browser_dom}
\alias{browser_dom}
\title{Execute JavaScript on an HTML page via a headless browser and dump the DOM}
\usage{
browser_dom(input, output = NULL, fragment = FALSE, browser = NULL)
}
\arguments{
\item{input}{Path or URL to the HTML page.}

\item{output}{An optional output file path. If not provided, the rendered
HTML is returned as a character string.}

\item{fragment}{Whether to return only the inner body content instead of the
full HTML document.}

\item{browser}{Path to the web browser. By default, the browser is found via
\code{xfun:::find_browser()}. If it cannot be found, you may set the global
option \code{options(xfun.browser = )} or environment variable \code{R_XFUN_BROWSER}
to the path of the browser, or simply pass the path to the \code{browser}
argument.}
}
\value{
If \code{output} is \code{NULL}, the rendered HTML as a character string;
otherwise the \code{output} path (invisibly).
}
\description{
Load an HTML page in a headless browser (Chromium/Chrome/Edge), execute all
JavaScript on it, and return (or save) the rendered DOM as HTML.
}
\examples{\dontshow{if (interactive()) \{}
f = tempfile(fileext = '.html')
writeLines(c(
  '<html><body><div id="out"></div>',
  '<script>document.getElementById("out").textContent = "hello";</script>',
  '</body></html>'
), f)
xfun::browser_dom(f)
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{strip_html}
\alias{strip_html}
\title{Strip HTML tags}
\usage{
strip_html(x)
}
\arguments{
\item{x}{A character vector.}
}
\value{
A character vector with HTML tags and comments stripped off.
}
\description{
Remove HTML tags and comments from text.
}
\examples{
xfun::strip_html("<a href=\"#\">Hello <!-- comment -->world!</a>")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{is_sub_path}
\alias{is_sub_path}
\title{Test if a path is a subpath of a dir}
\usage{
is_sub_path(x, dir, n = nchar(dir))
}
\arguments{
\item{x}{A vector of paths.}

\item{dir}{A vector of directory paths.}

\item{n}{The length of \code{dir} paths.}
}
\value{
A logical vector.
}
\description{
Check if the path starts with the dir path.
}
\note{
You may want to normalize the values of the \code{x} and \code{dir} arguments
first (with \code{\link[=normalize_path]{normalize_path()}}), to make sure the path separators
are consistent.
}
\examples{
xfun::is_sub_path("a/b/c.txt", "a/b")  # TRUE
xfun::is_sub_path("a/b/c.txt", "d/b")  # FALSE
xfun::is_sub_path("a/b/c.txt", "a\\\\b")  # FALSE (even on Windows)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{process_file}
\alias{process_file}
\alias{sort_file}
\title{Read a text file, process the text with a function, and write the text back}
\usage{
process_file(file, fun = identity, x = read_utf8(file))

sort_file(..., fun = sort)
}
\arguments{
\item{file}{Path to a text file.}

\item{fun}{A function to process the text.}

\item{x}{The content of the file.}

\item{...}{Arguments to be passed to \code{process_file()}.}
}
\value{
If \code{file} is provided, invisible \code{NULL} (the file is updated as a
side effect), otherwise the processed content (as a character vector).
}
\description{
Read a text file with the UTF-8 encoding, apply a function to the text, and
write back to the original file if the processed text is different with the
original input.
}
\details{
\code{sort_file()} is an application of \code{process_file()}, with the processing
function being \code{\link[=sort]{sort()}}, i.e., it sorts the text lines in a file and write
back the sorted text.
}
\examples{
f = tempfile()
xfun::write_utf8("Hello World", f)
xfun::process_file(f, function(x) gsub("World", "woRld", x))
xfun::read_utf8(f)  # see if it has been updated
file.remove(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{optipng}
\alias{optipng}
\title{Run OptiPNG on all PNG files under a directory}
\usage{
optipng(dir = ".")
}
\arguments{
\item{dir}{Path to a directory.}
}
\description{
Calls the command \command{optipng} to optimize all PNG files under a
directory.
}
\references{
OptiPNG: \url{http://optipng.sourceforge.net}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{retry}
\alias{retry}
\title{Retry calling a function for a number of times}
\usage{
retry(fun, ..., .times = 3, .pause = 5)
}
\arguments{
\item{fun}{A function.}

\item{...}{Arguments to be passed to the function.}

\item{.times}{The number of times.}

\item{.pause}{The number of seconds to wait before the next attempt.}
}
\description{
If the function returns an error, retry it for the specified number of
times, with a pause between attempts.
}
\details{
One application of this function is to download a web resource. Since the
download might fail sometimes, you may want to retry it for a few more times.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# read the Github releases info of the repo yihui/xfun
xfun::retry(xfun::github_releases, "yihui/xfun")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{optipng}
\alias{optipng}
\title{Run OptiPNG on all PNG files under a directory}
\usage{
optipng(dir = ".", files = all_files("[.]png$", dir), ...)
}
\arguments{
\item{dir}{Path to a directory.}

\item{files}{Alternatively, you can choose the specific files to optimize.}

\item{...}{Arguments to be passed to \code{system2()}.}
}
\description{
Call the command \command{optipng} via \code{system2()} to optimize all PNG
files under a directory.
}
\references{
OptiPNG: \url{https://optipng.sourceforge.net}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{tabset}
\alias{tabset}
\alias{tab_content}
\title{Represent a (recursive) list with (nested) tabsets}
\usage{
tabset(x, value = str)

tab_content(x)
}
\arguments{
\item{x}{A list for \code{tabset()}, and a character vector for \code{tab_content()}.}

\item{value}{A function to print the value of a list member. By default,
\code{\link[=str]{str()}} is used to print the structure of the value. You may also use
\code{\link[=dput]{dput()}} to output the full value, but it may be slow when the size of the
value is too big. If the function prints the value to the console, the
output will be captured and shown in a code block. If you want construct
the tab content by yourself instead of capturing console output as a code
block, you may pass the Markdown content to \code{\link[=tab_content]{tab_content()}} and return it
in this \code{value()} function.}
}
\value{
A character vector of Markdown that can be rendered to HTML with
\code{\link[litedown:mark]{litedown::mark()}}.
}
\description{
The tab titles are names of list members, and the tab content contains the
values of list members. If a list member is also a list, it will be
represented recursively with a child tabset.
}
\examples{
xfun::tabset(iris)
xfun::tabset(iris, dput)
xfun::tabset(iris, print)

# a deeply nested list
plot(1:10)
p = recordPlot()
xfun::tabset(p)

# custom tab content
xfun::tabset(iris, function(x) {
    if (is.factor(x)) {
        res = c("A factor with levels: ", xfun::join_words(levels(x), before = "`"))
        xfun::tab_content(res)
    } else print(summary(x))
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{msg_cat}
\alias{msg_cat}
\title{Generate a message with \code{cat()}}
\usage{
msg_cat(...)
}
\arguments{
\item{...}{Character strings of messages, which will be concatenated into one
string via \code{paste(c(...), collapse = '')}.}
}
\value{
Invisible \code{NULL}, with the side-effect of printing the message.
}
\description{
This function is similar to \code{\link[=message]{message()}}, and the difference is
that \code{msg_cat()} uses \code{\link[=cat]{cat()}} to write out the message,
which is sent to \code{\link[=stdout]{stdout()}} instead of \code{\link[=stderr]{stderr()}}. The
message can be suppressed by \code{\link[=suppressMessages]{suppressMessages()}}.
}
\note{
By default, a newline will not be appended to the message. If you need
a newline, you have to explicitly add it to the message (see
\sQuote{Examples}).
}
\examples{
{
    # a message without a newline at the end
    xfun::msg_cat("Hello world!")
    # add a newline at the end
    xfun::msg_cat(" This message appears right after the previous one.\n")
}
suppressMessages(xfun::msg_cat("Hello world!"))
}
\seealso{
This function was inspired by \code{rlang::inform()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{retry}
\alias{retry}
\title{Retry calling a function for a number of times}
\usage{
retry(fun, ..., .times = 3, .pause = 5)
}
\arguments{
\item{fun}{A function.}

\item{...}{Arguments to be passed to the function.}

\item{.times}{The number of times.}

\item{.pause}{The number of seconds to wait before the next attempt.}
}
\description{
If the function returns an error, retry it for the specified number of
times, with a pause between attempts.
}
\details{
One application of this function is to download a web resource. Since the
download might fail sometimes, you may want to retry it for a few more times.
}
\examples{\dontshow{if (interactive()) \{}
# read the GitHub releases info of the repo yihui/xfun
xfun::retry(xfun::github_releases, "yihui/xfun")
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{alnum_id}
\alias{alnum_id}
\title{Generate ID strings}
\usage{
alnum_id(x, exclude = "[^[:alnum:]]+")
}
\arguments{
\item{x}{A character vector.}

\item{exclude}{A (Perl) regular expression to detect characters to be
replaced by dashes. By default, non-alphanumeric characters are replaced.}
}
\value{
A character vector of IDs.
}
\description{
Substitute certain (by default, non-alphanumeric) characters with dashes and
remove extra dashes at both ends to generate ID strings.
}
\examples{
x = c("Hello world 123!", "a  &b*^##c 456")
xfun::alnum_id(x)
xfun::alnum_id(x, "[^[:alpha:]]+")  # only keep alphabetical chars
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{md_table}
\alias{md_table}
\title{Generate a simple Markdown pipe table}
\usage{
md_table(x, digits = NULL, na = NULL, newline = NULL, limit = NULL)
}
\arguments{
\item{x}{A 2-dimensional object (e.g., a matrix or data frame).}

\item{digits}{The number of decimal places to be passed to \code{\link[=round]{round()}}. It can
be a integer vector of the same length as the number of columns in \code{x} to
round columns separately. The default is \code{3}.}

\item{na}{A character string to represent \code{NA} values. The default is an
empty string.}

\item{newline}{A character string to substitute \verb{\\n} in \code{x} (because pipe
tables do not support line breaks in cells). The default is a space.}

\item{limit}{The maximum number of rows to show in the table. If it is
smaller than the number of rows, the data in the middle will be omitted. If
it is of length 2, the second number will be used to limit the number of
columns. Zero and negative values are ignored.}
}
\value{
A character vector.
}
\description{
A minimal Markdown table generator using the pipe \code{|} as column separators.
}
\details{
The default argument values can be set via global options with the prefix
\code{xfun.md_table.}, e.g., \verb{options(xfun.md_table.digits 2, xfun.md_table.na = 'n/a')}.
}
\examples{
xfun::md_table(head(iris))
xfun::md_table(mtcars, limit = c(10, 6))
}
\seealso{
\code{\link[knitr:kable]{knitr::kable()}} (which supports more features)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{split_source}
\alias{split_source}
\title{Split source lines into complete expressions}
\usage{
split_source(x)
}
\arguments{
\item{x}{A character vector of R source code.}
}
\value{
A list of character vectors, and each vector contains a complete R
  expression.
}
\description{
Parse the lines of code one by one to find complete expressions in the code,
and put them in a list.
}
\examples{
xfun::split_source(c("if (TRUE) {", "1 + 1", "}", "print(1:5)"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github.R
\name{post_release}
\alias{post_release}
\title{Perform post-CRAN-release routine tasks}
\usage{
post_release()
}
\description{
After a CRAN release is accepted, commit and tag the release, bump the
package version to the next patch dev version, prepend a new header to
\file{NEWS.md}, push to the remote repository, and create a GitHub release
with the news items of this version.
}
\details{
The version number \code{X.Y} or \code{X.Y.0} is extracted from the \code{Version} field in
\file{DESCRIPTION}. For example, if the current version is \code{0.58} (or
\verb{0.58.0}), the repo will be tagged as \code{v0.58} (or \code{v0.58.0}). The patch
version is then bumped to \verb{0.58.1}, and a new header for the next minor
version \code{0.59} (or \verb{0.59.0}) is prepended to \file{NEWS.md}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{split_source}
\alias{split_source}
\title{Split source lines into complete expressions}
\usage{
split_source(x)
}
\arguments{
\item{x}{A character vector of R source code.}
}
\value{
A list of character vectors, and each vector contains a complete R
expression.
}
\description{
Parse the lines of code one by one to find complete expressions in the code,
and put them in a list.
}
\examples{
xfun::split_source(c("if (TRUE) {", "1 + 1", "}", "print(1:5)"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{normalize_path}
\alias{normalize_path}
\title{Normalize paths}
\usage{
normalize_path(x, winslash = "/", must_work = FALSE, resolve_symlink = TRUE)
}
\arguments{
\item{x, winslash, must_work}{Arguments passed to
\code{\link[=normalizePath]{normalizePath()}}.}

\item{resolve_symlink}{Whether to resolve symbolic links.}
}
\description{
A wrapper function of \code{normalizePath()} with different defaults.
}
\examples{
library(xfun)
normalize_path("~")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{alnum_id}
\alias{alnum_id}
\title{Generate ID strings}
\usage{
alnum_id(x, exclude = "[^[:alnum:]]+")
}
\arguments{
\item{x}{A character vector.}

\item{exclude}{A (Perl) regular expression to detect characters to be
replaced by dashes. By default, non-alphanumeric characters are replaced.}
}
\value{
A character vector of IDs.
}
\description{
Substitute certain (by default, non-alphanumeric) characters with dashes and
remove extra dashes at both ends to generate ID strings. This function is
intended for generating IDs for HTML elements, so HTML tags in the input text
will be removed first.
}
\examples{
x = c("Hello world 123!", "a  &b*^##c 456")
xfun::alnum_id(x)
xfun::alnum_id(x, "[^[:alpha:]]+")  # only keep alphabetical chars
# when text contains HTML tags
xfun::alnum_id("<h1>Hello <strong>world</strong>!")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{is_abs_path}
\alias{is_abs_path}
\alias{is_rel_path}
\title{Test if paths are relative or absolute}
\usage{
is_abs_path(x)

is_rel_path(x)
}
\arguments{
\item{x}{A vector of paths.}
}
\value{
A logical vector.
}
\description{
On Unix, check if the paths start with \file{/} or \file{~} (if they do, they
are absolute paths). On Windows, check if a path remains the same (via
\code{xfun::\link{same_path}()}) if it is prepended with \file{./} (if it
does, it is a relative path).
}
\examples{
xfun::is_abs_path(c("C:/foo", "foo.txt", "/Users/john/", tempdir()))
xfun::is_rel_path(c("C:/foo", "foo.txt", "/Users/john/", tempdir()))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{split_source}
\alias{split_source}
\title{Split source lines into complete expressions}
\usage{
split_source(x, merge_comments = FALSE, line_number = FALSE)
}
\arguments{
\item{x}{A character vector of R source code.}

\item{merge_comments}{Whether to merge consecutive lines of comments as a
single expression to be combined with the next non-comment expression (if
any).}

\item{line_number}{Whether to store the line numbers of each expression in
the returned value.}
}
\value{
A list of character vectors, and each vector contains a complete R
expression, with an attribute \code{lines} indicating the starting and ending
line numbers of the expression if the argument \code{line_number = TRUE}.
}
\description{
Parse the lines of code one by one to find complete expressions in the code,
and put them in a list.
}
\examples{
code = c("# comment 1", "# comment 2", "if (TRUE) {", "1 + 1", "}", "print(1:5)")
xfun::split_source(code)
xfun::split_source(code, merge_comments = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{install_dir}
\alias{install_dir}
\title{Install a source package from a directory}
\usage{
install_dir(pkg, build = TRUE, build_opts = NULL, install_opts = NULL)
}
\arguments{
\item{pkg}{The package source directory.}

\item{build}{Whether to build a tarball from the source directory. If
\code{FALSE}, run \command{R CMD INSTALL} on the directory directly (note
that vignettes will not be automatically built).}

\item{build_opts}{The options for \command{R CMD build}.}

\item{install_opts}{The options for \command{R CMD INSTALL}.}
}
\value{
Invisible status from \command{R CMD INSTALL}.
}
\description{
Run \command{R CMD build} to build a tarball from a source directory, and run
\command{R CMD INSTALL} to install it.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{attr}
\alias{attr}
\title{Obtain an attribute of an object without partial matching}
\usage{
attr(...)
}
\arguments{
\item{...}{Passed to \code{base::\link[base]{attr}()} (without the
\code{exact} argument).}
}
\description{
An abbreviation of \code{base::\link[base]{attr}(exact = TRUE)}.
}
\examples{
z = structure(list(a = 1), foo = 2)
base::attr(z, "f")  # 2
xfun::attr(z, "f")  # NULL
xfun::attr(z, "foo")  # 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{install_dir}
\alias{install_dir}
\title{Install a source package from a directory}
\usage{
install_dir(pkg = ".", build = TRUE, build_opts = NULL, install_opts = NULL)
}
\arguments{
\item{pkg}{The package source directory.}

\item{build}{Whether to build a tarball from the source directory. If
\code{FALSE}, run \command{R CMD INSTALL} on the directory directly (note
that vignettes will not be automatically built).}

\item{build_opts}{The options for \command{R CMD build}.}

\item{install_opts}{The options for \command{R CMD INSTALL}.}
}
\value{
Invisible status from \command{R CMD INSTALL}.
}
\description{
Run \command{R CMD build} to build a tarball from a source directory, and run
\command{R CMD INSTALL} to install it.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{is_abs_path}
\alias{is_abs_path}
\alias{is_rel_path}
\title{Test if paths are relative or absolute}
\usage{
is_abs_path(x)

is_rel_path(x)
}
\arguments{
\item{x}{A vector of paths.}
}
\value{
A logical vector.
}
\description{
On Unix, check if the paths start with \file{/} or \file{~} (if they do, they
are absolute paths). On Windows, check if a path remains the same (via
\code{\link[=same_path]{same_path()}}) if it is prepended with \file{./} (if it does, it is a
relative path).
}
\examples{
xfun::is_abs_path(c("C:/foo", "foo.txt", "/Users/john/", tempdir()))
xfun::is_rel_path(c("C:/foo", "foo.txt", "/Users/john/", tempdir()))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rstudio.R
\name{rstudio_type}
\alias{rstudio_type}
\title{Type a character vector into the RStudio source editor}
\usage{
rstudio_type(x, pause = function() 0.1, mistake = 0, save = 0)
}
\arguments{
\item{x}{A character vector.}

\item{pause}{A function to return a number in seconds to pause after typing
each character.}

\item{mistake}{The probability of making random mistakes when typing the next
character. A random mistake is a random string typed into the editor and
deleted immediately.}

\item{save}{The probability of saving the document after typing each
character. Note that If a document is not opened from a file, it will never
be saved.}
}
\description{
Use the \pkg{rstudioapi} package to insert characters one by one into the
RStudio source editor, as if they were typed by a human.
}
\examples{
library(xfun)
if (loadable("rstudioapi") && rstudioapi::isAvailable()) {
    rstudio_type("Hello, RStudio! xfun::rstudio_type() looks pretty cool!",
        pause = function() runif(1, 0, 0.5), mistake = 0.1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R
\name{new_app}
\alias{new_app}
\title{Create a local web application}
\usage{
new_app(name, handler, open = interactive(), ports = 4321 + 1:10)
}
\arguments{
\item{name}{The app name (a character string, and each app should have a
unique name).}

\item{handler}{A function that takes the HTTP request information (the first
argument is the requested path) and returns a response.}

\item{open}{Whether to open the app, or a function to open the app URL.}

\item{ports}{A vector of ports to try for starting the server.}
}
\value{
The app URL of the form \verb{http://127.0.0.1:port/custom/name/}.
}
\description{
An experimental function to create a local web application based on R's
internal \code{httpd} server (which is primarily for running R's dynamic help
system).
}
\note{
This function is not based on base R's public API, and is possible to
break in the future, which is also why the documentation here is terse.
Please avoid creating public-facing web apps with it. You may consider
packages like \pkg{httpuv} and \pkg{Rserve} for production web apps.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session.R
\name{do_once}
\alias{do_once}
\title{Perform a task once in an R session}
\usage{
do_once(
  task,
  option,
  hint = c("You will not see this message again in this R session.",
    "If you never want to see this message,",
    sprintf("you may set options(\%s = FALSE) in your .Rprofile.", option))
)
}
\arguments{
\item{task}{Any R code expression to be evaluated once to perform a task,
e.g., \code{warning('Danger!')} or \code{message('Today is ', Sys.Date())}.}

\item{option}{An R option name. This name should be as unique as possible in
\code{\link[=options]{options()}}. After the task has been successfully performed,
this option will be set to \code{FALSE} in the current R session, to
prevent the task from being performed again the next time when
\code{do_once()} is called.}

\item{hint}{A character vector to provide a hint to users on how not to
perform the task or see the message again in the current R session. Set
\code{hint = ""} if you do not want to provide the hint.}
}
\value{
The value returned by the \code{task}, invisibly.
}
\description{
Perform a task once in an R session, e.g., emit a message or warning. Then
give users an optional hint on how not to perform this task at all.
}
\examples{
do_once(message("Today's date is ", Sys.Date()), "xfun.date.reminder")
# if you run it again, it will not emit the message again
do_once(message("Today's date is ", Sys.Date()), "xfun.date.reminder")

do_once({
    Sys.sleep(2)
    1 + 1
}, "xfun.task.1plus1")
do_once({
    Sys.sleep(2)
    1 + 1
}, "xfun.task.1plus1")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R
\name{new_app}
\alias{new_app}
\alias{stop_app}
\title{Create or stop a local web application}
\usage{
new_app(name, handler, open = interactive(), host = "127.0.0.1", port = NULL)

stop_app(name = names(.proxy$apps))
}
\arguments{
\item{name}{App name (a character string). Use \code{name = ''} to serve the app
at the root URL \verb{http://host:PORT/} via the proxy. If \code{name} is non-empty,
the app is served directly by R's internal httpd at \verb{/custom/name/}. Each
name must be unique; calling \code{new_app()} with an existing name replaces
that app. For \code{stop_app()}, a character vector of app names to stop;
defaults to all running apps.}

\item{handler}{A function with signature \verb{function(path, query, post, headers)} that handles HTTP requests and returns a response list.}

\item{open}{Whether to open the app URL in a browser, or a function to open
it. In non-interactive sessions this also controls whether the call
blocks: passing \code{open = FALSE} explicitly suppresses both browser-opening
and blocking.}

\item{host}{Bind address for the proxy (\code{"127.0.0.1"} or \code{"0.0.0.0"}). Only
used when \code{name = ''}.}

\item{port}{Candidate proxy ports when \code{name = ''}. The first available port
is selected (falling back to \code{\link[=random_port]{random_port()}} if needed). Ignored when
\code{name != ''}.}
}
\value{
\code{new_app()} returns the app URL invisibly. \code{stop_app()} returns
nothing.
}
\description{
Create a local web application based on R's internal httpd.
}
\details{
\code{new_app()} has two modes:
\itemize{
\item \code{name = ''}: start an app behind a lightweight proxy at
\verb{http://host:PORT/}, forwarding to R's internal httpd path
\verb{/custom/xfun:PORT/}.
\item \code{name != ''}: register a legacy app directly on R's internal httpd at
\verb{http://127.0.0.1:BACKEND/custom/name/} (compatible with older versions).
}

\code{stop_app()} deregisters one or more running apps.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{attr}
\alias{attr}
\title{Obtain an attribute of an object without partial matching}
\usage{
attr(...)
}
\arguments{
\item{...}{Passed to \code{\link[base:attr]{base::attr()}} (without the \code{exact} argument).}
}
\description{
An abbreviation of \code{\link[base:attr]{base::attr}}\code{(exact = TRUE)}.
}
\examples{
z = structure(list(a = 1), foo = 2)
base::attr(z, "f")  # 2
xfun::attr(z, "f")  # NULL
xfun::attr(z, "foo")  # 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/browser.R
\name{browser_print}
\alias{browser_print}
\title{Print a web page to PDF/PNG/JPEG}
\usage{
browser_print(
  url,
  output = ".pdf",
  args = "default",
  window_size = c(1280, 1024),
  browser = env_option("xfun.browser", find_browser())
)
}
\arguments{
\item{url}{Path or URL to the page to be printed.}

\item{output}{An output filename. If only an extension is provided, the
filename will be inferred from \code{url} via \code{\link[=url_filename]{url_filename()}}. Only \code{.pdf},
\code{.png}, and \code{.jpeg} are supported.}

\item{args}{Command-line arguments to be passed to the headless browser. The
default arguments can be found in \code{xfun:::browser_args()}. You may pass
additional arguments on top of these via, e.g., \code{c('default', '--no-pdf-header-footer')}, or completely override the default arguments by
providing a vector of other arguments.}

\item{window_size}{The browser window size when printing to PNG/JPEG.}

\item{browser}{Path to the web browser. By default, the browser is found via
\code{xfun:::find_browser()}. If it cannot be found, you may set the global
option \code{options(xfun.browser = )} or environment variable \code{R_XFUN_BROWSER}
to the path of the browser, or simply pass the path to the \code{browser}
argument.}
}
\value{
The \code{output} path if the web page is successfully printed.
}
\description{
Print a web page via a headless browser such as Chromium or Google Chrome.
}
\examples{\dontshow{if (interactive()) \{}
xfun::browser_print("https://www.r-project.org")
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{try_error}
\alias{try_error}
\title{Try an expression and see if it throws an error}
\usage{
try_error(expr)
}
\arguments{
\item{expr}{An R expression.}
}
\value{
\code{TRUE} (error) or \code{FALSE} (success).
}
\description{
Use \code{\link[=tryCatch]{tryCatch()}} to check if an expression throws an error.
}
\examples{
xfun::try_error(stop("foo"))  # TRUE
xfun::try_error(1:10)  # FALSE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{file_string}
\alias{file_string}
\title{Read a text file and concatenate the lines by \code{'\\n'}}
\usage{
file_string(file)
}
\arguments{
\item{file}{Path to a text file (should be encoded in UTF-8).}
}
\value{
A character string of text lines concatenated by \code{'\\n'}.
}
\description{
The source code of this function should be self-explanatory.
}
\examples{
xfun::file_string(system.file("DESCRIPTION", package = "xfun"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image.R
\name{shrink_images}
\alias{shrink_images}
\title{Shrink images to a maximum width}
\usage{
shrink_images(
  width = 800,
  dir = ".",
  files = all_files("[.](png|jpe?g|webp)$", dir),
  tinify = FALSE
)
}
\arguments{
\item{width}{The desired maximum width of images.}

\item{dir}{The directory of images.}

\item{files}{A vector of image file paths. By default, this is all
\file{.png}, \file{.jpeg}, and \file{.webp} images under \code{dir}.}

\item{tinify}{Whether to compress images using \code{\link{tinify}()}.}
}
\description{
Use \code{\link[magick:image_resize]{magick::image_resize}()} to shrink an
image if its width is larger than the value specified by the argument
\code{width}, and optionally call \code{\link{tinify}()} to compress it.
}
\examples{
f = xfun:::all_files("[.](png|jpe?g)$", R.home("doc"))
file.copy(f, tempdir())
f = file.path(tempdir(), basename(f))
magick::image_info(magick::image_read(f))  # some widths are larger than 300
xfun::shrink_images(300, files = f)
magick::image_info(magick::image_read(f))  # all widths <= 300 now
file.remove(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/encoding.R
\name{is_ascii}
\alias{is_ascii}
\title{Check if a character vector consists of entirely ASCII characters}
\usage{
is_ascii(x)
}
\arguments{
\item{x}{A character vector.}
}
\value{
A logical vector indicating whether each element of the character
  vector is ASCII.
}
\description{
Converts the encoding of a character vector to \code{'ascii'}, and check if
the result is \code{NA}.
}
\examples{
library(xfun)
is_ascii(letters)  # yes
is_ascii(intToUtf8(8212))  # no
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{dir_exists}
\alias{dir_exists}
\alias{file_exists}
\title{Test the existence of files and directories}
\usage{
dir_exists(x)

file_exists(x)
}
\arguments{
\item{x}{A vector of paths.}
}
\value{
A logical vector.
}
\description{
These are wrapper functions of [\verb{utils::file_test()]} to test the
existence of directories and files. Note that \code{file_exists()} only tests
files but not directories, which is the main difference between
\code{\link[=file.exists]{file.exists()}} in base R. If you use are using the R version
3.2.0 or above, \code{dir_exists()} is the same as \code{\link[=dir.exists]{dir.exists()}}
in base R.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{read_all}
\alias{read_all}
\title{Read all text files and concatenate their content}
\usage{
read_all(files, before = function(f) NULL, after = function(f) NULL)
}
\arguments{
\item{files}{A vector of file paths.}

\item{before, after}{A function that takes one file path as the input and
returns values to be added before or after the content of the file.
Alternatively, they can be constant values to be added.}
}
\value{
A character vector.
}
\description{
Read files one by one, and optionally add text before/after the content. Then
combine all content into one character vector.
}
\examples{
# two files in this package
fs = system.file("scripts", c("call-fun.R", "child-pids.sh"), package = "xfun")
xfun::read_all(fs)

# add file paths before file content and an empty line after content
xfun::read_all(fs, before = function(f) paste("#-----", f, "-----"), after = "")

# add constants
xfun::read_all(fs, before = "/*", after = c("*/", ""))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/os.R
\name{is_windows}
\alias{is_windows}
\alias{is_unix}
\alias{is_macos}
\alias{is_linux}
\alias{is_arm64}
\title{Test for types of operating systems}
\usage{
is_windows()

is_unix()

is_macos()

is_linux()

is_arm64()
}
\description{
Functions based on \code{.Platform$OS.type} and \code{Sys.info()} to test if
the current operating system is Windows, macOS, Unix, or Linux.
}
\examples{
library(xfun)
# only one of the following statements should be true
is_windows()
is_unix() && is_macos()
is_linux()
# In newer Macs, CPU can be either Intel or Apple
is_arm64()  # TRUE on Apple silicone machines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{relative_path}
\alias{relative_path}
\title{Get the relative path of a path relative to a directory}
\usage{
relative_path(x, dir = ".", use.. = TRUE, error = TRUE)
}
\arguments{
\item{x}{A vector of paths to be converted to relative paths.}

\item{dir}{Path to a directory.}

\item{use..}{Whether to use double-dots (\file{..}) in the relative path. A
double-dot indicates the parent directory (starting from the directory
provided by the \code{dir} argument).}

\item{error}{Whether to signal an error if a path cannot be converted to a
relative path.}
}
\value{
A vector of relative paths if the conversion succeeded; otherwise the
original paths when \code{error = FALSE}, and an error when \code{error = TRUE}.
}
\description{
Given a directory, return the relative path that is relative to this
directory. For example, the path \file{foo/bar.txt} relative to the directory
\file{foo/} is \file{bar.txt}, and the path \file{/a/b/c.txt} relative to
\file{/d/e/} is \file{../../a/b/c.txt}.
}
\examples{
xfun::relative_path("foo/bar.txt", "foo/")
xfun::relative_path("foo/bar/a.txt", "foo/haha/")
xfun::relative_path(getwd())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{set_envvar}
\alias{set_envvar}
\title{Set environment variables}
\usage{
set_envvar(vars)
}
\arguments{
\item{vars}{A named character vector of the form \code{c(VARIABLE = VALUE)}.
If any value is \code{NA}, this function will try to unset the variable.}
}
\value{
Old values of the variables (if not set, \code{NA}).
}
\description{
Set environment variables from a named character vector, and return the old
values of the variables, so they could be restored later.
}
\details{
The motivation of this function is that \code{\link{Sys.setenv}()} does not
return the old values of the environment variables, so it is not
straightforward to restore the variables later.
}
\examples{
vars = xfun::set_envvar(c(FOO = "1234"))
Sys.getenv("FOO")
xfun::set_envvar(vars)
Sys.getenv("FOO")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{bump_version}
\alias{bump_version}
\title{Bump version numbers}
\usage{
bump_version(x)
}
\arguments{
\item{x}{A vector of version numbers (of the class \code{"numeric_version"}),
or values that can be coerced to version numbers via
\code{as.numeric_version()}.}
}
\value{
A vector of new version numbers.
}
\description{
Increase the last digit of version numbers, e.g., from \code{0.1} to
\code{0.2}, or \code{7.23.9} to \code{7.23.10}.
}
\examples{
xfun::bump_version(c("0.1", "91.2.14"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{magic_path}
\alias{magic_path}
\title{Find a file or directory under a root directory}
\usage{
magic_path(
  ...,
  root = proj_root(),
  relative = TRUE,
  error = TRUE,
  message = getOption("xfun.magic_path.message", TRUE),
  n_dirs = getOption("xfun.magic_path.n_dirs", 10000)
)
}
\arguments{
\item{...}{A character vector of path components.}

\item{root}{The root directory under which to search for the path. If
\code{NULL}, the current working directory is used.}

\item{relative}{Whether to return a relative path.}

\item{error}{Whether to signal an error if the path is not found, or multiple
paths are found.}

\item{message}{Whether to emit a message when multiple paths are found and
\code{error = FALSE}.}

\item{n_dirs}{The number of subdirectories to recursively search. The
recursive search may be time-consuming when there are a large number of
subdirectories under the root directory. If you really want to search for
all subdirectories, you may try \code{n_dirs = Inf}.}
}
\value{
The path found under the root directory, or an error when \code{error
  = TRUE} and the path is not found (or multiple paths are found).
}
\description{
Given a path, try to find it recursively under a root directory. The input
path can be an incomplete path, e.g., it can be a base filename, and
\code{magic_path()} will try to find this file under subdirectories.
}
\examples{
\dontrun{
xfun::magic_path("mtcars.csv")  # find any file that has the base name mtcars.csv
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{str_wrap}
\alias{str_wrap}
\title{Wrap character vectors}
\usage{
str_wrap(...)
}
\arguments{
\item{...}{Arguments passed to \code{\link[=strwrap]{strwrap()}}.}
}
\value{
A character vector.
}
\description{
A wrapper function to make \code{\link[=strwrap]{strwrap()}} return a character vector of the same
length as the input vector; each element of the output vector is a string
formed by concatenating wrapped strings by \verb{\\n}.
}
\examples{
x = sample(c(letters, " "), 200, TRUE, c(rep(0.5/26, 26), 0.5))
x = rep(paste(x, collapse = ""), 2)
strwrap(x, width = 30)
xfun::str_wrap(x, width = 30)  # same length as x
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{bg_process}
\alias{bg_process}
\title{Start a background process}
\usage{
bg_process(
  command,
  args = character(),
  verbose = getOption("xfun.bg_process.verbose", FALSE)
)
}
\arguments{
\item{command, args}{The system command and its arguments. They do not need to
be quoted, since they will be quoted via \code{\link{shQuote}()}
internally.}

\item{verbose}{If \code{FALSE}, suppress the output from \verb{stdout} (and
also \verb{stderr} on Windows). The default value of this argument can be
set via a global option, e.g., \code{options(xfun.bg_process.verbose =
TRUE)}.}
}
\value{
The process ID as a character string.
}
\description{
Start a background process using the PowerShell cmdlet \command{Start-Process
-PassThru} on Windows or the ampersand \command{&} on Unix, and return the
process ID.
}
\note{
On Windows, if PowerShell is not available, try to use
  \code{\link{system2}(wait = FALSE)} to start the background process
  instead. The process ID will be identified from the output of the command
  \command{tasklist}. This method of looking for the process ID may not be
  reliable. If the search is not successful in 30 seconds, it will throw an
  error (timeout). If a longer time is needed, you may set
  \code{options(xfun.bg_process.timeout)} to a larger value, but it should be
  very rare that a process cannot be started in 30 seconds. When you reach
  the timeout, it is more likely that the command actually failed.
}
\seealso{
\code{\link{proc_kill}()} to kill a process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image.R
\name{shrink_images}
\alias{shrink_images}
\title{Shrink images to a maximum width}
\usage{
shrink_images(
  width = 800,
  dir = ".",
  files = all_files("[.](png|jpe?g|webp)$", dir),
  tinify = FALSE
)
}
\arguments{
\item{width}{The desired maximum width of images.}

\item{dir}{The directory of images.}

\item{files}{A vector of image file paths. By default, this is all
\file{.png}, \file{.jpeg}, and \file{.webp} images under \code{dir}.}

\item{tinify}{Whether to compress images using \code{\link[=tinify]{tinify()}}.}
}
\description{
Use \code{\link[magick:transform]{magick::image_resize()}} to shrink an
image if its width is larger than the value specified by the argument
\code{width}, and optionally call \code{\link[=tinify]{tinify()}} to compress it.
}
\examples{\dontshow{if (xfun::loadable('magick')) \{}
f = xfun:::all_files("[.](png|jpe?g)$", R.home("doc"))
file.copy(f, tempdir())
f = file.path(tempdir(), basename(f))
magick::image_info(magick::image_read(f))  # some widths are larger than 300
xfun::shrink_images(300, files = f)
magick::image_info(magick::image_read(f))  # all widths <= 300 now
file.remove(f)
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R, R/github.R
\name{rest_api}
\alias{rest_api}
\alias{rest_api_raw}
\alias{github_api}
\title{Get data from a REST API}
\usage{
rest_api(...)

rest_api_raw(root, endpoint, token = "", params = list(), headers = NULL)

github_api(
  endpoint,
  token = "",
  params = list(),
  headers = NULL,
  raw = !loadable("jsonlite")
)
}
\arguments{
\item{...}{Arguments to be passed to \code{rest_api_raw()}.}

\item{root}{The API root URL.}

\item{endpoint}{The API endpoint.}

\item{token}{A named character string (e.g., \code{c(token = "xxxx")}), which
will be used to create an authorization header of the form
\samp{Authorization: NAME TOKEN} for the API call, where \samp{NAME} is the
name of the string and \samp{TOKEN} is the string. If the string does not
have a name, \samp{Basic} will be used as the default name.}

\item{params}{A list of query parameters to be sent with the API call.}

\item{headers}{A named character vector of HTTP headers, e.g., \code{c(Accept
= "application/vnd.github.v3+json")}.}

\item{raw}{Whether to return the raw response or parse the response with
\pkg{jsonlite}.}
}
\value{
A character vector (the raw JSON response) or an R object parsed from
  the JSON text.
}
\description{
Read data from a REST API and optionally with an authorization token in the
request header. The function \code{rest_api_raw()} returns the raw text of
the response, and \code{rest_api()} will parse the response with
\code{jsonlite::fromJSON()} (assuming that the response is in the JSON
format).
}
\details{
These functions are simple wrappers based on \code{\link{url}()} and
\code{\link{read_utf8}()}. Specifically, the \code{headers} argument is
passed to \code{url()}, and \code{read_utf8()} will send a \samp{GET} request
to the API server. This means these functions only support the \samp{GET}
method. If you need to use other HTTP methods (such as \samp{POST}), you have
to use other packages such as \pkg{curl} and \pkg{httr}.

\code{github_api()} is a wrapper function based on
  \code{rest_api_raw()} to obtain data from the Github API:
  \url{https://docs.github.com/en/rest}. You can provide a personal access
  token (PAT) via the \code{token} argument, or via one of the environment
  variables \var{GITHUB_PAT}, \var{GITHUB_TOKEN}, \var{GH_TOKEN}. A PAT
  allows for a much higher rate limit in API calls. Without a token, you can
  only make 60 calls in an hour.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# a normal GET request
xfun::rest_api("https://httpbin.org", "/get")
xfun::rest_api_raw("https://httpbin.org", "/get")

# send the request with an auth header
xfun::rest_api("https://httpbin.org", "/headers", "OPEN SESAME!")

# with query parameters
xfun::rest_api("https://httpbin.org", "/response-headers", params = list(foo = "bar"))

# get the rate limit info from Github
xfun::github_api("/rate_limit")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/encoding.R
\name{is_ascii}
\alias{is_ascii}
\title{Check if a character vector consists of entirely ASCII characters}
\usage{
is_ascii(x)
}
\arguments{
\item{x}{A character vector.}
}
\value{
A logical vector indicating whether each element of the character
vector is ASCII.
}
\description{
Converts the encoding of a character vector to \code{'ascii'}, and check if
the result is \code{NA}.
}
\examples{
library(xfun)
is_ascii(letters)  # yes
is_ascii(intToUtf8(8212))  # no
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{pkg_attach}
\alias{pkg_attach}
\alias{pkg_load}
\alias{loadable}
\alias{pkg_available}
\alias{pkg_attach2}
\alias{pkg_load2}
\title{Attach or load packages, and automatically install missing packages if
requested}
\usage{
pkg_attach(
  ...,
  install = FALSE,
  message = getOption("xfun.pkg_attach.message", TRUE)
)

pkg_load(..., error = TRUE, install = FALSE)

loadable(pkg, strict = TRUE, new_session = FALSE)

pkg_available(pkg, version = NULL)

pkg_attach2(...)

pkg_load2(...)
}
\arguments{
\item{...}{Package names (character vectors, and must always be quoted).}

\item{install}{Whether to automatically install packages that are not
available using \code{\link{install.packages}()}. Besides \code{TRUE} and
\code{FALSE}, the value of this argument can also be a function to install
packages (\code{install = TRUE} is equivalent to \code{install =
install.packages}), or a character string \code{"pak"} (equivalent to
\code{install = pak::pkg_install}, which requires the \pkg{pak} package).
You are recommended to set a CRAN mirror in the global option \code{repos}
via \code{\link{options}()} if you want to automatically install packages.}

\item{message}{Whether to show the package startup messages (if any startup
messages are provided in a package).}

\item{error}{Whether to signal an error when certain packages cannot be loaded.}

\item{pkg}{A single package name.}

\item{strict}{If \code{TRUE}, use \code{\link{requireNamespace}()} to test if
a package is loadable; otherwise only check if the package is in
\code{\link{.packages}(TRUE)} (this does not really load the package, so it
is less rigorous but on the other hand, it can keep the current R session
clean).}

\item{new_session}{Whether to test if a package is loadable in a new R
session. Note that \code{new_session = TRUE} implies \code{strict = TRUE}.}

\item{version}{A minimal version number. If \code{NULL}, only test if a
package is available and do not check its version.}
}
\value{
\code{pkg_attach()} returns \code{NULL} invisibly. \code{pkg_load()}
  returns a logical vector, indicating whether the packages can be loaded.
}
\description{
\code{pkg_attach()} is a vectorized version of \code{\link{library}()} over
the \code{package} argument to attach multiple packages in a single function
call. \code{pkg_load()} is a vectorized version of
\code{\link{requireNamespace}()} to load packages (without attaching them).
The functions \code{pkg_attach2()} and \code{pkg_load2()} are wrappers of
\code{pkg_attach(install = TRUE)} and \code{pkg_load(install = TRUE)},
respectively. \code{loadable()} is an abbreviation of
\code{requireNamespace(quietly = TRUE)}. \code{pkg_available()} tests if a
package with a minimal version is available.
}
\details{
These are convenience functions that aim to solve these common problems: (1)
We often need to attach or load multiple packages, and it is tedious to type
several \code{library()} calls; (2) We are likely to want to install the
packages when attaching/loading them but they have not been installed.
}
\examples{
library(xfun)
pkg_attach("stats", "graphics")
# pkg_attach2('servr') # automatically install servr if it is not installed

(pkg_load("stats", "graphics"))
}
\seealso{
\code{pkg_attach2()} is similar to \code{pacman::p_load()}, but does
  not allow non-standard evaluation (NSE) of the \code{...} argument, i.e.,
  you must pass a real character vector of package names to it, and all names
  must be quoted. Allowing NSE adds too much complexity with too little gain
  (the only gain is that it saves your effort in typing two quotes).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{read_all}
\alias{read_all}
\title{Read all text files and concatenate their content}
\usage{
read_all(files, before = function(f, x) NULL, after = function(f, x) NULL)
}
\arguments{
\item{files}{A vector of file paths.}

\item{before, after}{A function that takes one file path and its content as
the input and returns values to be added before or after the content of the
file. Alternatively, they can be constant values to be added.}
}
\value{
A character vector.
}
\description{
Read files one by one, and optionally add text before/after the content. Then
combine all content into one character vector.
}
\examples{
# two files in this package
fs = system.file("scripts", c("call-fun.R", "child-pids.sh"), package = "xfun")
xfun::read_all(fs)

# add file paths before file content and an empty line after content
xfun::read_all(fs, before = function(f) paste("#-----", f, "-----"), after = "")

# add constants
xfun::read_all(fs, before = "/*", after = c("*/", ""))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{set_envvar}
\alias{set_envvar}
\title{Set environment variables}
\usage{
set_envvar(vars)
}
\arguments{
\item{vars}{A named character vector of the form \code{c(VARIABLE = VALUE)}.
If any value is \code{NA}, this function will try to unset the variable.}
}
\value{
Old values of the variables (if not set, \code{NA}).
}
\description{
Set environment variables from a named character vector, and return the old
values of the variables, so they could be restored later.
}
\details{
The motivation of this function is that \code{\link[=Sys.setenv]{Sys.setenv()}} does not
return the old values of the environment variables, so it is not
straightforward to restore the variables later.
}
\examples{
vars = xfun::set_envvar(c(FOO = "1234"))
Sys.getenv("FOO")
xfun::set_envvar(vars)
Sys.getenv("FOO")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/encoding.R
\name{native_encode}
\alias{native_encode}
\title{Try to use the system native encoding to represent a character vector}
\usage{
native_encode(x, windows_only = is_windows())
}
\arguments{
\item{x}{A character vector.}

\item{windows_only}{Whether to make the attempt on Windows only. On Unix,
characters are typically encoded in the native encoding (UTF-8), so there
is no need to do the conversion.}
}
\description{
Apply \code{enc2native()} to the character vector, and check if
\code{enc2utf8()} can convert it back without a loss. If it does, return
\code{enc2native(x)}, otherwise return the original vector with a warning.
}
\examples{
library(xfun)
s = intToUtf8(c(20320, 22909))
Encoding(s)

s2 = native_encode(s)
Encoding(s2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{broken_packages}
\alias{broken_packages}
\title{Find out broken packages and reinstall them}
\usage{
broken_packages(reinstall = TRUE)
}
\arguments{
\item{reinstall}{Whether to reinstall the broken packages, or only list their
names.}
}
\value{
A character vector of names of broken package.
}
\description{
If a package is broken (i.e., not \code{\link{loadable}()}), reinstall it.
}
\details{
Installed R packages could be broken for several reasons. One common reason
is that you have upgraded R to a newer \code{x.y} version, e.g., from
\code{4.0.5} to \code{4.1.0}, in which case you need to reinstall previously
installed packages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{bump_version}
\alias{bump_version}
\title{Bump version numbers}
\usage{
bump_version(x)
}
\arguments{
\item{x}{A vector of version numbers (of the class \code{"numeric_version"}),
or values that can be coerced to version numbers via
\code{as.numeric_version()}.}
}
\value{
A vector of new version numbers.
}
\description{
Increase the last digit of version numbers, e.g., from \code{0.1} to
\code{0.2}, or \verb{7.23.9} to \verb{7.23.10}.
}
\examples{
xfun::bump_version(c("0.1", "91.2.14"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{bg_process}
\alias{bg_process}
\title{Start a background process}
\usage{
bg_process(
  command,
  args = character(),
  verbose = getOption("xfun.bg_process.verbose", FALSE)
)
}
\arguments{
\item{command, args}{The system command and its arguments. They do not need to
be quoted, since they will be quoted via \code{\link[=shQuote]{shQuote()}} internally.}

\item{verbose}{If \code{FALSE}, suppress the output from \code{stdout} (and also
\code{stderr} on Windows). The default value of this argument can be set via a
global option, e.g., \code{options(xfun.bg_process.verbose = TRUE)}.}
}
\value{
The process ID as a character string.
}
\description{
Start a background process using the PowerShell cmdlet
\command{Start-Process-PassThru} on Windows or the ampersand \command{&} on
Unix, and return the process ID.
}
\note{
On Windows, if PowerShell is not available, try to use
\code{\link{system2}}\code{(wait = FALSE)} to start the background process instead. The
process ID will be identified from the output of the command
\command{tasklist}. This method of looking for the process ID may not be
reliable. If the search is not successful in 30 seconds, it will throw an
error (timeout). If a longer time is needed, you may set
\code{options(xfun.bg_process.timeout)} to a larger value, but it should be very
rare that a process cannot be started in 30 seconds. When you reach the
timeout, it is more likely that the command actually failed.
}
\seealso{
\code{\link[=proc_kill]{proc_kill()}} to kill a process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{magic_path}
\alias{magic_path}
\title{Find a file or directory under a root directory}
\usage{
magic_path(
  ...,
  root = proj_root(),
  relative = TRUE,
  error = TRUE,
  message = getOption("xfun.magic_path.message", TRUE),
  n_dirs = getOption("xfun.magic_path.n_dirs", 10000)
)
}
\arguments{
\item{...}{A character vector of path components.}

\item{root}{The root directory under which to search for the path. If
\code{NULL}, the current working directory is used.}

\item{relative}{Whether to return a relative path.}

\item{error}{Whether to signal an error if the path is not found, or multiple
paths are found.}

\item{message}{Whether to emit a message when multiple paths are found and
\code{error = FALSE}.}

\item{n_dirs}{The number of subdirectories to recursively search. The
recursive search may be time-consuming when there are a large number of
subdirectories under the root directory. If you really want to search for
all subdirectories, you may try \code{n_dirs = Inf}.}
}
\value{
The path found under the root directory, or an error when \code{error = TRUE} and the path is not found (or multiple paths are found).
}
\description{
Given a path, try to find it recursively under a root directory. The input
path can be an incomplete path, e.g., it can be a base filename, and
\code{magic_path()} will try to find this file under subdirectories.
}
\examples{
\dontrun{
xfun::magic_path("mtcars.csv")  # find any file that has the base name mtcars.csv
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R, R/github.R
\name{rest_api}
\alias{rest_api}
\alias{rest_api_raw}
\alias{github_api}
\title{Get data from a REST API}
\usage{
rest_api(...)

rest_api_raw(root, endpoint, token = "", params = list(), headers = NULL)

github_api(
  endpoint,
  token = "",
  params = list(),
  headers = NULL,
  raw = !loadable("jsonlite")
)
}
\arguments{
\item{...}{Arguments to be passed to \code{rest_api_raw()}.}

\item{root}{The API root URL.}

\item{endpoint}{The API endpoint.}

\item{token}{A named character string (e.g., \code{c(token = "xxxx")}), which
will be used to create an authorization header of the form
\samp{Authorization: NAME TOKEN} for the API call, where \samp{NAME} is the
name of the string and \samp{TOKEN} is the string. If the string does not
have a name, \samp{Basic} will be used as the default name.}

\item{params}{A list of query parameters to be sent with the API call.}

\item{headers}{A named character vector of HTTP headers, e.g., \code{c(Accept = "application/vnd.github.v3+json")}.}

\item{raw}{Whether to return the raw response or parse the response with
\pkg{jsonlite}.}
}
\value{
A character vector (the raw JSON response) or an R object parsed from
the JSON text.
}
\description{
Read data from a REST API and optionally with an authorization token in the
request header. The function \code{rest_api_raw()} returns the raw text of
the response, and \code{rest_api()} will parse the response with
\code{jsonlite::fromJSON()} (assuming that the response is in the JSON
format).
}
\details{
These functions are simple wrappers based on \code{\link[=url]{url()}} and
\code{\link[=read_utf8]{read_utf8()}}. Specifically, the \code{headers} argument is
passed to \code{url()}, and \code{read_utf8()} will send a \samp{GET} request
to the API server. This means these functions only support the \samp{GET}
method. If you need to use other HTTP methods (such as \samp{POST}), you have
to use other packages such as \pkg{curl} and \pkg{httr}.

\code{github_api()} is a wrapper function based on
\code{rest_api_raw()} to obtain data from the GitHub API:
\url{https://docs.github.com/en/rest}. You can provide a personal access
token (PAT) via the \code{token} argument, or via one of the environment
variables \var{GITHUB_PAT}, \var{GITHUB_TOKEN}, \var{GH_TOKEN}. A PAT
allows for a much higher rate limit in API calls. Without a token, you can
only make 60 calls in an hour.
}
\examples{\dontshow{if (interactive()) \{}
# a normal GET request
xfun::rest_api("https://mockhttp.org", "/get")
xfun::rest_api_raw("https://mockhttp.org", "/get")

# send the request with an auth header
xfun::rest_api("https://mockhttp.org", "/headers", "OPEN SESAME!")

# with query parameters
xfun::rest_api("https://mockhttp.org", "/response-headers", params = list(foo = "bar"))

# get the rate limit info from GitHub
xfun::github_api("/rate_limit")
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revcheck.R
\name{rev_check}
\alias{rev_check}
\alias{compare_Rcheck}
\title{Run \command{R CMD check} on the reverse dependencies of a package}
\usage{
rev_check(
  pkg,
  which = "all",
  recheck = NULL,
  ignore = NULL,
  update = TRUE,
  timeout = getOption("xfun.rev_check.timeout", 15 * 60),
  src = file.path(src_dir, pkg),
  src_dir = getOption("xfun.rev_check.src_dir")
)

compare_Rcheck(status_only = TRUE, output = "00check_diffs.md")
}
\arguments{
\item{pkg}{The package name.}

\item{which}{Which types of reverse dependencies to check. See
\code{tools::\link[tools]{package_dependencies}()} for possible values. The
special value \code{'hard'} means the hard dependencies, i.e.,
\code{c('Depends', 'Imports', 'LinkingTo')}.}

\item{recheck}{A vector of package names to be (re)checked. If not provided
and there are any \file{*.Rcheck} directories left by certain packages
(this often means these packages failed the last time), \code{recheck} will
be these packages; if there are no \file{*.Rcheck} directories but a text
file \file{recheck} exists, \code{recheck} will be the character vector
read from this file. This provides a way for you to manually specify the
packages to be checked. If there are no packages to be rechecked, all
reverse dependencies will be checked.}

\item{ignore}{A vector of package names to be ignored in \command{R CMD
check}. If this argument is missing and a file \file{00ignore} exists, the
file will be read as a character vector and passed to this argument.}

\item{update}{Whether to update all packages before the check.}

\item{timeout}{Timeout in seconds for \command{R CMD check} to check each
package. The (approximate) total time can be limited by the global option
\code{xfun.rev_check.timeout_total}.}

\item{src}{The path of the source package directory.}

\item{src_dir}{The parent directory of the source package directory. This can
be set in a global option if all your source packages are under a common
parent directory.}

\item{status_only}{If \code{TRUE}, only compare the final statuses of the
checks (the last line of \file{00check.log}), and delete \file{*.Rcheck}
and \file{*.Rcheck2} if the statuses are identical, otherwise write out the
full diffs of the logs. If \code{FALSE}, compare the full logs under
\file{*.Rcheck} and \file{*.Rcheck2}.}

\item{output}{The output Markdown file to which the diffs in check logs will
be written. If the \pkg{markdown} package is available, the Markdown file
will be converted to HTML, so you can see the diffs more clearly.}
}
\value{
A named numeric vector with the names being package names of reverse
  dependencies; \code{0} indicates check success, \code{1} indicates failure,
  and \code{2} indicates that a package was not checked due to global
  timeout.
}
\description{
Install the source package, figure out the reverse dependencies on CRAN,
download all of their source packages, and run \command{R CMD check} on them
in parallel.
}
\details{
Everything occurs under the current working directory, and you are
recommended to call this function under a designated directory, especially
when the number of reverse dependencies is large, because all source packages
will be downloaded to this directory, and all \file{*.Rcheck} directories
will be generated under this directory, too.

If a source tarball of the expected version has been downloaded before (under
the \file{tarball} directory), it will not be downloaded again (to save time
and bandwidth).

After a package has been checked, the associated \file{*.Rcheck} directory
will be deleted if the check was successful (no warnings or errors or notes),
which means if you see a \file{*.Rcheck} directory, it means the check
failed, and you need to take a look at the log files under that directory.

The time to finish the check is recorded for each package. As the check goes
on, the total remaining time will be roughly estimated via \code{n *
mean(times)}, where \code{n} is the number of packages remaining to be
checked, and \code{times} is a vector of elapsed time of packages that have
been checked.

If a check on a reverse dependency failed, its \file{*.Rcheck} directory will
be renamed to \file{*.Rcheck2}, and another check will be run against the
CRAN version of the package unless \code{options(xfun.rev_check.compare =
FALSE)} is set. If the logs of the two checks are the same, it means no new
problems were introduced in the package, and you can probably ignore this
particular reverse dependency. The function \code{compare_Rcheck()} can be
used to create a summary of all the differences in the check logs under
\file{*.Rcheck} and \file{*.Rcheck2}. This will be done automatically if
\code{options(xfun.rev_check.summary = TRUE)} has been set.

A recommended workflow is to use a special directory to run
\code{rev_check()}, set the global \code{\link{options}}
\code{xfun.rev_check.src_dir} and \code{repos} in the R startup (see
\code{?\link{Startup}}) profile file \code{.Rprofile} under this directory,
and (optionally) set \code{R_LIBS_USER} in \file{.Renviron} to use a special
library path (so that your usual library will not be cluttered). Then run
\code{xfun::rev_check(pkg)} once, investigate and fix the problems or (if you
believe it was not your fault) ignore broken packages in the file
\file{00ignore}, and run \code{xfun::rev_check(pkg)} again to recheck the
failed packages. Repeat this process until all \file{*.Rcheck} directories
are gone.

As an example, I set \code{options(repos = c(CRAN =
'https://cran.rstudio.com'), xfun.rev_check.src_dir = '~/Dropbox/repo')} in
\file{.Rprofile}, and \code{R_LIBS_USER=~/R-tmp} in \file{.Renviron}. Then I
can run, for example, \code{xfun::rev_check('knitr')} repeatedly under a
special directory \file{~/Downloads/revcheck}. Reverse dependencies and their
dependencies will be installed to \file{~/R-tmp}, and \pkg{knitr} will be
installed from \file{~/Dropbox/repo/kintr}.
}
\seealso{
\code{devtools::revdep_check()} is more sophisticated, but currently
  has a few major issues that affect me: (1) It always deletes the
  \file{*.Rcheck} directories
  (\url{https://github.com/r-lib/devtools/issues/1395}), which makes it
  difficult to know more information about the failures; (2) It does not
  fully install the source package before checking its reverse dependencies
  (\url{https://github.com/r-lib/devtools/pull/1397}); (3) I feel it is
  fairly difficult to iterate the check (ignore the successful packages and
  only check the failed packages); by comparison, \code{xfun::rev_check()}
  only requires you to run a short command repeatedly (failed packages are
  indicated by the existing \file{*.Rcheck} directories, and automatically
  checked again the next time).

  \code{xfun::rev_check()} borrowed a very nice feature from
  \code{devtools::revdep_check()}: estimating and displaying the remaining
  time. This is particularly useful for packages with huge numbers of reverse
  dependencies.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{pkg_attach}
\alias{pkg_attach}
\alias{pkg_load}
\alias{loadable}
\alias{pkg_available}
\alias{pkg_attach2}
\alias{pkg_load2}
\title{Attach or load packages, and automatically install missing packages if
requested}
\usage{
pkg_attach(
  ...,
  install = FALSE,
  message = getOption("xfun.pkg_attach.message", TRUE)
)

pkg_load(..., error = TRUE, install = FALSE)

loadable(pkg, strict = TRUE, new_session = FALSE)

pkg_available(pkg, version = NULL)

pkg_attach2(...)

pkg_load2(...)
}
\arguments{
\item{...}{Package names (character vectors, and must always be quoted).}

\item{install}{Whether to automatically install packages that are not
available using \code{\link[=install.packages]{install.packages()}}. Besides \code{TRUE} and
\code{FALSE}, the value of this argument can also be a function to install
packages (\code{install = TRUE} is equivalent to \code{install = install.packages}), or a character string \code{"pak"} (equivalent to
\code{install = pak::pkg_install}, which requires the \pkg{pak} package).
You are recommended to set a CRAN mirror in the global option \code{repos}
via \code{\link[=options]{options()}} if you want to automatically install packages.}

\item{message}{Whether to show the package startup messages (if any startup
messages are provided in a package).}

\item{error}{Whether to signal an error when certain packages cannot be loaded.}

\item{pkg}{A single package name.}

\item{strict}{If \code{TRUE}, use \code{\link[=requireNamespace]{requireNamespace()}} to test if
a package is loadable; otherwise only check if the package is in
\code{\link{.packages}}\code{(TRUE)} (this does not really load the package, so it
is less rigorous but on the other hand, it can keep the current R session
clean).}

\item{new_session}{Whether to test if a package is loadable in a new R
session. Note that \code{new_session = TRUE} implies \code{strict = TRUE}.}

\item{version}{A minimal version number. If \code{NULL}, only test if a
package is available and do not check its version.}
}
\value{
\code{pkg_attach()} returns \code{NULL} invisibly. \code{pkg_load()}
returns a logical vector, indicating whether the packages can be loaded.
}
\description{
\code{pkg_attach()} is a vectorized version of \code{\link[=library]{library()}} over
the \code{package} argument to attach multiple packages in a single function
call. \code{pkg_load()} is a vectorized version of
\code{\link[=requireNamespace]{requireNamespace()}} to load packages (without attaching them).
The functions \code{pkg_attach2()} and \code{pkg_load2()} are wrappers of
\code{pkg_attach(install = TRUE)} and \code{pkg_load(install = TRUE)},
respectively. \code{loadable()} is an abbreviation of
\code{requireNamespace(quietly = TRUE)}. \code{pkg_available()} tests if a
package with a minimal version is available.
}
\details{
These are convenience functions that aim to solve these common problems: (1)
We often need to attach or load multiple packages, and it is tedious to type
several \code{library()} calls; (2) We are likely to want to install the
packages when attaching/loading them but they have not been installed.
}
\examples{
library(xfun)
pkg_attach("stats", "graphics")
# pkg_attach2('servr') # automatically install servr if it is not installed

(pkg_load("stats", "graphics"))
}
\seealso{
\code{pkg_attach2()} is similar to \code{pacman::p_load()}, but does
not allow non-standard evaluation (NSE) of the \code{...} argument, i.e.,
you must pass a real character vector of package names to it, and all names
must be quoted. Allowing NSE adds too much complexity with too little gain
(the only gain is that it saves your effort in typing two quotes).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{mark_dirs}
\alias{mark_dirs}
\title{Mark some paths as directories}
\usage{
mark_dirs(x)
}
\arguments{
\item{x}{Character vector of paths to files and directories.}
}
\description{
Add a trailing backlash to a file path if this is a directory. This is useful
in messages to the console for example to quickly identify directories from
files.
}
\details{
If \code{x} is a vector of relative paths, directory test is done with path
relative to the current working dir. Use \code{\link[=in_dir]{in_dir()}} or use absolute
paths.
}
\examples{
mark_dirs(list.files(find.package("xfun"), full.names = TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{broken_packages}
\alias{broken_packages}
\title{Find out broken packages and reinstall them}
\usage{
broken_packages(reinstall = TRUE)
}
\arguments{
\item{reinstall}{Whether to reinstall the broken packages, or only list their
names.}
}
\value{
A character vector of names of broken package.
}
\description{
If a package is broken (i.e., not \code{\link[=loadable]{loadable()}}), reinstall it.
}
\details{
Installed R packages could be broken for several reasons. One common reason
is that you have upgraded R to a newer \code{x.y} version, e.g., from \verb{4.0.5} to
\verb{4.1.0}, in which case you need to reinstall previously installed packages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{dir_create}
\alias{dir_create}
\title{Create a directory recursively by default}
\usage{
dir_create(x, recursive = TRUE, ...)
}
\arguments{
\item{x}{A path name.}

\item{recursive}{Whether to create all directory components in the path.}

\item{...}{Other arguments to be passed to \code{\link{dir.create}()}.}
}
\value{
A logical value indicating if the directory either exists or is
  successfully created.
}
\description{
First check if a directory exists. If it does, return \code{TRUE}, otherwise
create it with \code{\link{dir.create}(recursive = TRUE)} by default.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image.R
\name{upload_imgur}
\alias{upload_imgur}
\title{Upload an image to imgur.com}
\usage{
upload_imgur(
  file,
  key = env_option("xfun.upload_imgur.key", "9f3460e67f308f6"),
  use_curl = loadable("curl"),
  include_xml = FALSE
)
}
\arguments{
\item{file}{Path to the image file to be uploaded.}

\item{key}{Client ID for Imgur. It can be set via either the global option
\code{xfun.upload_imgur.key} or the environment variable
\code{R_XFUN_UPLOAD_IMGUR_KEY} (see \code{\link[=env_option]{env_option()}}). If neither is set,
this uses a client ID registered by Yihui Xie.}

\item{use_curl}{Whether to use the R package \pkg{curl} to upload the image.
If \code{FALSE}, the system command \code{curl} will be used.}

\item{include_xml}{Whether to include the XML response in the returned value.}
}
\value{
A character string of the link to the image. If \code{include_xml = TRUE},
this string carries an attribute named \code{XML}, which is the XML response
from Imgur (it will be parsed by \pkg{xml2} if available). See Imgur API in
the references.
}
\description{
This function uses the \pkg{curl} package or the system command \code{curl}
(whichever is available) to upload a image to \url{https://imgur.com}.
}
\details{
One application is to upload local image files to Imgur when knitting a
document with \pkg{knitr}: you can set the \verb{knitr::opts_knit$set(upload.fun = xfun::upload_imgur}, so the output document does not need local image files
any more, and it is ready to be published online.
}
\note{
Please register your own Imgur application to get your client ID; you
can certainly use mine, but this ID is in the public domain so everyone has
access to all images associated to it.
}
\examples{
\dontrun{
f = tempfile(fileext = ".png")
png(f)
plot(rnorm(100), main = R.version.string)
dev.off()

res = imgur_upload(f, include_xml = TRUE)
res  # link to original URL of the image
attr(res, "XML")  # all information
if (interactive())
    browseURL(res)

# to use your own key
options(xfun.upload_imgur.key = "your imgur key")
}
}
\references{
A demo: \url{https://yihui.org/knitr/demo/upload/}
}
\author{
Yihui Xie, adapted from the \pkg{imguR} package by Aaron Statham
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{gsub_file}
\alias{gsub_file}
\alias{gsub_files}
\alias{gsub_dir}
\alias{gsub_ext}
\title{Search and replace strings in files}
\usage{
gsub_file(file, ..., rw_error = TRUE)

gsub_files(files, ...)

gsub_dir(..., dir = ".", recursive = TRUE, ext = NULL, mimetype = ".*")

gsub_ext(ext, ..., dir = ".", recursive = TRUE)
}
\arguments{
\item{file}{Path of a single file.}

\item{...}{For \code{gsub_file()}, arguments passed to \code{gsub()}. For
other functions, arguments passed to \code{gsub_file()}. Note that the
argument \code{x} of \code{gsub()} is the content of the file.}

\item{rw_error}{Whether to signal an error if the file cannot be read or
written. If \code{FALSE}, the file will be ignored (with a warning).}

\item{files}{A vector of file paths.}

\item{dir}{Path to a directory (all files under this directory will be
replaced).}

\item{recursive}{Whether to find files recursively under a directory.}

\item{ext}{A vector of filename extensions (without the leading periods).}

\item{mimetype}{A regular expression to filter files based on their MIME
types, e.g., \code{'^text/'} for plain text files. This requires the
\pkg{mime} package.}
}
\description{
These functions provide the "file" version of \code{\link{gsub}()}, i.e.,
they perform searching and replacement in files via \code{gsub()}.
}
\note{
These functions perform in-place replacement, i.e., the files will be
  overwritten. Make sure you backup your files in advance, or use version
  control!
}
\examples{
library(xfun)
f = tempfile()
writeLines(c("hello", "world"), f)
gsub_file(f, "world", "woRld", fixed = TRUE)
readLines(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image.R
\name{upload_imgur}
\alias{upload_imgur}
\title{Upload an image to imgur.com}
\usage{
upload_imgur(
  file,
  key = env_option("xfun.upload_imgur.key"),
  use_curl = loadable("curl"),
  include_xml = FALSE
)
}
\arguments{
\item{file}{Path to the image file to be uploaded.}

\item{key}{Client ID for Imgur. It can be set via either the global option
\code{xfun.upload_imgur.key} or the environment variable
\code{R_XFUN_UPLOAD_IMGUR_KEY} (see \code{\link[=env_option]{env_option()}}). If neither is set, this
uses a client ID registered by Yihui Xie.}

\item{use_curl}{Whether to use the R package \pkg{curl} to upload the image.
If \code{FALSE}, the system command \code{curl} will be used.}

\item{include_xml}{Whether to include the XML response in the returned value.}
}
\value{
A character string of the link to the image. If \code{include_xml = TRUE},
this string carries an attribute named \code{XML}, which is the XML response
from Imgur (it will be parsed by \pkg{xml2} if available). See Imgur API in
the references.
}
\description{
This function uses the \pkg{curl} package or the system command \code{curl}
(whichever is available) to upload a image to \url{https://imgur.com}.
}
\details{
One application is to upload local image files to Imgur when knitting a
document with \pkg{knitr}: you can set the \verb{knitr::opts_knit$set(upload.fun = xfun::upload_imgur}, so the output document does not need local image files
any more, and it is ready to be published online.
}
\note{
Please register your own Imgur application to get your client ID. You
can certainly use mine, but this ID is in the public domain and it might
reach Imgur's rate limit if too many people are using it in the same time
period or someone is trying to upload too many images at once.
}
\examples{
\dontrun{
f = tempfile(fileext = ".png")
png(f)
plot(rnorm(100), main = R.version.string)
dev.off()

res = imgur_upload(f, include_xml = TRUE)
res  # link to original URL of the image
xfun::attr2(res, "XML")  # all information
if (interactive())
    browseURL(res)

# to use your own key
options(xfun.upload_imgur.key = "your imgur key")
}
}
\references{
A demo: \url{https://yihui.org/knitr/demo/upload/}
}
\author{
Yihui Xie, adapted from the \pkg{imguR} package by Aaron Statham
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{exit_call}
\alias{exit_call}
\title{Call \code{on.exit()} in a parent function}
\usage{
exit_call(fun, n = 2, ...)
}
\arguments{
\item{fun}{A function to be called when the parent function exits.}

\item{n}{The parent frame number. For \code{n = 1}, \code{exit_call(fun)} is
the same as \code{on.exit(fun())}; \code{n = 2} means adding
\code{on.exit(fun())} in the parent function; \code{n = 3} means the
grandparent, etc.}

\item{...}{Other arguments to be passed to \code{on.exit()}.}
}
\description{
The function \code{\link{on.exit}()} is often used to perform tasks when the
current function exits. This \code{exit_call()} function allows calling a
function when a parent function exits (thinking of it as inserting an
\code{on.exit()} call into the parent function).
}
\examples{
f = function(x) {
    print(x)
    xfun::exit_call(function() print("The parent function is exiting!"))
}
g = function(y) {
    f(y)
    print("f() has been called!")
}
g("An argument of g()!")
}
\references{
This function was inspired by Kevin Ushey:
  \url{https://yihui.org/en/2017/12/on-exit-parent/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{is_web_path}
\alias{is_web_path}
\title{Test if a path is a web path}
\usage{
is_web_path(x)
}
\arguments{
\item{x}{A vector of paths.}
}
\value{
A logical vector.
}
\description{
Check if a path starts with \file{http://} or \file{https://} or
\file{ftp://} or \file{ftps://}.
}
\examples{
xfun::is_web_path("https://www.r-project.org")  # TRUE
xfun::is_web_path("www.r-project.org")  # FALSE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/httpd.R
\name{serve_dir}
\alias{serve_dir}
\title{Serve a static directory}
\usage{
serve_dir(dir = ".", name = "xfun-dir", ...)
}
\arguments{
\item{dir}{The directory to serve.}

\item{name}{The app name passed to \code{\link[=new_app]{new_app()}}. Use \code{''} to serve the
directory at the root URL via the proxy.}

\item{...}{Other arguments passed to \code{\link[=new_app]{new_app()}}.}
}
\value{
The app URL, returned invisibly.
}
\description{
Serve a directory of files via \code{\link[=new_app]{new_app()}}. If \code{index.html} exists under the
directory, it is shown; otherwise a directory listing is shown. This is
similar to \code{servr::httd()} and Python's \code{http.server} module.
}
\examples{\dontshow{if (interactive()) \{}
xfun::serve_dir(tempdir())
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran.R
\name{submit_cran}
\alias{submit_cran}
\title{Submit a source package to CRAN}
\usage{
submit_cran(file = pkg_build(), comment = "")
}
\arguments{
\item{file}{The path to the source package tarball. By default, the current
working directory is treated as the package root directory, and
automatically built into a tarball, which is deleted after submission. This
means you should run \code{xfun::submit_cran()} in the root directory of a
package project, unless you want to pass a path explicitly to the
\code{file} argument.}

\item{comment}{Submission comments for CRAN. By default, if a file
\file{cran-comments.md} exists, its content will be read and used as the
comment.}
}
\description{
Build a source package and submit it to CRAN with the \pkg{curl} package.
}
\seealso{
\code{devtools::submit_cran()} does the same job, with a few more
dependencies in addition to \pkg{curl} (such as \pkg{cli});
\code{xfun::submit_cran()} only depends on \pkg{curl}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/algorithm.R
\name{rand_unit}
\alias{rand_unit}
\title{Pseudo-random numbers on \verb{[0, 1)} based on a linear congruential generator}
\usage{
rand_unit(n = 1, a = 55797, c = 0, m = 4294967296, seed = NULL)
}
\arguments{
\item{n}{The desired length of the sequence.}

\item{a, c, m}{Parameters for the generator (see References for the default
values).}

\item{seed}{The seed. The default is the system time when this function is
called for the first time in the current session. For subsequent calls, the
last \eqn{X_i}{X[i]} of the previous call will be used as the default seed.}
}
\value{
Random numbers on \verb{[0, 1)} (i.e., \eqn{X/m} instead of \eqn{X}). Note
the unit interval is open on the right (excluding 1).
}
\description{
Generate pseudo-random numbers from \eqn{X_{i+1} = (aX_i + c) \bmod{m}}{X\link[=i +
1]{i + 1} = (a * X\link{i} + c) \%\% m}, where \eqn{X_1}{X[1]} is the initial value (seed).
}
\note{
All argument values must be smaller than \eqn{2^{64}}{2^64} as they
will be coerced to 64-bit integers.
}
\examples{
rand_unit(10)
rand_unit(10, seed = 0)
rand_unit(10, seed = 0)  # identical results
rand_unit(10, seed = Sys.getpid())
}
\references{
Steele, Guy L. Jr.; Vigna, Sebastiano (2022). "Computationally
easy, spectrally good multipliers for congruential pseudorandom number
generators". \emph{Software: Practice and Experience.} 52 (2): 443–458.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revcheck.R
\name{rev_check}
\alias{rev_check}
\alias{compare_Rcheck}
\title{Run \command{R CMD check} on the reverse dependencies of a package}
\usage{
rev_check(
  pkg,
  which = "all",
  recheck = NULL,
  ignore = NULL,
  update = TRUE,
  timeout = getOption("xfun.rev_check.timeout", 15 * 60),
  src = file.path(src_dir, pkg),
  src_dir = getOption("xfun.rev_check.src_dir")
)

compare_Rcheck(status_only = TRUE, output = "00check_diffs.md")
}
\arguments{
\item{pkg}{The package name.}

\item{which}{Which types of reverse dependencies to check. See
\code{\link[tools:package_dependencies]{tools::package_dependencies()}} for possible values. The
special value \code{'hard'} means the hard dependencies, i.e.,
\code{c('Depends', 'Imports', 'LinkingTo')}.}

\item{recheck}{A vector of package names to be (re)checked. If not provided
and there are any \file{*.Rcheck} directories left by certain packages
(this often means these packages failed the last time), \code{recheck} will
be these packages; if there are no \file{*.Rcheck} directories but a text
file \file{recheck} exists, \code{recheck} will be the character vector
read from this file. This provides a way for you to manually specify the
packages to be checked. If there are no packages to be rechecked, all
reverse dependencies will be checked.}

\item{ignore}{A vector of package names to be ignored in \command{R CMD
  check}. If this argument is missing and a file \file{00ignore} exists, the
file will be read as a character vector and passed to this argument.}

\item{update}{Whether to update all packages before the check.}

\item{timeout}{Timeout in seconds for \command{R CMD check} to check each
package. The (approximate) total time can be limited by the global option
\code{xfun.rev_check.timeout_total}.}

\item{src}{The path of the source package directory.}

\item{src_dir}{The parent directory of the source package directory. This can
be set in a global option if all your source packages are under a common
parent directory.}

\item{status_only}{If \code{TRUE}, only compare the final statuses of the
checks (the last line of \file{00check.log}), and delete \file{*.Rcheck}
and \file{*.Rcheck2} if the statuses are identical, otherwise write out the
full diffs of the logs. If \code{FALSE}, compare the full logs under
\file{*.Rcheck} and \file{*.Rcheck2}.}

\item{output}{The output Markdown file to which the diffs in check logs will
be written. If the \pkg{markdown} package is available, the Markdown file
will be converted to HTML, so you can see the diffs more clearly.}
}
\value{
A named numeric vector with the names being package names of reverse
dependencies; \code{0} indicates check success, \code{1} indicates failure,
and \code{2} indicates that a package was not checked due to global
timeout.
}
\description{
Install the source package, figure out the reverse dependencies on CRAN,
download all of their source packages, and run \command{R CMD check} on them
in parallel.
}
\details{
Everything occurs under the current working directory, and you are
recommended to call this function under a designated directory, especially
when the number of reverse dependencies is large, because all source packages
will be downloaded to this directory, and all \file{*.Rcheck} directories
will be generated under this directory, too.

If a source tarball of the expected version has been downloaded before (under
the \file{tarball} directory), it will not be downloaded again (to save time
and bandwidth).

After a package has been checked, the associated \file{*.Rcheck} directory
will be deleted if the check was successful (no warnings or errors or notes),
which means if you see a \file{*.Rcheck} directory, it means the check
failed, and you need to take a look at the log files under that directory.

The time to finish the check is recorded for each package. As the check goes
on, the total remaining time will be roughly estimated via \code{n * mean(times)}, where \code{n} is the number of packages remaining to be
checked, and \code{times} is a vector of elapsed time of packages that have
been checked.

If a check on a reverse dependency failed, its \file{*.Rcheck} directory will
be renamed to \file{*.Rcheck2}, and another check will be run against the
CRAN version of the package unless \code{options(xfun.rev_check.compare = FALSE)} is set. If the logs of the two checks are the same, it means no new
problems were introduced in the package, and you can probably ignore this
particular reverse dependency. The function \code{compare_Rcheck()} can be
used to create a summary of all the differences in the check logs under
\file{*.Rcheck} and \file{*.Rcheck2}. This will be done automatically if
\code{options(xfun.rev_check.summary = TRUE)} has been set.

A recommended workflow is to use a special directory to run
\code{rev_check()}, set the global \code{\link[=options]{options()}}
\code{xfun.rev_check.src_dir} and \code{repos} in the R startup (see
\verb{?}\code{\link{Startup}}) profile file \code{.Rprofile} under this directory,
and (optionally) set \code{R_LIBS_USER} in \file{.Renviron} to use a special
library path (so that your usual library will not be cluttered). Then run
\code{xfun::rev_check(pkg)} once, investigate and fix the problems or (if you
believe it was not your fault) ignore broken packages in the file
\file{00ignore}, and run \code{xfun::rev_check(pkg)} again to recheck the
failed packages. Repeat this process until all \file{*.Rcheck} directories
are gone.

As an example, I set \code{options(repos = c(CRAN = 'https://cran.rstudio.com'), xfun.rev_check.src_dir = '~/Dropbox/repo')} in
\file{.Rprofile}, and \verb{R_LIBS_USER=~/R-tmp} in \file{.Renviron}. Then I
can run, for example, \code{xfun::rev_check('knitr')} repeatedly under a
special directory \file{~/Downloads/revcheck}. Reverse dependencies and their
dependencies will be installed to \file{~/R-tmp}, and \pkg{knitr} will be
installed from \file{~/Dropbox/repo/kintr}.
}
\seealso{
\code{devtools::revdep_check()} is more sophisticated, but currently
has a few major issues that affect me: (1) It always deletes the
\file{*.Rcheck} directories
(\url{https://github.com/r-lib/devtools/issues/1395}), which makes it
difficult to know more information about the failures; (2) It does not
fully install the source package before checking its reverse dependencies
(\url{https://github.com/r-lib/devtools/pull/1397}); (3) I feel it is
fairly difficult to iterate the check (ignore the successful packages and
only check the failed packages); by comparison, \code{xfun::rev_check()}
only requires you to run a short command repeatedly (failed packages are
indicated by the existing \file{*.Rcheck} directories, and automatically
checked again the next time).

\code{xfun::rev_check()} borrowed a very nice feature from
\code{devtools::revdep_check()}: estimating and displaying the remaining
time. This is particularly useful for packages with huge numbers of reverse
dependencies.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{isFALSE}
\alias{isFALSE}
\title{Test if an object is identical to \code{FALSE}}
\usage{
isFALSE(x)
}
\arguments{
\item{x}{An R object.}
}
\description{
A simple abbreviation of \code{identical(x, FALSE)}.
}
\examples{
library(xfun)
isFALSE(TRUE)  # false
isFALSE(FALSE)  # true
isFALSE(c(FALSE, FALSE))  # false
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{isFALSE}
\alias{isFALSE}
\title{Test if an object is \code{FALSE}}
\usage{
isFALSE(x)
}
\arguments{
\item{x}{An R object.}
}
\description{
For R versions lower than 3.5.0, this function is a simple abbreviation of
\code{identical(x, FALSE)}. For higher R versions, this function calls
\code{base::isFALSE()}.
}
\note{
This function will be deprecated in the future. We recommend that you
use \code{\link[base:Logic]{base::isFALSE()}} instead. If you have to support R versions lower
than 3.5.0, you may use \code{identical(x, FALSE)}, but please note that it is
not equivalent to \code{base::isFALSE()}.
}
\examples{\dontshow{if (getRversion() < '3.5.0') (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(xfun)
isFALSE(TRUE)  # false
isFALSE(FALSE)  # true
isFALSE(c(FALSE, FALSE))  # false
\dontshow{\}) # examplesIf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/encoding.R
\name{native_encode}
\alias{native_encode}
\title{Try to use the system native encoding to represent a character vector}
\usage{
native_encode(x)
}
\arguments{
\item{x}{A character vector.}
}
\description{
Apply \code{enc2native()} to the character vector, and check if \code{enc2utf8()} can
convert it back without a loss. If it does, return \code{enc2native(x)}, otherwise
return the original vector with a warning.
}
\note{
On platforms that supports UTF-8 as the native encoding
(\code{\link[=l10n_info]{l10n_info()}}\verb{[['UTF-8']]} returns \code{TRUE}), the conversion will be
skipped.
}
\examples{
library(xfun)
s = intToUtf8(c(20320, 22909))
Encoding(s)

s2 = native_encode(s)
Encoding(s2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{url_filename}
\alias{url_filename}
\title{Extract filenames from a URLs}
\usage{
url_filename(x, default = "index.html")
}
\arguments{
\item{x}{A character vector of URLs.}

\item{default}{The default filename when it cannot be determined from the
URL, e.g., when the URL ends with a slash.}
}
\value{
A character vector of filenames at the end of URLs.
}
\description{
Get the base names of URLs via \code{\link[=basename]{basename()}}, and remove the
possible query parameters or hash from the names.
}
\examples{
xfun::url_filename("https://yihui.org/images/logo.png")
xfun::url_filename("https://yihui.org/index.html")
xfun::url_filename("https://yihui.org/index.html?foo=bar")
xfun::url_filename("https://yihui.org/index.html#about")
xfun::url_filename("https://yihui.org")
xfun::url_filename("https://yihui.org/")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{join_words}
\alias{join_words}
\title{Join multiple words into a single string}
\usage{
join_words(
  words,
  sep = ", ",
  and = " and ",
  before = "",
  after = before,
  oxford_comma = TRUE
)
}
\arguments{
\item{words}{A character vector.}

\item{sep}{Separator to be inserted between words.}

\item{and}{Character string to be prepended to the last word.}

\item{before, after}{A character string to be added before/after each word.}

\item{oxford_comma}{Whether to insert the separator between the last two
elements in the list.}
}
\value{
A character string marked by \code{\link[=raw_string]{raw_string()}}.
}
\description{
If \code{words} is of length 2, the first word and second word are joined by the
\code{and} string; if \code{and} is blank, \code{sep} is used. When the length is greater
than 2, \code{sep} is used to separate all words, and the \code{and} string is
prepended to the last word.
}
\examples{
join_words("a")
join_words(c("a", "b"))
join_words(c("a", "b", "c"))
join_words(c("a", "b", "c"), sep = " / ", and = "")
join_words(c("a", "b", "c"), and = "")
join_words(c("a", "b", "c"), before = "\"", after = "\"")
join_words(c("a", "b", "c"), before = "\"", after = "\"", oxford_comma = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{file_ext}
\alias{file_ext}
\alias{sans_ext}
\alias{with_ext}
\title{Manipulate filename extensions}
\usage{
file_ext(x, extra = "")

sans_ext(x, extra = "")

with_ext(x, ext, extra = "")
}
\arguments{
\item{x}{A character of file paths.}

\item{extra}{Extra characters to be allowed in the extensions. By default,
only alphanumeric characters are allowed (and also some special cases in
\sQuote{Details}). If other characters should be allowed, they can be
specified in a character string, e.g., \code{"-+!_#"}.}

\item{ext}{A vector of new extensions. It must be either of length 1, or the
same length as \code{x}.}
}
\value{
A character vector of the same length as \code{x}.
}
\description{
Functions to obtain (\code{file_ext()}), remove (\code{sans_ext()}), and
change (\code{with_ext()}) extensions in filenames.
}
\details{
\code{file_ext()} is similar to \code{tools::\link{file_ext}()}, and
\code{sans_ext()} is similar to \code{tools::\link{file_path_sans_ext}()}.
The main differences are that they treat \code{tar.(gz|bz2|xz)} and
\code{nb.html} as extensions (but functions in the \pkg{tools} package
doesn't allow double extensions by default), and allow characters \code{~}
and \code{#} to be present at the end of a filename.
}
\examples{
library(xfun)
p = c("abc.doc", "def123.tex", "path/to/foo.Rmd", "backup.ppt~", "pkg.tar.xz")
file_ext(p)
sans_ext(p)
with_ext(p, ".txt")
with_ext(p, c(".ppt", ".sty", ".Rnw", "doc", "zip"))
with_ext(p, "html")

# allow for more characters in extensions
p = c("a.c++", "b.c--", "c.e##")
file_ext(p)  # -/+/# not recognized by default
file_ext(p, extra = "-+#")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{dir_create}
\alias{dir_create}
\title{Create a directory recursively by default}
\usage{
dir_create(x, recursive = TRUE, ...)
}
\arguments{
\item{x}{A path name.}

\item{recursive}{Whether to create all directory components in the path.}

\item{...}{Other arguments to be passed to \code{\link[=dir.create]{dir.create()}}.}
}
\value{
A logical value indicating if the directory either exists or is
successfully created.
}
\description{
First check if a directory exists. If it does, return \code{TRUE}, otherwise
create it with \code{\link{dir.create}}\code{(recursive = TRUE)} by default.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session.R
\name{session_info}
\alias{session_info}
\title{An alternative to sessionInfo() to print session information}
\usage{
session_info(packages = NULL, dependencies = TRUE)
}
\arguments{
\item{packages}{A character vector of package names, of which the versions
will be printed. If not specified, it means all loaded and attached
packages in the current R session.}

\item{dependencies}{Whether to print out the versions of the recursive
dependencies of packages.}
}
\value{
A character vector of the session information marked as
\code{\link[=raw_string]{raw_string()}}.
}
\description{
This function tweaks the output of \code{\link[=sessionInfo]{sessionInfo()}}: (1) It adds
the RStudio version information if running in the RStudio IDE; (2) It removes
the information about matrix products, BLAS, and LAPACK; (3) It removes the
names of base R packages; (4) It prints out package versions in a single
group, and does not differentiate between loaded and attached packages.
}
\details{
It also allows you to only print out the versions of specified packages (via
the \code{packages} argument) and optionally their recursive dependencies.
For these specified packages (if provided), if a function
\code{xfun_session_info()} exists in a package, it will be called and
expected to return a character vector to be appended to the output of
\code{session_info()}. This provides a mechanism for other packages to inject
more information into the \code{session_info} output. For example,
\pkg{rmarkdown} (>= 1.20.2) has a function \code{xfun_session_info()} that
returns the version of Pandoc, which can be very useful information for
diagnostics.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
xfun::session_info()
if (xfun::loadable("MASS")) xfun::session_info("MASS")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/base64.R
\name{base64_uri}
\alias{base64_uri}
\title{Generate the Data URI for a file}
\usage{
base64_uri(x, type = mime_type(x))
}
\arguments{
\item{x}{A file path.}

\item{type}{The MIME type of the file, e.g., \code{"image/png"} for a PNG image
file.}
}
\value{
A string of the form \verb{data:<media type>;base64,<data>}.
}
\description{
Encode the file in the base64 encoding, and add the media type. The data URI
can be used to embed data in HTML documents, e.g., in the \code{src} attribute of
the \verb{<img />} tag.
}
\examples{
logo = xfun:::R_logo()
img = xfun::html_tag("img", src = xfun::base64_uri(logo), alt = "R logo")
if (interactive()) xfun::html_view(img)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{gsub_file}
\alias{gsub_file}
\alias{gsub_files}
\alias{gsub_dir}
\alias{gsub_ext}
\title{Search and replace strings in files}
\usage{
gsub_file(file, ..., rw_error = TRUE)

gsub_files(files, ...)

gsub_dir(..., dir = ".", recursive = TRUE, ext = NULL, mimetype = ".*")

gsub_ext(ext, ..., dir = ".", recursive = TRUE)
}
\arguments{
\item{file}{Path of a single file.}

\item{...}{For \code{gsub_file()}, arguments passed to \code{gsub()}. For
other functions, arguments passed to \code{gsub_file()}. Note that the
argument \code{x} of \code{gsub()} is the content of the file.}

\item{rw_error}{Whether to signal an error if the file cannot be read or
written. If \code{FALSE}, the file will be ignored (with a warning).}

\item{files}{A vector of file paths.}

\item{dir}{Path to a directory (all files under this directory will be
replaced).}

\item{recursive}{Whether to find files recursively under a directory.}

\item{ext}{A vector of filename extensions (without the leading periods).}

\item{mimetype}{A regular expression to filter files based on their MIME
types, e.g., \code{'^text/'} for plain text files.}
}
\description{
These functions provide the "file" version of \code{\link[=gsub]{gsub()}}, i.e.,
they perform searching and replacement in files via \code{gsub()}.
}
\note{
These functions perform in-place replacement, i.e., the files will be
overwritten. Make sure you backup your files in advance, or use version
control!
}
\examples{
library(xfun)
f = tempfile()
writeLines(c("hello", "world"), f)
gsub_file(f, "world", "woRld", fixed = TRUE)
readLines(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{read_utf8}
\alias{read_utf8}
\alias{write_utf8}
\alias{append_utf8}
\alias{append_unique}
\title{Read / write files encoded in UTF-8}
\usage{
read_utf8(con, error = FALSE)

write_utf8(text, con, ...)

append_utf8(text, con, sort = TRUE)

append_unique(text, con, sort = function(x) base::sort(unique(x)))
}
\arguments{
\item{con}{A connection or a file path.}

\item{error}{Whether to signal an error when non-UTF8 characters are detected
(if \code{FALSE}, only a warning message is issued).}

\item{text}{A character vector (will be converted to UTF-8 via
\code{\link{enc2utf8}()}).}

\item{...}{Other arguments passed to \code{\link{writeLines}()} (except
\code{useBytes}, which is \code{TRUE} in \code{write_utf8()}).}

\item{sort}{Logical (\code{FALSE} means not to sort the content) or a
function to sort the content; \code{TRUE} is equivalent to
\code{base::sort}.}
}
\description{
Read or write files, assuming they are encoded in UTF-8. \code{read_utf8()}
is roughly \code{readLines(encoding = 'UTF-8')} (a warning will be issued if
non-UTF8 lines are found), and \code{write_utf8()} calls
\code{writeLines(enc2utf8(text), useBytes = TRUE)}.
}
\details{
The function \code{append_utf8()} appends UTF-8 content to a file or
connection based on \code{read_utf8()} and \code{write_utf8()}, and
optionally sort the content. The function \code{append_unique()} appends
unique lines to a file or connection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{exit_call}
\alias{exit_call}
\title{Call \code{on.exit()} in a parent function}
\usage{
exit_call(fun, n = 2, ...)
}
\arguments{
\item{fun}{A function to be called when the parent function exits.}

\item{n}{The parent frame number. For \code{n = 1}, \code{exit_call(fun)} is
the same as \code{on.exit(fun())}; \code{n = 2} means adding
\code{on.exit(fun())} in the parent function; \code{n = 3} means the
grandparent, etc.}

\item{...}{Other arguments to be passed to \code{on.exit()}.}
}
\description{
The function \code{\link[=on.exit]{on.exit()}} is often used to perform tasks when the
current function exits. This \code{exit_call()} function allows calling a
function when a parent function exits (thinking of it as inserting an
\code{on.exit()} call into the parent function).
}
\examples{
f = function(x) {
    print(x)
    xfun::exit_call(function() print("The parent function is exiting!"))
}
g = function(y) {
    f(y)
    print("f() has been called!")
}
g("An argument of g()!")
}
\references{
This function was inspired by Kevin Ushey:
\url{https://yihui.org/en/2017/12/on-exit-parent/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{isFALSE}
\alias{isFALSE}
\title{Test if an object is \code{FALSE}}
\usage{
isFALSE(x)
}
\arguments{
\item{x}{An R object.}
}
\description{
For R versions lower than 3.5.0, this function is a simple abbreviation of
\code{identical(x, FALSE)}. For higher R versions, this function calls
\code{base::isFALSE()}.
}
\note{
This function will be deprecated in the future. We recommend that you
use \code{\link[base:Logic]{base::isFALSE()}} instead. If you have to support R versions lower
than 3.5.0, you may use \code{identical(x, FALSE)}, but please note that it is
not equivalent to \code{base::isFALSE()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{read_utf8}
\alias{read_utf8}
\alias{write_utf8}
\alias{append_utf8}
\alias{append_unique}
\title{Read / write files encoded in UTF-8}
\usage{
read_utf8(con, error = FALSE)

write_utf8(text, con, ...)

append_utf8(text, con, sort = TRUE)

append_unique(text, con, sort = function(x) base::sort(unique(x)))
}
\arguments{
\item{con}{A connection or a file path.}

\item{error}{Whether to signal an error when non-UTF8 characters are detected
(if \code{FALSE}, only a warning message is issued).}

\item{text}{A character vector (will be converted to UTF-8 via
\code{\link[=enc2utf8]{enc2utf8()}}).}

\item{...}{Other arguments passed to \code{\link[=writeLines]{writeLines()}} (except
\code{useBytes}, which is \code{TRUE} in \code{write_utf8()}).}

\item{sort}{Logical (\code{FALSE} means not to sort the content) or a
function to sort the content; \code{TRUE} is equivalent to
\code{base::sort}.}
}
\description{
Read or write files, assuming they are encoded in UTF-8. \code{read_utf8()}
is roughly \code{readLines(encoding = 'UTF-8')} (a warning will be issued if
non-UTF8 lines are found), and \code{write_utf8()} calls
\code{writeLines(enc2utf8(text), useBytes = TRUE)}.
}
\details{
The function \code{append_utf8()} appends UTF-8 content to a file or
connection based on \code{read_utf8()} and \code{write_utf8()}, and
optionally sort the content. The function \code{append_unique()} appends
unique lines to a file or connection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revcheck.R
\name{crandalf_check}
\alias{crandalf_check}
\alias{crandalf_results}
\title{Submit check jobs to crandalf}
\usage{
crandalf_check(pkg, size = 400, jobs = Inf, which = "all")

crandalf_results(pkg, repo = NA, limit = 200, wait = 5 * 60)
}
\arguments{
\item{pkg}{The package name of which the reverse dependencies are to be
checked.}

\item{size}{The number of reverse dependencies to be checked in each job.}

\item{jobs}{The number of jobs to run in GitHub Actions (by default, all jobs
are submitted, but you can choose to submit the first few jobs).}

\item{which}{The type of dependencies (see \code{\link[=rev_check]{rev_check()}}).}

\item{repo}{The crandalf repo on GitHub (of the form \code{user/repo} such as
\code{"yihui/crandalf"}). Usually you do not need to specify it, unless you
are not calling this function inside the crandalf project, because
\command{gh} should be able to figure out the repo automatically.}

\item{limit}{The maximum of records for \command{gh run list} to retrieve.
You only need a larger number if the check results are very early in the
GitHub Action history.}

\item{wait}{Number of seconds to wait if not all jobs have been completed on
GitHub. By default, this function checks the status every 5 minutes until
all jobs are completed. Set \code{wait} to 0 to disable waiting (and throw
an error immediately when any jobs are not completed).}
}
\description{
Check the reverse dependencies of a package using the crandalf service:
\url{https://github.com/yihui/crandalf}. If the number of reverse
dependencies is large, they will be split into batches and pushed to crandalf
one by one.
}
\details{
Due to the time limit of a single job on GitHub Actions (6 hours), you will
have to split the large number of reverse dependencies into batches and check
them sequentially on GitHub (at most 5 jobs in parallel). The function
\code{crandalf_check()} does this automatically when necessary. It requires
the \command{git} command to be available.

The function \code{crandalf_results()} fetches check results from GitHub
after all checks are completed, merge the results, and show a full summary of
check results. It requires \code{gh} (GitHub CLI:
\url{https://cli.github.com/manual/}) to be installed and you also need to
authenticate with your GitHub account beforehand.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\docType{data}
\name{proj_root}
\alias{proj_root}
\alias{root_rules}
\title{Return the (possible) root directory of a project}
\format{
An object of class \code{matrix} (inherits from \code{array}) with 2 rows and 2 columns.
}
\usage{
proj_root(path = "./", rules = root_rules)

root_rules
}
\arguments{
\item{path}{The initial path to start the search. If it is a file path, its
parent directory will be used.}

\item{rules}{A matrix of character strings of two columns: the first column
contains regular expressions to look for filenames that match the patterns,
and the second column contains regular expressions to match the content of
the matched files. The regular expression can be an empty string, meaning
that it will match anything.}
}
\value{
Path to the root directory if found, otherwise \code{NULL}.
}
\description{
Given a path of a file (or dir) in a potential project (e.g., an R package or
an RStudio project), return the path to the project root directory.
}
\details{
The search for the root directory is performed by a series of tests,
currently including looking for a \file{DESCRIPTION} file that contains
\verb{Package: *} (which usually indicates an R package), and a
\file{*.Rproj} file that contains \verb{Version: *} (which usually indicates
an RStudio project). If files with the expected patterns are not found in the
initial directory, the search will be performed recursively in upper-level
directories.
}
\note{
This function was inspired by the \pkg{rprojroot} package, but is much
less sophisticated. It is a rather simple function designed to be used in
some of packages that I maintain, and may not meet the need of general
users until this note is removed in the future (which should be unlikely).
If you are sure that you are working on the types of projects mentioned in
the \sQuote{Details} section, this function may be helpful to you,
otherwise please consider using \pkg{rprojroot} instead.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session.R
\name{session_info}
\alias{session_info}
\title{An alternative to sessionInfo() to print session information}
\usage{
session_info(packages = NULL, dependencies = TRUE)
}
\arguments{
\item{packages}{A character vector of package names, of which the versions
will be printed. If not specified, it means all loaded and attached
packages in the current R session.}

\item{dependencies}{Whether to print out the versions of the recursive
dependencies of packages.}
}
\value{
A character vector of the session information marked as
\code{\link[=raw_string]{raw_string()}}.
}
\description{
This function tweaks the output of \code{\link[=sessionInfo]{sessionInfo()}}: (1) It adds
the RStudio version information if running in the RStudio IDE; (2) It removes
the information about matrix products, BLAS, and LAPACK; (3) It removes the
names of base R packages; (4) It prints out package versions in a single
group, and does not differentiate between loaded and attached packages.
}
\details{
It also allows you to only print out the versions of specified packages (via
the \code{packages} argument) and optionally their recursive dependencies.
For these specified packages (if provided), if a function
\code{xfun_session_info()} exists in a package, it will be called and
expected to return a character vector to be appended to the output of
\code{session_info()}. This provides a mechanism for other packages to inject
more information into the \code{session_info} output. For example,
\pkg{rmarkdown} (>= 1.20.2) has a function \code{xfun_session_info()} that
returns the version of Pandoc, which can be very useful information for
diagnostics.
}
\examples{\dontshow{if (interactive()) \{}
xfun::session_info()
if (xfun::loadable("MASS")) xfun::session_info("MASS")
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R
\name{random_port}
\alias{random_port}
\title{Find a random available TCP port}
\usage{
random_port(port = 4321L, n = 20L, exclude = NULL, error = TRUE)
}
\arguments{
\item{port}{Default port to try first.}

\item{n}{Number of additional random ports to try.}

\item{exclude}{Integer vector of ports to exclude from the search.}

\item{error}{Whether to signal an error (default) or return \code{NULL} when no
port is found.}
}
\value{
An integer port number. When \code{error = TRUE} (default), signals an
error if no port is found; when \code{error = FALSE}, returns \code{NULL}.
}
\description{
Find an available TCP port, starting with \code{port}, then sampling from
3000--8000 (excluding ports known to be blocked by Chrome).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{install_github}
\alias{install_github}
\title{An alias of \code{remotes::install_github()}}
\usage{
install_github(...)
}
\arguments{
\item{...}{Arguments to be passed to
\code{remotes::\link[remotes]{install_github}()}.}
}
\description{
This alias is to make autocomplete faster via \code{xfun::install_github},
because most \code{remotes::install_*} functions are never what I want. I
only use \code{install_github} and it is inconvenient to autocomplete it,
e.g. \code{install_git} always comes before \code{install_github}, but I
never use it. In RStudio, I only need to type \code{xfun::ig} to get
\code{xfun::install_github}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{format_bytes}
\alias{format_bytes}
\title{Format numbers of bytes using a specified unit}
\usage{
format_bytes(x, units = "auto", ...)
}
\arguments{
\item{x}{A numeric vector (each element represents a number of bytes).}

\item{units, ...}{Passed to \code{\link[=format.object_size]{format()}}.}
}
\value{
A character vector.
}
\description{
Call the S3 method \code{format.object_size()} to format numbers of bytes.
}
\examples{
xfun::format_bytes(c(1, 1024, 2000, 1e+06, 2e+08))
xfun::format_bytes(c(1, 1024, 2000, 1e+06, 2e+08), units = "KB")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{attr2}
\alias{attr2}
\alias{attr}
\title{Obtain an attribute of an object without partial matching}
\usage{
attr2(...)

attr(...)
}
\arguments{
\item{...}{Passed to \code{\link[base:attr]{base::attr()}} (without the \code{exact} argument).}
}
\description{
An abbreviation of \code{\link[base:attr]{base::attr}}\code{(exact = TRUE)}.
}
\note{
\code{xfun::attr()} is deprecated. Please use \code{xfun::attr2()} instead.
}
\examples{
z = structure(list(a = 1), foo = 2)
base::attr(z, "f")  # 2
xfun::attr2(z, "f")  # NULL
xfun::attr2(z, "foo")  # 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{is_blank}
\alias{is_blank}
\title{Test if a character vector consists of blank strings}
\usage{
is_blank(x)
}
\arguments{
\item{x}{A character vector.}
}
\value{
\code{TRUE} if the input is blank, or \code{FALSE} otherwise.
}
\description{
Return \code{TRUE} if all elements of a character vector are blank (white
spaces or empty strings).
}
\examples{
xfun::is_blank("")
xfun::is_blank("abc")
xfun::is_blank(c("", "  ", "\n\t"))
xfun::is_blank(c("", " ", "abc"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{read_utf8}
\alias{read_utf8}
\alias{write_utf8}
\alias{append_utf8}
\alias{append_unique}
\title{Read / write files encoded in UTF-8}
\usage{
read_utf8(con, error = FALSE)

write_utf8(text, con, ...)

append_utf8(text, con, sort = TRUE)

append_unique(text, con, sort = function(x) base::sort(unique(x)))
}
\arguments{
\item{con}{A connection or a file path.}

\item{error}{Whether to signal an error when non-UTF8 characters are detected
(if \code{FALSE}, only a warning message is issued).}

\item{text}{A character vector (will be converted to UTF-8 via \code{\link[=enc2utf8]{enc2utf8()}}).}

\item{...}{Other arguments passed to \code{\link[=writeLines]{writeLines()}} (except \code{useBytes}, which
is \code{TRUE} in \code{write_utf8()}).}

\item{sort}{Logical (\code{FALSE} means not to sort the content) or a
function to sort the content; \code{TRUE} is equivalent to
\code{base::sort}.}
}
\value{
\code{read_utf8()} returns a character vector of the file content;
\code{write_utf8()} returns the \code{con} argument (invisibly).
}
\description{
Read or write files, assuming they are encoded in UTF-8. \code{read_utf8()} is
roughly \code{readLines(encoding = 'UTF-8')} (a warning will be issued if non-UTF8
lines are found), and \code{write_utf8()} calls \code{writeLines(enc2utf8(text), useBytes = TRUE)}.
}
\details{
The function \code{append_utf8()} appends UTF-8 content to a file or connection
based on \code{read_utf8()} and \code{write_utf8()}, and optionally sort the content.
The function \code{append_unique()} appends unique lines to a file or connection.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{base_pkgs}
\alias{base_pkgs}
\title{Get base R package names}
\usage{
base_pkgs()
}
\value{
A character vector of base R package names.
}
\description{
Return names of packages from \code{\link[=installed.packages]{installed.packages()}} of which the
priority is \code{"base"}.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
xfun::base_pkgs()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json.R
\name{tojson}
\alias{tojson}
\alias{json_vector}
\title{A simple JSON serializer}
\usage{
tojson(x)

json_vector(x, to_array = FALSE, quote = TRUE)
}
\arguments{
\item{x}{An R object.}

\item{to_array}{Whether to convert a vector to a JSON array (use \code{[]}).}

\item{quote}{Whether to double quote the elements.}
}
\value{
A character string.
}
\description{
A JSON serializer that only works on a limited types of R data (\code{NULL},
lists, logical scalars, character/numeric vectors). A character string of the
class \code{JS_EVAL} is treated as raw JavaScript, so will not be quoted. The
function \code{json_vector()} converts an atomic R vector to JSON.
}
\examples{
library(xfun)
tojson(NULL)
tojson(1:10)
tojson(TRUE)
tojson(FALSE)
cat(tojson(list(a = 1, b = list(c = 1:3, d = "abc"))))
cat(tojson(list(c("a", "b"), 1:5, TRUE)))

# the class JS_EVAL is originally from htmlwidgets::JS()
JS = function(x) structure(x, class = "JS_EVAL")
cat(tojson(list(a = 1:5, b = JS("function() {return true;}"))))
}
\seealso{
The \pkg{jsonlite} package provides a full JSON serializer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{base_pkgs}
\alias{base_pkgs}
\title{Get base R package names}
\usage{
base_pkgs()
}
\value{
A character vector of base R package names.
}
\description{
Return base R package names.
}
\examples{
xfun::base_pkgs()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{find_globals}
\alias{find_globals}
\alias{find_locals}
\title{Find global/local variables in R code}
\usage{
find_globals(code, envir = parent.frame())

find_locals(code)
}
\arguments{
\item{code}{Either a character vector of R source code, or an R expression.}

\item{envir}{The global environment in which global variables are to be
found.}
}
\value{
A character vector of the variable names. If the source code contains
syntax errors, an empty character vector will be returned.
}
\description{
Use \code{\link[codetools:findGlobals]{codetools::findGlobals()}} and \code{\link[codetools:codetools]{codetools::findLocalsList()}} to find
global and local variables in a piece of code. Global variables are defined
outside the code, and local variables are created inside the code.
}
\note{
Due to the flexibility of creating and getting variables in R, these
functions are not guaranteed to find all possible variables in the code
(e.g., when the code is hidden behind \code{eval()}).
}
\examples{
x = 2
xfun::find_globals("y = x + 1")
xfun::find_globals("y = get('x') + 1")  # x is not recognized
xfun::find_globals("y = zzz + 1")  # zzz doesn't exist

xfun::find_locals("y = x + 1")
xfun::find_locals("assign('y', x + 1)")  # it works
xfun::find_locals("assign('y', x + 1, new.env())")  # still smart
xfun::find_locals("eval(parse(text = 'y = x + 1'))")  # no way
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{file_rename}
\alias{file_rename}
\title{Rename files and directories}
\usage{
file_rename(from, to)
}
\arguments{
\item{from, to}{Original and target paths, respectively.}
}
\value{
A logical vector (\code{TRUE} for success and \code{FALSE} for failure).
}
\description{
First try \code{\link[=file.rename]{file.rename()}}. If it fails (e.g., renaming a file from one volume
to another on disk is likely to fail), try \code{\link[=file.copy]{file.copy()}} instead, and clean
up the original files if the copy succeeds.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{Rscript}
\alias{Rscript}
\alias{Rcmd}
\title{Run the commands \command{Rscript} and \command{R CMD}}
\usage{
Rscript(args, ...)

Rcmd(args, ...)
}
\arguments{
\item{args}{A character vector of command-line arguments.}

\item{...}{Other arguments to be passed to \code{\link{system2}()}.}
}
\value{
A value returned by \code{system2()}.
}
\description{
Wrapper functions to run the commands \command{Rscript} and \command{R CMD}.
}
\examples{
library(xfun)
Rscript(c("-e", "1+1"))
Rcmd(c("build", "--help"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{parse_only}
\alias{parse_only}
\title{Parse R code and do not keep the source}
\usage{
parse_only(code)
}
\arguments{
\item{code}{A character vector of the R source code.}
}
\value{
R \code{\link[=expression]{expression()}}s.
}
\description{
An abbreviation of \code{parse(keep.source = FALSE)}.
}
\examples{
library(xfun)
parse_only("1+1")
parse_only(c("y~x", "1:5 # a comment"))
parse_only(character(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{file_ext}
\alias{file_ext}
\alias{sans_ext}
\alias{with_ext}
\title{Manipulate filename extensions}
\usage{
file_ext(x, extra = "")

sans_ext(x, extra = "")

with_ext(x, ext, extra = "")
}
\arguments{
\item{x}{A character of file paths.}

\item{extra}{Extra characters to be allowed in the extensions. By default,
only alphanumeric characters are allowed (and also some special cases in
\sQuote{Details}). If other characters should be allowed, they can be
specified in a character string, e.g., \code{"-+!_#"}.}

\item{ext}{A vector of new extensions. It must be either of length 1, or the
same length as \code{x}.}
}
\value{
A character vector of the same length as \code{x}.
}
\description{
Functions to obtain (\code{file_ext()}), remove (\code{sans_ext()}), and
change (\code{with_ext()}) extensions in filenames.
}
\details{
\code{file_ext()} is similar to \code{\link[tools:fileutils]{tools::file_ext()}}, and
\code{sans_ext()} is similar to \code{\link[tools:fileutils]{tools::file_path_sans_ext()}}.
The main differences are that they treat \code{tar.(gz|bz2|xz)} and
\code{nb.html} as extensions (but functions in the \pkg{tools} package
doesn't allow double extensions by default), and allow characters \code{~}
and \verb{#} to be present at the end of a filename.
}
\examples{
library(xfun)
p = c("abc.doc", "def123.tex", "path/to/foo.Rmd", "backup.ppt~", "pkg.tar.xz")
file_ext(p)
sans_ext(p)
with_ext(p, ".txt")
with_ext(p, c(".ppt", ".sty", ".Rnw", "doc", "zip"))
with_ext(p, "html")

# allow for more characters in extensions
p = c("a.c++", "b.c--", "c.e##")
file_ext(p)  # -/+/# not recognized by default
file_ext(p, extra = "-+#")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{install_github}
\alias{install_github}
\title{An alias of \code{remotes::install_github()}}
\usage{
install_github(...)
}
\arguments{
\item{...}{Arguments to be passed to \code{\link[remotes:install_github]{remotes::install_github()}}.}
}
\description{
This alias is to make autocomplete faster via \code{xfun::install_github}, because
most \verb{remotes::install_*} functions are never what I want. I only use
\code{install_github} and it is inconvenient to autocomplete it, e.g.
\code{install_git} always comes before \code{install_github}, but I never use it. In
RStudio, I only need to type \code{xfun::ig} to get \code{xfun::install_github}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{in_dir}
\alias{in_dir}
\title{Evaluate an expression under a specified working directory}
\usage{
in_dir(dir, expr)
}
\arguments{
\item{dir}{Path to a directory.}

\item{expr}{An R expression.}
}
\description{
Change the working directory, evaluate the expression, and restore the
working directory.
}
\examples{
library(xfun)
in_dir(tempdir(), {
    print(getwd())
    list.files()
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github.R
\name{github_releases}
\alias{github_releases}
\title{Get the tags of GitHub releases of a repository}
\usage{
github_releases(
  repo,
  tag = "",
  pattern = "v[0-9.]+",
  use_jsonlite = loadable("jsonlite")
)
}
\arguments{
\item{repo}{The repository name of the form \code{user/repo}, e.g.,
\code{"yihui/xfun"}.}

\item{tag}{A tag as a character string. If provided, it will be returned if
the tag exists. If \code{tag = "latest"}, the tag of the latest release is
returned.}

\item{pattern}{A regular expression to match the tags.}

\item{use_jsonlite}{Whether to use \pkg{jsonlite} to parse the releases info.}
}
\value{
A character vector of (GIT) tags.
}
\description{
Use the GitHub API (\code{\link[=github_api]{github_api()}}) to obtain the tags of the
releases.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
xfun::github_releases("yihui/xfun")
xfun::github_releases("gohugoio/hugo")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{grep_sub}
\alias{grep_sub}
\title{Perform replacement with \code{gsub()} on elements matched from \code{grep()}}
\usage{
grep_sub(pattern, replacement, x, ...)
}
\arguments{
\item{pattern, replacement, x, ...}{Passed to \code{\link{grep}()} and
\code{gsub()}.}
}
\value{
A character vector.
}
\description{
This function is a shorthand of \code{gsub(pattern, replacement,
grep(pattern, x, value = TRUE))}.
}
\examples{
# find elements that matches 'a[b]+c' and capitalize 'b' with perl regex
xfun::grep_sub("a([b]+)c", "a\\\\U\\\\1c", c("abc", "abbbc", "addc", "123"), perl = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{embed_file}
\alias{embed_file}
\alias{embed_dir}
\alias{embed_files}
\title{Embed a file, multiple files, or directory on an HTML page}
\usage{
embed_file(path, name = basename(path), text = paste("Download", name), ...)

embed_dir(path, name = paste0(normalize_path(path), ".zip"), ...)

embed_files(path, name = with_ext(basename(path[1]), ".zip"), ...)
}
\arguments{
\item{path}{Path to the file(s) or directory.}

\item{name}{The default filename to use when downloading the file. Note that
for \code{embed_dir()}, only the base name (of the zip filename) will be used.}

\item{text}{The text for the hyperlink.}

\item{...}{For \code{embed_file()}, additional arguments to be passed to
\code{\link[=html_tag]{html_tag()}} (e.g., \code{class = 'foo'}). For \code{embed_dir()} and
\code{embed_files()}, arguments passed to \code{embed_file()}.}
}
\value{
An HTML tag \samp{<a>} with the appropriate attributes.
}
\description{
For a file, first encode it into base64 data (a character string). Then
generate a hyperlink of the form \samp{<a href="base64 data"
download="filename">Download filename</a>}. The file can be downloaded when
the link is clicked in modern web browsers. For a directory, it will be
compressed as a zip archive first, and the zip file is passed to
\code{embed_file()}. For multiple files, they are also compressed to a zip file
first.
}
\details{
These functions can be called in R code chunks in R Markdown documents with
HTML output formats. You may embed an arbitrary file or directory in the HTML
output file, so that readers of the HTML page can download it from the
browser. A common use case is to embed data files for readers to download.
}
\note{
Windows users may need to install Rtools to obtain the \command{zip}
command to use \code{embed_dir()} and \code{embed_files()}.

Currently Internet Explorer does not support downloading embedded files
(\url{https://caniuse.com/download}). Chrome has a 2MB limit on the file size.
}
\examples{
logo = xfun:::R_logo()
link = xfun::embed_file(logo, text = "Download R logo")
link
if (interactive()) xfun::html_view(link)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/yaml.R
\name{taml_load}
\alias{taml_load}
\alias{taml_file}
\alias{taml_save}
\title{A simple YAML reader and writer}
\usage{
taml_load(x, envir = parent.frame())

taml_file(path)

taml_save(x, path = NULL, indent = "  ")
}
\arguments{
\item{x}{For \code{taml_load()}, a character vector of the YAML content. For
\code{taml_save()}, a list to be converted to YAML.}

\item{envir}{The environment in which R expressions in YAML are evaluated. To
disable the evaluation, use \code{envir = FALSE}.}

\item{path}{A file path to read from or write to.}

\item{indent}{A character string to indent sub-lists by one level.}
}
\value{
\code{taml_load()} and \code{taml_file()} return a list; if \code{path = NULL},
\code{taml_save()} returns a character vector, otherwise the vector is written
to the file specified by the \code{path}.
}
\description{
TAML is a tiny subset of YAML. See
\url{https://yihui.org/litedown/#sec:yaml-syntax} for its specifications.
}
\examples{
(res = taml_load("a: 1"))
taml_save(res)

(res = taml_load("a: 1\nb: \"foo\"\nc: null"))
taml_save(res)

(res = taml_load("a:\n  b: false\n  c: true\n  d: 1.234\ne: bar"))
taml_save(res)
taml_save(res, indent = "\t")

taml_load("a: !expr paste(1:10, collapse = \", \")")
taml_load("a: [1, 3, 4, 2]")
taml_load("a: [1, \"abc\", 4, 2]")
taml_load("a: [\"foo\", \"bar\"]")
taml_load("a: [true, false, true]")
# the other form of array is not supported
taml_load("a:\n  - b\n  - c")
# and you must use the yaml package
if (loadable("yaml")) yaml_load("a:\n  - b\n  - c")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{numbers_to_words}
\alias{numbers_to_words}
\alias{n2w}
\title{Convert numbers to English words}
\usage{
numbers_to_words(x, cap = FALSE, hyphen = TRUE, and = FALSE)

n2w(x, cap = FALSE, hyphen = TRUE, and = FALSE)
}
\arguments{
\item{x}{A numeric vector. Values should be integers. The absolute values
should be less than \code{1e15}.}

\item{cap}{Whether to capitalize the first letter of the word. This can be
useful when the word is at the beginning of a sentence. Default is
\code{FALSE}.}

\item{hyphen}{Whether to insert hyphen (-) when the number is between 21 and
99 (except 30, 40, etc.).}

\item{and}{Whether to insert \code{and} between hundreds and tens, e.g.,
write 110 as \dQuote{one hundred and ten} if \code{TRUE} instead of
\dQuote{one hundred ten}.}
}
\value{
A character vector.
}
\description{
This can be helpful when writing reports with \pkg{knitr}/\pkg{rmarkdown} if
we want to print numbers as English words in the output. The function
\code{n2w()} is an alias of \code{numbers_to_words()}.
}
\examples{
library(xfun)
n2w(0, cap = TRUE)
n2w(0:121, and = TRUE)
n2w(1e+06)
n2w(1e+11 + 12345678)
n2w(-987654321)
n2w(1e+15 - 1)
}
\author{
Daijiang Li
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{md5}
\alias{md5}
\title{Calculate the MD5 checksums of R objects}
\usage{
md5(...)
}
\arguments{
\item{...}{Any number of R objects.}
}
\value{
A character vector of the checksums of objects passed to \code{md5()}. If
the arguments are named, the results will also be named.
}
\description{
\link[=serialize]{Serialize} an object and calculate the checksum via
\code{\link[tools:md5sum]{tools::md5sum()}}. If \code{tools::md5sum()} does not have the argument \code{bytes},
the object will be first serialized to a temporary file, which will be
deleted after the checksum is calculated, otherwise the raw bytes of the
object will be passed to the \code{bytes} argument directly (which will be
faster than writing to a temporary file).
}
\examples{
x1 = 1
x2 = 1:10
x3 = seq(1, 10)
x4 = iris
x5 = paste
(m = xfun::md5(x1, x2, x3, x4, x5))
stopifnot(m[2] == m[3])  # x2 and x3 should be identical

xfun::md5(x1 = x1, x2 = x2)  # named arguments
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{url_accessible}
\alias{url_accessible}
\title{Test if a URL is accessible}
\usage{
url_accessible(x, use_curl = !capabilities("libcurl"), ...)
}
\arguments{
\item{x}{A URL as a character string.}

\item{use_curl}{Whether to use the \pkg{curl} package or the
\code{curlGetHeaders()} function in base R to send the request to the URL.
By default, \pkg{curl} will be used when base R does not have the
\command{libcurl} capability (which should be rare).}

\item{...}{Arguments to be passed to \code{curlGetHeaders()}.}
}
\value{
\code{TRUE} or \code{FALSE}.
}
\description{
Try to send a \code{HEAD} request to a URL using
\code{\link[=curlGetHeaders]{curlGetHeaders()}} or the \pkg{curl} package, and see if it
returns a successful status code.
}
\examples{
xfun::url_accessible("https://yihui.org")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{from_root}
\alias{from_root}
\title{Get the relative path of a path in a project relative to the current working
directory}
\usage{
from_root(..., root = proj_root(), error = TRUE)
}
\arguments{
\item{...}{A character vector of path components \emph{relative to the root
directory of the project}.}

\item{root}{The root directory of the project.}

\item{error}{Whether to signal an error if the path cannot be converted to a
relative path.}
}
\value{
A relative path, or an error when the project root directory cannot
  be determined or the conversion failed and \code{error = TRUE}.
}
\description{
First compose an absolute path using the project root directory and the
relative path components, i.e., \code{\link{file.path}(root, ...)}. Then
convert it to a relative path with \code{\link{relative_path}()}, which is
relative to the current working directory.
}
\details{
This function was inspired by \code{here::here()}, and the major difference
is that it returns a relative path by default, which is more portable.
}
\examples{
\dontrun{
xfun::from_root("data", "mtcars.csv")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/github.R
\name{github_releases}
\alias{github_releases}
\title{Get the tags of GitHub releases of a repository}
\usage{
github_releases(
  repo,
  tag = "",
  pattern = "v[0-9.]+",
  use_jsonlite = loadable("jsonlite")
)
}
\arguments{
\item{repo}{The repository name of the form \code{user/repo}, e.g.,
\code{"yihui/xfun"}.}

\item{tag}{A tag as a character string. If provided, it will be returned if
the tag exists. If \code{tag = "latest"}, the tag of the latest release is
returned.}

\item{pattern}{A regular expression to match the tags.}

\item{use_jsonlite}{Whether to use \pkg{jsonlite} to parse the releases info.}
}
\value{
A character vector of (GIT) tags.
}
\description{
Use the GitHub API (\code{\link[=github_api]{github_api()}}) to obtain the tags of the
releases.
}
\examples{\dontshow{if (interactive()) \{}
xfun::github_releases("yihui/xfun")
xfun::github_releases("gohugoio/hugo")
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_exec}
\alias{cache_exec}
\title{Cache the execution of an expression in memory or on disk}
\usage{
cache_exec(expr, path = "cache/", id = NULL, ...)
}
\arguments{
\item{expr}{An R expression to be cached.}

\item{path}{The path to save the cache. The special value \code{":memory:"} means
in-memory caching. If it is intended to be a directory path, please make
sure to add a trailing slash.}

\item{id}{A stable and unique string identifier for the expression to be used
to identify a unique copy of cache for the current expression from all
cache files (or in-memory elements). If not provided, an MD5 digest of the
\link{deparse}d expression will be used, which means if the expression does not
change (changes in comments or white spaces do not matter), the \code{id} will
remain the same. This may not be a good default is two identical
expressions are cached under the same \code{path}, because they could overwrite
each other's cache when one expression's cache is invalidated, which may or
may not be what you want. If you do not want that to happen, you need to
manually provide an \code{id}.}

\item{...}{More arguments to control the behavior of caching (see
\sQuote{Details}).}
}
\value{
If the cache is found, the cached value of the expression will be
loaded and returned (other local variables will also be lazy-loaded into
the current environment as a side-effect). If cache does not exist, the
expression is executed and its value is returned.
}
\description{
Caching is based on the assumption that if the input does not change, the
output will not change. After an expression is executed for the first time,
its result will be saved (either in memory or on disk). The next run will be
skipped and the previously saved result will be loaded directly if all
external inputs of the expression remain the same, otherwise the cache will
be invalidated and the expression will be re-executed.
}
\details{
Arguments supported in \code{...} include:
\itemize{
\item \code{vars}: Names of local variables (which are created inside the expression).
By default, local variables are automatically detected from the expression
via \code{\link[codetools:codetools]{codetools::findLocalsList()}}. Locally created variables are cached along
with the value of the expression.
\item \code{hash} and \code{extra}: R objects to be used to determine if cache should be
loaded or invalidated. If (the MD5 hash of) the objects is not changed, the
cache is loaded, otherwise the cache is invalidated and rebuilt. By default,
\code{hash} is a list of values of global variables in the expression (i.e.,
variables created outside the expression). Global variables are automatically
detected by \code{\link[codetools:findGlobals]{codetools::findGlobals()}}. You can provide a vector of names to
override the automatic detection if you want some specific global variables
to affect caching, or the automatic detection is not reliable. You can also
provide additional information via the \code{extra} argument. For example, if the
expression reads an external file \code{foo.csv}, and you want the cache to be
invalidated after the file is modified, you may use \code{extra = file.mtime("foo.csv")}.
\item \code{keep}: By default, only one copy of the cache corresponding to an \code{id}
under \code{path} is kept, and all other copies for this \code{id} is automatically
purged. If \code{TRUE}, all copies of the cache are kept. If \code{FALSE}, all copies
are removed, which means the cache is \emph{always} invalidated, and can be useful
to force re-executing the expression.
\item \code{rw}: A list of functions to read/write the cache files. The list is of the
form \code{list(name = 'xxx', load = function(file) {}, save = function(x, file) {})}. By default, \code{\link[=readRDS]{readRDS()}} and \code{\link[=saveRDS]{saveRDS()}} are used. This argument can
also take a character string to use some built-in read/write methods.
Currently available methods include \code{rds} (the default), \code{raw} (using
\code{\link[=serialize]{serialize()}} and \code{\link[=unserialize]{unserialize()}}), and \code{qs} (using \code{\link[qs:qread]{qs::qread()}} and
\code{\link[qs:qsave]{qs::qsave()}}). The \code{rds} and \code{raw} methods only use base R functions (the
\code{rds} method generates smaller files because it uses compression, but is
often slower than the \code{raw} method, which does not use compression). The \code{qs}
method requires the \pkg{qs} package, which can be much faster than base R
methods and also supports compression.
}
}
\examples{
# the first run takes about 1 second
y1 = xfun::cache_exec({
    x = rnorm(1e+05)
    Sys.sleep(1)
    x
}, path = ":memory:", id = "sim-norm")

# the second run takes almost no time
y2 = xfun::cache_exec({
    # comments won't affect caching
    x = rnorm(1e+05)
    Sys.sleep(1)
    x
}, path = ":memory:", id = "sim-norm")

# y1, y2, and x should be identical
stopifnot(identical(y1, y2), identical(y1, x))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-structure.R
\name{strict_list}
\alias{strict_list}
\alias{as_strict_list}
\alias{$.xfun_strict_list}
\alias{print.xfun_strict_list}
\title{Strict lists}
\usage{
strict_list(...)

as_strict_list(x)

\method{$}{xfun_strict_list}(x, name)

\method{print}{xfun_strict_list}(x, ...)
}
\arguments{
\item{...}{Objects (list elements), possibly named. Ignored in the
\code{print()} method.}

\item{x}{For \code{as_strict_list()}, the object to be coerced to a strict
  list.

  For \code{print()}, a strict list.}

\item{name}{The name (a character string) of the list element.}
}
\value{
Both \code{strict_list()} and \code{as_strict_list()} return a list
  with the class \code{xfun_strict_list}. Whereas \code{as_strict_list()}
  attempts to coerce its argument \code{x} to a list if necessary,
  \code{strict_list()} just wraps its argument \code{...} in a list, i.e., it
  will add another list level regardless if \code{...} already is of type
  list.
}
\description{
A strict list is essentially a normal \code{\link{list}()} but it does not
allow partial matching with \code{$}.
}
\details{
To me, partial matching is often more annoying and surprising than
convenient. It can lead to bugs that are very hard to discover, and I have
been bitten by it many times. When I write \code{x$name}, I always mean
precisely \code{name}. You should use a modern code editor to autocomplete
the \code{name} if it is too long to type, instead of using partial names.
}
\examples{
library(xfun)
(z = strict_list(aaa = "I am aaa", b = 1:5))
z$a  # NULL!
z$aaa  # I am aaa
z$b
z$c = "create a new element"

z2 = unclass(z)  # a normal list
z2$a  # partial matching

z3 = as_strict_list(z2)  # a strict list again
z3$a  # NULL again!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{embed_file}
\alias{embed_file}
\alias{embed_dir}
\alias{embed_files}
\title{Embed a file, multiple files, or directory on an HTML page}
\usage{
embed_file(path, name = basename(path), text = paste("Download", name), ...)

embed_dir(path, name = paste0(normalize_path(path), ".zip"), ...)

embed_files(path, name = with_ext(basename(path[1]), ".zip"), ...)
}
\arguments{
\item{path}{Path to the file(s) or directory.}

\item{name}{The default filename to use when downloading the file. Note that
for \code{embed_dir()}, only the base name (of the zip filename) will be used.}

\item{text}{The text for the hyperlink.}

\item{...}{For \code{embed_file()}, additional arguments to be passed to
\code{\link[=html_tag]{html_tag()}} (e.g., \code{class = 'foo'}). For \code{embed_dir()} and
\code{embed_files()}, arguments passed to \code{embed_file()}.}
}
\value{
An HTML tag \samp{<a>} with the appropriate attributes.
}
\description{
For a file, first encode it into base64 data (a character string). Then
generate a hyperlink of the form \samp{<a href="base64 data"
download="filename">Download filename</a>}. The file can be downloaded when
the link is clicked in modern web browsers. For a directory, it will be
compressed as a zip archive first, and the zip file is passed to
\code{embed_file()}. For multiple files, they are also compressed to a zip file
first.
}
\details{
These functions can be called in R code chunks in R Markdown documents with
HTML output formats. You may embed an arbitrary file or directory in the HTML
output file, so that readers of the HTML page can download it from the
browser. A common use case is to embed data files for readers to download.
}
\note{
Windows users may need to install Rtools to obtain the \command{zip}
command to use \code{embed_dir()} and \code{embed_files()}.

Internet Explorer does not support downloading embedded files. Chrome has a
2MB limit on the file size.
}
\examples{
logo = xfun:::R_logo()
link = xfun::embed_file(logo, text = "Download R logo")
link
if (interactive()) xfun::html_view(link)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{proc_kill}
\alias{proc_kill}
\title{Kill a process and (optionally) all its child processes}
\usage{
proc_kill(pid, recursive = TRUE, ...)
}
\arguments{
\item{pid}{The process ID.}

\item{recursive}{Whether to kill the child processes of the process.}

\item{...}{Arguments to be passed to \code{\link{system2}()} to run the
command to kill the process.}
}
\value{
The status code returned from \code{system2()}.
}
\description{
Run the command \command{taskkill /f /pid} on Windows and \command{kill} on
Unix, respectively, to kill a process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/citation.R
\name{pkg_bib}
\alias{pkg_bib}
\title{Generate BibTeX bibliography databases for R packages}
\usage{
pkg_bib(
  x = .packages(),
  file = "",
  tweak = TRUE,
  width = NULL,
  prefix = getOption("xfun.bib.prefix", "R-"),
  lib.loc = NULL,
  packageURL = TRUE,
  date = c("none", "package", "today")
)
}
\arguments{
\item{x}{Package names. Packages which are not installed are ignored.}

\item{file}{The (\file{.bib}) file to write. By default, or if \code{NULL}, output
is written to the R console.}

\item{tweak}{Whether to fix some known problems in the citations, especially
non-standard format of author names.}

\item{width}{Width of lines in bibliography entries. If \code{NULL}, lines will
not be wrapped.}

\item{prefix}{Prefix string for keys in BibTeX entries; by default, it is
\verb{R-} unless \code{option('xfun.bib.prefix')} has been set to another string.}

\item{lib.loc}{A vector of path names of R libraries.}

\item{packageURL}{Use the \code{URL} field from the \file{DESCRIPTION} file. See
Details below.}

\item{date}{The way to include a date of the form \verb{(retrieved YYYY-mm-dd)} in
the \code{note} field of bib entries. For \code{date = 'none'}, no date is included.
For \code{date = 'package'}, the \code{\link[=packageDate]{packageDate()}} of the package is included. For
\code{date = 'today'}, \code{\link[=Sys.Date]{Sys.Date()}} is included. Alternatively, you can provide
a date directly, e.g., \verb{date = '2025-04-01'.}}
}
\value{
A list containing the citations. Citations are also written to the
\code{file} as a side effect.
}
\description{
Call \code{\link[utils:citation]{utils::citation()}} and \code{\link[utils:toLatex]{utils::toBibtex()}} to create bib entries for R
packages and write them in a file. It can facilitate the auto-generation of
bibliography databases for R packages, and it is easy to regenerate all the
citations after updating R packages.
}
\details{
For a package, the keyword \code{R-pkgname} is used for its bib item, where
\code{pkgname} is the name of the package. Citation entries specified in the
\file{CITATION} file of the package are also included. The main purpose of
this function is to automate the generation of the package citation
information because it often changes (e.g., the author, year, package
version, and so on).

There are at least two different uses for the URL in a reference list. You
might want to tell users where to go for more information. In that case, use
the default \code{packageURL = TRUE}, and the first URL listed in the
\file{DESCRIPTION} file will be used. Be careful: some authors don't put the
most relevant URL first. Alternatively, you might want to identify exactly
which version of the package was used in the document. If it was installed
from CRAN or some other repositories, the version number identifies it, and
\code{packageURL = FALSE} will use the repository URL (as used by
\code{\link[utils:citation]{utils::citation()}}).
}
\note{
Some packages on CRAN do not have standard bib entries, which was once
reported by Michael Friendly at
\url{https://stat.ethz.ch/pipermail/r-devel/2010-November/058977.html}. I
find this a pain, and there are no easy solutions except contacting package
authors to modify their DESCRIPTION files. The argument \code{tweak} has
provided hacks to deal with known packages with non-standard bib entries;
\code{tweak = TRUE} is by no means intended to hide or modify the original
citation information. It is just due to the loose requirements on package
authors for the DESCRIPTION file. On one hand, I apologize if it really
mangles the information about certain packages; on the other, I strongly
recommend package authors to consider the \code{Authors@R} field (see the
manual \emph{Writing R Extensions}) to make it easier for other people to cite R
packages.
}
\examples{
pkg_bib(c("base", "MASS", "xfun"))
pkg_bib("cluster", prefix = "R-pkg-")  # a different prefix
pkg_bib("xfun", date = "package")
}
\author{
Yihui Xie and Michael Friendly
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{is_blank}
\alias{is_blank}
\title{Test if a character vector consists of blank strings}
\usage{
is_blank(x)
}
\arguments{
\item{x}{A character vector.}
}
\value{
\code{TRUE} for blank elements, or \code{FALSE} otherwise.
}
\description{
Return a logical vector indicating if elements of a character vector are
blank (white spaces or empty strings).
}
\examples{
xfun::is_blank("")
xfun::is_blank("abc")
xfun::is_blank(c("", "  ", "\n\t"))
xfun::is_blank(c("", " ", "abc"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/base64.R
\name{mime_type}
\alias{mime_type}
\title{Get the MIME types of files}
\usage{
mime_type(x, use_mime = loadable("mime"), empty = "text/plain")
}
\arguments{
\item{x}{A vector of file paths.}

\item{use_mime}{Whether to use the \pkg{mime} package.}

\item{empty}{The MIME type for files without extensions (e.g., \code{Makefile}).
If \code{NA}, the type will be obtained from system command. This argument is
used only for \code{use_mime = FALSE}.}
}
\value{
A character vector of MIME types.
}
\description{
If the \pkg{mime} package is installed, call \code{\link[mime:guess_type]{mime::guess_type()}}, otherwise
use the system command \code{file --mime-type} to obtain the MIME type of a file.
Typically, the \code{file} command exists on *nix. On Windows, the command should
exist if Cygwin or Rtools is installed. If it is not found, .NET's
\code{MimeMapping} class will be used instead (which requires the .NET framework).
}
\note{
When querying the MIME type via the system command, the result will be
cached to \code{xfun:::cache_dir()}. This will make future queries much faster,
since running the command in real time can be a little slow.
}
\examples{\dontshow{if (tolower(Sys.getenv('CI')) == 'true') (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
f = list.files(R.home("doc"), full.names = TRUE)
mime_type(f)
mime_type(f, FALSE)  # don't use mime
mime_type(f, FALSE, NA)  # run command for files without extension
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{Rscript_call}
\alias{Rscript_call}
\title{Call a function in a new R session via \code{Rscript()}}
\usage{
Rscript_call(
  fun,
  args = list(),
  options = NULL,
  ...,
  wait = TRUE,
  fail = sprintf("Failed to run '\%s' in a new R session.", deparse(substitute(fun))[1])
)
}
\arguments{
\item{fun}{A function, or a character string that can be parsed and evaluated
to a function.}

\item{args}{A list of argument values.}

\item{options}{A character vector of options to passed to
\code{\link{Rscript}}, e.g., \code{"--vanilla"}.}

\item{..., wait}{Arguments to be passed to \code{\link{system2}()}.}

\item{fail}{The desired error message when an error occurred in calling the
function.}
}
\value{
The returned value of the function in the new R session.
}
\description{
Save the argument values of a function in a temporary RDS file, open a new R
session via \code{\link{Rscript}()}, read the argument values, call the
function, and read the returned value back to the current R session.
}
\examples{
factorial(10)
# should return the same value
xfun::Rscript_call("factorial", list(10))

# the first argument can be either a character string or a function
xfun::Rscript_call(factorial, list(10))

# Run Rscript starting a vanilla R session
xfun::Rscript_call(factorial, list(10), options = c("--vanilla"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{upload_ftp}
\alias{upload_ftp}
\alias{upload_win_builder}
\title{Upload to an FTP server via \command{curl}}
\usage{
upload_ftp(file, server, dir = "")

upload_win_builder(
  file = pkg_build(),
  version = c("R-devel", "R-release", "R-oldrelease"),
  server = c("ftp", "https"),
  solaris = pkg_available("rhub")
)
}
\arguments{
\item{file}{Path to a local file.}

\item{server}{The address of the FTP server. For \code{upload_win_builder()},
\code{server = 'https'} means uploading to
\code{'https://win-builder.r-project.org/upload.aspx'}.}

\item{dir}{The remote directory to which the file should be uploaded.}

\item{version}{The R version(s) on win-builder.}

\item{solaris}{Whether to also upload the package to the Rhub server to check
it on Solaris.}
}
\value{
Status code returned from \code{\link{system2}()} or
  \code{curl::curl_fetch_memory()}.
}
\description{
The function \code{upload_ftp()} runs the command \command{curl -T file
server} to upload a file to an FTP server if the system command
\command{curl} is available, otherwise it uses the R package \pkg{curl}. The
function \code{upload_win_builder()} uses \code{upload_ftp()} to upload
packages to the win-builder server.
}
\details{
These functions were written mainly to save package developers the trouble of
going to the win-builder web page and uploading packages there manually.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{html_tag}
\alias{html_tag}
\alias{html_value}
\alias{html_escape}
\alias{html_view}
\title{Tools for HTML tags}
\usage{
html_tag(.name, .content = NULL, .attrs = NULL, ...)

html_value(x)

html_escape(x, attr = FALSE)

html_view(x, ...)
}
\arguments{
\item{.name}{The tag name.}

\item{.content}{The content between opening and closing tags. Ignored for
void tags such as \verb{<img>}. Special characters such as \code{&}, \code{<}, and \code{>}
will be escaped unless the value was generated from \code{\link[=html_value]{html_value()}}. The
content can be either a character vector or a list. If it is a list, it may
contain both normal text and HTML content.}

\item{.attrs}{A named list of attributes.}

\item{...}{For \code{html_tag()}, named arguments as an alternative way to provide
attributes. For \code{html_view()}, other arguments to be passed to \code{\link[=new_app]{new_app()}}.}

\item{x}{A character vector to be treated as raw HTML content for
\code{html_value()}, escaped for \code{html_escape()}, and viewed for \code{html_view()}.}

\item{attr}{Whether to escape \verb{\\r} and \verb{\\n} (which should be escaped for tag
attributes).}
}
\value{
A character string.
}
\description{
Given a tag name, generate an HTML tag with optional attributes and content.
\code{\link[=html_tag]{html_tag()}} can be viewed as a simplified version of \code{htmltools::tags},
\code{\link[=html_value]{html_value()}} adds classes on the value so that it will be treated as raw
HTML (not escaped by \code{html_tag()}), \code{\link[=html_escape]{html_escape()}} escapes special
characters in HTML, and \code{html_view()} launches a browser or viewer to view
the HTML content.
}
\examples{
xfun::html_tag("a", "<R Project>", href = "https://www.r-project.org", target = "_blank")
xfun::html_tag("br")
xfun::html_tag("a", xfun::html_tag("strong", "R Project"), href = "#")
xfun::html_tag("a", list("<text>", xfun::html_tag("b", "R Project")), href = "#")
xfun::html_escape("\" quotes \" & brackets < >")
xfun::html_escape("\" & < > \r \n", attr = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json.R
\name{tojson}
\alias{tojson}
\alias{json_vector}
\title{A simple JSON serializer}
\usage{
tojson(x)

json_vector(x, to_array = FALSE, quote = TRUE)
}
\arguments{
\item{x}{An R object.}

\item{to_array}{Whether to convert a vector to a JSON array (use \verb{[]}).}

\item{quote}{Whether to double quote the elements.}
}
\value{
A character string.
}
\description{
A JSON serializer that only works on a limited types of R data (\code{NULL},
lists, arrays, logical/character/numeric/date/time vectors). Other types of
data will be coerced to character. A character string of the class
\code{JS_LITERAL} is treated as raw JavaScript, so will not be quoted. The
function \code{json_vector()} converts an atomic R vector to JSON.
}
\details{
Both \code{NULL} and \code{NA} are converted to \code{null}. Named lists are converted to
objects of the form \verb{\{key1: value1, key2: value2, ...\}}. Unnamed lists are
converted to arrays of the form \verb{[[value1], [value2], ...]}. The same rules
apply to data frames since technically they are also lists. However, please
note that unnamed data frames (i.e., without column names) will be converted
to an array with each \emph{row} as an array element, whereas named data frames
will have each \emph{column} as an individual element. For matrices, the JSON
array will have each row as an individual element, and names are discarded.

Dates and times are coerced to character using UTC as the timezone, and
represented via the JavaScript expression \verb{new Date(value)} (which is not
standard JSON but practically more useful).
}
\examples{
library(xfun)
tojson(NULL)
tojson(1:10)
tojson(TRUE)
tojson(FALSE)
tojson(list(a = 1, b = list(c = 1:3, d = "abc")))
tojson(list(c("a", "b"), 1:5, TRUE, Sys.Date() + 1:3))
tojson(head(iris))  # each column is in an element
tojson(unname(head(iris)))  # each row is in an element
tojson(matrix(1:12, 3))

# literal JS code
JS = function(x) structure(x, class = "JS_LITERAL")
tojson(list(a = 1:5, b = JS("function() {return true;}")))
}
\seealso{
The \pkg{jsonlite} package provides a full JSON serializer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr.R
\name{csv_options}
\alias{csv_options}
\title{Parse comma-separated chunk options}
\usage{
csv_options(x)
}
\arguments{
\item{x}{The chunk options as a string.}
}
\value{
A list of chunk options.
}
\description{
For \pkg{knitr} and R Markdown documents, code chunk options can be written
using the comma-separated syntax (e.g., \verb{opt1=value1, opt2=value2}). This
function parses these options and returns a list. If an option is not named,
it will be treated as the chunk label.
}
\examples{
xfun::csv_options("foo, eval=TRUE, fig.width=5, echo=if (TRUE) FALSE")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{news2md}
\alias{news2md}
\title{Convert package news to the Markdown format}
\usage{
news2md(package, ..., output = "NEWS.md", category = TRUE)
}
\arguments{
\item{package, ...}{Arguments to be passed to \code{\link{news}()}.}

\item{output}{The output file path.}

\item{category}{Whether to keep the category names.}
}
\value{
If \code{output = NA}, returns the Markdown content as a character
  vector, otherwise the content is written to the output file.
}
\description{
Read the package news with \code{\link{news}()}, convert the result to
Markdown, and write to an output file (e.g., \file{NEWS.md}). Each package
version appears in a first-level header, each category (e.g., \samp{NEW
FEATURES} or \samp{BUG FIXES}) is in a second-level header, and the news
items are written into bullet lists.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# news for the current version of R
xfun::news2md("R", Version == getRversion(), output = NA)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{valid_syntax}
\alias{valid_syntax}
\title{Check if the syntax of the code is valid}
\usage{
valid_syntax(code, silent = TRUE)
}
\arguments{
\item{code}{A character vector of R source code.}

\item{silent}{Whether to suppress the error message when the code is not
valid.}
}
\value{
\code{TRUE} if the code could be parsed, otherwise \code{FALSE}.
}
\description{
Try to \code{\link[=parse]{parse()}} the code and see if an error occurs.
}
\examples{
xfun::valid_syntax("1+1")
xfun::valid_syntax("1+")
xfun::valid_syntax(c("if(T){1+1}", "else {2+2}"), silent = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{Rscript}
\alias{Rscript}
\alias{Rcmd}
\title{Run the commands \command{Rscript} and \command{R CMD}}
\usage{
Rscript(args, ...)

Rcmd(args, ...)
}
\arguments{
\item{args}{A character vector of command-line arguments.}

\item{...}{Other arguments to be passed to \code{\link[=system2]{system2()}}.}
}
\value{
A value returned by \code{system2()}.
}
\description{
Wrapper functions to run the commands \command{Rscript} and \command{R CMD}.
}
\examples{
library(xfun)
Rscript(c("-e", "1+1"))
Rcmd(c("build", "--help"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_exec}
\alias{cache_exec}
\title{Cache the execution of an expression in memory or on disk}
\usage{
cache_exec(expr, path = "cache/", id = NULL, ...)
}
\arguments{
\item{expr}{An R expression to be cached.}

\item{path}{The path to save the cache. The special value \code{":memory:"} means
in-memory caching. If it is intended to be a directory path, please make
sure to add a trailing slash.}

\item{id}{A stable and unique string identifier for the expression to be used
to identify a unique copy of cache for the current expression from all
cache files (or in-memory elements). If not provided, an MD5 digest of the
\link{deparse}d expression will be used, which means if the expression does not
change (changes in comments or white spaces do not matter), the \code{id} will
remain the same. This may not be a good default is two identical
expressions are cached under the same \code{path}, because they could overwrite
each other's cache when one expression's cache is invalidated, which may or
may not be what you want. If you do not want that to happen, you need to
manually provide an \code{id}.}

\item{...}{More arguments to control the behavior of caching (see
\sQuote{Details}).}
}
\value{
If the cache is found, the cached value of the expression will be
loaded and returned (other local variables will also be lazy-loaded into
the current environment as a side-effect). If cache does not exist, the
expression is executed and its value is returned.
}
\description{
Caching is based on the assumption that if the input does not change, the
output will not change. After an expression is executed for the first time,
its result will be saved (either in memory or on disk). The next run will be
skipped and the previously saved result will be loaded directly if all
external inputs of the expression remain the same, otherwise the cache will
be invalidated and the expression will be re-executed.
}
\details{
Arguments supported in \code{...} include:
\itemize{
\item \code{vars}: Names of local variables (which are created inside the expression).
By default, local variables are automatically detected from the expression
via \code{\link[=find_locals]{find_locals()}}. Locally created variables are cached along with the
value of the expression.
\item \code{hash} and \code{extra}: R objects to be used to determine if cache should be
loaded or invalidated. If (the MD5 hash of) the objects is not changed, the
cache is loaded, otherwise the cache is invalidated and rebuilt. By default,
\code{hash} is a list of values of global variables in the expression (i.e.,
variables created outside the expression). Global variables are automatically
detected by \code{\link[=find_globals]{find_globals()}}. You can provide a vector of names to override
the automatic detection if you want some specific global variables to affect
caching, or the automatic detection is not reliable. You can also provide
additional information via the \code{extra} argument. For example, if the
expression reads an external file \code{foo.csv}, and you want the cache to be
invalidated after the file is modified, you may use \code{extra = file.mtime("foo.csv")}.
\item \code{keep}: By default, only one copy of the cache corresponding to an \code{id}
under \code{path} is kept, and all other copies for this \code{id} is automatically
purged. If \code{TRUE}, all copies of the cache are kept. If \code{FALSE}, all copies
are removed, which means the cache is \emph{always} invalidated, and can be useful
to force re-executing the expression.
\item \code{rw}: A list of functions to read/write the cache files. The list is of the
form \code{list(name = 'xxx', load = function(file) {}, save = function(x, file) {})}. By default, \code{\link[=readRDS]{readRDS()}} and \code{\link[=saveRDS]{saveRDS()}} are used. This argument can
also take a character string to use some built-in read/write methods.
Currently available methods include \code{rds} (the default), \code{raw} (using
\code{\link[=serialize]{serialize()}} and \code{\link[=unserialize]{unserialize()}}), \code{qs2} (using \code{\link[qs2:qs_read]{qs2::qs_read()}} and
\code{\link[qs2:qs_save]{qs2::qs_save()}}), and \code{qdata} (using \code{\link[qs2:qd_read]{qs2::qd_read()}} and \code{\link[qs2:qd_save]{qs2::qd_save()}}).
The \code{rds} and \code{raw} methods only use base R functions (the \code{rds} method
generates smaller files because it uses compression, but is often slower than
the \code{raw} method, which does not use compression). The \code{qs2} and \code{qdata}
methods require the \pkg{qs2} package, which can be much faster than base R
methods and also supports compression.
}
}
\examples{
# the first run takes about 1 second
y1 = xfun::cache_exec({
    x = rnorm(1e+05)
    Sys.sleep(1)
    x
}, path = ":memory:", id = "sim-norm")

# the second run takes almost no time
y2 = xfun::cache_exec({
    # comments won't affect caching
    x = rnorm(1e+05)
    Sys.sleep(1)
    x
}, path = ":memory:", id = "sim-norm")

# y1, y2, and x should be identical
stopifnot(identical(y1, y2), identical(y1, x))
}
\references{
See \url{https://yihui.org/litedown/#sec:option-cache} for how it
works and an application in \pkg{litedown}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\docType{data}
\name{download_cache}
\alias{download_cache}
\title{Download a file from a URL and cache it on disk}
\format{
A list of methods:
\itemize{
\item \verb{$get(url, type, handler)} downloads a URL, caches it, and returns the file
content according to the value of \code{type} (possible values: \code{"text"} means
the text content; \code{"base64"} means the base64 encoded data; \code{"raw"} means
the raw binary content; \code{"auto"} is the default and means the type is
determined by the content type in the URL headers). Optionally a \code{handler}
function can be applied to the content.
\item \verb{$list()} gives the list of cache files.
\item \verb{$summary()} gives a summary of existing cache files.
\item \verb{$remove(url, type)} removes a single cache file.
\item \verb{$purge()} deletes all cache files.
}
}
\usage{
download_cache
}
\description{
This object provides methods to download files and cache them on disk.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# the first time it may take a few seconds
x1 = xfun::download_cache$get("https://www.r-project.org/")
head(x1)

# now you can get the cached content
x2 = xfun::download_cache$get("https://www.r-project.org/")
identical(x1, x2)  # TRUE

# a binary file
x3 = xfun::download_cache$get("https://yihui.org/images/logo.png", "raw")
length(x3)

# show a summary
xfun::download_cache$summary()
# remove a specific cache file
xfun::download_cache$remove("https://yihui.org/images/logo.png", "raw")
# remove all cache files
xfun::download_cache$purge()
\dontshow{\}) # examplesIf}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image.R
\name{tinify}
\alias{tinify}
\alias{tinify_dir}
\title{Use the Tinify API to compress PNG and JPEG images}
\usage{
tinify(
  input,
  output,
  quiet = FALSE,
  force = FALSE,
  key = env_option("xfun.tinify.key"),
  history = env_option("xfun.tinify.history")
)

tinify_dir(dir = ".", ...)
}
\arguments{
\item{input}{A vector of input paths of images.}

\item{output}{A vector of output paths or a function that takes \code{input} and
returns a vector of output paths (e.g., \verb{output = }\code{\link{identity}} means
\code{output = input}). By default, if the \code{history} argument is not a provided,
\code{output} is \code{input} with a suffix \code{-min} (e.g., when \code{input = 'foo.png'},
\code{output = 'foo-min.png'}), otherwise \code{output} is the same as \code{input}, which
means the original image files will be overwritten.}

\item{quiet}{Whether to suppress detailed information about the compression,
which is of the form \samp{input.png (10 Kb) ==> output.png (5 Kb, 50\%);
  compression count: 42}. The percentage after \code{output.png} stands for
the compression ratio, and the compression count shows the number of
compressions used for the current month.}

\item{force}{Whether to compress an image again when it appears to have been
compressed before. This argument only makes sense when the \code{history}
argument is provided.}

\item{key}{The Tinify API key. It can be set via either the global option
\code{xfun.tinify.key} or the environment variable \code{R_XFUN_TINIFY_KEY} (see
\code{\link[=env_option]{env_option()}}).}

\item{history}{Path to a history file to record the MD5 checksum of
compressed images. If the checksum of an expected output image exists in
this file and \code{force = FALSE}, the compression will be skipped. This can
help you avoid unnecessary API calls.}

\item{dir}{A directory under which all \file{.png}, \file{.jpeg}, and
\file{.webp} files are to be compressed.}

\item{...}{Arguments passed to \code{\link[=tinify]{tinify()}}.}
}
\value{
The output file paths.
}
\description{
Compress PNG/JPEG images with \samp{api.tinify.com}, and download the
compressed images. These functions require R packages \pkg{curl} and
\pkg{jsonlite}. \code{tinify_dir()} is a wrapper function of \code{tinify()} to
compress images under a directory.
}
\details{
You are recommended to set the API key in \file{.Rprofile} or
\file{.Renviron}. After that, the only required argument of this function is
\code{input}. If the original images can be overwritten by the compressed images,
you may either use \code{output = identity}, or set the value of the \code{history}
argument in \file{.Rprofile} or \file{.Renviron}.
}
\examples{\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
f = xfun:::R_logo("jpg$")
xfun::tinify(f)  # remember to set the API key before trying this
\dontshow{\}) # examplesIf}
}
\references{
Tinify API: \url{https://tinypng.com/developers}.
}
\seealso{
The \pkg{tinieR} package (\url{https://github.com/jmablog/tinieR/}) is a
more comprehensive implementation of the Tinify API, whereas
\code{xfun::tinify()} has only implemented the feature of shrinking images.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{grep_sub}
\alias{grep_sub}
\title{Perform replacement with \code{gsub()} on elements matched from \code{grep()}}
\usage{
grep_sub(pattern, replacement, x, ...)
}
\arguments{
\item{pattern, replacement, x, ...}{Passed to \code{\link[=grep]{grep()}} and
\code{gsub()}.}
}
\value{
A character vector.
}
\description{
This function is a shorthand of \code{gsub(pattern, replacement, grep(pattern, x, value = TRUE))}.
}
\examples{
# find elements that matches 'a[b]+c' and capitalize 'b' with perl regex
xfun::grep_sub("a([b]+)c", "a\\\\U\\\\1c", c("abc", "abbbc", "addc", "123"), perl = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{prose_index}
\alias{prose_index}
\title{Find the indices of lines in Markdown that are prose (not code blocks)}
\usage{
prose_index(x, warn = TRUE)
}
\arguments{
\item{x}{A character vector of text in Markdown.}

\item{warn}{Whether to emit a warning when code fences are not balanced.}
}
\value{
An integer vector of indices of lines that are prose in Markdown.
}
\description{
Filter out the indices of lines between code block fences such as \verb{```}
(could be three or four or more backticks).
}
\note{
If the code fences are not balanced (e.g., a starting fence without an
  ending fence), this function will treat all lines as prose.
}
\examples{
library(xfun)
prose_index(c("a", "```", "b", "```", "c"))
prose_index(c("a", "````", "```r", "1+1", "```", "````", "c"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{stringsAsStrings}
\alias{stringsAsStrings}
\alias{strings_please}
\title{Set the global option \code{\link{options}(stringsAsFactors = FALSE)} inside
a parent function and restore the option after the parent function exits}
\usage{
stringsAsStrings()

strings_please()
}
\description{
This is a shorthand of \code{opts = options(stringsAsFactors = FALSE);
on.exit(options(opts), add = TRUE)}; \code{strings_please()} is an alias of
\code{stringsAsStrings()}.
}
\examples{
f = function() {
    xfun::strings_please()
    data.frame(x = letters[1:4], y = factor(letters[1:4]))
}
str(f())  # the first column should be character
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/record.R
\name{record}
\alias{record}
\alias{format.xfun_record_results}
\alias{print.xfun_record_results}
\title{Run R code and record the results}
\usage{
record(
  code = NULL,
  dev = "png",
  dev.path = "xfun-record",
  dev.ext = dev_ext(dev),
  dev.args = list(),
  message = TRUE,
  warning = TRUE,
  error = NA,
  cache = list(),
  print = record_print,
  print.args = list(),
  verbose = getOption("xfun.record.verbose", 0),
  envir = parent.frame()
)

\method{format}{xfun_record_results}(x, to = c("text", "html"), encode = FALSE, template = FALSE, ...)

\method{print}{xfun_record_results}(
  x,
  browse = interactive(),
  to = if (browse) "html" else "text",
  template = TRUE,
  ...
)
}
\arguments{
\item{code}{A character vector of R source code.}

\item{dev}{A graphics device. It can be a function name, a function, or a
character string that can be evaluated to a function to open a graphics
device.}

\item{dev.path}{A base file path for plots. Actual plot filenames will be
this base path plus incremental suffixes. For example, if \code{dev.path = "foo"}, the plot files will be \verb{foo-1.png}, \verb{foo-2.png}, and so on. If
\code{dev.path} is not character (e.g., \code{FALSE}), plots will not be recorded.}

\item{dev.ext}{The file extension for plot files. By default, it will be
inferred from the first argument of the device function if possible.}

\item{dev.args}{Extra arguments to be passed to the device. The default
arguments are \code{list(units = 'in', onefile = FALSE, width = 7, height = 7, res = 96)}. If any of these arguments is not present in the device
function, it will be dropped.}

\item{message, warning, error}{If \code{TRUE}, record and store messages / warnings
/ errors in the output. If \code{FALSE}, suppress them. If \code{NA}, do not process
them (messages will be emitted to the console, and errors will halt the
execution).}

\item{cache}{A list of options for caching. See the \code{path}, \code{id}, and \code{...}
arguments of \code{\link[=cache_exec]{cache_exec()}}.}

\item{print}{A (typically S3) function that takes the value of an expression
in the code as input and returns output. The default is \code{\link[=record_print]{record_print()}}.}

\item{print.args}{A list of arguments for the \code{print} function. By default,
the whole list is not passed directly to the function, but only an element
in the list with a name identical to the first class name of the returned
value of the expression, e.g., \code{list(data.frame = list(digits = 3), matrix = list())}. This makes it possible to apply different print arguments to
objects of different classes. If the whole list is intended to be passed to
the print function directly, wrap the list in \code{\link[=I]{I()}}.}

\item{verbose}{\code{2} means to always print the value of each expression in the
code, no matter if the value is \code{\link[=invisible]{invisible()}} or not; \code{1} means to always
print the value of the last expression; \code{0} means no special handling
(i.e., print only when the value is visible).}

\item{envir}{An environment in which the code is evaluated.}

\item{x}{An object returned by \code{record()}.}

\item{to}{The output format (text or html).}

\item{encode}{For HTML output, whether to base64 encode plots.}

\item{template}{For HTML output, whether to embed the formatted results in an
HTML template. Alternatively, this argument can take a file path, i.e.,
path to an HTML template that contains the variable \verb{$body$}. If \code{TRUE},
the default template in this package will be used
(\code{xfun:::pkg_file('resources', 'record.html')}).}

\item{...}{Currently ignored.}

\item{browse}{Whether to browse the results on an HTML page.}
}
\value{
\code{record()} returns a list of the class \code{xfun_record_results} that
contains elements with these possible classes: \code{record_source} (source
code), \code{record_output} (text output), \code{record_plot} (plot file paths),
\code{record_message} (messages), \code{record_warning} (warnings), and
\code{record_error} (errors, only when the argument \code{error = TRUE}).

The \code{format()} method returns a character vector of plain-text output
or HTML code for displaying the results.

The \code{print()} method prints the results as plain text or HTML to the
console or displays the HTML page.
}
\description{
Run R code and capture various types of output, including text output, plots,
messages, warnings, and errors.
}
\examples{
code = c("# a warning test", "1:2 + 1:3", "par(mar = c(4, 4, 1, .2))",
    "barplot(5:1, col = 2:6, horiz = TRUE)", "head(iris)",
    "sunflowerplot(iris[, 3:4], seg.col = 'purple')",
    "if (TRUE) {\n  message('Hello, xfun::record()!')\n}",
    "# throw an error", "1 + 'a'")
res = xfun::record(code, dev.args = list(width = 9, height = 6.75),
    error = TRUE)
xfun::tree(res)
format(res)
# find and clean up plot files
plots = Filter(function(x) inherits(x, "record_plot"),
    res)
file.remove(unlist(plots))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{try_silent}
\alias{try_silent}
\title{Try to evaluate an expression silently}
\usage{
try_silent(expr)
}
\arguments{
\item{expr}{An R expression.}
}
\description{
An abbreviation of \code{try(silent = TRUE)}.
}
\examples{
library(xfun)
z = try_silent(stop("Wrong!"))
inherits(z, "try-error")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/base64.R
\name{mime_type}
\alias{mime_type}
\title{Get the MIME types of files}
\usage{
mime_type(x, use_mime = loadable("mime"), empty = "text/plain")
}
\arguments{
\item{x}{A vector of file paths.}

\item{use_mime}{Whether to use the \pkg{mime} package.}

\item{empty}{The MIME type for files without extensions (e.g., \code{Makefile}).
If \code{NA}, the type will be obtained from system command. This argument is
used only for \code{use_mime = FALSE}.}
}
\value{
A character vector of MIME types.
}
\description{
If the \pkg{mime} package is installed, call \code{\link[mime:guess_type]{mime::guess_type()}}, otherwise
use the system command \code{file --mime-type} to obtain the MIME type of a file.
Typically, the \code{file} command exists on *nix. On Windows, the command should
exist if Cygwin or Rtools is installed. If it is not found, .NET's
\code{MimeMapping} class will be used instead (which requires the .NET framework).
}
\note{
When querying the MIME type via the system command, the result will be
cached to \code{xfun:::cache_dir()}. This will make future queries much faster,
since running the command in real time can be a little slow.
}
\examples{\dontshow{if (tolower(Sys.getenv('CI')) == 'true') \{}
f = list.files(R.home("doc"), full.names = TRUE)
mime_type(f)
mime_type(f, FALSE)  # don't use mime
mime_type(f, FALSE, NA)  # run command for files without extension
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{html_tag}
\alias{html_tag}
\alias{html_value}
\alias{html_escape}
\alias{html_view}
\title{Tools for HTML tags}
\usage{
html_tag(.name, .content = NULL, .attrs = NULL, ...)

html_value(x)

html_escape(x, attr = FALSE)

html_view(x, ...)
}
\arguments{
\item{.name}{The tag name.}

\item{.content}{The content between opening and closing tags. Ignored for
void tags such as \verb{<img>}. Special characters such as \code{&}, \code{<}, and \code{>}
will be escaped unless the value was generated from \code{\link[=html_value]{html_value()}}. The
content can be either a character vector or a list. If it is a list, it may
contain both normal text and HTML content.}

\item{.attrs}{A named list of attributes.}

\item{...}{For \code{html_tag()}, named arguments as an alternative way to provide
attributes. For \code{html_view()}, other arguments to be passed to \code{\link[=new_app]{new_app()}}.}

\item{x}{A character vector to be treated as raw HTML content for
\code{html_value()}, escaped for \code{html_escape()}, and viewed for \code{html_view()}.}

\item{attr}{Whether to escape \verb{"}, \verb{\\r}, and \verb{\\n} (which should be escaped
for tag attributes).}
}
\value{
A character string.
}
\description{
Given a tag name, generate an HTML tag with optional attributes and content.
\code{\link[=html_tag]{html_tag()}} can be viewed as a simplified version of \code{htmltools::tags},
\code{\link[=html_value]{html_value()}} adds classes on the value so that it will be treated as raw
HTML (not escaped by \code{html_tag()}), \code{\link[=html_escape]{html_escape()}} escapes special
characters in HTML, and \code{html_view()} launches a browser or viewer to view
the HTML content.
}
\examples{
xfun::html_tag("a", "<R Project>", href = "https://www.r-project.org", target = "_blank")
xfun::html_tag("br")
xfun::html_tag("a", xfun::html_tag("strong", "R Project"), href = "#")
xfun::html_tag("a", list("<text>", xfun::html_tag("b", "R Project")), href = "#")
xfun::html_escape("\" quotes \" & brackets < >")
xfun::html_escape("\" & < > \r \n", attr = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cran.R
\name{is_R_CMD_check}
\alias{is_R_CMD_check}
\alias{is_CRAN_incoming}
\alias{check_package_name}
\alias{check_old_package}
\title{Some utility functions for checking packages}
\usage{
is_R_CMD_check()

is_CRAN_incoming()

check_package_name()

check_old_package(name, version)
}
\description{
Miscellaneous utility functions to obtain information about the package
checking environment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/base64.R
\name{base64_encode}
\alias{base64_encode}
\alias{base64_decode}
\title{Encode/decode data into/from base64 encoding.}
\usage{
base64_encode(x)

base64_decode(x, from = NA)
}
\arguments{
\item{x}{For \code{base64_encode()}, a raw vector. If not raw, it is assumed
to be a file or a connection to be read via \code{readBin()}. For
\code{base64_decode()}, a string.}

\item{from}{If provided (and \code{x} is not provided), a connection or file
to be read via \code{readChar()}, and the result will be passed to the
argument \code{x}.}
}
\value{
\code{base64_encode()} returns a character string.
  \code{base64_decode()} returns a raw vector.
}
\description{
The function \code{base64_encode()} encodes a file or a raw vector into the
base64 encoding. The function \code{base64_decode()} decodes data from the
base64 encoding.
}
\examples{
xfun::base64_encode(as.raw(1:10))
logo = xfun:::R_logo()
xfun::base64_encode(logo)
xfun::base64_decode("AQIDBAUGBwgJCg==")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{Rscript_call}
\alias{Rscript_call}
\title{Call a function in a new R session via \code{Rscript()}}
\usage{
Rscript_call(
  fun,
  args = list(),
  options = NULL,
  ...,
  wait = TRUE,
  fail = sprintf("Failed to run '\%s' in a new R session", deparse(substitute(fun))[1])
)
}
\arguments{
\item{fun}{A function, or a character string that can be parsed and evaluated
to a function.}

\item{args}{A list of argument values.}

\item{options}{A character vector of options to passed to \code{\link[=Rscript]{Rscript()}}, e.g.,
\code{"--vanilla"}.}

\item{..., wait}{Arguments to be passed to \code{\link[=system2]{system2()}}.}

\item{fail}{The desired error message when an error occurred in calling the
function. If the actual error message during running the function is
available, it will be appended to this message.}
}
\value{
If \code{wait = TRUE}, the returned value of the function in the new R
session. If \code{wait = FALSE}, three file paths will be returned: the first
one stores \code{fun} and \code{args} (as a list), the second one is supposed to
store the returned value of the function, and the third one stores the
possible error message.
}
\description{
Save the argument values of a function in a temporary RDS file, open a new R
session via \code{\link[=Rscript]{Rscript()}}, read the argument values, call the function, and
read the returned value back to the current R session.
}
\examples{
factorial(10)
# should return the same value
xfun::Rscript_call("factorial", list(10))

# the first argument can be either a character string or a function
xfun::Rscript_call(factorial, list(10))

# Run Rscript starting a vanilla R session
xfun::Rscript_call(factorial, list(10), options = c("--vanilla"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{numbers_to_words}
\alias{numbers_to_words}
\alias{n2w}
\title{Convert numbers to English words}
\usage{
numbers_to_words(x, cap = FALSE, hyphen = TRUE, and = FALSE)

n2w(x, cap = FALSE, hyphen = TRUE, and = FALSE)
}
\arguments{
\item{x}{A numeric vector. The absolute values should be less than \code{1e15}.}

\item{cap}{Whether to capitalize the first letter of the word. This can be
useful when the word is at the beginning of a sentence. Default is \code{FALSE}.}

\item{hyphen}{Whether to insert hyphen (-) when the number is between 21 and
99 (except 30, 40, etc.).}

\item{and}{Whether to insert \code{and} between hundreds and tens, e.g., write 110
as \dQuote{one hundred and ten} if \code{TRUE} instead of \dQuote{one hundred
ten}.}
}
\value{
A character vector.
}
\description{
This can be helpful when writing reports with \pkg{knitr}/\pkg{rmarkdown} if
we want to print numbers as English words in the output. The function \code{n2w()}
is an alias of \code{numbers_to_words()}.
}
\examples{
library(xfun)
n2w(0, cap = TRUE)
n2w(0:121, and = TRUE)
n2w(1e+06)
n2w(1e+11 + 12345678)
n2w(-987654321)
n2w(1e+15 - 1)
n2w(123.456)
n2w(123.45678901)
n2w(123.456789098765)
}
\author{
Daijiang Li
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{download_file}
\alias{download_file}
\title{Try various methods to download a file}
\usage{
download_file(
  url,
  output = url_filename(url),
  ...,
  .error = "No download method works (auto/wininet/wget/curl/lynx)"
)
}
\arguments{
\item{url}{The URL of the file.}

\item{output}{Path to the output file. By default, it is determined by
\code{\link[=url_filename]{url_filename()}}.}

\item{...}{Other arguments to be passed to \code{\link[=download.file]{download.file()}}
(except \code{method}).}

\item{.error}{An error message to signal when the download fails.}
}
\value{
The integer code \code{0} for success, or an error if none of the
methods work.
}
\description{
Try all possible methods in \code{\link[=download.file]{download.file()}} (e.g.,
\code{libcurl}, \code{curl}, \code{wget}, and \code{wininet}) and see if any
method can succeed. The reason to enumerate all methods is that sometimes the
default method does not work, e.g.,
\url{https://stat.ethz.ch/pipermail/r-devel/2016-June/072852.html}.
}
\note{
To allow downloading large files, the \code{timeout} option in
\code{\link[=options]{options()}} will be temporarily set to one hour (3600 seconds)
inside this function when this option has the default value of 60 seconds.
If you want a different \code{timeout} value, you may set it via
\code{options(timeout = N)}, where \code{N} is the number of seconds (not
60).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\docType{data}
\name{download_cache}
\alias{download_cache}
\title{Download a file from a URL and cache it on disk}
\format{
A list of methods:
\itemize{
\item \verb{$get(url, type, handler)} downloads a URL, caches it, and returns the file
content according to the value of \code{type} (possible values: \code{"text"} means
the text content; \code{"base64"} means the base64 encoded data; \code{"raw"} means
the raw binary content; \code{"auto"} is the default and means the type is
determined by the content type in the URL headers). Optionally a \code{handler}
function can be applied to the content.
\item \verb{$list()} gives the list of cache files.
\item \verb{$summary()} gives a summary of existing cache files.
\item \verb{$remove(url, type)} removes a single cache file.
\item \verb{$purge()} deletes all cache files.
}
}
\usage{
download_cache
}
\description{
This object provides methods to download files and cache them on disk.
}
\examples{\dontshow{if (interactive()) \{}
# the first time it may take a few seconds
x1 = xfun::download_cache$get("https://www.r-project.org/")
head(x1)

# now you can get the cached content
x2 = xfun::download_cache$get("https://www.r-project.org/")
identical(x1, x2)  # TRUE

# a binary file
x3 = xfun::download_cache$get("https://yihui.org/images/logo.png", "raw")
length(x3)

# show a summary
xfun::download_cache$summary()
# remove a specific cache file
xfun::download_cache$remove("https://yihui.org/images/logo.png", "raw")
# remove all cache files
xfun::download_cache$purge()
\dontshow{\}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{existing_files}
\alias{existing_files}
\title{Find file paths that exist}
\usage{
existing_files(x, first = FALSE, error = TRUE)
}
\arguments{
\item{x}{A vector of file paths.}

\item{first}{Whether to return the first existing path. If \code{TRUE} and no
specified files exist, it will signal an error unless the argument
\code{error = FALSE}.}

\item{error}{Whether to throw an error when \code{first = TRUE} but no files
exist. It can also take a character value, which will be used as the error
message.}
}
\value{
A vector of existing file paths.
}
\description{
This is a shorthand of \code{x[file.exists(x)]}, and optionally returns the
first existing file path.
}
\examples{
xfun::existing_files(c("foo.txt", system.file("DESCRIPTION", package = "xfun")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/record.R
\name{record}
\alias{record}
\alias{format.xfun_record_results}
\alias{print.xfun_record_results}
\title{Run R code and record the results}
\usage{
record(
  code = NULL,
  dev = "png",
  dev.path = "xfun-record",
  dev.ext = dev_ext(dev),
  dev.args = list(),
  dev.keep = TRUE,
  message = TRUE,
  warning = TRUE,
  error = NA,
  cache = list(),
  print = record_print,
  print.args = list(),
  verbose = getOption("xfun.record.verbose", 0),
  envir = parent.frame()
)

\method{format}{xfun_record_results}(
  x,
  to = c("text", "markdown", "html"),
  encode = FALSE,
  template = FALSE,
  ...
)

\method{print}{xfun_record_results}(
  x,
  browse = interactive(),
  to = if (browse) "html" else "text",
  template = TRUE,
  ...
)
}
\arguments{
\item{code}{A character vector of R source code.}

\item{dev}{A graphics device. It can be a function name, a function, or a
character string that can be evaluated to a function to open a graphics
device.}

\item{dev.path}{A base file path for plots. Actual plot filenames will be
this base path plus incremental suffixes. For example, if \code{dev.path = "foo"}, the plot files will be \verb{foo-1.png}, \verb{foo-2.png}, and so on. If
\code{dev.path} is not character (e.g., \code{FALSE}), plots will not be recorded.}

\item{dev.ext}{The file extension for plot files. By default, it will be
inferred from the first argument of the device function if possible.}

\item{dev.args}{Extra arguments to be passed to the device. The default
arguments are \code{list(units = 'in', onefile = FALSE, width = 7, height = 7, res = 96)}. If any of these arguments is not present in the device
function, it will be dropped.}

\item{dev.keep}{Indices of plots to be kept. The indices can be either
numeric (positive or negative integers) or logical. Negative integers and
false values will remove the corresponding plots. For example, if the code
generated 3 plots, \code{dev.keep = c(1, 3)}, \code{-2}, and \code{c(T, F, T)} are
equivalent ways to remove the second plot. The special value \code{dev.keep = 'last'} means to keep the last plot only.}

\item{message, warning, error}{If \code{TRUE}, record and store messages / warnings
/ errors in the output. If \code{FALSE}, suppress them. If \code{NA}, do not process
them (messages will be emitted to the console, and errors will halt the
execution).}

\item{cache}{A list of options for caching. See the \code{path}, \code{id}, and \code{...}
arguments of \code{\link[=cache_exec]{cache_exec()}}.}

\item{print}{A (typically S3) function that takes the value of an expression
in the code as input and returns output. The default is \code{\link[=record_print]{record_print()}}.
If a non-function value (e.g., \code{NA}) is passed to this argument, \code{\link[=print]{print()}}
(or \code{\link[=show]{show()}} for S4 objects) will be used.}

\item{print.args}{A list of arguments for the \code{print} function. By default,
the whole list is not passed directly to the function, but only an element
in the list with a name identical to the first class name of the returned
value of the expression, e.g., \code{list(data.frame = list(digits = 3), matrix = list())}. This makes it possible to apply different print arguments to
objects of different classes. If the whole list is intended to be passed to
the print function directly, wrap the list in \code{\link[=I]{I()}}.}

\item{verbose}{\code{2} means to always print the value of each expression in the
code, no matter if the value is \code{\link[=invisible]{invisible()}} or not; \code{1} means to always
print the value of the last expression; \code{0} means no special handling
(i.e., print only when the value is visible).}

\item{envir}{An environment in which the code is evaluated.}

\item{x}{An object returned by \code{record()}.}

\item{to}{The output format (text, markdown, or html).}

\item{encode}{For HTML output, whether to base64 encode plots.}

\item{template}{For HTML output, whether to embed the formatted results in an
HTML template. Alternatively, this argument can take a file path, i.e.,
path to an HTML template that contains the variable \verb{$body$}. If \code{TRUE},
the default template in this package will be used
(\code{xfun:::pkg_file('resources', 'record.html')}).}

\item{...}{Currently ignored.}

\item{browse}{Whether to browse the results on an HTML page.}
}
\value{
\code{record()} returns a list of the class \code{xfun_record_results} that
contains elements with these possible classes: \code{record_source} (source
code), \code{record_output} (text output), \code{record_plot} (plot file paths),
\code{record_message} (messages), \code{record_warning} (warnings), and
\code{record_error} (errors, only when the argument \code{error = TRUE}).

The \code{format()} method returns a character vector of plain-text output
or HTML code for displaying the results.

The \code{print()} method prints the results as plain text or HTML to the
console or displays the HTML page.
}
\description{
Run R code and capture various types of output, including text output, plots,
messages, warnings, and errors.
}
\examples{
code = c("# a warning test", "1:2 + 1:3", "par(mar = c(4, 4, 1, .2))",
    "barplot(5:1, col = 2:6, horiz = TRUE)", "head(iris)",
    "sunflowerplot(iris[, 3:4], seg.col = 'purple')",
    "if (TRUE) {\n  message('Hello, xfun::record()!')\n}",
    "# throw an error", "1 + 'a'")
res = xfun::record(code, dev.args = list(width = 9, height = 6.75),
    error = TRUE)
xfun::tree(res)
format(res)
# find and clean up plot files
plots = Filter(function(x) inherits(x, "record_plot"),
    res)
file.remove(unlist(plots))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{decimal_dot}
\alias{decimal_dot}
\title{Evaluate an expression after forcing the decimal point to be a dot}
\usage{
decimal_dot(x)
}
\arguments{
\item{x}{An expression.}
}
\value{
The value of \code{x}.
}
\description{
Sometimes it is necessary to use the dot character as the decimal separator.
In R, this could be affected by two settings: the global option
\code{options(OutDec)} and the \code{LC_NUMERIC} locale. This function sets the former
to \code{.} and the latter to \code{C} before evaluating an expression, such as
coercing a number to character.
}
\examples{
opts = options(OutDec = ",")
as.character(1.234)  # using ',' as the decimal separator
print(1.234)  # same
xfun::decimal_dot(as.character(1.234))  # using dot
xfun::decimal_dot(print(1.234))  # using dot
options(opts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{from_root}
\alias{from_root}
\title{Get the relative path of a path in a project relative to the current working
directory}
\usage{
from_root(..., root = proj_root(), error = TRUE)
}
\arguments{
\item{...}{A character vector of path components \emph{relative to the root
directory of the project}.}

\item{root}{The root directory of the project.}

\item{error}{Whether to signal an error if the path cannot be converted to a
relative path.}
}
\value{
A relative path, or an error when the project root directory cannot
be determined or the conversion failed and \code{error = TRUE}.
}
\description{
First compose an absolute path using the project root directory and the
relative path components, i.e., \code{\link{file.path}}\verb{(root, ...)}. Then
convert it to a relative path with \code{\link[=relative_path]{relative_path()}}, which is
relative to the current working directory.
}
\details{
This function was inspired by \code{here::here()}, and the major difference
is that it returns a relative path by default, which is more portable.
}
\examples{
\dontrun{
xfun::from_root("data", "mtcars.csv")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/image.R
\name{tinify}
\alias{tinify}
\alias{tinify_dir}
\title{Use the Tinify API to compress PNG and JPEG images}
\usage{
tinify(
  input,
  output,
  quiet = FALSE,
  force = FALSE,
  key = env_option("xfun.tinify.key"),
  history = env_option("xfun.tinify.history")
)

tinify_dir(dir = ".", ...)
}
\arguments{
\item{input}{A vector of input paths of images.}

\item{output}{A vector of output paths or a function that takes \code{input} and
returns a vector of output paths (e.g., \verb{output = }\code{\link{identity}} means
\code{output = input}). By default, if the \code{history} argument is not a provided,
\code{output} is \code{input} with a suffix \code{-min} (e.g., when \code{input = 'foo.png'},
\code{output = 'foo-min.png'}), otherwise \code{output} is the same as \code{input}, which
means the original image files will be overwritten.}

\item{quiet}{Whether to suppress detailed information about the compression,
which is of the form \samp{input.png (10 Kb) ==> output.png (5 Kb, 50\%);
  compression count: 42}. The percentage after \code{output.png} stands for
the compression ratio, and the compression count shows the number of
compressions used for the current month.}

\item{force}{Whether to compress an image again when it appears to have been
compressed before. This argument only makes sense when the \code{history}
argument is provided.}

\item{key}{The Tinify API key. It can be set via either the global option
\code{xfun.tinify.key} or the environment variable \code{R_XFUN_TINIFY_KEY} (see
\code{\link[=env_option]{env_option()}}).}

\item{history}{Path to a history file to record the MD5 checksum of
compressed images. If the checksum of an expected output image exists in
this file and \code{force = FALSE}, the compression will be skipped. This can
help you avoid unnecessary API calls.}

\item{dir}{A directory under which all \file{.png}, \file{.jpeg}, and
\file{.webp} files are to be compressed.}

\item{...}{Arguments passed to \code{\link[=tinify]{tinify()}}.}
}
\value{
The output file paths.
}
\description{
Compress PNG/JPEG images with \samp{api.tinify.com}, and download the
compressed images. These functions require R packages \pkg{curl} and
\pkg{jsonlite}. \code{tinify_dir()} is a wrapper function of \code{tinify()} to
compress images under a directory.
}
\details{
You are recommended to set the API key in \file{.Rprofile} or
\file{.Renviron}. After that, the only required argument of this function is
\code{input}. If the original images can be overwritten by the compressed images,
you may either use \code{output = identity}, or set the value of the \code{history}
argument in \file{.Rprofile} or \file{.Renviron}.
}
\examples{
\dontrun{
f = xfun:::R_logo("jpg$")
xfun::tinify(f)  # remember to set the API key before trying this
}
}
\references{
Tinify API: \url{https://tinypng.com/developers}.
}
\seealso{
The \pkg{tinieR} package (\url{https://github.com/jmablog/tinieR/}) is a
more comprehensive implementation of the Tinify API, whereas
\code{xfun::tinify()} has only implemented the feature of shrinking images.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{protect_math}
\alias{protect_math}
\title{Protect math expressions in pairs of backticks in Markdown}
\usage{
protect_math(x, token = "")
}
\arguments{
\item{x}{A character vector of text in Markdown.}

\item{token}{A character string to wrap math expressions at both ends. This
can be a unique token so that math expressions can be reliably identified
and restored after the Markdown text is converted.}
}
\value{
A character vector with math expressions in backticks.
}
\description{
For Markdown renderers that do not support LaTeX math, we need to protect
math expressions as verbatim code (in a pair of backticks), because some
characters in the math expressions may be interpreted as Markdown syntax
(e.g., a pair of underscores may make text italic). This function detects
math expressions in Markdown (by heuristics), and wrap them in backticks.
}
\details{
Expressions in pairs of dollar signs or double dollar signs are treated as
math, if there are no spaces after the starting dollar sign, or before the
ending dollar sign. There should be spaces before the starting dollar sign,
unless the math expression starts from the very beginning of a line. For a
pair of single dollar signs, the ending dollar sign should not be followed by
a number. With these assumptions, there should not be too many false
positives when detecing math expressions.

Besides, LaTeX environments (\verb{\begin{*}} and \verb{\end{*}}) are also
protected in backticks.
}
\note{
If you are using Pandoc or the \pkg{rmarkdown} package, there is no
need to use this function, because Pandoc's Markdown can recognize math
expressions.
}
\examples{
library(xfun)
protect_math(c("hi $a+b$", "hello $$\\\\alpha$$", "no math here: $x is $10 dollars"))
protect_math(c("hi $$", "\\\\begin{equation}", "x + y = z", "\\\\end{equation}"))
protect_math("$a+b$", "===")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-structure.R
\name{strict_list}
\alias{strict_list}
\alias{as_strict_list}
\alias{$.xfun_strict_list}
\alias{print.xfun_strict_list}
\title{Strict lists}
\usage{
strict_list(...)

as_strict_list(x)

\method{$}{xfun_strict_list}(x, name)

\method{print}{xfun_strict_list}(x, ...)
}
\arguments{
\item{...}{Objects (list elements), possibly named. Ignored in the
\code{print()} method.}

\item{x}{For \code{as_strict_list()}, the object to be coerced to a strict
list.

For \code{print()}, a strict list.}

\item{name}{The name (a character string) of the list element.}
}
\value{
Both \code{strict_list()} and \code{as_strict_list()} return a list
with the class \code{xfun_strict_list}. Whereas \code{as_strict_list()}
attempts to coerce its argument \code{x} to a list if necessary,
\code{strict_list()} just wraps its argument \code{...} in a list, i.e., it
will add another list level regardless if \code{...} already is of type
list.
}
\description{
A strict list is essentially a normal \code{\link[=list]{list()}} but it does not
allow partial matching with \code{$}.
}
\details{
To me, partial matching is often more annoying and surprising than
convenient. It can lead to bugs that are very hard to discover, and I have
been bitten by it many times. When I write \code{x$name}, I always mean
precisely \code{name}. You should use a modern code editor to autocomplete
the \code{name} if it is too long to type, instead of using partial names.
}
\examples{
library(xfun)
(z = strict_list(aaa = "I am aaa", b = 1:5))
z$a  # NULL!
z$aaa  # I am aaa
z$b
z$c = "create a new element"

z2 = unclass(z)  # a normal list
z2$a  # partial matching

z3 = as_strict_list(z2)  # a strict list again
z3$a  # NULL again!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{read_bin}
\alias{read_bin}
\title{Read all records of a binary file as a raw vector by default}
\usage{
read_bin(file, what = "raw", n = file.info(file)$size, ...)
}
\arguments{
\item{file, what, n, ...}{Arguments to be passed to \code{readBin()}.}
}
\value{
A vector returned from \code{readBin()}.
}
\description{
This is a wrapper function of \code{\link[=readBin]{readBin()}} with default arguments
\code{what = "raw"} and \verb{n = }\code{\link{file.size}}\code{(file)}, which means it
will read the full content of a binary file as a raw vector by default.
}
\examples{
f = tempfile()
cat("abc", file = f)
xfun::read_bin(f)
unlink(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{download_file}
\alias{download_file}
\title{Try various methods to download a file}
\usage{
download_file(
  url,
  output = url_filename(url),
  ...,
  .error = "No download method works (auto/wininet/wget/curl/lynx)"
)
}
\arguments{
\item{url}{The URL of the file.}

\item{output}{Path to the output file. By default, it is determined by
\code{\link[=url_filename]{url_filename()}}.}

\item{...}{Other arguments to be passed to \code{\link[=download.file]{download.file()}} (except
\code{method}).}

\item{.error}{An error message to signal when the download fails.}
}
\value{
The \code{output} file path if the download succeeded, or an error if none
of the download methods worked.
}
\description{
Try all possible methods in \code{\link[=download.file]{download.file()}} (e.g., \code{libcurl}, \code{curl},
\code{wget}, and \code{wininet}) and see if any method can succeed. The reason to
enumerate all methods is that sometimes the default method does not work,
e.g., \url{https://stat.ethz.ch/pipermail/r-devel/2016-June/072852.html}.
}
\note{
To allow downloading large files, the \code{timeout} option in \code{\link[=options]{options()}}
will be temporarily set to one hour (3600 seconds) inside this function
when this option has the default value of 60 seconds. If you want a
different \code{timeout} value, you may set it via \code{options(timeout = N)}, where
\code{N} is the number of seconds (not 60).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{fenced_block}
\alias{fenced_block}
\alias{fenced_div}
\alias{make_fence}
\title{Create a fenced block in Markdown}
\usage{
fenced_block(x, attrs = NULL, fence = make_fence(x, char), char = "`")

fenced_div(...)

make_fence(x, char = "`")
}
\arguments{
\item{x}{A character vector of the block content.}

\item{attrs}{A vector of block attributes.}

\item{fence}{The fence string, e.g., \code{:::} or \verb{```}. This will be
generated from the \code{char} argument by default.}

\item{char}{The fence character to be used to generate the fence string by
default.}

\item{...}{Arguments to be passed to \code{fenced_block()}.}
}
\value{
\code{fenced_block()} returns a character vector that contains both the
fences and content.

\code{make_fence()} returns a character string. If the block content
contains \code{N} fence characters (e.g., backticks), use \code{N + 1} characters as
the fence.
}
\description{
Wrap content with fence delimiters such as backticks (code blocks) or colons
(fenced Div). Optionally the fenced block can have attributes. The function
\code{fenced_div()} is a shorthand of \code{fenced_block(char = ':')}.
}
\examples{
# code block with class 'r' and ID 'foo'
xfun::fenced_block("1+1", c(".r", "#foo"))
# fenced Div
xfun::fenced_block("This is a **Div**.", char = ":")
# three backticks by default
xfun::make_fence("1+1")
# needs five backticks for the fences because content has four
xfun::make_fence(c("````r", "1+1", "````"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{proc_kill}
\alias{proc_kill}
\title{Kill a process and (optionally) all its child processes}
\usage{
proc_kill(pid, recursive = TRUE, ...)
}
\arguments{
\item{pid}{The process ID.}

\item{recursive}{Whether to kill the child processes of the process.}

\item{...}{Arguments to be passed to \code{\link[=system2]{system2()}} to run the
command to kill the process.}
}
\value{
The status code returned from \code{system2()}.
}
\description{
Run the command \command{taskkill /f /pid} on Windows and \command{kill} on
Unix, respectively, to kill a process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_rds}
\alias{cache_rds}
\title{Cache the value of an R expression to an RDS file}
\usage{
cache_rds(
  expr = {
 },
  rerun = FALSE,
  file = "cache.rds",
  dir = "cache/",
  hash = NULL,
  clean = getOption("xfun.cache_rds.clean", TRUE),
  ...
)
}
\arguments{
\item{expr}{An R expression.}

\item{rerun}{Whether to delete the RDS file, rerun the expression, and save
the result again (i.e., invalidate the cache if it exists).}

\item{file}{The \emph{base} (see Details) cache filename under the directory
specified by the \code{dir} argument. If not specified and this function is
called inside a code chunk of a \pkg{knitr} document (e.g., an R Markdown
document), the default is the current chunk label plus the extension
\file{.rds}.}

\item{dir}{The path of the RDS file is partially determined by \code{paste0(dir, file)}. If not specified and the \pkg{knitr} package is available, the
default value of \code{dir} is the \pkg{knitr} chunk option \code{cache.path} (so if
you are compiling a \pkg{knitr} document, you do not need to provide this
\code{dir} argument explicitly), otherwise the default is \file{cache/}. If you
do not want to provide a \code{dir} but simply a valid path to the \code{file}
argument, you may use \code{dir = ""}.}

\item{hash}{A \code{list} object that contributes to the MD5 hash of the cache
filename (see Details). It can also take a special character value
\code{"auto"}. Other types of objects are ignored.}

\item{clean}{Whether to clean up the old cache files automatically when
\code{expr} has changed.}

\item{...}{Other arguments to be passed to \code{\link[=saveRDS]{saveRDS()}}.}
}
\value{
If the cache file does not exist, run the expression and save the
result to the file, otherwise read the cache file and return the value.
}
\description{
Save the value of an expression to a cache file (of the RDS format). Next
time the value is loaded from the file if it exists.
}
\details{
Note that the \code{file} argument does not provide the full cache filename. The
actual name of the cache file is of the form \file{BASENAME_HASH.rds}, where
\file{BASENAME} is the base name provided via the \file{file} argument (e.g.,
if \code{file = 'foo.rds'}, \code{BASENAME} would be \file{foo}), and \file{HASH} is
the MD5 hash (also called the \sQuote{checksum}) calculated from the R code
provided to the \code{expr} argument and the value of the \code{hash} argument, which
means when the code or the \code{hash} argument changes, the \file{HASH} string
may also change, and the old cache will be invalidated (if it exists). If you
want to find the cache file, look for \file{.rds} files that contain 32
hexadecimal digits (consisting of 0-9 and a-z) at the end of the filename.

The possible ways to invalidate the cache are: 1) change the code in \code{expr}
argument; 2) delete the cache file manually or automatically through the
argument \code{rerun = TRUE}; and 3) change the value of the \code{hash} argument. The
first two ways should be obvious. For the third way, it makes it possible to
automatically invalidate the cache based on changes in certain R objects. For
example, when you run \code{cache_rds({ x + y })}, you may want to invalidate the
cache to rerun \code{{ x + y }} when the value of \code{x} or \code{y} has been changed, and
you can tell \code{cache_rds()} to do so by \code{cache_rds({ x + y }, hash = list(x, y))}. The value of the argument \code{hash} is expected to be a list, but it can
also take a special value, \code{"auto"}, which means \code{cache_rds(expr)} will try
to automatically figure out the global variables in \code{expr}, return a list of
their values, and use this list as the actual value of \code{hash}. This behavior
is most likely to be what you really want: if the code in \code{expr} uses an
external global variable, you may want to invalidate the cache if the value
of the global variable has changed. Here a \dQuote{global variable} means a
variable not created locally in \code{expr}, e.g., for \code{cache_rds({ x <- 1; x + y })}, \code{x} is a local variable, and \code{y} is (most likely to be) a global
variable, so changes in \code{y} should invalidate the cache. However, you know
your own code the best. If you want to be completely sure when to invalidate
the cache, you can always provide a list of objects explicitly rather than
relying on \code{hash = "auto"}.

By default (the argument \code{clean = TRUE}), old cache files will be
automatically cleaned up. Sometimes you may want to use \code{clean = FALSE} (set
the R global option \code{options(xfun.cache_rds.clean = FALSE)} if you want
\code{FALSE} to be the default). For example, you may not have decided which
version of code to use, and you can keep the cache of both versions with
\code{clean = FALSE}, so when you switch between the two versions of code, it will
still be fast to run the code.
}
\note{
Changes in the code in the \code{expr} argument do not necessarily always
invalidate the cache, if the changed code is \code{\link{parse}}\code{d} to the same
expression as the previous version of the code. For example, if you have
run \code{cache_rds({Sys.sleep(5);1+1})} before, running \code{cache_rds({ Sys.sleep( 5 ) ; 1 + 1 })} will use the cache, because the two expressions are
essentially the same (they only differ in white spaces). Usually you can
add/delete white spaces or comments to your code in \code{expr} without
invalidating the cache. See the package vignette \code{vignette('xfun', package = 'xfun')} for more examples.

When this function is called in a code chunk of a \pkg{knitr} document, you
may not want to provide the filename or directory of the cache file,
because they have reasonable defaults.

Side-effects (such as plots or printed output) will not be cached. The
cache only stores the last value of the expression in \code{expr}.
}
\examples{
f = tempfile()  # the cache file
compute = function(...) {
    res = xfun::cache_rds({
        Sys.sleep(1)
        1:10
    }, file = f, dir = "", ...)
    res
}
compute()  # takes one second
compute()  # returns 1:10 immediately
compute()  # fast again
compute(rerun = TRUE)  # one second to rerun
compute()
unlink(paste0(f, "_*.rds"))
}
\seealso{
\code{\link[=cache_exec]{cache_exec()}}, which is more flexible (e.g., it supports in-memory
caching and different read/write methods for cache files).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/command.R
\name{upload_ftp}
\alias{upload_ftp}
\alias{upload_win_builder}
\title{Upload to an FTP server via \command{curl}}
\usage{
upload_ftp(file, server, dir = "")

upload_win_builder(
  file = pkg_build(),
  version = c("R-devel", "R-release", "R-oldrelease"),
  server = c("ftp", "https")
)
}
\arguments{
\item{file}{Path to a local file.}

\item{server}{The address of the FTP server. For \code{upload_win_builder()},
\code{server = 'https'} means uploading to
\code{'https://win-builder.r-project.org/upload.aspx'}.}

\item{dir}{The remote directory to which the file should be uploaded.}

\item{version}{The R version(s) on win-builder.}
}
\value{
Status code returned from \code{\link[=system2]{system2()}} or
\code{curl::curl_fetch_memory()}.
}
\description{
The function \code{upload_ftp()} runs the command \command{curl -T file
server} to upload a file to an FTP server if the system command
\command{curl} is available, otherwise it uses the R package \pkg{curl}. The
function \code{upload_win_builder()} uses \code{upload_ftp()} to upload
packages to the win-builder server.
}
\details{
These functions were written mainly to save package developers the trouble of
going to the win-builder web page and uploading packages there manually.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/yaml.R
\name{yaml_load}
\alias{yaml_load}
\title{Read YAML data}
\usage{
yaml_load(
  x,
  ...,
  handlers = NULL,
  envir = parent.frame(),
  use_yaml = loadable("yaml")
)
}
\arguments{
\item{x}{A character vector of YAML data.}

\item{..., handlers}{Arguments to be passed to \code{\link[yaml:yaml.load]{yaml::yaml.load()}}.}

\item{envir}{The environment in which R expressions in YAML are evaluated. To
disable the evaluation, use \code{envir = FALSE}.}

\item{use_yaml}{Whether to use the \pkg{yaml} package.}
}
\value{
An R object (typically a list).
}
\description{
If the \pkg{yaml} package is installed, use \code{\link[yaml:yaml.load]{yaml::yaml.load()}} to read the
data. If not, use a simple parser instead, which only supports a limited
number of data types (see \dQuote{Examples}). In particular, it does not
support values that span across multiple lines (such as multi-line text).
}
\note{
R expressions in YAML will be returned as \link{expression}s when they are
not evaluated. This is different with \code{\link[yaml:yaml.load]{yaml::yaml.load()}}, which returns
character strings for expressions.
}
\examples{
# test the simple parser without using the yaml package
read_yaml = function(...) xfun::yaml_load(..., use_yaml = FALSE)
read_yaml("a: 1")
read_yaml("a: 1\nb: \"foo\"\nc: null")
read_yaml("a:\n  b: false\n  c: true\n  d: 1.234\ne: bar")
read_yaml("a: !expr paste(1:10, collapse = \", \")")
read_yaml("a: [1, 3, 4, 2]")
read_yaml("a: [1, \"abc\", 4, 2]")
read_yaml("a: [\"foo\", \"bar\"]")
read_yaml("a: [true, false, true]")
# the other form of array is not supported
read_yaml("a:\n  - b\n  - c")
# and you must use the yaml package
if (loadable("yaml")) yaml_load("a:\n  - b\n  - c")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr.R
\name{divide_chunk}
\alias{divide_chunk}
\title{Divide chunk options from the code chunk body}
\usage{
divide_chunk(engine, code)
}
\arguments{
\item{engine}{The name of the language engine (to determine the appropriate
comment character).}

\item{code}{A character vector (lines of code).}
}
\value{
A list with the following items:
\itemize{
\item \code{options}: The parsed options (if there are any) as a list.
\item \code{src}: The part of the input that contains the options.
\item \code{code}: The part of the input that contains the code.
}
}
\description{
Chunk options can be written in special comments (e.g., after \verb{#|} for R code
chunks) inside a code chunk. This function partitions these options from the
chunk body.
}
\note{
Chunk options must be written on \emph{continuous} lines (i.e., all lines
must start with the special comment prefix such as \verb{#|}) at the beginning
of the chunk body.
}
\examples{
# parse yaml-like items
yaml_like = c("#| label: mine", "#| echo: true", "#| fig.width: 8", "#| foo: bar",
    "1 + 1")
writeLines(yaml_like)
xfun::divide_chunk("r", yaml_like)

# parse CSV syntax
csv_like = c("#| mine, echo = TRUE, fig.width = 8, foo = 'bar'", "1 + 1")
writeLines(csv_like)
xfun::divide_chunk("r", csv_like)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packages.R
\name{news2md}
\alias{news2md}
\title{Convert package news to the Markdown format}
\usage{
news2md(package, ..., output = "NEWS.md", category = TRUE)
}
\arguments{
\item{package, ...}{Arguments to be passed to \code{\link[=news]{news()}}.}

\item{output}{The output file path.}

\item{category}{Whether to keep the category names.}
}
\value{
If \code{output = NA}, returns the Markdown content as a character
vector, otherwise the content is written to the output file.
}
\description{
Read the package news with \code{\link[=news]{news()}}, convert the result to
Markdown, and write to an output file (e.g., \file{NEWS.md}). Each package
version appears in a first-level header, each category (e.g., \samp{NEW
FEATURES} or \samp{BUG FIXES}) is in a second-level header, and the news
items are written into bullet lists.
}
\examples{\dontshow{if (interactive()) \{}
# news for the current version of R
xfun::news2md("R", Version == getRversion(), output = NA)
\dontshow{\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string.R
\name{split_lines}
\alias{split_lines}
\title{Split a character vector by line breaks}
\usage{
split_lines(x)
}
\arguments{
\item{x}{A character vector.}
}
\value{
All elements of the character vector are split by \code{'\\n'} into
lines.
}
\description{
Call \code{unlist(strsplit(x, '\\n'))} on the character vector \code{x} and
make sure it works in a few edge cases: \code{split_lines('')} returns
\code{''} instead of \code{character(0)} (which is the returned value of
\code{strsplit('', '\\n')}); \code{split_lines('a\\n')} returns \code{c('a', '')} instead of \code{c('a')} (which is the returned value of
\code{strsplit('a\\n', '\\n')}.
}
\examples{
xfun::split_lines(c("a", "b\nc"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/yaml.R
\name{yaml_body}
\alias{yaml_body}
\title{Partition the YAML metadata and the body in a document}
\usage{
yaml_body(x, ...)
}
\arguments{
\item{x}{A character vector of the document content.}

\item{...}{Arguments to be passed to \code{yaml_load()}.}
}
\value{
A list of components \code{yaml} (the parsed YAML data), \code{lines} (starting
and ending line numbers of YAML), and \code{body} (a character vector of the
body text). If YAML metadata does not exist in the document, the components
\code{yaml} and \code{lines} will be missing.
}
\description{
Split a document into the YAML metadata (which starts with \verb{---} in the
beginning of the document) and the body. The YAML metadata will be parsed.
}
\examples{
xfun::yaml_body(c("---", "title: Hello", "output: markdown::html_document", "---",
    "", "Content."))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/paths.R
\name{rename_seq}
\alias{rename_seq}
\title{Rename files with a sequential numeric prefix}
\usage{
rename_seq(
  pattern = "^[0-9]+-.+[.]Rmd$",
  format = "auto",
  replace = TRUE,
  start = 1,
  dry_run = TRUE
)
}
\arguments{
\item{pattern}{A regular expression for \code{\link[=list.files]{list.files()}} to obtain
the files to be renamed. For example, to rename \code{.jpeg} files, use
\code{pattern = "[.]jpeg$"}.}

\item{format}{The format for the numeric prefix. This is passed to
\code{\link[=sprintf]{sprintf()}}. The default format is \code{"\\\%0Nd"} where \code{N = floor(log10(n)) + 1} and \code{n} is the number of files, which means the
prefix may be padded with zeros. For example, if there are 150 files to be
renamed, the format will be \code{"\\\%03d"} and the prefixes will be
\code{001}, \code{002}, ..., \code{150}.}

\item{replace}{Whether to remove existing numeric prefixes in filenames.}

\item{start}{The starting number for the prefix (it can start from 0).}

\item{dry_run}{Whether to not really rename files. To be safe, the default is
\code{TRUE}. If you have looked at the new filenames and are sure the new
names are what you want, you may rerun \code{rename_seq()} with
\code{dry_run = FALSE} to actually rename files.}
}
\value{
A named character vector. The names are original filenames, and the
vector itself is the new filenames.
}
\description{
Rename a series of files and add an incremental numeric prefix to the
filenames. For example, files \file{a.txt}, \file{b.txt}, and \file{c.txt}
can be renamed to \file{1-a.txt}, \file{2-b.txt}, and \file{3-c.txt}.
}
\examples{
xfun::rename_seq()
xfun::rename_seq("[.](jpeg|png)$", format = "\%04d")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{fenced_block}
\alias{fenced_block}
\alias{fenced_div}
\alias{make_fence}
\title{Create a fenced block in Markdown}
\usage{
fenced_block(x, attrs = NULL, fence = make_fence(x, char), char = "`")

fenced_div(...)

make_fence(x, char = "`", start = 3)
}
\arguments{
\item{x}{A character vector of the block content.}

\item{attrs}{A vector of block attributes.}

\item{fence}{The fence string, e.g., \code{:::} or \verb{```}. This will be
generated from the \code{char} argument by default.}

\item{char}{The fence character to be used to generate the fence string by
default.}

\item{...}{Arguments to be passed to \code{fenced_block()}.}

\item{start}{The number of characters to start searching \code{x} with. If the
string of this number of characters is found, add one more character, and
repeat the search.}
}
\value{
\code{fenced_block()} returns a character vector that contains both the
fences and content.

\code{make_fence()} returns a character string. If the block content
contains \code{N} fence characters (e.g., backticks), use \code{N + 1} characters as
the fence.
}
\description{
Wrap content with fence delimiters such as backticks (code blocks) or colons
(fenced Div). Optionally the fenced block can have attributes. The function
\code{fenced_div()} is a shorthand of \code{fenced_block(char = ':')}.
}
\examples{
# code block with class 'r' and ID 'foo'
xfun::fenced_block("1+1", c(".r", "#foo"))
# fenced Div
xfun::fenced_block("This is a **Div**.", char = ":")
# three backticks by default
xfun::make_fence("1+1")
# needs five backticks for the fences because content has four
xfun::make_fence(c("````r", "1+1", "````"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{lazy_save}
\alias{lazy_save}
\alias{lazy_load}
\title{Save objects to files and lazy-load them}
\usage{
lazy_save(list = NULL, path = "./", method = "auto", envir = parent.frame())

lazy_load(path = "./", method = "auto", envir = parent.frame())
}
\arguments{
\item{list}{A character vector of object names. This list will be written to
an index file with \code{0} as the base name (e.g., \verb{0.rds}).}

\item{path}{The path to write files to / read files from.}

\item{method}{The file save/load method. It can be a string (e.g., \code{rds},
\code{raw}, or \code{qs}) or a list. See the \code{rw} argument of \code{\link[=cache_exec]{cache_exec()}}. By
default, it is automatically detected by checking the existence of the
index file (e.g., \verb{0.rds}, \verb{0.raw}, or \verb{0.qs}).}

\item{envir}{The environment to \link{get} or \link{assign} objects.}
}
\value{
\code{\link[=lazy_save]{lazy_save()}} returns invisible \code{NULL}; \code{\link[=lazy_load]{lazy_load()}} returns the
object names invisibly.
}
\description{
The function \code{\link[=lazy_save]{lazy_save()}} saves objects to files with incremental integer
names (e.g., the first object is saved to \verb{1.rds}, and the second object is
saved to \verb{2.rds}, etc.). The function \code{\link[=lazy_load]{lazy_load()}} lazy-load objects from
files saved via \code{\link[=lazy_save]{lazy_save()}}, i.e., a file will not be read until the object
is used.
}
\seealso{
\code{\link[=delayedAssign]{delayedAssign()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{protect_math}
\alias{protect_math}
\title{Protect math expressions in pairs of backticks in Markdown}
\usage{
protect_math(x, token = "", use_block = FALSE)
}
\arguments{
\item{x}{A character vector of text in Markdown.}

\item{token}{A character string to wrap math expressions at both ends. This
can be a unique token so that math expressions can be reliably identified
and restored after the Markdown text is converted.}

\item{use_block}{Whether to use code blocks (\verb{```md-math}) to protect
\verb{$$ $$} expressions that span across multiple lines. This is necessary when
a certain line in the math expression starts with a special character that
can accidentally start a new element (e.g., a leading \code{+} may start a
bullet list). Only code blocks can prevent this case.}
}
\value{
A character vector with math expressions in backticks.
}
\description{
For Markdown renderers that do not support LaTeX math, we need to protect
math expressions as verbatim code (in a pair of backticks), because some
characters in the math expressions may be interpreted as Markdown syntax
(e.g., a pair of underscores may make text italic). This function detects
math expressions in Markdown (by heuristics), and wrap them in backticks.
}
\details{
Expressions in pairs of dollar signs or double dollar signs are treated as
math, if there are no spaces after the starting dollar sign, or before the
ending dollar sign. There should be a space or \code{(} before the starting dollar
sign, unless the math expression starts from the very beginning of a line.
For a pair of single dollar signs, the ending dollar sign should not be
followed by a number, and the inner math expression should not be wrapped in
backticks. With these assumptions, there should not be too many false
positives when detecing math expressions.

Besides, LaTeX environments (\verb{\begin{*}} and \verb{\end{*}}) are also
protected in backticks.
}
\note{
If you are using Pandoc or the \pkg{rmarkdown} package, there is no
need to use this function, because Pandoc's Markdown can recognize math
expressions.
}
\examples{
library(xfun)
protect_math(c("hi $a+b$", "hello $$\\\\alpha$$", "no math here: $x is $10 dollars"))
protect_math(c("hi $$", "\\\\begin{equation}", "x + y = z", "\\\\end{equation}"))
protect_math("$a+b$", "===")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{tree}
\alias{tree}
\title{Turn the output of \code{\link[=str]{str()}} into a tree diagram}
\usage{
tree(...)
}
\arguments{
\item{...}{Arguments to be passed to \code{\link[=str]{str()}} (note that the
\code{comp.str} is hardcoded inside this function, and it is the only
argument that you cannot customize).}
}
\value{
A character string as a \code{\link[=raw_string]{raw_string()}}.
}
\description{
The super useful function \code{str()} uses \verb{..} to indicate the level
of sub-elements of an object, which may be difficult to read. This function
uses vertical pipes to connect all sub-elements on the same level, so it is
clearer which elements belong to the same parent element in an object with a
nested structure (such as a nested list).
}
\examples{
fit = lsfit(1:9, 1:9)
str(fit)
xfun::tree(fit)

fit = lm(dist ~ speed, data = cars)
str(fit)
xfun::tree(fit)

# some trivial examples
xfun::tree(1:10)
xfun::tree(iris)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_rds}
\alias{cache_rds}
\title{Cache the value of an R expression to an RDS file}
\usage{
cache_rds(
  expr = {
 },
  rerun = FALSE,
  file = "cache.rds",
  dir = "cache/",
  hash = NULL,
  clean = getOption("xfun.cache_rds.clean", TRUE),
  ...
)
}
\arguments{
\item{expr}{An R expression.}

\item{rerun}{Whether to delete the RDS file, rerun the expression, and save
the result again (i.e., invalidate the cache if it exists).}

\item{file}{The \emph{base} (see Details) cache filename under the directory
specified by the \code{dir} argument. If not specified and this function is
called inside a code chunk of a \pkg{knitr} document (e.g., an R Markdown
document), the default is the current chunk label plus the extension
\file{.rds}. The file can also be other formats (see Details).}

\item{dir}{The path of the RDS file is partially determined by \code{paste0(dir, file)}. If not specified and the \pkg{knitr} package is available, the
default value of \code{dir} is the \pkg{knitr} chunk option \code{cache.path} (so if
you are compiling a \pkg{knitr} document, you do not need to provide this
\code{dir} argument explicitly), otherwise the default is \file{cache/}. If you
do not want to provide a \code{dir} but simply a valid path to the \code{file}
argument, you may use \code{dir = ""}.}

\item{hash}{A \code{list} object that contributes to the MD5 hash of the cache
filename (see Details). It can also take a special character value
\code{"auto"}. Other types of objects are ignored.}

\item{clean}{Whether to clean up the old cache files automatically when
\code{expr} has changed.}

\item{...}{Other arguments to be passed to the saving method such as
\code{\link[=saveRDS]{saveRDS()}}.}
}
\value{
If the cache file does not exist, run the expression and save the
result to the file, otherwise read the cache file and return the value.
}
\description{
Save the value of an expression to a cache file. Next time the value is
loaded from the file if it exists. Please consider using \code{\link[=cache_exec]{cache_exec()}}
instead, which is more flexible and intelligent.
}
\details{
Note that the \code{file} argument does not provide the full cache filename. The
actual name of the cache file is of the form \file{BASENAME_HASH.rds}, where
\file{BASENAME} is the base name provided via the \file{file} argument (e.g.,
if \code{file = 'foo.rds'}, \code{BASENAME} would be \file{foo}), and \file{HASH} is
the MD5 hash (also called the \sQuote{checksum}) calculated from the R code
provided to the \code{expr} argument and the value of the \code{hash} argument, which
means when the code or the \code{hash} argument changes, the \file{HASH} string
may also change, and the old cache will be invalidated (if it exists). If you
want to find the cache file, look for \file{.rds} files that contain 32
hexadecimal digits (consisting of 0-9 and a-z) at the end of the filename.

The file can be other formats such as \code{.qs2}. See the \code{rw} argument of
\code{\link[=cache_exec]{cache_exec()}}.

The possible ways to invalidate the cache are: 1) change the code in \code{expr}
argument; 2) delete the cache file manually or automatically through the
argument \code{rerun = TRUE}; and 3) change the value of the \code{hash} argument. The
first two ways should be obvious. For the third way, it makes it possible to
automatically invalidate the cache based on changes in certain R objects. For
example, when you run \code{cache_rds({ x + y })}, you may want to invalidate the
cache to rerun \code{{ x + y }} when the value of \code{x} or \code{y} has been changed, and
you can tell \code{cache_rds()} to do so by \code{cache_rds({ x + y }, hash = list(x, y))}. The value of the argument \code{hash} is expected to be a list, but it can
also take a special value, \code{"auto"}, which means \code{cache_rds(expr)} will try
to automatically figure out the global variables in \code{expr}, return a list of
their values, and use this list as the actual value of \code{hash}. This behavior
is most likely to be what you really want: if the code in \code{expr} uses an
external global variable, you may want to invalidate the cache if the value
of the global variable has changed. Here a \dQuote{global variable} means a
variable not created locally in \code{expr}, e.g., for \code{cache_rds({ x <- 1; x + y })}, \code{x} is a local variable, and \code{y} is (most likely to be) a global
variable, so changes in \code{y} should invalidate the cache. However, you know
your own code the best. If you want to be completely sure when to invalidate
the cache, you can always provide a list of objects explicitly rather than
relying on \code{hash = "auto"}.

By default (the argument \code{clean = TRUE}), old cache files will be
automatically cleaned up. Sometimes you may want to use \code{clean = FALSE} (set
the R global option \code{options(xfun.cache_rds.clean = FALSE)} if you want
\code{FALSE} to be the default). For example, you may not have decided which
version of code to use, and you can keep the cache of both versions with
\code{clean = FALSE}, so when you switch between the two versions of code, it will
still be fast to run the code.
}
\note{
Changes in the code in the \code{expr} argument do not necessarily always
invalidate the cache, if the changed code is \code{\link{parse}}\code{d} to the same
expression as the previous version of the code. For example, if you have
run \code{cache_rds({Sys.sleep(5);1+1})} before, running \code{cache_rds({ Sys.sleep( 5 ) ; 1 + 1 })} will use the cache, because the two expressions are
essentially the same (they only differ in white spaces). Usually you can
add/delete white spaces or comments to your code in \code{expr} without
invalidating the cache. See the package vignette \code{vignette('xfun', package = 'xfun')} for more examples.

When this function is called in a code chunk of a \pkg{knitr} document, you
may not want to provide the filename or directory of the cache file,
because they have reasonable defaults.

Side-effects (such as plots or printed output) will not be cached. The
cache only stores the last value of the expression in \code{expr}.

The hash is calculated by saving the hash object to a temporary file via
\code{saveRDS()} and getting the MD5 digest of the file. Since the file
generated by \code{saveRDS()} may be platform-dependent, the cache may not be
platform-agnostic. In other words, the cache saved on one platform may be
invalidated on another platform even if nothing in the hash object has
changed. By comparison, \code{\link[=cache_exec]{cache_exec()}} doesn't have this problem.
}
\examples{
f = tempfile()  # the cache file
compute = function(...) {
    res = xfun::cache_rds({
        Sys.sleep(1)
        1:10
    }, file = f, dir = "", ...)
    res
}
compute()  # takes one second
compute()  # returns 1:10 immediately
compute()  # fast again
compute(rerun = TRUE)  # one second to rerun
compute()
unlink(paste0(f, "_*.rds"))
}
\seealso{
\code{\link[=cache_exec]{cache_exec()}}, which is more flexible (e.g., it supports in-memory
caching and different read/write methods for cache files).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/yaml.R
\name{yaml_load}
\alias{yaml_load}
\title{Read YAML data}
\usage{
yaml_load(
  x,
  ...,
  handlers = NULL,
  envir = parent.frame(),
  use_yaml = loadable("yaml")
)
}
\arguments{
\item{x}{A character vector of YAML data.}

\item{..., handlers}{Arguments to be passed to \code{\link[yaml:yaml.load]{yaml::yaml.load()}}.}

\item{envir}{The environment in which R expressions in YAML are evaluated. To
disable the evaluation, use \code{envir = FALSE}.}

\item{use_yaml}{Whether to use the \pkg{yaml} package.}
}
\value{
An R object (typically a list).
}
\description{
If the \pkg{yaml} package is installed, use \code{\link[yaml:yaml.load]{yaml::yaml.load()}} to read the
data. If not, use the simple parser \code{\link[=taml_load]{taml_load()}} instead.
}
\note{
R expressions in YAML will be returned as \link{expression}s when they are
not evaluated. This is different with \code{\link[yaml:yaml.load]{yaml::yaml.load()}}, which returns
character strings for expressions.
}
\examples{
yaml_load("a: 1")
yaml_load("a: 1", use_yaml = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{env_option}
\alias{env_option}
\title{Retrieve a global option from both \code{options()} and environment variables}
\usage{
env_option(name, default = NULL)
}
\arguments{
\item{name}{The option name.}

\item{default}{The default value if the option is not found in \code{\link[=options]{options()}} or
environment variables.}
}
\value{
The option value.
}
\description{
If the option exists in \code{\link[=options]{options()}}, use its value. If not, query the
environment variable with the name \code{R_NAME} where \code{NAME} is the capitalized
option name with dots substituted by underscores. For example, for an option
\code{xfun.foo}, first we try \code{getOption('xfun.foo')}; if it does not exist, we
check the environment variable \code{R_XFUN_FOO}.
}
\details{
This provides two possible ways, whichever is more convenient, for users to
set an option. For example, global options can be set in the \link{.Rprofile}
file, and environment variables can be set in the \link{.Renviron} file.
}
\examples{
xfun::env_option("xfun.test.option")  # NULL

Sys.setenv(R_XFUN_TEST_OPTION = "1234")
xfun::env_option("xfun.test.option")  # 1234

options(xfun.test.option = TRUE)
xfun::env_option("xfun.test.option")  # TRUE (from options())
options(xfun.test.option = NULL)  # reset the option
xfun::env_option("xfun.test.option")  # 1234 (from env var)

Sys.unsetenv("R_XFUN_TEST_OPTION")
xfun::env_option("xfun.test.option")  # NULL again

xfun::env_option("xfun.test.option", FALSE)  # use default
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown.R
\name{prose_index}
\alias{prose_index}
\title{Find the indices of lines in Markdown that are prose (not code blocks)}
\usage{
prose_index(x, warn = TRUE)
}
\arguments{
\item{x}{A character vector of text in Markdown.}

\item{warn}{Whether to emit a warning when code fences are not balanced.}
}
\value{
An integer vector of indices of lines that are prose in Markdown.
}
\description{
Filter out the indices of lines between code block fences such as \verb{```}
(could be three or four or more backticks).
}
\note{
If the code fences are not balanced (e.g., a starting fence without an
ending fence), this function will treat all lines as prose.
}
\examples{
library(xfun)
prose_index(c("a", "```", "b", "```", "c"))
prose_index(c("a", "````", "```r", "1+1", "```", "````", "c"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knitr.R
\name{divide_chunk}
\alias{divide_chunk}
\title{Divide chunk options from the code chunk body}
\usage{
divide_chunk(engine, code, strict = FALSE, ...)
}
\arguments{
\item{engine}{The name of the language engine (to determine the appropriate
comment character).}

\item{code}{A character vector (lines of code).}

\item{strict}{If \code{FALSE}, allow chunk options to be written after \verb{#|} even
if \verb{#} is not the comment character of the engine (e.g., when \code{engine = 'js'}), otherwise throw an error if \verb{#|} is detected but \verb{#} is not the
comment character.}

\item{...}{Arguments to be passed to \code{\link[=yaml_load]{yaml_load()}}.}
}
\value{
A list with the following items:
\itemize{
\item \code{options}: The parsed options (if there are any) as a list.
\item \code{src}: The part of the input that contains the options.
\item \code{code}: The part of the input that contains the code.
}
}
\description{
Chunk options can be written in special comments (e.g., after \verb{#|} for R code
chunks) inside a code chunk. This function partitions these options from the
chunk body.
}
\note{
Chunk options must be written on \emph{continuous} lines (i.e., all lines
must start with the special comment prefix such as \verb{#|}) at the beginning
of the chunk body.
}
\examples{
# parse yaml-like items
yaml_like = c("#| label: mine", "#| echo: true", "#| fig.width: 8", "#| foo: bar",
    "1 + 1")
writeLines(yaml_like)
xfun::divide_chunk("r", yaml_like)

# parse CSV syntax
csv_like = c("#| mine, echo = TRUE, fig.width = 8, foo = 'bar'", "1 + 1")
writeLines(csv_like)
xfun::divide_chunk("r", csv_like)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/base64.R
\name{base64_encode}
\alias{base64_encode}
\alias{base64_decode}
\title{Encode/decode data into/from base64 encoding.}
\usage{
base64_encode(x)

base64_decode(x, from = NA)
}
\arguments{
\item{x}{For \code{base64_encode()}, a raw vector. If not raw, it is assumed
to be a file or a connection to be read via \code{readBin()}. For
\code{base64_decode()}, a string.}

\item{from}{If provided (and \code{x} is not provided), a connection or file
to be read via \code{readChar()}, and the result will be passed to the
argument \code{x}.}
}
\value{
\code{base64_encode()} returns a character string.
\code{base64_decode()} returns a raw vector.
}
\description{
The function \code{base64_encode()} encodes a file or a raw vector into the
base64 encoding. The function \code{base64_decode()} decodes data from the
base64 encoding.
}
\examples{
xfun::base64_encode(as.raw(1:10))
logo = xfun:::R_logo()
xfun::base64_encode(logo)
xfun::base64_decode("AQIDBAUGBwgJCg==")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/yaml.R
\name{yaml_body}
\alias{yaml_body}
\title{Partition the YAML metadata and the body in a document}
\usage{
yaml_body(x, ..., parse = TRUE)
}
\arguments{
\item{x}{A character vector of the document content.}

\item{...}{Arguments to be passed to \code{yaml_load()}.}

\item{parse}{Whether to parse the YAML data.}
}
\value{
A list of components \code{yaml} (the YAML data), \code{lines} (starting and
ending line numbers of YAML), and \code{body} (a character vector of the body
text). If YAML metadata does not exist in the document, the components
\code{yaml} and \code{lines} will be missing.
}
\description{
Split a document into the YAML metadata (which starts with \verb{---} in the
beginning of the document) and the body.
}
\examples{
xfun::yaml_body(c("---", "title: Hello", "output: litedown::html_format", "---",
    "", "Content."))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io.R
\name{url_destination}
\alias{url_destination}
\title{Get the final destination of a URL}
\usage{
url_destination(x, force = FALSE)
}
\arguments{
\item{x}{A character vector of URLs.}

\item{force}{By default, the query is cached in the current R session. To
bypass the cache, use \code{force = TRUE}.}
}
\value{
The final destination(s). If the URL returns a status code greater
than or equal to 400, it will throw an error.
}
\description{
If a URL is redirected, query the new location via \code{\link[=curlGetHeaders]{curlGetHeaders()}},
otherwise return the original URL.
}
\examples{
u = "https://tinytex.yihui.org"  # redirected to https://yihui.org/tinytex/
if (url_accessible(u)) url_destination(u)
}
